@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
+ * Copyright (c) 1997 by Silicon Graphics.  All rights reserved.
  *
  * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
  * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
@@ -13,7 +14,6 @@
  * Original author: Bill Janssen
  * Heavily modified by Hans Boehm and others
  */
-/* Boehm, September 21, 1995 5:57 pm PDT */
 
 /*
  * This is incredibly OS specific code for tracking down data sections in
@@ -32,7 +32,9 @@
 #include "gc_priv.h"
 
 /* BTL: avoid circular redefinition of dlopen if SOLARIS_THREADS defined */
-# if defined(SOLARIS_THREADS) && defined(dlopen)
+# if (defined(LINUX_THREADS) || defined(SOLARIS_THREADS) \
+      || defined(HPUX_THREADS) || defined(IRIX_THREADS)) && defined(dlopen) \
+     && !defined(USE_LD_WRAP)
     /* To support threads in Solaris, gc.h interposes on dlopen by       */
     /* defining "dlopen" to be "GC_dlopen", which is implemented below.  */
     /* However, both GC_FirstDLOpenedLinkMap() and GC_dlopen() use the   */
@@ -45,7 +47,10 @@
 # endif
 
 #if (defined(DYNAMIC_LOADING) || defined(MSWIN32)) && !defined(PCR)
-#if !defined(SUNOS4) && !defined(SUNOS5DL) && !defined(IRIX5) && !defined(MSWIN32) && !(defined(ALPHA) && defined(OSF1)) &&!defined(HP_PA) && (!defined(LINUX) && !defined(__ELF__))
+#if !defined(SUNOS4) && !defined(SUNOS5DL) && !defined(IRIX5) && \
+    !defined(MSWIN32) && !(defined(ALPHA) && defined(OSF1)) && \
+    !defined(HPUX) && !(defined(LINUX) && defined(__ELF__)) && \
+    !defined(RS6000) && !defined(SCO_ELF) && !defined(FREEBSD)
  --> We only know how to find data segments of dynamic libraries for the
  --> above.  Additional SVR4 variants might not be too
  --> hard to add.
@@ -68,7 +73,7 @@
 #endif
 
 
-#ifdef SUNOS5DL
+#if defined(SUNOS5DL) && !defined(USE_PROC_FOR_LIBRARIES)
 
 #ifdef LINT
     Elf32_Dyn _DYNAMIC;
@@ -114,9 +119,9 @@ GC_FirstDLOpenedLinkMap()
     return cachedResult;
 }
 
-#endif
+#endif /* SUNOS5DL ... */
 
-#ifdef SUNOS4
+#if defined(SUNOS4) && !defined(USE_PROC_FOR_LIBRARIES)
 
 #ifdef LINT
     struct link_dynamic _DYNAMIC;
@@ -154,7 +159,70 @@ static ptr_t GC_first_common()
     return(result);
 }
 
+#endif  /* SUNOS4 ... */
+
+# if defined(LINUX_THREADS) || defined(SOLARIS_THREADS) \
+     || defined(HPUX_THREADS) || defined(IRIX_THREADS)
+  /* Make sure we're not in the middle of a collection, and make	*/
+  /* sure we don't start any.	Returns previous value of GC_dont_gc.	*/
+  /* This is invoked prior to a dlopen call to avoid synchronization	*/
+  /* issues.  We can't just acquire the allocation lock, since startup 	*/
+  /* code in dlopen may try to allocate.				*/
+  /* This solution risks heap growth in the presence of many dlopen	*/
+  /* calls in either a multithreaded environment, or if the library	*/
+  /* initialization code allocates substantial amounts of GC'ed memory.	*/
+  /* But I don't know of a better solution.				*/
+  /* This can still deadlock if the client explicitly starts a GC 	*/
+  /* during the dlopen.  He shouldn't do that.				*/
+  static GC_bool disable_gc_for_dlopen()
+  {
+    GC_bool result;
+    LOCK();
+    result = GC_dont_gc;
+    while (GC_incremental && GC_collection_in_progress()) {
+	GC_collect_a_little_inner(1000);
+    }
+    GC_dont_gc = TRUE;
+    UNLOCK();
+    return(result);
+  }
+
+  /* Redefine dlopen to guarantee mutual exclusion with	*/
+  /* GC_register_dynamic_libraries.			*/
+  /* Should probably happen for other operating	systems, too. */
+
+#include <dlfcn.h>
+
+#ifdef USE_LD_WRAP
+  void * __wrap_dlopen(const char *path, int mode)
+#else
+  void * GC_dlopen(path, mode)
+  GC_CONST char * path;
+  int mode;
 #endif
+{
+    void * result;
+    GC_bool dont_gc_save;
+    
+#   ifndef USE_PROC_FOR_LIBRARIES
+      dont_gc_save = disable_gc_for_dlopen();
+#   endif
+#   ifdef USE_LD_WRAP
+      result = __real_dlopen(path, mode);
+#   else
+      result = dlopen(path, mode);
+#   endif
+#   ifndef USE_PROC_FOR_LIBRARIES
+      GC_dont_gc = dont_gc_save;
+#   endif
+    return(result);
+}
+# endif  /* SOLARIS_THREADS */
+
+/* BTL: added to fix circular dlopen definition if SOLARIS_THREADS defined */
+# if defined(GC_must_restore_redefined_dlopen)
+#   define dlopen GC_dlopen
+# endif
 
 # if defined(SUNOS4) || defined(SUNOS5DL)
 /* Add dynamic library data sections to the root set.		*/
@@ -164,30 +232,7 @@ static ptr_t GC_first_common()
 #   endif  /* We assume M3 programs don't call dlopen for now */
 # endif
 
-# ifdef SOLARIS_THREADS
-  /* Redefine dlopen to guarantee mutual exclusion with	*/
-  /* GC_register_dynamic_libraries.			*/
-  /* assumes that dlopen doesn't need to call GC_malloc	*/
-  /* and friends.					*/
-# include <thread.h>
-# include <synch.h>
-  
-void * GC_dlopen(const char *path, int mode)
-{
-    void * result;
-    
-    mutex_lock(&GC_allocate_ml);
-    result = dlopen(path, mode);
-    mutex_unlock(&GC_allocate_ml);
-    return(result);
-}
-# endif
-
-/* BTL: added to fix circular dlopen definition if SOLARIS_THREADS defined */
-# if defined(GC_must_restore_redefined_dlopen)
-#   define dlopen GC_dlopen
-# endif
-
+# ifndef USE_PROC_FOR_LIBRARIES
 void GC_register_dynamic_libraries()
 {
   struct link_map *lm = GC_FirstDLOpenedLinkMap();
@@ -249,9 +294,10 @@ void GC_register_dynamic_libraries()
 #   endif
 }
 
+# endif /* !USE_PROC ... */
 # endif /* SUNOS */
 
-#if defined(LINUX) && defined(__ELF__)
+#if defined(LINUX) && defined(__ELF__) || defined(SCO_ELF) || defined(FREEBSD)
 
 /* Dynamic loading code for Linux running ELF. Somewhat tested on
  * Linux/x86, untested but hopefully should work on Linux/Alpha. 
@@ -274,6 +320,9 @@ void GC_register_dynamic_libraries()
 static struct link_map *
 GC_FirstDLOpenedLinkMap()
 {
+#   ifdef __GNUC__
+#     pragma weak _DYNAMIC
+#   endif
     extern ElfW(Dyn) _DYNAMIC[];
     ElfW(Dyn) *dp;
     struct r_debug *r;
@@ -332,7 +381,7 @@ void GC_register_dynamic_libraries()
 
 #endif
 
-#ifdef IRIX5
+#if defined(IRIX5) || defined(USE_PROC_FOR_LIBRARIES)
 
 #include <sys/procfs.h>
 #include <sys/stat.h>
@@ -341,8 +390,12 @@ void GC_register_dynamic_libraries()
 #include <errno.h>
 
 extern void * GC_roots_present();
+	/* The type is a lie, since the real type doesn't make sense here, */
+	/* and we only test for NULL.					   */
 
+#ifndef GC_scratch_last_end_ptr   /* Never an extern any more? */
 extern ptr_t GC_scratch_last_end_ptr; /* End of GC_scratch_alloc arena	*/
+#endif
 
 /* We use /proc to track down all parts of the address space that are	*/
 /* mapped by the process, and throw out regions we know we shouldn't	*/
@@ -361,8 +414,14 @@ void GC_register_dynamic_libraries()
     ptr_t heap_start = (ptr_t)HEAP_START;
     ptr_t heap_end = heap_start;
 
+#   ifdef SUNOS5DL
+#     define MA_PHYS 0
+#   endif /* SUNOS5DL */
+
     if (fd < 0) {
       sprintf(buf, "/proc/%d", getpid());
+	/* The above generates a lint complaint, since pid_t varies.	*/
+	/* It's unclear how to improve this.				*/
       fd = open(buf, O_RDONLY);
       if (fd < 0) {
     	ABORT("/proc open failed");
@@ -375,10 +434,12 @@ void GC_register_dynamic_libraries()
     if (needed_sz >= current_sz) {
         current_sz = needed_sz * 2 + 1;
         		/* Expansion, plus room for 0 record */
-        addr_map = (prmap_t *)GC_scratch_alloc(current_sz * sizeof(prmap_t));
+        addr_map = (prmap_t *)GC_scratch_alloc((word)
+						(current_sz * sizeof(prmap_t)));
     }
     if (ioctl(fd, PIOCMAP, addr_map) < 0) {
-	GC_err_printf2("fd = %d, errno = %d\n", fd, errno);
+        GC_err_printf4("fd = %d, errno = %d, needed_sz = %d, addr_map = 0x%X\n",
+                        fd, errno, needed_sz, addr_map);
     	ABORT("/proc PIOCMAP ioctl failed");
     };
     if (GC_n_heap_sects > 0) {
@@ -399,6 +460,10 @@ void GC_register_dynamic_libraries()
         if (GC_roots_present(start)) goto irrelevant;
         if (start < heap_end && start >= heap_start)
         	goto irrelevant;
+#	ifdef MMAP_STACKS
+	  if (GC_is_thread_stack(start)) goto irrelevant;
+#	endif /* MMAP_STACKS */
+
         limit = start + addr_map[i].pr_size;
 	if (addr_map[i].pr_off == 0 && strncmp(start, ELFMAG, 4) == 0) {
 	    /* Discard text segments, i.e. 0-offset mappings against	*/
@@ -437,7 +502,7 @@ void GC_register_dynamic_libraries()
 	fd = -1;
 }
 
-#endif  /* IRIX5 */
+# endif /* USE_PROC || IRIX5 */
 
 # ifdef MSWIN32
 
@@ -450,7 +515,7 @@ void GC_register_dynamic_libraries()
   /* that could possibly have been written to.				*/
   DWORD GC_allocation_granularity;
   
-  extern bool GC_is_heap_base (ptr_t p);
+  extern GC_bool GC_is_heap_base (ptr_t p);
 
 # ifdef WIN32_THREADS
     extern void GC_get_next_stack(char *start, char **lo, char **hi);
@@ -474,7 +539,7 @@ void GC_register_dynamic_libraries()
 	    GC_add_roots_inner(curr_base, next_stack_lo, TRUE);
 	    curr_base = next_stack_hi;
 	}
-	if (curr_base < limit) GC_add_roots_inner(curr_base, limit);
+	if (curr_base < limit) GC_add_roots_inner(curr_base, limit, TRUE);
     }
 #   else
         if (limit > stack_top && base < GC_stackbottom) {
@@ -485,7 +550,7 @@ void GC_register_dynamic_libraries()
 #   endif
   }
   
-  extern bool GC_win32s;
+  extern GC_bool GC_win32s;
   
   void GC_register_dynamic_libraries()
   {
@@ -632,8 +697,9 @@ void GC_register_dynamic_libraries()
 }
 #endif
 
-#if defined(HP_PA)
+#if defined(HPUX)
 
+#include <errno.h>
 #include <dl.h>
 
 extern int errno;
@@ -654,6 +720,11 @@ void GC_register_dynamic_libraries()
 
       /* Check if this is the end of the list or if some error occured */
         if (status != 0) {
+#	 ifdef HPUX_THREADS
+	   /* I've seen errno values of 0.  The man page is not clear	*/
+	   /* as to whether errno should get set on a -1 return.	*/
+	   break;
+#	 else
           if (errno == EINVAL) {
               break; /* Moved past end of shared library list --> finished */
           } else {
@@ -664,6 +735,7 @@ void GC_register_dynamic_libraries()
 	      }
               ABORT("shl_get failed");
           }
+#	 endif
         }
 
 #     ifdef VERBOSE
@@ -686,7 +758,41 @@ void GC_register_dynamic_libraries()
         index++;
     }
 }
-#endif /* HP_PA */
+#endif /* HPUX */
+
+#ifdef RS6000
+#pragma alloca
+#include <sys/ldr.h>
+#include <sys/errno.h>
+void GC_register_dynamic_libraries()
+{
+	int len;
+	char *ldibuf;
+	int ldibuflen;
+	struct ld_info *ldi;
+
+	ldibuf = alloca(ldibuflen = 8192);
+
+	while ( (len = loadquery(L_GETINFO,ldibuf,ldibuflen)) < 0) {
+		if (errno != ENOMEM) {
+			ABORT("loadquery failed");
+		}
+		ldibuf = alloca(ldibuflen *= 2);
+	}
+
+	ldi = (struct ld_info *)ldibuf;
+	while (ldi) {
+		len = ldi->ldinfo_next;
+		GC_add_roots_inner(
+				ldi->ldinfo_dataorg,
+				(unsigned long)ldi->ldinfo_dataorg
+			        + ldi->ldinfo_datasize,
+				TRUE);
+		ldi = len ? (struct ld_info *)((char *)ldi + len) : 0;
+	}
+}
+#endif /* RS6000 */
+
 
 
 #else /* !DYNAMIC_LOADING */
