@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1998-1999 The University of Melbourne.
+% Copyright (C) 1998-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,7 +13,7 @@
 %
 % This module corresponds to what is called the "Query Handler" in Opium.
 
-:- module debugger_interface.
+:- module mdb__debugger_interface.
 :- interface. 
 
 % This module exports the following C functions:
@@ -32,7 +32,7 @@
 :- implementation.
 :- import_module io, require.
 :- import_module list, bool, std_util.
-:- import_module interactive_query, util.
+:- import_module mdb__interactive_query, mdb__util.
 
 dummy_pred_to_avoid_warning_about_nothing_exported.
 
@@ -44,14 +44,6 @@ dummy_pred_to_avoid_warning_about_nothing_exported.
 	% encoded as specified in ../runtime/mercury_stack_layout.h
 	% and ../compiler/stack_layout.m.
 
-% This enumeration must be EXACTLY the same as the MR_PredFunc enum in
-% runtime/mercury_stack_layout.h, and in the same order, since the code
-% assumes the representation is the same.
-
-:- type pred_or_func
-	--->	predicate
-	;	function.
-
 
 % Depending whether the Opium side is requesting for a user defined procedure
 % or a compiler generated one, the event has not exactly the same structure.
