@@ -1,10 +1,11 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
 %
-% This module generates the bytecode used by the debugger.
+% This module generates bytecode, which is intended to be used by a
+% (not yet implemented) bytecode interpreter/debugger.
 %
 % Author: zs.
 %
@@ -21,16 +22,35 @@
 	io__state::di, io__state::uo) is det.
 
 %---------------------------------------------------------------------------%
+%---------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds, arg_info.
-:- import_module passes_aux, call_gen, mode_util, code_util, goal_util.
-:- import_module globals, tree, varset, term.
+% We make use of some stuff from the LLDS back-end, in particular the stuff
+% relating to the argument passing convention in arg_info.m and call_gen.m.
+% The intent here is to use the same argument passing convention as for
+% the LLDS, to allow interoperability between code compiled to bytecode
+% and code compiled to machine code.
+%
+% XXX It might be nice to move the argument passing related stuff
+% in call_gen.m that we use here into arg_info.m, and to then rework
+% arg_info.m so that it didn't depend on the LLDS.
+
+:- import_module arg_info, call_gen. % XXX for arg passing convention
+:- import_module code_util.	% XXX for cons_id_to_tag
+
+:- import_module prog_data.
+:- import_module hlds_pred, hlds_goal, hlds_data.
+:- import_module type_util, mode_util, goal_util.
+:- import_module builtin_ops, code_model, passes_aux.
+:- import_module globals, tree.
+:- import_module prog_out.
 
 :- import_module bool, int, string, list, assoc_list, set, map, varset.
 :- import_module std_util, require, term.
 
+%---------------------------------------------------------------------------%
+
 bytecode_gen__module(ModuleInfo, Code) -->
 	{ module_info_predids(ModuleInfo, PredIds) },
 	bytecode_gen__preds(PredIds, ModuleInfo, CodeTree),
@@ -53,14 +73,7 @@ bytecode_gen__preds([PredId | PredIds], ModuleInfo, Code) -->
 		{ predicate_name(ModuleInfo, PredId, PredName) },
 		{ list__length(ProcIds, ProcsCount) },
 		{ pred_info_arity(PredInfo, Arity) },
-		{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
-		{ 
-			(PredOrFunc = predicate ->
-				IsFunc = 0
-			;
-				IsFunc = 1
-			)
-		},
+		{ bytecode_gen__get_is_func(PredInfo, IsFunc) },
 		{ EnterCode = node([enter_pred(PredName, Arity, IsFunc,
 			ProcsCount)]) },
 		{ EndofCode = node([endof_pred]) },
@@ -113,7 +126,17 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 	call_gen__output_arg_locs(Args, OutputArgs),
 	bytecode_gen__gen_places(OutputArgs, ByteInfo, PlaceCode),
 
-	bytecode_gen__goal(Goal, ByteInfo1, ByteInfo, GoalCode),
+	% If semideterministic, reserve temp slot 0 for the return value
+	( CodeModel = model_semi ->
+		bytecode_gen__get_next_temp(ByteInfo1, _FrameTemp, ByteInfo2)
+	;
+		ByteInfo2 = ByteInfo1
+	),
+
+	bytecode_gen__goal(Goal, ByteInfo2, ByteInfo3, GoalCode),
+
+	bytecode_gen__get_next_label(ByteInfo3, EndLabel, ByteInfo),
+	
 	bytecode_gen__get_counts(ByteInfo, LabelCount, TempCount),
 
 	ZeroLabelCode = node([label(ZeroLabel)]),
@@ -130,12 +153,12 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 		BodyCode = node(BodyCode0)
 	),
 	proc_id_to_int(ProcId, ProcInt),
-	EnterCode = node([enter_proc(ProcInt, Detism, LabelCount, TempCount,
-		VarInfos)]),
+	EnterCode = node([enter_proc(ProcInt, Detism, LabelCount, EndLabel,
+		TempCount, VarInfos)]),
 	( CodeModel = model_semi ->
-		EndofCode = node([semidet_succeed, endof_proc])
+		EndofCode = node([semidet_succeed, label(EndLabel), endof_proc])
 	;
-		EndofCode = node([endof_proc])
+		EndofCode = node([label(EndLabel), endof_proc])
 	),
 	Code = tree(EnterCode, tree(BodyCode, EndofCode)).
 
@@ -164,11 +187,13 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 			ByteInfo = ByteInfo0
 		;
 			% XXX
-			functor(GenericCallType, GenericCallFunctor, _),
-			string__append_list([
+			functor(GenericCallType, _GenericCallFunctor, _),
+			/*string__append_list([
 				"sorry: bytecode not yet implemented for ",
 				GenericCallFunctor, " calls"], Msg),
-			error(Msg)
+			error(Msg)*/
+			Code = node([not_supported]),
+			ByteInfo = ByteInfo0
 		)
 	;
 		GoalExpr = call(PredId, ProcId, ArgVars, BuiltinState, _, _),
@@ -188,10 +213,14 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 	;
 		GoalExpr = not(Goal),
 		bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, SomeCode),
-		bytecode_gen__get_next_label(ByteInfo1, EndLabel, ByteInfo),
-		EnterCode = node([enter_negation(EndLabel)]),
-		EndofCode = node([endof_negation, label(EndLabel)]),
-		Code = tree(EnterCode, tree(SomeCode, EndofCode))
+		bytecode_gen__get_next_label(ByteInfo1, EndLabel, ByteInfo2),
+		bytecode_gen__get_next_temp(ByteInfo2, FrameTemp, ByteInfo),
+		EnterCode = node([enter_negation(FrameTemp, EndLabel)]),
+		EndofCode = node([endof_negation_goal(FrameTemp),
+			label(EndLabel), endof_negation]),
+		Code =  tree(EnterCode,
+			tree(SomeCode,
+			     EndofCode))
 	;
 		GoalExpr = some(_, _, Goal),
 		bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, SomeCode),
@@ -238,7 +267,8 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 		bytecode_gen__goal(Else, ByteInfo5, ByteInfo, ElseCode),
 		EnterIfCode = node([enter_if(ElseLabel, EndLabel, FrameTemp)]),
 		EnterThenCode = node([enter_then(FrameTemp)]),
-		EndofThenCode = node([endof_then(EndLabel), label(ElseLabel)]),
+		EndofThenCode = node([endof_then(EndLabel), label(ElseLabel),
+			enter_else(FrameTemp)]),
 		EndofIfCode = node([endof_if, label(EndLabel)]),
 		Code =
 			tree(EnterIfCode,
@@ -249,7 +279,7 @@ bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 			tree(ElseCode,
 			     EndofIfCode))))))
 	;
-		GoalExpr = pragma_c_code(_, _, _, _, _, _, _),
+		GoalExpr = pragma_foreign_code(_, _, _, _, _, _, _),
 		Code = node([not_supported]),
 		ByteInfo = ByteInfo0
 	;
@@ -293,7 +323,7 @@ bytecode_gen__higher_order_call(PredVar, ArgVars, ArgModes, Detism,
 	make_arg_infos(ArgTypes, ArgModes, CodeModel, ModuleInfo, ArgInfo),
 	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, ArgVarsInfos),
 
-	call_gen__partition_args(ArgVarsInfos, InVars, OutVars),
+	arg_info__partition_args(ArgVarsInfos, InVars, OutVars),
 	list__length(InVars, NInVars),
 	list__length(OutVars, NOutVars),
 
@@ -324,6 +354,9 @@ bytecode_gen__call(PredId, ProcId, ArgVars, Detism, ByteInfo, Code) :-
 	proc_info_arg_info(ProcInfo, ArgInfo),
 	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, ArgVarsInfos),
 
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	bytecode_gen__get_is_func(PredInfo, IsFunc),
+
 	call_gen__input_arg_locs(ArgVarsInfos, InputArgs),
 	bytecode_gen__gen_places(InputArgs, ByteInfo, PlaceArgs),
 
@@ -332,7 +365,7 @@ bytecode_gen__call(PredId, ProcId, ArgVars, Detism, ByteInfo, Code) :-
 
 	predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
 	proc_id_to_int(ProcId, ProcInt),
-	Call = node([call(ModuleName, PredName, Arity, ProcInt)]),
+	Call = node([call(ModuleName, PredName, Arity, IsFunc, ProcInt)]),
 	determinism_to_code_model(Detism, CodeModel),
 	( CodeModel = model_semi ->
 		Check = node([semidet_success_check])
@@ -351,73 +384,76 @@ bytecode_gen__builtin(PredId, ProcId, Args, ByteInfo, Code) :-
 	predicate_module(ModuleInfo, PredId, ModuleName),
 	predicate_name(ModuleInfo, PredId, PredName),
 	(
-		code_util__translate_builtin(ModuleName, PredName, ProcId,
-			Args, MaybeTest, MaybeAssign)
+		builtin_ops__translate_builtin(ModuleName, PredName, ProcId,
+			Args, SimpleCode)
 	->
-		( MaybeTest = yes(Test) ->
-			bytecode_gen__map_test(ByteInfo, Test, TestCode)
+		(
+			SimpleCode = test(Test),
+			bytecode_gen__map_test(ByteInfo, Test, Code)
 		;
-			TestCode = empty
-		),
-		( MaybeAssign = yes(Var - Rval) ->
-			bytecode_gen__map_assign(ByteInfo, Var, Rval,
-				AssignCode)
-		;
-			AssignCode = empty
-		),
-		Code = tree(TestCode, AssignCode)
+			SimpleCode = assign(Var, Expr),
+			bytecode_gen__map_assign(ByteInfo, Var, Expr,
+				Code)
+		)
 	;
 		string__append("unknown builtin predicate ", PredName, Msg),
 		error(Msg)
 	).
 
-:- pred bytecode_gen__map_test(byte_info::in, rval::in, byte_tree::out) is det.
+:- pred bytecode_gen__map_test(byte_info::in,
+		simple_expr(prog_var)::in(simple_test_expr),
+		byte_tree::out) is det.
 
-bytecode_gen__map_test(ByteInfo, Rval, Code) :-
-	( Rval = binop(Binop, X, Y) ->
+bytecode_gen__map_test(ByteInfo, TestExpr, Code) :-
+	(
+		TestExpr = binary(Binop, X, Y),
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		bytecode_gen__map_arg(ByteInfo, Y, ByteY),
 		Code = node([builtin_bintest(Binop, ByteX, ByteY)])
-	; Rval = unop(Unop, X) ->
+	;
+		TestExpr = unary(Unop, X),
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		Code = node([builtin_untest(Unop, ByteX)])
-	;
-		error("builtin test is not in a recognized form")
 	).
 
-:- pred bytecode_gen__map_assign(byte_info::in, prog_var::in, rval::in,
-	byte_tree::out) is det.
+:- pred bytecode_gen__map_assign(byte_info::in, prog_var::in,
+		simple_expr(prog_var)::in(simple_assign_expr),
+		byte_tree::out) is det.
 
-bytecode_gen__map_assign(ByteInfo, Var, Rval, Code) :-
-	( Rval = binop(Binop, X, Y) ->
+bytecode_gen__map_assign(ByteInfo, Var, Expr, Code) :-
+	(
+		Expr = binary(Binop, X, Y),
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		bytecode_gen__map_arg(ByteInfo, Y, ByteY),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 		Code = node([builtin_binop(Binop, ByteX, ByteY, ByteVar)])
-	; Rval = unop(Unop, X) ->
+	;
+		Expr = unary(Unop, X),
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 		Code = node([builtin_unop(Unop, ByteX, ByteVar)])
-	; Rval = var(X) ->
+	;
+		Expr = leaf(X),
 		bytecode_gen__map_var(ByteInfo, X, ByteX),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 		Code = node([assign(ByteVar, ByteX)])
-	;
-		error("builtin assignment is not in a recognized form")
 	).
 
-:- pred bytecode_gen__map_arg(byte_info::in, rval::in, byte_arg::out) is det.
+:- pred bytecode_gen__map_arg(byte_info::in,
+		simple_expr(prog_var)::in(simple_arg_expr),
+		byte_arg::out) is det.
 
-bytecode_gen__map_arg(ByteInfo, Rval, ByteArg) :-
-	( Rval = var(Var) ->
+bytecode_gen__map_arg(ByteInfo, Expr, ByteArg) :-
+	(
+		Expr = leaf(Var),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 		ByteArg = var(ByteVar)
-	; Rval = const(int_const(IntVal)) ->
-		ByteArg = int_const(IntVal)
-	; Rval = const(float_const(FloatVal)) ->
-		ByteArg = float_const(FloatVal)
 	;
-		error("unknown kind of builtin argument")
+		Expr = int_const(IntVal),
+		ByteArg = int_const(IntVal)
+	;
+		Expr = float_const(FloatVal),
+		ByteArg = float_const(FloatVal)
 	).
 
 %---------------------------------------------------------------------------%
@@ -433,7 +469,7 @@ bytecode_gen__unify(construct(Var, ConsId, Args, UniModes, _, _, _),
 	bytecode_gen__map_vars(ByteInfo, Args, ByteArgs),
 	bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId),
 	(
-		ByteConsId = pred_const(_, _, _, _)
+		ByteConsId = pred_const(_, _, _, _, _)
 	->
 		Code = node([construct(ByteVar, ByteConsId, ByteArgs)])
 	;
@@ -452,8 +488,8 @@ bytecode_gen__unify(construct(Var, ConsId, Args, UniModes, _, _, _),
 				Pairs)])
 		)
 	).
-bytecode_gen__unify(deconstruct(Var, ConsId, Args, UniModes, _), _, _, ByteInfo,
-		Code) :-
+bytecode_gen__unify(deconstruct(Var, ConsId, Args, UniModes, _, _), _, _,
+		ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 	bytecode_gen__map_vars(ByteInfo, Args, ByteArgs),
 	bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId),
@@ -473,7 +509,51 @@ bytecode_gen__unify(assign(Target, Source), _, _, ByteInfo, Code) :-
 bytecode_gen__unify(simple_test(Var1, Var2), _, _, ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var1, ByteVar1),
 	bytecode_gen__map_var(ByteInfo, Var2, ByteVar2),
-	Code = node([test(ByteVar1, ByteVar2)]).
+	bytecode_gen__get_var_type(ByteInfo, Var1, Var1Type),
+	bytecode_gen__get_var_type(ByteInfo, Var2, Var2Type),
+
+	(	type_to_type_id(Var1Type, TypeId1, _),
+		type_to_type_id(Var2Type, TypeId2, _)
+	->	(	TypeId2 = TypeId1
+		->	TypeId = TypeId1
+		;	error("unexpected simple_test between different types")
+		)
+	;	error("failed lookup of type id")
+	),
+
+	ByteInfo = byte_info(_, _, ModuleInfo, _, _),
+
+	classify_type_id(ModuleInfo, TypeId, BuiltinType),
+
+	(	BuiltinType = int_type,
+		TestId = int_test
+	
+	;	BuiltinType = char_type,
+		TestId = char_test
+	
+	;	BuiltinType = str_type,
+		TestId = string_test
+	
+	;	BuiltinType = float_type,
+		TestId = float_test 
+	
+	;	BuiltinType = enum_type,
+		TestId = enum_test 
+
+	;	BuiltinType = pred_type,
+		error("unexpected pred_type in simple_test")
+	
+	;	BuiltinType = tuple_type,
+		error("unexpected tuple_type in simple_test")
+
+	;	BuiltinType = user_type,
+		error("unexpected user_type in simple_test")
+
+	;	BuiltinType = polymorphic_type,
+		error("unexpected polymorphic_type in simple_test")
+
+	),
+	Code = node([test(ByteVar1, ByteVar2, TestId)]).
 bytecode_gen__unify(complicated_unify(_,_,_), _Var, _RHS, _ByteInfo, _Code) :-
 	error("complicated unifications should have been handled by polymorphism.m").
 
@@ -590,10 +670,13 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 	bytecode_gen__get_module_info(ByteInfo, ModuleInfo),
 	(
 		ConsId = cons(Functor, Arity),
+		bytecode_gen__get_var_type(ByteInfo, Var, Type),
 		(
-			% Everything other than characters should
+			% Everything other than characters and tuples should
 			% be module qualified.
 			Functor = unqualified(FunctorName),
+			\+ type_is_tuple(Type, _)
+		->
 			string__to_char_list(FunctorName, FunctorList),
 			( FunctorList = [Char] ->
 				ByteConsId = char_const(Char)
@@ -601,8 +684,12 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 				error("bytecode_gen__map_cons_id: unqualified cons_id is not a char_const")
 			)
 		;
-			Functor = qualified(ModuleName, FunctorName),
-			bytecode_gen__get_var_type(ByteInfo, Var, Type),
+			(
+				Functor = unqualified(FunctorName),
+				ModuleName = unqualified("builtin")
+			;
+				Functor = qualified(ModuleName, FunctorName)
+			),
 			code_util__cons_id_to_tag(ConsId,
 				Type, ModuleInfo, ConsTag),
 			bytecode_gen__map_cons_tag(ConsTag, ByteConsTag),
@@ -623,9 +710,13 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 		( EvalMethod = normal ->
 			predicate_id(ModuleInfo, PredId,
 				ModuleName, PredName, Arity),
+
+			module_info_pred_info(ModuleInfo, PredId, PredInfo),
+			bytecode_gen__get_is_func(PredInfo, IsFunc),
+				
 			proc_id_to_int(ProcId, ProcInt),
 			ByteConsId = pred_const(ModuleName,
-				PredName, Arity, ProcInt)
+				PredName, Arity, IsFunc, ProcInt)
 		;
 			% XXX
 			error(
@@ -760,3 +851,14 @@ bytecode_gen__get_next_temp(ByteInfo0, Temp0, ByteInfo) :-
 bytecode_gen__get_counts(byte_info(_, _, _, Label, Temp), Label, Temp).
 
 %---------------------------------------------------------------------------%
+
+:- pred bytecode_gen__get_is_func(pred_info, byte_is_func).
+:- mode bytecode_gen__get_is_func(in, out) is det.
+
+bytecode_gen__get_is_func(PredInfo, IsFunc) :-
+	pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
+	(	PredOrFunc = predicate
+	->	IsFunc = 0
+	;	IsFunc = 1
+	).
+
