@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -49,37 +49,45 @@
 
 :- interface.
 
-:- import_module hlds_module, bool, io.
+:- import_module hlds_module, make_hlds.
+:- import_module bool, io.
 
-:- pred check_typeclass__check_instance_decls(module_info, module_info, bool,
-	io__state, io__state).
-:- mode check_typeclass__check_instance_decls(in, out, out, di, uo) is det.
+:- pred check_typeclass__check_instance_decls(module_info, qual_info,
+	module_info, bool, io__state, io__state).
+:- mode check_typeclass__check_instance_decls(in, in, out, out, di, uo) is det.
 
 :- implementation.
 
-:- import_module map, list, std_util, hlds_pred, hlds_data, prog_data, require.
-:- import_module type_util, assoc_list, mode_util, inst_match, hlds_module.
-:- import_module typecheck, int, globals, options, make_hlds, error_util. 
-:- import_module base_typeclass_info, string, hlds_goal, set, prog_out.
-:- import_module mercury_to_mercury, varset, term.
+:- import_module prog_data, prog_out.
+:- import_module hlds_pred, hlds_data, hlds_goal, hlds_out.
+:- import_module type_util, typecheck, mode_util, inst_match.
+:- import_module base_typeclass_info.
+:- import_module mercury_to_mercury, error_util.
+:- import_module globals, options. 
+
+:- import_module int, string.
+:- import_module list, assoc_list, map, set, term, varset.
+:- import_module std_util, require. 
 
 :- type error_message == pair(prog_context, list(format_component)).
 :- type error_messages == list(error_message).
 
-check_typeclass__check_instance_decls(ModuleInfo0, ModuleInfo, FoundError, 
-		IO0, IO) :-
+check_typeclass__check_instance_decls(ModuleInfo0, QualInfo0,
+		ModuleInfo, FoundError, IO0, IO) :-
 	module_info_classes(ModuleInfo0, ClassTable),
 	module_info_instances(ModuleInfo0, InstanceTable0),
 	map__to_assoc_list(InstanceTable0, InstanceList0),
-	list__map_foldl(check_one_class(ClassTable), InstanceList0,
-		InstanceList, [] - ModuleInfo0, Errors - ModuleInfo1),
+	list_map_foldl2(check_one_class(ClassTable), InstanceList0,
+		InstanceList, check_tc_info([], ModuleInfo0, QualInfo0),
+		check_tc_info(Errors, ModuleInfo1, _QualInfo),
+		IO0, IO1),
 	(
 		Errors = []
 	->
 		map__from_assoc_list(InstanceList, InstanceTable),
 		module_info_set_instances(ModuleInfo1, InstanceTable,
 			ModuleInfo),
-		IO = IO0,
+		IO = IO1,
 		FoundError = no
 	;
 		ModuleInfo = ModuleInfo1,
@@ -90,21 +98,40 @@ check_typeclass__check_instance_decls(ModuleInfo0, ModuleInfo, FoundError,
 				write_error_pieces(ErrorContext, 0, 
 					ErrorPieces, TheIO0, TheIO)
 			)),
-		list__foldl(WriteError, ErrorList, IO0, IO1),
-		io__set_exit_status(1, IO1, IO),
+		list__foldl(WriteError, ErrorList, IO1, IO2),
+		io__set_exit_status(1, IO2, IO),
 		FoundError = yes
 	).  
+
+:- type check_tc_info
+	--->	check_tc_info(error_messages, module_info, qual_info).
+
+	% list__map_foldl2(Pred, InList, OutList, StartA, EndA, StartB, EndB)
+	% calls Pred with two accumulator (with the initial values of
+	% StartA and StartB respectively) on each element of InList
+	% (working left-to-right) to transform InList into OutList.
+	% The final values of the accumulators are returned in EndA
+	% and EndB respectively.
+:- pred list_map_foldl2(pred(X, Y, Z, Z, W, W), list(X), list(Y), Z, Z, W, W).
+:- mode list_map_foldl2(pred(in, out, in, out, di, uo) is det,
+			     in, out, in, out, di, uo) is det.
+
+list_map_foldl2(_, [],  [], A, A) -->
+        [].
+list_map_foldl2(P, [H0|T0], [H|T], A0, A) -->
+        call(P, H0, H, A0, A1),
+        list_map_foldl2(P, T0, T, A1, A).
 		
 	% check all the instances of one class.
 :- pred check_one_class(class_table,
 	pair(class_id, list(hlds_instance_defn)), 
 	pair(class_id, list(hlds_instance_defn)), 
-	pair(error_messages, module_info), 
-	pair(error_messages, module_info)).
-:- mode check_one_class(in, in, out, in, out) is det.
+	check_tc_info, check_tc_info,
+	io__state, io__state).
+:- mode check_one_class(in, in, out, in, out, di, uo) is det.
 
 check_one_class(ClassTable, ClassId - InstanceDefns0, 
-	ClassId - InstanceDefns, ModuleInfo0, ModuleInfo) :-
+	ClassId - InstanceDefns, CheckTCInfo0, CheckTCInfo, IO0, IO) :-
 
 	map__lookup(ClassTable, ClassId, ClassDefn),
 	ClassDefn = hlds_class_defn(_, SuperClasses, ClassVars, _,
@@ -116,89 +143,85 @@ check_one_class(ClassTable, ClassId - InstanceDefns0,
 				ClassProc = hlds_class_proc(PredId, _)
 			)),
 		PredIds),
-	list__map_foldl(check_class_instance(ClassId, SuperClasses, ClassVars,
+	list_map_foldl2(check_class_instance(ClassId, SuperClasses, ClassVars,
 				ClassInterface, ClassVarSet,
 				PredIds),
 		InstanceDefns0, InstanceDefns, 
-		ModuleInfo0, ModuleInfo).
+		CheckTCInfo0, CheckTCInfo,
+		IO0, IO).
 
 	% check one instance of one class
 :- pred check_class_instance(class_id, list(class_constraint), list(tvar),
 	hlds_class_interface, tvarset, list(pred_id), 
 	hlds_instance_defn, hlds_instance_defn, 
-	pair(error_messages, module_info), 
-	pair(error_messages, module_info)).
-:- mode check_class_instance(in, in, in, in, in, in, in, out, in, out) is det.
+	check_tc_info, check_tc_info, 
+	io__state, io__state).
+:- mode check_class_instance(in, in, in, in, in, in, in, out, in, out,
+	di, uo) is det.
 
 check_class_instance(ClassId, SuperClasses, Vars, ClassInterface, ClassVarSet,
 		PredIds, InstanceDefn0, InstanceDefn, 
-		Errors0 - ModuleInfo0, Errors - ModuleInfo):-
+		check_tc_info(Errors0, ModuleInfo0, QualInfo0),
+		check_tc_info(Errors, ModuleInfo, QualInfo),
+		IO0, IO):-
 		
 		% check conformance of the instance body
-	InstanceDefn0 = hlds_instance_defn(_, _, _, _, InstanceBody, _, _, _),
+	InstanceDefn0 = hlds_instance_defn(_, _, _, _, _, InstanceBody,
+		_, _, _),
 	(
 		InstanceBody = abstract,
 		InstanceDefn2 = InstanceDefn0,
 		ModuleInfo1 = ModuleInfo0,
-		Errors2 = Errors0
+		QualInfo = QualInfo0,
+		Errors2 = Errors0,
+		IO = IO0
 	;
 		InstanceBody = concrete(InstanceMethods),
+		InstanceCheckInfo0 = instance_check_info(InstanceDefn0,
+				[], Errors0, ModuleInfo0, QualInfo0),
 		list__foldl2(
 			check_instance_pred(ClassId, Vars, ClassInterface), 
-			PredIds, InstanceDefn0, InstanceDefn1,
-			Errors0 - ModuleInfo0, Errors1 - ModuleInfo1),
+			PredIds, InstanceCheckInfo0, InstanceCheckInfo,
+			IO0, IO),
+		InstanceCheckInfo = instance_check_info(InstanceDefn1,
+				RevInstanceMethods, Errors1, ModuleInfo1,
+				QualInfo), 
+			
 		%
 		% We need to make sure that the MaybePredProcs field is
 		% set to yes(_) after this pass.  Normally that will be
 		% handled by check_instance_pred, but we also need to handle
 		% it below, in case the class has no methods.
 		%
-		InstanceDefn1 = hlds_instance_defn(A, B, C, D, E, 
-				MaybePredProcs1, G, H),
+		InstanceDefn1 = hlds_instance_defn(A, B, C, D, E, _, 
+				MaybePredProcs1, H, I),
 		(
 			MaybePredProcs1 = yes(_),
-			MaybePredProcs2 = MaybePredProcs1
+			MaybePredProcs = MaybePredProcs1
 		;
 			MaybePredProcs1 = no,
-			MaybePredProcs2 = yes([])
+			MaybePredProcs = yes([])
 		),
-		InstanceDefn2 = hlds_instance_defn(A, B, C, D, E, 
-				MaybePredProcs2, G, H),
+
+		%
+		% Make sure the list of instance methods is in the same
+		% order as the methods in the class definition. intermod.m
+		% relies on this
+		OrderedInstanceMethods = list__reverse(RevInstanceMethods),
+
+		InstanceDefn2 = hlds_instance_defn(A, B, C, D, E,
+				concrete(OrderedInstanceMethods),
+				MaybePredProcs, H, I),
+
 		%
 		% Check if there are any instance methods left over,
-		% for which we did not produce a pred_id/proc_id;
-		% if there are any, the instance declaration must have
-		% specified some methods that don't occur in the class.
+		% which did not match any of the methods from the
+		% class interface.
 		%
-		InstanceDefn2 = hlds_instance_defn(_, Context, _, _,
-				_, MaybePredProcs, _, _),
-		(
-			MaybePredProcs = yes(PredProcs),
-
-				% Check that we wind with a procedure for each
-				% proc in the type class interface.
-			list__same_length(PredProcs, ClassInterface),
-
-				% Check that we wind with a pred for each
-				% pred in the instance class interface.
-			list__map((pred(PP::in, P::out) is det :-
-				PP = hlds_class_proc(P, _)), PredProcs, Preds0),
-			list__remove_dups(Preds0, Preds),
-			list__same_length(Preds, InstanceMethods)
-		->
-			Errors2 = Errors1
-		;
-			ClassId = class_id(ClassName, ClassArity),
-			prog_out__sym_name_to_string(ClassName,
-				ClassNameString),
-			string__int_to_string(ClassArity, ClassArityString),
-			string__append_list([
-				"In instance declaration for `",
-				ClassNameString, "/", ClassArityString, "': ",
-				"incorrect method name(s)."],
-				NewError),
-			Errors2 = [Context - [words(NewError)] | Errors1]
-		)
+		InstanceDefn2 = hlds_instance_defn(_, _, Context,
+			_, _, _, _, _, _),
+		check_for_bogus_methods(InstanceMethods, ClassId, PredIds,
+			Context, ModuleInfo1, Errors1, Errors2)
 	),
 
 		% check that the superclass constraints are satisfied for the
@@ -207,23 +230,107 @@ check_class_instance(ClassId, SuperClasses, Vars, ClassInterface, ClassVarSet,
 		InstanceDefn2, InstanceDefn,
 		Errors2 - ModuleInfo1, Errors - ModuleInfo).
 
+		%
+		% Check if there are any instance methods left over,
+		% which did not match any of the methods from the
+		% class interface.  If so, add an appropriate error
+		% message to the list of error messages.
+		%
+:- pred check_for_bogus_methods(list(instance_method), class_id, list(pred_id),
+		prog_context, module_info, error_messages, error_messages).
+:- mode check_for_bogus_methods(in, in, in, in, in, in, out) is det.
+
+check_for_bogus_methods(InstanceMethods, ClassId, ClassPredIds, Context,
+		ModuleInfo1, Errors0, Errors) :-
+	module_info_get_predicate_table(ModuleInfo1, PredTable),
+	DefnIsOK = (pred(Method::in) is semidet :-
+		% Find this method definition's p/f, name, arity
+		Method = instance_method(MethodPredOrFunc,
+			MethodName, _MethodDefn,
+			MethodArity, _Context),
+		% Search for pred_ids matching that p/f, name, arity,
+		% and succeed if the method definition p/f, name, and
+		% arity matches at least one of the methods from the
+		% class interface
+		adjust_func_arity(MethodPredOrFunc, MethodArity,
+			MethodPredArity),
+		predicate_table_search_pf_sym_arity(PredTable,
+			MethodPredOrFunc, MethodName, MethodPredArity,
+			MatchingPredIds),
+		some [PredId] (
+			list__member(PredId, MatchingPredIds),
+			list__member(PredId, ClassPredIds)
+		)
+	),
+	list__filter(DefnIsOK, InstanceMethods, _OKInstanceMethods,
+		BogusInstanceMethods),
+	(
+		BogusInstanceMethods = []
+	->
+		Errors = Errors0
+	;
+		%
+		% There were one or more bogus methods.
+		% Construct an appropriate error message.
+		%
+		ClassId = class_id(ClassName, ClassArity),
+		prog_out__sym_name_to_string(ClassName,
+			ClassNameString),
+		string__int_to_string(ClassArity, ClassArityString),
+		string__append_list([
+			"In instance declaration for `",
+			ClassNameString, "/", ClassArityString, "': ",
+			"incorrect method name(s): "],
+			ErrorMsgStart),
+		BogusInstanceMethodNames = list__map(format_method_name,
+			BogusInstanceMethods),
+		error_util__list_to_pieces(BogusInstanceMethodNames,
+			ErrorMsgBody0),
+		ErrorMsgBody = list__append(ErrorMsgBody0, [words(".")]),
+		NewError = Context - [words(ErrorMsgStart) | ErrorMsgBody],
+		Errors = [NewError | Errors0]
+	).
+
+:- func format_method_name(instance_method) = string.
+format_method_name(Method) = StringName :-
+	Method = instance_method(PredOrFunc, Name, _Defn, Arity, _Context),
+	adjust_func_arity(PredOrFunc, Arity, PredArity),
+	hlds_out__simple_call_id_to_string(
+		PredOrFunc - Name/PredArity, StringName).
+
 %----------------------------------------------------------------------------%
 
+:- type instance_check_info
+	---> instance_check_info(
+		hlds_instance_defn,
+		instance_methods,	% The instance methods in reverse
+					% order of the methods in the class
+					% declaration.
+		error_messages,
+		module_info,
+		qual_info
+	).
+
 	% This structure holds the information about a particular instance
 	% method
 :- type instance_method_info ---> instance_method_info(
 		module_info,
+		qual_info,
 		sym_name,				% Name that the
 							% introduced pred
 							% should be given.
 		arity,					% Arity of the method.
+							% (For funcs, this is
+							% the original arity,
+							% not the arity as a
+							% predicate.)
 		existq_tvars,				% Existentially quant.
 							% type variables
 		list(type),				% Expected types of
 							% arguments.
 		class_constraints,			% Constraints from
 							% class method.
-		list(pair(list(mode), determinism)),	% Modes and 
+		list(modes_and_detism),			% Modes and 
 							% determinisms of the
 							% required procs.
 		error_messages,				% Error messages
@@ -240,13 +347,16 @@ check_class_instance(ClassId, SuperClasses, Vars, ClassInterface, ClassVarSet,
 
 	% check one pred in one instance of one class
 :- pred check_instance_pred(class_id, list(tvar), hlds_class_interface, 
-	pred_id, hlds_instance_defn, hlds_instance_defn,
-	pair(error_messages, module_info), pair(error_messages, module_info)).
-:- mode check_instance_pred(in,in, in, in, in, out, in, out) is det.
+	pred_id, instance_check_info, instance_check_info,
+	io__state, io__state).
+:- mode check_instance_pred(in,in, in, in, in, out, di, uo) is det.
 
 check_instance_pred(ClassId, ClassVars, ClassInterface, PredId,
-		InstanceDefn0, InstanceDefn, 
-		Errors0 - ModuleInfo0, Errors - ModuleInfo):-
+		InstanceCheckInfo0, InstanceCheckInfo, IO0, IO) :-
+
+	InstanceCheckInfo0 = instance_check_info(InstanceDefn0,
+				OrderedMethods0, Errors0, ModuleInfo0,
+				QualInfo0),
 	solutions(
 		lambda([ProcId::out] is nondet, 
 			(
@@ -257,6 +367,8 @@ check_instance_pred(ClassId, ClassVars, ClassInterface, PredId,
 	module_info_pred_info(ModuleInfo0, PredId, PredInfo),
 	pred_info_arg_types(PredInfo, ArgTypeVars, ExistQVars, ArgTypes),
 	pred_info_get_class_context(PredInfo, ClassContext0),
+	pred_info_get_markers(PredInfo, Markers0),
+	remove_marker(Markers0, class_method, Markers),
 		% The first constraint in the class context of a class method
 		% is always the constraint for the class of which it is
 		% a member. Seeing that we are checking an instance 
@@ -276,6 +388,7 @@ check_instance_pred(ClassId, ClassVars, ClassInterface, PredId,
 	MethodName = qualified(PredModule, MethodName0),
 	pred_info_arity(PredInfo, PredArity),
 	pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
+	adjust_func_arity(PredOrFunc, Arity, PredArity),
 	pred_info_procedures(PredInfo, ProcTable),
 	list__map(
 		lambda([TheProcId::in, ModesAndDetism::out] is det, 
@@ -284,54 +397,72 @@ check_instance_pred(ClassId, ClassVars, ClassInterface, PredId,
 				proc_info_argmodes(ProcInfo, Modes),
 				proc_info_interface_determinism(ProcInfo, 
 					Detism),
-				ModesAndDetism = Modes - Detism
+				proc_info_inst_varset(ProcInfo, InstVarSet),
+				ModesAndDetism = modes_and_detism(Modes,
+						InstVarSet, Detism)
 			)),
 		ProcIds, 
 		ArgModes),
 	
-	InstanceDefn0 = hlds_instance_defn(Status, _, _, InstanceTypes, 
+	InstanceDefn0 = hlds_instance_defn(_, Status, _, _, InstanceTypes, 
 		_, _, _, _),
 
 		% Work out the name of the predicate that we will generate
 		% to check this instance method.
-	make_introduced_pred_name(ClassId, MethodName, PredArity, 
+	make_introduced_pred_name(ClassId, MethodName, Arity, 
 		InstanceTypes, PredName),
 	
-	Info0 = instance_method_info(ModuleInfo0, PredName, PredArity, 
-		ExistQVars, ArgTypes, ClassContext, ArgModes, Errors0,
-		ArgTypeVars, Status, PredOrFunc),
+	Info0 = instance_method_info(ModuleInfo0, QualInfo0, PredName,
+		Arity, ExistQVars, ArgTypes, ClassContext, ArgModes,
+		Errors0, ArgTypeVars, Status, PredOrFunc),
 
-	check_instance_pred_procs(ClassId, ClassVars, MethodName,
-		InstanceDefn0, InstanceDefn, Info0, Info),
+	check_instance_pred_procs(ClassId, ClassVars, MethodName, Markers,
+		InstanceDefn0, InstanceDefn, OrderedMethods0, OrderedMethods,
+		Info0, Info, IO0, IO),
 
-	Info = instance_method_info(ModuleInfo, _PredName, _PredArity, 
-		_ExistQVars, _ArgTypes, _ClassContext, _ArgModes, Errors,
-		_ArgTypeVars, _Status, _PredOrFunc).
+	Info = instance_method_info(ModuleInfo, QualInfo, _PredName,
+		_Arity, _ExistQVars, _ArgTypes, _ClassContext, _ArgModes,
+		Errors, _ArgTypeVars, _Status, _PredOrFunc),
 
-:- pred check_instance_pred_procs(class_id, list(tvar), sym_name,
+	InstanceCheckInfo = instance_check_info(InstanceDefn,
+				OrderedMethods, Errors, ModuleInfo, QualInfo).
+
+:- type modes_and_detism
+	--->	modes_and_detism(list(mode), inst_varset, determinism).
+
+:- pred check_instance_pred_procs(class_id, list(tvar), sym_name, pred_markers,
 	hlds_instance_defn, hlds_instance_defn, 
-	instance_method_info, instance_method_info).
-:- mode check_instance_pred_procs(in, in, in, in, out, in, out) is det.
+	instance_methods, instance_methods,
+	instance_method_info, instance_method_info,
+	io__state, io__state).
+:- mode check_instance_pred_procs(in, in, in, in, in, out,
+	in, out, in, out, di, uo) is det.
 
-check_instance_pred_procs(ClassId, ClassVars, MethodName, InstanceDefn0,
-		InstanceDefn, Info0, Info) :-
-	InstanceDefn0 = hlds_instance_defn(A, InstanceContext, 
+check_instance_pred_procs(ClassId, ClassVars, MethodName, Markers,
+		InstanceDefn0, InstanceDefn, OrderedInstanceMethods0,
+		OrderedInstanceMethods, Info0, Info, IO0, IO) :-
+	InstanceDefn0 = hlds_instance_defn(A, B, InstanceContext, 
 				InstanceConstraints, InstanceTypes,
 				InstanceBody, MaybeInstancePredProcs,
-				InstanceVarSet, H),
-	Info0 = instance_method_info(ModuleInfo, PredName, PredArity, 
+				InstanceVarSet, I),
+	Info0 = instance_method_info(ModuleInfo, QualInfo, PredName, Arity,
 		ExistQVars, ArgTypes, ClassContext, ArgModes, Errors0,
 		ArgTypeVars, Status, PredOrFunc),
-	get_matching_instance_names(InstanceBody, PredOrFunc, MethodName,
-		PredArity, InstanceNames),
+	get_matching_instance_defns(InstanceBody, PredOrFunc, MethodName,
+		Arity, MatchingInstanceMethods),
 	(
-		InstanceNames = [InstancePredName - Context]
+		MatchingInstanceMethods = [InstanceMethod]
 	->
-		produce_auxiliary_procs(ClassVars, 
+		OrderedInstanceMethods =
+			[InstanceMethod | OrderedInstanceMethods0],
+		InstanceMethod = instance_method(_, _, InstancePredDefn,
+					_, Context),
+		produce_auxiliary_procs(ClassId, ClassVars, Markers,
 			InstanceTypes, InstanceConstraints, 
 			InstanceVarSet, 
-			InstancePredName, Context,
-			InstancePredId, InstanceProcIds, Info0, Info),
+			InstancePredDefn, Context,
+			InstancePredId, InstanceProcIds, Info0, Info,
+			IO0, IO),
 
 		MakeClassProc = 
 			lambda([TheProcId::in, PredProcId::out] is det,
@@ -348,27 +479,22 @@ check_instance_pred_procs(ClassId, ClassVars, MethodName, InstanceDefn0,
 			MaybeInstancePredProcs = no,
 			InstancePredProcs = InstancePredProcs1
 		),
-		InstanceDefn = hlds_instance_defn(A, Context, 
+		InstanceDefn = hlds_instance_defn(A, B, Context, 
 			InstanceConstraints, InstanceTypes, InstanceBody,
-			yes(InstancePredProcs), InstanceVarSet, H)
+			yes(InstancePredProcs), InstanceVarSet, I)
 	;
-		InstanceNames = [I1, I2 | Is]
+		MatchingInstanceMethods = [I1, I2 | Is]
 	->
-			% one kind of error
+		%
+		% duplicate method definition error
+		%
+		OrderedInstanceMethods = OrderedInstanceMethods0,
 		InstanceDefn = InstanceDefn0,
 		ClassId = class_id(ClassName, _ClassArity),
 		prog_out__sym_name_to_string(MethodName, MethodNameString),
 		prog_out__sym_name_to_string(ClassName, ClassNameString),
-		(
-			PredOrFunc = predicate,
-			PredOrFuncString = "predicate",
-			RealPredArity = PredArity
-		;
-			PredOrFunc = function,
-			PredOrFuncString = "function",
-			RealPredArity = PredArity - 1
-		),
-		string__int_to_string(RealPredArity, PredArityString),
+		pred_or_func_to_string(PredOrFunc, PredOrFuncString),
+		string__int_to_string(Arity, ArityString),
 		mercury_type_list_to_string(InstanceVarSet, InstanceTypes,
 			InstanceTypesString),
 		string__append_list([
@@ -376,15 +502,15 @@ check_instance_pred_procs(ClassId, ClassVars, MethodName, InstanceDefn0,
 			ClassNameString, "(", InstanceTypesString, ")': ",
 			"multiple implementations of type class ",
 			PredOrFuncString, " method `",
-			MethodNameString, "/", PredArityString, "'."],
+			MethodNameString, "/", ArityString, "'."],
 			ErrorHeader),
-		I1 = _ - I1Context, 
+		I1 = instance_method(_, _, _, _, I1Context), 
 		Heading = 
 			[I1Context - [words("First definition appears here.")],
 			InstanceContext - [words(ErrorHeader)]],
 		list__map(lambda([Definition::in, ContextAndError::out] is det,
 		(
-			Definition = _ - TheContext,
+			Definition = instance_method(_, _, _, _, TheContext),
 			Error = [words("Subsequent definition appears here.")],
 			ContextAndError = TheContext - Error
 		)), [I2|Is], SubsequentErrors),
@@ -392,25 +518,21 @@ check_instance_pred_procs(ClassId, ClassVars, MethodName, InstanceDefn0,
 			% errors are built up in reverse.
 		list__append(SubsequentErrors, Heading, NewErrors),
 		list__append(NewErrors, Errors0, Errors),
-		Info = instance_method_info(ModuleInfo, PredName, PredArity,
-			ExistQVars, ArgTypes, ClassContext, ArgModes, Errors,
-			ArgTypeVars, Status, PredOrFunc)
+		Info = instance_method_info(ModuleInfo, QualInfo, PredName,
+			Arity, ExistQVars, ArgTypes, ClassContext,
+			ArgModes, Errors, ArgTypeVars, Status, PredOrFunc),
+		IO = IO0
 	;
-			% another kind of error
+		%
+		% undefined method error
+		%
+		OrderedInstanceMethods = OrderedInstanceMethods0,
 		InstanceDefn = InstanceDefn0,
 		ClassId = class_id(ClassName, _ClassArity),
 		prog_out__sym_name_to_string(MethodName, MethodNameString),
 		prog_out__sym_name_to_string(ClassName, ClassNameString),
-		(
-			PredOrFunc = predicate,
-			PredOrFuncString = "predicate",
-			RealPredArity = PredArity
-		;
-			PredOrFunc = function,
-			PredOrFuncString = "function",
-			RealPredArity = PredArity - 1
-		),
-		string__int_to_string(RealPredArity, PredArityString),
+		pred_or_func_to_string(PredOrFunc, PredOrFuncString),
+		string__int_to_string(Arity, ArityString),
 		mercury_type_list_to_string(InstanceVarSet, InstanceTypes,
 			InstanceTypesString),
 		string__append_list([
@@ -418,90 +540,105 @@ check_instance_pred_procs(ClassId, ClassVars, MethodName, InstanceDefn0,
 			ClassNameString, "(", InstanceTypesString, ")': ",
 			"no implementation for type class ",
 			PredOrFuncString, " method `",
-			MethodNameString, "/", PredArityString, "'."],
+			MethodNameString, "/", ArityString, "'."],
 			NewError),
 		Errors = [InstanceContext - [words(NewError)] | Errors0],
-		Info = instance_method_info(ModuleInfo, PredName, PredArity,
-			ExistQVars, ArgTypes, ClassContext, ArgModes, Errors,
-			ArgTypeVars, Status, PredOrFunc)
+		Info = instance_method_info(ModuleInfo, QualInfo, PredName,
+			Arity, ExistQVars, ArgTypes, ClassContext,
+			ArgModes, Errors,
+			ArgTypeVars, Status, PredOrFunc),
+		IO = IO0
 	).
 
-:- pred get_matching_instance_names(instance_body, pred_or_func,
-	sym_name, arity, list(pair(sym_name, prog_context))).
-:- mode get_matching_instance_names(in, in, in, in, out) is det.
+	%
+	% Get all the instance definitions which match the specified
+	% predicate/function name/arity, with multiple clause definitions
+	% being combined into a single definition.
+	%
+:- pred get_matching_instance_defns(instance_body, pred_or_func,
+	sym_name, arity, list(instance_method)).
+:- mode get_matching_instance_defns(in, in, in, in, out) is det.
 
-get_matching_instance_names(InstanceBody, PredOrFunc, PredName,
-	PredArity, InstanceNames) :-
+get_matching_instance_defns(abstract, _, _, _, []).
+get_matching_instance_defns(concrete(InstanceMethods), PredOrFunc, MethodName,
+		MethodArity, ResultList) :-
+	%
+	% First find the instance method definitions that match this
+	% predicate/function's name and arity
+	%
+	list__filter(
+		(pred(Method::in) is semidet :-
+			Method = instance_method(PredOrFunc,
+				MethodName, _MethodDefn,
+				MethodArity, _Context)
+		),
+		InstanceMethods, MatchingMethods),
 	(
-		PredOrFunc = predicate,
-		solutions(
-			lambda([Pair::out] is nondet, 
-				(
-					InstanceBody =
-						concrete(InstanceMethods),
-					list__member(Method, InstanceMethods),
-					Method = pred_instance(PredName, 
-							SymName, PredArity,
-							Context),
-					Pair = SymName - Context
-				)),
-			InstanceNames)
+		MatchingMethods = [First, _Second | _],
+		First = instance_method(_, _, _, _, FirstContext),
+		\+ (
+			list__member(DefnViaName, MatchingMethods),
+			DefnViaName = instance_method(_, _, name(_), _, _)
+		)
+	->
+		%
+		% If all of the instance method definitions for this
+		% pred/func are clauses, and there are more than one
+		% of them, then we must combine them all into a
+		% single definition.
+		%
+		MethodToClause = (pred(Method::in, Clauses::out) is semidet :-
+			Method = instance_method(_, _, Defn, _, _),
+			Defn = clauses(Clauses)),
+		list__filter_map(MethodToClause, MatchingMethods, ClausesList),
+		list__condense(ClausesList, FlattenedClauses),
+		CombinedMethod = instance_method(PredOrFunc,
+			MethodName, clauses(FlattenedClauses),
+			MethodArity, FirstContext),
+		ResultList = [CombinedMethod]
 	;
-		PredOrFunc = function,
-		FuncArity is PredArity - 1,
-		solutions(
-			lambda([Pair::out] is nondet, 
-				(
-					InstanceBody =
-						concrete(InstanceMethods),
-					list__member(Method, InstanceMethods),
-					Method = func_instance(PredName, 
-							SymName, FuncArity,
-							Context),
-					Pair = SymName - Context
-				)),
-			InstanceNames)
+		%
+		% If there are less than two matching method definitions,
+		% or if any of the instance method definitions is a method
+		% name, then we're done.
+		%
+		ResultList = MatchingMethods
 	).
+	
+:- pred pred_or_func_to_string(pred_or_func::in, string::out) is det.
+pred_or_func_to_string(predicate, "predicate").
+pred_or_func_to_string(function, "function").
 
-	% Just a bit simpler than using a pair of pairs
-:- type triple(T1, T2, T3) ---> triple(T1, T2, T3).
+:- pred produce_auxiliary_procs(class_id, list(tvar), pred_markers, list(type),
+	list(class_constraint), tvarset, instance_proc_def, prog_context,
+	pred_id, list(proc_id), instance_method_info, instance_method_info,
+	io__state, io__state).
+:- mode produce_auxiliary_procs(in, in, in, in, in, in, in, in, out, out, 
+	in, out, di, uo) is det.
 
-:- pred produce_auxiliary_procs(list(tvar), 
-	list(type), list(class_constraint), tvarset, sym_name, prog_context,
-	pred_id, list(proc_id), instance_method_info, instance_method_info).
-:- mode produce_auxiliary_procs(in, in, in, in, in, in, out, out, 
-	in, out) is det.
-
-produce_auxiliary_procs(ClassVars, 
+produce_auxiliary_procs(ClassId, ClassVars, Markers0,
 		InstanceTypes0, InstanceConstraints0, InstanceVarSet,
-		InstancePredName, Context, PredId,
-		InstanceProcIds, Info0, Info) :-
+		InstancePredDefn, Context, PredId,
+		InstanceProcIds, Info0, Info, IO0, IO) :-
 
-	Info0 = instance_method_info(ModuleInfo0, PredName, PredArity, 
-		ExistQVars0, ArgTypes0, ClassContext0, ArgModes, Errors,
-		ArgTypeVars0, Status0, PredOrFunc),
+	Info0 = instance_method_info(ModuleInfo0, QualInfo0, PredName,
+		Arity, ExistQVars0, ArgTypes0, ClassMethodClassContext0,
+		ArgModes, Errors, ArgTypeVars0, Status0, PredOrFunc),
 
 		% Rename the instance variables apart from the class variables
 	varset__merge_subst(ArgTypeVars0, InstanceVarSet, ArgTypeVars1,
 		RenameSubst),
 	term__apply_substitution_to_list(InstanceTypes0, RenameSubst,
-		InstanceTypes),
+		InstanceTypes1),
 	apply_subst_to_constraint_list(RenameSubst, InstanceConstraints0,
-		InstanceConstraints),
+		InstanceConstraints1),
 
 		% Work out what the type variables are bound to for this
 		% instance, and update the class types appropriately.
-	map__from_corresponding_lists(ClassVars, InstanceTypes, TypeSubst),
+	map__from_corresponding_lists(ClassVars, InstanceTypes1, TypeSubst),
 	term__apply_substitution_to_list(ArgTypes0, TypeSubst, ArgTypes1),
-	apply_subst_to_constraints(TypeSubst, ClassContext0, ClassContext1),
-
-		% Add the constraints from the instance declaration to the 
-		% constraints from the class method. This allows an instance
-		% method to have constraints on it which are not part of the
-		% instance declaration as a whole.
-	ClassContext1 = constraints(UnivConstraints1, ExistConstraints),
-	list__append(InstanceConstraints, UnivConstraints1, UnivConstraints),
-	ClassContext2 = constraints(UnivConstraints, ExistConstraints),
+	apply_subst_to_constraints(TypeSubst, ClassMethodClassContext0,
+		ClassMethodClassContext1),
 
 		% Get rid of any unwanted type variables
 	term__vars_list(ArgTypes1, VarsToKeep0),
@@ -510,8 +647,21 @@ produce_auxiliary_procs(ClassVars,
 	term__apply_variable_renaming_to_list(ArgTypes1, SquashSubst, 
 		ArgTypes),
 	apply_variable_renaming_to_constraints(SquashSubst,
-		ClassContext2, ClassContext),
+		ClassMethodClassContext1, ClassMethodClassContext),
 	apply_partial_map_to_list(ExistQVars0, SquashSubst, ExistQVars),
+	apply_variable_renaming_to_list(InstanceTypes1, SquashSubst,
+		InstanceTypes),
+	apply_variable_renaming_to_constraint_list(SquashSubst,
+		InstanceConstraints1, InstanceConstraints),
+
+		% Add the constraints from the instance declaration to the 
+		% constraints from the class method. This allows an instance
+		% method to have constraints on it which are not part of the
+		% instance declaration as a whole.
+	ClassMethodClassContext = constraints(UnivConstraints1,
+					ExistConstraints),
+	list__append(InstanceConstraints, UnivConstraints1, UnivConstraints),
+	ClassContext = constraints(UnivConstraints, ExistConstraints),
 
 		% Introduce a new predicate which calls the implementation
 		% given in the instance declaration.
@@ -519,22 +669,24 @@ produce_auxiliary_procs(ClassVars,
 
 	Cond = true,
 	map__init(Proofs),
-	init_markers(Markers0),
-	add_marker(Markers0, class_instance_method, Markers),
+	add_marker(Markers0, class_instance_method, Markers1),
+	( InstancePredDefn = name(_) ->
+		% For instance methods which are defined using the named
+		% syntax (e.g. "pred(...) is ...") rather than the clauses
+		% syntax, we record an additional marker; the only effect
+		% of this marker is that we output slightly different
+		% error messages for such predicates.
+		add_marker(Markers1, named_class_instance_method, Markers)
+	;
+		Markers = Markers1
+	),
 	module_info_globals(ModuleInfo0, Globals),
 	globals__lookup_string_option(Globals, aditi_user, User),
 
-		% We have to add the actual clause after we have added the
-		% procs because we need a list of proc numbers for which the
-		% clauses holds.
-	DummyClause = [],
-	varset__init(VarSet0),
-	make_n_fresh_vars("HeadVar__", PredArity, VarSet0, HeadVars, VarSet), 
-	map__from_corresponding_lists(HeadVars, ArgTypes, VarTypes),
-	map__init(TI_VarMap),
-	map__init(TCI_VarMap),
-	ClausesInfo0 = clauses_info(VarSet, VarTypes, VarTypes, HeadVars,
-		DummyClause, TI_VarMap, TCI_VarMap),
+	adjust_func_arity(PredOrFunc, Arity, PredArity),
+	produce_instance_method_clauses(InstancePredDefn, PredOrFunc,
+		PredArity, ArgTypes, Markers, Context, ClausesInfo,
+		ModuleInfo0, ModuleInfo1, QualInfo0, QualInfo, IO0, IO),
 
 	( status_is_imported(Status0, yes) ->
 		Status = opt_imported
@@ -543,63 +695,42 @@ produce_auxiliary_procs(ClassVars,
 	),
 
 	pred_info_init(ModuleName, PredName, PredArity, ArgTypeVars, 
-		ExistQVars, ArgTypes, Cond, Context, ClausesInfo0, Status,
+		ExistQVars, ArgTypes, Cond, Context, ClausesInfo, Status,
 		Markers, none, PredOrFunc, ClassContext, Proofs, User,
 		PredInfo0),
+	pred_info_set_clauses_info(PredInfo0, ClausesInfo, PredInfo1),
+
+	% Fill in some information in the pred_info which is
+	% used by polymorphism to make sure the type-infos
+	% and typeclass-infos are added in the correct order.
+	MethodConstraints = instance_method_constraints(ClassId,
+			InstanceTypes, InstanceConstraints,
+			ClassMethodClassContext),
+	pred_info_set_maybe_instance_method_constraints(PredInfo1,
+		yes(MethodConstraints), PredInfo2),
 
 		% Add procs with the expected modes and determinisms
 	AddProc = lambda([ModeAndDet::in, NewProcId::out,
 			OldPredInfo::in, NewPredInfo::out] is det,
 	(
-		ModeAndDet = Modes - Det,
-		add_new_proc(OldPredInfo, PredArity, Modes, yes(Modes), no,
-			yes(Det), Context, address_is_taken,
+		ModeAndDet = modes_and_detism(Modes, InstVarSet, Det),
+		add_new_proc(OldPredInfo, InstVarSet, PredArity, Modes,
+			yes(Modes), no, yes(Det), Context, address_is_taken,
 			NewPredInfo, NewProcId)
 	)),
 	list__map_foldl(AddProc, ArgModes, InstanceProcIds, 
-		PredInfo0, PredInfo1),
+		PredInfo2, PredInfo),
 
-		% Add the body of the introduced pred
-
-		% First the goal info
-	goal_info_init(GoalInfo0),
-	goal_info_set_context(GoalInfo0, Context, GoalInfo1),
-	set__list_to_set(HeadVars, NonLocals),
-	goal_info_set_nonlocals(GoalInfo1, NonLocals, GoalInfo),
-
-		% Then the goal itself
-	invalid_pred_id(InvalidPredId),
-	invalid_proc_id(InvalidProcId),
-	(
-		PredOrFunc = predicate,
-		Call = call(InvalidPredId, InvalidProcId, HeadVars, not_builtin,
-			no, InstancePredName),
-		IntroducedGoal = Call - GoalInfo
-	;
-		PredOrFunc = function,
-		pred_args_to_func_args(HeadVars, RealHeadVars, ReturnVar),
-		create_atomic_unification(ReturnVar, 
-			functor(cons(InstancePredName, PredArity),
-				RealHeadVars), 
-			Context, explicit, [], IntroducedGoal0),
-		% set the goal_info
-		IntroducedGoal0 = IntroducedGoalExpr - _,
-		IntroducedGoal = IntroducedGoalExpr - GoalInfo
-	),
-	IntroducedClause = clause(InstanceProcIds, IntroducedGoal, Context),
-	clauses_info_set_clauses(ClausesInfo0, [IntroducedClause], ClausesInfo),
-	pred_info_set_clauses_info(PredInfo1, ClausesInfo, PredInfo),
-
-	module_info_get_predicate_table(ModuleInfo0, PredicateTable0),
-	module_info_get_partial_qualifier_info(ModuleInfo0, PQInfo),
+	module_info_get_predicate_table(ModuleInfo1, PredicateTable1),
+	module_info_get_partial_qualifier_info(ModuleInfo1, PQInfo),
 	% XXX why do we need to pass may_be_unqualified here,
 	%     rather than passing must_be_qualified or calling the /4 version?
-	predicate_table_insert(PredicateTable0, PredInfo,
+	predicate_table_insert(PredicateTable1, PredInfo,
 		may_be_unqualified, PQInfo, PredId, PredicateTable),
-	module_info_set_predicate_table(ModuleInfo0, PredicateTable,
+	module_info_set_predicate_table(ModuleInfo1, PredicateTable,
 		ModuleInfo),
 
-	Info = instance_method_info(ModuleInfo, PredName, PredArity,
+	Info = instance_method_info(ModuleInfo, QualInfo, PredName, Arity,
 		ExistQVars, ArgTypes, ClassContext, ArgModes, Errors,
 		ArgTypeVars, Status, PredOrFunc).
 
@@ -621,14 +752,14 @@ apply_substitution_to_var_list(Vars0, RenameSubst, Vars) :-
 	sym_name).
 :- mode make_introduced_pred_name(in, in, in, in, out) is det.
 
-make_introduced_pred_name(ClassId, MethodName, PredArity, 
+make_introduced_pred_name(ClassId, MethodName, Arity, 
 		InstanceTypes, PredName) :-
 	ClassId = class_id(ClassName, _ClassArity),
 	prog_out__sym_name_to_string(ClassName, "__", ClassNameString),
 	prog_out__sym_name_to_string(MethodName, "__", MethodNameString),
-		% Perhaps we should include the pred arity in this mangled
+		% Perhaps we should include the arity in this mangled
 		% string?
-	string__int_to_string(PredArity, PredArityString),
+	string__int_to_string(Arity, ArityString),
 	base_typeclass_info__make_instance_string(InstanceTypes, 
 		InstanceString),
 	string__append_list(
@@ -636,7 +767,7 @@ make_introduced_pred_name(ClassId, MethodName, PredArity,
 		ClassNameString, "__",
 		InstanceString, "____",
 		MethodNameString, "_",
-		PredArityString], 
+		ArityString], 
 		PredNameString),
 	PredName = unqualified(PredNameString).
 
@@ -660,8 +791,8 @@ check_superclass_conformance(ClassId, SuperClasses0, ClassVars0, ClassVarSet,
 		InstanceDefn0, InstanceDefn, 
 		Errors0 - ModuleInfo, Errors - ModuleInfo) :-
 
-	InstanceDefn0 = hlds_instance_defn(A, Context, InstanceConstraints,
-		InstanceTypes, E, F, InstanceVarSet0, Proofs0),
+	InstanceDefn0 = hlds_instance_defn(A, B, Context, InstanceConstraints,
+		InstanceTypes, F, G, InstanceVarSet0, Proofs0),
 	varset__merge_subst(InstanceVarSet0, ClassVarSet, InstanceVarSet1,
 		Subst),
 
@@ -696,8 +827,8 @@ check_superclass_conformance(ClassId, SuperClasses0, ClassVars0, ClassVarSet,
 		UnprovenConstraints = []
 	->
 		Errors = Errors0,
-		InstanceDefn = hlds_instance_defn(A, Context, 
-			InstanceConstraints, InstanceTypes, E, F, 
+		InstanceDefn = hlds_instance_defn(A, B, Context, 
+			InstanceConstraints, InstanceTypes, F, G,
 			InstanceVarSet2, Proofs1)
 	;
 		ClassId = class_id(ClassName, _ClassArity),
