@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -138,21 +138,21 @@
 		exprn_info, exprn_info).
 :- mode code_exprn__produce_var(in, out, out, in, out) is det.
 
-%	code_exprn__produce_var_in_reg(Var, Rval, Code, ExprnInfo0, ExprnInfo)
+%	code_exprn__produce_var_in_reg(Var, Lval, Code, ExprnInfo0, ExprnInfo)
 %		Produces a code fragment Code to evaluate Var and
-%		provide it as an Rval of the form lval(reg(_)).
+%		provide it as an Lval of the form reg(_).
 
-:- pred code_exprn__produce_var_in_reg(prog_var, rval, code_tree,
+:- pred code_exprn__produce_var_in_reg(prog_var, lval, code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg(in, out, out, in, out) is det.
 
-%	code_exprn__produce_var_in_reg_or_stack(Var, FollowVars, Rval, Code,
+%	code_exprn__produce_var_in_reg_or_stack(Var, FollowVars, Lval, Code,
 %			ExprnInfo0, ExprnInfo)
 %		Produces a code fragment Code to evaluate Var and
-%		provide it as an Rval of the form lval(reg(_)),
-%		lval(stackvar(_)), or lval(framevar(_)).
+%		provide it as an Lval of the form reg(_), stackvar(_),
+%		or framevar(_).
 
-:- pred code_exprn__produce_var_in_reg_or_stack(prog_var, rval, code_tree,
+:- pred code_exprn__produce_var_in_reg_or_stack(prog_var, lval, code_tree,
 	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg_or_stack(in, out, out, in, out) is det.
 
@@ -414,8 +414,8 @@ code_exprn__rval_depends_on_reg(var(Var), Vars) :-
 		code_exprn__filter_out_reg_depending(Rvals0, Vars, Rvals),
 		set__empty(Rvals)
 	).
-code_exprn__rval_depends_on_reg(create(_, Rvals, _, _, _, _), Vars) :-
-	code_exprn__args_depend_on_reg(Rvals, Vars).
+code_exprn__rval_depends_on_reg(create(_, Rvals, _, _, _, _, Reuse), Vars) :-
+	code_exprn__args_depend_on_reg([Reuse | Rvals], Vars).
 code_exprn__rval_depends_on_reg(mkword(_Tag, Rval), Vars) :-
 	code_exprn__rval_depends_on_reg(Rval, Vars).
 code_exprn__rval_depends_on_reg(const(_Const), _Vars) :-
@@ -616,8 +616,8 @@ code_exprn__add_rval_list_reg_dependencies([R | Rs]) -->
 code_exprn__add_rval_reg_dependencies(lval(Lval)) -->
 	code_exprn__add_lval_reg_dependencies(Lval).
 code_exprn__add_rval_reg_dependencies(var(_Var)) --> [].
-code_exprn__add_rval_reg_dependencies(create(_, Rvals, _, _, _, _)) -->
-	code_exprn__add_arg_reg_dependencies(Rvals).
+code_exprn__add_rval_reg_dependencies(create(_, Rvals, _, _, _, _, Reuse)) -->
+	code_exprn__add_arg_reg_dependencies([Reuse | Rvals]).
 code_exprn__add_rval_reg_dependencies(mkword(_Tag, Rval)) -->
 	code_exprn__add_rval_reg_dependencies(Rval).
 code_exprn__add_rval_reg_dependencies(const(_Const)) --> [].
@@ -689,8 +689,8 @@ code_exprn__rem_rval_list_reg_dependencies([R | Rs]) -->
 code_exprn__rem_rval_reg_dependencies(lval(Lval)) -->
 	code_exprn__rem_lval_reg_dependencies(Lval).
 code_exprn__rem_rval_reg_dependencies(var(_Var)) --> [].
-code_exprn__rem_rval_reg_dependencies(create(_, Rvals, _, _, _, _)) -->
-	code_exprn__rem_arg_reg_dependencies(Rvals).
+code_exprn__rem_rval_reg_dependencies(create(_, Rvals, _, _, _, _, Reuse)) -->
+	code_exprn__rem_arg_reg_dependencies([Reuse | Rvals]).
 code_exprn__rem_rval_reg_dependencies(mkword(_Tag, Rval)) -->
 	code_exprn__rem_rval_reg_dependencies(Rval).
 code_exprn__rem_rval_reg_dependencies(const(_Const)) --> [].
@@ -917,15 +917,19 @@ code_exprn__expr_is_constant(mkword(Tag, Expr0), Vars, ExprnOpts,
 		mkword(Tag, Expr)) :-
 	code_exprn__expr_is_constant(Expr0, Vars, ExprnOpts, Expr).
 
-code_exprn__expr_is_constant(create(Tag, Args0, ArgTypes, StatDyn, Label, Msg),
+code_exprn__expr_is_constant(create(Tag, Args0, ArgTypes, StatDyn,
+		Label, Msg, Reuse),
 		Vars, ExprnOpts, NewRval) :-
+	Reuse = no,
 	( StatDyn = must_be_static ->
-		NewRval = create(Tag, Args0, ArgTypes, StatDyn, Label, Msg)
+		NewRval = create(Tag, Args0, ArgTypes, StatDyn,
+			Label, Msg, Reuse)
 	;
 		ExprnOpts = nlg_asm_sgt_ubf(_, _, StaticGroundTerms, _),
 		StaticGroundTerms = yes,
 		code_exprn__args_are_constant(Args0, Vars, ExprnOpts, Args),
-		NewRval = create(Tag, Args, ArgTypes, StatDyn, Label, Msg)
+		NewRval = create(Tag, Args, ArgTypes, StatDyn,
+			Label, Msg, Reuse)
 	).
 
 code_exprn__expr_is_constant(var(Var), Vars, ExprnOpts, Rval) :-
@@ -1254,7 +1258,7 @@ code_exprn__find_real_creates([MaybeRval | MaybeRvals], SubCreates) -->
 
 code_exprn__rval_is_real_create(Rval) -->
 	(
-		{ Rval = create(_, _, _, _, _, _) },
+		{ Rval = create(_, _, _, _, _, _, _) },
 		code_exprn__get_vars(Vars0),
 		code_exprn__get_options(ExprnOpts),
 		{ \+ code_exprn__expr_is_constant(Rval, Vars0, ExprnOpts, _) }
@@ -1280,8 +1284,12 @@ code_exprn__rval_is_real_create(Rval) -->
 code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 	{ exprn_aux__simplify_rval(Rval0, Rval) },
 	(
-		{ Rval = create(Tag, Rvals, ArgTypes, _StatDyn, _Label, Msg) }
+		{ Rval = create(Tag, Rvals, ArgTypes, _StatDyn, _Label,
+			Msg, Reuse) }
 	->
+		{ require(unify(Reuse, no),
+	"code_exprn__construct_code: structure reuse not yet implemented") },
+
 		{ require(lambda([] is semidet,
 			(llds__all_args_are_word_size(ArgTypes, yes))),
 		"trying to construct heap cell with non-word-size arg(s)") },
@@ -1389,7 +1397,7 @@ code_exprn__produce_var(Var, Rval, Code) -->
 			;
 				RvalX = unop(_, _)
 			;
-				RvalX = create(_, _, _, _, _, _)
+				RvalX = create(_, _, _, _, _, _, _)
 			;
 				RvalX = mkword(_, _)
 			}
@@ -1406,59 +1414,63 @@ code_exprn__produce_var(Var, Rval, Code) -->
 
 %------------------------------------------------------------------------------%
 
-code_exprn__produce_var_in_reg(Var, Rval, Code) -->
+code_exprn__produce_var_in_reg(Var, Lval, Code) -->
 	code_exprn__get_var_status(Var, Stat),
 	(
 		{ Stat = evaled(Rvals) },
 		{ set__to_sorted_list(Rvals, RvalList) },
-		{ code_exprn__select_reg_rval(RvalList, Rval0) }
+		{ code_exprn__select_reg_rval(RvalList, Lval0) }
 	->
 		{ Code = empty },
-		{ Rval = Rval0 }
+		{ Lval = Lval0 }
 	;
 		code_exprn__select_preferred_reg(Var, Lval),
-		code_exprn__place_var(Var, Lval, Code),
-		{ Rval = lval(Lval) }
+		code_exprn__place_var(Var, Lval, Code)
 	).
 
-code_exprn__produce_var_in_reg_or_stack(Var, Rval, Code) -->
+code_exprn__produce_var_in_reg_or_stack(Var, Lval, Code) -->
 	code_exprn__get_var_status(Var, Stat),
 	(
 		{ Stat = evaled(Rvals) },
 		{ set__to_sorted_list(Rvals, RvalList) },
-		{ code_exprn__select_reg_or_stack_rval(RvalList, Rval0) }
+		{ code_exprn__select_reg_or_stack_rval(RvalList, Lval0) }
 	->
 		{ Code = empty },
-		{ Rval = Rval0 }
+		{ Lval = Lval0 }
 	;
 		code_exprn__select_preferred_lval(Var, Lval),
-		code_exprn__place_var(Var, Lval, Code),
-		{ Rval = lval(Lval) }
+		code_exprn__place_var(Var, Lval, Code)
 	).
 
 %------------------------------------------------------------------------------%
 
-:- pred code_exprn__select_reg_rval(list(rval), rval).
+:- pred code_exprn__select_reg_rval(list(rval), lval).
 :- mode code_exprn__select_reg_rval(in, out) is semidet.
 
-code_exprn__select_reg_rval([Rval0 | Rvals0], Rval) :-
-	( Rval0 = lval(reg(_, _)) ->
-		Rval = Rval0
+code_exprn__select_reg_rval([Rval0 | Rvals0], Lval) :-
+	(
+		Rval0 = lval(Lval0),
+		Lval0 = reg(_, _)
+	->
+		Lval = Lval0
 	;
-		code_exprn__select_reg_rval(Rvals0, Rval)
+		code_exprn__select_reg_rval(Rvals0, Lval)
 	).
 
-:- pred code_exprn__select_reg_or_stack_rval(list(rval), rval).
+:- pred code_exprn__select_reg_or_stack_rval(list(rval), lval).
 :- mode code_exprn__select_reg_or_stack_rval(in, out) is semidet.
 
-code_exprn__select_reg_or_stack_rval([Rval0 | Rvals0], Rval) :-
+code_exprn__select_reg_or_stack_rval([Rval0 | Rvals0], Lval) :-
 	(
-		Rval0 = lval(Lval),
-		( Lval = reg(_, _) ; Lval = stackvar(_) ; Lval = framevar(_) )
+		Rval0 = lval(Lval0),
+		( Lval0 = reg(_, _)
+		; Lval0 = stackvar(_)
+		; Lval0 = framevar(_)
+		)
 	->
-		Rval = Rval0
+		Lval = Lval0
 	;
-		code_exprn__select_reg_or_stack_rval(Rvals0, Rval)
+		code_exprn__select_reg_or_stack_rval(Rvals0, Lval)
 	).
 
 %------------------------------------------------------------------------------%
@@ -1469,8 +1481,9 @@ code_exprn__select_reg_or_stack_rval([Rval0 | Rvals0], Rval) :-
 
 code_exprn__select_preferred_lval(Var, Lval) -->
 	code_exprn__get_follow_vars(FollowVars),
+	{ FollowVars = follow_vars(FollowVarsMap, NextNonReserved) },
 	(
-		{ map__search(FollowVars, Var, PrefLval) }
+		{ map__search(FollowVarsMap, Var, PrefLval) }
 	->
 		(
 			\+ { unreal_lval(PrefLval) },
@@ -1478,10 +1491,11 @@ code_exprn__select_preferred_lval(Var, Lval) -->
 		->
 			{ Lval = PrefLval }
 		;
-			code_exprn__get_spare_reg(r, Lval)
+			code_exprn__get_spare_reg_start_from(r,
+				NextNonReserved, Lval)
 		)
 	;
-		code_exprn__get_spare_reg(r, Lval)
+		code_exprn__get_spare_reg_start_from(r, NextNonReserved, Lval)
 	).
 
 :- pred code_exprn__select_preferred_reg(prog_var, lval,
@@ -1490,8 +1504,9 @@ code_exprn__select_preferred_lval(Var, Lval) -->
 
 code_exprn__select_preferred_reg(Var, Lval) -->
 	code_exprn__get_follow_vars(FollowVars),
+	{ FollowVars = follow_vars(FollowVarsMap, NextNonReserved) },
 	(
-		{ map__search(FollowVars, Var, PrefLval) },
+		{ map__search(FollowVarsMap, Var, PrefLval) },
 		{ PrefLval = reg(_, _) }
 	->
 		(
@@ -1500,10 +1515,11 @@ code_exprn__select_preferred_reg(Var, Lval) -->
 		->
 			{ Lval = PrefLval }
 		;
-			code_exprn__get_spare_reg(r, Lval)
+			code_exprn__get_spare_reg_start_from(r,
+				NextNonReserved, Lval)
 		)
 	;
-		code_exprn__get_spare_reg(r, Lval)
+		code_exprn__get_spare_reg_start_from(r, NextNonReserved, Lval)
 	).
 
 :- pred unreal_lval(lval).
@@ -1695,24 +1711,23 @@ code_exprn__maybe_add_evaled(yes(Var), NewRval) -->
 
 	% Warning: if you get a reg, you must mark it as in use yourself.
 
+:- pred code_exprn__get_spare_reg_start_from(reg_type, int, lval,
+	exprn_info, exprn_info).
+:- mode code_exprn__get_spare_reg_start_from(in, in, out, in, out) is det.
+
+code_exprn__get_spare_reg_start_from(RegType, N0, Lval) -->
+	{ TrialLval = reg(RegType, N0) },
+	( code_exprn__lval_in_use(TrialLval) ->
+		code_exprn__get_spare_reg_start_from(RegType, N0 + 1, Lval)
+	;
+		{ Lval = TrialLval }
+	).
+
 :- pred code_exprn__get_spare_reg(reg_type, lval, exprn_info, exprn_info).
 :- mode code_exprn__get_spare_reg(in, out, in, out) is det.
 
 code_exprn__get_spare_reg(RegType, Lval) -->
-	code_exprn__get_regs(Regs),
-	{ code_exprn__get_spare_reg_2(RegType, 1, Regs, Lval) }.
-
-:- pred code_exprn__get_spare_reg_2(reg_type, int, bag(lval), lval).
-:- mode code_exprn__get_spare_reg_2(in, in, in, out) is det.
-
-code_exprn__get_spare_reg_2(RegType, N0, Regs, Lval) :-
-	TrialLval = reg(RegType, N0),
-	( bag__contains(Regs, TrialLval) ->
-		N1 is N0 + 1,
-		code_exprn__get_spare_reg_2(RegType, N1, Regs, Lval)
-	;
-		Lval = TrialLval
-	).
+	code_exprn__get_spare_reg_start_from(RegType, 1, Lval).
 
 %------------------------------------------------------------------------------%
 
