@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -30,9 +30,11 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, llds, code_info.
-:- import_module globals.
-:- import_module list, io.
+:- import_module hlds_module, hlds_pred, hlds_goal.
+:- import_module code_model.
+:- import_module llds, code_info.
+
+:- import_module list, io, counter.
 
 		% Translate a HLDS module to LLDS.
 
@@ -46,8 +48,8 @@
 		% of cell numbers.
 
 :- pred generate_proc_code(pred_info::in, proc_info::in,
-	proc_id::in, pred_id::in, module_info::in, globals::in,
-	global_data::in, global_data::out, int::in, int::out,
+	proc_id::in, pred_id::in, module_info::in,
+	global_data::in, global_data::out, counter::in, counter::out,
 	c_procedure::out) is det.
 
 		% Translate a HLDS goal to LLDS.
@@ -60,12 +62,23 @@
 
 :- implementation.
 
+% Parse tree modules
+:- import_module prog_data, prog_out, prog_util.
+
+% HLDS modules
+:- import_module hlds_out, instmap, type_util, mode_util, goal_util.
+
+% LLDS code generator modules.
 :- import_module call_gen, unify_gen, ite_gen, switch_gen, disj_gen.
 :- import_module par_conj_gen, pragma_c_gen, commit_gen.
-:- import_module continuation_info, trace, options, hlds_out.
-:- import_module code_aux, middle_rec, passes_aux, llds_out.
-:- import_module code_util, type_util, mode_util.
-:- import_module prog_data, prog_out, instmap.
+:- import_module continuation_info, trace, trace_params.
+:- import_module code_aux, code_util, middle_rec, llds_out.
+
+% Misc compiler modules
+:- import_module builtin_ops, passes_aux.
+:- import_module globals, options.
+
+% Standard library modules
 :- import_module bool, char, int, string.
 :- import_module map, assoc_list, set, term, tree, std_util, require, varset.
 
@@ -86,10 +99,26 @@ generate_code(ModuleInfo0, ModuleInfo, GlobalData0, GlobalData, Procedures) -->
 	list(pred_id)::in, list(c_procedure)::out,
 	io__state::di, io__state::uo) is det.
 
+:- pred generate_maybe_pred_code(module_info::in, module_info::out,
+	global_data::in, global_data::out, pred_id::in, list(c_procedure)::out,
+	io__state::di, io__state::uo) is det.
+
 generate_pred_list_code(ModuleInfo, ModuleInfo, GlobalData, GlobalData,
 		[], []) --> [].
 generate_pred_list_code(ModuleInfo0, ModuleInfo, GlobalData0, GlobalData,
 		[PredId | PredIds], Predicates) -->
+	generate_maybe_pred_code(ModuleInfo0, ModuleInfo1,
+		GlobalData0, GlobalData1, PredId, Predicates0),
+	generate_pred_list_code(ModuleInfo1, ModuleInfo,
+		GlobalData1, GlobalData, PredIds, Predicates1),
+	{ list__append(Predicates0, Predicates1, Predicates) }.
+
+	% Note that some of the logic of generate_maybe_pred_code is duplicated
+	% by mercury_compile__backend_pass_by_preds, so modifications here may
+	% also need to be repeated there.
+
+generate_maybe_pred_code(ModuleInfo0, ModuleInfo, GlobalData0, GlobalData,
+		PredId, Predicates) -->
 	{ module_info_preds(ModuleInfo0, PredInfos) },
 		% get the pred_info structure for this predicate
 	{ map__lookup(PredInfos, PredId, PredInfo) },
@@ -101,64 +130,86 @@ generate_pred_list_code(ModuleInfo0, ModuleInfo, GlobalData0, GlobalData,
 		; hlds_pred__pred_info_is_aditi_relation(PredInfo)
 		}
 	->
-		{ Predicates0 = [] },
-		{ ModuleInfo1 = ModuleInfo0 },
-		{ GlobalData1 = GlobalData0 }
+		{ Predicates = [] },
+		{ ModuleInfo = ModuleInfo0 },
+		{ GlobalData = GlobalData0 }
 	;
-		generate_pred_code(ModuleInfo0, ModuleInfo1,
-				GlobalData0, GlobalData1,
-				PredId, PredInfo, ProcIds, Predicates0)
-	),
-	{ list__append(Predicates0, Predicates1, Predicates) },
-		% and generate the code for the rest of the predicates
-	generate_pred_list_code(ModuleInfo1, ModuleInfo,
-		GlobalData1, GlobalData, PredIds, Predicates1).
+		{ module_info_globals(ModuleInfo0, Globals0) },
+		{ globals__lookup_bool_option(Globals0, very_verbose,
+			VeryVerbose) },
+		( { VeryVerbose = yes } ->
+			io__write_string("% Generating code for "),
+			hlds_out__write_pred_id(ModuleInfo0, PredId),
+			io__write_string("\n"),
+			{ globals__lookup_bool_option(Globals0,
+				statistics, Statistics) },
+			maybe_report_stats(Statistics)
+		;
+			[]
+		),
+		{
+			pred_info_module(PredInfo, PredModule),
+			pred_info_name(PredInfo, PredName),
+			pred_info_arity(PredInfo, PredArity),
+			no_type_info_builtin(PredModule, PredName, PredArity)
+		->
+				% These predicates should never be traced,
+				% since they do not obey typeinfo_liveness.
+				% Since they may be opt_imported into other
+				% modules, we must switch off the tracing
+				% of such preds on a pred-by-pred basis.
+			globals__get_trace_level(Globals0, TraceLevel),
+			globals__set_trace_level_none(Globals0, Globals1),
+			module_info_set_globals(ModuleInfo0, Globals1,
+				ModuleInfo1),
+			generate_pred_code(ModuleInfo1, ModuleInfo2,
+				GlobalData0, GlobalData,
+				PredId, PredInfo, ProcIds, Predicates),
+			module_info_globals(ModuleInfo2, Globals2),
+			globals__set_trace_level(Globals2, TraceLevel,
+				Globals),
+			module_info_set_globals(ModuleInfo2, Globals,
+				ModuleInfo)
+		;
+			generate_pred_code(ModuleInfo0, ModuleInfo,
+				GlobalData0, GlobalData,
+				PredId, PredInfo, ProcIds, Predicates)
+		}
+	).
 
 	% Translate a HLDS predicate to LLDS.
 
 :- pred generate_pred_code(module_info::in, module_info::out,
-	global_data::in, global_data::out,
-	pred_id::in, pred_info::in, list(proc_id)::in, list(c_procedure)::out,
-	io__state::di, io__state::uo) is det.
+	global_data::in, global_data::out, pred_id::in, pred_info::in,
+	list(proc_id)::in, list(c_procedure)::out) is det.
 
 generate_pred_code(ModuleInfo0, ModuleInfo, GlobalData0, GlobalData,
-		PredId, PredInfo, ProcIds, Code) -->
-	globals__io_lookup_bool_option(very_verbose, VeryVerbose),
-	( { VeryVerbose = yes } ->
-		io__write_string("% Generating code for "),
-		hlds_out__write_pred_id(ModuleInfo0, PredId),
-		io__write_string("\n"),
-		globals__io_lookup_bool_option(statistics, Statistics),
-		maybe_report_stats(Statistics)
-	;
-		[]
-	),
-	{ module_info_get_cell_count(ModuleInfo0, CellCount0) },
-	globals__io_get_globals(Globals),
-	{ generate_proc_list_code(ProcIds, PredId, PredInfo, ModuleInfo0,
-		Globals, GlobalData0, GlobalData, CellCount0, CellCount,
-		[], Code) },
-	{ module_info_set_cell_count(ModuleInfo0, CellCount, ModuleInfo) }.
+		PredId, PredInfo, ProcIds, Code) :-
+	module_info_get_cell_counter(ModuleInfo0, CellCounter0),
+	generate_proc_list_code(ProcIds, PredId, PredInfo, ModuleInfo0,
+		GlobalData0, GlobalData, CellCounter0, CellCounter,
+		[], Code),
+	module_info_set_cell_counter(ModuleInfo0, CellCounter, ModuleInfo).
 
 	% Translate all the procedures of a HLDS predicate to LLDS.
 
 :- pred generate_proc_list_code(list(proc_id)::in, pred_id::in, pred_info::in,
-	module_info::in, globals::in, global_data::in, global_data::out,
-	int::in, int::out, list(c_procedure)::in, list(c_procedure)::out)
-	is det.
+	module_info::in, global_data::in, global_data::out,
+	counter::in, counter::out,
+	list(c_procedure)::in, list(c_procedure)::out) is det.
 
-generate_proc_list_code([], _PredId, _PredInfo, _ModuleInfo, _Globals,
-		GlobalData, GlobalData, CellCount, CellCount, Procs, Procs).
+generate_proc_list_code([], _PredId, _PredInfo, _ModuleInfo,
+		GlobalData, GlobalData, CellCounter, CellCounter,
+		Procs, Procs).
 generate_proc_list_code([ProcId | ProcIds], PredId, PredInfo, ModuleInfo0,
-		Globals, GlobalData0, GlobalData, CellCount0, CellCount,
+		GlobalData0, GlobalData, CellCounter0, CellCounter,
 		Procs0, Procs) :-
 	pred_info_procedures(PredInfo, ProcInfos),
 	map__lookup(ProcInfos, ProcId, ProcInfo),
 	generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo0,
-		Globals, GlobalData0, GlobalData1, CellCount0, CellCount1,
-		Proc),
+		GlobalData0, GlobalData1, CellCounter0, CellCounter1, Proc),
 	generate_proc_list_code(ProcIds, PredId, PredInfo, ModuleInfo0,
-		Globals, GlobalData1, GlobalData, CellCount1, CellCount,
+		GlobalData1, GlobalData, CellCounter1, CellCounter,
 		[Proc | Procs0], Procs).
 
 %---------------------------------------------------------------------------%
@@ -180,8 +231,8 @@ generate_proc_list_code([ProcId | ProcIds], PredId, PredInfo, ModuleInfo0,
 
 %---------------------------------------------------------------------------%
 
-generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo, Globals,
-		GlobalData0, GlobalData, CellCount0, CellCount, Proc) :-
+generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo,
+		GlobalData0, GlobalData, CellCounter0, CellCounter, Proc) :-
 	proc_info_interface_determinism(ProcInfo, Detism),
 	proc_info_interface_code_model(ProcInfo, CodeModel),
 	proc_info_goal(ProcInfo, Goal),
@@ -191,10 +242,12 @@ generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo, Globals,
 		MaybeFollowVars = yes(FollowVars)
 	;
 		MaybeFollowVars = no,
-		map__init(FollowVars)
+		map__init(FollowVarsMap),
+		FollowVars = follow_vars(FollowVarsMap, 1)
 	),
+	module_info_globals(ModuleInfo, Globals),
 	continuation_info__basic_stack_layout_for_proc(PredInfo, Globals,
-	BasicStackLayout, ForceProcId),
+		BasicStackLayout, ForceProcId),
 	( BasicStackLayout = yes ->
 		SaveSuccip = yes
 	;
@@ -209,7 +262,7 @@ generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo, Globals,
 		% needed for model_non procedures only if we are doing
 		% execution tracing.
 	code_info__init(SaveSuccip, Globals, PredId, ProcId, ProcInfo,
-		FollowVars, ModuleInfo, CellCount0, OutsideResumePoint,
+		FollowVars, ModuleInfo, CellCounter0, OutsideResumePoint,
 		TraceSlotInfo, CodeInfo0),
 
 		% Generate code for the procedure.
@@ -217,7 +270,29 @@ generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo, Globals,
 		TraceSlotInfo, CodeTree, MaybeTraceCallLabel, FrameInfo,
 		CodeInfo0, CodeInfo),
 	code_info__get_max_reg_in_use_at_trace(MaxTraceReg, CodeInfo, _),
-	code_info__get_cell_count(CellCount, CodeInfo, _),
+	code_info__get_cell_counter(CellCounter, CodeInfo, _),
+
+	globals__get_trace_level(Globals, TraceLevel),
+	code_info__get_created_temp_frame(CreatedTempFrame, CodeInfo, _),
+
+	(
+		trace_level_is_none(TraceLevel) = no,
+		CreatedTempFrame = yes,
+		CodeModel \= model_non
+	->
+			% If tracing is enabled, the procedure lives on
+			% the det stack and the code created any temporary
+			% nondet stack frames, then we must have reserved a
+			% stack slot for storing the value of maxfr; if we
+			% didn't, a retry command in the debugger from a point
+			% in the middle of this procedure will do the wrong
+			% thing.
+		proc_info_get_need_maxfr_slot(ProcInfo, HaveMaxfrSlot),
+		require(unify(HaveMaxfrSlot, yes),
+			"should have reserved a slot for maxfr, but didn't")
+	;
+		true
+	),
 
 		% Turn the code tree into a list.
 	tree__flatten(CodeTree, FragmentList),
@@ -243,38 +318,70 @@ generate_proc_code(PredInfo, ProcInfo, ProcId, PredId, ModuleInfo, Globals,
 		code_info__get_layout_info(InternalMap, CodeInfo, _),
 		code_util__make_local_entry_label(ModuleInfo, PredId, ProcId,
 			no, EntryLabel),
+		proc_info_eval_method(ProcInfo, EvalMethod),
+		proc_info_get_initial_instmap(ProcInfo, ModuleInfo, InstMap0),
+		proc_info_varset(ProcInfo, VarSet),
+		proc_info_vartypes(ProcInfo, VarTypes),
 		ProcLayout = proc_layout_info(EntryLabel, Detism, TotalSlots,
-			MaybeSuccipSlot, MaybeTraceCallLabel, MaxTraceReg,
-			TraceSlotInfo, ForceProcId, InternalMap),
+			MaybeSuccipSlot, EvalMethod, MaybeTraceCallLabel,
+			MaxTraceReg, Goal, InstMap0, TraceSlotInfo,
+			ForceProcId, VarSet, VarTypes, InternalMap),
 		global_data_add_new_proc_layout(GlobalData0,
 			proc(PredId, ProcId), ProcLayout, GlobalData1)
 	;
 		GlobalData1 = GlobalData0
 	),
 
-	code_info__get_non_common_static_data(NonCommonStatics, CodeInfo, _),
-	global_data_add_new_non_common_static_datas(GlobalData1,
-		NonCommonStatics, GlobalData2),
+	code_info__get_closure_layouts(ClosureLayouts, CodeInfo, _),
+	global_data_add_new_closure_layouts(GlobalData1, ClosureLayouts,
+		GlobalData2),
+	code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel),
 	maybe_add_tabling_pointer_var(ModuleInfo, PredId, ProcId, ProcInfo,
-		GlobalData2, GlobalData),
+		ProcLabel, GlobalData2, GlobalData),
 
 	pred_info_name(PredInfo, Name),
 	pred_info_arity(PredInfo, Arity),
-		% Construct a c_procedure structure with all the information.
-	Proc = c_procedure(Name, Arity, proc(PredId, ProcId), Instructions).
 
-:- pred maybe_add_tabling_pointer_var(module_info, pred_id, proc_id, proc_info,
-		global_data, global_data).
-:- mode maybe_add_tabling_pointer_var(in, in, in, in, in, out) is det.
+	( goal_contains_reconstruction(Goal) ->
+		ContainsReconstruction = contains_reconstruction
+	;
+		ContainsReconstruction = does_not_contain_reconstruction
+	),
 
-maybe_add_tabling_pointer_var(ModuleInfo, PredId, ProcId, ProcInfo,
+	code_info__get_label_counter(LabelCounter, CodeInfo, _),
+
+	globals__lookup_bool_option(Globals, generate_bytecode, GenBytecode),
+	(
+		% XXX: There is a mass of calls above that the bytecode
+		% doesn't need; work out which is and isn't needed and put
+		% inside the else case below
+		GenBytecode = yes,
+		% We don't generate bytecode for __Unify__, __Compare__ etc
+		% Since we will assume this code is already correct
+		\+ code_util__compiler_generated(PredInfo),
+		% Don't generate bytecode for procs with foreign code
+		goal_has_foreign(Goal) = no
+	->
+		EmptyLabelCounter = counter__init(0),
+		code_gen__bytecode_stub(ModuleInfo, PredId, ProcId,
+			BytecodeInstructions),
+		Proc = c_procedure(Name, Arity, proc(PredId, ProcId),
+			BytecodeInstructions, ProcLabel, EmptyLabelCounter,
+			ContainsReconstruction)
+	;	
+		Proc = c_procedure(Name, Arity, proc(PredId, ProcId),
+			Instructions, ProcLabel, LabelCounter,
+			ContainsReconstruction)
+	).
+
+:- pred maybe_add_tabling_pointer_var(module_info::in,
+	pred_id::in, proc_id::in, proc_info::in, proc_label::in,
+	global_data::in, global_data::out) is det.
+
+maybe_add_tabling_pointer_var(ModuleInfo, PredId, ProcId, ProcInfo, ProcLabel,
 		GlobalData0, GlobalData) :-
 	proc_info_eval_method(ProcInfo, EvalMethod),
-	(
-		EvalMethod \= eval_normal
-	->
-		code_util__make_proc_label(ModuleInfo, PredId, ProcId,
-			ProcLabel),
+	( eval_method_has_per_proc_tabling_pointer(EvalMethod) = yes ->
 		module_info_name(ModuleInfo, ModuleName),
 		Var = tabling_pointer_var(ModuleName, ProcLabel),
 		global_data_add_new_proc_var(GlobalData0,
@@ -351,8 +458,15 @@ generate_category_code(model_det, Goal, ResumePoint, TraceSlotInfo, Code,
 		code_info__get_maybe_trace_info(MaybeTraceInfo),
 		( { MaybeTraceInfo = yes(TraceInfo) } ->
 			trace__generate_external_event_code(call, TraceInfo,
-				BodyContext, TraceCallLabel, _TypeInfos,
-				TraceCallCode),
+				BodyContext, MaybeCallExternalInfo),
+			{
+				MaybeCallExternalInfo = yes(CallExternalInfo),
+				CallExternalInfo = external_event_info(
+					TraceCallLabel, _, TraceCallCode)
+			;
+				MaybeCallExternalInfo = no,
+				error("generate_category_code: call events suppressed")
+			},
 			{ MaybeTraceCallLabel = yes(TraceCallLabel) }
 		;
 			{ TraceCallCode = empty },
@@ -384,8 +498,15 @@ generate_category_code(model_semi, Goal, ResumePoint, TraceSlotInfo, Code,
 	code_info__get_maybe_trace_info(MaybeTraceInfo),
 	( { MaybeTraceInfo = yes(TraceInfo) } ->
 		trace__generate_external_event_code(call, TraceInfo,
-			BodyContext, TraceCallLabel, _TypeInfos,
-			TraceCallCode),
+			BodyContext, MaybeCallExternalInfo),
+		{
+			MaybeCallExternalInfo = yes(CallExternalInfo),
+			CallExternalInfo = external_event_info(
+				TraceCallLabel, _, TraceCallCode)
+		;
+			MaybeCallExternalInfo = no,
+			error("generate_category_code: call events suppressed")
+		},
 		{ MaybeTraceCallLabel = yes(TraceCallLabel) },
 		code_gen__generate_goal(model_semi, Goal, BodyCode),
 		code_gen__generate_entry(model_semi, Goal, ResumePoint,
@@ -400,7 +521,15 @@ generate_category_code(model_semi, Goal, ResumePoint, TraceSlotInfo, Code,
 			% XXX A context that gives the end of the procedure
 			% definition would be better than BodyContext.
 		trace__generate_external_event_code(fail, TraceInfo,
-			BodyContext, _, _, TraceFailCode),
+			BodyContext, MaybeFailExternalInfo),
+		{
+			MaybeFailExternalInfo = yes(FailExternalInfo),
+			FailExternalInfo = external_event_info(
+				_, _, TraceFailCode)
+		;
+			MaybeFailExternalInfo = no,
+			TraceFailCode = empty
+		},
 		{ Code =
 			tree(EntryCode,
 			tree(TraceCallCode,
@@ -436,8 +565,15 @@ generate_category_code(model_non, Goal, ResumePoint, TraceSlotInfo, Code,
 	{ goal_info_get_context(GoalInfo, BodyContext) },
 	( { MaybeTraceInfo = yes(TraceInfo) } ->
 		trace__generate_external_event_code(call, TraceInfo,
-			BodyContext, TraceCallLabel, _TypeInfos,
-			TraceCallCode),
+			BodyContext, MaybeCallExternalInfo),
+		{
+			MaybeCallExternalInfo = yes(CallExternalInfo),
+			CallExternalInfo = external_event_info(
+				TraceCallLabel, _, TraceCallCode)
+		;
+			MaybeCallExternalInfo = no,
+			error("generate_category_code: call events suppressed")
+		},
 		{ MaybeTraceCallLabel = yes(TraceCallLabel) },
 		code_gen__generate_goal(model_non, Goal, BodyCode),
 		code_gen__generate_entry(model_non, Goal, ResumePoint,
@@ -452,14 +588,43 @@ generate_category_code(model_non, Goal, ResumePoint, TraceSlotInfo, Code,
 			% XXX A context that gives the end of the procedure
 			% definition would be better than BodyContext.
 		trace__generate_external_event_code(fail, TraceInfo,
-			BodyContext, _, _, TraceFailCode),
-		{ TraceSlotInfo = trace_slot_info(_, _, yes(_)) ->
-			DiscardTraceTicketCode = node([
-				discard_ticket - "discard retry ticket"
-			])
+			BodyContext, MaybeFailExternalInfo),
+		{
+			MaybeFailExternalInfo = yes(FailExternalInfo),
+			FailExternalInfo = external_event_info(
+				_, _, TraceFailCode)
 		;
-			DiscardTraceTicketCode = empty
+			MaybeFailExternalInfo = no,
+			TraceFailCode = empty
 		},
+		( { TraceSlotInfo ^ slot_trail = yes(_) } ->
+			( { TraceSlotInfo ^ slot_from_full =
+				yes(FromFullSlot) }
+			->
+				%
+				% Generate code which discards the ticket
+				% only if it was allocated, i.e. only if
+				% MR_trace_from_full was true on entry.
+				%
+				{ FromFullSlotLval =
+					llds__stack_slot_num_to_lval(
+						model_non, FromFullSlot) },
+				code_info__get_next_label(SkipLabel),
+				{ DiscardTraceTicketCode = node([
+					if_val(unop(not,
+						lval(FromFullSlotLval)),
+						label(SkipLabel)) - "",
+					discard_ticket - "discard retry ticket",
+					label(SkipLabel) - ""
+				]) }
+			;
+				{ DiscardTraceTicketCode = node([
+					discard_ticket - "discard retry ticket"
+				]) }
+			)
+		;
+			{ DiscardTraceTicketCode = empty }
+		),
 		{ FailCode = node([
 			goto(do_fail) - "fail after fail trace port"
 		]) },
@@ -569,7 +734,8 @@ code_gen__generate_entry(CodeModel, Goal, OutsideResumePoint, FrameInfo,
 		{ code_info__resume_point_stack_addr(OutsideResumePoint,
 			OutsideResumeAddress) },
 		(
-			{ Goal = pragma_c_code(_,_,_,_,_,_, PragmaCode) - _},
+			{ Goal = pragma_foreign_code(_, _, _, _, _, _,
+				PragmaCode) - _},
 			{ PragmaCode = nondet(Fields, FieldsContext,
 				_,_,_,_,_,_,_) }
 		->
@@ -586,7 +752,7 @@ code_gen__generate_entry(CodeModel, Goal, OutsideResumePoint, FrameInfo,
 				mkframe(NondetFrameInfo, OutsideResumeAddress)
 					- "Allocate stack frame",
 				pragma_c([], DefineComponents,
-					will_not_call_mercury, no, no, no)
+					will_not_call_mercury, no, no, no, no)
 					- ""
 			]) },
 			{ NondetPragma = yes }
@@ -673,7 +839,7 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 		{ UndefComponents = [pragma_c_raw_code(UndefStr)] },
 		{ UndefCode = node([
 			pragma_c([], UndefComponents,
-				will_not_call_mercury, no, no, no)
+				will_not_call_mercury, no, no, no, no)
 				- ""
 		]) },
 		{ RestoreDeallocCode = empty },	% always empty for nondet code
@@ -691,9 +857,10 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 		(
 			{ instmap__is_unreachable(Instmap) }
 		->
+			{ OutLvals = set__init },
 			{ FlushCode = empty }
 		;
-			code_info__setup_call(Args, callee, FlushCode)
+			code_info__setup_return(Args, OutLvals, FlushCode)
 		),
 		{
 			MaybeSuccipSlot = yes(SuccipSlot)
@@ -714,21 +881,65 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 				decr_sp(TotalSlots) - "Deallocate stack frame"
 			])
 		},
-		{
-			TraceSlotInfo = trace_slot_info(_, _, yes(_)),
-			CodeModel \= model_non
+		(
+			{ TraceSlotInfo ^ slot_trail = yes(_) },
+			{ CodeModel \= model_non }
 		->
-			DiscardTraceTicketCode = node([
-				discard_ticket - "discard retry ticket"
-			])
+			(
+				{ TraceSlotInfo ^ slot_from_full =
+					yes(FromFullSlot) }
+			->
+				%
+				% Generate code which prunes the ticket
+				% only if it was allocated, i.e. only if
+				% MR_trace_from_full was true on entry.
+				%
+				% Note that to avoid duplicating label names,
+				% we need to generate two different copies
+				% of this with different labels; this is
+				% needed for semidet code, which will get one
+				% copy in the success epilogue and one copy
+				% in the failure epilogue
+				%
+				{ FromFullSlotLval =
+					llds__stack_slot_num_to_lval(
+						CodeModel, FromFullSlot) },
+				code_info__get_next_label(SkipLabel),
+				code_info__get_next_label(SkipLabelCopy),
+				{ PruneTraceTicketCode = node([
+					if_val(unop(not,
+						lval(FromFullSlotLval)),
+						label(SkipLabel)) - "",
+					prune_ticket - "prune retry ticket",
+					label(SkipLabel) - ""
+				]) },
+				{ PruneTraceTicketCodeCopy = node([
+					if_val(unop(not,
+						lval(FromFullSlotLval)),
+						label(SkipLabelCopy)) - "",
+					prune_ticket - "prune retry ticket",
+					label(SkipLabelCopy) - ""
+				]) }
+			;
+				{ PruneTraceTicketCode = node([
+					prune_ticket - "prune retry ticket"
+				]) },
+				{ PruneTraceTicketCodeCopy = PruneTraceTicketCode }
+			)
 		;
-			DiscardTraceTicketCode = empty
-		},
+			{ PruneTraceTicketCode = empty },
+			{ PruneTraceTicketCodeCopy = empty }
+		),
 
 		{ RestoreDeallocCode =
 			tree(RestoreSuccipCode,
-			tree(DeallocCode,
-			     DiscardTraceTicketCode))
+			tree(PruneTraceTicketCode,
+			     DeallocCode))
+		},
+		{ RestoreDeallocCodeCopy =
+			tree(RestoreSuccipCode,
+			tree(PruneTraceTicketCodeCopy,
+			     DeallocCode))
 		},
 
 		code_info__get_maybe_trace_info(MaybeTraceInfo),
@@ -737,7 +948,16 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 				% procedure definition would be better than
 				% CallContext.
 			trace__generate_external_event_code(exit, TraceInfo,
-				BodyContext, _, TypeInfoDatas, TraceExitCode),
+				BodyContext, MaybeExitExternalInfo),
+			{
+				MaybeExitExternalInfo = yes(ExitExternalInfo),
+				ExitExternalInfo = external_event_info(
+					_, TypeInfoDatas, TraceExitCode)
+			;
+				MaybeExitExternalInfo = no,
+				TypeInfoDatas = map__init,
+				TraceExitCode = empty
+			},
 			{ map__values(TypeInfoDatas, TypeInfoLocnSets) },
 			{ FindBaseLvals = lambda([Lval::out] is nondet, (
 				list__member(LocnSet, TypeInfoLocnSets),
@@ -748,34 +968,28 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 					Locn = indirect(Lval, _)
 				)
 			)) },
-			{ solutions(FindBaseLvals, TypeInfoLvals) }
+			{ solutions(FindBaseLvals, TypeInfoLvals) },
+			{ set__insert_list(OutLvals, TypeInfoLvals,
+				LiveLvals) }
 		;
 			{ TraceExitCode = empty },
-			{ TypeInfoLvals = [] }
+			{ LiveLvals = OutLvals }
 		),
 
-			% Find out which locations should be mentioned
-			% in the success path livevals(...) annotation,
-			% so that value numbering doesn't optimize them away.
-		{ code_gen__select_args_with_mode(Args, top_out, _OutVars,
-			OutLvals) },
-		{ list__append(TypeInfoLvals, OutLvals, LiveArgLvals) },
-		{ set__list_to_set(LiveArgLvals, LiveArgs) },
-
 		(
 			{ CodeModel = model_det },
 			{ SuccessCode = node([
-				livevals(LiveArgs) - "",
+				livevals(LiveLvals) - "",
 				goto(succip) - "Return from procedure call"
 			]) },
 			{ AllSuccessCode =
 				tree(TraceExitCode,
-				tree(RestoreDeallocCode,
+				tree(RestoreDeallocCodeCopy,
 				     SuccessCode))
 			}
 		;
 			{ CodeModel = model_semi },
-			{ set__insert(LiveArgs, reg(r, 1), SuccessLiveRegs) },
+			{ set__insert(LiveLvals, reg(r, 1), SuccessLiveRegs) },
 			{ SuccessCode = node([
 				assign(reg(r, 1), const(true)) - "Succeed",
 				livevals(SuccessLiveRegs) - "",
@@ -783,7 +997,7 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 			]) },
 			{ AllSuccessCode =
 				tree(TraceExitCode,
-				tree(RestoreDeallocCode,
+				tree(RestoreDeallocCodeCopy,
 				     SuccessCode))
 			}
 		;
@@ -795,7 +1009,7 @@ code_gen__generate_exit(CodeModel, FrameInfo, TraceSlotInfo, BodyContext,
 				SetupRedoCode = empty
 			},
 			{ SuccessCode = node([
-				livevals(LiveArgs) - "",
+				livevals(LiveLvals) - "",
 				goto(do_succeed(no))
 					- "Return from procedure call"
 			]) },
@@ -853,7 +1067,32 @@ code_gen__generate_goal(ContextModel, Goal - GoalInfo, Code) -->
 			)
 		},
 
-		code_gen__generate_goal_2(Goal, GoalInfo, CodeModel, Code),
+		code_gen__generate_goal_2(Goal, GoalInfo, CodeModel, GoalCode),
+
+			% If the predicate's evaluation method is memo,
+			% loopcheck or minimal model, the goal generated
+			% the variable that represents the call table tip,
+			% *and* tracing is enabled, then we save this variable
+			% to its stack slot. This is necessary to enable
+			% retries across this procedure to reset the call table
+			% entry to uninitialized, effectively removing the
+			% call table entry.
+		(
+			{ goal_info_get_features(GoalInfo, Features) },
+			{ set__member(call_table_gen, Features) },
+			code_info__get_proc_info(ProcInfo),
+			{ proc_info_get_call_table_tip(ProcInfo,
+				MaybeCallTableVar) },
+				% MaybeCallTableVar will be `no' unless
+				% tracing is enabled.
+			{ MaybeCallTableVar = yes(CallTableVar) }
+		->
+			code_info__save_variables_on_stack([CallTableVar],
+				SaveCode),
+			{ Code = tree(GoalCode, SaveCode) }
+		;
+			{ Code = GoalCode }
+		),
 
 			% Make live any variables which subsequent goals
 			% will expect to be live, but were not generated
@@ -861,16 +1100,16 @@ code_gen__generate_goal(ContextModel, Goal - GoalInfo, Code) -->
 		code_info__post_goal_update(GoalInfo)
 	;
 		{ Code = empty }
-	),
-	!.
+	).
 
 %---------------------------------------------------------------------------%
 
 :- pred code_gen__generate_goal_2(hlds_goal_expr::in, hlds_goal_info::in,
 	code_model::in, code_tree::out, code_info::in, code_info::out) is det.
 
-code_gen__generate_goal_2(unify(_, _, _, Uni, _), _, CodeModel, Code) -->
-	unify_gen__generate_unification(CodeModel, Uni, Code).
+code_gen__generate_goal_2(unify(_, _, _, Uni, _), GoalInfo, CodeModel, Code)
+		-->
+	unify_gen__generate_unification(CodeModel, Uni, GoalInfo, Code).
 code_gen__generate_goal_2(conj(Goals), _GoalInfo, CodeModel, Code) -->
 	code_gen__generate_goals(Goals, CodeModel, Code).
 code_gen__generate_goal_2(par_conj(Goals, _SM), GoalInfo, CodeModel, Code) -->
@@ -904,12 +1143,18 @@ code_gen__generate_goal_2(call(PredId, ProcId, Args, BuiltinState, _, _),
 		call_gen__generate_builtin(CodeModel, PredId, ProcId, Args,
 			Code)
 	).
-code_gen__generate_goal_2(pragma_c_code(Attributes,
+code_gen__generate_goal_2(pragma_foreign_code(Attributes,
 		PredId, ModeId, Args, ArgNames, OrigArgTypes, PragmaCode),
 		GoalInfo, CodeModel, Instr) -->
-	pragma_c_gen__generate_pragma_c_code(CodeModel, Attributes,
-		PredId, ModeId, Args, ArgNames, OrigArgTypes, GoalInfo,
-		PragmaCode, Instr).
+	( 
+		{ foreign_language(Attributes, c) } 
+	->
+		pragma_c_gen__generate_pragma_c_code(CodeModel, Attributes,
+			PredId, ModeId, Args, ArgNames, OrigArgTypes,
+			GoalInfo, PragmaCode, Instr)
+	;
+		{ error("code_gen__generate_goal_2: foreign code other than C unexpected") }
+	).
 code_gen__generate_goal_2(bi_implication(_, _), _, _, _) -->
 	% these should have been expanded out by now
 	{ error("code_gen__generate_goal_2: unexpected bi_implication") }.
@@ -977,15 +1222,75 @@ code_gen__add_saved_succip([Instrn0 - Comment | Instrns0 ], StackLoc,
 		set__insert(LiveVals0, stackvar(StackLoc), LiveVals1),
 		Instrn = livevals(LiveVals1)
         ;
-		Instrn0 = call(Target, ReturnLabel, LiveVals0, Context, CM)
+		Instrn0 = call(Target, ReturnLabel, LiveVals0, Context, GP, CM)
 	->
 		map__init(Empty),
 		LiveVals  = [live_lvalue(direct(stackvar(StackLoc)),
 				succip, Empty) | LiveVals0],
-		Instrn = call(Target, ReturnLabel, LiveVals, Context, CM)
+		Instrn = call(Target, ReturnLabel, LiveVals, Context, GP, CM)
 	;
 		Instrn = Instrn0
 	),
 	code_gen__add_saved_succip(Instrns0, StackLoc, Instrns).
 
 %---------------------------------------------------------------------------%
+
+:- pred code_gen__bytecode_stub(module_info::in, pred_id::in, proc_id::in,
+	list(instruction)::out) is det.
+
+code_gen__bytecode_stub(ModuleInfo, PredId, ProcId, BytecodeInstructions) :-
+
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	pred_info_module(PredInfo, ModuleSymName),
+
+	prog_out__sym_name_to_string(ModuleSymName, "__", ModuleName),
+	
+	code_util__make_local_entry_label(ModuleInfo, PredId,
+		ProcId, no, Entry),
+
+	pred_info_name(PredInfo, PredName),
+	proc_id_to_int(ProcId, ProcNum),
+	string__int_to_string(ProcNum, ProcStr),
+	pred_info_arity(PredInfo, Arity),
+	int_to_string(Arity, ArityStr),
+	pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
+
+	CallStructName = "bytecode_call_info",
+
+	append_list([
+		"\t\tstatic MB_Call ", CallStructName, " = {\n",
+		"\t\t\t(MB_Word)NULL,\n",
+		"\t\t\t""", ModuleName, """,\n",
+		"\t\t\t""", PredName, """,\n",
+		"\t\t\t", ProcStr, ",\n",
+		"\t\t\t", ArityStr, ",\n",
+		"\t\t\t", (PredOrFunc = function -> "TRUE" ; "FALSE"), "\n",
+		"\t\t};\n"
+		], CallStruct),
+
+	append_list([
+		"\t\tMB_Native_Addr return_addr;\n",
+		"\t\tMR_save_registers();\n",
+		"\t\treturn_addr = MB_bytecode_call_entry(",
+			"&",CallStructName,");\n",
+		"\t\tMR_restore_registers();\n",
+		"\t\tMR_GOTO(return_addr);\n"
+		], BytecodeCall),
+
+		
+	BytecodeInstructions = [
+		label(Entry) - "Procedure entry point",
+
+		pragma_c(
+		[],
+		[
+		pragma_c_raw_code("\t{\n"),
+		pragma_c_raw_code(CallStruct),
+		pragma_c_raw_code(BytecodeCall),
+		pragma_c_raw_code("\t}\n")
+		],
+		may_call_mercury, no, no, no, no
+		) - "Entry stub"
+	].
+
+%---------------------------------------------------------------------------%
