@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,12 +17,14 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, hlds_data, prog_data, llds.
-:- import_module list, assoc_list, set, std_util.
+:- import_module prog_data, hlds_module, hlds_pred, hlds_goal, hlds_data.
+:- import_module rtti, llds.
+
+:- import_module bool, list, std_util.
 
 	% Create a code address which holds the address of the specified
 	% procedure.
-	% The fourth argument should be `no' if the the caller wants the
+	% The `immed' argument should be `no' if the the caller wants the
 	% returned address to be valid from everywhere in the program.
 	% If being valid from within the current procedure is enough,
 	% this argument should be `yes' wrapped around the value of the
@@ -30,10 +32,16 @@
 	% Using an address that is only valid from within the current
 	% procedure may make jumps more efficient.
 
+:- type immed == maybe(pair(int, pred_proc_id)).
+
 :- pred code_util__make_entry_label(module_info, pred_id, proc_id, 
-	maybe(pair(int, pred_proc_id)), code_addr).
+		immed, code_addr).
 :- mode code_util__make_entry_label(in, in, in, in, out) is det.
 
+:- pred code_util__make_entry_label_from_rtti(rtti_proc_label, immed,
+		code_addr).
+:- mode code_util__make_entry_label_from_rtti(in, in, out) is det.
+
 	% Create a label which holds the address of the specified procedure,
 	% which must be defined in the current module (procedures that are
 	% imported from other modules have representations only as code_addrs,
@@ -43,17 +51,29 @@
 	% code_util__make_entry_label.
 
 :- pred code_util__make_local_entry_label(module_info, pred_id, proc_id,
-	maybe(pair(int, pred_proc_id)), label).
+		immed, label).
 :- mode code_util__make_local_entry_label(in, in, in, in, out) is det.
 
 	% Create a label internal to a Mercury procedure.
 :- pred code_util__make_internal_label(module_info, pred_id, proc_id, int,
-	label).
+		label).
 :- mode code_util__make_internal_label(in, in, in, in, out) is det.
 
 :- pred code_util__make_proc_label(module_info, pred_id, proc_id, proc_label).
 :- mode code_util__make_proc_label(in, in, in, out) is det.
 
+	% code_util__make_user_proc_label(ModuleName, PredIsImported,
+	%	PredOrFunc, ModuleName, PredName, Arity, ProcId, Label):
+	% Make a proc_label for a user-defined procedure.
+	%
+	% The PredIsImported argument should be the result of
+	% calling pred_info_is_imported.
+
+:- pred code_util__make_user_proc_label(module_name, bool,
+	pred_or_func, module_name, string, arity, proc_id, proc_label).
+:- mode code_util__make_user_proc_label(in, in,
+	in, in, in, in, in, out) is det.
+
 :- pred code_util__make_uni_label(module_info, type_id, proc_id, proc_label).
 :- mode code_util__make_uni_label(in, in, in, out) is det.
 
@@ -66,6 +86,9 @@
 :- pred code_util__arg_loc_to_register(arg_loc, lval).
 :- mode code_util__arg_loc_to_register(in, out) is det.
 
+:- pred code_util__max_mentioned_reg(list(lval), int).
+:- mode code_util__max_mentioned_reg(in, out) is det.
+
 	% Determine whether a goal might allocate some heap space,
 	% i.e. whether it contains any construction unifications
 	% or predicate calls.  BEWARE that this predicate is only
@@ -80,6 +103,9 @@
 :- pred code_util__goal_list_may_allocate_heap(list(hlds_goal)).
 :- mode code_util__goal_list_may_allocate_heap(in) is semidet.
 
+:- pred code_util__goal_may_alloc_temp_frame(hlds_goal).
+:- mode code_util__goal_may_alloc_temp_frame(in) is semidet.
+
 	% Negate a condition.
 	% This is used mostly just to make the generated code more readable.
 
@@ -89,6 +115,18 @@
 :- pred code_util__negate_the_test(list(instruction), list(instruction)).
 :- mode code_util__negate_the_test(in, out) is det.
 
+	% code_util__compiler_generated(PredInfo) succeeds iff
+	% the PredInfo is for a compiler generated instance of a
+	% type-specific special_pred (i.e. one of the __Unify__,
+	% __Index__, or __Compare__ predicates generated as a
+	% type-specific instance of unify/2, index/2, or compare/3).
+	% 
+	% XXX The name of this predicate is misleading, because there
+	% are other kinds of compiler-generated predicates, e.g. those
+	% for lambda expressions, those generated by higher-order
+	% specialization, ordinary type specialization, deforestation,
+	% etc., for which this predicate does not succeed.
+
 :- pred code_util__compiler_generated(pred_info).
 :- mode code_util__compiler_generated(in) is semidet.
 
@@ -98,28 +136,6 @@
 :- pred code_util__builtin_state(module_info, pred_id, proc_id, builtin_state).
 :- mode code_util__builtin_state(in, in, in, out) is det.
 
-	% Given a module name, a predicate name, a proc_id and a list of
-	% variables as the arguments, find out if that procedure of that
-	% predicate is an inline builtin. If yes, the last two arguments
-	% return two things:
-	%
-	% - an rval to execute as a test if the builtin is semidet; and
-	%
-	% - an rval to assign to a variable if the builtin calls for this.
-	%
-	% At least one of these will be present.
-	%
-	% Each test rval returned is guaranteed to be either a unop or a binop,
-	% applied to arguments that are either variables (from the argument
-	% list) or constants.
-	%
-	% Each to be assigned rval is guaranteed to be either in a form
-	% acceptable for a test rval, or in the form of a variable.
-
-:- pred code_util__translate_builtin(module_name, string, proc_id,
-	list(prog_var), maybe(rval), maybe(pair(prog_var, rval))).
-:- mode code_util__translate_builtin(in, in, in, in, out, out) is semidet.
-
 	% Find out how a function symbol (constructor) is represented
 	% in the given type.
 
@@ -153,11 +169,6 @@
 	int, int).
 :- mode code_util__count_recursive_calls(in, in, in, out, out) is det.
 
-	% Return the set of locations occupied by output arguments.
-
-:- pred code_util__output_args(assoc_list(prog_var, arg_info), set(lval)).
-:- mode code_util__output_args(in, out) is det.
-
 	% These predicates return the set of lvals referenced in an rval
 	% and an lval respectively. Lvals referenced indirectly through
 	% lvals of the form var(_) are not counted.
@@ -168,65 +179,69 @@
 :- pred code_util__lvals_in_lval(lval, list(lval)).
 :- mode code_util__lvals_in_lval(in, out) is det.
 
+:- pred code_util__lvals_in_lvals(list(lval), list(lval)).
+:- mode code_util__lvals_in_lvals(in, out) is det.
+
 %---------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module prog_data, builtin_ops, type_util, special_pred.
+:- import_module prog_util, type_util, special_pred, builtin_ops, code_model.
 
-:- import_module bool, char, int, string, set, map, term, varset.
+:- import_module char, int, string, set, map, term, varset.
 :- import_module require, std_util, assoc_list.
 
 %---------------------------------------------------------------------------%
 
-code_util__make_entry_label(ModuleInfo, PredId, ProcId, Immed, PredAddress) :-
-	module_info_preds(ModuleInfo, Preds),
-	map__lookup(Preds, PredId, PredInfo),
+code_util__make_entry_label(ModuleInfo, PredId, ProcId, Immed, ProcAddr) :-
+	RttiProcLabel = rtti__make_proc_label(ModuleInfo, PredId, ProcId),
+	code_util__make_entry_label_from_rtti(RttiProcLabel, Immed, ProcAddr).
+
+code_util__make_entry_label_from_rtti(RttiProcLabel, Immed, ProcAddr) :-
 	(
 		(
-			pred_info_is_imported(PredInfo)
+			RttiProcLabel^is_imported = yes
 		;
-			pred_info_is_pseudo_imported(PredInfo),
+			RttiProcLabel^is_pseudo_imported = yes,
 			% only the (in, in) mode of unification is imported
-			hlds_pred__in_in_unification_proc_id(ProcId)
+			hlds_pred__in_in_unification_proc_id(
+				RttiProcLabel^proc_id)
 		)
 	->
-		code_util__make_proc_label(ModuleInfo, PredId, ProcId,
-			ProcLabel),
-		PredAddress = imported(ProcLabel)
+		code_util__make_proc_label_from_rtti(RttiProcLabel, ProcLabel),
+		ProcAddr = imported(ProcLabel)
 	;
-		code_util__make_local_entry_label(ModuleInfo, PredId, ProcId,
+		code_util__make_local_entry_label_from_rtti(RttiProcLabel,
 			Immed, Label),
-		PredAddress = label(Label)
+		ProcAddr = label(Label)
 	).
 
 code_util__make_local_entry_label(ModuleInfo, PredId, ProcId, Immed, Label) :-
-	code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel),
-	module_info_preds(ModuleInfo, Preds),
-	map__lookup(Preds, PredId, PredInfo),
+	RttiProcLabel = rtti__make_proc_label(ModuleInfo, PredId, ProcId),
+	code_util__make_local_entry_label_from_rtti(RttiProcLabel,
+		Immed, Label).
+
+:- pred code_util__make_local_entry_label_from_rtti(rtti_proc_label, immed,
+		label).
+:- mode code_util__make_local_entry_label_from_rtti(in, in, out) is det.
+
+code_util__make_local_entry_label_from_rtti(RttiProcLabel, Immed, Label) :-
+	code_util__make_proc_label_from_rtti(RttiProcLabel, ProcLabel),
 	(
-		procedure_is_exported(PredInfo, ProcId)
-	->
-		(
-			Immed = no,
+		Immed = no,
+		% If we want to define the label or use it to put it
+		% into a data structure, a label that is usable only
+		% within the current C module won't do.
+		( RttiProcLabel^is_exported = yes ->
 			Label = exported(ProcLabel)
 		;
-			Immed = yes(ProcsPerFunc - proc(CurPredId, CurProcId)),
-			choose_local_label_type(ProcsPerFunc, CurPredId,
-				CurProcId, PredId, ProcId, ProcLabel, Label)
+			Label = local(ProcLabel)
 		)
 	;
-		(
-			% If we want to define the label or use it to put it
-			% into a data structure, a label that is usable only
-			% within the current C module won't do.
-			Immed = no,
-			Label = local(ProcLabel)
-		;
-			Immed = yes(ProcsPerFunc - proc(CurPredId, CurProcId)),
-			choose_local_label_type(ProcsPerFunc, CurPredId,
-				CurProcId, PredId, ProcId, ProcLabel, Label)
-		)
+		Immed = yes(ProcsPerFunc - proc(CurPredId, CurProcId)),
+		choose_local_label_type(ProcsPerFunc, CurPredId, CurProcId,
+			RttiProcLabel^pred_id, RttiProcLabel^proc_id,
+			ProcLabel, Label)
 	).
 
 
@@ -257,25 +272,39 @@ choose_local_label_type(ProcsPerFunc, CurPredId, CurProcId,
 
 code_util__make_internal_label(ModuleInfo, PredId, ProcId, LabelNum, Label) :-
 	code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel),
-	Label = local(ProcLabel, LabelNum).
+	Label = local(LabelNum, ProcLabel).
 
 code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel) :-
-	module_info_pred_info(ModuleInfo, PredId, PredInfo),
-	pred_info_module(PredInfo, PredModule),
-	pred_info_name(PredInfo, PredName),
-	module_info_name(ModuleInfo, ThisModule),
+	RttiProcLabel = rtti__make_proc_label(ModuleInfo, PredId, ProcId),
+	code_util__make_proc_label_from_rtti(RttiProcLabel, ProcLabel).
+
+:- pred code_util__make_proc_label_from_rtti(rtti_proc_label, proc_label).
+:- mode code_util__make_proc_label_from_rtti(in, out) is det.
+
+code_util__make_proc_label_from_rtti(RttiProcLabel, ProcLabel) :-
+	RttiProcLabel = rtti_proc_label(PredOrFunc, ThisModule,
+		PredModule, PredName, PredArity, ArgTypes, _PredId, ProcId,
+		_VarSet, _HeadVars, _ArgModes, _CodeModel,
+		IsImported, _IsPseudoImported, _IsExported,
+		IsSpecialPredInstance),
 	(
-		code_util__compiler_generated(PredInfo)
+		IsSpecialPredInstance = yes
 	->
-		pred_info_arg_types(PredInfo, ArgTypes),
 		(
 			special_pred_get_type(PredName, ArgTypes, Type),
 			type_to_type_id(Type, TypeId, _),
-			% All type_ids here should be module qualified,
-			% since builtin types are handled separately in
-			% polymorphism.m.
-			TypeId = qualified(TypeModule, TypeName) - Arity
+			% All type_ids other than tuples here should be
+			% module qualified, since builtin types are
+			% handled separately in polymorphism.m.
+			(
+				TypeId = unqualified(TypeName) - _,
+				type_id_is_tuple(TypeId),
+				mercury_public_builtin_module(TypeModule)
+			;
+				TypeId = qualified(TypeModule, TypeName) - _
+			)
 		->
+			TypeId = _ - TypeArity,
 			(
 				ThisModule \= TypeModule,
 				PredName = "__Unify__",
@@ -286,7 +315,7 @@ code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel) :-
 				DefiningModule = TypeModule
 			),
 			ProcLabel = special_proc(DefiningModule, PredName,
-				TypeModule, TypeName, Arity, ProcId)
+				TypeModule, TypeName, TypeArity, ProcId)
 		;
 			string__append_list(["code_util__make_proc_label:\n",
 				"cannot make label for special pred `",
@@ -294,24 +323,30 @@ code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel) :-
 			error(ErrorMessage)
 		)
 	;
-		(
-			% Work out which module supplies the code for
-			% the predicate.
-			ThisModule \= PredModule,
-			\+ pred_info_is_imported(PredInfo)
-		->
-			% This predicate is a specialized version of 
-			% a pred from a `.opt' file.
-			DefiningModule = ThisModule
-		;	
-			DefiningModule = PredModule
-		),
-		pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
-		pred_info_arity(PredInfo, Arity),
-		ProcLabel = proc(DefiningModule, PredOrFunc,
-			PredModule, PredName, Arity, ProcId)
+		code_util__make_user_proc_label(ThisModule, IsImported,
+			PredOrFunc, PredModule, PredName, PredArity,
+			ProcId, ProcLabel)
 	).
 
+code_util__make_user_proc_label(ThisModule, PredIsImported,
+		PredOrFunc, PredModule, PredName, PredArity,
+		ProcId, ProcLabel) :-
+
+	(
+		% Work out which module supplies the code for
+		% the predicate.
+		ThisModule \= PredModule,
+		PredIsImported = no
+	->
+		% This predicate is a specialized version of 
+		% a pred from a `.opt' file.
+		DefiningModule = ThisModule
+	;	
+		DefiningModule = PredModule
+	),
+	ProcLabel = proc(DefiningModule, PredOrFunc,
+		PredModule, PredName, PredArity, ProcId).
+
 code_util__make_uni_label(ModuleInfo, TypeId, UniModeNum, ProcLabel) :-
 	module_info_name(ModuleInfo, ModuleName),
 	( TypeId = qualified(TypeModule, TypeName) - Arity ->
@@ -344,7 +379,7 @@ code_util__proc_label_from_code_addr(CodeAddr, ProcLabel) :-
 		CodeAddr = imported(ProcLabel)
 	).
 
-code_util__extract_proc_label_from_label(local(ProcLabel, _), ProcLabel).
+code_util__extract_proc_label_from_label(local(_, ProcLabel), ProcLabel).
 code_util__extract_proc_label_from_label(c_local(ProcLabel), ProcLabel).
 code_util__extract_proc_label_from_label(local(ProcLabel), ProcLabel).
 code_util__extract_proc_label_from_label(exported(ProcLabel), ProcLabel).
@@ -355,233 +390,57 @@ code_util__arg_loc_to_register(ArgLoc, reg(r, ArgLoc)).
 
 %-----------------------------------------------------------------------------%
 
+code_util__max_mentioned_reg(Lvals, MaxRegNum) :-
+	code_util__max_mentioned_reg_2(Lvals, 0, MaxRegNum).
+
+:- pred code_util__max_mentioned_reg_2(list(lval)::in, int::in, int::out)
+	is det.
+
+code_util__max_mentioned_reg_2([], MaxRegNum, MaxRegNum).
+code_util__max_mentioned_reg_2([Lval | Lvals], MaxRegNum0, MaxRegNum) :-
+	( Lval = reg(r, N) ->
+		int__max(MaxRegNum0, N, MaxRegNum1)
+	;
+		MaxRegNum1 = MaxRegNum0
+	),
+	code_util__max_mentioned_reg_2(Lvals, MaxRegNum1, MaxRegNum).
+
+%-----------------------------------------------------------------------------%
+
 code_util__predinfo_is_builtin(PredInfo) :-
 	pred_info_module(PredInfo, ModuleName),
 	pred_info_name(PredInfo, PredName),
-%	code_util__translate_builtin(ModuleName, PredName, _, _, _, _).
 	pred_info_arity(PredInfo, Arity),
-	ProcId = 0,
-	code_util__inline_builtin(ModuleName, PredName, ProcId, Arity).
+	hlds_pred__initial_proc_id(ProcId),
+	code_util__is_inline_builtin(ModuleName, PredName, ProcId, Arity).
 
-code_util__builtin_state(ModuleInfo, PredId0, ProcId, BuiltinState) :-
-	predicate_module(ModuleInfo, PredId0, ModuleName),
-	predicate_name(ModuleInfo, PredId0, PredName),
-	predicate_arity(ModuleInfo, PredId0, Arity),
-	proc_id_to_int(ProcId, ProcInt),
-	( code_util__inline_builtin(ModuleName, PredName, ProcInt, Arity) ->
+code_util__builtin_state(ModuleInfo, PredId, ProcId, BuiltinState) :-
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	pred_info_module(PredInfo, ModuleName),
+	pred_info_name(PredInfo, PredName),
+	pred_info_arity(PredInfo, Arity),
+	( code_util__is_inline_builtin(ModuleName, PredName, ProcId, Arity) ->
 		BuiltinState = inline_builtin
 	;
 		BuiltinState = not_builtin
 	).
 
-:- pred code_util__inline_builtin(module_name, string, int, int).
-:- mode code_util__inline_builtin(in, in, in, in) is semidet.
+:- pred code_util__is_inline_builtin(module_name, string, proc_id, arity).
+:- mode code_util__is_inline_builtin(in, in, in, in) is semidet.
 
-code_util__inline_builtin(FullyQualifiedModule, PredName, ProcId, Arity) :-
+code_util__is_inline_builtin(ModuleName, PredName, ProcId, Arity) :-
 	Arity =< 3,
-	varset__init(VarSet),
+	prog_varset_init(VarSet),
 	varset__new_vars(VarSet, Arity, Args, _),
-	% --- not yet:
-	% FullyQualifiedModule = qualified(unqualified("std"), ModuleName),
-	FullyQualifiedModule = unqualified(ModuleName),
-	code_util__translate_builtin_2(ModuleName, PredName, ProcId, Args,
-		_, _).
+	builtin_ops__translate_builtin(ModuleName, PredName, ProcId, Args, _).
 
-code_util__translate_builtin(FullyQualifiedModule, PredName, ProcId, Args,
-		BinOp, AsgOp) :-
-	proc_id_to_int(ProcId, ProcInt),
-	% -- not yet:
-	% FullyQualifiedModule = qualified(unqualified("std"), ModuleName),
-	FullyQualifiedModule = unqualified(ModuleName),
-	code_util__translate_builtin_2(ModuleName, PredName, ProcInt, Args,
-		BinOp, AsgOp).
-
-:- pred code_util__translate_builtin_2(string, string, int, list(prog_var),
-	maybe(rval), maybe(pair(prog_var, rval))).
-:- mode code_util__translate_builtin_2(in, in, in, in, out, out) is semidet.
-
-% WARNING: any changes here will probably require similar changes
-% in ml_code_gen:ml_translate_builtin_2 and vice versa.
-
-code_util__translate_builtin_2("private_builtin", "unsafe_type_cast", 0,
-		[X, Y], no, yes(Y - var(X))).
-code_util__translate_builtin_2("builtin", "unsafe_promise_unique", 0,
-		[X, Y], no, yes(Y - var(X))).
-
-code_util__translate_builtin_2("private_builtin", "builtin_int_gt", 0, [X, Y],
-	yes(binop((>), var(X), var(Y))), no).
-code_util__translate_builtin_2("private_builtin", "builtin_int_lt", 0, [X, Y],
-	yes(binop((<), var(X), var(Y))), no).
-
-code_util__translate_builtin_2("int", "builtin_plus", 0, [X, Y, Z],
-	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_plus", 1, [X, Y, Z],
-	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_plus", 2, [X, Y, Z],
-	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "+", 0, [X, Y, Z],
-	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "+", 1, [X, Y, Z],
-	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "+", 2, [X, Y, Z],
-	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "builtin_minus", 0, [X, Y, Z],
-	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_minus", 1, [X, Y, Z],
-	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_minus", 2, [X, Y, Z],
-	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "-", 0, [X, Y, Z],
-	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "-", 1, [X, Y, Z],
-	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "-", 2, [X, Y, Z],
-	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_times", 0, [X, Y, Z],
-	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_times", 1, [X, Y, Z],
-	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_times", 2, [X, Y, Z],
-	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "*", 0, [X, Y, Z],
-	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "*", 1, [X, Y, Z],
-	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin_2("int", "*", 2, [X, Y, Z],
-	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin_2("int", "builtin_div", 0, [X, Y, Z],
-	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_div", 1, [X, Y, Z],
-	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_div", 2, [X, Y, Z],
-	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "//", 0, [X, Y, Z],
-	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "//", 1, [X, Y, Z],
-	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin_2("int", "//", 2, [X, Y, Z],
-	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin_2("int", "builtin_mod", 0, [X, Y, Z],
-	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "rem", 0, [X, Y, Z],
-	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_left_shift", 0, [X, Y, Z],
-	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "unchecked_left_shift", 0, [X, Y, Z],
-	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_right_shift", 0, [X, Y, Z],
-	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "unchecked_right_shift", 0, [X, Y, Z],
-	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_and", 0, [X, Y, Z],
-	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "/\\", 0, [X, Y, Z],
-	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_or", 0, [X, Y, Z],
-	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "\\/", 0, [X, Y, Z],
-	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_bit_xor", 0, [X, Y, Z],
-	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "^", 0, [X, Y, Z],
-	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "xor", 0, [X, Y, Z],
-	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin_2("int", "builtin_unary_plus", 0, [X, Y],
-	no, yes(Y - var(X))).
-code_util__translate_builtin_2("int", "+", 0, [X, Y],
-	no, yes(Y - var(X))).
-code_util__translate_builtin_2("int", "builtin_unary_minus", 0, [X, Y],
-	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin_2("int", "-", 0, [X, Y],
-	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin_2("int", "builtin_bit_neg", 0, [X, Y],
-	no, yes(Y - unop(bitwise_complement, var(X)))).
-code_util__translate_builtin_2("int", "\\", 0, [X, Y],
-	no, yes(Y - unop(bitwise_complement, var(X)))).
-code_util__translate_builtin_2("int", ">", 0, [X, Y],
-	yes(binop((>), var(X), var(Y))), no).
-code_util__translate_builtin_2("int", "<", 0, [X, Y],
-	yes(binop((<), var(X), var(Y))), no).
-code_util__translate_builtin_2("int", ">=", 0, [X, Y],
-	yes(binop((>=), var(X), var(Y))), no).
-code_util__translate_builtin_2("int", "=<", 0, [X, Y],
-	yes(binop((<=), var(X), var(Y))), no).
-
-code_util__translate_builtin_2("float", "builtin_float_plus", 0, [X, Y, Z],
-	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_plus", 1, [X, Y, Z],
-	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_plus", 2, [X, Y, Z],
-	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "+", 0, [X, Y, Z],
-	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "+", 1, [X, Y, Z],
-	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "+", 2, [X, Y, Z],
-	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 0, [X, Y, Z],
-	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 1, [X, Y, Z],
-	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_minus", 2, [X, Y, Z],
-	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "-", 0, [X, Y, Z],
-	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "-", 1, [X, Y, Z],
-	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "-", 2, [X, Y, Z],
-	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 0, [X, Y, Z],
-	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 1, [X, Y, Z],
-	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_times", 2, [X, Y, Z],
-	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "*", 0, [X, Y, Z],
-	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "*", 1, [X, Y, Z],
-	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin_2("float", "*", 2, [X, Y, Z],
-	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 0, [X, Y, Z],
-	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 1, [X, Y, Z],
-	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "builtin_float_divide", 2, [X, Y, Z],
-	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "/", 0, [X, Y, Z],
-	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin_2("float", "/", 1, [X, Y, Z],
-	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin_2("float", "/", 2, [X, Y, Z],
-	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin_2("float", "+", 0, [X, Y],
-	no, yes(Y - var(X))).
-code_util__translate_builtin_2("float", "-", 0, [X, Y],
-	no, yes(Y - binop(float_minus, const(float_const(0.0)), var(X)))).
-code_util__translate_builtin_2("float", "builtin_float_gt", 0, [X, Y],
-	yes(binop(float_gt, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", ">", 0, [X, Y],
-	yes(binop(float_gt, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", "builtin_float_lt", 0, [X, Y],
-	yes(binop(float_lt, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", "<", 0, [X, Y],
-	yes(binop(float_lt, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", "builtin_float_ge", 0, [X, Y],
-	yes(binop(float_ge, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", ">=", 0, [X, Y],
-	yes(binop(float_ge, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", "builtin_float_le", 0, [X, Y],
-	yes(binop(float_le, var(X), var(Y))), no).
-code_util__translate_builtin_2("float", "=<", 0, [X, Y],
-	yes(binop(float_le, var(X), var(Y))), no).
+:- pred prog_varset_init(prog_varset::out) is det.
+prog_varset_init(VarSet) :- varset__init(VarSet).
 
 %-----------------------------------------------------------------------------%
 
-	% code_util__compiler_generated(PredInfo) should succeed iff
-	% the PredInfo is for a compiler generated predicate.
-
+	% XXX The name of this predicate is misleading -- see the comment
+	% in the declaration.
 code_util__compiler_generated(PredInfo) :-
 	pred_info_name(PredInfo, PredName),
 	pred_info_arity(PredInfo, PredArity),
@@ -589,52 +448,162 @@ code_util__compiler_generated(PredInfo) :-
 
 %-----------------------------------------------------------------------------%
 
-	% This code may _look_ nondeterministic, but it's really semidet,
-	% and Mercury is smart enough to know this.
+code_util__goal_may_allocate_heap(Goal) :-
+	code_util__goal_may_allocate_heap(Goal, yes).
 
-code_util__goal_may_allocate_heap(Goal - _GoalInfo) :-
-	code_util__goal_may_allocate_heap_2(Goal).
+code_util__goal_list_may_allocate_heap(Goals) :-
+	code_util__goal_list_may_allocate_heap(Goals, yes).
 
-:- pred code_util__goal_may_allocate_heap_2(hlds_goal_expr).
-:- mode code_util__goal_may_allocate_heap_2(in) is semidet.
+:- pred code_util__goal_may_allocate_heap(hlds_goal::in, bool::out) is det.
 
-code_util__goal_may_allocate_heap_2(generic_call(_, _, _, _)).
-code_util__goal_may_allocate_heap_2(call(_, _, _, Builtin, _, _)) :-
-	Builtin \= inline_builtin.
-code_util__goal_may_allocate_heap_2(
-		unify(_, _, _, construct(_,_,Args,_,_,_,_), _)) :-
-	Args = [_|_].
-code_util__goal_may_allocate_heap_2(some(_Vars, _, Goal)) :-
-	code_util__goal_may_allocate_heap(Goal).
-code_util__goal_may_allocate_heap_2(not(Goal)) :-
-	code_util__goal_may_allocate_heap(Goal).
-code_util__goal_may_allocate_heap_2(conj(Goals)) :-
-	code_util__goal_list_may_allocate_heap(Goals).
-code_util__goal_may_allocate_heap_2(disj(Goals, _)) :-
-	code_util__goal_list_may_allocate_heap(Goals).
-code_util__goal_may_allocate_heap_2(switch(_Var, _Det, Cases, _)) :-
-	code_util__cases_may_allocate_heap(Cases).
-code_util__goal_may_allocate_heap_2(if_then_else(_Vars, A, B, C, _)) :-
-	(
-		code_util__goal_may_allocate_heap(A)
-	;	
-		code_util__goal_may_allocate_heap(B)
+code_util__goal_may_allocate_heap(Goal - _GoalInfo, May) :-
+	code_util__goal_may_allocate_heap_2(Goal, May).
+
+:- pred code_util__goal_may_allocate_heap_2(hlds_goal_expr::in, bool::out)
+	is det.
+
+code_util__goal_may_allocate_heap_2(generic_call(_, _, _, _), yes).
+code_util__goal_may_allocate_heap_2(call(_, _, _, Builtin, _, _), May) :-
+	( Builtin = inline_builtin ->
+		May = no
 	;
-		code_util__goal_may_allocate_heap(C)
+		May = yes
+	).
+code_util__goal_may_allocate_heap_2(unify(_, _, _, Unification, _), May) :-
+	( Unification = construct(_,_,Args,_,_,_,_), Args = [_|_] ->
+		May = yes
+	;
+		May = no
+	).
+	% We cannot safely say that a C code fragment does not allocate memory
+	% without knowing all the #defined macros that expand to incr_hp and
+	% variants thereof.
+	% XXX although you could make it an attribute of the C code and
+	% trust the programmer
+code_util__goal_may_allocate_heap_2(pragma_foreign_code(_,_,_,_,_,_,_), yes).
+code_util__goal_may_allocate_heap_2(some(_Vars, _, Goal), May) :-
+	code_util__goal_may_allocate_heap(Goal, May).
+code_util__goal_may_allocate_heap_2(not(Goal), May) :-
+	code_util__goal_may_allocate_heap(Goal, May).
+code_util__goal_may_allocate_heap_2(conj(Goals), May) :-
+	code_util__goal_list_may_allocate_heap(Goals, May).
+code_util__goal_may_allocate_heap_2(par_conj(_, _), yes).
+code_util__goal_may_allocate_heap_2(disj(Goals, _), May) :-
+	code_util__goal_list_may_allocate_heap(Goals, May).
+code_util__goal_may_allocate_heap_2(switch(_Var, _Det, Cases, _), May) :-
+	code_util__cases_may_allocate_heap(Cases, May).
+code_util__goal_may_allocate_heap_2(if_then_else(_Vars, C, T, E, _), May) :-
+	( code_util__goal_may_allocate_heap(C, yes) ->
+		May = yes
+	; code_util__goal_may_allocate_heap(T, yes) ->
+		May = yes
+	;
+		code_util__goal_may_allocate_heap(E, May)
+	).
+code_util__goal_may_allocate_heap_2(bi_implication(G1, G2), May) :-
+	( code_util__goal_may_allocate_heap(G1, yes) ->
+		May = yes
+	;
+		code_util__goal_may_allocate_heap(G2, May)
 	).
 
-:- pred code_util__cases_may_allocate_heap(list(case)).
-:- mode code_util__cases_may_allocate_heap(in) is semidet.
+:- pred code_util__goal_list_may_allocate_heap(list(hlds_goal)::in, bool::out)
+	is det.
 
-code_util__cases_may_allocate_heap([case(_, Goal) | _]) :-
-	code_util__goal_may_allocate_heap(Goal).
-code_util__cases_may_allocate_heap([_ | Cases]) :-
-	code_util__cases_may_allocate_heap(Cases).
+code_util__goal_list_may_allocate_heap([], no).
+code_util__goal_list_may_allocate_heap([Goal | Goals], May) :-
+	( code_util__goal_may_allocate_heap(Goal, yes) ->
+		May = yes
+	;
+		code_util__goal_list_may_allocate_heap(Goals, May)
+	).
 
-code_util__goal_list_may_allocate_heap([Goal | _]) :-
-	code_util__goal_may_allocate_heap(Goal).
-code_util__goal_list_may_allocate_heap([_ | Goals]) :-
-	code_util__goal_list_may_allocate_heap(Goals).
+:- pred code_util__cases_may_allocate_heap(list(case)::in, bool::out) is det.
+
+code_util__cases_may_allocate_heap([], no).
+code_util__cases_may_allocate_heap([case(_, Goal) | Cases], May) :-
+	( code_util__goal_may_allocate_heap(Goal, yes) ->
+		May = yes
+	;
+		code_util__cases_may_allocate_heap(Cases, May)
+	).
+
+%-----------------------------------------------------------------------------%
+
+code_util__goal_may_alloc_temp_frame(Goal) :-
+	code_util__goal_may_alloc_temp_frame(Goal, yes).
+
+:- pred code_util__goal_may_alloc_temp_frame(hlds_goal::in, bool::out) is det.
+
+code_util__goal_may_alloc_temp_frame(Goal - _GoalInfo, May) :-
+	code_util__goal_may_alloc_temp_frame_2(Goal, May).
+
+:- pred code_util__goal_may_alloc_temp_frame_2(hlds_goal_expr::in, bool::out)
+	is det.
+
+code_util__goal_may_alloc_temp_frame_2(generic_call(_, _, _, _), no).
+code_util__goal_may_alloc_temp_frame_2(call(_, _, _, _, _, _), no).
+code_util__goal_may_alloc_temp_frame_2(unify(_, _, _, _, _), no).
+	% We cannot safely say that a C code fragment does not allocate
+	% temporary nondet frames without knowing all the #defined macros
+	% that expand to mktempframe and variants thereof. The performance
+	% impact of being too conservative is probably not too bad.
+code_util__goal_may_alloc_temp_frame_2(pragma_foreign_code(_,_,_,_,_,_,_),
+		yes).
+code_util__goal_may_alloc_temp_frame_2(some(_Vars, _, Goal), May) :-
+	Goal = _ - GoalInfo,
+	goal_info_get_code_model(GoalInfo, CodeModel),
+	( CodeModel = model_non ->
+		May = yes
+	;
+		code_util__goal_may_alloc_temp_frame(Goal, May)
+	).
+code_util__goal_may_alloc_temp_frame_2(not(Goal), May) :-
+	code_util__goal_may_alloc_temp_frame(Goal, May).
+code_util__goal_may_alloc_temp_frame_2(conj(Goals), May) :-
+	code_util__goal_list_may_alloc_temp_frame(Goals, May).
+code_util__goal_may_alloc_temp_frame_2(par_conj(Goals, _), May) :-
+	code_util__goal_list_may_alloc_temp_frame(Goals, May).
+code_util__goal_may_alloc_temp_frame_2(disj(Goals, _), May) :-
+	code_util__goal_list_may_alloc_temp_frame(Goals, May).
+code_util__goal_may_alloc_temp_frame_2(switch(_Var, _Det, Cases, _), May) :-
+	code_util__cases_may_alloc_temp_frame(Cases, May).
+code_util__goal_may_alloc_temp_frame_2(if_then_else(_Vars, C, T, E, _), May) :-
+	( code_util__goal_may_alloc_temp_frame(C, yes) ->
+		May = yes
+	; code_util__goal_may_alloc_temp_frame(T, yes) ->
+		May = yes
+	;
+		code_util__goal_may_alloc_temp_frame(E, May)
+	).
+code_util__goal_may_alloc_temp_frame_2(bi_implication(G1, G2), May) :-
+	( code_util__goal_may_alloc_temp_frame(G1, yes) ->
+		May = yes
+	;
+		code_util__goal_may_alloc_temp_frame(G2, May)
+	).
+
+:- pred code_util__goal_list_may_alloc_temp_frame(list(hlds_goal)::in,
+	bool::out) is det.
+
+code_util__goal_list_may_alloc_temp_frame([], no).
+code_util__goal_list_may_alloc_temp_frame([Goal | Goals], May) :-
+	( code_util__goal_may_alloc_temp_frame(Goal, yes) ->
+		May = yes
+	;
+		code_util__goal_list_may_alloc_temp_frame(Goals, May)
+	).
+
+:- pred code_util__cases_may_alloc_temp_frame(list(case)::in, bool::out)
+	is det.
+
+code_util__cases_may_alloc_temp_frame([], no).
+code_util__cases_may_alloc_temp_frame([case(_, Goal) | Cases], May) :-
+	( code_util__goal_may_alloc_temp_frame(Goal, yes) ->
+		May = yes
+	;
+		code_util__cases_may_alloc_temp_frame(Cases, May)
+	).
 
 %-----------------------------------------------------------------------------%
 
@@ -716,6 +685,13 @@ code_util__cons_id_to_tag(cons(Name, Arity), Type, ModuleInfo, Tag) :-
 	->
 		char__to_int(Char, CharCode),
 		Tag = int_constant(CharCode)
+	;
+		% Tuples do not need a tag. Note that unary tuples are not
+		% treated as no_tag types. There's no reason why they
+		% couldn't be, it's just not worth the effort.
+		type_is_tuple(Type, _)
+	->
+		Tag = unshared_tag(0)
 	;
 			% Use the type to determine the type_id
 		( type_to_type_id(Type, TypeId0, _) ->
@@ -835,8 +811,8 @@ code_util__count_recursive_calls_2(some(_, _, Goal),
 code_util__count_recursive_calls_2(unify(_, _, _, _, _), _, _, 0, 0).
 code_util__count_recursive_calls_2(generic_call(_, _, _, _), _, _,
 		0, 0).
-code_util__count_recursive_calls_2(pragma_c_code(_,_,_, _, _, _, _), _, _,
-		0, 0).
+code_util__count_recursive_calls_2(pragma_foreign_code(_, _, _, _, _, _, _),
+		_, _, 0, 0).
 code_util__count_recursive_calls_2(call(CallPredId, CallProcId, _, _, _, _),
 		PredId, ProcId, Count, Count) :-
 	(
@@ -924,25 +900,18 @@ code_util__count_recursive_calls_cases([case(_, Goal) | Cases], PredId, ProcId,
 		int__max(Max0, Max1, Max)
 	).
 
-code_util__output_args([], LiveVals) :-
-	set__init(LiveVals).
-code_util__output_args([_V - arg_info(Loc, Mode) | Args], Vs) :-
-	code_util__output_args(Args, Vs0),
-	(
-		Mode = top_out
-	->
-		code_util__arg_loc_to_register(Loc, Reg),
-		set__insert(Vs0, Reg, Vs)
-	;
-		Vs = Vs0
-	).
-
 %-----------------------------------------------------------------------------%
 
+code_util__lvals_in_lvals([], []).
+code_util__lvals_in_lvals([First | Rest], Lvals) :-
+	code_util__lvals_in_lval(First, FirstLvals),
+	code_util__lvals_in_lvals(Rest, RestLvals),
+	list__append(FirstLvals, RestLvals, Lvals).
+
 code_util__lvals_in_rval(lval(Lval), [Lval | Lvals]) :-
 	code_util__lvals_in_lval(Lval, Lvals).
 code_util__lvals_in_rval(var(_), []).
-code_util__lvals_in_rval(create(_, _, _, _, _, _), []).
+code_util__lvals_in_rval(create(_, _, _, _, _, _, _), []).
 code_util__lvals_in_rval(mkword(_, Rval), Lvals) :-
 	code_util__lvals_in_rval(Rval, Lvals).
 code_util__lvals_in_rval(const(_), []).
