@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1997-1999 The University of Melbourne.
+% Copyright (C) 1997-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -61,26 +61,50 @@
 	%
 :- type proc_layout_info
 	--->	proc_layout_info(
-			label,		% The entry label.
-			determinism,	% Determines which stack is used.
-			int,		% Number of stack slots.
-			maybe(int),	% Location of succip on stack.
-			maybe(label),	% If the trace level is not none,
+			entry_label	:: label,
+					% Determines which stack is used.
+			detism		:: determinism,
+					% Number of stack slots.
+			stack_slot_count :: int,
+					% Location of succip on stack.
+			succip_slot	:: maybe(int),
+					% If the trace level is not none, this
+					% contains the label associated with
+					% the call event, whose stack layout
+					% gives the locations of the input
+					% arguments on procedure entry, for
+					% use in implementing retry in the
+					% debugger.
+			eval_method	:: eval_method,
+					% Of the procedure.
+			call_label	:: maybe(label),
+					% If the trace level is not none,
 					% this contains the label associated
 					% with the call event, whose stack
 					% layout says which variables were
 					% live and where on entry.
-			int,		% The number of the highest numbered
+			max_trace_reg	:: int,
+					% The number of the highest numbered
 					% rN register that can contain useful
 					% information during a call to MR_trace
 					% from within this procedure.
-			trace_slot_info,% Info about the stack slots used
+			proc_body	:: hlds_goal,
+					% The body of the procedure.
+			initial_instmap	:: instmap,
+					% The instmap at the start of the
+					% procedure body.
+			trace_slot_info	:: trace_slot_info,
+					% Info about the stack slots used
 					% for tracing.
-			bool,		% Do we require the procedure id
+			need_proc_id	:: bool,
+					% Do we require the procedure id
 					% section of the procedure layout
 					% to be present, even if the option
 					% procid_stack_layout is not set?
-			proc_label_layout_info
+			varset		:: prog_varset,
+					% The names of all the variables.
+			vartypes	:: vartypes,
+			internal_map	:: proc_label_layout_info
 					% Info for each internal label,
 					% needed for basic_stack_layouts.
 		).
@@ -183,7 +207,7 @@
 
 :- type return_layout_info
 	--->	return_layout_info(
-			list(pair(code_addr, prog_context)),
+			assoc_list(code_addr, pair(prog_context, goal_path)),
 			layout_label_info
 		).
 
@@ -263,13 +287,13 @@
 	% of a call.
 :- pred continuation_info__generate_return_live_lvalues(
 	assoc_list(prog_var, arg_loc)::in, instmap::in, list(prog_var)::in,
-	map(prog_var, set(rval))::in, assoc_list(lval, slot_contents)::in,
+	map(prog_var, set(lval))::in, assoc_list(lval, slot_contents)::in,
 	proc_info::in, module_info::in, globals::in, list(liveinfo)::out)
 	is det.
 
 	% Generate the layout information we need for a resumption point,
 	% a label where forward execution can restart after backtracking.
-:- pred continuation_info__generate_resume_layout(map(prog_var, set(rval))::in,
+:- pred continuation_info__generate_resume_layout(map(prog_var, set(lval))::in,
 	assoc_list(lval, slot_contents)::in, instmap::in, proc_info::in,
 	module_info::in, layout_label_info::out) is det.
 
@@ -280,7 +304,7 @@
 	% For each type variable in the given list, find out where the
 	% typeinfo var for that type variable is.
 :- pred continuation_info__find_typeinfos_for_tvars(list(tvar)::in,
-	map(prog_var, set(rval))::in, proc_info::in,
+	map(prog_var, set(lval))::in, proc_info::in,
 	map(tvar, set(layout_locn))::out) is det.
 
 %-----------------------------------------------------------------------------%
@@ -289,7 +313,7 @@
 :- implementation.
 
 :- import_module hlds_goal, code_util, type_util, inst_match, options.
-:- import_module string, require, varset, term.
+:- import_module string, require, term, varset.
 
 %-----------------------------------------------------------------------------%
 
@@ -297,7 +321,7 @@
 
 continuation_info__maybe_process_llds([], _) --> [].
 continuation_info__maybe_process_llds([Proc | Procs], ModuleInfo) -->
-	{ Proc = c_procedure(_, _, PredProcId, Instrs) },
+	{ Proc = c_procedure(_, _, PredProcId, Instrs, _, _, _) },
 	continuation_info__maybe_process_proc_llds(Instrs, PredProcId,
 		ModuleInfo),
 	continuation_info__maybe_process_llds(Procs, ModuleInfo).
@@ -322,7 +346,9 @@ continuation_info__maybe_process_proc_llds(Instructions, PredProcId,
 			label,		% the return label
 			code_addr,	% the target of the call
 			list(liveinfo),	% what is live on return
-			term__context	% the position of the call in source
+			term__context,	% the position of the call in source
+			goal_path	% the position of the call in the body;
+					% meaningful only if tracing is enabled
 		).
 
 	%
@@ -338,11 +364,12 @@ continuation_info__process_proc_llds(PredProcId, Instructions,
 
 		% Get all the continuation info from the call instructions.
 	global_data_get_proc_layout(GlobalData0, PredProcId, ProcLayoutInfo0),
-	ProcLayoutInfo0 = proc_layout_info(A, B, C, D, E, F, G, H, Internals0),
+	Internals0 = ProcLayoutInfo0^internal_map,
 	GetCallInfo = lambda([Instr::in, Call::out] is semidet, (
-		Instr = call(Target, label(ReturnLabel), LiveInfo, Context, _)
-			- _Comment,
-		Call = call_info(ReturnLabel, Target, LiveInfo, Context)
+		Instr = call(Target, label(ReturnLabel), LiveInfo, Context,
+			GoalPath, _) - _Comment,
+		Call = call_info(ReturnLabel, Target, LiveInfo, Context,
+			GoalPath)
 	)),
 	list__filter_map(GetCallInfo, Instructions, Calls),
 
@@ -350,7 +377,7 @@ continuation_info__process_proc_llds(PredProcId, Instructions,
 	list__foldl(continuation_info__process_continuation(WantReturnInfo),
 		Calls, Internals0, Internals),
 
-	ProcLayoutInfo = proc_layout_info(A, B, C, D, E, F, G, H, Internals),
+	ProcLayoutInfo = ProcLayoutInfo0^internal_map := Internals,
 	global_data_update_proc_layout(GlobalData0, PredProcId, ProcLayoutInfo,
 		GlobalData).
 
@@ -365,7 +392,8 @@ continuation_info__process_proc_llds(PredProcId, Instructions,
 
 continuation_info__process_continuation(WantReturnInfo, CallInfo,
 		Internals0, Internals) :-
-	CallInfo = call_info(ReturnLabel, Target, LiveInfoList, Context),
+	CallInfo = call_info(ReturnLabel, Target, LiveInfoList,
+		Context, MaybeGoalPath),
 	( map__search(Internals0, ReturnLabel, Internal0) ->
 		Internal0 = internal_layout_info(Port0, Resume0, Return0)
 	;
@@ -379,7 +407,8 @@ continuation_info__process_continuation(WantReturnInfo, CallInfo,
 		(
 			Return0 = no,
 			Layout = layout_label_info(VarInfoSet, TypeInfoMap),
-			ReturnInfo = return_layout_info([Target - Context],
+			ReturnInfo = return_layout_info(
+				[Target - (Context - MaybeGoalPath)],
 				Layout),
 			Return = yes(ReturnInfo)
 		;
@@ -395,7 +424,8 @@ continuation_info__process_continuation(WantReturnInfo, CallInfo,
 			set__intersect(LV0, VarInfoSet, LV),
 			map__intersect(set__intersect, TV0, TypeInfoMap, TV),
 			Layout = layout_label_info(LV, TV),
-			TargetContexts = [Target - Context | TargetsContexts0],
+			TargetContexts = [Target - (Context - MaybeGoalPath)
+				| TargetsContexts0],
 			ReturnInfo = return_layout_info(TargetContexts,
 				Layout),
 			Return = yes(ReturnInfo)
@@ -516,7 +546,7 @@ continuation_info__generate_temp_live_lvalues([Temp | Temps], [Live | Lives]) :-
 
 :- pred continuation_info__generate_var_live_lvalues(
 	assoc_list(prog_var, lval)::in, instmap::in,
-	map(prog_var, set(rval))::in, proc_info::in, module_info::in,
+	map(prog_var, set(lval))::in, proc_info::in, module_info::in,
 	bool::in, list(liveinfo)::out) is det.
 
 continuation_info__generate_var_live_lvalues([], _, _, _, _, _, []).
@@ -554,37 +584,32 @@ continuation_info__generate_resume_layout(ResumeMap, Temps, InstMap,
 	Layout = layout_label_info(AllInfoSet, TVarInfoMap).
 
 :- pred continuation_info__generate_resume_layout_for_vars(
-	assoc_list(prog_var, set(rval))::in, instmap::in, proc_info::in,
+	assoc_list(prog_var, set(lval))::in, instmap::in, proc_info::in,
 	module_info::in, list(var_info)::out, set(tvar)::in,
 	set(tvar)::out) is det.
 
 continuation_info__generate_resume_layout_for_vars([], _, _, _, [],
 		TVars, TVars).
-continuation_info__generate_resume_layout_for_vars([Var - RvalSet | VarRvals],
+continuation_info__generate_resume_layout_for_vars([Var - LvalSet | VarLvals],
 		InstMap, ProcInfo, ModuleInfo, [VarInfo | VarInfos],
 		TVars0, TVars) :-
-	continuation_info__generate_resume_layout_for_var(Var, RvalSet,
+	continuation_info__generate_resume_layout_for_var(Var, LvalSet,
 		InstMap, ProcInfo, ModuleInfo, VarInfo, TypeVars),
 	set__insert_list(TVars0, TypeVars, TVars1),
-	continuation_info__generate_resume_layout_for_vars(VarRvals,
+	continuation_info__generate_resume_layout_for_vars(VarLvals,
 		InstMap, ProcInfo, ModuleInfo, VarInfos, TVars1, TVars).
 
 :- pred continuation_info__generate_resume_layout_for_var(prog_var::in,
-	set(rval)::in, instmap::in, proc_info::in, module_info::in,
+	set(lval)::in, instmap::in, proc_info::in, module_info::in,
 	var_info::out, list(tvar)::out) is det.
 
-continuation_info__generate_resume_layout_for_var(Var, RvalSet, InstMap,
+continuation_info__generate_resume_layout_for_var(Var, LvalSet, InstMap,
 		ProcInfo, ModuleInfo, VarInfo, TypeVars) :-
-	set__to_sorted_list(RvalSet, RvalList),
-	( RvalList = [RvalPrime] ->
-		Rval = RvalPrime
-	;
-		error("var has more than one rval in stack resume map")
-	),
-	( Rval = lval(LvalPrime) ->
+	set__to_sorted_list(LvalSet, LvalList),
+	( LvalList = [LvalPrime] ->
 		Lval = LvalPrime
 	;
-		error("var rval is not lval in stack resume map")
+		error("var has more than one lval in stack resume map")
 	),
 	continuation_info__generate_layout_for_var(Var, InstMap, ProcInfo,
 		ModuleInfo, LiveValueType, TypeVars),
@@ -623,7 +648,7 @@ continuation_info__generate_layout_for_var(Var, InstMap, ProcInfo, ModuleInfo,
 		LldsInst = partial(Inst)
 	),
 	LiveValueType = var(Var, Name, Type, LldsInst),
-	type_util__vars(Type, TypeVars).
+	type_util__real_vars(Type, TypeVars).
 
 %---------------------------------------------------------------------------%
 
@@ -653,8 +678,8 @@ continuation_info__generate_closure_layout(ModuleInfo, PredId, ProcId,
 
 :- pred continuation_info__build_closure_info(list(prog_var)::in,
 	list(type)::in, list(arg_info)::in,  list(closure_arg_info)::out,
-	instmap::in, map(prog_var, set(rval))::in,
-	map(prog_var, set(rval))::out, set(tvar)::in, set(tvar)::out)
+	instmap::in, map(prog_var, set(lval))::in,
+	map(prog_var, set(lval))::out, set(tvar)::in, set(tvar)::out)
 	is semidet.
 
 continuation_info__build_closure_info([], [], [], [], _, VarLocs, VarLocs,
@@ -665,9 +690,9 @@ continuation_info__build_closure_info([Var | Vars], [Type | Types],
 	ArgInfo = arg_info(ArgLoc, _ArgMode),
 	instmap__lookup_var(InstMap, Var, Inst),
 	Layout = closure_arg_info(Type, Inst),
-	set__singleton_set(Locations, lval(reg(r, ArgLoc))),
+	set__singleton_set(Locations, reg(r, ArgLoc)),
 	map__det_insert(VarLocs0, Var, Locations, VarLocs1),
-	type_util__vars(Type, VarTypeVars),
+	type_util__real_vars(Type, VarTypeVars),
 	set__insert_list(TypeVars0, VarTypeVars, TypeVars1),
 	continuation_info__build_closure_info(Vars, Types, ArgInfos, Layouts,
 		InstMap, VarLocs1, VarLocs, TypeVars1, TypeVars).
@@ -682,11 +707,10 @@ continuation_info__find_typeinfos_for_tvars(TypeVars, VarLocs, ProcInfo,
 	FindLocn = lambda([TypeInfoLocn::in, Locns::out] is det, (
 		type_info_locn_var(TypeInfoLocn, TypeInfoVar),
 		(
-			map__search(VarLocs, TypeInfoVar, TypeInfoRvalSet)
+			map__search(VarLocs, TypeInfoVar, TypeInfoLvalSet)
 		->
-			ConvertRval = lambda([Locn::out] is nondet, (
-				set__member(Rval, TypeInfoRvalSet),
-				Rval = lval(Lval),
+			ConvertLval = lambda([Locn::out] is nondet, (
+				set__member(Lval, TypeInfoLvalSet),
 				(
 					TypeInfoLocn = typeclass_info(_,
 						FieldNum),
@@ -696,7 +720,7 @@ continuation_info__find_typeinfos_for_tvars(TypeVars, VarLocs, ProcInfo,
 					Locn = direct(Lval)
 				)
 			)),
-			solutions_set(ConvertRval, Locns)
+			solutions_set(ConvertLval, Locns)
 		;
 			varset__lookup_name(VarSet, TypeInfoVar,
 				VarString),
