@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1999 University of Melbourne.
+% Copyright (C) 1999-2000 University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -122,8 +122,9 @@ deforest__proc(proc(PredId, ProcId), CostDelta, SizeDelta) -->
 		{ proc_info_get_initial_instmap(ProcInfo3,
 			ModuleInfo2, InstMap0) },
 		{ proc_info_vartypes(ProcInfo3, VarTypes) },
-		{ recompute_instmap_delta(yes, Goal3, Goal, 
-			VarTypes, InstMap0, ModuleInfo2, ModuleInfo3) },
+		{ proc_info_inst_varset(ProcInfo3, InstVarSet) },
+		{ recompute_instmap_delta(yes, Goal3, Goal, VarTypes,
+			InstVarSet, InstMap0, ModuleInfo2, ModuleInfo3) },
 		pd_info_set_module_info(ModuleInfo3),
 
 		pd_info_get_pred_info(PredInfo),
@@ -200,7 +201,7 @@ deforest__goal(switch(Var, CanFail, Cases0, SM) - Info,
 	deforest__cases(Var, Cases0, Cases).
 
 deforest__goal(Goal, Goal) -->
-	{ Goal = pragma_c_code(_, _, _, _, _, _, _) - _ }.
+	{ Goal = pragma_foreign_code(_, _, _, _, _, _, _) - _ }.
 
 deforest__goal(Goal, Goal) -->
 	{ Goal = generic_call(_, _, _, _) - _ }.
