@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,8 +14,10 @@
 
 :- interface.
 
+:- import_module prog_data.
 :- import_module hlds_module, hlds_pred, hlds_goal.
-:- import_module det_util, prog_data.
+:- import_module det_util.
+
 :- import_module io, list.
 
 :- type det_msg	--->
@@ -120,10 +122,13 @@
 
 :- implementation.
 
+:- import_module prog_out. 
 :- import_module hlds_data, type_util, mode_util, inst_match.
-:- import_module globals, options, prog_out, hlds_out, mercury_to_mercury.
-:- import_module passes_aux, term, varset.
+:- import_module hlds_out, mercury_to_mercury.
+:- import_module code_util, passes_aux.
+:- import_module globals, options.
 
+:- import_module term, varset.
 :- import_module bool, int, map, set, std_util, require, string.
 
 %-----------------------------------------------------------------------------%
@@ -134,7 +139,7 @@ global_checking_pass([proc(PredId, ProcId) | Rest], ModuleInfo0, ModuleInfo) -->
 		PredInfo, ProcInfo) },
 	check_determinism(PredId, ProcId, PredInfo, ProcInfo,
 		ModuleInfo0, ModuleInfo1),
-	check_if_main_can_fail(PredId, ProcId, PredInfo, ProcInfo,
+	check_determinism_of_main(PredId, ProcId, PredInfo, ProcInfo,
 		ModuleInfo1, ModuleInfo2),
 	check_for_multisoln_func(PredId, ProcId, PredInfo, ProcInfo,
 		ModuleInfo2, ModuleInfo3),
@@ -162,7 +167,27 @@ check_determinism(PredId, ProcId, PredInfo0, ProcInfo0,
 			globals__io_lookup_bool_option(
 				warn_det_decls_too_lax,
 				ShouldIssueWarning),
-			( { ShouldIssueWarning = yes } ->
+			{ pred_info_get_markers(PredInfo0, Markers) },
+			(
+				{ ShouldIssueWarning = yes },
+
+				% Don't report warnings for class method
+				% implementations -- the determinism in the
+				% `:- typeclass' declaration will be
+				% the loosest of all possible instances.
+				% This is similar to the reason we don't
+				% report warnings for lambda expressions.
+				{ \+ check_marker(Markers,
+					class_instance_method) },
+
+				% Don't report warnings for compiler-generated
+				% Unify, Compare or Index procedures, since the
+				% user has no way to shut these up. These can
+				% happen for the Unify pred for the unit type,
+				% if such types are not boxed (as they are not
+				% boxed for the IL backend).
+				{ \+ code_util__compiler_generated(PredInfo0) }
+			->
 				{ Message = "  warning: determinism declaration could be tighter.\n" },
 				report_determinism_problem(PredId,
 					ProcId, ModuleInfo0, Message,
@@ -179,9 +204,10 @@ check_determinism(PredId, ProcId, PredInfo0, ProcInfo0,
 				ProcId, ModuleInfo1, Message,
 				DeclaredDetism, InferredDetism),
 			{ proc_info_goal(ProcInfo0, Goal) },
+			{ proc_info_vartypes(ProcInfo0, VarTypes) },
 			globals__io_get_globals(Globals),
-			{ det_info_init(ModuleInfo1, PredId, ProcId, Globals,
-				DetInfo) },
+			{ det_info_init(ModuleInfo1, VarTypes, PredId, ProcId,
+				Globals, DetInfo) },
 			det_diagnose_goal(Goal, DeclaredDetism, [], DetInfo, _)
 			% XXX with the right verbosity options, we want to
 			% call report_determinism_problem only if diagnose
@@ -191,9 +217,8 @@ check_determinism(PredId, ProcId, PredInfo0, ProcInfo0,
 	
 	% make sure the code model is valid given the eval method
 	{ proc_info_eval_method(ProcInfo0, EvalMethod) },
-	{ determinism_to_code_model(InferredDetism, CodeMod) },
 	( 
-		{ valid_code_model_for_eval_method(EvalMethod, CodeMod) }
+		{ valid_determinism_for_eval_method(EvalMethod, InferredDetism) }
 	->
 		{
 		    proc_info_set_eval_method(ProcInfo0, EvalMethod, ProcInfo),
@@ -227,11 +252,25 @@ check_determinism(PredId, ProcId, PredInfo0, ProcInfo0,
 	).
 
 :- pred get_valid_dets(eval_method, determinism).
-:- mode get_valid_dets(in, out) is multidet.
+:- mode get_valid_dets(in, out) is nondet.
 
-get_valid_dets(EvalMethod, Det) :-
-	valid_code_model_for_eval_method(EvalMethod, CodeModel),
-	determinism_to_code_model(Det, CodeModel).
+get_valid_dets(EvalMethod, Detism) :-
+	determinism(Detism),
+	valid_determinism_for_eval_method(EvalMethod, Detism).
+
+	% generate all the possible determinisms
+:- pred determinism(determinism).
+:- mode determinism(out) is multi.
+:- mode determinism(in) is det. % to ensure we don't forget any
+
+determinism(det).
+determinism(semidet).
+determinism(multidet).
+determinism(nondet).
+determinism(cc_multidet).
+determinism(cc_nondet).
+determinism(erroneous).
+determinism(failure).
 
 :- pred print_dets(list(determinism), io__state, io__state).
 :- mode print_dets(in, di, uo) is det.
@@ -243,31 +282,26 @@ print_dets([D|Rest]) -->
 	io__nl,
 	print_dets(Rest).
 	
-:- pred check_if_main_can_fail(pred_id, proc_id, pred_info, proc_info,
+:- pred check_determinism_of_main(pred_id, proc_id, pred_info, proc_info,
 		module_info, module_info, io__state, io__state).
-:- mode check_if_main_can_fail(in, in, in, in, in, out, di, uo) is det.
+:- mode check_determinism_of_main(in, in, in, in, in, out, di, uo) is det.
 
-check_if_main_can_fail(_PredId, _ProcId, PredInfo, ProcInfo,
+check_determinism_of_main(_PredId, _ProcId, PredInfo, ProcInfo,
 		ModuleInfo0, ModuleInfo) -->
+	%
+	% check that `main/2' has determinism `det' or `cc_multi',
+	% as required by the language reference manual
+	%
 	{ proc_info_declared_determinism(ProcInfo, MaybeDetism) },
-	{ proc_info_inferred_determinism(ProcInfo, InferredDetism) },
-	% check that `main/2' cannot fail
 	( 
 		{ pred_info_name(PredInfo, "main") },
 		{ pred_info_arity(PredInfo, 2) },
 		{ pred_info_is_exported(PredInfo) },
-		{
-		  determinism_components(InferredDetism, can_fail, _)
-		;
-		  MaybeDetism = yes(DeclaredDeterminism),
-		  determinism_components(DeclaredDeterminism, can_fail, _)
-		}
+		{ MaybeDetism = yes(DeclaredDetism) },
+		{ DeclaredDetism \= det, DeclaredDetism \= cc_multidet }
 	->
 		{ proc_info_context(ProcInfo, Context1) },
 		prog_out__write_context(Context1),
-			% The error message is actually a lie -
-			% main/2 can also be `erroneous'.  But mentioning
-			% that would probably just confuse people.
 		io__write_string(
 			"Error: main/2 must be `det' or `cc_multi'.\n"),
 		{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }
@@ -606,8 +640,8 @@ det_diagnose_goal_2(some(_Vars, _, Goal), _, Desired, Actual,
 	det_diagnose_goal(Goal, InternalDesired, SwitchContext, DetInfo,
 		Diagnosed).
 
-det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _, _), GoalInfo, Desired, 
-		_, _, _, yes) -->
+det_diagnose_goal_2(pragma_foreign_code(_, _, _, _, _, _, _), GoalInfo,
+		Desired, _, _, _, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
 	io__write_string("  Determinism declaration not satisfied. Desired \n"),
@@ -1262,7 +1296,10 @@ det_report_msg(error_in_lambda(DeclaredDetism, InferredDetism, Goal, GoalInfo,
 	hlds_out__write_determinism(InferredDetism),
 	io__write_string("'.\n"),
 	globals__io_get_globals(Globals),
-	{ det_info_init(ModuleInfo, PredId, ProcId, Globals, DetInfo) },
+	{ module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _, ProcInfo) },
+	{ proc_info_vartypes(ProcInfo, VarTypes) },
+	{ det_info_init(ModuleInfo, VarTypes, PredId, ProcId, Globals,
+		DetInfo) },
 	det_diagnose_goal(Goal, DeclaredDetism, [], DetInfo, _),
 	io__set_exit_status(1).
 det_report_msg(par_conj_not_det(InferredDetism, PredId,
@@ -1289,7 +1326,10 @@ det_report_msg(par_conj_not_det(InferredDetism, PredId,
 	prog_out__write_context(Context),
 	io__write_string("  non-failing parallel conjunctions.\n"),
 	globals__io_get_globals(Globals),
-	{ det_info_init(ModuleInfo, PredId, ProcId, Globals, DetInfo) },
+	{ module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _, ProcInfo) },
+	{ proc_info_vartypes(ProcInfo, VarTypes) },
+	{ det_info_init(ModuleInfo, VarTypes, PredId, ProcId, Globals,
+		DetInfo) },
 	det_diagnose_conj(Goals, det, [], DetInfo, _),
 	io__set_exit_status(1).
 det_report_msg(pragma_c_code_without_det_decl(PredId, ProcId),
