@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -53,13 +53,14 @@ equiv_type__expand_eqv_types(Items0, Items, CircularTypes, EqvMap) -->
 	{ map__init(EqvMap0) },
 	{ equiv_type__build_eqv_map(Items0, EqvMap0, EqvMap) },
 	{ equiv_type__replace_in_item_list(Items0, EqvMap,
-		Items, [], CircularTypeList0) },
-	{ list__reverse(CircularTypeList0, CircularTypeList) },
+		[], RevItems, [], RevCircularTypeList) },
+	{ list__reverse(RevItems, Items) },
 	(
-		{ CircularTypeList = [] }
+		{ RevCircularTypeList = [] }
 	->
 		{ CircularTypes = no }	
 	;
+		{ list__reverse(RevCircularTypeList, CircularTypeList) },
 		equiv_type__report_circular_types(CircularTypeList),
 		{ CircularTypes = yes },
 		io__set_exit_status(1)
@@ -88,24 +89,28 @@ equiv_type__build_eqv_map([Item - _Context | Items], EqvMap0, EqvMap) :-
 	% follow perform substitution of equivalence types on <foo>s.
 
 :- pred equiv_type__replace_in_item_list(list(item_and_context), eqv_map,
-	list(item_and_context), list(item_and_context), list(item_and_context)).
-:- mode equiv_type__replace_in_item_list(in, in, out, in, out) is det.
+	list(item_and_context), list(item_and_context),
+	list(item_and_context), list(item_and_context)).
+:- mode equiv_type__replace_in_item_list(in, in, in, out, in, out) is det.
 
-equiv_type__replace_in_item_list([], _, [], Circ, Circ).
-equiv_type__replace_in_item_list([Item0 - Context | Items0], EqvMap,
-		[Item - Context | Items], Circ0, Circ) :-
-	( equiv_type__replace_in_item(Item0, EqvMap, Item1, ContainsCirc) ->
-		Item = Item1,
+equiv_type__replace_in_item_list([], _, Items, Items, Circ, Circ).
+equiv_type__replace_in_item_list([ItemAndContext0 | Items0], EqvMap,
+		ReplItems0, ReplItems, Circ0, Circ) :-
+	ItemAndContext0 = Item0 - Context,
+	( equiv_type__replace_in_item(Item0, EqvMap, Item, ContainsCirc) ->
+		ItemAndContext = Item - Context,
 		( ContainsCirc = yes ->
-			Circ1 = [Item - Context | Circ0]
+			Circ1 = [ItemAndContext | Circ0]
 		;
 			Circ1 = Circ0
 		)
 	;
-		Item = Item0,
+		ItemAndContext = ItemAndContext0,
 		Circ1 = Circ0
 	),
-	equiv_type__replace_in_item_list(Items0, EqvMap, Items, Circ1, Circ).
+	ReplItems1 = [ItemAndContext | ReplItems0],
+	equiv_type__replace_in_item_list(Items0, EqvMap, ReplItems1, ReplItems,
+		Circ1, Circ).
 
 :- pred equiv_type__replace_in_item(item, eqv_map, item, bool).
 :- mode equiv_type__replace_in_item(in, in, out, out) is semidet.
@@ -157,10 +162,10 @@ equiv_type__replace_in_item(
 
 equiv_type__replace_in_item(
 			instance(Constraints0, ClassName, Ts0, 
-				InstanceBody, VarSet0),
+				InstanceBody, VarSet0, ModName),
 			EqvMap,
 			instance(Constraints, ClassName, Ts, 
-				InstanceBody, VarSet),
+				InstanceBody, VarSet, ModName),
 			no) :-
 	equiv_type__replace_in_class_constraint_list(Constraints0, VarSet0, 
 				EqvMap, Constraints, VarSet1),
@@ -248,10 +253,11 @@ equiv_type__replace_in_class_interface(ClassInterface0, EqvMap,
 
 equiv_type__replace_in_class_method(EqvMap,
 			pred(TypeVarSet0, InstVarSet, ExistQVars, PredName,
-				TypesAndModes0, Det, Cond, ClassContext0,
-				Context),
+				TypesAndModes0, Det, Cond, Purity,
+				ClassContext0, Context),
 			pred(TypeVarSet, InstVarSet, ExistQVars, PredName,
-				TypesAndModes, Det, Cond, ClassContext, Context)
+				TypesAndModes, Det, Cond, Purity,
+				ClassContext, Context)
 			) :-
 	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
 				EqvMap, ClassContext, TypeVarSet1),
@@ -261,10 +267,10 @@ equiv_type__replace_in_class_method(EqvMap,
 equiv_type__replace_in_class_method(EqvMap,
 			func(TypeVarSet0, InstVarSet, ExistQVars, PredName,
 				TypesAndModes0, RetTypeAndMode0, Det, Cond,
-				ClassContext0, Context),
+				Purity, ClassContext0, Context),
 			func(TypeVarSet, InstVarSet, ExistQVars, PredName,
 				TypesAndModes, RetTypeAndMode, Det, Cond,
-				ClassContext, Context)
+				Purity, ClassContext, Context)
 			) :-
 	equiv_type__replace_in_class_constraints(ClassContext0, TypeVarSet0, 
 				EqvMap, ClassContext, TypeVarSet1),
