@@ -1,11 +1,11 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
 
 % This module defines predicates to produce the functions which are
-% exported to C via a `pragma export' declaration.
+% exported to a foreign language via a `pragma export' declaration.
 
 % Note: any changes here might also require similar changes to the handling
 % of `pragma import' declarations, which are handled in make_hlds.m.
@@ -18,31 +18,38 @@
 
 :- interface.
 
-:- import_module hlds_module, prog_data, llds.
+:- import_module prog_data, hlds_module, llds.
 :- import_module io.
 
-	% From the module_info, get a list of c_export_decls,
+	% From the module_info, get a list of foreign_export_decls,
 	% each of which holds information about the declaration
-	% of a C function named in a `pragma export' declaration,
+	% of a foreign function named in a `pragma export' declaration,
 	% which is used to allow a call to be made to a Mercury
-	% procedure from C.
-:- pred export__get_c_export_decls(module_info, c_export_decls).
-:- mode export__get_c_export_decls(in, out) is det.
+	% procedure from the foreign language.
+:- pred export__get_foreign_export_decls(module_info, foreign_export_decls).
+:- mode export__get_foreign_export_decls(in, out) is det.
 
-	% From the module_info, get a list of c_export_defns,
-	% each of which is a string containing the C code
-	% for defining a C function named in a `pragma export' decl.
-:- pred export__get_c_export_defns(module_info, c_export_defns).
-:- mode export__get_c_export_defns(in, out) is det.
+	% From the module_info, get a list of foreign_export_defns,
+	% each of which is a string containing the foreign code
+	% for defining a foreign function named in a `pragma export' decl.
+:- pred export__get_foreign_export_defns(module_info, foreign_export_defns).
+:- mode export__get_foreign_export_defns(in, out) is det.
 
-	% Produce a header file containing prototypes for the exported C
-	% functions
-:- pred export__produce_header_file(c_export_decls, module_name,
+	% Produce an interface file containing declarations for the
+	% exported foreign functions (if required in this foreign
+	% language).
+:- pred export__produce_header_file(foreign_export_decls, module_name,
 					io__state, io__state).
 :- mode export__produce_header_file(in, in, di, uo) is det.
 
-	% Convert the type, to a string corresponding to its C type.
-	% (Defaults to Word).
+
+%-----------------------------------------------------------------------------%
+
+% Utilities for generating C code which interfaces with Mercury.  
+% The {MLDS,LLDS}->C backends and fact tables use this code.
+
+	% Convert the type to a string corresponding to its C type.
+	% (Defaults to MR_Word).
 :- pred export__type_to_type_string(type, string).
 :- mode export__type_to_type_string(in, out) is det.
 
@@ -63,8 +70,11 @@
 
 :- implementation.
 
-:- import_module code_gen, code_util, hlds_pred, llds_out, modules.
-:- import_module type_util.
+:- import_module modules.
+:- import_module hlds_pred, type_util.
+:- import_module code_model.
+:- import_module code_gen, code_util, llds_out.
+:- import_module globals, options.
 
 :- import_module term, varset.
 :- import_module library, map, int, string, std_util, assoc_list, require.
@@ -72,30 +82,34 @@
 
 %-----------------------------------------------------------------------------%
 
-export__get_c_export_decls(HLDS, C_ExportDecls) :-
+export__get_foreign_export_decls(HLDS, C_ExportDecls) :-
 	module_info_get_predicate_table(HLDS, PredicateTable),
 	predicate_table_get_preds(PredicateTable, Preds),
 	module_info_get_pragma_exported_procs(HLDS, ExportedProcs),
-	export__get_c_export_decls_2(Preds, ExportedProcs, C_ExportDecls).
+	module_info_globals(HLDS, Globals),
+	export__get_foreign_export_decls_2(Preds, ExportedProcs, Globals,
+		C_ExportDecls).
 
-:- pred export__get_c_export_decls_2(pred_table, list(pragma_exported_proc),
-	list(c_export_decl)).
-:- mode export__get_c_export_decls_2(in, in, out) is det.
+:- pred export__get_foreign_export_decls_2(pred_table,
+	list(pragma_exported_proc), globals, list(foreign_export_decl)).
+:- mode export__get_foreign_export_decls_2(in, in, in, out) is det.
 
-export__get_c_export_decls_2(_Preds, [], []).
-export__get_c_export_decls_2(Preds, [E|ExportedProcs], C_ExportDecls) :-
-	E = pragma_exported_proc(PredId, ProcId, C_Function),
-	get_export_info(Preds, PredId, ProcId, _Exported, C_RetType,
-		_DeclareReturnVal, _FailureAction, _SuccessAction,
+export__get_foreign_export_decls_2(_Preds, [], _, []).
+export__get_foreign_export_decls_2(Preds, [E|ExportedProcs], Globals,
+		C_ExportDecls) :-
+	E = pragma_exported_proc(PredId, ProcId, C_Function, _Ctxt),
+	get_export_info(Preds, PredId, ProcId, Globals, _HowToDeclare,
+		C_RetType, _DeclareReturnVal, _FailureAction, _SuccessAction,
 		HeadArgInfoTypes),
 	get_argument_declarations(HeadArgInfoTypes, no, ArgDecls),
-	C_ExportDecl = c_export_decl(C_RetType, C_Function, ArgDecls),
-	export__get_c_export_decls_2(Preds, ExportedProcs, C_ExportDecls0),
+	C_ExportDecl = foreign_export_decl(c, C_RetType, C_Function, ArgDecls),
+	export__get_foreign_export_decls_2(Preds, ExportedProcs, Globals,
+		C_ExportDecls0),
 	C_ExportDecls = [C_ExportDecl | C_ExportDecls0].
 
 %-----------------------------------------------------------------------------%
 
-export__get_c_export_defns(Module, ExportedProcsCode) :-
+export__get_foreign_export_defns(Module, ExportedProcsCode) :-
 	module_info_get_pragma_exported_procs(Module, ExportedProcs),
 	module_info_get_predicate_table(Module, PredicateTable),
 	predicate_table_get_preds(PredicateTable, Preds),
@@ -104,54 +118,77 @@ export__get_c_export_defns(Module, ExportedProcsCode) :-
 	% For each exported procedure, produce a C function.
 	% The code we generate is in the form
 	%
-	% Declare_entry(<label of called proc>); /* or Declare_static */
+	% MR_declare_entry(<label of called proc>); /* or MR_declare_static */
 	%
 	% #if SEMIDET
 	%   bool
 	% #elif FUNCTION
-	%   Word
+	%   MR_Word
 	% #else
 	%   void
 	% #endif
-	% <function name>(Word Mercury__Argument1, Word *Mercury__Argument2...)
+	% <function name>(MR_Word Mercury__Argument1, 
+	%			MR_Word *Mercury__Argument2...)
 	%			/* Word for input, Word* for output */
 	% {
-	% #if NUM_REAL_REGS > 0
-	%	Word c_regs[NUM_REAL_REGS];
+	% #if MR_NUM_REAL_REGS > 0
+	%	MR_Word c_regs[MR_NUM_REAL_REGS];
 	% #endif
 	% #if FUNCTION
-	%	Word retval;
+	%	MR_Word retval;
 	% #endif
+	% #if MR_THREAD_SAFE
+	% 	MR_Bool must_finalize_engine;
+	% #endif 
 	%
 	%		/* save the registers that our C caller may be using */
-	%	save_regs_to_mem(c_regs);
+	%	MR_save_regs_to_mem(c_regs);
 	%
-	%		/* restore Mercury's registers that were saved as */
-	%		/* we entered C from Mercury (the process must    */
-	%		/* always start in Mercury so that we can 	  */
-	%		/* init_engine() etc.)				  */
-	%	restore_registers();
+	%		/* 
+	%		** start a new Mercury engine inside this POSIX 
+	%		** thread, if necessary (the C code may be 
+	%		** multi-threaded itself).
+	%		*/
+	%
+	% #if MR_THREAD_SAFE
+	% 	must_finalize_engine = MR_init_thread(MR_use_now);
+	% #endif 
+	%
+	%		/* 
+	%		** restore Mercury's registers that were saved as
+	%		** we entered C from Mercury.  For single threaded
+	%		** programs the process must always start in Mercury
+	%		** so that we can MR_init_engine() etc.  For
+	%		** multi-threaded MR_init_thread (above) takes care
+	%		** of making a new engine if required.
+	%		*/
+	%	MR_restore_registers();
 	%	<copy input arguments from Mercury__Arguments into registers>
 	%		/* save the registers which may be clobbered      */
 	%		/* by the C function call MR_call_engine().       */
-	%	save_transient_registers();
+	%	MR_save_transient_registers();
 	%
-	%	(void) MR_call_engine(ENTRY(<label of called proc>), FALSE);
+	%	(void) MR_call_engine(MR_ENTRY(<label of called proc>), FALSE);
 	%
 	%		/* restore the registers which may have been      */
 	%		/* clobbered by the return from the C function    */
 	%		/* MR_call_engine()				  */
-	%	restore_transient_registers();
+	%	MR_restore_transient_registers();
 	% #if SEMIDET
-	%	if (!r1) {
-	%		restore_regs_from_mem(c_regs);
+	%	if (!MR_r1) {
+	%		MR_restore_regs_from_mem(c_regs);
 	%		return FALSE;
 	%	}
 	% #elif FUNCTION
 	%	<copy return value register into retval>
 	% #endif
 	%	<copy output args from registers into *Mercury__Arguments>
-	%	restore_regs_from_mem(c_regs);
+	% #if MR_THREAD_SAFE
+	% 	if (must_finalize_engine) {
+	% 		MR_finalize_thread_engine();
+	% 	}
+	% #endif 
+	%	MR_restore_regs_from_mem(c_regs);
 	% #if SEMIDET
 	%	return TRUE;
 	% #elif FUNCTION
@@ -164,8 +201,9 @@ export__get_c_export_defns(Module, ExportedProcsCode) :-
 
 export__to_c(_Preds, [], _Module, []).
 export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
-	E = pragma_exported_proc(PredId, ProcId, C_Function),
-	get_export_info(Preds, PredId, ProcId, Exported,
+	E = pragma_exported_proc(PredId, ProcId, C_Function, _Ctxt),
+	module_info_globals(Module, Globals),
+	get_export_info(Preds, PredId, ProcId, Globals, DeclareString,
 		C_RetType, MaybeDeclareRetval, MaybeFail, MaybeSucceed,
 		ArgInfoTypes),
 	get_argument_declarations(ArgInfoTypes, yes, ArgDecls),
@@ -178,32 +216,37 @@ export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 	code_util__make_proc_label(Module, PredId, ProcId, ProcLabel),
 	llds_out__get_proc_label(ProcLabel, yes, ProcLabelString),
 
-	( Exported = yes ->
-		DeclareString = "Declare_entry"
-	;
-		DeclareString = "Declare_static"
-	),
-
 	string__append_list([	"\n",
 				DeclareString, "(", ProcLabelString, ");\n",
 				"\n",
 				C_RetType, "\n", 
 				C_Function, "(", ArgDecls, ")\n{\n",
-				"#if NUM_REAL_REGS > 0\n",
-				"\tWord c_regs[NUM_REAL_REGS];\n",
+				"#if MR_NUM_REAL_REGS > 0\n",
+				"\tMR_Word c_regs[MR_NUM_REAL_REGS];\n",
+				"#endif\n",
+				"#if MR_THREAD_SAFE\n",
+				"\tMR_Bool must_finalize_engine;\n", 
 				"#endif\n",
 				MaybeDeclareRetval,
 				"\n",
-				"\tsave_regs_to_mem(c_regs);\n", 
-				"\trestore_registers();\n", 
+				"\tMR_save_regs_to_mem(c_regs);\n", 
+				"#if MR_THREAD_SAFE\n",
+				"\tmust_finalize_engine = MR_init_thread(MR_use_now);\n", 
+				"#endif\n",
+				"\tMR_restore_registers();\n", 
 				InputArgs,
-				"\tsave_transient_registers();\n",
-				"\t(void) MR_call_engine(ENTRY(",
+				"\tMR_save_transient_registers();\n",
+				"\t(void) MR_call_engine(MR_ENTRY(",
 					ProcLabelString, "), FALSE);\n",
-				"\trestore_transient_registers();\n",
+				"\tMR_restore_transient_registers();\n",
 				MaybeFail,
 				OutputArgs,
-				"\trestore_regs_from_mem(c_regs);\n", 
+				"#if MR_THREAD_SAFE\n",
+				"\tif (must_finalize_engine) {\n", 
+				"\t\t MR_finalize_thread_engine();\n", 
+				"\t}\n", 
+				"#endif\n",
+				"\tMR_restore_regs_from_mem(c_regs);\n", 
 				MaybeSucceed,
 				"}\n\n"],
 				Code),
@@ -212,24 +255,37 @@ export__to_c(Preds, [E|ExportedProcs], Module, ExportedProcsCode) :-
 	ExportedProcsCode = [Code|TheRest].
 
 
-	% get_export_info(Preds, PredId, ProcId,
+	% get_export_info(Preds, PredId, ProcId, Globals, DeclareString,
 	%		C_RetType, MaybeDeclareRetval, MaybeFail, MaybeSuccess,
 	%		ArgInfoTypes):
-	%	Figure out the C return type, the actions on success
-	%	and failure, and the argument locations/modes/types
-	%	for a given procedure.
-:- pred get_export_info(pred_table, pred_id, proc_id, bool,
-			string, string, string, string,
-			assoc_list(arg_info, type)).
-:- mode get_export_info(in, in, in, out, out, out, out, out, out) is det.
+	%	For a given procedure, figure out the information about
+	%	that procedure that is needed to export it:
+	%	- how to declare the procedure's entry label,
+	%	- the C return type, and the C declaration for the variable
+	%	  holding the return value (if any),
+	%	- the actions on success and failure, and
+	%	- the argument locations/modes/types.
 
-get_export_info(Preds, PredId, ProcId, Exported, C_RetType,
+:- pred get_export_info(pred_table, pred_id, proc_id, globals,
+			string, string, string, string, string,
+			assoc_list(arg_info, type)).
+:- mode get_export_info(in, in, in, in, out, out, out, out, out, out) is det.
+
+get_export_info(Preds, PredId, ProcId, Globals, HowToDeclareLabel, C_RetType,
 		MaybeDeclareRetval, MaybeFail, MaybeSucceed, ArgInfoTypes) :-
 	map__lookup(Preds, PredId, PredInfo),
-	( procedure_is_exported(PredInfo, ProcId) ->
-		Exported = yes
+	pred_info_import_status(PredInfo, Status),
+	(
+		( procedure_is_exported(PredInfo, ProcId)
+		; status_defined_in_this_module(Status, no)
+		  % for --split-c-files, we need to treat
+		  % all procedures as if they were exported
+		; globals__lookup_bool_option(Globals, split_c_files, yes)
+		)
+	->
+		HowToDeclareLabel = "MR_declare_entry"
 	;
-		Exported = no
+		HowToDeclareLabel = "MR_declare_static"
 	),
 	pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
 	pred_info_procedures(PredInfo, ProcTable),
@@ -277,8 +333,8 @@ get_export_info(Preds, PredId, ProcId, Exported, C_RetType,
 		C_RetType = "bool",
 		MaybeDeclareRetval = "",
 		string__append_list([
-			"\tif (!r1) {\n",
-			"\t\trestore_regs_from_mem(c_regs);\n",
+			"\tif (!MR_r1) {\n",
+			"\t\tMR_restore_regs_from_mem(c_regs);\n",
 			"\treturn FALSE;\n",
 			"\t}\n"
 				], MaybeFail),
@@ -424,20 +480,20 @@ argloc_to_string(RegNum, RegName) :-
 			% XXX This magic number can't be good
 		RegNum > 32 
 	->
-		string__append_list(["r(", RegNumString, ")"], RegName)
+		string__append_list(["MR_r(", RegNumString, ")"], RegName)
 	;
-		string__append("r", RegNumString, RegName)
+		string__append("MR_r", RegNumString, RegName)
 	).
 
 convert_type_to_mercury(Rval, Type, ConvertedRval) :-	
 	(
         	Type = term__functor(term__atom("string"), [], _)
 	->
-		string__append("(Word) ", Rval, ConvertedRval)
+		string__append("(MR_Word) ", Rval, ConvertedRval)
 	;
         	Type = term__functor(term__atom("float"), [], _)
 	->
-		string__append_list(["float_to_word(", Rval, ")" ],
+		string__append_list(["MR_float_to_word(", Rval, ")" ],
 			ConvertedRval)
 	;
         	Type = term__functor(term__atom("character"), [], _)
@@ -454,11 +510,11 @@ convert_type_from_mercury(Rval, Type, ConvertedRval) :-
 	(
         	Type = term__functor(term__atom("string"), [], _)
 	->
-		string__append("(String) ", Rval, ConvertedRval)
+		string__append("(MR_String) ", Rval, ConvertedRval)
 	;
         	Type = term__functor(term__atom("float"), [], _)
 	->
-		string__append_list(["word_to_float(", Rval, ")" ],
+		string__append_list(["MR_word_to_float(", Rval, ")" ],
 			ConvertedRval)
 	;
 		ConvertedRval = Rval
@@ -480,7 +536,7 @@ export__produce_header_file(C_ExportDecls, ModuleName) -->
 		{ library__version(Version) },
 		io__write_strings(["/*\n** Automatically generated from `", 
 			SourceFileName,
-			".m' by the\n** Mercury compiler, version ", Version,
+			"' by the\n** Mercury compiler, version ", Version,
 			".  Do not edit.\n*/\n"]),
 		{ llds_out__sym_name_mangle(ModuleName, MangledModuleName) },
 		{ string__to_upper(MangledModuleName, UppercaseModuleName) },
@@ -516,36 +572,41 @@ export__produce_header_file(C_ExportDecls, ModuleName) -->
 		io__set_exit_status(1)
 	).
 
-:- pred export__produce_header_file_2(c_export_decls, io__state, io__state).
+:- pred export__produce_header_file_2(foreign_export_decls, 
+		io__state, io__state).
 :- mode export__produce_header_file_2(in, di, uo) is det.
 export__produce_header_file_2([]) --> [].
 export__produce_header_file_2([E|ExportedProcs]) -->
-	{ E = c_export_decl(C_RetType, C_Function, ArgDecls) },
-
-		% output the function header
-	io__write_string(C_RetType),
-	io__write_string(" "),
-	io__write_string(C_Function),
-	io__write_string("("),
-	io__write_string(ArgDecls),
-	io__write_string(");\n"),
-
+	{ E = foreign_export_decl(Lang, C_RetType, C_Function, ArgDecls) },
+	( 
+		{ Lang = c }
+	->
+			% output the function header
+		io__write_string(C_RetType),
+		io__write_string(" "),
+		io__write_string(C_Function),
+		io__write_string("("),
+		io__write_string(ArgDecls),
+		io__write_string(");\n")
+	;
+		{ error("export__produce_header_file_2: foreign languages other than C unimplemented") }
+	),
 	export__produce_header_file_2(ExportedProcs).
 
 	% Convert a term representation of a variable type to a string which
 	% represents the C type of the variable
-	% Apart from special cases, local variables become Words
+	% Apart from special cases, local variables become MR_Words
 export__type_to_type_string(Type, Result) :-
 	( Type = term__functor(term__atom("int"), [], _) ->
-		Result = "Integer"
+		Result = "MR_Integer"
 	; Type = term__functor(term__atom("float"), [], _) ->
-		Result = "Float"
+		Result = "MR_Float"
 	; Type = term__functor(term__atom("string"), [], _) ->
-		Result = "String"
+		Result = "MR_String"
 	; Type = term__functor(term__atom("character"), [], _) ->
-		Result = "Char"
+		Result = "MR_Char"
 	;
-		Result = "Word"
+		Result = "MR_Word"
 	).
 
 %-----------------------------------------------------------------------------%
