@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %------------------------------------------------------------------------------%
@@ -45,6 +45,9 @@
 :- mode exprn_aux__args_contain_rval(in, in) is semidet.
 :- mode exprn_aux__args_contain_rval(in, out) is nondet.
 
+:- pred exprn_aux__substitute_lval_in_lval(lval, lval, lval, lval).
+:- mode exprn_aux__substitute_lval_in_lval(in, in, in, out) is det.
+
 :- pred exprn_aux__substitute_lval_in_rval(lval, lval, rval, rval).
 :- mode exprn_aux__substitute_lval_in_rval(in, in, in, out) is det.
 
@@ -88,6 +91,10 @@
 	list(code_addr), list(data_addr)).
 :- mode exprn_aux__maybe_rval_list_addrs(in, out, out) is det.
 
+:- func exprn_aux__var_lval_to_rval(prog_var, lval) = rval.
+
+:- func exprn_aux__lval_to_rval(lval) = rval.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
@@ -160,7 +167,7 @@ exprn_aux__addr_is_constant(do_aditi_insert, _, no).
 exprn_aux__addr_is_constant(do_aditi_delete, _, no).
 exprn_aux__addr_is_constant(do_aditi_bulk_insert, _, no).
 exprn_aux__addr_is_constant(do_aditi_bulk_delete, _, no).
-exprn_aux__addr_is_constant(do_aditi_modify, _, no).
+exprn_aux__addr_is_constant(do_aditi_bulk_modify, _, no).
 exprn_aux__addr_is_constant(do_not_reached, _, no).
 
 :- pred exprn_aux__label_is_constant(label, bool, bool, bool).
@@ -198,8 +205,8 @@ exprn_aux__imported_is_constant(NonLocalGotos, AsmLabels, IsConst) :-
 
 exprn_aux__rval_contains_lval(lval(Lval0), Lval) :-
 	exprn_aux__lval_contains_lval(Lval0, Lval).
-exprn_aux__rval_contains_lval(create(_, Rvals, _, _, _, _), Lval) :-
-	exprn_aux__args_contain_lval(Rvals, Lval).
+exprn_aux__rval_contains_lval(create(_, Rvals, _, _, _, _, Reuse), Lval) :-
+	exprn_aux__args_contain_lval([Reuse | Rvals], Lval).
 exprn_aux__rval_contains_lval(mkword(_, Rval), Lval) :-
 	exprn_aux__rval_contains_lval(Rval, Lval).
 exprn_aux__rval_contains_lval(unop(_, Rval), Lval) :-
@@ -258,8 +265,8 @@ exprn_aux__rval_contains_rval(Rval0, Rval) :-
 			Rval0 = lval(Lval),
 			exprn_aux__lval_contains_rval(Lval, Rval)
 		;
-			Rval0 = create(_, Rvals, _, _, _, _),
-			exprn_aux__args_contain_rval(Rvals, Rval)
+			Rval0 = create(_, Rvals, _, _, _, _, Reuse),
+			exprn_aux__args_contain_rval([Reuse | Rvals], Rval)
 		;
 			Rval0 = mkword(_, Rval1),
 			exprn_aux__rval_contains_rval(Rval1, Rval)
@@ -300,8 +307,8 @@ exprn_aux__args_contain_rval([M | Ms], Rval) :-
 exprn_aux__vars_in_rval(lval(Lval), Vars) :-
 	exprn_aux__vars_in_lval(Lval, Vars).
 exprn_aux__vars_in_rval(var(Var), [Var]).
-exprn_aux__vars_in_rval(create(_, Rvals, _, _, _, _), Vars) :-
-	exprn_aux__vars_in_args(Rvals, Vars).
+exprn_aux__vars_in_rval(create(_, Rvals, _, _, _, _, Reuse), Vars) :-
+	exprn_aux__vars_in_args([Reuse | Rvals], Vars).
 exprn_aux__vars_in_rval(mkword(_, Rval), Vars) :-
 	exprn_aux__vars_in_rval(Rval, Vars).
 exprn_aux__vars_in_rval(const(_Conts), []).
@@ -376,10 +383,13 @@ exprn_aux__substitute_lval_in_rval(OldLval, NewLval, Rval0, Rval) :-
 		Rval0 = var(_Var),
 		Rval = Rval0
 	;
-		Rval0 = create(Tag, Rvals0, ArgTypes, StatDyn, Num, Msg),
+		Rval0 = create(Tag, Rvals0, ArgTypes, StatDyn,
+				Num, Msg, Reuse0),
 		exprn_aux__substitute_lval_in_args(OldLval, NewLval,
 			Rvals0, Rvals),
-		Rval = create(Tag, Rvals, ArgTypes, StatDyn, Num, Msg)
+		exprn_aux__substitute_lval_in_arg(OldLval, NewLval,
+			Reuse0, Reuse),
+		Rval = create(Tag, Rvals, ArgTypes, StatDyn, Num, Msg, Reuse)
 	;
 		Rval0 = mkword(Tag, Rval1),
 		exprn_aux__substitute_lval_in_rval(OldLval, NewLval,
@@ -424,9 +434,6 @@ exprn_aux__substitute_lval_in_mem_ref(OldLval, NewLval, MemRef0, MemRef) :-
 		MemRef = heap_ref(Rval, Tag, N)
 	).
 
-:- pred exprn_aux__substitute_lval_in_lval(lval, lval, lval, lval).
-:- mode exprn_aux__substitute_lval_in_lval(in, in, in, out) is det.
-
 exprn_aux__substitute_lval_in_lval(OldLval, NewLval, Lval0, Lval) :-
 	(
 		Lval0 = OldLval
@@ -516,6 +523,14 @@ exprn_aux__substitute_lval_in_lval_2(OldLval, NewLval, Lval0, Lval) :-
 
 exprn_aux__substitute_lval_in_args(_OldLval, _NewLval, [], []).
 exprn_aux__substitute_lval_in_args(OldLval, NewLval, [M0 | Ms0], [M | Ms]) :-
+	exprn_aux__substitute_lval_in_arg(OldLval, NewLval, M0, M),
+	exprn_aux__substitute_lval_in_args(OldLval, NewLval, Ms0, Ms).
+
+:- pred exprn_aux__substitute_lval_in_arg(lval, lval,
+				maybe(rval), maybe(rval)).
+:- mode exprn_aux__substitute_lval_in_arg(in, in, in, out) is det.
+
+exprn_aux__substitute_lval_in_arg(OldLval, NewLval, M0, M) :-
 	(
 		M0 = yes(Rval0)
 	->
@@ -524,8 +539,7 @@ exprn_aux__substitute_lval_in_args(OldLval, NewLval, [M0 | Ms0], [M | Ms]) :-
 		M = yes(Rval)
 	;
 		M = M0
-	),
-	exprn_aux__substitute_lval_in_args(OldLval, NewLval, Ms0, Ms).
+	).
 
 exprn_aux__substitute_rval_in_rval(OldRval, NewRval, Rval0, Rval) :-
 	(
@@ -542,10 +556,14 @@ exprn_aux__substitute_rval_in_rval(OldRval, NewRval, Rval0, Rval) :-
 			Rval0 = var(_),
 			Rval = Rval0
 		;
-			Rval0 = create(Tag, Rvals0, ATs, StatDyn, Num, Msg),
+			Rval0 = create(Tag, Rvals0, ATs, StatDyn,
+					Num, Msg, Reuse0),
 			exprn_aux__substitute_rval_in_args(OldRval, NewRval,
 				Rvals0, Rvals),
-			Rval = create(Tag, Rvals, ATs, StatDyn, Num, Msg)
+			exprn_aux__substitute_rval_in_arg(OldRval, NewRval,
+				Reuse0, Reuse),
+			Rval = create(Tag, Rvals, ATs, StatDyn,
+					Num, Msg, Reuse)
 		;
 			Rval0 = mkword(Tag, Rval1),
 			exprn_aux__substitute_rval_in_rval(OldRval, NewRval,
@@ -670,6 +688,14 @@ exprn_aux__substitute_rval_in_lval(OldRval, NewRval, Lval0, Lval) :-
 
 exprn_aux__substitute_rval_in_args(_OldRval, _NewRval, [], []).
 exprn_aux__substitute_rval_in_args(OldRval, NewRval, [M0 | Ms0], [M | Ms]) :-
+	exprn_aux__substitute_rval_in_arg(OldRval, NewRval, M0, M),
+	exprn_aux__substitute_rval_in_args(OldRval, NewRval, Ms0, Ms).
+
+:- pred exprn_aux__substitute_rval_in_arg(rval, rval,
+				maybe(rval), maybe(rval)).
+:- mode exprn_aux__substitute_rval_in_arg(in, in, in, out) is det.
+
+exprn_aux__substitute_rval_in_arg(OldRval, NewRval, M0, M) :-
 	(
 		M0 = yes(Rval0)
 	->
@@ -678,8 +704,7 @@ exprn_aux__substitute_rval_in_args(OldRval, NewRval, [M0 | Ms0], [M | Ms]) :-
 		M = yes(Rval)
 	;
 		M = M0
-	),
-	exprn_aux__substitute_rval_in_args(OldRval, NewRval, Ms0, Ms).
+	).
 
 %------------------------------------------------------------------------------%
 
@@ -738,7 +763,7 @@ exprn_aux__simplify_rval(Rval0, Rval) :-
 exprn_aux__simplify_rval_2(Rval0, Rval) :-
 	(
 		Rval0 = lval(field(MaybeTag, Base, Field)),
-		Base = create(Tag, Args, _, _, _, _),
+		Base = create(Tag, Args, _, _, _, _, _),
 		(
 			MaybeTag = yes(Tag)
 		;
@@ -753,11 +778,15 @@ exprn_aux__simplify_rval_2(Rval0, Rval) :-
 	->
 		Rval = lval(field(MaybeTag, Rval2, Num))
 	;
-		Rval0 = create(Tag, Args0, ArgTypes, StatDyn, CNum, Msg),
+		Rval0 = create(Tag, Args0, ArgTypes, StatDyn,
+				CNum, Msg, Reuse0),
 		exprn_aux__simplify_args(Args0, Args),
-		Args \= Args0
+		exprn_aux__simplify_arg(Reuse0, Reuse),
+		( Args \= Args0
+		; Reuse \= Reuse0
+		)
 	->
-		Rval = create(Tag, Args, ArgTypes, StatDyn, CNum, Msg)
+		Rval = create(Tag, Args, ArgTypes, StatDyn, CNum, Msg, Reuse)
 	;
 		Rval0 = unop(UnOp, Rval1),
 		exprn_aux__simplify_rval_2(Rval1, Rval2)
@@ -783,6 +812,12 @@ exprn_aux__simplify_rval_2(Rval0, Rval) :-
 exprn_aux__simplify_args([], []).
 exprn_aux__simplify_args([MR0 | Ms0], [MR | Ms]) :-
 	exprn_aux__simplify_args(Ms0, Ms),
+	exprn_aux__simplify_arg(MR0, MR).
+
+:- pred exprn_aux__simplify_arg(maybe(rval), maybe(rval)).
+:- mode exprn_aux__simplify_arg(in, out) is det.
+
+exprn_aux__simplify_arg(MR0, MR) :-
 	(
 		MR0 = yes(Rval0),
 		exprn_aux__simplify_rval_2(Rval0, Rval)
@@ -800,8 +835,10 @@ exprn_aux__simplify_args([MR0 | Ms0], [MR | Ms]) :-
 exprn_aux__rval_addrs(lval(Lval), CodeAddrs, DataAddrs) :-
 	exprn_aux__lval_addrs(Lval, CodeAddrs, DataAddrs).
 exprn_aux__rval_addrs(var(_), [], []).
-exprn_aux__rval_addrs(create(_, MaybeRvals, _,_,_,_), CodeAddrs, DataAddrs) :-
-	exprn_aux__maybe_rval_list_addrs(MaybeRvals, CodeAddrs, DataAddrs).
+exprn_aux__rval_addrs(create(_, MaybeRvals, _,_,_,_, Reuse),
+		CodeAddrs, DataAddrs) :-
+	exprn_aux__maybe_rval_list_addrs([Reuse | MaybeRvals],
+		CodeAddrs, DataAddrs).
 exprn_aux__rval_addrs(mkword(_Tag, Rval), CodeAddrs, DataAddrs) :-
 	exprn_aux__rval_addrs(Rval, CodeAddrs, DataAddrs).
 exprn_aux__rval_addrs(const(Const), CodeAddrs, DataAddrs) :-
@@ -901,5 +938,9 @@ exprn_aux__maybe_rval_list_addrs([MaybeRval | MaybeRvals],
 			CodeAddrs, DataAddrs)
 	).
 
+exprn_aux__var_lval_to_rval(_, Lval) = lval(Lval).
+
+exprn_aux__lval_to_rval(Lval) = lval(Lval).
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
