@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -45,7 +45,8 @@
 
 :- implementation.
 
-:- import_module options, globals, prog_io_util.
+:- import_module options, globals, prog_io_util, trace_params, unify_proc.
+:- import_module prog_data.
 :- import_module char, int, string, map, set, getopt, library.
 
 handle_options(MaybeError, Args, Link) -->
@@ -80,14 +81,20 @@ handle_options(MaybeError, Args, Link) -->
 			ConvertToGoedel),
 		globals__io_lookup_bool_option(typecheck_only, TypecheckOnly),
 		globals__io_lookup_bool_option(errorcheck_only, ErrorcheckOnly),
-		globals__io_lookup_bool_option(compile_to_c, CompileToC),
+		globals__io_lookup_bool_option(target_code_only,
+			TargetCodeOnly),
+		globals__io_get_target(Target),
+		{ GenerateIL = (if Target = il then yes else no) },
+		{ GenerateJava = (if Target = java then yes else no) },
 		globals__io_lookup_bool_option(compile_only, CompileOnly),
 		globals__io_lookup_bool_option(aditi_only, AditiOnly),
 		{ bool__or_list([GenerateDependencies, MakeInterface,
 			MakePrivateInterface, MakeShortInterface,
 			MakeOptimizationInt, MakeTransOptInt,
 			ConvertToMercury, ConvertToGoedel, TypecheckOnly,
-			ErrorcheckOnly, CompileToC, CompileOnly, AditiOnly],
+			ErrorcheckOnly, TargetCodeOnly,
+			GenerateIL, GenerateJava,
+			CompileOnly, AditiOnly],
 			NotLink) },
 		{ bool__not(NotLink, Link) }
 	).
@@ -114,94 +121,117 @@ dump_arguments([Arg | Args]) -->
 
 postprocess_options(error(ErrorMessage), yes(ErrorMessage)) --> [].
 postprocess_options(ok(OptionTable), Error) -->
-    { map__lookup(OptionTable, gc, GC_Method0) },
+    { map__lookup(OptionTable, target, Target0) },
     (
-        { GC_Method0 = string(GC_MethodStr) },
-        { convert_gc_method(GC_MethodStr, GC_Method) }
+        { Target0 = string(TargetStr) },
+        { convert_target(TargetStr, Target) }
     ->
-        { map__lookup(OptionTable, tags, TagsMethod0) },
+        { map__lookup(OptionTable, gc, GC_Method0) },
         (
-            { TagsMethod0 = string(TagsMethodStr) },
-            { convert_tags_method(TagsMethodStr, TagsMethod) }
+            { GC_Method0 = string(GC_MethodStr) },
+            { convert_gc_method(GC_MethodStr, GC_Method) }
         ->
-            { map__lookup(OptionTable, prolog_dialect, PrologDialect0) },
+            { map__lookup(OptionTable, tags, TagsMethod0) },
             (
-                { PrologDialect0 = string(PrologDialectStr) },
-                { convert_prolog_dialect(PrologDialectStr, PrologDialect) }
+                { TagsMethod0 = string(TagsMethodStr) },
+                { convert_tags_method(TagsMethodStr, TagsMethod) }
             ->
-                { map__lookup(OptionTable,
-                    fact_table_hash_percent_full, PercentFull) },
+                { map__lookup(OptionTable, prolog_dialect, PrologDialect0) },
                 (
-                    { PercentFull = int(Percent) },
-                    { Percent >= 1 },
-                    { Percent =< 100 }
+                    { PrologDialect0 = string(PrologDialectStr) },
+                    { convert_prolog_dialect(PrologDialectStr, PrologDialect) }
                 ->
-                    { map__lookup(OptionTable, termination_norm,
-                        TermNorm0) },
+                    { map__lookup(OptionTable,
+                        fact_table_hash_percent_full, PercentFull) },
                     (
-                        { TermNorm0 = string(TermNormStr) },
-                        { convert_termination_norm(TermNormStr, TermNorm) }
+                        { PercentFull = int(Percent) },
+                        { Percent >= 1 },
+                        { Percent =< 100 }
                     ->
-                        { map__lookup(OptionTable, trace, Trace) },
-                        { map__lookup(OptionTable, require_tracing,
-                            RequireTracingOpt) },
+                        { map__lookup(OptionTable, termination_norm,
+                            TermNorm0) },
                         (
-                            { Trace = string(TraceStr) },
-                            { RequireTracingOpt = bool(RequireTracing) },
-                            { convert_trace_level(TraceStr, RequireTracing,
-                                TraceLevel) }
+                            { TermNorm0 = string(TermNormStr) },
+                            { convert_termination_norm(TermNormStr, TermNorm) }
                         ->
-                            { map__lookup(OptionTable, dump_hlds_alias,
-                                DumpAliasOption) },
+                            { map__lookup(OptionTable, trace, Trace) },
+                            { map__lookup(OptionTable, require_tracing,
+                                RequireTracingOpt) },
                             (
-                                { DumpAliasOption = string(DumpAlias) },
-                                { DumpAlias = "" }
+                                { Trace = string(TraceStr) },
+                                { RequireTracingOpt = bool(RequireTracing) },
+                                { convert_trace_level(TraceStr, RequireTracing,
+                                    TraceLevel) }
                             ->
-                                postprocess_options_2(OptionTable,
-                                    GC_Method, TagsMethod, PrologDialect,
-				    TermNorm, TraceLevel, Error)
+                                { map__lookup(OptionTable, suppress_trace,
+                                    Suppress) },
+                                (
+                                    { Suppress = string(SuppressStr) },
+                                    { convert_trace_suppress(SuppressStr,
+                                        TraceSuppress) }
+                                ->
+                                    { map__lookup(OptionTable, dump_hlds_alias,
+                                        DumpAliasOption) },
+                                    (
+                                        { DumpAliasOption = string(DumpAlias) },
+                                        { DumpAlias = "" }
+                                    ->
+                                        postprocess_options_2(OptionTable,
+                                            Target, GC_Method, TagsMethod,
+                                            PrologDialect, TermNorm, TraceLevel,
+                                            TraceSuppress, Error)
+                                    ;
+                                        { DumpAliasOption = string(DumpAlias) },
+                                        { convert_dump_alias(DumpAlias,
+                                            DumpOptions) }
+                                    ->
+                                        { map__set(OptionTable,
+                                            dump_hlds_options,
+                                            string(DumpOptions),
+                                            NewOptionTable) },
+                                        postprocess_options_2(NewOptionTable,
+                                            Target, GC_Method, TagsMethod,
+                                            PrologDialect, TermNorm,
+                                            TraceLevel, TraceSuppress, Error)
+                                    ;
+                                        { Error = yes("Invalid argument to option `--hlds-dump-alias'.") }
+                                    )
+                                ;
+                                    { Error = yes("Invalid argument to option `--suppress-trace'.") }
+				)
                             ;
-                                { DumpAliasOption = string(DumpAlias) },
-                                { convert_dump_alias(DumpAlias,
-                                    DumpOptions) }
-                            ->
-                                { map__set(OptionTable, dump_hlds_options,
-                                    string(DumpOptions), NewOptionTable) },
-                                postprocess_options_2(NewOptionTable,
-                                    GC_Method, TagsMethod, PrologDialect,
-				    TermNorm, TraceLevel, Error)
-                            ;
-                                { Error = yes("Invalid argument to option `--hlds-dump-alias'.") }
+                                { Error = yes("Invalid argument to option `--trace'\n\t(must be `minimum', `shallow', `deep', or `default').") }
                             )
                         ;
-                            { Error = yes("Invalid argument to option `--trace'\n\t(must be `minimum', `shallow', `deep', or `default').") }
+                            { Error = yes("Invalid argument to option `--termination-norm'\n\t(must be `simple', `total' or  `num-data-elems').") }
                         )
                     ;
-                        { Error = yes("Invalid argument to option `--termination-norm'\n\t(must be `simple', `total' or  `num-data-elems').") }
+                        { Error = yes("Invalid argument to option `--fact-table-hash-percent-full'\n\t(must be an integer between 1 and 100)") }
                     )
                 ;
-                    { Error = yes("Invalid argument to option `--fact-table-hash-percent-full'\n\t(must be an integer between 1 and 100)") }
+                    { Error = yes("Invalid prolog-dialect option (must be `sicstus', `nu', or `default')") }
                 )
             ;
-                { Error = yes("Invalid prolog-dialect option (must be `sicstus', `nu', or `default')") }
+                { Error = yes("Invalid tags option (must be `none', `low' or `high')") }
             )
         ;
-            { Error = yes("Invalid tags option (must be `none', `low' or `high')") }
-        )
+            { Error = yes("Invalid GC option (must be `none', `conservative' or `accurate')") }
+	)
     ;
-        { Error = yes("Invalid GC option (must be `none', `conservative' or `accurate')") }
+        { Error = yes("Invalid target option (must be `c', `il', or `java')") }
     ).
+    
 
-:- pred postprocess_options_2(option_table, gc_method, tags_method,
-	prolog_dialect, termination_norm, trace_level, maybe(string),
-	io__state, io__state).
-:- mode postprocess_options_2(in, in, in, in, in, in, out, di, uo) is det.
+:- pred postprocess_options_2(option_table::in, compilation_target::in,
+    gc_method::in, tags_method::in, prolog_dialect::in, termination_norm::in,
+    trace_level::in, trace_suppress_items::in, maybe(string)::out,
+    io__state::di, io__state::uo) is det.
 
-postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
-		TermNorm, TraceLevel, Error) -->
+postprocess_options_2(OptionTable, Target, GC_Method, TagsMethod,
+		PrologDialect, TermNorm, TraceLevel, TraceSuppress, Error) -->
 	{ unsafe_promise_unique(OptionTable, OptionTable1) }, % XXX
-	globals__io_init(OptionTable1, GC_Method, TagsMethod,
-		PrologDialect, TermNorm, TraceLevel),
+	globals__io_init(OptionTable1, Target, GC_Method, TagsMethod,
+		PrologDialect, TermNorm, TraceLevel, TraceSuppress),
 
 	% --gc conservative implies --no-reclaim-heap-*
 	( { GC_Method = conservative } ->
@@ -250,9 +280,65 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 
 	globals__io_set_option(num_tag_bits, int(NumTagBits)),
 
-	option_implies(highlevel_c, gcc_non_local_gotos, bool(no)),
-	option_implies(highlevel_c, gcc_global_registers, bool(no)),
-	option_implies(highlevel_c, asm_labels, bool(no)),
+	% Generating IL implies high-level code, turning off nested functions,
+	% using copy-out for nondet output arguments,
+	% using zero tags, boxing enums, disabling no_tag_types and no
+	% static ground terms.
+	( { Target = il } ->
+		globals__io_set_option(highlevel_code, bool(yes)),
+		globals__io_set_option(gcc_nested_functions, bool(no)),
+		globals__io_set_option(nondet_copy_out, bool(yes)),
+		globals__io_set_option(num_tag_bits, int(0)),
+		globals__io_set_option(unboxed_enums, bool(no)),
+		globals__io_set_option(unboxed_no_tag_types, bool(no)),
+		globals__io_set_option(static_ground_terms, bool(no))
+	;
+		[]
+	),
+	% Generating Java implies high-level code, turning off nested functions,
+	% using copy-out for both det and nondet output arguments,
+	% using no tags, not optimizing tailcalls and no static ground terms.
+	% XXX no static ground terms should be eliminated in a later
+	%     version.
+	% XXX The Java backend should eventually support optimizing tailcalls.
+	( { Target = java } ->
+		globals__io_set_option(highlevel_code, bool(yes)),
+		globals__io_set_option(gcc_nested_functions, bool(no)),
+		globals__io_set_option(nondet_copy_out, bool(yes)),
+		globals__io_set_option(det_copy_out, bool(yes)),
+		globals__io_set_option(num_tag_bits, int(0)),
+		globals__io_set_option(optimize_tailcalls, bool(no)),
+		globals__io_set_option(static_ground_terms, bool(no))
+	;
+		[]
+	),
+	% Generating assembler via the gcc back-end requires
+	% using high-level code.
+	( { Target = asm } ->
+		globals__io_set_option(highlevel_code, bool(yes))
+	;
+		[]
+	),
+
+	% Generating high-level C or asm code requires putting each commit
+	% in its own function, to avoid problems with setjmp() and
+	% non-volatile local variables.
+	( { Target = c ; Target = asm } ->
+		option_implies(highlevel_code, put_commit_in_own_func, bool(yes))
+	;
+		[]
+	),
+	
+	% --high-level-code disables the use of low-level gcc extensions
+	option_implies(highlevel_code, gcc_non_local_gotos, bool(no)),
+	option_implies(highlevel_code, gcc_global_registers, bool(no)),
+	option_implies(highlevel_code, asm_labels, bool(no)),
+
+	% --no-gcc-nested-functions implies --no-gcc-local-labels
+	option_neg_implies(gcc_nested_functions, gcc_local_labels, bool(no)),
+
+	% --no-mlds-optimize implies --no-optimize-tailcalls
+	option_neg_implies(optimize, optimize_tailcalls, bool(no)),
 
 	option_implies(verbose_check_termination, check_termination,bool(yes)),
 	option_implies(check_termination, termination, bool(yes)),
@@ -274,6 +360,16 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 	% --split-c-files implies --procs-per-c-function 1
 	option_implies(split_c_files, procs_per_c_function, int(1)),
 
+	% make_hlds contains an optimization which requires the value of the
+	% compare_specialization option to accurately specify the max number
+	% of constructors in a type whose comparison procedure is specialized
+	% and which therefore don't need index functions.
+	globals__io_lookup_int_option(compare_specialization, CompareSpec0),
+	{ int__min(unify_proc__max_exploited_compare_spec_value,
+		CompareSpec0, CompareSpec) },
+	globals__io_set_option(compare_specialization, int(CompareSpec)),
+
+
 	% Minimal model tabling is not compatible with trailing;
 	% see the comment in runtime/mercury_tabling.c.
 
@@ -311,7 +407,7 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 	% 	- enabling stack layouts
 	% 	- enabling typeinfo liveness
 	globals__io_lookup_bool_option(trace_optimized, TraceOptimized),
-	( { TraceLevel \= none } ->
+	( { trace_level_is_none(TraceLevel) = no } ->
 		( { TraceOptimized = no } ->
 			% The following options modify the structure
 			% of the program, which makes it difficult to
@@ -334,6 +430,31 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 		;
 			[]
 		),
+
+			% Disable hijacks if debugging is enabled. The
+			% code we now use to restore the stacks for
+			% direct retries works only if the retry does not
+			% "backtrack" over a hijacked nondet stack frame
+			% whose hijack has not been undone. Note that
+			% code compiled without debugging may still hijack
+			% nondet stack frames. Execution may reemerge from
+			% the nondebugged region in one of two ways. If
+			% the nondebugged code returns, then it will have
+			% undone hijack, and the retry code will work. If
+			% the nondebugged code calls debugged code, there
+			% will be a region on the stacks containing no
+			% debugging information, and the retry command will
+			% refuse to perform retries that go into or beyond
+			% this region. Both cases preserve correctness.
+			%
+			% An alternative solution would be to store everything
+			% on the nondet stack that may be hijacked in ordinary
+			% stack slots on entry to every procedure, but that
+			% would be not only more complex than simply disabling
+			% hijacks, it would be slower as well, except in
+			% procedures that would have many nested hijacks,
+			% and such code is extremely rare.
+		globals__io_set_option(allow_hijacks, bool(no)),
 			% The following option prevents useless variables
 			% from cluttering the trace. Its explicit setting
 			% removes a source of variability in the goal paths
@@ -354,6 +475,9 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 			% we don't have to discover this fact anew
 			% for each procedure.
 		globals__io_set_option(optimize_value_number, bool(no)),
+			% Without value numbering, the eager code generator
+			% generates better code than the lazy code generator.
+		globals__io_set_option(lazy_code, bool(no)),
 			% The following options cause the info required
 			% by tracing to be generated.
 		globals__io_set_option(trace_stack_layout, bool(yes)),
@@ -483,7 +607,77 @@ postprocess_options_2(OptionTable, GC_Method, TagsMethod, PrologDialect,
 	% we are expecting some to be missing.
 	option_implies(use_opt_files, warn_missing_opt_files, bool(no)),
 
-	% --optimize-frames requires --optimize-labels and --optimize-jumps
+
+	% The preferred backend foreign language depends on the target.
+	( 	
+		{ Target = c },
+		{ BackendForeignLanguage = foreign_language_string(c) }
+	;
+		{ Target = il },
+		{ BackendForeignLanguage =
+			foreign_language_string(managed_cplusplus) }
+	;
+		{ Target = asm },
+		% XXX This is wrong!  It should be asm.
+		{ BackendForeignLanguage = foreign_language_string(c) }
+	;
+		% XXX We don't generate java or handle it as a foreign
+		% language just yet, but if we did, we should fix this
+		{ Target = java },
+		{ BackendForeignLanguage = foreign_language_string(c) }
+	),
+	globals__io_set_option(backend_foreign_language,
+		string(BackendForeignLanguage)),
+	% The default foreign language we use is the same as the backend.
+	globals__io_lookup_string_option(use_foreign_language,
+		UseForeignLanguage),
+	( 
+		{ UseForeignLanguage = "" }
+	->
+		globals__io_set_option(use_foreign_language, 
+			string(BackendForeignLanguage))
+	; 
+		{ convert_foreign_language(UseForeignLanguage, FL) }
+	->
+		{ CanonicalLangName = foreign_language_string(FL) },
+		globals__io_set_option(use_foreign_language, 
+			string(CanonicalLangName))
+	;
+		usage_error(
+			string__format(
+			"unrecognized foreign language argument `%s' for --use-foreign-language",
+			[s(UseForeignLanguage)]))
+	),
+
+	globals__io_lookup_bool_option(highlevel_code, HighLevel),
+	( { HighLevel = no } ->
+		postprocess_options_lowlevel
+	;
+		[]
+	).
+
+	% These option implications only affect the low-level (LLDS) code
+	% generator.  They may in fact be harmful if set for the high-level
+	% code generator, because sometimes the same option has different
+	% meanings and implications in the two backends.
+	%
+:- pred postprocess_options_lowlevel(io__state::di, io__state::uo) is det.
+
+postprocess_options_lowlevel -->
+		% --no-lazy-code assumes that const(_) rvals are really
+		% constant, and that create(_) rvals with constant arguments
+		% can be materialized in an assignable rval without further
+		% code. For float_consts, the former is true only if either
+		% static_ground_terms or unboxed_floats is true, and the latter
+		% cannot be true without static_ground_terms.
+	option_neg_implies(lazy_code, static_ground_terms, bool(yes)),
+
+		% --no-lazy-code requires --follow-vars for acceptable
+		% performance.
+	option_neg_implies(lazy_code, follow_vars, bool(yes)),
+
+		% --optimize-frames requires --optimize-labels and
+		% --optimize-jumps
 	option_implies(optimize_frames, optimize_labels, bool(yes)),
 	option_implies(optimize_frames, optimize_jumps, bool(yes)).
 
@@ -531,7 +725,7 @@ usage -->
 	{ library__version(Version) },
  	io__write_strings(StdErr, [
 		"Mercury Compiler, version ", Version, "\n",
-		"Copyright (C) 1993-1999 The University of Melbourne\n",
+		"Copyright (C) 1993-2000 The University of Melbourne\n",
 		"Usage: mmc [<options>] <arguments>\n",
 		"Use `mmc --help' for more information.\n"
 	]).
@@ -539,7 +733,7 @@ usage -->
 long_usage -->
 	{ library__version(Version) },
  	io__write_strings(["Mercury Compiler, version ", Version, "\n"]),
- 	io__write_string("Copyright (C) 1993-1999 The University of Melbourne\n"),
+ 	io__write_string("Copyright (C) 1993-2000 The University of Melbourne\n"),
 	io__write_string("Usage: mmc [<options>] <arguments>\n"),
 	io__write_string("Arguments:\n"),
 	io__write_string("\tArguments ending in `.m' are assumed to be source file names.\n"),
@@ -571,7 +765,8 @@ long_usage -->
 	% actually matters is for constructing the pathname for the
 	% grade of the library, etc for linking (and installation).
 :- type grade_component
-	--->	gcc_ext		% gcc extensions -- see grade_component_table
+	--->	gcc_ext		% gcc extensions etc. -- see
+				% grade_component_table
 	;	par		% parallelism / multithreading
 	;	gc		% the kind of GC to use
 	;	prof		% what profiling options to use
@@ -659,19 +854,117 @@ compute_grade_components(Options, GradeComponents) :-
 :- mode grade_component_table(out, in, out) is multi.
 :- mode grade_component_table(out, out, out) is multi.
 
-	% GCC-hack components
-grade_component_table("none", gcc_ext, [asm_labels - bool(no),
-	gcc_non_local_gotos - bool(no), gcc_global_registers - bool(no)]).
-grade_component_table("reg", gcc_ext, [asm_labels - bool(no),
-	gcc_non_local_gotos - bool(no), gcc_global_registers - bool(yes)]).
-grade_component_table("jump", gcc_ext, [asm_labels - bool(no),
-	gcc_non_local_gotos - bool(yes), gcc_global_registers - bool(no)]).
-grade_component_table("asm_jump", gcc_ext, [asm_labels - bool(yes),
-	gcc_non_local_gotos - bool(yes), gcc_global_registers - bool(no)]).
-grade_component_table("fast", gcc_ext, [asm_labels - bool(no),
-	gcc_non_local_gotos - bool(yes), gcc_global_registers - bool(yes)]).
-grade_component_table("asm_fast", gcc_ext, [asm_labels - bool(yes),
-	gcc_non_local_gotos - bool(yes), gcc_global_registers - bool(yes)]).
+	% Base components
+	% These specify the basic compilation model we use,
+	% including the choice of back-end and the use of gcc extensions.
+grade_component_table("none", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("reg", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(yes),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("jump", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(yes),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("asm_jump", gcc_ext, [
+		asm_labels		- bool(yes),
+		gcc_non_local_gotos	- bool(yes),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("fast", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(yes),
+		gcc_global_registers	- bool(yes),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("asm_fast", gcc_ext, [
+		asm_labels		- bool(yes),
+		gcc_non_local_gotos	- bool(yes),
+		gcc_global_registers	- bool(yes),
+		highlevel_code		- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("c")]).
+grade_component_table("hl", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(yes)
+		% target can be either c or asm
+		]).
+grade_component_table("hlc", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no)
+		% target can be either c or asm
+		]).
+grade_component_table("hl_nest", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(yes),
+		highlevel_data		- bool(yes)
+		% target can be either c or asm
+		]).
+grade_component_table("hlc_nest", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(yes),
+		highlevel_data		- bool(no)
+		% target can be either c or asm
+		]).
+grade_component_table("il", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(yes),
+		target			- string("il")]).
+grade_component_table("ilc", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		highlevel_code		- bool(yes),
+		gcc_nested_functions	- bool(no),
+		highlevel_data		- bool(no),
+		target			- string("il")]).
+grade_component_table("java", gcc_ext, [
+		asm_labels		- bool(no),
+		gcc_non_local_gotos	- bool(no),
+		gcc_global_registers	- bool(no),
+		gcc_nested_functions	- bool(no),
+		highlevel_code		- bool(yes),
+		highlevel_data		- bool(yes),
+		target			- string("java")]).
 
 	% Parallelism/multithreading components.
 grade_component_table("par", par, [parallel - bool(yes)]).
@@ -733,6 +1026,9 @@ reset_grade_options(Options0, Options) :-
 grade_start_values(asm_labels - bool(no)).
 grade_start_values(gcc_non_local_gotos - bool(no)).
 grade_start_values(gcc_global_registers - bool(no)).
+grade_start_values(highlevel_code - bool(no)).
+grade_start_values(highlevel_data - bool(no)).
+grade_start_values(gcc_nested_functions - bool(no)).
 grade_start_values(parallel - bool(no)).
 grade_start_values(gc - string("none")).
 grade_start_values(profile_deep - bool(no)).
