@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,7 +13,7 @@
 
 :- interface.
 
-:- import_module hlds_pred, llds, prog_data, (inst), term.
+:- import_module hlds_pred, prog_data, (inst), term.
 :- import_module bool, list, map, std_util.
 
 %-----------------------------------------------------------------------------%
@@ -26,6 +26,9 @@
 :- type cons_table	==	map(cons_id, list(hlds_cons_defn)).
 
 :- type cons_id		--->	cons(sym_name, arity)	% name, arity
+				% Tuples have cons_id
+				% `cons(unqualified("{}"), Arity)'.
+
 			;	int_const(int)
 			;	string_const(string)
 			;	float_const(float)
@@ -62,8 +65,9 @@
 			% you can get the tvarset from the hlds__type_defn.
 			existq_tvars,		% existential type vars
 			list(class_constraint), % existential class constraints
-			list(type),		% The types of the arguments
-						% of this functor (if any)
+			list(constructor_arg),	% The field names and types of
+						% the arguments of this functor
+						% (if any)
 			type_id,		% The result type, i.e. the
 						% type to which this
 						% cons_defn belongs.
@@ -72,6 +76,36 @@
 						% original source code
 		).
 
+%-----------------------------------------------------------------------------%
+
+:- type ctor_field_table == map(ctor_field_name, list(hlds_ctor_field_defn)).
+
+:- type hlds_ctor_field_defn
+	---> hlds_ctor_field_defn(
+		prog_context,	% context of the field definition
+		import_status,
+		type_id,	% type containing the field
+		cons_id,	% constructor containing the field
+		int		% argument number (counting from 1)
+	).
+
+	%
+	% Field accesses are expanded into inline unifications by
+	% post_typecheck.m after typechecking has worked out which 
+	% field is being referred to.
+	%
+	% Function declarations and clauses are not generated for these
+	% because it would be difficult to work out how to mode them.
+	%
+	% Users can supply type and mode declarations, for example
+	% to export a field of an abstract data type or to allow
+	% taking the address of a field access function.
+	%
+:- type field_access_type
+	--->	get
+	;	set
+	.
+
 %-----------------------------------------------------------------------------%
 
 	% Various predicates for accessing the cons_id type.
@@ -102,6 +136,18 @@
 :- pred make_cons_id(sym_name, list(constructor_arg), type_id, cons_id).
 :- mode make_cons_id(in, in, in, out) is det.
 
+	% Another way of making a cons_id from a functor.
+	% Given the name, argument types, and type_id of a functor,
+	% create a cons_id for that functor.
+	%
+	% Differs from make_cons_id in that (a) it requires the sym_name
+	% to be already module qualified, which means that it does not
+	% need the module qualification of the type, (b) it can compute the
+	% arity from any list of the right length.
+
+:- pred make_cons_id_from_qualified_sym_name(sym_name, list(_), cons_id).
+:- mode make_cons_id_from_qualified_sym_name(in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 :- implementation.
@@ -162,6 +208,9 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 	),
 	list__length(Args, Arity).
 
+make_cons_id_from_qualified_sym_name(SymName, Args, cons(SymName, Arity)) :-
+	list__length(Args, Arity).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -206,6 +255,10 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 			hlds_type_defn).
 :- mode hlds_data__set_type_defn_status(in, in, out) is det.
 
+:- pred hlds_data__set_type_defn_body(hlds_type_defn, hlds_type_body,
+			hlds_type_defn).
+:- mode hlds_data__set_type_defn_body(in, in, out) is det.
+
 	% An `hlds_type_body' holds the body of a type definition:
 	% du = discriminated union, uu = undiscriminated union,
 	% eqv_type = equivalence type (a type defined to be equivalent
@@ -236,13 +289,13 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 
 :- type cons_tag
 	--->	string_constant(string)
-			% Strings are represented using the string_const()
+			% Strings are represented using the MR_string_const()
 			% macro; in the current implementation, Mercury
 			% strings are represented just as C null-terminated
 			% strings.
 	;	float_constant(float)
-			% Floats are represented using the float_to_word(),
-			% word_to_float(), and float_const() macros.
+			% Floats are represented using the MR_float_to_word(),
+			% MR_word_to_float(), and MR_float_const() macros.
 			% The default implementation of these is to
 			% use boxed double-precision floats.
 	;	int_constant(int)
@@ -317,6 +370,21 @@ make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
 
 :- type tag_bits	==	int.	% actually only 2 (or maybe 3) bits
 
+
+	% The type definitions for no_tag types have information
+	% mirrored in a separate table for faster lookups.
+	% mode_util__mode_to_arg_mode makes heavy use of
+	% type_util__type_is_no_tag_type.
+:- type no_tag_type
+	--->	no_tag_type(
+			list(type_param),	% Formal type parameters.
+			sym_name,		% Constructor name.
+			(type)			% Argument type.
+		).
+
+:- type no_tag_type_table == map(type_id, no_tag_type).
+
+
 :- implementation.
 
 :- type hlds_type_defn
@@ -352,6 +420,8 @@ hlds_data__get_type_defn_body(hlds_type_defn(_, _, Body, _, _), Body).
 hlds_data__get_type_defn_status(hlds_type_defn(_, _, _, Status, _), Status).
 hlds_data__get_type_defn_context(hlds_type_defn(_, _, _, _, Context), Context).
 
+hlds_data__set_type_defn_body(hlds_type_defn(A, B, _, D, E), Body,
+				hlds_type_defn(A, B, Body, D, E)).
 hlds_data__set_type_defn_status(hlds_type_defn(A, B, C, _, E), Status, 
 				hlds_type_defn(A, B, C, Status, E)).
 
@@ -399,7 +469,7 @@ hlds_data__set_type_defn_status(hlds_type_defn(A, B, C, _, E), Status,
 	--->	hlds_inst_defn(
 			inst_varset,		% The names of the inst
 						% parameters (if any).
-			list(inst_param),	% The inst parameters (if any).
+			list(inst_var),		% The inst parameters (if any).
 						% ([I] in the above example.)
 			hlds_inst_body,	% The definition of this inst.
 			condition,		% Unused (reserved for
@@ -596,7 +666,7 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 	--->	hlds_mode_defn(
 			inst_varset,		% The names of the inst
 						% parameters (if any).
-			list(inst_param),	% The list of the inst
+			list(inst_var),		% The list of the inst
 						% parameters (if any).
 						% (e.g. [I] for the second
 						% example above.)
@@ -672,6 +742,7 @@ mode_table_optimize(mode_table(ModeDefns0, ModeIds0),
 
 %
 % Types and procedures for decomposing and analysing determinism.
+% See also the `code_model' type in code_model.m.
 % The `determinism' type itself is defined in prog_data.m.
 %
 
@@ -693,10 +764,6 @@ mode_table_optimize(mode_table(ModeDefns0, ModeIds0),
 :- mode determinism_components(in, out, out) is det.
 :- mode determinism_components(out, in, in) is det.
 
-:- pred determinism_to_code_model(determinism, code_model).
-:- mode determinism_to_code_model(in, out) is det.
-:- mode determinism_to_code_model(out, in) is multidet.
-
 :- implementation.
 
 determinism_components(det,         cannot_fail, at_most_one).
@@ -708,15 +775,6 @@ determinism_components(cc_nondet,   can_fail,    at_most_many_cc).
 determinism_components(erroneous,   cannot_fail, at_most_zero).
 determinism_components(failure,     can_fail,    at_most_zero).
 
-determinism_to_code_model(det,         model_det).
-determinism_to_code_model(semidet,     model_semi).
-determinism_to_code_model(nondet,      model_non).
-determinism_to_code_model(multidet,    model_non).
-determinism_to_code_model(cc_nondet,   model_semi).
-determinism_to_code_model(cc_multidet, model_det).
-determinism_to_code_model(erroneous,   model_det).
-determinism_to_code_model(failure,     model_semi).
-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -757,6 +815,7 @@ determinism_to_code_model(failure,     model_semi).
 	% Information about a single `instance' declaration
 :- type hlds_instance_defn 
 	--->	hlds_instance_defn(
+			module_name,		% module of the instance decl
 			import_status,		% import status of the instance
 						% declaration
 			prog_context,		% context of declaration
