@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -36,7 +36,7 @@
 	% procedure may inlined while others are not.
 	%
 	% It builds the call-graph (if necessary) works from the bottom of
-	% the call-graph towards the top, first perfoming inlining on a
+	% the call-graph towards the top, first performing inlining on a
 	% procedure then deciding if calls to it (higher in the call-graph)
 	% should be inlined. SCCs get flattend and processed in the order
 	% returned by hlds_dependency_info_get_dependency_ordering.
@@ -58,7 +58,7 @@
 	%	  inlining them enables the code generator to avoid creating
 	%	  the intermediate structures which is often a win).
 	%
-	%	- arithmetic predicates where the as above, the cost of the
+	%	- arithmetic predicates where as above, the cost of the
 	%	  call will often outweigh the cost of the arithmetic.
 	%
 	%	- det or semi pragma C code, where often the C operation is
@@ -75,8 +75,9 @@
 	% procedure with so many variables that the back end of the compiler
 	% gets bogged down (for example in the pseudoknot benchmark).
 	%
-	% Due to the way in which we generate code for model_non pragma_c_code,
-	% procedures whose body is such a pragma_c_code must NOT be inlined.
+	% Due to the way in which we generate code for model_non
+	% pragma_foreign_code, procedures whose body is such a
+	% pragma_foreign_code must NOT be inlined.
 
 %-----------------------------------------------------------------------------%
 
@@ -137,20 +138,33 @@
 
 :- implementation.
 
-:- import_module globals, options, llds.
-:- import_module term, varset.
-:- import_module dead_proc_elim, type_util, mode_util, goal_util.
-:- import_module passes_aux, code_aux, quantification, det_analysis, prog_data.
+% Parse tree modules
+:- import_module prog_data.
 
-:- import_module bool, int, list, assoc_list, set, std_util.
-:- import_module require, hlds_data, dependency_graph.
+% HLDS modules
+:- import_module hlds_data, type_util, mode_util, goal_util, det_analysis.
+:- import_module quantification, code_aux, dead_proc_elim, dependency_graph.
+:- import_module passes_aux.
+
+% Misc
+:- import_module globals, options.
+
+% Standard library modules
+:- import_module bool, int, list, assoc_list, set, std_util, require.
+:- import_module term, varset.
 
 %-----------------------------------------------------------------------------%
 
-:- type inline_params	--->	params(bool, bool, int, int, int).
-				% simple, single_use,
-				%	size-threshold, simple-goal-threshold
-				%		var-threshold
+	% this structure holds option values, extracted from the globals
+:- type inline_params
+	--->	params(
+			simple				:: bool,
+			single_use			:: bool,
+			size_threshold			:: int,
+			simple_goal_threshold		:: int,
+			var_threshold			:: int,
+			highlevel_code			:: bool
+		).
 
 inlining(ModuleInfo0, ModuleInfo) -->
 		%
@@ -166,6 +180,7 @@ inlining(ModuleInfo0, ModuleInfo) -->
 		%   we want in procedures - if inlining a procedure
 		%   would cause the number of variables to exceed
 		%   this threshold then we don't inline it.
+		% - whether we're in an MLDS grade
 		%
 	globals__io_lookup_bool_option(inline_simple, Simple),
 	globals__io_lookup_bool_option(inline_single_use, SingleUse),
@@ -173,8 +188,9 @@ inlining(ModuleInfo0, ModuleInfo) -->
 							CompoundThreshold),
 	globals__io_lookup_int_option(inline_simple_threshold, SimpleThreshold),
 	globals__io_lookup_int_option(inline_vars_threshold, VarThreshold),
+	globals__io_lookup_bool_option(highlevel_code, HighLevelCode),
 	{ Params = params(Simple, SingleUse, CompoundThreshold,
-			SimpleThreshold, VarThreshold) },
+			SimpleThreshold, VarThreshold, HighLevelCode) },
 
 		%
 		% Get the usage counts for predicates
@@ -220,7 +236,7 @@ inlining(ModuleInfo0, ModuleInfo) -->
 :- mode inlining__do_inlining(in, in, in, in, in, out, di, uo) is det.
 
 inlining__do_inlining([], _Needed, _Params, _Inlined, Module, Module) --> [].
-inlining__do_inlining([PPId|PPIds], Needed, Params, Inlined0,
+inlining__do_inlining([PPId | PPIds], Needed, Params, Inlined0,
 		Module0, Module) -->
 	inlining__in_predproc(PPId, Inlined0, Params, Module0, Module1),
 	inlining__mark_predproc(PPId, Needed, Params, Module1,
@@ -232,11 +248,17 @@ inlining__do_inlining([PPId|PPIds], Needed, Params, Inlined0,
 		io__state, io__state).
 :- mode inlining__mark_predproc(in, in, in, in, in, out, di, uo) is det.
 
+%
+% This predicate effectively adds implicit `pragma inline'
+% directives for procedures that match its heuristic.
+%
+
 inlining__mark_predproc(PredProcId, NeededMap, Params, ModuleInfo, 
 		InlinedProcs0, InlinedProcs) -->
 	(
 		{ Params = params(Simple, SingleUse, CompoundThreshold,
-				SimpleThreshold, _VarThreshold) },
+				SimpleThreshold, _VarThreshold,
+				_HighLevelCode) },
 		{ PredProcId = proc(PredId, ProcId) },
 		{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
 		{ pred_info_procedures(PredInfo, Procs) },
@@ -265,18 +287,8 @@ inlining__mark_predproc(PredProcId, NeededMap, Params, ModuleInfo,
 			{ NumUses = 1 }
 		),
 		% Don't inline recursive predicates
-		{ \+ goal_calls(CalledGoal, PredProcId) },
-
-		% Under no circumstances inline model_non pragma c codes.
-		% The resulting code would not work properly.
-		\+ {
-			CalledGoal = pragma_c_code(_,_,_,_,_,_,_) - _,
-			proc_info_interface_code_model(ProcInfo, model_non)
-		},
-
-		% Don't inline memoed Aditi predicates.
-		{ pred_info_get_markers(PredInfo, Markers) },
-		{ \+ check_marker(Markers, aditi_memo) }
+		% (unless explicitly requested)
+		{ \+ goal_calls(CalledGoal, PredProcId) }
 	->
 		inlining__mark_proc_as_inlined(PredProcId, ModuleInfo,
 			InlinedProcs0, InlinedProcs)
@@ -364,6 +376,7 @@ inlining__mark_proc_as_inlined(proc(PredId, ProcId), ModuleInfo,
 :- type inline_info	
 	---> inline_info(
 		int,			% variable threshold for inlining
+		bool,			% highlevel_code option
 		set(pred_proc_id),	% inlined procs
 		module_info,		% module_info
 		list(tvar),		% universally quantified type vars
@@ -382,6 +395,9 @@ inlining__mark_proc_as_inlined(proc(PredId, ProcId), ModuleInfo,
 					% type variables to variables
 					% where their type_info is
 					% stored.
+		bool,			% Did we do any inlining in the proc?
+		bool,			% Does the goal need to be
+					% requantified?
 		bool			% Did we change the determinism
 					% of any subgoal?
 	).
@@ -392,8 +408,9 @@ inlining__mark_proc_as_inlined(proc(PredId, ProcId), ModuleInfo,
 
 inlining__in_predproc(PredProcId, InlinedProcs, Params,
 		ModuleInfo0, ModuleInfo, IoState0, IoState) :-
-	Params = params(_Simple, _SingleUse, _CompoundThreshold,
-			_SimpleThreshold, VarThresh),
+	VarThresh = Params^var_threshold,
+	HighLevelCode = Params^highlevel_code,
+
 	PredProcId = proc(PredId, ProcId),
 
 	module_info_preds(ModuleInfo0, PredTable0),
@@ -410,40 +427,63 @@ inlining__in_predproc(PredProcId, InlinedProcs, Params,
 	proc_info_vartypes(ProcInfo0, VarTypes0),
 	proc_info_typeinfo_varmap(ProcInfo0, TypeInfoVarMap0),
 
+	DidInlining0 = no,
+	Requantify0 = no,
 	DetChanged0 = no,
 
-	InlineInfo0 = inline_info(
-		VarThresh, InlinedProcs, ModuleInfo0, UnivQTVars, Markers,
-		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0, DetChanged0),
+	InlineInfo0 = inline_info(VarThresh, HighLevelCode,
+		InlinedProcs, ModuleInfo0, UnivQTVars, Markers,
+		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0,
+		DidInlining0, Requantify0, DetChanged0),
 
 	inlining__inlining_in_goal(Goal0, Goal, InlineInfo0, InlineInfo),
 
-	InlineInfo = inline_info(_, _, _, _, _, VarSet, VarTypes, TypeVarSet, 
-		TypeInfoVarMap, DetChanged),
+	InlineInfo = inline_info(_, _, _, _, _, _, VarSet, VarTypes,
+		TypeVarSet, TypeInfoVarMap, DidInlining, Requantify,
+		DetChanged),
 
 	pred_info_set_typevarset(PredInfo0, TypeVarSet, PredInfo1),
 
 	proc_info_set_varset(ProcInfo0, VarSet, ProcInfo1),
 	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
 	proc_info_set_typeinfo_varmap(ProcInfo2, TypeInfoVarMap, ProcInfo3),
-	proc_info_set_goal(ProcInfo3, Goal, ProcInfo),
+	proc_info_set_goal(ProcInfo3, Goal, ProcInfo4),
+
+	(
+		DidInlining = yes,
+		recompute_instmap_delta_proc(yes, ProcInfo4, ProcInfo5,
+			ModuleInfo0, ModuleInfo1)
+	;
+		DidInlining = no,
+		ProcInfo5 = ProcInfo4,
+		ModuleInfo1 = ModuleInfo0
+	),
+
+	globals__io_get_globals(Globals, IoState0, IoState),
+	(
+		Requantify = yes,
+		requantify_proc(ProcInfo5, ProcInfo)
+	;
+		Requantify = no,
+		ProcInfo = ProcInfo5
+	),
 
 	map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
 	pred_info_set_procedures(PredInfo1, ProcTable, PredInfo),
 	map__det_update(PredTable0, PredId, PredInfo, PredTable),
-	module_info_set_preds(ModuleInfo0, PredTable, ModuleInfo1),
+	module_info_set_preds(ModuleInfo1, PredTable, ModuleInfo2),
 
 		% If the determinism of some sub-goals has changed,
 		% then we re-run determinism analysis, because
 		% propagating the determinism information through
 		% the procedure may lead to more efficient code.
-	( DetChanged = yes,	
-		globals__io_get_globals(Globals, IoState0, IoState),
-		det_infer_proc(PredId, ProcId, ModuleInfo1, ModuleInfo,
+	(
+		DetChanged = yes,	
+		det_infer_proc(PredId, ProcId, ModuleInfo2, ModuleInfo,
 			Globals, _, _, _)
-	; DetChanged = no,
-		ModuleInfo = ModuleInfo1,
-		IoState = IoState0
+	;
+		DetChanged = no,
+		ModuleInfo = ModuleInfo2
 	).
 
 %-----------------------------------------------------------------------------%
@@ -484,14 +524,15 @@ inlining__inlining_in_goal(some(Vars, CanRemove, Goal0) - GoalInfo,
 inlining__inlining_in_goal(call(PredId, ProcId, ArgVars, Builtin, Context,
 		Sym) - GoalInfo0, Goal - GoalInfo, InlineInfo0, InlineInfo) :-
 
-	InlineInfo0 = inline_info(VarThresh, InlinedProcs, ModuleInfo,
-		HeadTypeParams, Markers,
-		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0, DetChanged0),
+	InlineInfo0 = inline_info(VarThresh, HighLevelCode,
+		InlinedProcs, ModuleInfo, HeadTypeParams, Markers,
+		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0,
+		DidInlining0, Requantify0, DetChanged0),
 
 	% should we inline this call?
 	(
 		inlining__should_inline_proc(PredId, ProcId, Builtin,
-				InlinedProcs, Markers, ModuleInfo),
+			HighLevelCode, InlinedProcs, Markers, ModuleInfo),
 			% okay, but will we exceed the number-of-variables
 			% threshold?
 		varset__vars(VarSet0, ListOfVars),
@@ -511,12 +552,24 @@ inlining__inlining_in_goal(call(PredId, ProcId, ArgVars, Builtin, Context,
 			TypeVarSet0, TypeVarSet, TypeInfoVarMap0, 
 			TypeInfoVarMap, Goal - GoalInfo),
 
+			%
+			% If some of the output variables are not used in
+			% the calling procedure, requantify the procedure.
+			%
+		goal_info_get_nonlocals(GoalInfo0, NonLocals),
+		( set__list_to_set(ArgVars) = NonLocals ->
+			Requantify = Requantify0
+		;
+			Requantify = yes
+		),
+
 			% If the inferred determinism of the called
 			% goal differs from the declared determinism,
 			% flag that we should re-run determinism analysis
 			% on this proc.
 		goal_info_get_determinism(GoalInfo0, Determinism0),
 		goal_info_get_determinism(GoalInfo, Determinism),
+		DidInlining = yes,
 		( Determinism0 = Determinism ->
 			DetChanged = DetChanged0
 		;
@@ -529,11 +582,14 @@ inlining__inlining_in_goal(call(PredId, ProcId, ArgVars, Builtin, Context,
 		VarTypes = VarTypes0,
 		TypeVarSet = TypeVarSet0,
 		TypeInfoVarMap = TypeInfoVarMap0,
+		DidInlining = DidInlining0,
+		Requantify = Requantify0,
 		DetChanged = DetChanged0
 	),
-	InlineInfo = inline_info(
-		VarThresh, InlinedProcs, ModuleInfo, HeadTypeParams, Markers,
-		VarSet, VarTypes, TypeVarSet, TypeInfoVarMap, DetChanged).
+	InlineInfo = inline_info(VarThresh, HighLevelCode,
+		InlinedProcs, ModuleInfo, HeadTypeParams, Markers,
+		VarSet, VarTypes, TypeVarSet, TypeInfoVarMap, DidInlining,
+		Requantify, DetChanged).
 
 inlining__inlining_in_goal(generic_call(A, B, C, D) - GoalInfo,
 		generic_call(A, B, C, D) - GoalInfo) --> [].
@@ -541,8 +597,9 @@ inlining__inlining_in_goal(generic_call(A, B, C, D) - GoalInfo,
 inlining__inlining_in_goal(unify(A, B, C, D, E) - GoalInfo,
 		unify(A, B, C, D, E) - GoalInfo) --> [].
 
-inlining__inlining_in_goal(pragma_c_code(A, B, C, D, E, F, G) - GoalInfo,
-		pragma_c_code(A, B, C, D, E, F, G) - GoalInfo) --> [].
+inlining__inlining_in_goal(
+		pragma_foreign_code(A, B, C, D, E, F, G) - GoalInfo,
+		pragma_foreign_code(A, B, C, D, E, F, G) - GoalInfo) --> [].
 
 inlining__inlining_in_goal(bi_implication(_, _) - _, _) -->
 	% these should have been expanded out by now
@@ -721,18 +778,21 @@ inlining__inlining_in_conj([Goal0 | Goals0], Goals) -->
 
 	% Check to see if we should inline a call.
 	%
-	% Fails if the called predicate is a builtin or is imported.
+	% Fails if the called predicate cannot be inlined,
+	% e.g. because it is a builtin, we don't have code for it,
+	% it uses nondet pragma c_code, etc.
 	%
-	% Succeeds if the called predicate has an annotation
-	% indicating that it should be inlined, or if the goal
-	% is a conjunction of builtins.
+	% It succeeds if the called procedure is inlinable,
+	% and in addition either there was a `pragma inline'
+	% for this procedure, or the procedure was marked by
+	% inlining__mark_predproc as having met its heuristic.
 
 :- pred inlining__should_inline_proc(pred_id, proc_id, builtin_state,
-	set(pred_proc_id), pred_markers, module_info).
-:- mode inlining__should_inline_proc(in, in, in, in, in, in) is semidet.
+	bool, set(pred_proc_id), pred_markers, module_info).
+:- mode inlining__should_inline_proc(in, in, in, in, in, in, in) is semidet.
 
-inlining__should_inline_proc(PredId, ProcId, BuiltinState, InlinedProcs,
-		CallingPredMarkers, ModuleInfo) :-
+inlining__should_inline_proc(PredId, ProcId, BuiltinState, HighLevelCode,
+		InlinedProcs, CallingPredMarkers, ModuleInfo) :-
 
 	% don't inline builtins, the code generator will handle them
 
@@ -764,6 +824,35 @@ inlining__should_inline_proc(PredId, ProcId, BuiltinState, InlinedProcs,
 
 	\+ pred_info_requested_no_inlining(PredInfo),
 
+	% For the LLDS back-end,
+	% under no circumstances inline model_non pragma c codes.
+	% The resulting code would not work properly.
+	proc_info_goal(ProcInfo, CalledGoal),
+	\+ (
+		HighLevelCode = no,
+		CalledGoal = pragma_foreign_code(_,_,_,_,_,_,_) - _,
+		proc_info_interface_determinism(ProcInfo, Detism),
+		( Detism = nondet ; Detism = multidet )
+	),
+
+	% only inline foreign_code if it is appropriate for
+	% the target language
+	module_info_globals(ModuleInfo, Globals),
+	globals__get_target(Globals, Target),
+	(
+		(
+		CalledGoal = pragma_foreign_code(ForeignAttributes,
+			_,_,_,_,_,_) - _,
+		foreign_language(ForeignAttributes, ForeignLanguage)
+		)
+	=>
+		ok_to_inline_language(ForeignLanguage, Target)
+	),
+
+	% Don't inline memoed Aditi predicates.
+	pred_info_get_markers(PredInfo, CalledPredMarkers),
+	\+ check_marker(CalledPredMarkers, aditi_memo),
+
 	% Don't inline Aditi procedures into non-Aditi procedures,
 	% since this could result in joins being performed by
 	% backtracking rather than by more efficient methods in
@@ -783,5 +872,19 @@ inlining__should_inline_proc(PredId, ProcId, BuiltinState, InlinedProcs,
 		set__member(proc(PredId, ProcId), InlinedProcs)
 	).
 
+	% Succeed iff it is appropriate to inline `pragma foreign_code'
+	% in the specified language for the given compilation_target.
+	% Generally that will only be the case if the target directly
+	% supports inline code in that language.
+:- pred ok_to_inline_language(foreign_language::in, compilation_target::in)
+	is semidet.
+ok_to_inline_language(c, c).
+% ok_to_inline_language(java, java). % foreign_language = java not implemented
+% ok_to_inline_language(asm, asm).   % foreign_language = asm not implemented
+% We could define a language "C/C++" (c_slash_cplusplus) which was the
+% intersection of "C" and "C++", and then we'd have
+%	ok_to_inline_language(c_slash_cplusplus, c).
+%	ok_to_inline_language(c_slash_cplusplus, cplusplus).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
