@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,8 +14,8 @@
 
 :- interface.
 
-:- import_module llds, globals.
-:- import_module list, set, bool.
+:- import_module llds, trace_params.
+:- import_module list, set, bool, counter.
 
 	% Take an instruction list and optimize jumps. This includes jumps
 	% implicit in procedure returns.
@@ -39,8 +39,9 @@
 	% by the optimization.
 
 :- pred jumpopt_main(list(instruction)::in, set(label)::in, trace_level::in,
-	bool::in, bool::in, bool::in, list(instruction)::out, bool::out)
-	is det.
+	proc_label::in, counter::in, counter::out,
+	bool::in, bool::in, bool::in,
+	list(instruction)::out, bool::out) is det.
 
 %-----------------------------------------------------------------------------%
 
@@ -75,8 +76,8 @@
 % numbering, which can do a better job of optimizing this block, have
 % been applied.
 
-jumpopt_main(Instrs0, LayoutLabels, TraceLevel, Blockopt, Recjump,
-		MostlyDetTailCall, Instrs, Mod) :-
+jumpopt_main(Instrs0, LayoutLabels, TraceLevel, ProcLabel, C0, C,
+		Blockopt, Recjump, MostlyDetTailCall, Instrs, Mod) :-
 	map__init(Instrmap0),
 	map__init(Lvalmap0),
 	map__init(Procmap0),
@@ -89,15 +90,24 @@ jumpopt_main(Instrs0, LayoutLabels, TraceLevel, Blockopt, Recjump,
 	map__init(Forkmap0),
 	jumpopt__build_forkmap(Instrs0, Sdprocmap, Forkmap0, Forkmap),
 	( MostlyDetTailCall = yes ->
-		opt_util__get_prologue(Instrs0, ProcLabel, _, _, _),
-		opt_util__new_label_no(Instrs0, 500, LabelNum),
-		CheckedNondetTailCallInfo = yes(ProcLabel - LabelNum)
+		CheckedNondetTailCallInfo0 = yes(ProcLabel - C0),
+		jumpopt__instr_list(Instrs0, comment(""), Instrmap, Blockmap,
+			Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap,
+			LayoutLabels, TraceLevel, CheckedNondetTailCallInfo0,
+			CheckedNondetTailCallInfo, Instrs1),
+		( CheckedNondetTailCallInfo = yes(_ - Cprime) ->
+			C = Cprime
+		;
+			error("jumpopt_main: lost the next label number")
+		)
 	;
-		CheckedNondetTailCallInfo = no
+		CheckedNondetTailCallInfo0 = no,
+		jumpopt__instr_list(Instrs0, comment(""), Instrmap, Blockmap,
+			Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap,
+			LayoutLabels, TraceLevel, CheckedNondetTailCallInfo0,
+			_, Instrs1),
+		C = C0
 	),
-	jumpopt__instr_list(Instrs0, comment(""), Instrmap, Blockmap, Lvalmap,
-		Procmap, Sdprocmap, Forkmap, Succmap, LayoutLabels,
-		TraceLevel, CheckedNondetTailCallInfo, _, Instrs1),
 	opt_util__filter_out_bad_livevals(Instrs1, Instrs),
 	( Instrs = Instrs0 ->
 		Mod = no
@@ -216,7 +226,7 @@ jumpopt__build_forkmap([Instr - _Comment|Instrs], Sdprocmap,
 
 :- pred jumpopt__instr_list(list(instruction), instr, instrmap, tailmap,
 	lvalmap, tailmap, tailmap, tailmap, tailmap, set(label), trace_level,
-	maybe(pair(proc_label, int)), maybe(pair(proc_label, int)),
+	maybe(pair(proc_label, counter)), maybe(pair(proc_label, counter)),
 	list(instruction)).
 :- mode jumpopt__instr_list(in, in, in, in, in, in, in, in, in, in, in,
 	in, out, out) is det.
@@ -229,10 +239,9 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		TraceLevel, CheckedNondetTailCallInfo0,
 		CheckedNondetTailCallInfo, Instrs) :-
 	Instr0 = Uinstr0 - Comment0,
-	string__append(Comment0, " (redirected return)", Redirect),
 	(
 		Uinstr0 = call(Proc, label(RetLabel), LiveInfos, Context,
-			CallModel)
+			GoalPath, CallModel)
 	->
 		(
 			% Look for det style tailcalls. We look for this
@@ -242,12 +251,13 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			( CallModel = det ; CallModel = semidet ),
 			map__search(Procmap, RetLabel, Between0),
 			PrevInstr = livevals(Livevals),
-			TraceLevel = none,
+			trace_level_is_none(TraceLevel) = yes,
 			not set__member(RetLabel, LayoutLabels)
 		->
 			opt_util__filter_out_livevals(Between0, Between1),
 			list__append(Between1, [livevals(Livevals) - "",
-				goto(Proc) - Redirect], NewInstrs),
+				goto(Proc) - redirect_comment(Comment0)],
+				NewInstrs),
 			RemainInstrs = Instrs0,
 			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
@@ -255,11 +265,12 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			CallModel = semidet,
 			map__search(Forkmap, RetLabel, Between),
 			PrevInstr = livevals(Livevals),
-			TraceLevel = none,
+			trace_level_is_none(TraceLevel) = yes,
 			not set__member(RetLabel, LayoutLabels)
 		->
 			list__append(Between, [livevals(Livevals) - "",
-				goto(Proc) - Redirect], NewInstrs),
+				goto(Proc) - redirect_comment(Comment0)],
+				NewInstrs),
 			RemainInstrs = Instrs0,
 			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
 		;
@@ -269,7 +280,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			map__search(Succmap, RetLabel, BetweenIncl),
 			BetweenIncl = [livevals(_) - _, goto(_) - _],
 			PrevInstr = livevals(Livevals),
-			TraceLevel = none,
+			trace_level_is_none(TraceLevel) = yes,
 			not set__member(RetLabel, LayoutLabels)
 		->
 			NewInstrs = [
@@ -280,7 +291,7 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				assign(curfr, lval(succfr(lval(curfr))))
 					- "setup curfr on return from tailcall",
 				livevals(Livevals) - "",
-				goto(Proc) - Redirect
+				goto(Proc) - redirect_comment(Comment0)
 			],
 			RemainInstrs = Instrs0,
 			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
@@ -289,14 +300,15 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			% a runtime check.
 			CallModel = nondet(checked_tail_call),
 			CheckedNondetTailCallInfo0 =
-				yes(ProcLabel - LabelNum0),
+				yes(ProcLabel - Counter0),
 			map__search(Succmap, RetLabel, BetweenIncl),
 			BetweenIncl = [livevals(_) - _, goto(_) - _],
 			PrevInstr = livevals(Livevals),
-			TraceLevel = none,
+			trace_level_is_none(TraceLevel) = yes,
 			not set__member(RetLabel, LayoutLabels)
 		->
-			NewLabel = local(ProcLabel, LabelNum0),
+			counter__allocate(LabelNum, Counter0, Counter1),
+			NewLabel = local(LabelNum, ProcLabel),
 			NewInstrs = [
 				if_val(binop(ne, lval(curfr), lval(maxfr)),
 					label(NewLabel))
@@ -308,17 +320,16 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				assign(curfr, lval(succfr(lval(curfr))))
 					- "setup curfr on return from tailcall",
 				livevals(Livevals) - "",
-				goto(Proc) - Redirect,
+				goto(Proc) - redirect_comment(Comment0),
 				label(NewLabel) - "non tail call",
 				Instr0
 			],
 			RemainInstrs = Instrs0,
-			LabelNum1 is LabelNum0 + 1,
-			CheckedNondetTailCallInfo1 = yes(ProcLabel - LabelNum1)
+			CheckedNondetTailCallInfo1 = yes(ProcLabel - Counter1)
 		;
 			% Short circuit the return label if possible.
 			map__search(Instrmap, RetLabel, RetInstr),
-			TraceLevel = none,
+			trace_level_is_none(TraceLevel) = yes,
 			not set__member(RetLabel, LayoutLabels)
 		->
 			jumpopt__final_dest(RetLabel, RetInstr, Instrmap,
@@ -328,8 +339,9 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				RemainInstrs = Instrs0
 			;
 				NewInstrs = [call(Proc, label(DestLabel),
-					LiveInfos, Context, CallModel)
-					- Redirect],
+					LiveInfos, Context, GoalPath,
+					CallModel)
+					- redirect_comment(Comment0)],
 				RemainInstrs = Instrs0
 			),
 			CheckedNondetTailCallInfo1 = CheckedNondetTailCallInfo0
@@ -617,6 +629,9 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		CheckedNondetTailCallInfo, Instrs9),
 	list__append(NewInstrs, Instrs9, Instrs).
 
+:- func redirect_comment(string) = string.
+redirect_comment(Comment0) = string__append(Comment0, " (redirected return)").
+
 % We avoid generating statements that redefine the value of a location
 % by comparing its old contents for non-equality with zero.
 %
@@ -741,9 +756,12 @@ jumpopt__short_labels_rval(lval(Lval0), Instrmap, lval(Lval)) :-
 	jumpopt__short_labels_lval(Lval0, Instrmap, Lval).
 jumpopt__short_labels_rval(var(_), _, _) :-
 	error("var rval in jumpopt__short_labels_rval").
-jumpopt__short_labels_rval(create(Tag, Rvals0, ArgTypes, StatDyn, Cell, Type),
-		Instrmap, create(Tag, Rvals, ArgTypes, StatDyn, Cell, Type)) :-
-	jumpopt__short_labels_maybe_rvals(Rvals0, Instrmap, Rvals).
+jumpopt__short_labels_rval(
+		create(Tag, Rvals0, ArgTypes, StatDyn, Cell, Type, Reuse0),
+		Instrmap,
+		create(Tag, Rvals, ArgTypes, StatDyn, Cell, Type, Reuse)) :-
+	jumpopt__short_labels_maybe_rvals(Rvals0, Instrmap, Rvals),
+	jumpopt__short_labels_maybe_rval(Reuse0, Instrmap, Reuse).
 jumpopt__short_labels_rval(mkword(Tag, Rval0), Instrmap,
 		mkword(Tag, Rval)) :-
 	jumpopt__short_labels_rval(Rval0, Instrmap, Rval).
@@ -787,6 +805,13 @@ jumpopt__short_labels_const(data_addr_const(D), _, data_addr_const(D)).
 jumpopt__short_labels_maybe_rvals([], _, []).
 jumpopt__short_labels_maybe_rvals([MaybeRval0 | MaybeRvals0], Instrmap,
 		[MaybeRval | MaybeRvals]) :-
+	jumpopt__short_labels_maybe_rval(MaybeRval0, Instrmap, MaybeRval),
+	jumpopt__short_labels_maybe_rvals(MaybeRvals0, Instrmap, MaybeRvals).
+
+:- pred jumpopt__short_labels_maybe_rval(maybe(rval), instrmap, maybe(rval)).
+:- mode jumpopt__short_labels_maybe_rval(in, in, out) is det.
+
+jumpopt__short_labels_maybe_rval(MaybeRval0, Instrmap, MaybeRval) :-
 	(
 		MaybeRval0 = no,
 		MaybeRval = no
@@ -794,8 +819,7 @@ jumpopt__short_labels_maybe_rvals([MaybeRval0 | MaybeRvals0], Instrmap,
 		MaybeRval0 = yes(Rval0),
 		jumpopt__short_labels_rval(Rval0, Instrmap, Rval),
 		MaybeRval = yes(Rval)
-	),
-	jumpopt__short_labels_maybe_rvals(MaybeRvals0, Instrmap, MaybeRvals).
+	).
 
 :- pred jumpopt__short_labels_lval(lval, instrmap, lval).
 :- mode jumpopt__short_labels_lval(in, in, out) is det.
