@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -28,7 +28,7 @@
 % as late as possible (but before the first possible use), and the
 % death should be as early as possible (but after the last possible use).
 %
-% We compute liveness related information in three distinct passes.
+% We compute liveness related information in four distinct passes.
 %
 % The first pass, detect_liveness_in_goal, finds the first value-giving
 % occurrence of each variable on each computation path. Goals containing
@@ -46,7 +46,18 @@
 % used at all include a pre-death set listing the variables that
 % have died in parallel branches.
 %
-% The third pass, detect_resume_points_in_goal, finds goals that
+% The third pass is optional: it delays the deaths of named variables until
+% the last possible moment. This can be useful if debugging is enabled, as it
+% allows the programmer to look at the values of such variables at as many
+% trace events as possible. If debugging is not enabled, this pass is a pure
+% pessimization (it increases stack slot pressure and thus probably increases
+% the size of the procedure's stack frame), and therefore should not be
+% enabled.
+%
+% The second and third passes cannot be combined, because the second pass
+% traverses goals backwards while the third pass traverses goals forwards.
+%
+% The fourth pass, detect_resume_points_in_goal, finds goals that
 % establish resume points and attaches to them a resume_point
 % annotation listing the variables that may be referenced by the
 % code at that resume point as well as the nature of the required
@@ -56,9 +67,12 @@
 %
 % When using accurate gc or execution tracing, liveness is computed
 % slightly differently.  The runtime system needs access to the
-% typeinfo variables of any variable that is live at a continuation.
+% typeinfo variables of any variable that is live at a continuation or event.
+% (This includes typeclass info variables that hold typeinfos;
+% in the following "typeinfo variables" also includes typeclass info
+% variables.)
 % 
-% Hence, the invariant needed for alternate liveness calculation:
+% Hence, the invariant needed for typeinfo-liveness calculation:
 % 	a variable holding a typeinfo must be live at any continuation
 % 	where any variable whose type is described (in whole or in part)
 % 	by that typeinfo is live.
@@ -82,12 +96,26 @@
 %	    variable.
 % 
 % (1) happens without any changes to the liveness computation (it is
-%     the normal condition for variables becoming dead). This more
+%     the normal condition for variables becoming dead). This is more
 %     conservative than what is required for the invariant, but is
 %     required for code generation, so we should keep it ;-)
 % (2) is implemented by adding the typeinfo variables for the types of the
 %     nonlocals to the nonlocals for the purposes of computing liveness.
 %
+% In some circumstances, one of which is tests/debugger/resume_typeinfos.m,
+% it is possible for a typeinfo variable to be born in a goal without that
+% typeinfo variable appearing anywhere else in the procedure body.
+% Nevertheless, with typeinfo liveness, we must consider such variables
+% to be born in such goals even though they do not appear in the nonlocals set
+% or in the instmap delta. (If they were not born, it would be an error for
+% them to die, and die they will, at the last occurrence of a variable whose
+% type they (partially) describe.) The special case solution we adopt for
+% such situations is that we consider the first appearance of a typeinfo
+% variable in the typeinfo-completed nonlocals set of a goal to be a value
+% giving occurrence, even if the typeinfo does not appear in the instmap delta.
+% This is safe, since with our current scheme for handling polymorphism,
+% the first appearance will in fact always ground the typeinfo.
+%
 % So typeinfo variables will always be born before they are needed, and
 % die only when no other variable needing them will be live, so the
 % invariant holds.
@@ -143,35 +171,34 @@
 
 :- implementation.
 
-:- import_module hlds_goal, hlds_data, llds, quantification, (inst), instmap.
-:- import_module hlds_out, mode_util, code_util, quantification, options.
-:- import_module trace, globals, polymorphism, passes_aux.
-:- import_module term, varset.
+% Parse tree modules
+:- import_module prog_util, (inst).
+% HLDS modules
+:- import_module hlds_goal, hlds_data, hlds_out, instmap, mode_util.
+:- import_module quantification, polymorphism.
+% Modules shared between different back-ends.
+:- import_module code_model, passes_aux.
+% LLDS modules
+:- import_module llds, code_util, trace_params, trace.
+% Misc
+:- import_module globals, options.
 
-:- import_module bool, map, std_util, list, assoc_list, require.
-:- import_module string.
+% Standard library modules
+:- import_module bool, string, map, std_util, list, assoc_list, require.
+:- import_module term, varset.
 
 detect_liveness_proc(ProcInfo0, PredId, ModuleInfo, ProcInfo) :-
 	requantify_proc(ProcInfo0, ProcInfo1),
-	proc_info_goal(ProcInfo1, Goal0),
-	proc_info_varset(ProcInfo1, Varset),
-	proc_info_vartypes(ProcInfo1, VarTypes),
-	module_info_globals(ModuleInfo, Globals),
 
+	proc_info_goal(ProcInfo1, Goal0),
+	proc_info_varset(ProcInfo1, VarSet),
+	proc_info_vartypes(ProcInfo1, VarTypes),
+	proc_info_typeinfo_varmap(ProcInfo1, TVarMap),
+	module_info_globals(ModuleInfo, Globals),
 	module_info_pred_info(ModuleInfo, PredId, PredInfo),
-	pred_info_module(PredInfo, PredModule),
-	pred_info_name(PredInfo, PredName),
-	pred_info_arity(PredInfo, PredArity),
-	(
-		polymorphism__no_type_info_builtin(PredModule,
-			PredName, PredArity)
-	->
-		TypeInfoLiveness = no
-	;
-		body_should_use_typeinfo_liveness(Globals, TypeInfoLiveness)
-	),
-	live_info_init(ModuleInfo, ProcInfo1, TypeInfoLiveness,
-		VarTypes, Varset, LiveInfo),
+	body_should_use_typeinfo_liveness(PredInfo, Globals, TypeInfoLiveness),
+	live_info_init(ModuleInfo, TypeInfoLiveness, VarTypes, TVarMap, VarSet,
+		LiveInfo),
 
 	initial_liveness(ProcInfo1, PredId, ModuleInfo, Liveness0),
 	detect_liveness_in_goal(Goal0, Liveness0, LiveInfo,
@@ -180,13 +207,25 @@ detect_liveness_proc(ProcInfo0, PredId, ModuleInfo, ProcInfo) :-
 	initial_deadness(ProcInfo1, LiveInfo, ModuleInfo, Deadness0),
 	detect_deadness_in_goal(Goal1, Deadness0, LiveInfo, _, Goal2),
 
+	(
+		globals__get_trace_level(Globals, TraceLevel),
+		AllowDelayDeath = trace_level_allows_delay_death(TraceLevel),
+		AllowDelayDeath = yes,
+		globals__lookup_bool_option(Globals, delay_death, DelayDeath),
+		DelayDeath = yes
+	->
+		delay_death_proc_body(Goal2, VarSet, Liveness0, Goal3)
+	;
+		Goal3 = Goal2
+	),
+
 	globals__get_trace_level(Globals, TraceLevel),
-	( TraceLevel \= none ->
+	( trace_level_is_none(TraceLevel) = no ->
 		trace__fail_vars(ModuleInfo, ProcInfo0, ResumeVars0)
 	;
 		set__init(ResumeVars0)
 	),
-	detect_resume_points_in_goal(Goal2, Liveness0, LiveInfo,
+	detect_resume_points_in_goal(Goal3, Liveness0, LiveInfo,
 		ResumeVars0, Goal, _),
 
 	proc_info_set_goal(ProcInfo1, Goal, ProcInfo2),
@@ -203,8 +242,9 @@ detect_liveness_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
 		Liveness, Goal - GoalInfo) :-
 
 		% work out which variables get born in this goal
-	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo0, NonLocals),
-	set__difference(NonLocals, Liveness0, NewVarsSet),
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo0,
+		BaseNonLocals, CompletedNonLocals),
+	set__difference(CompletedNonLocals, Liveness0, NewVarsSet),
 	set__to_sorted_list(NewVarsSet, NewVarsList),
 	goal_info_get_instmap_delta(GoalInfo0, InstMapDelta),
 	set__init(Empty),
@@ -213,12 +253,14 @@ detect_liveness_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
 	;
 		set__init(Births0),
 		find_value_giving_occurrences(NewVarsList, LiveInfo,
-			InstMapDelta, Births0, Births)
+			InstMapDelta, Births0, Births1),
+		set__difference(CompletedNonLocals, BaseNonLocals,
+			TypeInfos),
+		set__difference(TypeInfos, Liveness0, NewTypeInfos),
+		set__union(Births1, NewTypeInfos, Births)
 	),
 	set__union(Liveness0, Births, Liveness),
-	(
-		goal_is_atomic(Goal0)
-	->
+	( goal_is_atomic(Goal0) ->
 		PreDeaths = Empty,
 		PreBirths = Births,
 		PostDeaths = Empty,
@@ -227,9 +269,10 @@ detect_liveness_in_goal(Goal0 - GoalInfo0, Liveness0, LiveInfo,
 	;
 		PreDeaths = Empty,
 		PreBirths = Empty,
-		detect_liveness_in_goal_2(Goal0, Liveness0, NonLocals,
+		detect_liveness_in_goal_2(Goal0, Liveness0, CompletedNonLocals,
 			LiveInfo, ActualLiveness, Goal),
-		set__intersect(NonLocals, ActualLiveness, NonLocalLiveness),
+		set__intersect(CompletedNonLocals, ActualLiveness,
+			NonLocalLiveness),
 		set__union(NonLocalLiveness, Liveness0, FinalLiveness),
 		set__difference(FinalLiveness, Liveness, PostDeaths),
 		set__difference(Liveness, FinalLiveness, PostBirths)
@@ -322,8 +365,9 @@ detect_liveness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
 detect_liveness_in_goal_2(unify(_,_,_,_,_), _, _, _, _, _) :-
 	error("unify in detect_liveness_in_goal_2").
 
-detect_liveness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_), _, _, _, _, _) :-
-	error("pragma_c_code in detect_liveness_in_goal_2").
+detect_liveness_in_goal_2(pragma_foreign_code(_,_,_,_,_,_,_),
+		_, _, _, _, _) :-
+	error("pragma_foreign_code in detect_liveness_in_goal_2").
 
 detect_liveness_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
 	error("bi_implication in detect_liveness_in_goal_2").
@@ -422,11 +466,10 @@ detect_deadness_in_goal(Goal0 - GoalInfo0, Deadness0, LiveInfo, Deadness,
 	set__difference(Deadness0, PostBirths0, Deadness1),
 	set__union(Deadness1, PostDeaths0, Deadness2),
 
-	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo0, NonLocals),
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo0,
+		_BaseNonLocals, CompletedNonLocals),
 	set__init(Empty),
-	(
-		goal_is_atomic(Goal0)
-	->
+	( goal_is_atomic(Goal0) ->
 		% 
 		% The code below is slightly dodgy: the new postdeaths really
 		% ought to be computed as the difference between the liveness
@@ -444,7 +487,7 @@ detect_deadness_in_goal(Goal0 - GoalInfo0, Deadness0, LiveInfo, Deadness,
 		% some extra code in the third pass (detect_resume_points)
 		% to delete any such untimely deaths.]
 		% 
-		set__difference(NonLocals, Deadness2, NewPostDeaths),
+		set__difference(CompletedNonLocals, Deadness2, NewPostDeaths),
 		set__union(Deadness2, NewPostDeaths, Deadness3),
 		Goal = Goal0
 	;
@@ -472,23 +515,26 @@ detect_deadness_in_goal_2(conj(Goals0), _, Deadness0, LiveInfo,
 detect_deadness_in_goal_2(par_conj(Goals0, SM), GoalInfo, Deadness0, LiveInfo,
 		Deadness, par_conj(Goals, SM)) :-
 	set__init(Union0),
-	goal_info_get_nonlocals(GoalInfo, NonLocals),
-	detect_deadness_in_par_conj(Goals0, Deadness0, NonLocals,
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo,
+		_, CompletedNonLocals),
+	detect_deadness_in_par_conj(Goals0, Deadness0, CompletedNonLocals,
 		LiveInfo, Union0, Union, Goals),
 	set__union(Union, Deadness0, Deadness).
 
 detect_deadness_in_goal_2(disj(Goals0, SM), GoalInfo, Deadness0,
 		LiveInfo, Deadness, disj(Goals, SM)) :-
 	set__init(Union0),
-	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo, NonLocals),
-	detect_deadness_in_disj(Goals0, Deadness0, NonLocals,
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo,
+		_, CompletedNonLocals),
+	detect_deadness_in_disj(Goals0, Deadness0, CompletedNonLocals,
 		LiveInfo, Union0, Deadness, Goals).
 
 detect_deadness_in_goal_2(switch(Var, Det, Cases0, SM), GoalInfo, Deadness0,
 		LiveInfo, Deadness, switch(Var, Det, Cases, SM)) :-
 	set__init(Union0),
-	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo, NonLocals),
-	detect_deadness_in_cases(Var, Cases0, Deadness0, NonLocals,
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo,
+		_, CompletedNonLocals),
+	detect_deadness_in_cases(Var, Cases0, Deadness0, CompletedNonLocals,
 		LiveInfo, Union0, Deadness, Cases).
 
 detect_deadness_in_goal_2(not(Goal0), _, Deadness0, LiveInfo,
@@ -505,9 +551,10 @@ detect_deadness_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	detect_deadness_in_goal(Cond0, DeadnessThen, LiveInfo,
 		DeadnessCond, Cond1),
 
-	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo, NonLocals),
+	liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo,
+		_, CompletedNonLocals),
 	set__union(DeadnessCond, DeadnessElse, Deadness),
-	set__intersect(Deadness, NonLocals, NonLocalDeadness),
+	set__intersect(Deadness, CompletedNonLocals, NonLocalDeadness),
 
 	set__difference(NonLocalDeadness, DeadnessCond, ResidueCond),
 	set__difference(NonLocalDeadness, DeadnessElse, ResidueElse),
@@ -528,8 +575,9 @@ detect_deadness_in_goal_2(call(_,_,_,_,_,_), _, _, _, _, _) :-
 detect_deadness_in_goal_2(unify(_,_,_,_,_), _, _, _, _, _) :-
 	error("unify in detect_deadness_in_goal_2").
 
-detect_deadness_in_goal_2(pragma_c_code(_,_,_,_,_,_,_), _, _, _, _, _) :-
-	error("pragma_c_code in detect_deadness_in_goal_2").
+detect_deadness_in_goal_2(pragma_foreign_code(_, _, _, _, _, _, _),
+		_, _, _, _, _) :-
+	error("pragma_foreign_code in detect_deadness_in_goal_2").
 
 detect_deadness_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
 	error("bi_implication in detect_deadness_in_goal_2").
@@ -620,6 +668,288 @@ detect_deadness_in_par_conj([Goal0 | Goals0], Deadness, NonLocals, LiveInfo,
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
+% The delay_death pass works by maintaining a set of variables (all named)
+% that, according to the deadness pass should have died by now, but which
+% are being kept alive so that their values are accessible to the debugger.
+% Variables in this DelayedDead set are finally killed when we come to the end
+% of the goal in which they were born. This is because if a variable is born in
+% one arm of a branched control structure (e.g. a switch), it cannot live
+% beyond the control structure, because it is not given a value in other
+% branches.
+%
+% The correctness of this pass with typeinfo_liveness depends on the fact that
+% typeinfo and typeclass_info variables are all named. If they weren't, then it
+% would be possible for a (named) variable to have its death delayed without
+% the type(class)info variable describing part of its type having its death
+% delayed as well. In fact, its death will be delayed by at least as much,
+% since a variable cannot be live on entry to a branched control structure
+% without the type(class)info variables describing its type being live there as
+% well.
+
+:- pred delay_death_proc_body(hlds_goal::in, prog_varset::in, set(prog_var)::in,
+	hlds_goal::out) is det.
+
+delay_death_proc_body(Goal0, VarSet, BornVars0, Goal) :-
+	delay_death_goal(Goal0, BornVars0, set__init, VarSet,
+		Goal1, _, DelayedDead),
+	Goal1 = GoalExpr - GoalInfo1,
+	goal_info_get_post_deaths(GoalInfo1, PostDeaths1),
+	set__union(PostDeaths1, DelayedDead, PostDeaths),
+	goal_info_set_post_deaths(GoalInfo1, PostDeaths, GoalInfo),
+	Goal = GoalExpr - GoalInfo.
+
+:- pred delay_death_goal(hlds_goal::in, set(prog_var)::in, set(prog_var)::in,
+	prog_varset::in,
+	hlds_goal::out, set(prog_var)::out, set(prog_var)::out) is det.
+
+delay_death_goal(GoalExpr0 - GoalInfo0, BornVars0, DelayedDead0, VarSet,
+		GoalExpr - GoalInfo, BornVars, DelayedDead) :-
+	goal_info_get_pre_births(GoalInfo0, PreBirths),
+	goal_info_get_pre_deaths(GoalInfo0, PreDeaths0),
+
+	set__union(BornVars0, PreBirths, BornVars1),
+	set__divide(var_is_named(VarSet), PreDeaths0,
+		PreDelayedDead, UnnamedPreDeaths),
+	set__union(DelayedDead0, PreDelayedDead, DelayedDead1),
+	goal_info_set_pre_deaths(GoalInfo0, UnnamedPreDeaths, GoalInfo1),
+
+	delay_death_goal_expr(GoalExpr0, GoalInfo1, BornVars1, DelayedDead1,
+		VarSet, GoalExpr, GoalInfo2, BornVars2, DelayedDead2),
+
+	goal_info_get_post_births(GoalInfo2, PostBirths),
+	goal_info_get_post_deaths(GoalInfo2, PostDeaths2),
+
+	set__union(BornVars2, PostBirths, BornVars),
+	set__divide(var_is_named(VarSet), PostDeaths2,
+		PostDelayedDead, UnnamedPostDeaths),
+	set__union(DelayedDead2, PostDelayedDead, DelayedDead3),
+	set__divide(set__contains(BornVars0), DelayedDead3,
+		DelayedDead, ToBeKilled),
+	set__union(UnnamedPostDeaths, ToBeKilled, PostDeaths),
+	goal_info_set_post_deaths(GoalInfo2, PostDeaths, GoalInfo).
+
+:- pred var_is_named(prog_varset::in, prog_var::in) is semidet.
+
+var_is_named(VarSet, Var) :-
+	varset__search_name(VarSet, Var, _).
+
+:- pred delay_death_goal_expr(hlds_goal_expr::in, hlds_goal_info::in,
+	set(prog_var)::in, set(prog_var)::in, prog_varset::in,
+	hlds_goal_expr::out, hlds_goal_info::out,
+	set(prog_var)::out, set(prog_var)::out) is det.
+
+delay_death_goal_expr(GoalExpr0, GoalInfo0, BornVars0, DelayedDead0, VarSet,
+		GoalExpr, GoalInfo, BornVars, DelayedDead) :-
+	(
+		GoalExpr0 = call(_, _, _, _, _, _),
+		GoalExpr = GoalExpr0,
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0,
+		DelayedDead = DelayedDead0
+	;
+		GoalExpr0 = generic_call(_, _, _, _),
+		GoalExpr = GoalExpr0,
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0,
+		DelayedDead = DelayedDead0
+	;
+		GoalExpr0 = unify(_, _, _, _, _),
+		GoalExpr = GoalExpr0,
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0,
+		DelayedDead = DelayedDead0
+	;
+		GoalExpr0 = pragma_foreign_code(_, _, _, _, _, _, _),
+		GoalExpr = GoalExpr0,
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0,
+		DelayedDead = DelayedDead0
+	;
+		GoalExpr0 = conj(Goals0),
+		delay_death_conj(Goals0, BornVars0, DelayedDead0, VarSet,
+			Goals, BornVars, DelayedDead),
+		GoalExpr = conj(Goals),
+		GoalInfo = GoalInfo0
+	;
+		GoalExpr0 = par_conj(Goals0, SM),
+		delay_death_conj(Goals0, BornVars0, DelayedDead0, VarSet,
+			Goals, BornVars, DelayedDead),
+		GoalExpr = par_conj(Goals, SM),
+		GoalInfo = GoalInfo0
+	;
+		GoalExpr0 = disj(Goals0, SM),
+		delay_death_disj(Goals0, BornVars0, DelayedDead0, VarSet,
+			GoalDeaths, MaybeBornVarsDelayedDead),
+		(
+			MaybeBornVarsDelayedDead = yes(BornVars - DelayedDead),
+			Goals = list__map(
+				kill_excess_delayed_dead_goal(DelayedDead),
+				GoalDeaths),
+			GoalExpr = disj(Goals, SM),
+			GoalInfo = GoalInfo0
+		;
+			MaybeBornVarsDelayedDead = no,
+			% Empty disjunctions represent the goal `fail',
+			% so we process them as if they were primitive goals.
+			GoalExpr = GoalExpr0,
+			GoalInfo = GoalInfo0,
+			BornVars = BornVars0,
+			DelayedDead = DelayedDead0
+		)
+	;
+		GoalExpr0 = switch(Var, CanFail, Cases0, SM),
+		delay_death_cases(Cases0, BornVars0, DelayedDead0, VarSet,
+			CaseDeaths, MaybeBornVarsDelayedDead),
+		(
+			MaybeBornVarsDelayedDead = yes(BornVars - DelayedDead),
+			Cases = list__map(
+				kill_excess_delayed_dead_case(DelayedDead),
+				CaseDeaths),
+			GoalExpr = switch(Var, CanFail, Cases, SM),
+			GoalInfo = GoalInfo0
+		;
+			MaybeBornVarsDelayedDead = no,
+			error("delay_death_goal_expr: empty switch")
+		)
+	;
+		GoalExpr0 = not(Goal0),
+		delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+			Goal, _BornVars, DelayedDead),
+		GoalExpr = not(Goal),
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0
+	;
+		GoalExpr0 = if_then_else(QuantVars, Cond0, Then0, Else0, SM),
+		delay_death_goal(Cond0, BornVars0, DelayedDead0, VarSet,
+			Cond, BornVarsCond, DelayedDeadCond),
+		delay_death_goal(Then0, BornVarsCond, DelayedDeadCond, VarSet,
+			Then1, BornVarsThen, DelayedDeadThen),
+		delay_death_goal(Else0, BornVars0, DelayedDead0, VarSet,
+			Else1, BornVarsElse, DelayedDeadElse),
+		set__intersect(BornVarsThen, BornVarsElse, BornVars),
+		set__intersect(DelayedDeadThen, DelayedDeadElse, DelayedDead),
+		Then = kill_excess_delayed_dead_goal(DelayedDead,
+			Then1 - DelayedDeadThen),
+		Else = kill_excess_delayed_dead_goal(DelayedDead,
+			Else1 - DelayedDeadElse),
+		GoalExpr = if_then_else(QuantVars, Cond, Then, Else, SM),
+		GoalInfo = GoalInfo0
+	;
+		GoalExpr0 = some(QuantVars, CanRemove, Goal0),
+		delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+			Goal, _BornVars, DelayedDead),
+		GoalExpr = some(QuantVars, CanRemove, Goal),
+		GoalInfo = GoalInfo0,
+		BornVars = BornVars0
+	;
+		GoalExpr0 = bi_implication(_, _),
+		error("delay_death_goal_expr: bi_implication")
+	).
+
+:- pred delay_death_conj(list(hlds_goal)::in,
+	set(prog_var)::in, set(prog_var)::in, prog_varset::in,
+	list(hlds_goal)::out, set(prog_var)::out, set(prog_var)::out) is det.
+
+delay_death_conj([], BornVars, DelayedDead, _, [], BornVars, DelayedDead).
+delay_death_conj([Goal0 | Goals0], BornVars0, DelayedDead0, VarSet,
+		[Goal | Goals], BornVars, DelayedDead) :-
+	delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+		Goal, BornVars1, DelayedDead1),
+	delay_death_conj(Goals0, BornVars1, DelayedDead1, VarSet,
+		Goals, BornVars, DelayedDead).
+
+:- pred delay_death_par_conj(list(hlds_goal)::in,
+	set(prog_var)::in, set(prog_var)::in, prog_varset::in,
+	list(hlds_goal)::out, set(prog_var)::out, set(prog_var)::out) is det.
+
+delay_death_par_conj([], BornVars, DelayedDead, _, [], BornVars, DelayedDead).
+delay_death_par_conj([Goal0 | Goals0], BornVars0, DelayedDead0, VarSet,
+		[Goal | Goals], BornVars, DelayedDead) :-
+	delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+		Goal, BornVarsGoal, DelayedDeadGoal),
+	delay_death_par_conj(Goals0, BornVars0, DelayedDead0, VarSet,
+		Goals, BornVarsGoals, DelayedDeadGoals),
+	set__union(BornVarsGoal, BornVarsGoals, BornVars),
+	set__union(DelayedDeadGoal, DelayedDeadGoals, DelayedDead).
+
+:- pred delay_death_disj(list(hlds_goal)::in,
+	set(prog_var)::in, set(prog_var)::in, prog_varset::in,
+	assoc_list(hlds_goal, set(prog_var))::out,
+	maybe(pair(set(prog_var)))::out) is det.
+
+delay_death_disj([], _, _, _, [], no).
+delay_death_disj([Goal0 | Goals0], BornVars0, DelayedDead0, VarSet,
+		[Goal - DelayedDeadGoal | Goals],
+		yes(BornVars - DelayedDead)) :-
+	delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+		Goal, BornVarsGoal, DelayedDeadGoal),
+	delay_death_disj(Goals0, BornVars0, DelayedDead0, VarSet,
+		Goals, MaybeBornVarsDelayedDead),
+	(
+		MaybeBornVarsDelayedDead =
+			yes(BornVarsGoals - DelayedDeadGoals),
+		set__intersect(BornVarsGoal, BornVarsGoals, BornVars),
+		set__intersect(DelayedDeadGoal, DelayedDeadGoals, DelayedDead)
+	;
+		MaybeBornVarsDelayedDead = no,
+		BornVars = BornVarsGoal,
+		DelayedDead = DelayedDeadGoal
+	).
+
+:- pred delay_death_cases(list(case)::in,
+	set(prog_var)::in, set(prog_var)::in, prog_varset::in,
+	assoc_list(case, set(prog_var))::out, maybe(pair(set(prog_var)))::out)
+	is det.
+
+delay_death_cases([], _, _, _, [], no).
+delay_death_cases([case(ConsId, Goal0) | Cases0], BornVars0, DelayedDead0,
+		VarSet, [case(ConsId, Goal) - DelayedDeadGoal | Cases],
+		yes(BornVars - DelayedDead)) :-
+	delay_death_goal(Goal0, BornVars0, DelayedDead0, VarSet,
+		Goal, BornVarsGoal, DelayedDeadGoal),
+	delay_death_cases(Cases0, BornVars0, DelayedDead0, VarSet,
+		Cases, MaybeBornVarsDelayedDead),
+	(
+		MaybeBornVarsDelayedDead =
+			yes(BornVarsCases - DelayedDeadCases),
+		set__intersect(BornVarsGoal, BornVarsCases, BornVars),
+		set__intersect(DelayedDeadGoal, DelayedDeadCases, DelayedDead)
+	;
+		MaybeBornVarsDelayedDead = no,
+		BornVars = BornVarsGoal,
+		DelayedDead = DelayedDeadGoal
+	).
+
+% The kill_excess_delayed_dead_* functions are called on each branch of a
+% branched control structure to make sure that all branches kill the same
+% set of variables.
+
+:- func kill_excess_delayed_dead_goal(set(prog_var),
+	pair(hlds_goal, set(prog_var))) = hlds_goal.
+
+kill_excess_delayed_dead_goal(FinalDelayedDead, Goal0 - DelayedDead0) = Goal :-
+	set__difference(DelayedDead0, FinalDelayedDead, ToBeKilled),
+	Goal0 = GoalExpr - GoalInfo0,
+	goal_info_get_post_deaths(GoalInfo0, PostDeath0),
+	set__union(PostDeath0, ToBeKilled, PostDeath),
+	goal_info_set_post_deaths(GoalInfo0, PostDeath, GoalInfo),
+	Goal = GoalExpr - GoalInfo.
+
+:- func kill_excess_delayed_dead_case(set(prog_var),
+	pair(case, set(prog_var))) = case.
+
+kill_excess_delayed_dead_case(FinalDelayedDead,
+		case(ConsId, Goal0) - DelayedDead0) = case(ConsId, Goal) :-
+	set__difference(DelayedDead0, FinalDelayedDead, ToBeKilled),
+	Goal0 = GoalExpr - GoalInfo0,
+	goal_info_get_post_deaths(GoalInfo0, PostDeath0),
+	set__union(PostDeath0, ToBeKilled, PostDeath),
+	goal_info_set_post_deaths(GoalInfo0, PostDeath, GoalInfo),
+	Goal = GoalExpr - GoalInfo.
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
 :- pred detect_resume_points_in_goal(hlds_goal, set(prog_var), live_info,
 		set(prog_var), hlds_goal, set(prog_var)).
 :- mode detect_resume_points_in_goal(in, in, in, in, out, out) is det.
@@ -679,7 +1009,10 @@ detect_resume_points_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	Else0 = _ElseExpr0 - ElseInfo0,
 	goal_info_get_pre_deaths(ElseInfo0, ElsePreDeath0),
 	set__difference(Liveness0, ElsePreDeath0, CondResumeVars0),
-	set__union(CondResumeVars0, ResumeVars0, CondResumeVars),
+	liveness__maybe_complete_with_typeinfos(LiveInfo, CondResumeVars0,
+		CondResumeVars1),
+		% ResumeVars0 should already have been completed.
+	set__union(CondResumeVars1, ResumeVars0, CondResumeVars),
 
 	detect_resume_points_in_goal(Cond0, Liveness0, LiveInfo,
 		CondResumeVars, Cond1, LivenessCond),
@@ -722,7 +1055,10 @@ detect_resume_points_in_goal_2(not(Goal0), _, Liveness0, LiveInfo, ResumeVars0,
 		not(Goal), Liveness) :-
 	detect_resume_points_in_goal(Goal0, Liveness0, LiveInfo, ResumeVars0,
 		_, Liveness),
-	set__union(Liveness, ResumeVars0, ResumeVars1),
+	liveness__maybe_complete_with_typeinfos(LiveInfo, Liveness,
+		CompletedLiveness),
+		% ResumeVars0 should already have been completed.
+	set__union(CompletedLiveness, ResumeVars0, ResumeVars1),
 	detect_resume_points_in_goal(Goal0, Liveness0, LiveInfo, ResumeVars1,
 		Goal1, _Liveness),
 
@@ -751,9 +1087,8 @@ detect_resume_points_in_goal_2(call(A,B,C,D,E,F), _, Liveness, _, _,
 detect_resume_points_in_goal_2(unify(A,B,C,D,E), _, Liveness, _, _,
 		unify(A,B,C,D,E), Liveness).
 
-detect_resume_points_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _, Liveness,
-		_, _,
-		pragma_c_code(A,B,C,D,E,F,G), Liveness).
+detect_resume_points_in_goal_2(pragma_foreign_code(A,B,C,D,E,F,G), _,
+		Liveness, _, _, pragma_foreign_code(A,B,C,D,E,F,G), Liveness).
 
 detect_resume_points_in_goal_2(bi_implication(_, _), _, _, _, _, _, _) :-
 	% these should have been expanded out by now
@@ -777,6 +1112,14 @@ detect_resume_points_in_conj([Goal0 | Goals0], Liveness0, LiveInfo,
 	% we stop generating code after the first cannot_fail disjunct.
 	% Second, an empty pruned disjunction is legal, while an empty
 	% nondet disjunction isn't.
+	%
+	% For both kinds of disjunctions, the resume points to be attached to
+	% the non-last disjuncts must be completed with the required typeinfos
+	% if --typeinfo-liveness is set. ResumeVars0 should already be so
+	% completed, so we need only complete the sets added here. We therefore
+	% perform this completion when we return the set of variables needed by
+	% the last disjunct, and when we add to this set the set of variables
+	% needed by a non-last disjunct.
 
 :- pred detect_resume_points_in_non_disj(list(hlds_goal), set(prog_var),
 		live_info, set(prog_var), list(hlds_goal),
@@ -876,7 +1219,10 @@ detect_resume_points_in_non_last_disjunct(Goal0, MayUseOrigOnly,
 	Goal = _ - GoalInfo,
 	goal_info_get_pre_deaths(GoalInfo, PreDeaths),
 	set__difference(Liveness0, PreDeaths, NeededFirst),
-	set__union(NeededFirst, NeededRest, Needed),
+	liveness__maybe_complete_with_typeinfos(LiveInfo, NeededFirst,
+		CompletedNeededFirst),
+		% NeededRest has already been completed.
+	set__union(CompletedNeededFirst, NeededRest, Needed),
 
 	require_equal(Liveness, LivenessRest, "disjunction", LiveInfo).
 
@@ -886,12 +1232,14 @@ detect_resume_points_in_non_last_disjunct(Goal0, MayUseOrigOnly,
 	is det.
 
 detect_resume_points_in_last_disjunct(Goal0, Liveness0, LiveInfo,
-		ResumeVars0, Goal, Liveness, Needed) :-
+		ResumeVars0, Goal, Liveness, CompletedNeeded) :-
 	detect_resume_points_in_goal(Goal0, Liveness0, LiveInfo,
 		ResumeVars0, Goal, Liveness),
 	Goal = _ - GoalInfo,
 	goal_info_get_pre_deaths(GoalInfo, PreDeaths),
-	set__difference(Liveness0, PreDeaths, Needed).
+	set__difference(Liveness0, PreDeaths, Needed),
+	liveness__maybe_complete_with_typeinfos(LiveInfo, Needed,
+		CompletedNeeded).
 
 :- pred detect_resume_points_in_cases(list(case), set(prog_var), live_info,
 		set(prog_var), list(case), set(prog_var)).
@@ -933,12 +1281,12 @@ require_equal(LivenessFirst, LivenessRest, GoalType, LiveInfo) :-
 	->
 		true
 	;
-		live_info_get_varset(LiveInfo, Varset),
+		VarSet = LiveInfo ^ varset,
 		set__to_sorted_list(LivenessFirst, FirstVarsList),
 		set__to_sorted_list(LivenessRest, RestVarsList),
-		list__map(varset__lookup_name(Varset),
+		list__map(varset__lookup_name(VarSet),
 			FirstVarsList, FirstVarNames),
-		list__map(varset__lookup_name(Varset),
+		list__map(varset__lookup_name(VarSet),
 			RestVarsList, RestVarNames),
 		Pad = lambda([S0::in, S::out] is det,
 			string__append(S0, " ", S)),
@@ -979,23 +1327,12 @@ initial_liveness(ProcInfo, PredId, ModuleInfo, Liveness) :-
 		% typeinfo vars.
 	module_info_globals(ModuleInfo, Globals),
 	proc_info_goal(ProcInfo, _Goal - GoalInfo),
-	goal_info_get_nonlocals(GoalInfo, NonLocals0),
+	goal_info_get_code_gen_nonlocals(GoalInfo, NonLocals0),
 	module_info_pred_info(ModuleInfo, PredId, PredInfo),
-	body_should_use_typeinfo_liveness(Globals, TypeinfoLiveness),
-	pred_info_module(PredInfo, PredModule),
-	pred_info_name(PredInfo, PredName),
-	pred_info_arity(PredInfo, PredArity),
-	( 	
-		TypeinfoLiveness = yes,
-		\+ polymorphism__no_type_info_builtin(PredModule,
-			PredName, PredArity)
-	->
-		proc_info_get_typeinfo_vars_setwise(ProcInfo, NonLocals0,
-			TypeInfoNonLocals),
-		set__union(NonLocals0, TypeInfoNonLocals, NonLocals)
-	;
-		NonLocals = NonLocals0
-	),
+	proc_info_typeinfo_varmap(ProcInfo, TVarMap),
+	body_should_use_typeinfo_liveness(PredInfo, Globals, TypeinfoLiveness),
+	proc_info_maybe_complete_with_typeinfo_vars(NonLocals0,
+		TypeinfoLiveness, VarTypes, TVarMap, NonLocals),
 	set__intersect(Liveness2, NonLocals, Liveness).
 
 :- pred initial_liveness_2(list(prog_var), list(mode), list(type), module_info,
@@ -1037,16 +1374,9 @@ initial_deadness(ProcInfo, LiveInfo, ModuleInfo, Deadness) :-
 
 		% If doing alternate liveness, the corresponding
 		% typeinfos need to be added to these.
-	live_info_get_typeinfo_liveness(LiveInfo, TypeinfoLiveness),
-	( 
-		TypeinfoLiveness = yes
-	->
-		proc_info_get_typeinfo_vars_setwise(ProcInfo, Deadness2,
-			TypeInfoVars),
-		set__union(Deadness2, TypeInfoVars, Deadness)
-	;
-		Deadness = Deadness2
-	).
+	proc_info_typeinfo_varmap(ProcInfo, TVarMap),
+	proc_info_maybe_complete_with_typeinfo_vars(Deadness2,
+		LiveInfo ^ typeinfo_liveness, VarTypes, TVarMap, Deadness).
 
 :- pred initial_deadness_2(list(prog_var), list(mode), list(type),
 		module_info, set(prog_var), set(prog_var)).
@@ -1101,11 +1431,11 @@ add_deadness_before_goal(Goal - GoalInfo0, Residue, Goal - GoalInfo) :-
 find_value_giving_occurrences([], _, _, ValueVars, ValueVars).
 find_value_giving_occurrences([Var | Vars], LiveInfo, InstMapDelta,
 		ValueVars0, ValueVars) :-
-	live_info_get_var_types(LiveInfo, VarTypes),
-	live_info_get_module_info(LiveInfo, ModuleInfo),
+	VarTypes = LiveInfo ^ vartypes,
 	map__lookup(VarTypes, Var, Type),
 	(
 		instmap_delta_search_var(InstMapDelta, Var, Inst),
+		ModuleInfo = LiveInfo ^ module_info,
 		mode_to_arg_mode(ModuleInfo, (free -> Inst), Type, top_out)
 	->
 		set__insert(ValueVars0, Var, ValueVars1)
@@ -1116,73 +1446,44 @@ find_value_giving_occurrences([Var | Vars], LiveInfo, InstMapDelta,
 		ValueVars1, ValueVars).
 
 %-----------------------------------------------------------------------------%
-%-----------------------------------------------------------------------------%
 
-:- type live_info	--->	live_info(
-					module_info,
-					proc_info,
-					bool,		% Do we use typeinfo
-							% liveness for this
-							% proc?
-					map(prog_var, type),
-					prog_varset
-				).
 
-:- pred live_info_init(module_info, proc_info, bool, map(prog_var, type),
-	prog_varset, live_info).
-:- mode live_info_init(in, in, in, in, in, out) is det.
-
-live_info_init(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, Varset,
-	live_info(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, Varset)).
-
-:- pred live_info_get_module_info(live_info, module_info).
-:- mode live_info_get_module_info(in, out) is det.
-
-live_info_get_module_info(live_info(ModuleInfo, _, _, _, _), ModuleInfo).
-
-:- pred live_info_get_proc_info(live_info, proc_info).
-:- mode live_info_get_proc_info(in, out) is det.
-
-live_info_get_proc_info(live_info(_, ProcInfo, _, _, _), ProcInfo).
-
-:- pred live_info_get_typeinfo_liveness(live_info, bool).
-:- mode live_info_get_typeinfo_liveness(in, out) is det.
-
-live_info_get_typeinfo_liveness(live_info(_, _, TypeInfoLiveness, _, _),
-	TypeInfoLiveness).
-
-:- pred live_info_get_var_types(live_info, map(prog_var, type)).
-:- mode live_info_get_var_types(in, out) is det.
-
-live_info_get_var_types(live_info(_, _, _, VarTypes, _), VarTypes).
-
-:- pred live_info_get_varset(live_info, prog_varset).
-:- mode live_info_get_varset(in, out) is det.
-
-live_info_get_varset(live_info(_, _, _, _, Varset), Varset).
-
-%-----------------------------------------------------------------------------%
-
-	% Get the nonlocals, and, if doing alternate liveness, add the
+	% Get the nonlocals, and, if doing typeinfo liveness, add the
 	% typeinfo vars for the nonlocals.
 
-:- pred liveness__get_nonlocals_and_typeinfos(live_info, hlds_goal_info,
-		set(prog_var)).
-:- mode liveness__get_nonlocals_and_typeinfos(in, in, out) is det.
+:- pred liveness__get_nonlocals_and_typeinfos(live_info::in,
+	hlds_goal_info::in, set(prog_var)::out, set(prog_var)::out) is det.
 
 liveness__get_nonlocals_and_typeinfos(LiveInfo, GoalInfo, 
-		NonLocals) :-
-	goal_info_get_nonlocals(GoalInfo, NonLocals0),
-	live_info_get_typeinfo_liveness(LiveInfo, TypeinfoLiveness),
-	( 
-		TypeinfoLiveness = yes
-	->
-		live_info_get_proc_info(LiveInfo, ProcInfo),
-		proc_info_get_typeinfo_vars_setwise(ProcInfo, NonLocals0,
-			TypeInfoVarsNonLocals),
-		set__union(NonLocals0, TypeInfoVarsNonLocals, NonLocals)
-	;
-		NonLocals = NonLocals0
-	).
+		NonLocals, CompletedNonLocals) :-
+	goal_info_get_code_gen_nonlocals(GoalInfo, NonLocals),
+	liveness__maybe_complete_with_typeinfos(LiveInfo,
+		NonLocals, CompletedNonLocals).
+
+:- pred liveness__maybe_complete_with_typeinfos(live_info::in,
+	set(prog_var)::in, set(prog_var)::out) is det.
+
+liveness__maybe_complete_with_typeinfos(LiveInfo, Vars0, Vars) :-
+	proc_info_maybe_complete_with_typeinfo_vars(Vars0,
+		LiveInfo ^ typeinfo_liveness, LiveInfo ^ vartypes,
+		LiveInfo ^ type_info_varmap, Vars).
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
+:- type live_info
+	--->	live_info(
+			module_info		::	module_info,
+			typeinfo_liveness	::	bool,
+			vartypes		::	vartypes,
+			type_info_varmap	::	type_info_varmap,
+			varset			::	prog_varset
+		).
+
+:- pred live_info_init(module_info::in, bool::in, vartypes::in,
+	type_info_varmap::in, prog_varset::in, live_info::out) is det.
+
+live_info_init(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, VarSet,
+	live_info(ModuleInfo, ProcInfo, TypeInfoLiveness, VarTypes, VarSet)).
 
 %-----------------------------------------------------------------------------%
