@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1993-1999 The University of Melbourne.
+% Copyright (C) 1993-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -16,44 +16,44 @@
 
 :- interface.
 
-:- import_module hlds_pred, hlds_data, tree, prog_data, (inst).
-:- import_module builtin_ops.
+:- import_module prog_data, (inst).
+:- import_module hlds_pred, hlds_goal, hlds_data.
+:- import_module code_model, rtti, layout, builtin_ops.
+:- import_module tree.
 
-:- import_module bool, assoc_list, list, map, set, std_util.
+:- import_module bool, assoc_list, list, map, set, std_util, counter.
 
 %-----------------------------------------------------------------------------%
 
-:- type code_model
-	--->	model_det		% functional & total
-	;	model_semi		% just functional
-	;	model_non.		% not functional
-
-%-----------------------------------------------------------------------------%
-
-% c_interface_info holds information used when generating
-% code that uses the C interface.
-:- type c_interface_info
-	---> c_interface_info(
+% foreign_interface_info holds information used when generating
+% code that uses the foreign language interface.
+:- type foreign_interface_info
+	---> foreign_interface_info(
 		module_name,
 		% info about stuff imported from C:
-		c_header_info,
-		c_body_info,
+		foreign_decl_info,
+		foreign_body_info,
 		% info about stuff exported to C:
-		c_export_decls,
-		c_export_defns
+		foreign_export_decls,
+		foreign_export_defns
 	).
 
-:- type c_header_info 	==	list(c_header_code).	% in reverse order
-:- type c_body_info 	==	list(c_body_code).	% in reverse order
+:- type foreign_decl_info ==	list(foreign_decl_code).	
+		% in reverse order
+:- type foreign_body_info   ==	list(foreign_body_code).
+		% in reverse order
 
-:- type c_header_code	==	pair(string, prog_context).
-:- type c_body_code	==	pair(string, prog_context).
+:- type foreign_decl_code	--->	
+		foreign_decl_code(foreign_language, string, prog_context).
+:- type foreign_body_code	--->
+		foreign_body_code(foreign_language, string, prog_context).
 
-:- type c_export_defns == list(c_export).
-:- type c_export_decls == list(c_export_decl).
+:- type foreign_export_defns == list(foreign_export).
+:- type foreign_export_decls == list(foreign_export_decl).
 
-:- type c_export_decl
-	---> c_export_decl(
+:- type foreign_export_decl
+	---> foreign_export_decl(
+		foreign_language,	% language of the export
 		string,		% return type
 		string,		% function name
 		string		% argument declarations
@@ -61,7 +61,7 @@
 
 	% the code for `pragma export' is generated directly as strings
 	% by export.m.
-:- type c_export	==	string.
+:- type foreign_export	==	string.
 
 %-----------------------------------------------------------------------------%
 
@@ -80,6 +80,9 @@
 :- pred global_data_update_proc_layout(global_data::in,
 	pred_proc_id::in, proc_layout_info::in, global_data::out) is det.
 
+:- pred global_data_add_new_closure_layouts(global_data::in,
+	list(comp_gen_c_data)::in, global_data::out) is det.
+
 :- pred global_data_add_new_non_common_static_datas(global_data::in,
 	list(comp_gen_c_data)::in, global_data::out) is det.
 
@@ -95,6 +98,9 @@
 :- pred global_data_get_all_proc_layouts(global_data::in,
 	list(proc_layout_info)::out) is det.
 
+:- pred global_data_get_all_closure_layouts(global_data::in,
+	list(comp_gen_c_data)::out) is det.
+
 :- pred global_data_get_all_non_common_static_data(global_data::in,
 	list(comp_gen_c_data)::out) is det.
 
@@ -106,19 +112,20 @@
 :- type c_file	
 	--->	c_file(
 			module_name,
-			c_header_info,
-			list(user_c_code),
-			list(c_export),
+			foreign_decl_info,
+			list(user_foreign_code),
+			list(foreign_export),
 			list(comp_gen_c_var),
 			list(comp_gen_c_data),
 			list(comp_gen_c_module)
 		).
 
-	% Some C code from a `pragma c_code' declaration that is not
+	% Some code from a `pragma foreign_code' declaration that is not
 	% associated with a given procedure.
-:- type user_c_code
-	--->	user_c_code(
-			string,			% C code
+:- type user_foreign_code
+	--->	user_foreign_code(
+			foreign_language,	% language of this code
+			string,			% code
 			term__context		% source code location
 		).
 
@@ -150,6 +157,12 @@
 						% arguments of the create.
 			list(pred_proc_id)	% The procedures referenced.
 						% Used by dead_proc_elim.
+		)
+	;	rtti_data(
+			rtti_data
+		)
+	;	layout_data(
+			layout_data
 		).
 
 :- type comp_gen_c_module
@@ -163,9 +176,20 @@
 			string,			% predicate name
 			int,			% arity
 			pred_proc_id,		% the pred_proc_id this code
-			list(instruction)	% the code for this procedure
+			list(instruction),	% the code for this procedure
+			proc_label,		% proc_label of this procedure
+			counter,		% source for new label numbers
+			contains_reconstruction	% value numbering needs
+						% to handle goals that
+						% perform structure reuse
+						% specially.
 		).
 
+:- type contains_reconstruction
+	--->	contains_reconstruction
+	;	does_not_contain_reconstruction
+	.
+
 :- type llds_proc_id	==	int.
 
 	% we build up instructions as trees and then flatten
@@ -229,15 +253,18 @@
 			% specified by lval.
 
 	;	call(code_addr, code_addr, list(liveinfo), term__context,
-				call_model)
+				goal_path, call_model)
 			% call(Target, Continuation, _, _, _) is the same as
 			% succip = Continuation; goto(Target).
 			% The third argument is the live value info for the
 			% values live on return. The fourth argument gives
-			% the context of the call. The last gives the model
-			% of the called procedure, and if it is nondet,
-			% says whether tail recursion elimination is
-			% potentially applicable to the call.
+			% the context of the call. The fifth gives the goal
+			% path of the call in the body of the procedure; it is
+			% meaningful only if execution tracing is enabled.
+			% The last gives the code model of the called
+			% procedure, and if it is model_non, says whether
+			% tail recursion elimination is potentially applicable
+			% to the call.
 
 	;	mkframe(nondet_frame_info, code_addr)
 			% mkframe(NondetFrameInfo, CodeAddr) creates a nondet
@@ -293,12 +320,23 @@
 			% The effect is to deallocate all the memory which
 			% was allocated since that call to mark_hp.
 
+	;	free_heap(rval)
+			% Notify the garbage collector that the heap space
+			% associated with the top-level cell of the rval is
+			% no longer needed.
+			% `free' is useless but harmless without conservative
+			% garbage collection.
+
 	;	store_ticket(lval)
 			% Allocate a new "ticket" and store it in the lval.
+			%
+			% Operational semantics:
+			% 	MR_ticket_counter = ++MR_ticket_high_water;
+			% 	lval = MR_trail_ptr;
 
 	;	reset_ticket(rval, reset_trail_reason)
 			% The rval must specify a ticket allocated with
-			% `store_ticket' and not yet invalidated or
+			% `store_ticket' and not yet invalidated, pruned or
 			% deallocated.
 			% If reset_trail_reason is `undo', `exception', or
 			% `retry', restore any mutable global state to the
@@ -313,24 +351,58 @@
 			% Note that we do not discard trail entries after
 			% commits, because that would in general be unsafe.
 			%
-			% Any invalidated ticket is useless and should
-			% be deallocated with either `discard_ticket'
-			% or `discard_tickets_to'.
+			% Any invalidated ticket which has not yet been
+			% backtracked over should be pruned with
+			% `prune_ticket' or `prune_tickets_to'.
+			% Any invalidated ticket which has been backtracked
+			% over is useless and should be deallocated with
+			% `discard_ticket'.
+			%
+			% Operational semantics:
+			% 	MR_untrail_to(rval, reset_trail_reason);
+
+	;	prune_ticket
+			% Invalidates the most-recently allocated ticket.
+			%
+			% Operational semantics:
+			%	--MR_ticket_counter;
 
 	;	discard_ticket
 			% Deallocates the most-recently allocated ticket.
+			%
+			% Operational semantics:
+			%	MR_ticket_high_water = --MR_ticket_counter;
 
 	;	mark_ticket_stack(lval)
 			% Tell the trail sub-system to store a ticket counter
-			% (for later use in discard_tickets_upto)
+			% (for later use in prune_tickets_to)
 			% in the specified lval.
+			%
+			% Operational semantics:
+			%	lval = MR_ticket_counter;
 
-	;	discard_tickets_to(rval)
+	;	prune_tickets_to(rval)
+			% The rval must be a ticket counter obtained via
+			% `mark_ticket_stack' and not yet invalidated.
+			% Prunes any trail tickets allocated after
+			% the corresponding call to mark_ticket_stack.
+			% Invalidates any later ticket counters.
+			%
+			% Operational semantics:
+			%	MR_ticket_counter = rval;
+
+%	;	discard_tickets_to(rval)	% this is only used in
+						% the hand-written code in
+						% library/exception.m
 			% The rval must be a ticket counter obtained via
 			% `mark_ticket_stack' and not yet invalidated.
 			% Deallocates any trail tickets allocated after
 			% the corresponding call to mark_ticket_stack.
 			% Invalidates any later ticket counters.
+			%
+			% Operational semantics:
+			%	MR_ticket_counter = rval;
+			%	MR_ticket_high_water = MR_ticket_counter;
 
 	;	incr_sp(int, string)
 			% Increment the det stack pointer. The string is
@@ -343,7 +415,7 @@
 
 	;	pragma_c(list(pragma_c_decl), list(pragma_c_component),
 				may_call_mercury, maybe(label), maybe(label),
-				bool)
+				maybe(label), bool)
 			% The first argument says what local variable
 			% declarations are required for the following
 			% components, which in turn can specify how
@@ -365,16 +437,19 @@
 			% refer to a Mercury label. If they do, we must
 			% prevent the label from being optimized away.
 			% To make it known to labelopt, we mention it in
-			% the fourth or the fifth arg. The fourth argument
+			% the fourth, fifth or sixth arg. The fourth argument
 			% may give the name of a label whose name is fixed
-			% (e.g. because it embedded in raw C code or because it
-			% has associated an label layout structure), while
-			% the fifth may give the name of a label that can
-			% be changed (because it is not mentioned in C code
-			% and has no associated layout structure, being
-			% mentioned only in pragma_c_fail_to components).
+			% because it embedded in raw C code, and which does
+			% not have a layout structure. The fifth argument
+			% may give the name of a label whose name is fixed
+			% because it does have an associated label layout
+			% structure (it may appear in C code as well).
+			% The sixth argument may give the name of a label
+			% that can be changed (because it is not mentioned
+			% in C code and has no associated layout structure,
+			% being mentioned only in pragma_c_fail_to components).
 			%
-			% The sixth argument says whether the contents
+			% The seventh argument says whether the contents
 			% of the pragma C code can refer to stack slots.
 			% User-written shouldn't refer to stack slots,
 			% the question is whether the compiler-generated
@@ -497,7 +572,7 @@
 
 	% A pragma_c_output represents the code that stores one of
 	% of the outputs for a pragma_c instruction.
-:- type pragma_c_output  
+:- type pragma_c_output
 	--->	pragma_c_output(lval, type, string).
 				% where to put the output val, type and name
 				% of variable containing the output val
@@ -566,7 +641,7 @@
 
 	% live_value_type describes the different sorts of data that
 	% can be considered live.
-:- type live_value_type 
+:- type live_value_type
 	--->	succip				% A stored succip.
 	;	curfr				% A stored curfr.
 	;	maxfr				% A stored maxfr.
@@ -720,9 +795,9 @@
 		% stage after code generation.
 
 	;	create(tag, list(maybe(rval)), create_arg_types,
-			static_or_dynamic, int, string)
+			static_or_dynamic, int, string, maybe(rval))
 		% create(Tag, Arguments, MaybeArgTypes, StaticOrDynamic,
-		%	LabelNumber, CellKind):
+		%	LabelNumber, CellKind, CellToReuse):
 		% A `create' instruction is used during code generation
 		% for creating a term, either on the heap or
 		% (if the term is constant) as a static constant.
@@ -749,10 +824,13 @@
 		% we can construct the term at compile-time
 		% and just reference the label.
 		%
-		% The last argument gives the name of the type constructor
+		% The string argument gives the name of the type constructor
 		% of the function symbol of which this is a cell, for use
 		% in memory profiling.
 		%
+		% The maybe(rval) contains the location of a cell to reuse.
+		% This will always be `no' after code generation.
+		%
 		% For the time being, you must leave the argument types
 		% implicit if the cell is to be unique. This is because
 		% (a) the code generator assumes that each argument of a cell
@@ -821,45 +899,31 @@
 	;	code_addr_const(code_addr)
 	;	data_addr_const(data_addr)
 	;	label_entry(label).
-			% the address of the label (uses ENTRY macro).
+			% the address of the label (uses MR_ENTRY macro).
 
 :- type data_addr
-	--->	data_addr(module_name, data_name).
+	--->	data_addr(module_name, data_name)
 			% module name; which var
+	;	rtti_addr(rtti_type_id, rtti_name)
+			% type id; which var
+	;	layout_addr(layout_name).
 
 :- type data_name
 	--->	common(int)
-	;	type_ctor(base_data, string, arity)
-			% base_data, type name, type arity
 	;	base_typeclass_info(class_id, string)
 			% class name & class arity, names and arities of the
 			% types
-	;	module_layout
-			% Layout information for the current module.
-	;	proc_layout(label)
-			% Layout structure for the procedure with the given
-			% entry label.
-	;	internal_layout(label)
-			% Layout structure for the given internal label.
 	;	tabling_pointer(proc_label).
 			% A variable that contains a pointer that points to
 			% the table used to implement memoization, loopcheck
 			% or minimal model semantics for the given procedure.
 
-:- type base_data
-	--->	info
-			% basic information, including special preds
-	;	layout
-			% layout information
-	;	functors.
-			% information on functors
-
 :- type reg_type	
 	--->	r		% general-purpose (integer) regs
 	;	f.		% floating point regs
 
 :- type label
-	--->	local(proc_label, int)	% not proc entry; internal to a
+	--->	local(int, proc_label)	% not proc entry; internal to a
 					% procedure
 	;	c_local(proc_label)	% proc entry; internal to a C module
 	;	local(proc_label)	% proc entry; internal to a Mercury
@@ -897,7 +961,7 @@
 	;	do_aditi_delete
 	;	do_aditi_bulk_insert
 	;	do_aditi_bulk_delete
-	;	do_aditi_modify
+	;	do_aditi_bulk_modify
 	;	do_not_reached.		% We should never jump to this address.
 
 	% A proc_label is a label used for the entry point to a procedure.
@@ -935,53 +999,59 @@
 	% to avoid unnecessary boxing/unboxing of floats.
 :- type llds_type
 	--->	bool		% A boolean value
-				% represented using the C type `Integer'.
+				% represented using the C type `MR_Integer'.
 	;	int_least8	% A signed value that fits that contains
 				% at least eight bits, represented using the
-				% C type int_least8_t. Intended for use in
+				% C type MR_int_least8_t. Intended for use in
 				% static data declarations, not for data
 				% that gets stored in registers, stack slots
 				% etc.
 	;	uint_least8	% An unsigned version of int_least8,
-				% represented using the C type uint_least8_t.
+				% represented using the C type
+				% MR_uint_least8_t.
 	;	int_least16	% A signed value that fits that contains
 				% at least sixteen bits, represented using the
-				% C type int_least16_t. Intended for use in
+				% C type MR_int_least16_t. Intended for use in
 				% static data declarations, not for data
 				% that gets stored in registers, stack slots
 				% etc.
 	;	uint_least16	% An unsigned version of int_least16,
-				% represented using the C type uint_least16_t.
+				% represented using the C type
+				% MR_uint_least16_t.
 	;	int_least32	% A signed value that fits that contains
 				% at least 32 bits, represented using the
-				% C type int_least32_t. Intended for use in
+				% C type MR_int_least32_t. Intended for use in
 				% static data declarations, not for data
 				% that gets stored in registers, stack slots
 				% etc.
 	;	uint_least32	% An unsigned version of intleast_32,
 				% represented using the C type uint_least32_t.
 	;	integer		% A Mercury `int', represented in C as a
-				% value of type `Integer' (which is
+				% value of type `MR_Integer' (which is
 				% a signed integral type of the same
 				% size as a pointer).
-	;	unsigned	% Something whose C type is `Unsigned'
-				% (the unsigned equivalent of `Integer').
+	;	unsigned	% Something whose C type is `MR_Unsigned'
+				% (the unsigned equivalent of `MR_Integer').
 	;	float		% A Mercury `float', represented in C as a
-				% value of type `Float' (which may be either
+				% value of type `MR_Float' (which may be either
 				% `float' or `double', but is usually
 				% `double').
 	;	string		% A Mercury string; represented in C as a
-				% value of type `String'.
+				% value of type `MR_String'.
 	;	data_ptr	% A pointer to data; represented in C
-				% as a value of C type `Word *'.
+				% as a value of C type `MR_Word *'.
 	;	code_ptr	% A pointer to code; represented in C
-				% as a value of C type `Code *'.
+				% as a value of C type `MR_Code *'.
 	;	word.		% Something that can be assigned to a value
-				% of C type `Word', i.e., something whose
+				% of C type `MR_Word', i.e., something whose
 				% size is a word but which may be either
 				% signed or unsigned
 				% (used for registers, stack slots, etc).
 
+:- func llds__stack_slot_num_to_lval(code_model, int) = lval.
+
+:- pred llds__wrap_rtti_data(rtti_data::in, comp_gen_c_data::out) is det.
+
 	% given a non-var rval, figure out its type
 :- pred llds__rval_type(rval::in, llds_type::out) is det.
 
@@ -1010,9 +1080,23 @@
 	% (floats may be bigger than a word, but if so, they are boxed)
 :- pred llds__type_is_word_size_as_arg(llds_type::in, bool::out) is det.
 
+:- func get_proc_label(label) = proc_label.
+
+:- func get_defining_module_name(proc_label) = module_name.
+
 :- implementation.
+
 :- import_module require.
 
+llds__stack_slot_num_to_lval(CodeModel, SlotNum) =
+	(if CodeModel = model_non then
+		framevar(SlotNum)
+	else
+		stackvar(SlotNum)
+	).
+
+llds__wrap_rtti_data(RttiData, rtti_data(RttiData)).
+
 llds__lval_type(reg(RegType, _), Type) :-
 	llds__register_type(RegType, Type).
 llds__lval_type(succip, code_ptr).
@@ -1038,7 +1122,7 @@ llds__rval_type(lval(Lval), Type) :-
 	llds__lval_type(Lval, Type).
 llds__rval_type(var(_), _) :-
 	error("var unexpected in llds__rval_type").
-llds__rval_type(create(_, _, _, _, _, _), data_ptr).
+llds__rval_type(create(_, _, _, _, _, _, _), data_ptr).
 	%
 	% Note that create and mkword must both be of type data_ptr,
 	% not of type word, to ensure that static consts containing
@@ -1075,8 +1159,6 @@ llds__unop_return_type(tag, word).
 llds__unop_return_type(unmktag, word).
 llds__unop_return_type(mkbody, word).
 llds__unop_return_type(unmkbody, word).
-llds__unop_return_type(body, word).
-llds__unop_return_type(cast_to_unsigned, unsigned).
 llds__unop_return_type(hash_string, integer).
 llds__unop_return_type(bitwise_complement, integer).
 llds__unop_return_type(not, bool).
@@ -1086,8 +1168,6 @@ llds__unop_arg_type(tag, word).
 llds__unop_arg_type(unmktag, word).
 llds__unop_arg_type(mkbody, word).
 llds__unop_arg_type(unmkbody, word).
-llds__unop_arg_type(body, word).
-llds__unop_arg_type(cast_to_unsigned, word).
 llds__unop_arg_type(hash_string, word).
 llds__unop_arg_type(bitwise_complement, integer).
 llds__unop_arg_type(not, bool).
@@ -1117,6 +1197,7 @@ llds__binop_return_type((<), bool).
 llds__binop_return_type((>), bool).
 llds__binop_return_type((<=), bool).
 llds__binop_return_type((>=), bool).
+llds__binop_return_type(unsigned_le, bool).
 llds__binop_return_type(float_plus, float).
 llds__binop_return_type(float_minus, float).
 llds__binop_return_type(float_times, float).
@@ -1127,6 +1208,7 @@ llds__binop_return_type(float_lt, bool).
 llds__binop_return_type(float_gt, bool).
 llds__binop_return_type(float_le, bool).
 llds__binop_return_type(float_ge, bool).
+llds__binop_return_type(body, word).
 
 llds__register_type(r, word).
 llds__register_type(f, float).
@@ -1161,6 +1243,14 @@ llds__type_is_word_size_as_arg(data_ptr,     yes).
 llds__type_is_word_size_as_arg(code_ptr,     yes).
 llds__type_is_word_size_as_arg(word,         yes).
 
+get_proc_label(exported(ProcLabel)) = ProcLabel.
+get_proc_label(local(ProcLabel)) = ProcLabel.
+get_proc_label(c_local(ProcLabel)) = ProcLabel.
+get_proc_label(local(_, ProcLabel)) = ProcLabel.
+
+get_defining_module_name(proc(ModuleName, _, _, _, _, _)) = ModuleName.
+get_defining_module_name(special_proc(ModuleName, _, _, _, _, _)) = ModuleName.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -1169,105 +1259,85 @@ llds__type_is_word_size_as_arg(word,         yes).
 
 :- type global_data
 	--->	global_data(
-			proc_var_map,		% Information about the global
+			proc_var_map		:: proc_var_map,
+						% Information about the global
 						% variables defined by each
 						% procedure.
-			proc_layout_map,	% Information about the
+			proc_layout_map		:: proc_layout_map,
+						% Information about the
 						% layout structures defined
 						% by each procedure.
-			list(comp_gen_c_data)	% The list of global data
+			closure_layouts		:: list(comp_gen_c_data),
+						% The list of all closure
+						% layouts generated in this
+						% module. While all closure
+						% layouts are different from
+						% all other comp_gen_c_datas,
+						% it is possible, although
+						% unlikely, for two closures
+						% to have the same layout.
+			non_common_data		:: list(comp_gen_c_data)
+						% The list of global data
 						% structures that do not need
 						% to be checked by llds_common,
 						% because their construction
 						% ensures no overlaps.
 		).
 
-global_data_init(global_data(EmptyDataMap, EmptyLayoutMap, [])) :-
+global_data_init(global_data(EmptyDataMap, EmptyLayoutMap, [], [])) :-
 	map__init(EmptyDataMap),
 	map__init(EmptyLayoutMap).
 
-global_data_add_new_proc_var(GlobalData0, PredProcId, ProcVar,
-		GlobalData) :-
-	global_data_get_proc_var_map(GlobalData0, ProcVarMap0),
+global_data_add_new_proc_var(GlobalData0, PredProcId, ProcVar, GlobalData) :-
+	ProcVarMap0 = GlobalData0 ^ proc_var_map,
 	map__det_insert(ProcVarMap0, PredProcId, ProcVar, ProcVarMap),
-	global_data_set_proc_var_map(GlobalData0, ProcVarMap,
-		GlobalData).
+	GlobalData = GlobalData0 ^ proc_var_map := ProcVarMap.
 
 global_data_add_new_proc_layout(GlobalData0, PredProcId, ProcLayout,
 		GlobalData) :-
-	global_data_get_proc_layout_map(GlobalData0, ProcLayoutMap0),
+	ProcLayoutMap0 = GlobalData0 ^ proc_layout_map,
 	map__det_insert(ProcLayoutMap0, PredProcId, ProcLayout, ProcLayoutMap),
-	global_data_set_proc_layout_map(GlobalData0, ProcLayoutMap,
-		GlobalData).
+	GlobalData = GlobalData0 ^ proc_layout_map := ProcLayoutMap.
 
 global_data_update_proc_layout(GlobalData0, PredProcId, ProcLayout,
 		GlobalData) :-
-	global_data_get_proc_layout_map(GlobalData0, ProcLayoutMap0),
+	ProcLayoutMap0 = GlobalData0 ^ proc_layout_map,
 	map__det_update(ProcLayoutMap0, PredProcId, ProcLayout, ProcLayoutMap),
-	global_data_set_proc_layout_map(GlobalData0, ProcLayoutMap,
-		GlobalData).
+	GlobalData = GlobalData0 ^ proc_layout_map := ProcLayoutMap.
+
+global_data_add_new_closure_layouts(GlobalData0, NewClosureLayouts,
+		GlobalData) :-
+	ClosureLayouts0 = GlobalData0 ^ closure_layouts,
+	list__append(NewClosureLayouts, ClosureLayouts0, ClosureLayouts),
+	GlobalData = GlobalData0 ^ closure_layouts := ClosureLayouts.
 
 global_data_add_new_non_common_static_datas(GlobalData0, NewNonCommonStatics,
 		GlobalData) :-
-	global_data_get_non_common_static_data(GlobalData0, NonCommonStatics0),
+	NonCommonStatics0 = GlobalData0 ^ non_common_data,
 	list__append(NewNonCommonStatics, NonCommonStatics0, NonCommonStatics),
-	global_data_set_non_common_static_data(GlobalData0, NonCommonStatics,
-		GlobalData).
+	GlobalData = GlobalData0 ^ non_common_data := NonCommonStatics.
 
-global_data_maybe_get_proc_layout(GlobalData0, PredProcId, ProcLayout) :-
-	global_data_get_proc_layout_map(GlobalData0, ProcLayoutMap),
+global_data_maybe_get_proc_layout(GlobalData, PredProcId, ProcLayout) :-
+	ProcLayoutMap = GlobalData ^ proc_layout_map,
 	map__search(ProcLayoutMap, PredProcId, ProcLayout).
 
-global_data_get_proc_layout(GlobalData0, PredProcId, ProcLayout) :-
-	global_data_get_proc_layout_map(GlobalData0, ProcLayoutMap),
+global_data_get_proc_layout(GlobalData, PredProcId, ProcLayout) :-
+	ProcLayoutMap = GlobalData ^ proc_layout_map,
 	map__lookup(ProcLayoutMap, PredProcId, ProcLayout).
 
 global_data_get_all_proc_vars(GlobalData, ProcVars) :-
-	global_data_get_proc_var_map(GlobalData, ProcVarMap),
+	ProcVarMap = GlobalData ^ proc_var_map,
 	map__values(ProcVarMap, ProcVars).
 
 global_data_get_all_proc_layouts(GlobalData, ProcLayouts) :-
-	global_data_get_proc_layout_map(GlobalData, ProcLayoutMap),
+	ProcLayoutMap = GlobalData ^ proc_layout_map,
 	map__values(ProcLayoutMap, ProcLayouts).
 
+global_data_get_all_closure_layouts(GlobalData, ClosureLayouts) :-
+	ClosureLayouts = GlobalData ^ closure_layouts.
+
 global_data_get_all_non_common_static_data(GlobalData, NonCommonStatics) :-
-	global_data_get_non_common_static_data(GlobalData, NonCommonStatics).
-
-%-----------------------------------------------------------------------------%
-
-:- pred global_data_get_proc_var_map(global_data::in, proc_var_map::out)
-	is det.
-:- pred global_data_get_proc_layout_map(global_data::in, proc_layout_map::out)
-	is det.
-:- pred global_data_get_non_common_static_data(global_data::in,
-	list(comp_gen_c_data)::out) is det.
-:- pred global_data_set_proc_var_map(global_data::in, proc_var_map::in,
-	global_data::out) is det.
-:- pred global_data_set_proc_layout_map(global_data::in, proc_layout_map::in,
-	global_data::out) is det.
-:- pred global_data_set_non_common_static_data(global_data::in,
-	list(comp_gen_c_data)::in, global_data::out) is det.
-
-global_data_get_proc_var_map(GD, A) :-
-	GD = global_data(A, _, _).
-
-global_data_get_proc_layout_map(GD, B) :-
-	GD = global_data(_, B, _).
-
-global_data_get_non_common_static_data(GD, C) :-
-	GD = global_data(_, _, C).
-
-global_data_set_proc_var_map(GD0, A, GD) :-
-	GD0 = global_data(_, B, C),
-	GD  = global_data(A, B, C).
-
-global_data_set_proc_layout_map(GD0, B, GD) :-
-	GD0 = global_data(A, _, C),
-	GD  = global_data(A, B, C).
-
-global_data_set_non_common_static_data(GD0, C, GD) :-
-	GD0 = global_data(A, B, _),
-	GD  = global_data(A, B, C).
+	NonCommonStatics = GlobalData ^ non_common_data.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
