@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -27,6 +27,7 @@
 
 :- import_module builtin_ops, hlds_module, hlds_data, prog_data, prog_out.
 :- import_module code_gen, unify_gen, code_util, code_aux, opt_util.
+:- import_module code_model.
 
 :- import_module bool, set, int, std_util, tree, list, assoc_list, require.
 :- import_module string.
@@ -140,9 +141,7 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	code_info__get_arginfo(ArgModes),
 	code_info__get_headvars(HeadVars),
 	{ assoc_list__from_corresponding_lists(HeadVars, ArgModes, Args) },
-	code_info__setup_call(Args, callee, EpilogCode),
-
-	{ code_util__output_args(Args, LiveArgs) },
+	code_info__setup_return(Args, LiveArgs, EpilogCode),
 
 	{ BaseCode = tree(BaseGoalCode, tree(BaseSaveCode, EpilogCode)) },
 	{ RecCode = tree(RecGoalCode, tree(RecSaveCode, EpilogCode)) },
@@ -320,7 +319,7 @@ middle_rec__generate_downloop_test([Instr0 | Instrs0], Target, Instrs) :-
 middle_rec__split_rec_code([], _, _) :-
 	error("did not find call in middle_rec__split_rec_code").
 middle_rec__split_rec_code([Instr0 | Instrs1], Before, After) :-
-	( Instr0 = call(_, _, _, _, _) - _ ->
+	( Instr0 = call(_, _, _, _, _, _) - _ ->
 		(
 			opt_util__skip_comments(Instrs1, Instrs2),
 			Instrs2 = [Instr2 | Instrs3],
@@ -397,7 +396,7 @@ middle_rec__find_used_registers_instr(block(_, _, Instrs), Used0, Used) :-
 middle_rec__find_used_registers_instr(assign(Lval, Rval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used1),
 	middle_rec__find_used_registers_rval(Rval, Used1, Used).
-middle_rec__find_used_registers_instr(call(_, _, _, _, _), Used, Used).
+middle_rec__find_used_registers_instr(call(_, _, _, _, _, _), Used, Used).
 middle_rec__find_used_registers_instr(mkframe(_, _), Used, Used).
 middle_rec__find_used_registers_instr(label(_), Used, Used).
 middle_rec__find_used_registers_instr(goto(_), Used, Used).
@@ -413,18 +412,21 @@ middle_rec__find_used_registers_instr(mark_hp(Lval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used).
 middle_rec__find_used_registers_instr(restore_hp(Rval), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
+middle_rec__find_used_registers_instr(free_heap(Rval), Used0, Used) :-
+	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(store_ticket(Lval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used).
 middle_rec__find_used_registers_instr(reset_ticket(Rval, _Rsn), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(discard_ticket, Used, Used).
+middle_rec__find_used_registers_instr(prune_ticket, Used, Used).
 middle_rec__find_used_registers_instr(mark_ticket_stack(Lval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used).
-middle_rec__find_used_registers_instr(discard_tickets_to(Rval), Used0, Used) :-
+middle_rec__find_used_registers_instr(prune_tickets_to(Rval), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(incr_sp(_, _), Used, Used).
 middle_rec__find_used_registers_instr(decr_sp(_), Used, Used).
-middle_rec__find_used_registers_instr(pragma_c(_, Components, _, _, _, _),
+middle_rec__find_used_registers_instr(pragma_c(_, Components, _, _, _, _, _),
 		Used0, Used) :-
 	middle_rec__find_used_registers_components(Components, Used0, Used).
 middle_rec__find_used_registers_instr(init_sync_term(Lval, _), Used0, Used) :-
@@ -492,9 +494,9 @@ middle_rec__find_used_registers_rval(Rval, Used0, Used) :-
 		Rval = var(_),
 		error("var found in middle_rec__find_used_registers_rval")
 	;
-		Rval = create(_, MaybeRvals, _, _, _, _),
-		middle_rec__find_used_registers_maybe_rvals(MaybeRvals,
-			Used0, Used)
+		Rval = create(_, MaybeRvals, _, _, _, _, Reuse),
+		middle_rec__find_used_registers_maybe_rvals(
+			[Reuse | MaybeRvals], Used0, Used)
 	;
 		Rval = mkword(_, Rval1),
 		middle_rec__find_used_registers_rval(Rval1, Used0, Used)
