@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1999 The University of Melbourne.
+% Copyright (C) 1999-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -66,17 +66,23 @@
 % MLDS function signatures are determined by the HLDS procedure's
 % argument types, modes, and determinism.
 % Procedures arguments with type `io__state' or `store(_)' are not passed.
+% Procedure arguments with top_unused modes are not passed.
 % Procedures arguments with top_in modes are passed as input.
-% Procedures arguments with top_out modes may be returned by value
-% or alternatively may be passed by reference.
-% Procedure arguments with top_unused modes should not be passed
-% (except possibly in the case where they are polymorphically typed,
-% and even then, depending on how polymorphism is handled).]
+% Procedures arguments with top_out modes are normally passed by reference.
+% However, several alternative approaches are also supported (see below).
+%
 % Procedures with determinism model_det need no special handling.
 % Procedures with determinism model_semi must return a boolean.
 % Procedures with determinism model_non get passed a continuation;
 % if the procedure succeeds, it must call the continuation, and if
 % it fails, it must return.
+%
+% With the `--copy-out' option, arguments with top_out modes will be returned
+% by value.  This requires the target language to support multiple return
+% values.  The MLDS->target code generator can of course handle that by mapping
+% functions with multiple return values into functions that returns a struct.
+% With the `--nondet-copy-out' option, arguments for nondet procedures with
+% top_out modes will be passed as arguments to the continuation.
 
 % 4. Variables
 %
@@ -261,9 +267,12 @@
 
 :- interface.
 
-:- import_module hlds_pred, hlds_data, prog_data, builtin_ops.
+:- import_module hlds_module, hlds_pred, hlds_data.
+:- import_module prog_data, builtin_ops, rtti.
+:- import_module type_util.
 
-% To avoid duplication, we use a few things from the LLDS.
+% To avoid duplication, we use a few things from the LLDS
+% (specifically stuff for the C interface).
 % It would be nice to avoid this dependency...
 :- import_module llds.
 
@@ -314,6 +323,11 @@
 % MLDS package.
 :- func mlds_module_name_to_sym_name(mlds__package_name) = sym_name.
 
+% Given an MLDS module name (e.g. `foo.bar'), append another class qualifier
+% (e.g. for a class `baz'), and return the result (e.g. `foo.bar.baz').
+% The `arity' argument specifies the arity of the class.
+:- func mlds__append_class_qualifier(mlds_module_name, mlds__class_name, arity) =
+	mlds_module_name.
 
 :- type mlds__defns == list(mlds__defn).
 :- type mlds__defn
@@ -363,6 +377,7 @@
 				% for additional information.
 			pred_id			% Specifies the HLDS pred_id.
 		)
+	;	export(string)	% A pragma export name.
 	.
 
 :- type mlds__func_sequence_num == int.
@@ -380,7 +395,7 @@
 		% constants or variables
 	--->	mlds__data(
 			mlds__type,
-			maybe(mlds__initializer)
+			mlds__initializer
 		)
 		% functions
 	;	mlds__function(
@@ -389,29 +404,42 @@
 			mlds__func_params,	% the arguments & return types
 			maybe(mlds__statement)	% the function body, or `no'
 						% if the function is abstract
+						% or if the function is defined
+						% externally (i.e. the original
+						% Mercury procedure was declared
+						% `:- external').
 		)
 		% packages, classes, interfaces, structs, enums
 	;	mlds__class(
 			mlds__class_defn
 		).
 
-:- type mlds__initializer == list(mlds__rval).
+	% Note that `one_copy' variables *must* have an initializer
+	% (the GCC back-end relies on this).
+:- type mlds__initializer
+	--->	init_obj(mlds__rval)
+	;	init_struct(list(mlds__initializer))
+	;	init_array(list(mlds__initializer))
+	;	no_initializer
+	.
 
 :- type mlds__func_params
 	---> mlds__func_params(
 		mlds__arguments,	% names and types of arguments (inputs)
-		list(mlds__type)	% types of return values (outputs)
+		mlds__return_types	% types of return values (outputs)
 	).
 
 :- type mlds__arguments == assoc_list(mlds__entity_name, mlds__type).
+:- type mlds__arg_types == list(mlds__type).
+:- type mlds__return_types == list(mlds__type).
 
 	% An mlds__func_signature is like an mlds__func_params
 	% except that it only includes the function's type, not
 	% the parameter names.
 :- type mlds__func_signature
 	---> mlds__func_signature(
-		list(mlds__type),	% argument types
-		list(mlds__type)	% return types
+		mlds__arg_types,	% argument types
+		mlds__return_types	% return types
 	).
 
 :- func mlds__get_func_signature(mlds__func_params) = mlds__func_signature.
@@ -442,12 +470,14 @@
 
 :- type mlds__class_defn
 	---> mlds__class_defn(
-		mlds__class_kind,
-		mlds__imports,			% imports these classes (or
+		kind	::	mlds__class_kind,
+		imports	::	mlds__imports,	% imports these classes (or
 						% modules, packages, ...)
-		list(mlds__class_id),		% inherits these base classes
-		list(mlds__interface_id),	% implements these interfaces
-		mlds__defns			% contains these members
+		inherits ::	list(mlds__class_id),
+						% inherits these base classes
+		implements ::	list(mlds__interface_id),
+						% implements these interfaces
+		members ::	mlds__defns	% contains these members
 	).
 
 	% Note: the definition of the `mlds__type' type is subject to change.
@@ -455,11 +485,15 @@
 	% switching on this type.
 :- type mlds__type
 	--->	% Mercury data types
-		mercury_type(prog_data__type)
+		mercury_type(
+			prog_data__type,	% the exact Mercury type
+			builtin_type		% what kind of type it is:
+						% enum, float, etc.
+		)
 
 		% The type for the continuation functions used
 		% to handle nondeterminism
-	;	mlds__cont_type
+	;	mlds__cont_type(mlds__return_types)
 
 		% The type used for storing information about a commit.
 		% This may be `jmp_buf' or `__label__'.
@@ -469,19 +503,35 @@
 		% These are the builtin types of the MLDS target language,
 		% whatever that may be.
 		% Currently we don't actually use many of these.
-	;	mlds__bool_type
-	;	mlds__int_type
-	;	mlds__float_type
-	;	mlds__char_type
+	;	mlds__native_bool_type
+	;	mlds__native_int_type
+	;	mlds__native_float_type
+	;	mlds__native_char_type
 
 		% MLDS types defined using mlds__class_defn
-	;	mlds__class_type(mlds__class, arity)	% name, arity
+	;	mlds__class_type(
+			mlds__class,		% name
+			arity,
+			mlds__class_kind
+		)
+
+		% MLDS array types.
+		% These are single-dimensional, and can be indexed
+		% using the `field' lval with an `offset' field_id;
+		% indices start at zero.
+		% Currently these are used for static constants
+		% that would otherwise be allocated with a `new_object'
+		% statement.
+	;	mlds__array_type(mlds__type)
 
 		% Pointer types.
 		% Currently these are used for handling output arguments.
 	;	mlds__ptr_type(mlds__type)
 
 		% Function types.
+		% For the C back-end, these are mapped to function
+		% pointer types, since C doesn't have first-class
+		% function types.
 	;	mlds__func_type(mlds__func_params)
 
 		% A generic type (e.g. `Word') that can hold any Mercury value.
@@ -492,22 +542,27 @@
 		% that can be used to point to the environment
 		% (set of local variables) of the containing function.
 		% This is used for handling nondeterminism,
-		% if the target language doesn't supported
+		% if the target language doesn't support
 		% nested functions, and also for handling
 		% closures for higher-order code.
 	;	mlds__generic_env_ptr_type
 
-	;	mlds__base_type_info_type.
+	;	mlds__pseudo_type_info_type
+	
+	;	mlds__rtti_type(rtti_name).
 
 :- type mercury_type == prog_data__type.
 
-:- func mercury_type_to_mlds_type(mercury_type) = mlds__type.
+:- func mercury_type_to_mlds_type(module_info, mercury_type) = mlds__type.
 
 % Hmm... this is tentative.
 :- type mlds__class_id == mlds__type.
 :- type mlds__interface_id == mlds__type.
 
 %-----------------------------------------------------------------------------%
+%
+% Declaration flags
+%
 
 :- type mlds__decl_flags.
 
@@ -522,6 +577,9 @@
 	--->	one_copy	% i.e. "static" storage duration
 				% (but not necessarily static linkage)
 				% or static member function
+				% Note that `one_copy' variables
+				% *must* have an initializer
+				% (the GCC back-end relies on this.)
 	;	per_instance.	% i.e. "auto" local variable in function,
 				% or non-static member of class.
 
@@ -559,23 +617,40 @@
 		constness, abstractness) = mlds__decl_flags.
 
 %-----------------------------------------------------------------------------%
+%
+% Foreign code interfacing
+%
 
 	%
-	% C code required for the C interface.
-	% When compiling to a language other than C,
-	% this part still needs to be generated as C code
-	% and compiled with a C compiler.
+	% Foreign code required for the foreign language interface.
+	% When compiling to a language other than the foreign language,
+	% this part still needs to be generated as C (or whatever) code
+	% and compiled with a C (or whatever) compiler.
 	%
 :- type mlds__foreign_code
 	---> mlds__foreign_code(
-		c_header_info,
-		list(user_c_code),
-		list(c_export)		% XXX we will need to modify
-					% export.m to handle different
-					% target languages
+		foreign_decl_info,
+		list(user_foreign_code),
+		list(mlds__pragma_export)
 	).
 
+	%
+	% Information required to generate code for each
+	% `pragma export'.
+	%
+:- type mlds__pragma_export
+	---> ml_pragma_export(
+		string,			% Exported name
+		mlds__qualified_entity_name, % MLDS name for exported entity
+		mlds__func_params,	% MLDS function parameters
+		mlds__context
+	).
+
+
 %-----------------------------------------------------------------------------%
+%
+% Contexts (i.e. source code locations)
+%
 
 	% mlds__context is probably == prog_context,
 	% but might also contain goal_path or other information.
@@ -586,6 +661,9 @@
 :- func mlds__get_prog_context(mlds__context) = prog_context.
 
 %-----------------------------------------------------------------------------%
+%
+% Statements
+%
 
 :- type mlds__statements == list(mlds__statement).
 
@@ -615,19 +693,37 @@
 	%
 	% selection (see also computed_goto)
 	%
+
 	;	if_then_else(mlds__rval, mlds__statement,
 			maybe(mlds__statement))
 
-/******
-	% Is it worth including this?  We already have `computed_goto'...
+		% This representation for switches is very general:
+		% it allows switching on any type, and cases can match
+		% on ranges as well as on values.
+		% Many target languages only allow switches on ints or enums.
+		% Some (e.g. C#) also allow switches on strings.
+		% Most target languages only allow matching on values;
+		% only some (e.g. GNU C) allow matching on ranges.
+		% The MLDS code generator should only generate switches
+		% that the target will support.
+		%
+		% Note that unlike C, MLDS cases do NOT fall through; if you
+		% want to achieve that effect, you need to use an explicit goto.
 	;	switch(
+			% The value to switch on
+			mlds__type,
 			mlds__rval,
 
-			% other representations might be better...
-			assoc_list(mlds__rval, mlds__statement),
-			mlds__statement		% the default case
+			% The range of possible values which the
+			% value might take (if known)
+			mlds__switch_range,
+
+			% The different cases
+			mlds__switch_cases,
+
+			% What to do if none of the cases match
+			mlds__switch_default
 		)
-******/
 
 	%
 	% transfer of control
@@ -688,6 +784,16 @@
 		% do_commit(Ref) instructions should only be used
 		% in goals called from the GoalToTry goal in the
 		% try_commit instruction with the same Ref.
+		%
+		% The C and GCC back-ends require each try_commit
+		% to be put in its own nested function, to avoid
+		% problems with setjmp() and local vars not declared
+		% volatile.  They also require each do_commit
+		% to be put in its own function -- this is needed when
+		% using __builtin_setjmp()/__builtin_longjmp() to
+		% ensure that the call to __builtin_longjmp() is
+		% not in the same function as the call to
+		% __builtin_setjmp().
 		%	
 	;	try_commit(mlds__lval, mlds__statement, mlds__statement)
 	;	do_commit(mlds__rval)
@@ -729,14 +835,79 @@ XXX Full exception handling support is not yet implemented.
 	
 	.
 
+%-----------------------------------------------------------------------------%
+%
+% Extra info for switches
+%
+	% The range of possible values which the
+	% switch variable might take (if known)
+:- type mlds__switch_range
+	--->	range_unknown
+	;	range(range_min::int, range_max::int).
+			% From range_min to range_max, inclusive.
+
+	% Each switch case consists of the conditions to match against,
+	% and the statement to execute if the match succeeds.
+	% Unlike C, cases do NOT fall through; if you want to achieve that
+	% effect, you need to use an explicit goto.
+:- type mlds__switch_cases == list(mlds__switch_case).
+:- type mlds__switch_case == pair(mlds__case_match_conds, mlds__statement).
+
+	% case_match_conds should be a _non-empty_ list of conditions;
+	% if _any_ of the conditions match, this case will be selected.
+:- type mlds__case_match_conds == list(mlds__case_match_cond).
+
+	% A case_match_cond specifies when a switch case will be selected
+:- type mlds__case_match_cond
+	--->	match_value(mlds__rval)		% match_value(Val) matches if
+						% the switch value is equal to
+						% the specified Val
+
+	;	match_range(mlds__rval, mlds__rval).  % match_range(Min, Max)
+						% matches if the switch value
+						% is between Min and Max,
+						% inclusive.
+						% Note that this should only be
+						% used if the target supports
+						% it; currently the C back-end
+						% supports this only if you're
+						% using the GNU C compiler.
+
+	% The switch_default specifies what to do if none of the switch
+	% conditions match.
+:- type mlds__switch_default
+	--->	default_is_unreachable		% The switch is exhaustive,
+						% so the default case should
+						% never be reached.
+
+	;	default_do_nothing		% The default action is to
+						% just fall through to the
+						% statement after the switch.
+
+	;	default_case(mlds__statement).	% The default is to execute
+						% the specified statement.
+
+%-----------------------------------------------------------------------------%
+%
+% Extra info for labels
+%
 
 :- type mlds__label == string.
 
+%-----------------------------------------------------------------------------%
+%
+% Extra info for calls
+%
+
 :- type is_tail_call
 	--->	tail_call	% a tail call
 	;	call		% just an ordinary call
 	.
 
+%-----------------------------------------------------------------------------%
+%
+% Extra info for exception handling
+%
 
 	% XXX This is tentative -- the current definition may be
 	% a bit too specific to C++-style exceptions.
@@ -753,6 +924,7 @@ XXX Full exception handling support is not yet implemented.
 				% if `no', then exception value will not be used
 		).
 
+%-----------------------------------------------------------------------------%
 
 	%
 	% atomic statements
@@ -771,7 +943,10 @@ XXX Full exception handling support is not yet implemented.
 	% heap management
 	%
 
-		% XXX the following is still quite tentative
+	;	delete_object(mlds__lval)
+			% Compile time garbage collect (ie explicitly
+			% deallocate) the memory used by the lval.
+
 			% new_object(Target, Tag, Type,
 			%	Size, CtorName, Args, ArgTypes):
 			% Allocate a memory block of the given size,
@@ -792,10 +967,7 @@ XXX Full exception handling support is not yet implemented.
 			maybe(mlds__rval),
 					% The amount of memory that needs to
 					% be allocated for the new object,
-					% measured in bytes.
-					% (XXX would it be better to measure
-					% this in bits or words rather than
-					% bytes?)
+					% measured in words (NOT bytes!).
 			maybe(ctor_name),
 					% The name of the constructor to
 					% invoke.
@@ -803,7 +975,17 @@ XXX Full exception handling support is not yet implemented.
 					% The arguments to the constructor.
 			list(mlds__type)
 					% The types of the arguments to the
-					% constructor.
+					% constructor. 
+					%
+					% Note that currently we store all 
+					% fields as type mlds__generic_type.
+					% But the type here is the actual
+					% argument type, which does not
+					% have to be mlds__generic_type.
+					% It is the responsibility of the
+					% MLDS->target code output phase
+					% to box the arguments if necessary.
+					% 
 		)
 
 	;	mark_hp(mlds__lval)
@@ -832,15 +1014,14 @@ XXX Full exception handling support is not yet implemented.
 	% foreign language interfacing
 	%
 
-	;	target_code(target_lang, string)
-			% Do whatever is specified by the string,
-			% which can be any piece of code in the specified
-			% target language (C, assembler, or whatever)
-			% that does not have any non-local flow of control.
-
+	;	target_code(target_lang, list(target_code_component))
+			% Do whatever is specified by the
+			% target_code_components, which can be any piece
+			% of code in the specified target language (C,
+			% assembler, or whatever) that does not have any
+			% non-local flow of control.
 	.
 
-
 	%
 	% This is just a random selection of possible languages
 	% that we might want to target...
@@ -854,50 +1035,40 @@ XXX Full exception handling support is not yet implemented.
 	;	lang_java_bytecode
 	.
 
+:- type target_code_component
+	--->	user_target_code(string, maybe(prog_context))
+			% user_target_code holds C code from
+			% the user's `pragma c_code' declaration
+	;	raw_target_code(string)
+			% raw_target_code holds C code that the
+			% compiler has generated.  To ensure that
+			% following `#line' directives work OK,
+			% either the string in a raw_target_code must
+			% end in `\n' (or `\n' followed by whitespace),
+			% or the following target_code_component must be 
+			% a `name(Name)' component, for which we do not
+			% output #line directives.
+	;	target_code_input(mlds__rval)
+	;	target_code_output(mlds__lval)
+	;	name(mlds__qualified_entity_name)
+	.
+
 	% XXX I'm not sure what representation we should use here
 :- type ctor_name == string.
 
 	%
 	% trail management
+	% For documentation, see the corresponding LLDS instructions
+	% in llds.m.
 	%
 :- type trail_op
-
 	--->	store_ticket(mlds__lval)
-			% Allocate a new "ticket" and store it in the lval.
-
 	;	reset_ticket(mlds__rval, mlds__reset_trail_reason)
-			% The rval must specify a ticket allocated with
-			% `store_ticket' and not yet invalidated or
-			% deallocated.
-			% If undo_reason is `undo' or `exception', restore
-			% any mutable global state to the state it was in when
-			% the ticket was obtained with store_ticket();
-			% invalidates any tickets allocated after this one.
-			% If undo_reason is `commit' or `solve', leave the state
-			% unchanged, just check that it is safe to commit
-			% to this solution (i.e. that there are no outstanding
-			% delayed goals -- this is the "floundering" check).
-			% Note that we do not discard trail entries after
-			% commits, because that would in general be unsafe.
-			%
-			% Any invalidated ticket is useless and should
-			% be deallocated with either `discard_ticket'
-			% or `discard_tickets_to'.
-
 	;	discard_ticket
-			% Deallocates the most-recently allocated ticket.
-
+	;	prune_ticket
 	;	mark_ticket_stack(mlds__lval)
-			% Tell the trail sub-system to store a ticket counter
-			% (for later use in discard_tickets_upto)
-			% in the specified lval.
-
-	;	discard_tickets_to(mlds__rval)
-			% The rval must be a ticket counter obtained via
-			% `mark_ticket_stack' and not yet invalidated.
-			% Deallocates any trail tickets allocated after
-			% the corresponding call to mark_ticket_stack.
-			% Invalidates any later ticket counters.
+	;	prune_tickets_to(mlds__rval)
+% 	;	discard_tickets_to(mlds__rval)	% used only by the library
 	.
 
 %-----------------------------------------------------------------------------%
@@ -910,9 +1081,17 @@ XXX Full exception handling support is not yet implemented.
 	--->		% offset(N) represents the field
 			% at offset N Words.
 	 	offset(mlds__rval)
-	;		% named_field(Name) represents the field
-			% with the specified name.
-		named_field(mlds__fully_qualified_name(field_name))
+	;		% named_field(Name, CtorType) represents the field
+			% with the specified name.  The CtorType gives the
+			% MLDS type for this particular constructor.
+			% The type of the object is given by the PtrType
+			% in the field(..) lval; CtorType may either be
+			% the same as PtrType, or it may be a pointer to
+			% a derived class.  In the latter case, the
+			% MLDS->target code back-end is responsible
+			% for inserting a downcast from PtrType to CtorType
+			% before accessing the field.
+		named_field(mlds__fully_qualified_name(field_name), mlds__type)
 	.
 
 :- type field_name == string.
@@ -933,8 +1112,10 @@ XXX Full exception handling support is not yet implemented.
 	% values on the heap
 	% or fields of a structure
 	%
-	--->	field(maybe(mlds__tag), mlds__rval, field_id)
-				% field(Tag, Address, FieldName)
+	--->	field(maybe(mlds__tag), mlds__rval, field_id, 
+			mlds__type, mlds__type)
+				% field(Tag, Address, FieldName, FieldType,
+				%	PtrType)
 				% selects a field of a compound term.
 				% Address is a tagged pointer to a cell
 				% on the heap; the offset into the cell
@@ -944,13 +1125,29 @@ XXX Full exception handling support is not yet implemented.
 				% The value of the tag should be given if
 				% it is known, since this will lead to
 				% faster code.
+				% The FieldType is the type of the field.
+				% The PtrType is the type of the pointer
+				% from which we are fetching the field.
+				%
+				% Note that currently we store all fields
+				% of objects created with new_object
+				% as type mlds__generic_type. For such objects,
+				% the type here should be mlds__generic_type,
+				% not the actual type of the field.
+				% If the actual type is different, then it
+				% is the HLDS->MLDS code generator's
+				% responsibility to insert the necessary
+				% code to handle boxing/unboxing.
 
 	%
 	% values somewhere in memory
 	% this is the deference operator (e.g. unary `*' in C)
 	%
-	;	mem_ref(mlds__rval)	% The rval should have
-				% originally come from a mem_addr rval.
+	;	mem_ref(mlds__rval, mlds__type)	
+				% The rval should have originally come
+				% from a mem_addr rval.
+				% The type is the type of the value being
+				% dereferenced
 
 	%
 	% variables
@@ -990,6 +1187,11 @@ XXX Full exception handling support is not yet implemented.
 :- type mlds__unary_op
 	--->	box(mlds__type)
 	;	unbox(mlds__type)
+			% cast(MLDSType):
+			% Coerce the type of the rval to be MLDSType.
+			% XXX it might be worthwhile adding the 
+			% type that we cast from.
+	;	cast(mlds__type)
 	;	std_unop(builtin_ops__unary_op).
 
 :- type mlds__rval_const
@@ -1004,11 +1206,21 @@ XXX Full exception handling support is not yet implemented.
 			% first null character.
 	;	multi_string_const(int, string)
 	;	code_addr_const(mlds__code_addr)
-	;	data_addr_const(mlds__data_addr).
+	;	data_addr_const(mlds__data_addr)
+		% A null value, of the given type. 
+		% Usually the type will be a pointer (mlds__ptr_type)
+		% but it could also be string or a func_type.
+		% (Null is not a valid value of type string
+		% or func_type, but null values of those types
+		% may be useful as placeholders in cases where the
+		% value will never be used.)
+	; 	null(mlds__type).
+
 
 :- type mlds__code_addr
-	--->	proc(mlds__qualified_proc_label)
-	;	internal(mlds__qualified_proc_label, mlds__func_sequence_num).
+	--->	proc(mlds__qualified_proc_label, mlds__func_signature)
+	;	internal(mlds__qualified_proc_label, mlds__func_sequence_num,
+			mlds__func_signature).
 
 :- type mlds__data_addr
 	--->	data_addr(mlds_module_name, mlds__data_name).
@@ -1022,13 +1234,16 @@ XXX Full exception handling support is not yet implemented.
 			% global constants.  These are called "common"
 			% because they may be common sub-expressions.
 	%
-	% Stuff for handling polymorphism and type classes
+	% Stuff for handling polymorphism/RTTI and type classes.
 	%
-	;	type_ctor(mlds__base_data, string, arity)
-			% base_data, type name, type arity
-	;	base_typeclass_info(hlds_data__class_id, string)
-			% class name & class arity, names and arities of the
-			% types
+	;	rtti(rtti_type_id, rtti_name)
+	;	base_typeclass_info(
+			hlds_data__class_id,	% class name & class arity,
+			string			% a mangled string that encodes
+						% the names and arities of the
+						% types in the instance
+						% declaration
+		)
 	%
 	% Stuff for handling debugging and accurate garbage collection.
 	% (Those features are not yet implemented for the MLDS back-end,
@@ -1051,10 +1266,10 @@ XXX Full exception handling support is not yet implemented.
 %-----------------------------------------------------------------------------%
 
 %
-% Note: the types `tag', `base_data', and `reset_trail_reason' here are all
+% Note: the types `tag' and `reset_trail_reason' here are all
 % defined exactly the same as the ones in llds.m.  The definitions are
 % duplicated here because we don't want mlds.m to depend on llds.m.
-% (Alternatively, we could move all these definitions into a new module
+% (Alternatively, we could move both these definitions into a new module
 % imported by both mlds.m and llds.m, but these definitions are small enough
 % and simple enough that I don't think it is worth creating a new module
 % just for them.)
@@ -1063,18 +1278,6 @@ XXX Full exception handling support is not yet implemented.
 	% A tag should be a small non-negative integer.
 :- type tag == int.
 
-	% See the definition in llds.m for comments about the meaning
-	% of the `base_data' type.
-	% For some targets, the target language and runtime system might
-	% provide all the necessary information about type layouts,
-	% in which case we won't need to define the type_functors and
-	% type_layout stuff, and we may also be able to use the language's
-	% RTTI rather than defining the type_infos ourselves.
-:- type base_data
-	--->	info
-	;	functors
-	;	layout.
-
 	% see runtime/mercury_trail.h
 :- type reset_trail_reason
 	--->	undo
@@ -1127,7 +1330,8 @@ XXX Full exception handling support is not yet implemented.
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module int, term, require.
+:- import_module modules.
+:- import_module int, term, string, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -1145,10 +1349,15 @@ mlds__get_prog_context(mlds__context(Context)) = Context.
 
 %-----------------------------------------------------------------------------%
 
-% Currently mlds__types are just the same as Mercury types.
-% XXX something more complicated may be needed here...
+% Currently we return mlds__types that are just the same as Mercury types,
+% except that we also store the type category, so that we
+% can tell if the type is an enumeration or not, without
+% needing to refer to the HLDS type_table.
+% XXX It might be a better idea to get rid of the mercury_type/2
+% MLDS type and instead fully convert all Mercury types to MLDS types.
 
-mercury_type_to_mlds_type(Type) = mercury_type(Type).
+mercury_type_to_mlds_type(ModuleInfo, Type) = mercury_type(Type, Category) :-
+	classify_type(Type, ModuleInfo, Category).
 
 %-----------------------------------------------------------------------------%
 
@@ -1175,63 +1384,13 @@ mercury_module_name_to_mlds(MercuryModule) = MLDS_Package :-
 		MLDS_Package = MercuryModule
 	).
 
-:- pred mercury_std_library_module(string::in) is semidet.
-mercury_std_library_module("array").
-mercury_std_library_module("assoc_list").
-mercury_std_library_module("bag").
-mercury_std_library_module("benchmarking").
-mercury_std_library_module("bimap").
-mercury_std_library_module("bintree").
-mercury_std_library_module("bintree_set").
-mercury_std_library_module("bool").
-mercury_std_library_module("bt_array").
-mercury_std_library_module("builtin").
-mercury_std_library_module("char").
-mercury_std_library_module("dir").
-mercury_std_library_module("eqvclass").
-mercury_std_library_module("exception").
-mercury_std_library_module("float").
-mercury_std_library_module("gc").
-mercury_std_library_module("getopt").
-mercury_std_library_module("graph").
-mercury_std_library_module("group").
-mercury_std_library_module("int").
-mercury_std_library_module("integer").
-mercury_std_library_module("io").
-mercury_std_library_module("lexer").
-mercury_std_library_module("library").
-mercury_std_library_module("list").
-mercury_std_library_module("map").
-mercury_std_library_module("math").
-mercury_std_library_module("mercury_builtin").
-mercury_std_library_module("multi_map").
-mercury_std_library_module("ops").
-mercury_std_library_module("parser").
-mercury_std_library_module("pqueue").
-mercury_std_library_module("private_builtin").
-mercury_std_library_module("prolog").
-mercury_std_library_module("queue").
-mercury_std_library_module("random").
-mercury_std_library_module("rational").
-mercury_std_library_module("rbtree").
-mercury_std_library_module("relation").
-mercury_std_library_module("require").
-mercury_std_library_module("set").
-mercury_std_library_module("set_bbbtree").
-mercury_std_library_module("set_ordlist").
-mercury_std_library_module("set_unordlist").
-mercury_std_library_module("stack").
-mercury_std_library_module("std_util").
-mercury_std_library_module("store").
-mercury_std_library_module("string").
-mercury_std_library_module("term").
-mercury_std_library_module("term_io").
-mercury_std_library_module("time").
-mercury_std_library_module("tree234").
-mercury_std_library_module("varset").
-
 mlds_module_name_to_sym_name(MLDS_Package) = MLDS_Package.
 
+mlds__append_class_qualifier(Package, ClassName, ClassArity) =
+		qualified(Package, ClassQualifier) :-
+	string__format("%s_%d", [s(ClassName), i(ClassArity)],
+		ClassQualifier).
+
 %-----------------------------------------------------------------------------%
 
 %
