@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -133,7 +133,8 @@ modecheck_unification(X0, functor(ConsId0, ArgVars0), Unification0,
 		% fields created by polymorphism.m
 		Unification0 \= construct(_, code_addr_const(_, _),
 			_, _, _, _, _),
-		Unification0 \= deconstruct(_, code_addr_const(_, _), _, _, _)
+		Unification0 \= deconstruct(_,
+				code_addr_const(_, _), _, _, _, _)
 	->
 		%
 		% convert the pred term to a lambda expression
@@ -355,7 +356,7 @@ modecheck_unify_lambda(X, PredOrFunc, ArgVars, LambdaModes,
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
 	mode_info_get_instmap(ModeInfo0, InstMap0),
 	instmap__lookup_var(InstMap0, X, InstOfX),
-	InstOfY = ground(unique, yes(LambdaPredInfo)),
+	InstOfY = ground(unique, higher_order(LambdaPredInfo)),
 	LambdaPredInfo = pred_inst_info(PredOrFunc, LambdaModes, LambdaDet),
 	(
 		abstractly_unify_inst(dead, InstOfX, InstOfY, real_unify,
@@ -623,7 +624,7 @@ split_complicated_subunifies(Unification0, ArgVars0,
 				Unification, ArgVars, ExtraGoals) -->
 	(
 		{ Unification0 = deconstruct(X, ConsId, ArgVars0, ArgModes0,
-			Det) }
+			Det, CanCGC) }
 	->
 		(
 			split_complicated_subunifies_2(ArgVars0, ArgModes0,
@@ -632,7 +633,7 @@ split_complicated_subunifies(Unification0, ArgVars0,
 			{ ExtraGoals = ExtraGoals1 },
 			{ ArgVars = ArgVars1 },
 			{ Unification = deconstruct(X, ConsId, ArgVars,
-							ArgModes0, Det) }
+					ArgModes0, Det, CanCGC) }
 		;
 			{ error("split_complicated_subunifies_2 failed") }
 		)
@@ -877,13 +878,17 @@ modecheck_complicated_unify(X, Y, Type, ModeOfX, ModeOfY, Det, UnifyContext,
 		ModeInfo2 = ModeInfo0
 	;
 		list__length(UnifyTypeInfoVars, NumTypeInfoVars),
-		list__duplicate(NumTypeInfoVars, ground(shared, no),
+		list__duplicate(NumTypeInfoVars, ground(shared, none),
 			ExpectedInsts),
 		mode_info_set_call_context(unify(UnifyContext),
 			ModeInfo0, ModeInfo1),
+		NeedExactMatch = no,
 		InitialArgNum = 0,
 		modecheck_var_has_inst_list(UnifyTypeInfoVars, ExpectedInsts,
-			InitialArgNum, ModeInfo1, ModeInfo2)
+			NeedExactMatch, InitialArgNum, _InstVarSub,
+			ModeInfo1, ModeInfo2)
+			% We can ignore _InstVarSub since type_info variables
+			% should not have variable insts.
 	),
 
 	mode_info_get_module_info(ModeInfo2, ModuleInfo2),
@@ -957,7 +962,8 @@ modecheck_complicated_unify(X, Y, Type, ModeOfX, ModeOfY, Det, UnifyContext,
 		type_to_type_id(Type, TypeId, _)
 	->
 		mode_info_get_context(ModeInfo2, Context),
-		unify_proc__request_unify(TypeId - UniMode,
+		mode_info_get_instvarset(ModeInfo2, InstVarSet),
+		unify_proc__request_unify(TypeId - UniMode, InstVarSet,
 			Det, Context, ModuleInfo2, ModuleInfo),
 		mode_info_set_module_info(ModeInfo2, ModuleInfo,
 			ModeInfo)
@@ -985,7 +991,7 @@ categorize_unify_var_lambda(ModeOfX, ArgModes0, X, ArgVars,
 	( Unification0 = construct(_, ConsId0, _, _, _, _, AditiInfo0) ->
 		AditiInfo = AditiInfo0,
 		ConsId = ConsId0
-	; Unification0 = deconstruct(_, ConsId1, _, _, _) ->
+	; Unification0 = deconstruct(_, ConsId1, _, _, _, _) ->
 		AditiInfo = no,
 		ConsId = ConsId1
 	;
@@ -1038,7 +1044,7 @@ categorize_unify_var_lambda(ModeOfX, ArgModes0, X, ArgVars,
 			RHS = RHS0
 		),
 		Unification = construct(X, ConsId, ArgVars, ArgModes,
-			no, cell_is_unique, AditiInfo),
+			construct_dynamically, cell_is_unique, AditiInfo),
 		ModeInfo = ModeInfo0
 	;
 		instmap__is_reachable(InstMap)
@@ -1082,7 +1088,7 @@ categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
 	% if we are re-doing mode analysis, preserve the existing cons_id
 	( Unification0 = construct(_, ConsId0, _, _, _, _, _) ->
 		ConsId = ConsId0
-	; Unification0 = deconstruct(_, ConsId1, _, _, _) ->
+	; Unification0 = deconstruct(_, ConsId1, _, _, _, _) ->
 		ConsId = ConsId1
 	;
 		ConsId = NewConsId
@@ -1093,10 +1099,9 @@ categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
 		mode_is_output(ModuleInfo, ModeOfX)
 	->
 		% It's a construction.
-		ReuseVar = no,
 		RLExprnId = no,
 		Unification = construct(X, ConsId, ArgVars, ArgModes,
-			ReuseVar, cell_is_unique, RLExprnId),
+			construct_dynamically, cell_is_unique, RLExprnId),
 
 		% For existentially quantified data types,
 		% check that any type_info or type_class_info variables in the
@@ -1149,7 +1154,8 @@ categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
 				ModeInfo = ModeInfo0
 			)
 		),
-		Unification = deconstruct(X, ConsId, ArgVars, ArgModes, CanFail)
+		Unification = deconstruct(X, ConsId, ArgVars,
+				ArgModes, CanFail, no)
 	).
 
 	% Check that any type_info or type_class_info variables
@@ -1167,9 +1173,10 @@ check_type_info_args_are_ground([ArgVar | ArgVars], VarTypes, UnifyContext)
 		{ is_introduced_type_info_type(ArgType) }
 	->
 		mode_info_set_call_context(unify(UnifyContext)),
+		{ NeedExactMatch = no },
 		{ InitialArgNum = 0 },
-		modecheck_var_has_inst_list([ArgVar], [ground(shared, no)],
-			InitialArgNum),
+		modecheck_var_has_inst_list([ArgVar], [ground(shared, none)],
+			NeedExactMatch, InitialArgNum, _InstVarSub),
 		check_type_info_args_are_ground(ArgVars, VarTypes,
 			UnifyContext)
 	;
@@ -1184,7 +1191,7 @@ check_type_info_args_are_ground([ArgVar | ArgVars], VarTypes, UnifyContext)
 bind_args(not_reached, _) -->
 	{ instmap__init_unreachable(InstMap) },
 	mode_info_set_instmap(InstMap).
-bind_args(ground(Uniq, no), Args) -->
+bind_args(ground(Uniq, none), Args) -->
 	ground_args(Uniq, Args).
 bind_args(bound(_Uniq, List), Args) -->
 	( { List = [] } ->
@@ -1209,7 +1216,7 @@ bind_args_2([Arg | Args], [Inst | Insts]) -->
 
 ground_args(_Uniq, []) --> [].
 ground_args(Uniq, [Arg | Args]) -->
-	modecheck_set_var_inst(Arg, ground(Uniq, no)),
+	modecheck_set_var_inst(Arg, ground(Uniq, none)),
 	ground_args(Uniq, Args).
 
 %-----------------------------------------------------------------------------%
@@ -1227,8 +1234,8 @@ get_mode_of_args(not_reached, ArgInsts, ArgModes) :-
 	mode_set_args(ArgInsts, not_reached, ArgModes).
 get_mode_of_args(any(Uniq), ArgInsts, ArgModes) :-
 	mode_set_args(ArgInsts, any(Uniq), ArgModes).
-get_mode_of_args(ground(Uniq, no), ArgInsts, ArgModes) :-
-	mode_set_args(ArgInsts, ground(Uniq, no), ArgModes).
+get_mode_of_args(ground(Uniq, none), ArgInsts, ArgModes) :-
+	mode_set_args(ArgInsts, ground(Uniq, none), ArgModes).
 get_mode_of_args(bound(_Uniq, List), ArgInstsA, ArgModes) :-
 	( List = [] ->
 		% the code is unreachable
