@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -146,7 +146,8 @@ a variable live if its value will be used later on in the computation.
 
 	% Mode-check or unique-mode-check the code for all the predicates
 	% in a module.
-:- pred check_pred_modes(how_to_check_goal, may_change_called_proc,				module_info, module_info, bool, io__state, io__state).
+:- pred check_pred_modes(how_to_check_goal, may_change_called_proc,
+		module_info, module_info, bool, io__state, io__state).
 :- mode check_pred_modes(in, in, in, out, out, di, uo) is det.
 
 	% Mode-check or unique-mode-check the code for single predicate.
@@ -190,12 +191,6 @@ a variable live if its value will be used later on in the computation.
 
 	% Modecheck a unification.
 
- 	% given the right-hand-side of a unification, return a list of
-	% the potentially non-local variables of that unification.
-	%
-:- pred unify_rhs_vars(unify_rhs, list(prog_var)).
-:- mode unify_rhs_vars(in, out) is det.
-
 	% Given a list of variables, and a list of livenesses,
 	% select the live variables.
 	%
@@ -214,20 +209,23 @@ a variable live if its value will be used later on in the computation.
 
 	% Given a list of variables and a list of expected liveness, ensure
 	% that the inst of each variable satisfies the corresponding expected
-	% liveness.
+	% liveness.  If the bool argument is `yes', then require an exact
+	% match.
 	%
-:- pred modecheck_var_list_is_live(list(prog_var), list(is_live), int,
+:- pred modecheck_var_list_is_live(list(prog_var), list(is_live), bool, int,
 		mode_info, mode_info).
-:- mode modecheck_var_list_is_live(in, in, in, mode_info_di, mode_info_uo)
+:- mode modecheck_var_list_is_live(in, in, in, in, mode_info_di, mode_info_uo)
 	is det.
 
 	% Given a list of variables and a list of initial insts, ensure
 	% that the inst of each variable matches the corresponding initial
-	% inst.
+	% inst.  If the bool argument is `yes', then we require an exact
+	% match (using inst_matches_final), otherwise we allow the var
+	% to be more instantiated than the inst (using inst_matches_initial).
 	%
-:- pred modecheck_var_has_inst_list(list(prog_var), list(inst), int,
-		mode_info, mode_info).
-:- mode modecheck_var_has_inst_list(in, in, in, mode_info_di, mode_info_uo)
+:- pred modecheck_var_has_inst_list(list(prog_var), list(inst), bool, int,
+		inst_var_sub, mode_info, mode_info).
+:- mode modecheck_var_has_inst_list(in, in, in, in, out, mode_info_di, mode_info_uo)
 	is det.
 
 :- pred modecheck_set_var_inst(prog_var, inst, mode_info, mode_info).
@@ -259,6 +257,16 @@ a variable live if its value will be used later on in the computation.
 :- pred modecheck_functor_test(prog_var, cons_id, mode_info, mode_info).
 :- mode modecheck_functor_test(in, in, mode_info_di, mode_info_uo) is det.
 
+	% compute_goal_instmap_delta(InstMap0, Goal,
+	%	GoalInfo0, GoalInfo, ModeInfo0, ModeInfo).
+	%
+	% Work out the instmap_delta for a goal from
+	% the instmaps before and after the goal.
+:- pred compute_goal_instmap_delta(instmap, hlds_goal_expr,
+		hlds_goal_info, hlds_goal_info, mode_info, mode_info).
+:- mode compute_goal_instmap_delta(in, in, in, out,
+		mode_info_di, mode_info_uo) is det.
+
 %-----------------------------------------------------------------------------%
 
 % The following predicates are used by modecheck_unify.m.
@@ -316,7 +324,7 @@ a variable live if its value will be used later on in the computation.
 
 :- import_module make_hlds, hlds_data, unique_modes, mode_debug.
 :- import_module mode_info, delay_info, mode_errors, inst_match, instmap.
-:- import_module type_util, mode_util, code_util, unify_proc.
+:- import_module type_util, mode_util, code_util, unify_proc, special_pred.
 :- import_module globals, options, mercury_to_mercury, hlds_out, int, set.
 :- import_module passes_aux, typecheck, module_qual, clause_to_proc.
 :- import_module modecheck_unify, modecheck_call, inst_util, purity.
@@ -354,14 +362,14 @@ check_pred_modes(WhatToCheck, MayChangeCalledProc,
 		UnsafeToContinue),
 	( { WhatToCheck = check_unique_modes },
 		write_mode_inference_messages(PredIds, yes, ModuleInfo1),
-		{ ModuleInfo2 = ModuleInfo1 }
+		check_eval_methods(ModuleInfo1, ModuleInfo2)
 	; { WhatToCheck = check_modes },
 		( { UnsafeToContinue = yes } ->
 			write_mode_inference_messages(PredIds, no, ModuleInfo1)
 		;
 			[]
 		),
-		check_eval_methods(ModuleInfo1, ModuleInfo2)
+		{ ModuleInfo2 = ModuleInfo1 }
 	),
 	{ ModuleInfo = ModuleInfo2 }.
 
@@ -604,13 +612,33 @@ modecheck_pred_mode(PredId, PredInfo0, WhatToCheck, MayChangeCalledProc,
 
 modecheck_pred_mode_2(PredId, PredInfo0, WhatToCheck, MayChangeCalledProc,
 		ModuleInfo0, ModuleInfo, Changed0, Changed, NumErrors) -->
-	{ pred_info_procedures(PredInfo0, Procs0) },
-	{ map__keys(Procs0, ProcIds) },
+	% Note that we use pred_info_procids rather than
+	% pred_info_all_procids here, which means that we
+	% don't process modes that have already been inferred
+	% as invalid.
+	{ pred_info_procids(PredInfo0, ProcIds) },
 	( { WhatToCheck = check_modes } ->
-		( { ProcIds = [] } ->
+		(
+			{ ProcIds = [] }
+		->
 			maybe_report_error_no_modes(PredId, PredInfo0,
 					ModuleInfo0),
 			{ NumErrors0 = 0 }
+		;
+			{ special_pred_name_arity(unify, _, PredName,
+				PredArity) },
+			{ pred_info_name(PredInfo0, PredName) },
+			{ pred_info_arity(PredInfo0, PredArity) }
+		->
+			% Don't check for indistinguishable modes in unification
+			% predicates.  The default (in, in) mode must be
+			% semidet, but for single-value types we also want to
+			% create a det mode which will be indistinguishable
+			% from the semidet mode.
+			% (When the type is known, the det mode is called,
+			% but the polymorphic unify needs to be able to call
+			% the semidet mode.)
+			{ NumErrors0 = 0 }
 		;
 			check_for_indistinguishable_modes(ProcIds, PredId,
 				PredInfo0, ModuleInfo0, 0, NumErrors0)
@@ -811,19 +839,33 @@ modecheck_proc_3(ProcId, PredId, WhatToCheck, MayChangeCalledProc,
 	modecheck_final_insts_2(HeadVars, ArgFinalInsts0, ModeInfo2,
 			InferModes, ArgFinalInsts, ModeInfo3),
 
-		% report any errors we found, and save away the results
-	report_mode_errors(ModeInfo3, ModeInfo),
+	( InferModes = yes ->
+		% For inferred predicates, we don't report the
+		% error(s) here; instead we just save them in the
+		% proc_info, thus marking that procedure as invalid.
+		ModeInfo = ModeInfo3,
+		% This is sometimes handy for debugging:
+		% report_mode_errors(ModeInfo3, ModeInfo),
+		mode_info_get_errors(ModeInfo, ModeErrors),
+		ProcInfo1 = ProcInfo0 ^ mode_errors := ModeErrors,
+		NumErrors = 0
+	;
+		% report any errors we found
+		report_mode_errors(ModeInfo3, ModeInfo),
+		mode_info_get_num_errors(ModeInfo, NumErrors),
+		ProcInfo1 = ProcInfo0
+	),
+	% save away the results
 	inst_lists_to_mode_list(ArgInitialInsts, ArgFinalInsts, ArgModes),
 	mode_info_get_changed_flag(ModeInfo, Changed),
 	mode_info_get_module_info(ModeInfo, ModuleInfo),
-	mode_info_get_num_errors(ModeInfo, NumErrors),
 	mode_info_get_io_state(ModeInfo, IOState),
 	mode_info_get_varset(ModeInfo, VarSet),
 	mode_info_get_var_types(ModeInfo, VarTypes),
-	proc_info_set_goal(ProcInfo0, Body, ProcInfo1),
-	proc_info_set_varset(ProcInfo1, VarSet, ProcInfo2),
-	proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo3),
-	proc_info_set_argmodes(ProcInfo3, ArgModes, ProcInfo).
+	proc_info_set_goal(ProcInfo1, Body, ProcInfo2),
+	proc_info_set_varset(ProcInfo2, VarSet, ProcInfo3),
+	proc_info_set_vartypes(ProcInfo3, VarTypes, ProcInfo4),
+	proc_info_set_argmodes(ProcInfo4, ArgModes, ProcInfo).
 
 	% modecheck_final_insts for a lambda expression
 modecheck_final_insts(HeadVars, ArgFinalInsts, ModeInfo0, ModeInfo) :-
@@ -844,7 +886,24 @@ modecheck_final_insts(HeadVars, ArgFinalInsts, ModeInfo0, ModeInfo) :-
 modecheck_final_insts_2(HeadVars, FinalInsts0, ModeInfo0, InferModes,
 			FinalInsts, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
-	mode_info_get_instmap(ModeInfo0, InstMap),
+	mode_info_get_errors(ModeInfo0, Errors),
+	% If there were any mode errors, use an unreachable instmap.
+	% This ensures that we don't get unwanted flow-on errors.
+	% This is not strictly necessary, since we only report the
+	% first mode error anyway, and the resulting FinalInsts
+	% will not be used; but it improves the readability of the
+	% rejected modes.
+	( Errors \= [] ->
+		% If there were any mode errors, something must have
+		% changed, since if the procedure had mode errors
+		% in a previous pass then it wouldn't have been
+		% processed at all in this pass.
+		Changed0 = yes,
+		instmap__init_unreachable(InstMap)
+	;
+		Changed0 = no,
+		mode_info_get_instmap(ModeInfo0, InstMap)
+	),
 	mode_info_get_var_types(ModeInfo0, VarTypes),
 	instmap__lookup_vars(HeadVars, InstMap, VarFinalInsts1),
 	map__apply_to_list(HeadVars, VarTypes, ArgTypes),
@@ -867,8 +926,8 @@ modecheck_final_insts_2(HeadVars, FinalInsts0, ModeInfo0, InferModes,
 		check_final_insts(HeadVars, FinalInsts0, FinalInsts,
 			InferModes, 1, ModuleInfo, no, Changed1,
 			ModeInfo0, ModeInfo1),
-		mode_info_get_changed_flag(ModeInfo1, Changed0),
-		bool__or(Changed0, Changed1, Changed),
+		mode_info_get_changed_flag(ModeInfo1, Changed2),
+		bool__or_list([Changed0, Changed1, Changed2], Changed),
 		mode_info_set_changed_flag(Changed, ModeInfo1, ModeInfo)
 	;
 		check_final_insts(HeadVars, FinalInsts0, VarFinalInsts1,
@@ -887,7 +946,7 @@ maybe_clobber_insts([_|_], [], _) :-
 maybe_clobber_insts([], [], []).
 maybe_clobber_insts([Inst0 | Insts0], [IsLive | IsLives], [Inst | Insts]) :-
 	( IsLive = dead ->
-		Inst = ground(clobbered, no)
+		Inst = ground(clobbered, none)
 	;
 		Inst = Inst0
 	),
@@ -904,7 +963,10 @@ check_final_insts(Vars, Insts, VarInsts, InferModes, ArgNum, ModuleInfo,
 		{ Changed = Changed0 }
 	; { Vars = [Var|Vars1], Insts = [Inst|Insts1],
 	    VarInsts = [VarInst|VarInsts1] } ->
-		( { inst_matches_final(VarInst, Inst, ModuleInfo) } ->
+		=(ModeInfo),
+		{ mode_info_get_var_types(ModeInfo, VarTypes) },
+		{ map__lookup(VarTypes, Var, Type) },
+		( { inst_matches_final(VarInst, Inst, Type, ModuleInfo) } ->
 			{ Changed1 = Changed0 }
 		;
 			{ Changed1 = yes },
@@ -917,10 +979,10 @@ check_final_insts(Vars, Insts, VarInsts, InferModes, ArgNum, ModuleInfo,
 				% XXX this might need to be reconsidered now
 				% we have unique modes
 				( { inst_matches_initial(VarInst, Inst,
-				    ModuleInfo) } ->
+					    Type, ModuleInfo) } ->
 					{ Reason = too_instantiated }
 				; { inst_matches_initial(Inst, VarInst,
-				    ModuleInfo) } ->
+					    Type, ModuleInfo) } ->
 					{ Reason = not_instantiated_enough }
 				;
 					% I don't think this can happen. 
@@ -977,11 +1039,31 @@ modecheck_goal(Goal0 - GoalInfo0, Goal - GoalInfo, ModeInfo0, ModeInfo) :-
 		%
 	mode_info_get_instmap(ModeInfo1, InstMap0),
 
-	modecheck_goal_expr(Goal0, GoalInfo0, Goal, ModeInfo1, ModeInfo),
+	modecheck_goal_expr(Goal0, GoalInfo0, Goal, ModeInfo1, ModeInfo2),
 
-	mode_info_get_instmap(ModeInfo, InstMap),
-	goal_info_get_nonlocals(GoalInfo0, NonLocals),
-	compute_instmap_delta(InstMap0, InstMap, NonLocals, DeltaInstMap),
+	compute_goal_instmap_delta(InstMap0, Goal, GoalInfo0, GoalInfo,
+		ModeInfo2, ModeInfo).
+
+compute_goal_instmap_delta(InstMap0, Goal,
+		GoalInfo0, GoalInfo, ModeInfo0, ModeInfo) :-
+	( Goal = conj([]) ->
+		%
+		% When modecheck_unify.m replaces a unification with a
+		% dead variable with `true', make sure the instmap_delta
+		% of the goal is empty. The code generator and
+		% mode_util__recompute_instmap_delta can be confused
+		% by references to the dead variable in the instmap_delta,
+		% resulting in calls to error/1.
+		%
+		instmap_delta_init_reachable(DeltaInstMap),
+		mode_info_set_instmap(InstMap0, ModeInfo0, ModeInfo)
+	;
+		ModeInfo = ModeInfo0,
+		goal_info_get_nonlocals(GoalInfo0, NonLocals),
+		mode_info_get_instmap(ModeInfo, InstMap),
+		compute_instmap_delta(InstMap0, InstMap,
+			NonLocals, DeltaInstMap)
+	),
 	goal_info_set_instmap_delta(GoalInfo0, DeltaInstMap, GoalInfo).
 
 modecheck_goal_expr(conj(List0), GoalInfo0, Goal) -->
@@ -1111,7 +1193,8 @@ modecheck_goal_expr(call(PredId, ProcId0, Args0, _, Context, PredName),
 	{ mode_info_get_call_id(ModeInfo0, PredId, CallId) },
 	mode_info_set_call_context(call(call(CallId))),
 
-	{ mode_info_get_instmap(ModeInfo0, InstMap0) },
+	=(ModeInfo1),
+	{ mode_info_get_instmap(ModeInfo1, InstMap0) },
 	{ DeterminismKnown = no },
 	modecheck_call_pred(PredId, ProcId0, Args0, DeterminismKnown,
 				Mode, Args, ExtraGoals),
@@ -1186,9 +1269,11 @@ modecheck_goal_expr(switch(Var, CanFail, Cases0, SM), GoalInfo0,
 
 	% to modecheck a pragma_c_code, we just modecheck the proc for 
 	% which it is the goal.
-modecheck_goal_expr(pragma_c_code(IsRecursive, PredId, ProcId0, Args0,
-		ArgNameMap, OrigArgTypes, PragmaCode), GoalInfo, Goal) -->
-	mode_checkpoint(enter, "pragma_c_code"),
+modecheck_goal_expr(pragma_foreign_code(Attributes, PredId, ProcId0,
+		Args0, ArgNameMap, OrigArgTypes, PragmaCode),
+		GoalInfo, Goal) -->
+	mode_checkpoint(enter, "pragma_foreign_code"),
+
 	=(ModeInfo0),
 	{ mode_info_get_call_id(ModeInfo0, PredId, CallId) },
 
@@ -1198,30 +1283,18 @@ modecheck_goal_expr(pragma_c_code(IsRecursive, PredId, ProcId0, Args0,
 	modecheck_call_pred(PredId, ProcId0, Args0, DeterminismKnown,
 				ProcId, Args, ExtraGoals),
 
-	{ Pragma = pragma_c_code(IsRecursive, PredId, ProcId, Args0,
-			ArgNameMap, OrigArgTypes, PragmaCode) },
+	{ Pragma = pragma_foreign_code(Attributes, PredId, ProcId,
+			Args0, ArgNameMap, OrigArgTypes, PragmaCode) },
 	handle_extra_goals(Pragma, ExtraGoals, GoalInfo, Args0, Args,
 			InstMap0, Goal),
 
 	mode_info_unset_call_context,
-	mode_checkpoint(exit, "pragma_c_code").
+	mode_checkpoint(exit, "pragma_foreign_code").
 
 modecheck_goal_expr(bi_implication(_, _), _, _) -->
 	% these should have been expanded out by now
 	{ error("modecheck_goal_expr: unexpected bi_implication") }.
 
- 	% given the right-hand-side of a unification, return a list of
-	% the potentially non-local variables of that unification.
-
-unify_rhs_vars(var(Var), [Var]).
-unify_rhs_vars(functor(_Functor, Vars), Vars).
-unify_rhs_vars(lambda_goal(_PredOrFunc, _EvalMethod, _Fix, LambdaNonLocals,
-		LambdaVars, _Modes, _Det, _Goal - GoalInfo), Vars) :-
-	goal_info_get_nonlocals(GoalInfo, NonLocals0),
-	set__delete_list(NonLocals0, LambdaVars, NonLocals1),
-	set__insert_list(NonLocals1, LambdaNonLocals, NonLocals),
-	set__to_sorted_list(NonLocals, Vars).
-
 append_extra_goals(no_extra_goals, ExtraGoals, ExtraGoals).
 append_extra_goals(extra_goals(BeforeGoals, AfterGoals),
 		no_extra_goals, extra_goals(BeforeGoals, AfterGoals)).
@@ -1325,12 +1398,6 @@ handle_extra_goals_contexts([Goal0 | Goals0], Context, [Goal | Goals]) :-
 	goal_info_set_context(GoalInfo0, Context, GoalInfo),
 	handle_extra_goals_contexts(Goals0, Context, Goals).
 
-:- pred goal_get_nonlocals(hlds_goal, set(prog_var)).
-:- mode goal_get_nonlocals(in, out) is det.
-
-goal_get_nonlocals(_Goal - GoalInfo, NonLocals) :-
-	goal_info_get_nonlocals(GoalInfo, NonLocals).
-
 %-----------------------------------------------------------------------------%
 
 	% Modecheck a conjunction without doing any reordering.
@@ -1416,14 +1483,21 @@ modecheck_conj_list(Goals0, Goals) -->
 
 mode_info_add_goals_live_vars([]) --> [].
 mode_info_add_goals_live_vars([Goal | Goals]) -->
-	{ goal_get_nonlocals(Goal, Vars) },
-	mode_info_add_live_vars(Vars),
-	mode_info_add_goals_live_vars(Goals).
+	% We add the live vars for the goals in the goal list
+	% in reverse order, because this ensures that in the
+	% common case (where there is no delaying), when we come
+	% to remove the live vars for the first goal
+	% they will have been added last and will thus be
+	% at the start of the list of live vars sets, which
+	% makes them cheaper to remove.
+	mode_info_add_goals_live_vars(Goals),
+	{ goal_get_nonlocals(Goal, NonLocals) },
+	mode_info_add_live_vars(NonLocals).
 
 mode_info_remove_goals_live_vars([]) --> [].
 mode_info_remove_goals_live_vars([Goal | Goals]) -->
-	{ goal_get_nonlocals(Goal, Vars) },
-	mode_info_remove_live_vars(Vars),
+	{ goal_get_nonlocals(Goal, NonLocals) },
+	mode_info_remove_live_vars(NonLocals),
 	mode_info_remove_goals_live_vars(Goals).
 
 :- type impurity_errors == list(mode_error_info).
@@ -1692,8 +1766,8 @@ modecheck_functor_test(Var, ConsId) -->
 		mode_info_di, mode_info_uo) is det.
 
 modecheck_par_conj_list([], [], _NonLocals, []) --> [].
-modecheck_par_conj_list([Goal0|Goals0], [Goal|Goals], NonLocals, 
-		[InstMap - GoalNonLocals|InstMaps]) -->
+modecheck_par_conj_list([Goal0 | Goals0], [Goal|Goals], NonLocals, 
+		[InstMap - GoalNonLocals | InstMaps]) -->
 	mode_info_dcg_get_instmap(InstMap0),
 	{ Goal0 = _ - GoalInfo },
 	{ goal_info_get_nonlocals(GoalInfo, GoalNonLocals) },
@@ -1725,18 +1799,6 @@ modecheck_par_conj_list([Goal0|Goals0], [Goal|Goals], NonLocals,
 	modecheck_par_conj_list(Goals0, Goals, NonLocals, InstMaps),
 	mode_info_unlock_vars(par_conj, Bound1).
 
-:- pred get_all_conjunct_nonlocals(list(hlds_goal), set(prog_var),
-		set(prog_var)).
-:- mode get_all_conjunct_nonlocals(in, in, out) is det.
-
-get_all_conjunct_nonlocals([], NonLocals, NonLocals).
-get_all_conjunct_nonlocals([G|Gs], NonLocals0, NonLocals) :-
-	G = _ - GoalInfo,
-	goal_info_get_nonlocals(GoalInfo, GoalNonLocals),
-	set__union(GoalNonLocals, NonLocals0, NonLocals1),
-	get_all_conjunct_nonlocals(Gs, NonLocals1, NonLocals).
-
-
 %-----------------------------------------------------------------------------%
 
 	%
@@ -1758,34 +1820,41 @@ compute_arg_offset(PredInfo, ArgOffset) :-
 	% ensure the liveness of each variable satisfies the corresponding
 	% expected liveness.
 
-modecheck_var_list_is_live([_|_], [], _) -->
+modecheck_var_list_is_live([_|_], [], _, _) -->
 	{ error("modecheck_var_list_is_live: length mismatch") }.
-modecheck_var_list_is_live([], [_|_], _) -->
+modecheck_var_list_is_live([], [_|_], _, _) -->
 	{ error("modecheck_var_list_is_live: length mismatch") }.
-modecheck_var_list_is_live([], [], _ArgNum) --> [].
-modecheck_var_list_is_live([Var|Vars], [IsLive|IsLives], ArgNum0) -->
+modecheck_var_list_is_live([], [], _NeedExactMatch, _ArgNum) --> [].
+modecheck_var_list_is_live([Var|Vars], [IsLive|IsLives], NeedExactMatch,
+		ArgNum0) -->
 	{ ArgNum is ArgNum0 + 1 },
 	mode_info_set_call_arg_context(ArgNum),
-	modecheck_var_is_live(Var, IsLive),
-	modecheck_var_list_is_live(Vars, IsLives, ArgNum).
+	modecheck_var_is_live(Var, IsLive, NeedExactMatch),
+	modecheck_var_list_is_live(Vars, IsLives, NeedExactMatch, ArgNum).
 
-:- pred modecheck_var_is_live(prog_var, is_live, mode_info, mode_info).
-:- mode modecheck_var_is_live(in, in, mode_info_di, mode_info_uo) is det.
+:- pred modecheck_var_is_live(prog_var, is_live, bool, mode_info, mode_info).
+:- mode modecheck_var_is_live(in, in, in, mode_info_di, mode_info_uo) is det.
 
 	% `live' means possibly used later on, and
 	% `dead' means definitely not used later on.
-	% The only time you get an error is if you pass a variable
+	% If you don't need an exact match, then
+	% the only time you get an error is if you pass a variable
 	% which is live to a predicate that expects the variable to
 	% be dead; the predicate may use destructive update to clobber
 	% the variable, so we must be sure that it is dead after the call.
 
-modecheck_var_is_live(VarId, ExpectedIsLive, ModeInfo0, ModeInfo) :-
+modecheck_var_is_live(VarId, ExpectedIsLive, NeedExactMatch,
+		ModeInfo0, ModeInfo) :-
 	mode_info_var_is_live(ModeInfo0, VarId, VarIsLive),
-	( ExpectedIsLive = dead, VarIsLive = live ->
+	( 
+		( ExpectedIsLive = dead, VarIsLive = live
+		; NeedExactMatch = yes, VarIsLive \= ExpectedIsLive
+		)
+	->
 		set__singleton_set(WaitingVars, VarId),
 		mode_info_error(WaitingVars, mode_error_var_is_live(VarId),
 			ModeInfo0, ModeInfo)
-	;
+	; 
 		ModeInfo = ModeInfo0
 	).
 
@@ -1795,28 +1864,70 @@ modecheck_var_is_live(VarId, ExpectedIsLive, ModeInfo0, ModeInfo) :-
 	% that the inst of each variable matches the corresponding initial
 	% inst.
 
-modecheck_var_has_inst_list([_|_], [], _) -->
+modecheck_var_has_inst_list(Vars, Insts, NeedEaxctMatch, ArgNum, Subst) -->
+	{ map__init(Subst0) },
+	modecheck_var_has_inst_list_2(Vars, Insts, NeedEaxctMatch, ArgNum,
+		Subst0, Subst).
+
+:- pred modecheck_var_has_inst_list_2(list(prog_var), list(inst), bool, int,
+		inst_var_sub, inst_var_sub, mode_info, mode_info).
+:- mode modecheck_var_has_inst_list_2(in, in, in, in, in, out,
+		mode_info_di, mode_info_uo) is det.
+
+modecheck_var_has_inst_list_2([_|_], [], _, _, _, _) -->
 	{ error("modecheck_var_has_inst_list: length mismatch") }.
-modecheck_var_has_inst_list([], [_|_], _) -->
+modecheck_var_has_inst_list_2([], [_|_], _, _, _, _) -->
 	{ error("modecheck_var_has_inst_list: length mismatch") }.
-modecheck_var_has_inst_list([], [], _ArgNum) --> [].
-modecheck_var_has_inst_list([Var|Vars], [Inst|Insts], ArgNum0) -->
+modecheck_var_has_inst_list_2([], [], _Exact, _ArgNum, Subst, Subst) --> [].
+modecheck_var_has_inst_list_2([Var|Vars], [Inst|Insts],
+		NeedExactMatch, ArgNum0, Subst0, Subst) -->
 	{ ArgNum is ArgNum0 + 1 },
 	mode_info_set_call_arg_context(ArgNum),
-	modecheck_var_has_inst(Var, Inst),
-	modecheck_var_has_inst_list(Vars, Insts, ArgNum).
+	modecheck_var_has_inst(Var, Inst, NeedExactMatch, Subst0, Subst1),
+	modecheck_var_has_inst_list_2(Vars, Insts,
+		NeedExactMatch, ArgNum, Subst1, Subst).
 
-:- pred modecheck_var_has_inst(prog_var, inst, mode_info, mode_info).
-:- mode modecheck_var_has_inst(in, in, mode_info_di, mode_info_uo) is det.
+:- pred modecheck_var_has_inst(prog_var, inst, bool,
+		inst_var_sub, inst_var_sub, mode_info, mode_info).
+:- mode modecheck_var_has_inst(in, in, in,
+		in, out, mode_info_di, mode_info_uo) is det.
 
-modecheck_var_has_inst(VarId, Inst, ModeInfo0, ModeInfo) :-
+modecheck_var_has_inst(VarId, Inst, NeedExactMatch, Subst0, Subst,
+		ModeInfo0, ModeInfo) :-
 	mode_info_get_instmap(ModeInfo0, InstMap),
 	instmap__lookup_var(InstMap, VarId, VarInst),
+	mode_info_get_var_types(ModeInfo0, VarTypes),
+	map__lookup(VarTypes, VarId, Type),
 
-	mode_info_get_module_info(ModeInfo0, ModuleInfo),
-	( inst_matches_initial(VarInst, Inst, ModuleInfo) ->
-		ModeInfo = ModeInfo0
+	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
+	(
+		(
+			NeedExactMatch = no,
+			inst_matches_initial(VarInst, Inst, Type, ModuleInfo0,
+				ModuleInfo, Subst0, Subst1)
+		;
+			NeedExactMatch = yes,
+			inst_matches_final(VarInst, Inst, Type, ModuleInfo0),
+			ModuleInfo = ModuleInfo0,
+			Subst1 = Subst0
+			% WARNING:
+			% The code above (Subst1 = Subst0) assumes that there
+			% are no inst variables in the mode of the callee.
+			% Currently this will always true, since
+			% `NeedExactMatch' will be `yes' only if we are
+			% doing mode inference on the callee, and mode
+			% inference currently will not infer polymorphic modes.
+			% But that assumption might not always hold in future.
+			% An alternative would be to call inst_matches_initial
+			% here too, just to calculate the inst substitution.
+			% But that would be less efficient, so (at least
+			% for now) we don't do it.
+		)
+	->
+		Subst = Subst1,
+		mode_info_set_module_info(ModeInfo0, ModuleInfo, ModeInfo)
 	;
+		Subst = Subst0,
 		set__singleton_set(WaitingVars, VarId),
 		mode_info_error(WaitingVars,
 			mode_error_var_has_inst(VarId, VarInst, Inst),
@@ -1916,7 +2027,9 @@ modecheck_set_var_inst(Var0, FinalInst, ModeInfo00, ModeInfo) :-
 			% If we haven't added any information and
 			% we haven't bound any part of the var, then
 			% the only thing we can have done is lose uniqueness.
-			inst_matches_initial(Inst0, Inst, ModuleInfo)
+			mode_info_get_var_types(ModeInfo1, VarTypes),
+			map__lookup(VarTypes, Var0, Type),
+			inst_matches_initial(Inst0, Inst, Type, ModuleInfo)
 		->
 			instmap__set(InstMap0, Var0, Inst, InstMap),
 			mode_info_set_instmap(InstMap, ModeInfo1, ModeInfo3)
@@ -1925,7 +2038,9 @@ modecheck_set_var_inst(Var0, FinalInst, ModeInfo00, ModeInfo) :-
 			% lost some uniqueness, or bound part of the var.
 			% The call to inst_matches_binding will succeed
 			% only if we haven't bound any part of the var.
-			inst_matches_binding(Inst, Inst0, ModuleInfo)
+			mode_info_get_var_types(ModeInfo1, VarTypes),
+			map__lookup(VarTypes, Var0, Type),
+			inst_matches_binding(Inst, Inst0, Type, ModuleInfo)
 		->
 			% We've just added some information
 			% or lost some uniqueness.
@@ -1984,12 +2099,15 @@ handle_implied_mode(Var0, VarInst0, InitialInst0, Var,
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
 	inst_expand(ModuleInfo0, InitialInst0, InitialInst),
 	inst_expand(ModuleInfo0, VarInst0, VarInst1),
+
+	mode_info_get_var_types(ModeInfo0, VarTypes0),
+	map__lookup(VarTypes0, Var0, VarType),
 	(
 		% If the initial inst of the variable matches_final
 		% the initial inst specified in the pred's mode declaration,
 		% then it's not a call to an implied mode, it's an exact
 		% match with a genuine mode.
-		inst_matches_final(VarInst1, InitialInst, ModuleInfo0)
+		inst_matches_final(VarInst1, InitialInst, VarType, ModuleInfo0)
 	->
 		Var = Var0,
 		ExtraGoals = ExtraGoals0,
@@ -2016,9 +2134,6 @@ handle_implied_mode(Var0, VarInst0, InitialInst0, Var,
 			% XXX We ought to use a more elegant method
 			% XXX than hard-coding the name `<foo>_init_any'.
 
-			mode_info_get_var_types(ModeInfo0, VarTypes0),
-			map__lookup(VarTypes0, Var, VarType),
-
 			mode_info_get_context(ModeInfo0, Context),
 			mode_info_get_mode_context(ModeInfo0, ModeContext),
 			mode_context_to_unify_context(ModeContext, ModeInfo0,
@@ -2077,9 +2192,7 @@ handle_implied_mode(Var0, VarInst0, InitialInst0, Var,
 
 			% Introduce a new variable
 			mode_info_get_varset(ModeInfo0, VarSet0),
-			mode_info_get_var_types(ModeInfo0, VarTypes0),
 			varset__new_var(VarSet0, Var, VarSet),
-			map__lookup(VarTypes0, Var0, VarType),
 			map__set(VarTypes0, Var, VarType, VarTypes),
 			mode_info_set_varset(VarSet, ModeInfo0, ModeInfo1),
 			mode_info_set_var_types(VarTypes, ModeInfo1, ModeInfo),
@@ -2125,6 +2238,9 @@ mode_context_to_unify_context(uninitialized, _, _) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
+% check that the evaluation method is OK for the given mode(s).
+% we also check the mode of main/2 here.
+
 :- pred check_eval_methods(module_info, module_info, io__state, io__state).
 :- mode check_eval_methods(in, out, di, uo) is det.
 
@@ -2151,69 +2267,39 @@ pred_check_eval_methods([PredId|Rest], ModuleInfo0, ModuleInfo) -->
 proc_check_eval_methods([], _, M, M) --> [].
 proc_check_eval_methods([ProcId|Rest], PredId, ModuleInfo0, ModuleInfo) --> 
 	{ module_info_pred_proc_info(ModuleInfo0, PredId, ProcId, 
-		_, ProcInfo) },
+		PredInfo, ProcInfo) },
 	{ proc_info_eval_method(ProcInfo, EvalMethod) },
-	( { EvalMethod \= eval_normal } ->
-		{ proc_info_context(ProcInfo, Context) },
-		{ eval_method_to_string(EvalMethod, EvalMethodS) },
-		globals__io_lookup_bool_option(verbose_errors, 
-			VerboseErrors),
-		{ proc_info_argmodes(ProcInfo, Modes) },
-		( 
-			\+ { only_fully_in_out_modes(Modes, 
-				ModuleInfo0) } 
-		->
-			prog_out__write_context(Context),
-			io__write_string("Sorry, not implemented: `pragma "),
-			io__write_string(EvalMethodS),
-			io__write_string("'\n"),
-			prog_out__write_context(Context),
-			io__write_string(
-			    "  declaration not allowed for procedure with\n"),
-			prog_out__write_context(Context),
-			io__write_string(
-			    "  partially instantiated modes.\n"), 
-			( { VerboseErrors = yes } ->
-				io__write_string(
-"	Tabling of predicates/functions with partially instantiated modes
-	is not currently implemented.\n")
-			;
-				[]
-			),
-			{ module_info_incr_errors(ModuleInfo0, ModuleInfo1) }
-		;
-			{ ModuleInfo1 = ModuleInfo0 }	
-		),	
-		( 
-			\+ { only_nonunique_modes(Modes, 
-				ModuleInfo1) } 
-		->
-			prog_out__write_context(Context),
-			io__write_string("Error: `pragma "),
-			io__write_string(EvalMethodS),
-			io__write_string("'\n"),
-			prog_out__write_context(Context),
-			io__write_string(
-			    "  declaration not allowed for procedure with\n"),
-			prog_out__write_context(Context),
-			io__write_string("  unique modes.\n"), 
-			( { VerboseErrors = yes } ->
-				io__write_string(
-"	Tabling of predicates/functions with unique modes is not allowed
-	as this would lead to a copying of the unique arguments which 
-	would result in them no longer being unique.\n")
-			;
-				[]
-			),
-			{ module_info_incr_errors(ModuleInfo1, ModuleInfo2) }
-		;
-			{ ModuleInfo2 = ModuleInfo1 }	
-		)
+	{ proc_info_argmodes(ProcInfo, Modes) },
+	( 
+		{ eval_method_requires_ground_args(EvalMethod) = yes },
+		\+ { only_fully_in_out_modes(Modes, ModuleInfo0) } 
+	->
+		report_eval_method_requires_ground_args(ProcInfo,
+			ModuleInfo0, ModuleInfo1)
 	;
-		{ ModuleInfo2 = ModuleInfo0 }	
-		
+		{ ModuleInfo1 = ModuleInfo0 }	
+	),	
+	( 
+		{ eval_method_destroys_uniqueness(EvalMethod) = yes },
+		\+ { only_nonunique_modes(Modes, ModuleInfo1) } 
+	->
+		report_eval_method_destroys_uniqueness(ProcInfo,
+			ModuleInfo1, ModuleInfo2)
+	;
+		{ ModuleInfo2 = ModuleInfo1 }	
 	),
-	proc_check_eval_methods(Rest, PredId, ModuleInfo2, ModuleInfo).
+	(
+		{ pred_info_name(PredInfo, "main") },
+		{ pred_info_arity(PredInfo, 2) },
+		{ pred_info_is_exported(PredInfo) },
+		{ \+ check_mode_of_main(Modes, ModuleInfo2) }
+	->
+		report_wrong_mode_for_main(ProcInfo,
+			ModuleInfo2, ModuleInfo3)
+	;
+		{ ModuleInfo3 = ModuleInfo2 }
+	),
+	proc_check_eval_methods(Rest, PredId, ModuleInfo3, ModuleInfo).
 
 :- pred only_fully_in_out_modes(list(mode), module_info).
 :- mode only_fully_in_out_modes(in, in) is semidet.
@@ -2243,6 +2329,92 @@ only_nonunique_modes([Mode|Rest], ModuleInfo) :-
 	inst_is_not_partly_unique(ModuleInfo, FinalInst),
 	only_nonunique_modes(Rest, ModuleInfo).
 
+:- pred check_mode_of_main(list(mode), module_info).
+:- mode check_mode_of_main(in, in) is semidet.
+
+check_mode_of_main([Di, Uo], ModuleInfo) :-
+	mode_get_insts(ModuleInfo, Di, DiInitialInst, DiFinalInst),
+	mode_get_insts(ModuleInfo, Uo, UoInitialInst, UoFinalInst),
+	%
+	% Note that we hard-code these tests,
+	% rather than using `inst_is_free', `inst_is_unique', etc.,
+	% since for main/2 we're looking for an exact match
+	% (modulo inst synonyms) with what the language reference
+	% manual specifies, rather than looking for a particular
+	% abstract property.
+	%
+	inst_expand(ModuleInfo, DiInitialInst, ground(unique, none)),
+	inst_expand(ModuleInfo, DiFinalInst, ground(clobbered, none)),
+	inst_expand(ModuleInfo, UoInitialInst, Free),
+	( Free = free ; Free = free(_Type) ),
+	inst_expand(ModuleInfo, UoFinalInst, ground(unique, none)).
+
+:- pred report_eval_method_requires_ground_args(proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode report_eval_method_requires_ground_args(in, in, out, di, uo) is det.
+
+report_eval_method_requires_ground_args(ProcInfo, ModuleInfo0, ModuleInfo) -->
+	{ proc_info_eval_method(ProcInfo, EvalMethod) },
+	{ proc_info_context(ProcInfo, Context) },
+	{ eval_method_to_string(EvalMethod, EvalMethodS) },
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	prog_out__write_context(Context),
+	io__write_string("Sorry, not implemented: `pragma "),
+	io__write_string(EvalMethodS),
+	io__write_string("'\n"),
+	prog_out__write_context(Context),
+	io__write_string(
+	    "  declaration not allowed for procedure with\n"),
+	prog_out__write_context(Context),
+	io__write_string(
+	    "  partially instantiated modes.\n"), 
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"	Tabling of predicates/functions with partially instantiated modes
+	is not currently implemented.\n")
+	;
+		[]
+	),
+	{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }.
+
+:- pred report_eval_method_destroys_uniqueness(proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode report_eval_method_destroys_uniqueness(in, in, out, di, uo) is det.
+
+report_eval_method_destroys_uniqueness(ProcInfo, ModuleInfo0, ModuleInfo) -->
+	{ proc_info_eval_method(ProcInfo, EvalMethod) },
+	{ proc_info_context(ProcInfo, Context) },
+	{ eval_method_to_string(EvalMethod, EvalMethodS) },
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	prog_out__write_context(Context),
+	io__write_string("Error: `pragma "),
+	io__write_string(EvalMethodS),
+	io__write_string("'\n"),
+	prog_out__write_context(Context),
+	io__write_string(
+	    "  declaration not allowed for procedure with\n"),
+	prog_out__write_context(Context),
+	io__write_string("  unique modes.\n"), 
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"	Tabling of predicates/functions with unique modes is not allowed
+	as this would lead to a copying of the unique arguments which 
+	would result in them no longer being unique.\n")
+	;
+		[]
+	),
+	{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }.
+
+:- pred report_wrong_mode_for_main(proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode report_wrong_mode_for_main(in, in, out, di, uo) is det.
+
+report_wrong_mode_for_main(ProcInfo, ModuleInfo0, ModuleInfo) -->
+	{ proc_info_context(ProcInfo, Context) },
+	prog_out__write_context(Context),
+	io__write_string("Error: main/2 must have mode `(di, uo)'.\n"),
+	{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
