@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -107,7 +107,8 @@
 
 module_qual__module_qualify_items(Items0, Items, ModuleName, ReportErrors,
 			Info, NumErrors, UndefTypes, UndefModes) -->
-	{ init_mq_info(ReportErrors, Info0) },
+	globals__io_get_globals(Globals),
+	{ init_mq_info(Items0, Globals, ReportErrors, ModuleName, Info0) },
 	{ collect_mq_info(Items0, Info0, Info1) },
 	do_module_qualify_items(Items0, Items, Info1, Info),
 	{ mq_info_get_type_error_flag(Info, UndefTypes) },
@@ -135,27 +136,39 @@ module_qual__qualify_type_qualification(Type0, Type, Context, Info0, Info) -->
 				% i.e. ones for which there was a
 				% `:- import_module' or `:- use_module'
 				% declaration in this module.
-			set(module_name),
+			imported_modules::set(module_name),
 
 				% Sets of all modules, types, insts, modes,
 				% and typeclasses visible in this module.
-			module_id_set,
-			type_id_set,
-			inst_id_set,
-			mode_id_set,
-			class_id_set,
+			modules::module_id_set,
+			types::type_id_set,
+			insts::inst_id_set,
+			modes::mode_id_set,
+			classes::class_id_set,
 
-			set(module_name), % modules imported in the
+			unused_interface_modules::set(module_name), 
+				% modules imported in the
 				% interface that are not definitely
 				% needed in the interface.
-			import_status, % import status of the current item.
-			int,	% number of errors found.
-			bool,	% are there any undefined types or typeclasses.
-			bool,	% are there any undefined insts or modes.
-			bool, 	% do we want to report errors.
-			error_context,	% context of the current item.
-			need_qualifier	% must uses of the current item be 
+
+				% import status of the current item.
+			import_status::import_status, 
+
+			num_errors::int,% number of errors found.
+
+				% are there any undefined types or typeclasses.
+			type_error_flag::bool,	
+				% are there any undefined insts or modes.
+			mode_error_flag::bool,	
+				% do we want to report errors.
+			report_error_flag::bool, 	
+				% context of the current item.
+			error_context::error_context,	
+				% name of the current module
+			this_module::module_name,  
+				% must uses of the current item be 
 				% explicitly module qualified.
+			need_qual_flag::need_qualifier	
 	).
 
 :- type partial_qualifier_info --->
@@ -227,7 +240,7 @@ collect_mq_info_2(assertion(Goal, _ProgVarSet), Info0, Info) :-
 collect_mq_info_2(nothing, Info, Info).
 collect_mq_info_2(typeclass(_, Name, Vars, _, _), Info0, Info) :-
 	add_typeclass_defn(Name, Vars, Info0, Info).
-collect_mq_info_2(instance(_,_,_,_,_), Info, Info).
+collect_mq_info_2(instance(_,_,_,_,_,_), Info, Info).
 
 
 % Predicates to add the type, inst, mode and typeclass ids visible
@@ -403,7 +416,7 @@ process_assert(if_then_else(_, GA, GB, GC) - _, Symbols, Success) :-
 	list__append(Symbols0, SymbolsC, Symbols),
 	bool__and(SuccessA, SuccessB, Success0),
 	bool__and(Success0, SuccessC, Success).
-process_assert(call(SymName, Args0, _) - _, Symbols, Success) :-
+process_assert(call(SymName, Args0, _Purity) - _, Symbols, Success) :-
 	(
 		SymName = qualified(_, _)
 	->
@@ -421,7 +434,7 @@ process_assert(call(SymName, Args0, _) - _, Symbols, Success) :-
 		Symbols = [],
 		Success = no
 	).
-process_assert(unify(LHS0, RHS0) - _, Symbols, Success) :-
+process_assert(unify(LHS0, RHS0, _Purity) - _, Symbols, Success) :-
 	term__coerce(LHS0, LHS),
 	term__coerce(RHS0, RHS),
 	(
@@ -570,9 +583,9 @@ module_qualify_item(typeclass(Constraints0, Name, Vars, Interface0, VarSet) -
 	qualify_class_constraint_list(Constraints0, Constraints, Info1, Info2),
 	qualify_class_interface(Interface0, Interface, Info2, Info).
 
-module_qualify_item(instance(Constraints0, Name0, Types0, Body0, VarSet) -
-			Context, 
-		instance(Constraints, Name, Types, Body, VarSet) -
+module_qualify_item(instance(Constraints0, Name0, Types0, Body0, VarSet,
+		ModName) - Context, 
+		instance(Constraints, Name, Types, Body, VarSet, ModName) -
 			Context, 
 		Info0, Info, yes) -->
 	{ list__length(Types0, Arity) },
@@ -749,15 +762,20 @@ qualify_inst(free(_), _, _, _) -->
 qualify_inst(bound(Uniq, BoundInsts0), bound(Uniq, BoundInsts),
 				Info0, Info) -->
 	qualify_bound_inst_list(BoundInsts0, BoundInsts, Info0, Info).
-qualify_inst(ground(Uniq, MaybePredInstInfo0), ground(Uniq, MaybePredInstInfo),
+qualify_inst(ground(Uniq, GroundInstInfo0), ground(Uniq, GroundInstInfo),
 				Info0, Info) -->
 	(
-		{ MaybePredInstInfo0 = yes(pred_inst_info(A, Modes0, Det)) },
+		{ GroundInstInfo0 = higher_order(pred_inst_info(A, Modes0,
+				Det)) },
 		qualify_mode_list(Modes0, Modes, Info0, Info),
-		{ MaybePredInstInfo = yes(pred_inst_info(A, Modes, Det)) }
+		{ GroundInstInfo = higher_order(pred_inst_info(A, Modes, Det)) }
 	;
-		{ MaybePredInstInfo0 = no },
-		{ MaybePredInstInfo = no },
+		{ GroundInstInfo0 = constrained_inst_var(Var) },
+		{ GroundInstInfo = constrained_inst_var(Var) },
+		{ Info = Info0 }
+	;
+		{ GroundInstInfo0 = none },
+		{ GroundInstInfo = none },
 		{ Info = Info0 }
 	).
 qualify_inst(inst_var(Var), inst_var(Var), Info, Info) --> [].
@@ -838,6 +856,9 @@ qualify_type(Type0, Type, Info0, Info) -->
 		; { type_id_is_higher_order(TypeId0, _, _) } ->
 			{ TypeId = TypeId0 },
 			{ Info1 = Info0 }
+		; { type_id_is_tuple(TypeId0) } ->
+			{ TypeId = TypeId0 },
+			{ Info1 = Info0 }
 		;
 			{ mq_info_get_types(Info0, Types) },
 			find_unique_match(TypeId0, TypeId, Types,
@@ -878,10 +899,11 @@ qualify_type(Type0, Type, Info0, Info) -->
 		mq_info::in, mq_info::out, io__state::di, io__state::uo) is det.
 
 qualify_pragma(source_file(File), source_file(File), Info, Info) --> [].
-qualify_pragma(c_header_code(Code), c_header_code(Code), Info, Info) --> [].
-qualify_pragma(c_code(Code), c_code(Code), Info, Info) --> [].
-qualify_pragma(c_code(Rec, SymName, PredOrFunc, PragmaVars0, Varset, CCode),
-		c_code(Rec, SymName, PredOrFunc, PragmaVars, Varset, CCode), 
+qualify_pragma(foreign_decl(L, Code), foreign_decl(L, Code), Info, Info) --> [].
+qualify_pragma(foreign(L, C), foreign(L, C), Info, Info) --> [].
+qualify_pragma(
+	    foreign(Rec, SymName, PredOrFunc, PragmaVars0, Varset, Code),
+	    foreign(Rec, SymName, PredOrFunc, PragmaVars, Varset, Code), 
 		Info0, Info) -->
 	qualify_pragma_vars(PragmaVars0, PragmaVars, Info0, Info).
 qualify_pragma(tabled(A, B, C, D, MModes0), tabled(A, B, C, D, MModes), 
@@ -1022,9 +1044,9 @@ qualify_class_interface([M0|M0s], [M|Ms], MQInfo0, MQInfo) -->
 	% done when the item is parsed.
 qualify_class_method(
 		pred(TypeVarset, InstVarset, ExistQVars, Name, TypesAndModes0,
-			MaybeDet, Cond, ClassContext0, Context), 
+			MaybeDet, Cond, Purity, ClassContext0, Context), 
 		pred(TypeVarset, InstVarset, ExistQVars, Name, TypesAndModes,
-			MaybeDet, Cond, ClassContext, Context), 
+			MaybeDet, Cond, Purity, ClassContext, Context), 
 		MQInfo0, MQInfo
 		) -->
 	qualify_types_and_modes(TypesAndModes0, TypesAndModes, 
@@ -1033,9 +1055,11 @@ qualify_class_method(
 		MQInfo1, MQInfo).
 qualify_class_method(
 		func(TypeVarset, InstVarset, ExistQVars, Name, TypesAndModes0,
-			ReturnMode0, MaybeDet, Cond, ClassContext0, Context), 
+			ReturnMode0, MaybeDet, Cond, Purity, ClassContext0,
+			Context), 
 		func(TypeVarset, InstVarset, ExistQVars, Name, TypesAndModes,
-			ReturnMode, MaybeDet, Cond, ClassContext, Context), 
+			ReturnMode, MaybeDet, Cond, Purity, ClassContext,
+			Context), 
 		MQInfo0, MQInfo
 		) -->
 	qualify_types_and_modes(TypesAndModes0, TypesAndModes, 
@@ -1068,16 +1092,11 @@ qualify_instance_body(ClassName, concrete(M0s), concrete(Ms)) :-
 		Ms = M0s
 	;
 		sym_name_get_module_name(ClassName, unqualified(""), Module),
-		Qualify = lambda([M0::in, M::out] is det,
-			(
-				M0 = pred_instance(Method0, A, B, C),
-				add_module_qualifier(Module, Method0, Method),
-				M = pred_instance(Method, A, B, C)
-			;
-				M0 = func_instance(Method0, A, B, C),
-				add_module_qualifier(Module, Method0, Method),
-				M = func_instance(Method, A, B, C)
-			)),
+		Qualify = lambda([M0::in, M::out] is det, (
+			M0 = instance_method(A, Method0, C, D, E),
+			add_module_qualifier(Module, Method0, Method),
+			M = instance_method(A, Method, C, D, E)
+		)),
 		list__map(Qualify, M0s, Ms)
 	).
 
@@ -1201,7 +1220,8 @@ report_undefined(Info, Id, IdType) -->
 		%
 		{ Id = qualified(ModuleName, _) - _Arity },
 		{ mq_info_get_imported_modules(Info, ImportedModules) },
-		{ \+ set__member(ModuleName, ImportedModules) }
+		{ \+ set__member(ModuleName, ImportedModules) },
+		{ \+ ModuleName = Info^this_module }
 	->
 		io__write_string("\n"),
 		prog_out__write_context(Context),
@@ -1305,6 +1325,7 @@ write_id(SymName - Arity) -->
 
 maybe_warn_unused_interface_imports(ModuleName, UnusedImports) -->
 	globals__io_lookup_bool_option(warn_interface_imports, Warn),
+	globals__io_lookup_bool_option(halt_at_warn, HaltAtWarn),
 	(
 		{ UnusedImports = []
 		; Warn = no
@@ -1335,7 +1356,14 @@ maybe_warn_unused_interface_imports(ModuleName, UnusedImports) -->
 			IsOrAre, " not\n"
 		]),
 		prog_out__write_context(Context),
-		io__write_string("  used in the interface.\n")
+		io__write_string("  used in the interface.\n"),
+		(
+			{ HaltAtWarn = yes }
+		->
+			io__set_exit_status(1)
+		;
+			[]
+		)
 	).
 
 :- pred is_or_are(list(T)::in, string::out) is det.
@@ -1378,20 +1406,20 @@ is_builtin_atomic_type(unqualified("character") - 0).
 %-----------------------------------------------------------------------------%
 % Access and initialisation predicates.
 
-:- pred init_mq_info(bool::in, mq_info::out) is det.
+:- pred init_mq_info(item_list::in, globals::in, bool::in, module_name::in,
+	mq_info::out) is det.
 
-init_mq_info(ReportErrors, Info0) :-
+init_mq_info(Items, Globals, ReportErrors, ModuleName, Info0) :-
 	term__context_init(Context),
 	ErrorContext = type(unqualified("") - 0) - Context,
 	set__init(InterfaceModules0),
-	mercury_public_builtin_module(BuiltinModule),
-	mercury_private_builtin_module(PrivateBuiltinModule),
-	set__list_to_set([BuiltinModule, PrivateBuiltinModule],
-		ImportedModules),
+	get_implicit_dependencies(Items, Globals, ImportDeps, UseDeps),
+	set__list_to_set(ImportDeps `list__append` UseDeps, ImportedModules),
 	id_set_init(Empty),
 	Info0 = mq_info(ImportedModules, Empty, Empty, Empty, Empty,
 			Empty, InterfaceModules0, not_exported, 0, no, no,
-			ReportErrors, ErrorContext, may_be_unqualified).
+			ReportErrors, ErrorContext, ModuleName,
+			may_be_unqualified).
 
 :- pred mq_info_get_imported_modules(mq_info::in, set(module_name)::out) is det.
 :- pred mq_info_get_modules(mq_info::in, module_id_set::out) is det.
@@ -1408,28 +1436,20 @@ init_mq_info(ReportErrors, Info0) :-
 :- pred mq_info_get_report_error_flag(mq_info::in, bool::out) is det.
 :- pred mq_info_get_error_context(mq_info::in, error_context::out) is det.
 
-mq_info_get_imported_modules(mq_info(ImportedModules, _,_,_,_,_,_,_,_,_,_,_,
-	_,_), ImportedModules).
-mq_info_get_modules(mq_info(_, Modules, _,_,_,_,_,_,_,_,_,_,_,_), Modules).
-mq_info_get_types(mq_info(_,_, Types, _,_,_,_,_,_,_,_,_,_,_), Types).
-mq_info_get_insts(mq_info(_,_,_, Insts, _,_,_,_,_,_,_,_,_,_), Insts).
-mq_info_get_modes(mq_info(_,_,_,_, Modes, _,_,_,_,_,_,_,_,_), Modes).
-mq_info_get_classes(mq_info(_,_,_,_,_, Classes, _,_,_,_,_,_,_,_), Classes).
-mq_info_get_unused_interface_modules(mq_info(_,_,_,_,_,_, Modules, _,_,_,_,_,
-	_,_), Modules).
-mq_info_get_import_status(mq_info(_,_,_,_,_,_,_, Status, _,_,_,_,_,_), Status).
-mq_info_get_num_errors(mq_info(_,_,_,_,_,_,_,_, NumErrors, _,_,_,_,_),
-	NumErrors).
-mq_info_get_type_error_flag(mq_info(_,_,_,_,_,_,_,_,_, TypeErrs, _,_,_,_),
-	TypeErrs).
-mq_info_get_mode_error_flag(mq_info(_,_,_,_,_,_,_,_,_,_, ModeError, _,_,_),
-	ModeError).
-mq_info_get_report_error_flag(mq_info(_,_,_,_,_,_,_,_,_,_,_, Report, _,_),
-	Report).
-mq_info_get_error_context(mq_info(_,_,_,_,_,_,_,_,_,_,_,_, Context, _),
-	Context).
-mq_info_get_need_qual_flag(mq_info(_,_,_,_,_,_,_,_,_,_,_,_,_, UseModule),
-	UseModule).
+mq_info_get_imported_modules(MQInfo, MQInfo^imported_modules).
+mq_info_get_modules(MQInfo, MQInfo^modules).
+mq_info_get_types(MQInfo, MQInfo^types).
+mq_info_get_insts(MQInfo, MQInfo^insts).
+mq_info_get_modes(MQInfo, MQInfo^modes).
+mq_info_get_classes(MQInfo, MQInfo^classes).
+mq_info_get_unused_interface_modules(MQInfo, MQInfo^unused_interface_modules).
+mq_info_get_import_status(MQInfo, MQInfo^import_status).
+mq_info_get_num_errors(MQInfo, MQInfo^num_errors).
+mq_info_get_type_error_flag(MQInfo, MQInfo^type_error_flag).
+mq_info_get_mode_error_flag(MQInfo, MQInfo^mode_error_flag).
+mq_info_get_report_error_flag(MQInfo, MQInfo^report_error_flag).
+mq_info_get_error_context(MQInfo, MQInfo^error_context).
+mq_info_get_need_qual_flag(MQInfo, MQInfo^need_qual_flag).
 
 :- pred mq_info_set_imported_modules(mq_info::in, set(module_name)::in,
 		mq_info::out) is det.
@@ -1448,38 +1468,24 @@ mq_info_get_need_qual_flag(mq_info(_,_,_,_,_,_,_,_,_,_,_,_,_, UseModule),
 :- pred mq_info_set_error_context(mq_info::in, error_context::in,
 						mq_info::out) is det.
 
-mq_info_set_imported_modules(mq_info(_, B,C,D,E,F,G,H,I,J,K,L,M,N),
-		ImportedModules,
-		mq_info(ImportedModules, B,C,D,E,F,G,H,I,J,K,L,M,N)).
-mq_info_set_modules(mq_info(A, _, C,D,E,F,G,H,I,J,K,L,M,N), Modules,
-		mq_info(A, Modules, C,D,E,F,G,H,I,J,K,L,M,N)).
-mq_info_set_types(mq_info(A,B, _, D,E,F,G,H,I,J,K,L,M,N), Types,
-		mq_info(A,B, Types, D,E,F,G,H,I,J,K,L,M,N)).
-mq_info_set_insts(mq_info(A,B,C, _, E,F,G,H,I,J,K,L,M,N), Insts,
-		mq_info(A,B,C, Insts, E,F,G,H,I,J,K,L,M,N)).
-mq_info_set_modes(mq_info(A,B,C,D, _, F,G,H,I,J,K,L,M,N), Modes,
-		mq_info(A,B,C,D, Modes, F,G,H,I,J,K,L,M,N)).
-mq_info_set_classes(mq_info(A,B,C,D,E, _, G,H,I,J,K,L,M,N), Classes,
-		mq_info(A,B,C,D,E, Classes, G,H,I,J,K,L,M,N)).
-mq_info_set_unused_interface_modules(mq_info(A,B,C,D,E,F, _, H,I,J,K,L,M,N),
-		Modules,
-		mq_info(A,B,C,D,E,F, Modules, H,I,J,K,L,M,N)).
-mq_info_set_import_status(mq_info(A,B,C,D,E,F,G, _, I,J,K,L,M,N), Status,
-		mq_info(A,B,C,D,E,F,G, Status, I,J,K,L,M,N)).
-mq_info_set_type_error_flag(mq_info(A,B,C,D,E,F,G,H,I, _, K,L,M,N),
-		mq_info(A,B,C,D,E,F,G,H,I, yes, K,L,M,N)).
-mq_info_set_mode_error_flag(mq_info(A,B,C,D,E,F,G,H,I,J, _, L,M,N),
-		mq_info(A,B,C,D,E,F,G,H,I,J, yes, L,M,N)).
-mq_info_set_error_context(mq_info(A,B,C,D,E,F,G,H,I,J,K,L, _, N), Context,
-		mq_info(A,B,C,D,E,F,G,H,I,J,K,L, Context, N)).
-mq_info_set_need_qual_flag(mq_info(A,B,C,D,E,F,G,H,I,J,K,L,M, _), Flag,
-		mq_info(A,B,C,D,E,F,G,H,I,J,K,L,M, Flag)).
+mq_info_set_imported_modules(MQInfo,
+		ImportedModules, MQInfo^imported_modules := ImportedModules).
+mq_info_set_modules(MQInfo, Modules, MQInfo^modules := Modules).
+mq_info_set_types(MQInfo, Types, MQInfo^types := Types).
+mq_info_set_insts(MQInfo, Insts, MQInfo^insts := Insts).
+mq_info_set_modes(MQInfo, Modes, MQInfo^modes := Modes).
+mq_info_set_classes(MQInfo, Classes, MQInfo^classes := Classes).
+mq_info_set_unused_interface_modules(MQInfo, 
+		Modules, MQInfo^unused_interface_modules := Modules).
+mq_info_set_import_status(MQInfo, Status, MQInfo^import_status := Status).
+mq_info_set_type_error_flag(MQInfo, MQInfo^type_error_flag := yes).
+mq_info_set_mode_error_flag(MQInfo, MQInfo^mode_error_flag := yes).
+mq_info_set_error_context(MQInfo, Context, MQInfo^error_context := Context).
+mq_info_set_need_qual_flag(MQInfo, Flag, MQInfo^need_qual_flag := Flag).
 
 :- pred mq_info_incr_errors(mq_info::in, mq_info::out) is det.
 
-mq_info_incr_errors(mq_info(A,B,C,D,E,F,G,H, NumErrors0, J,K,L,M,N), 
-		mq_info(A,B,C,D,E,F,G,H, NumErrors, J,K,L,M,N)) :-
-	NumErrors is NumErrors0 + 1.
+mq_info_incr_errors(MQInfo, MQInfo^num_errors := (MQInfo^num_errors +1)).
 
 :- pred mq_info_set_error_flag(mq_info::in, id_type::in, mq_info::out) is det.
 
