@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,13 +14,14 @@
 
 :- interface.
 
-:- import_module vn_type, vn_table, livemap.
-:- import_module llds, builtin_ops, atsort.
+:- import_module llds, vn_type, vn_table, livemap.
+:- import_module code_model, rtti, layout, builtin_ops.
+:- import_module atsort.
 
 :- import_module io, bool, list, assoc_list, std_util.
 
-:- pred opt_debug__msg(bool, string, io__state, io__state).
-:- mode opt_debug__msg(in, in, di, uo) is det.
+:- pred opt_debug__msg(bool, int, string, io__state, io__state).
+:- mode opt_debug__msg(in, in, in, di, uo) is det.
 
 :- pred opt_debug__dump_instrs(bool, list(instruction), io__state, io__state).
 :- mode opt_debug__dump_instrs(in, in, di, uo) is det.
@@ -136,9 +137,21 @@
 :- pred opt_debug__dump_const(rval_const, string).
 :- mode opt_debug__dump_const(in, out) is det.
 
+:- pred opt_debug__dump_data_addr(data_addr, string).
+:- mode opt_debug__dump_data_addr(in, out) is det.
+
 :- pred opt_debug__dump_data_name(data_name, string).
 :- mode opt_debug__dump_data_name(in, out) is det.
 
+:- pred opt_debug__dump_rtti_type_id(rtti_type_id, string).
+:- mode opt_debug__dump_rtti_type_id(in, out) is det.
+
+:- pred opt_debug__dump_rtti_name(rtti_name, string).
+:- mode opt_debug__dump_rtti_name(in, out) is det.
+
+:- pred opt_debug__dump_layout_name(layout_name, string).
+:- mode opt_debug__dump_layout_name(in, out) is det.
+
 :- pred opt_debug__dump_unop(unary_op, string).
 :- mode opt_debug__dump_unop(in, out) is det.
 
@@ -185,16 +198,24 @@
 
 :- implementation.
 
-:- import_module llds_out, opt_util, vn_util, hlds_pred, globals, options.
 :- import_module prog_out.
+:- import_module hlds_pred.
+:- import_module llds_out, opt_util, vn_util.
+:- import_module globals, options.
 
 :- import_module int, set, map, string.
 
-opt_debug__msg(OptDebug, Msg) -->
+opt_debug__msg(OptDebug, LabelNo, Msg) -->
 	(
 		{ OptDebug = yes },
 		io__write_string("\n"),
 		io__write_string(Msg),
+		( { LabelNo >= 0 } ->
+			io__write_string(", next label no: "),
+			io__write_int(LabelNo)
+		;
+			[]
+		),
 		io__write_string("\n")
 	;
 		{ OptDebug = no }
@@ -294,7 +315,7 @@ opt_debug__dump_ctrl_list([N - VnInstr | Ctrllist], Str) :-
 
 opt_debug__dump_vninstr(vn_livevals(_), Str) :-
 	string__append_list(["livevals(...)"], Str).
-opt_debug__dump_vninstr(vn_call(Proc, Ret, _, _, _), Str) :-
+opt_debug__dump_vninstr(vn_call(Proc, Ret, _, _, _, _), Str) :-
 	opt_debug__dump_code_addr(Proc, P_str),
 	opt_debug__dump_code_addr(Ret, R_str),
 	string__append_list(["call(", P_str, ", ", R_str, ")"], Str).
@@ -318,6 +339,9 @@ opt_debug__dump_vninstr(vn_mark_hp(Vnlval), Str) :-
 opt_debug__dump_vninstr(vn_restore_hp(Vn), Str) :-
 	opt_debug__dump_vn(Vn, Vn_str),
 	string__append_list(["restore_hp(", Vn_str, ")"], Str).
+opt_debug__dump_vninstr(vn_free_heap(Vn), Str) :-
+	opt_debug__dump_vn(Vn, Vn_str),
+	string__append_list(["free_heap(", Vn_str, ")"], Str).
 opt_debug__dump_vninstr(vn_store_ticket(Vnlval), Str) :-
 	opt_debug__dump_vnlval(Vnlval, V_str),
 	string__append_list(["store_ticket(", V_str, ")"], Str).
@@ -325,12 +349,13 @@ opt_debug__dump_vninstr(vn_reset_ticket(Vn, _Reason), Str) :-
 	opt_debug__dump_vn(Vn, Vn_str),
 	string__append_list(["reset_ticket(", Vn_str, ", _)"], Str).
 opt_debug__dump_vninstr(vn_discard_ticket, "discard_ticket").
+opt_debug__dump_vninstr(vn_prune_ticket, "prune_ticket").
 opt_debug__dump_vninstr(vn_mark_ticket_stack(Vnlval), Str) :-
 	opt_debug__dump_vnlval(Vnlval, V_str),
 	string__append_list(["mark_ticket_stack(", V_str, ")"], Str).
-opt_debug__dump_vninstr(vn_discard_tickets_to(Vn), Str) :-
+opt_debug__dump_vninstr(vn_prune_tickets_to(Vn), Str) :-
 	opt_debug__dump_vn(Vn, Vn_str),
-	string__append_list(["discard_tickets_to(", Vn_str, ", _)"], Str).
+	string__append_list(["prune_tickets_to(", Vn_str, ", _)"], Str).
 opt_debug__dump_vninstr(vn_incr_sp(N, _), Str) :-
 	string__int_to_string(N, N_str),
 	string__append_list(["incr_sp(", N_str, ")"], Str).
@@ -634,7 +659,7 @@ opt_debug__dump_rval(mkword(T, N), Str) :-
 opt_debug__dump_rval(const(C), Str) :-
 	opt_debug__dump_const(C, C_str),
 	string__append_list(["const(", C_str, ")"], Str).
-opt_debug__dump_rval(create(T, MA, _, U, L, _), Str) :-
+opt_debug__dump_rval(create(T, MA, _, U, L, _, _), Str) :-
 	string__int_to_string(T, T_str),
 	opt_debug__dump_maybe_rvals(MA, 3, MA_str),
 	(
@@ -697,42 +722,150 @@ opt_debug__dump_const(multi_string_const(L, _S), Str) :-
 opt_debug__dump_const(code_addr_const(CodeAddr), Str) :-
 	opt_debug__dump_code_addr(CodeAddr, C_str),
 	string__append_list(["code_addr_const(", C_str, ")"], Str).
-opt_debug__dump_const(data_addr_const(data_addr(BaseName, VarName)), Str) :-
-	opt_debug__dump_data_name(VarName, N_str),
-	prog_out__sym_name_to_string(BaseName, BaseName_str),
+opt_debug__dump_const(data_addr_const(DataAddr), Str) :-
+	opt_debug__dump_data_addr(DataAddr, DataAddr_str),
 	string__append_list(
-		["data_addr_const(", BaseName_str, ", ", N_str, ")"], Str).
+		["data_addr_const(", DataAddr_str, ")"], Str).
 opt_debug__dump_const(label_entry(Label), Str) :-
 	opt_debug__dump_label(Label, LabelStr),
 	string__append_list(["label_entry(", LabelStr, ")"], Str).
+
+opt_debug__dump_data_addr(data_addr(ModuleName, DataName), Str) :-
+	prog_out__sym_name_to_string(ModuleName, ModuleName_str),
+	opt_debug__dump_data_name(DataName, DataName_str),
+	string__append_list(
+		["data_addr(", ModuleName_str, ", ", DataName_str, ")"], Str).
+opt_debug__dump_data_addr(rtti_addr(RttiTypeId, DataName), Str) :-
+	opt_debug__dump_rtti_type_id(RttiTypeId, RttiTypeId_str),
+	opt_debug__dump_rtti_name(DataName, DataName_str),
+	string__append_list(
+		["rtti_addr(", RttiTypeId_str, ", ", DataName_str, ")"], Str).
+opt_debug__dump_data_addr(layout_addr(LayoutName), Str) :-
+	opt_debug__dump_layout_name(LayoutName, LayoutName_str),
+	string__append_list(["layout_addr(", LayoutName_str, ")"], Str).
+
 opt_debug__dump_data_name(common(N), Str) :-
 	string__int_to_string(N, N_str),
 	string__append("common", N_str, Str).
-opt_debug__dump_data_name(type_ctor(BaseData, TypeName, TypeArity), Str) :-
-	llds_out__make_type_ctor_name(BaseData, TypeName, TypeArity, Str).
 opt_debug__dump_data_name(base_typeclass_info(ClassId, InstanceNum), Str) :-
 	llds_out__make_base_typeclass_info_name(ClassId, InstanceNum, Str).
-opt_debug__dump_data_name(module_layout, "module_layout").
-opt_debug__dump_data_name(proc_layout(Label), Str) :-
-	opt_debug__dump_label(Label, LabelStr),
-	string__append_list(["proc_layout(", LabelStr, ")"], Str).
-opt_debug__dump_data_name(internal_layout(Label), Str) :-
-	opt_debug__dump_label(Label, LabelStr),
-	string__append_list(["internal_layout(", LabelStr, ")"], Str).
 opt_debug__dump_data_name(tabling_pointer(ProcLabel), Str) :-
 	opt_debug__dump_proclabel(ProcLabel, ProcLabelStr),
 	string__append_list(["tabling_pointer(", ProcLabelStr, ")"], Str).
 
+opt_debug__dump_rtti_type_id(rtti_type_id(ModuleName, TypeName, Arity), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleName_str),
+	llds_out__name_mangle(TypeName, TypeName_str),
+	string__int_to_string(Arity, Arity_str),
+	string__append_list(["rtti_type_id(", ModuleName_str, ", ",
+		TypeName_str, Arity_str, ")"], Str).
+
+opt_debug__dump_rtti_name(exist_locns(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("exist_locns_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(exist_info(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("exist_info_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(field_names(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("field_names_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(field_types(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("field_types_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(enum_functor_desc(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("enum_functor_desc_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(notag_functor_desc, Str) :-
+	Str = "notag_functor_desc_".
+opt_debug__dump_rtti_name(du_functor_desc(Ordinal), Str) :-
+	string__int_to_string(Ordinal, Ordinal_str),
+	string__append("du_functor_desc_", Ordinal_str, Str).
+opt_debug__dump_rtti_name(enum_name_ordered_table, Str) :-
+	Str = "enum_name_ordered_table".
+opt_debug__dump_rtti_name(enum_value_ordered_table, Str) :-
+	Str = "enum_value_ordered_table".
+opt_debug__dump_rtti_name(du_name_ordered_table, Str) :-
+	Str = "du_name_ordered_table".
+opt_debug__dump_rtti_name(du_stag_ordered_table(Ptag), Str) :-
+	string__int_to_string(Ptag, Ptag_str),
+	string__append("du_stag_ordered_table_", Ptag_str, Str).
+opt_debug__dump_rtti_name(du_ptag_ordered_table, Str) :-
+	Str = "du_ptag_ordered_table".
+opt_debug__dump_rtti_name(type_ctor_info, Str) :-
+	Str = "type_ctor_info".
+opt_debug__dump_rtti_name(base_typeclass_info(_ModuleName, ClassId,
+		InstanceStr), Str) :-
+	llds_out__make_base_typeclass_info_name(ClassId, InstanceStr, Str).
+opt_debug__dump_rtti_name(pseudo_type_info(_Pseudo), Str) :-
+	% XXX should give more info than this
+	Str = "pseudo_type_info".
+opt_debug__dump_rtti_name(type_hashcons_pointer, Str) :-
+	Str = "type_hashcons_pointer".
+
+opt_debug__dump_layout_name(label_layout(Label, LabelVars), Str) :-
+	opt_debug__dump_label(Label, LabelStr),
+	(
+		LabelVars = label_has_var_info,
+		LabelVarsStr = "label_has_var_info"
+	;
+		LabelVars = label_has_no_var_info,
+		LabelVarsStr = "label_has_no_var_info"
+	),
+	string__append_list(["label_layout(", LabelStr, ", ",
+		LabelVarsStr, ")"], Str).
+opt_debug__dump_layout_name(proc_layout(ProcLabel, _), Str) :-
+	opt_debug__dump_proclabel(ProcLabel, ProcLabelStr),
+	string__append_list(["proc_layout(", ProcLabelStr, ")"], Str).
+opt_debug__dump_layout_name(proc_layout_var_names(ProcLabel), Str) :-
+	opt_debug__dump_proclabel(ProcLabel, ProcLabelStr),
+	string__append_list(["proc_layout_var_names(", ProcLabelStr, ")"],
+		Str).
+opt_debug__dump_layout_name(closure_proc_id(ProcLabel, SeqNo, _), Str) :-
+	opt_debug__dump_proclabel(ProcLabel, ProcLabelStr),
+	string__int_to_string(SeqNo, SeqNoStr),
+	string__append_list(["closure_proc_id(", ProcLabelStr, ", ",
+		SeqNoStr, ")"], Str).
+opt_debug__dump_layout_name(file_layout(ModuleName, FileNum), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__int_to_string(FileNum, FileNumStr),
+	string__append_list(["file_layout(", ModuleNameStr, ", ",
+		FileNumStr, ")"], Str).
+opt_debug__dump_layout_name(file_layout_line_number_vector(ModuleName,
+		FileNum), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__int_to_string(FileNum, FileNumStr),
+	string__append_list(["file_layout_line_number_vector(", ModuleNameStr,
+		", ", FileNumStr, ")"], Str).
+opt_debug__dump_layout_name(file_layout_label_layout_vector(ModuleName,
+		FileNum), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__int_to_string(FileNum, FileNumStr),
+	string__append_list(["file_layout_label_layout_vector(", ModuleNameStr,
+		", ", FileNumStr, ")"], Str).
+opt_debug__dump_layout_name(module_layout_string_table(ModuleName), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__append_list(["module_layout_string_table(", ModuleNameStr,
+		")"], Str).
+opt_debug__dump_layout_name(module_layout_file_vector(ModuleName), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__append_list(["module_layout_file_vector(", ModuleNameStr, ")"],
+		Str).
+opt_debug__dump_layout_name(module_layout_proc_vector(ModuleName), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__append_list(["module_layout_proc_vector(", ModuleNameStr, ")"],
+		Str).
+opt_debug__dump_layout_name(module_layout(ModuleName), Str) :-
+	llds_out__sym_name_mangle(ModuleName, ModuleNameStr),
+	string__append_list(["module_layout(", ModuleNameStr, ")"], Str).
+
 opt_debug__dump_unop(mktag, "mktag").
 opt_debug__dump_unop(tag, "tag").
 opt_debug__dump_unop(unmktag, "unmktag").
 opt_debug__dump_unop(mkbody, "mkbody").
-opt_debug__dump_unop(body, "body").
 opt_debug__dump_unop(unmkbody, "unmkbody").
 opt_debug__dump_unop(not, "not").
 opt_debug__dump_unop(hash_string, "hash_string").
 opt_debug__dump_unop(bitwise_complement, "bitwise_complement").
-opt_debug__dump_unop(cast_to_unsigned, "cast_to_unsigned").
 
 opt_debug__dump_binop(Op, String) :-
 	llds_out__binary_op_to_string(Op, String).
@@ -779,7 +912,7 @@ opt_debug__dump_code_addr(do_aditi_insert, "do_aditi_insert").
 opt_debug__dump_code_addr(do_aditi_delete, "do_aditi_delete").
 opt_debug__dump_code_addr(do_aditi_bulk_insert, "do_aditi_bulk_insert").
 opt_debug__dump_code_addr(do_aditi_bulk_delete, "do_aditi_bulk_delete").
-opt_debug__dump_code_addr(do_aditi_modify, "do_aditi_modify").
+opt_debug__dump_code_addr(do_aditi_bulk_modify, "do_aditi_bulk_modify").
 opt_debug__dump_code_addr(do_not_reached, "do_not_reached").
 
 opt_debug__dump_code_addrs([], "").
@@ -788,7 +921,7 @@ opt_debug__dump_code_addrs([Addr | Addrs], Str) :-
 	opt_debug__dump_code_addrs(Addrs, A2_str),
 	string__append_list([" ", A_str, A2_str], Str).
 
-opt_debug__dump_label(local(ProcLabel, N), Str) :-
+opt_debug__dump_label(local(N, ProcLabel), Str) :-
 	opt_debug__dump_proclabel(ProcLabel, P_str),
 	string__int_to_string(N, N_str),
 	string__append_list([P_str, "_", N_str], Str).
@@ -857,7 +990,7 @@ opt_debug__dump_instr(assign(Lval, Rval), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	opt_debug__dump_rval(Rval, R_str),
 	string__append_list(["assign(", L_str, ", ", R_str, ")"], Str).
-opt_debug__dump_instr(call(Proc, Ret, _, _, _), Str) :-
+opt_debug__dump_instr(call(Proc, Ret, _, _, _, _), Str) :-
 	opt_debug__dump_code_addr(Proc, P_str),
 	opt_debug__dump_code_addr(Ret, R_str),
 	string__append_list(["call(", P_str, ", ", R_str, ", ...)"], Str).
@@ -919,6 +1052,9 @@ opt_debug__dump_instr(mark_hp(Lval), Str) :-
 opt_debug__dump_instr(restore_hp(Rval), Str) :-
 	opt_debug__dump_rval(Rval, R_str),
 	string__append_list(["restore_hp(", R_str, ")"], Str).
+opt_debug__dump_instr(free_heap(Rval), Str) :-
+	opt_debug__dump_rval(Rval, R_str),
+	string__append_list(["free_heap(", R_str, ")"], Str).
 opt_debug__dump_instr(store_ticket(Lval), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	string__append_list(["store_ticket(", L_str, ")"], Str).
@@ -926,12 +1062,13 @@ opt_debug__dump_instr(reset_ticket(Rval, _Reason), Str) :-
 	opt_debug__dump_rval(Rval, R_str),
 	string__append_list(["reset_ticket(", R_str, ", _)"], Str).
 opt_debug__dump_instr(discard_ticket, "discard_ticket").
+opt_debug__dump_instr(prune_ticket, "prune_ticket").
 opt_debug__dump_instr(mark_ticket_stack(Lval), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	string__append_list(["mark_ticket_stack(", L_str, ")"], Str).
-opt_debug__dump_instr(discard_tickets_to(Rval), Str) :-
+opt_debug__dump_instr(prune_tickets_to(Rval), Str) :-
 	opt_debug__dump_rval(Rval, R_str),
-	string__append_list(["discard_tickets_to(", R_str, ")"], Str).
+	string__append_list(["prune_tickets_to(", R_str, ")"], Str).
 opt_debug__dump_instr(incr_sp(Size, _), Str) :-
 	string__int_to_string(Size, S_str),
 	string__append_list(["incr_sp(", S_str, ")"], Str).
@@ -956,7 +1093,7 @@ opt_debug__dump_instr(join_and_continue(Lval, Label), Str) :-
 	opt_debug__dump_label(Label, LabelStr),
 	string__append_list(["join(", LvalStr, ", ", LabelStr, ")"], Str).
 % XXX  should probably give more info than this
-opt_debug__dump_instr(pragma_c(_, Comps, _, _, _, _), Str) :-
+opt_debug__dump_instr(pragma_c(_, Comps, _, _, _, _, _), Str) :-
 	opt_debug__dump_components(Comps, C_str),
 	string__append_list(["pragma_c(", C_str, ")"], Str).
 
