@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -34,7 +34,7 @@
 				io__state, io__state))
 		;	update_pred_error(pred_error_task)
 		;	update_module(pred(
-				proc_info, proc_info,
+				pred_info, proc_info, proc_info,
 				module_info, module_info))
 		;	update_module_io(pred(
 				pred_id, proc_id, proc_info, proc_info,
@@ -93,7 +93,7 @@ about unbound type variables.
 				out, out, di, uo) is det)
 		;	update_pred_error(pred(in, in, out, in, out,
 				out, out, di, uo) is det)
-		;	update_module(pred(in, out, in, out) is det)
+		;	update_module(pred(in, in, out, in, out) is det)
 		;	update_module_io(pred(in, in, in, out,
 				in, out, di, uo) is det)
 		;	update_module_cookie(pred(in, in, in, out, in, out,
@@ -146,13 +146,9 @@ about unbound type variables.
 
 :- pred report_error(string::in, io__state::di, io__state::uo) is det.
 
-:- pred invoke_system_command(string::in, bool::out,
-	io__state::di, io__state::uo) is det.
-
 :- pred maybe_report_sizes(module_info::in, io__state::di, io__state::uo)
 	is det.
 
-
 :- pred report_pred_proc_id(module_info, pred_id, proc_id, 
 		maybe(prog_context), prog_context, io__state, io__state).
 :- mode report_pred_proc_id(in, in, in, in, out, di, uo) is det.
@@ -163,12 +159,30 @@ about unbound type variables.
 	
 %-----------------------------------------------------------------------------%
 
+:- type quote_char
+	--->	forward		% '
+	;	double.		% "
+
+	% Invoke a shell script.
+:- pred invoke_shell_command(string::in, bool::out,
+	io__state::di, io__state::uo) is det.
+
+	% Invoke an executable.
+:- pred invoke_system_command(string::in, bool::out,
+	io__state::di, io__state::uo) is det.
+
+	% Make a command string, which needs to be invoked in a shell
+	% environment.
+:- pred make_command_string(string::in, quote_char::in, string::out) is det.
+
+%-----------------------------------------------------------------------------%
+
 :- implementation.
 
 :- import_module options, globals, hlds_out, prog_out, mode_util.
 :- import_module mercury_to_mercury.
 :- import_module varset.
-:- import_module int, map, tree234, require.
+:- import_module int, map, tree234, require, string.
 
 process_all_nonimported_procs(Task, ModuleInfo0, ModuleInfo) -->
 	{ True = lambda([_PredInfo::in] is semidet, true) },
@@ -274,7 +288,7 @@ process_nonimported_procs([ProcId | ProcIds], PredId, Task0, Task,
 
 	(
 		Task0 = update_module(Closure),
-		call(Closure, Proc0, Proc, ModuleInfo0, ModuleInfo8),
+		call(Closure, Pred0, Proc0, Proc, ModuleInfo0, ModuleInfo8),
 		Task1 = Task0,
 		State9 = State0
 	;
@@ -398,6 +412,10 @@ passes_aux__handle_errors(WarnCnt, ErrCnt, ModuleInfo1, ModuleInfo8,
 		State9 = State2
 	).
 
+invoke_shell_command(Command0, Succeeded) -->
+	{ make_command_string(Command0, forward, Command) },
+	invoke_system_command(Command, Succeeded).
+
 invoke_system_command(Command, Succeeded) -->
 	globals__io_lookup_bool_option(verbose, Verbose),
 	( { Verbose = yes } ->
@@ -420,6 +438,32 @@ invoke_system_command(Command, Succeeded) -->
 		{ Succeeded = no }
 	).
 
+make_command_string(String0, QuoteType, String) :-
+	( use_win32 ->
+		(
+			QuoteType = forward,
+			Quote = " '"
+		;
+			QuoteType = double,
+			Quote = " \""
+		),
+		string__append_list(["sh -c ", Quote, String0, Quote], String)
+	;
+		String = String0
+	).
+
+	% Are we compiling in a win32 environment?
+:- pred use_win32 is semidet.
+:- pragma c_code(use_win32,
+	[will_not_call_mercury, thread_safe],
+"
+#ifdef MR_WIN32
+	SUCCESS_INDICATOR = 1;
+#else
+	SUCCESS_INDICATOR = 0;
+#endif
+").
+
 maybe_report_sizes(HLDS) -->
 	globals__io_lookup_bool_option(statistics, Statistics),
 	( { Statistics = yes } ->
