@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1998-1999 University of Melbourne.
+% Copyright (C) 1998-2000 University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -151,14 +151,17 @@ pd_util__simplify_goal(Simplifications, Goal0, Goal) -->
 	pd_info_get_module_info(ModuleInfo0),
 	{ module_info_globals(ModuleInfo0, Globals) },
 	pd_info_get_pred_proc_id(proc(PredId, ProcId)),
-	{ det_info_init(ModuleInfo0, PredId, ProcId,
+	{ proc_info_vartypes(ProcInfo0, VarTypes0) },
+	{ det_info_init(ModuleInfo0, VarTypes0, PredId, ProcId,
 		Globals, DetInfo0) },
 	pd_info_get_instmap(InstMap0),
 	pd_info_get_proc_info(ProcInfo0),
 	{ proc_info_varset(ProcInfo0, VarSet0) },
-	{ proc_info_vartypes(ProcInfo0, VarTypes0) },
+	{ proc_info_inst_varset(ProcInfo0, InstVarSet0) },
+	{ proc_info_typeinfo_varmap(ProcInfo0, TVarMap0) },
+	{ proc_info_typeclass_info_varmap(ProcInfo0, TCVarMap0) },
 	{ simplify_info_init(DetInfo0, Simplifications, InstMap0,
-		VarSet0, VarTypes0, SimplifyInfo0) },
+		VarSet0, InstVarSet0, TVarMap0, TCVarMap0, SimplifyInfo0) },
 
 	{ simplify__process_goal(Goal0, Goal, SimplifyInfo0, SimplifyInfo) },
 
@@ -169,9 +172,13 @@ pd_util__simplify_goal(Simplifications, Goal0, Goal) -->
 	{ simplify_info_get_varset(SimplifyInfo, VarSet) },
 	{ simplify_info_get_var_types(SimplifyInfo, VarTypes) },
 	{ simplify_info_get_cost_delta(SimplifyInfo, CostDelta) },
+	{ simplify_info_get_type_info_varmap(SimplifyInfo, TVarMap) },
+	{ simplify_info_get_typeclass_info_varmap(SimplifyInfo, TCVarMap) },
 	pd_info_get_proc_info(ProcInfo1),
 	{ proc_info_set_varset(ProcInfo1, VarSet, ProcInfo2) },
-	{ proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo) },
+	{ proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo3) },
+	{ proc_info_set_typeinfo_varmap(ProcInfo3, TVarMap, ProcInfo4) },
+	{ proc_info_set_typeclass_info_varmap(ProcInfo4, TCVarMap, ProcInfo) },
 	pd_info_set_proc_info(ProcInfo),
 	pd_info_incr_cost_delta(CostDelta),
 	pd_info_set_module_info(ModuleInfo).
@@ -667,7 +674,7 @@ pd_util__requantify_goal(Goal0, NonLocals, Goal) -->
 	{ proc_info_varset(ProcInfo0, VarSet0) },
 	{ proc_info_vartypes(ProcInfo0, VarTypes0) },
 	{ implicitly_quantify_goal(Goal0, VarSet0, VarTypes0, NonLocals,
-			Goal, VarSet, VarTypes, _) },
+		Goal, VarSet, VarTypes, _) },
 	{ proc_info_set_varset(ProcInfo0, VarSet, ProcInfo1) },
 	{ proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo) },
 	pd_info_set_proc_info(ProcInfo).
@@ -677,8 +684,9 @@ pd_util__recompute_instmap_delta(Goal0, Goal) -->
 	pd_info_get_instmap(InstMap),
 	pd_info_get_proc_info(ProcInfo),
 	{ proc_info_vartypes(ProcInfo, VarTypes) },
-	{ recompute_instmap_delta(yes, Goal0, Goal, VarTypes, InstMap, 
-		ModuleInfo0, ModuleInfo) },
+	{ proc_info_inst_varset(ProcInfo, InstVarSet) },
+	{ recompute_instmap_delta(yes, Goal0, Goal, VarTypes, InstVarSet,
+		InstMap, ModuleInfo0, ModuleInfo) },
 	pd_info_set_module_info(ModuleInfo).
 
 %-----------------------------------------------------------------------------%
@@ -795,7 +803,7 @@ bound_inst_list_MSG(Xs, Ys, Expansions, ModuleInfo, Uniq, List, Inst) :-
 			Uniq = unique,
 			inst_is_unique(ModuleInfo, bound(unique, List))
 		),		
-		Inst = ground(Uniq, no)
+		Inst = ground(Uniq, none)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -933,9 +941,9 @@ pd_util__goals_match_2([OldGoal | OldGoals], [NewGoal | NewGoals],
 				NewArgs = [NewVar | NewArgs1]
 			;
 				OldUnification = deconstruct(OldVar, ConsId,
-							OldArgs1, _, _),
+							OldArgs1, _, _, _),
 				NewUnification = deconstruct(NewVar, ConsId,
-							NewArgs1, _, _),
+							NewArgs1, _, _, _),
 				OldArgs = [OldVar | OldArgs1],
 				NewArgs = [NewVar | NewArgs1]
 			)	
@@ -1009,13 +1017,11 @@ match_generic_call(aditi_builtin(Builtin1, CallId),
 	% The other fields are all implied by the pred_proc_id.
 match_aditi_builtin(aditi_call(PredProcId, _, _, _),
 		aditi_call(PredProcId, _, _, _)).
-match_aditi_builtin(aditi_insert(PredId), aditi_insert(PredId)).
+match_aditi_builtin(aditi_tuple_insert_delete(InsertDelete, PredId),
+		aditi_tuple_insert_delete(InsertDelete, PredId)).
 	% The syntax used does not change the result of the call.
-match_aditi_builtin(aditi_delete(PredId, _), aditi_delete(PredId, _)).
-match_aditi_builtin(aditi_bulk_operation(Op, PredId),
-		aditi_bulk_operation(Op, PredId)).
-	% The syntax used does not change the result of the call.
-match_aditi_builtin(aditi_modify(PredId, _), aditi_modify(PredId, _)).
+match_aditi_builtin(aditi_insert_delete_modify(Op, PredId, _),
+		aditi_insert_delete_modify(Op, PredId, _)).
 
 %-----------------------------------------------------------------------------%
 
