@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -21,14 +21,17 @@
 
 :- interface.
 
-:- import_module hlds_goal, hlds_pred, prog_data.
+:- import_module prog_data.
+:- import_module hlds_goal, hlds_pred.
+:- import_module code_model.
 :- import_module llds, code_info.
+
 :- import_module list, std_util.
 
 :- pred pragma_c_gen__generate_pragma_c_code(code_model::in,
-	pragma_c_code_attributes::in, pred_id::in, proc_id::in,
+	pragma_foreign_code_attributes::in, pred_id::in, proc_id::in,
 	list(prog_var)::in, list(maybe(pair(string, mode)))::in, list(type)::in,
-	hlds_goal_info::in, pragma_c_code_impl::in, code_tree::out,
+	hlds_goal_info::in, pragma_foreign_code_impl::in, code_tree::out,
 	code_info::in, code_info::out) is det.
 
 :- pred pragma_c_gen__struct_name(module_name::in, string::in, int::in,
@@ -38,9 +41,10 @@
 
 :- implementation.
 
-:- import_module hlds_module, hlds_pred, call_gen, llds_out, trace, tree.
+:- import_module hlds_module, hlds_pred, llds_out, trace, tree.
 :- import_module code_util.
 :- import_module options, globals.
+
 :- import_module bool, string, int, assoc_list, set, map, require, term.
 
 % The code we generate for an ordinary (model_det or model_semi) pragma_c_code
@@ -53,11 +57,11 @@
 %	#define MR_PROC_LABEL <procedure label> /* see note (5) below */
 %
 %	<assignment of input values from registers to local variables>
-%	save_registers(); /* see notes (1) and (2) below */
+%	MR_save_registers(); /* see notes (1) and (2) below */
 %	{ <the c code itself> }
 %	<for semidet code, check of r1>
 %	#ifndef CONSERVATIVE_GC
-%	  restore_registers(); /* see notes (1) and (3) below */
+%	  MR_restore_registers(); /* see notes (1) and (3) below */
 %	#endif
 %	<assignment of the output values from local variables to registers>
 %
@@ -73,7 +77,7 @@
 %
 % and the <check of r1> is of the form
 %
-%	if (!r1) GOTO_LABEL(fail_label);
+%	if (!r1) MR_GOTO_LABEL(fail_label);
 %
 % The code we generate for nondet pragma_c_code assumes that this code is
 % the only thing between the procedure prolog and epilog; such pragma_c_codes
@@ -92,7 +96,7 @@
 %	<declaration of one local variable to point to save struct>
 %	<assignment of input values from registers to local variables>
 %	<assignment to save struct pointer>
-%	save_registers(); /* see notes (1) and (2) below */
+%	MR_save_registers(); /* see notes (1) and (2) below */
 %	#define MR_PROC_LABEL <procedure label> /* see note (5) below */
 %	#define SUCCEED()	goto callsuccesslabel
 %	#define SUCCEED_LAST()	goto calllastsuccesslabel
@@ -100,11 +104,11 @@
 %	{ <the user-written call c code> }
 %	{ <the user-written shared c code> }
 % callsuccesslabel:
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed()
 % calllastsuccesslabel: /* see note (4) below) */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed_discard()
 % 	#undef SUCCEED
@@ -112,13 +116,13 @@
 % 	#undef FAIL
 %	#undef MR_PROC_LABEL /* see note (5) below */
 % }
-% Define_label(xxx_i1)
+% MR_define_label(xxx_i1)
 % <code for entry to a later disjunct>
 % {
 %	<declaration of one local variable for each output arg>
 %	<declaration of one local variable to point to save struct>
 %	<assignment to save struct pointer>
-%	save_registers(); /* see notes (1) and (2) below */
+%	MR_save_registers(); /* see notes (1) and (2) below */
 %	#define MR_PROC_LABEL <procedure label> /* see note (5) below */
 %	#define SUCCEED()	goto retrysuccesslabel
 %	#define SUCCEED_LAST()	goto retrylastsuccesslabel
@@ -126,11 +130,11 @@
 %	{ <the user-written retry c code> }
 %	{ <the user-written shared c code> }
 % retrysuccesslabel:
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed()
 % retrylastsuccesslabel: /* see note (4) below) */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed_discard()
 % 	#undef SUCCEED
@@ -153,19 +157,19 @@
 %	<declaration of one local variable to point to save struct>
 %	<assignment of input values from registers to local variables>
 %	<assignment to save struct pointer>
-%	save_registers(); /* see notes (1) and (2) below */
+%	MR_save_registers(); /* see notes (1) and (2) below */
 %	#define MR_PROC_LABEL <procedure label> /* see note (5) below */
 %	#define SUCCEED()	goto callsuccesslabel
 %	#define SUCCEED_LAST()	goto calllastsuccesslabel
 %	#define FAIL()		fail()
 %	{ <the user-written call c code> }
-%	GOTO_LABEL(xxx_i2)
+%	MR_GOTO_LABEL(xxx_i2)
 % callsuccesslabel: /* see note (4) below */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed()
 % calllastsuccesslabel: /* see note (4) below */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed_discard()
 % 	#undef SUCCEED
@@ -173,25 +177,25 @@
 % 	#undef FAIL
 %	#undef MR_PROC_LABEL /* see note (5) below */
 % }
-% Define_label(xxx_i1)
+% MR_define_label(xxx_i1)
 % <code for entry to a later disjunct>
 % {
 %	<declaration of one local variable for each output arg>
 %	<declaration of one local variable to point to save struct>
 %	<assignment to save struct pointer>
-%	save_registers(); /* see notes (1) and (2) below */
+%	MR_save_registers(); /* see notes (1) and (2) below */
 %	#define MR_PROC_LABEL <procedure label> /* see note (5) below */
 %	#define SUCCEED()	goto retrysuccesslabel
 %	#define SUCCEED_LAST()	goto retrylastsuccesslabel
 %	#define FAIL()		fail()
 %	{ <the user-written retry c code> }
-%	GOTO_LABEL(xxx_i2)
+%	MR_GOTO_LABEL(xxx_i2)
 % retrysuccesslabel: /* see note (4) below */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed()
 % retrylastsuccesslabel: /* see note (4) below */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed_discard()
 % 	#undef SUCCEED
@@ -199,7 +203,7 @@
 % 	#undef FAIL
 %	#undef MR_PROC_LABEL /* see note (5) below */
 % }
-% Define_label(xxx_i2)
+% MR_define_label(xxx_i2)
 % {
 %	<declaration of one local variable for each output arg>
 %	<declaration of one local variable to point to save struct>
@@ -210,11 +214,11 @@
 %	#define FAIL()		fail()
 %	{ <the user-written shared c code> }
 % sharedsuccesslabel:
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed()
 % sharedlastsuccesslabel: /* see note (4) below */
-%	restore_registers(); /* see notes (1) and (3) below */
+%	MR_restore_registers(); /* see notes (1) and (3) below */
 %	<assignment of the output values from local variables to registers>
 %	succeed_discard()
 % 	#undef SUCCEED
@@ -264,13 +268,13 @@
 %	If a pragma c_code(will_not_call_mercury, ...) declaration was used,
 %	they will not be emitted.
 %
-% (2)	The call to save_registers() is needed so that if the
-%	C code calls Mercury code, we can call restore_registers()
+% (2)	The call to MR_save_registers() is needed so that if the
+%	C code calls Mercury code, we can call MR_restore_registers()
 %	on entry to the Mercury code (see export.m) to get the
 %	right values of `sp', `hp', `curfr' and `maxfr' for the
 %	recursive invocation of Mercury.
 %
-% (3)	The call to restore_registers() is needed in case the
+% (3)	The call to MR_restore_registers() is needed in case the
 %	C code calls Mercury code which allocates some data
 %	on the heap, and this data is returned from Mercury
 %	through C back to Mercury.  In that case, we need to
@@ -281,7 +285,7 @@
 %	using conservative gc, there is nothing that needs restoring.
 %
 %	When calling nondet code, maxfr may be changed. This is why
-%	we must call restore_registers() from the code we generate for
+%	we must call MR_restore_registers() from the code we generate for
 %	nondet pragma C codes even if we are not using conservative gc.
 %
 % (4)	These labels and the code following them can be optimized away
@@ -316,12 +320,18 @@ pragma_c_gen__generate_pragma_c_code(CodeModel, Attributes,
 			Fields, FieldsContext, First, FirstContext,
 			Later, LaterContext, Treat, Shared, SharedContext,
 			Code)
+	;	{ PragmaImpl = import(Name, HandleReturn, Vars, Context) },
+		{ C_Code = string__append_list([HandleReturn, " ",
+				Name, "(", Vars, ");"]) },
+		pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
+			PredId, ProcId, ArgVars, ArgDatas, OrigArgTypes,
+			C_Code, Context, Code)
 	).
 
 %---------------------------------------------------------------------------%
 
 :- pred pragma_c_gen__ordinary_pragma_c_code(code_model::in,
-	pragma_c_code_attributes::in, pred_id::in, proc_id::in,
+	pragma_foreign_code_attributes::in, pred_id::in, proc_id::in,
 	list(prog_var)::in, list(maybe(pair(string, mode)))::in, list(type)::in,
 	string::in, maybe(prog_context)::in, code_tree::out,
 	code_info::in, code_info::out) is det.
@@ -344,6 +354,9 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 	{ pragma_select_in_args(Args, InArgs) },
 	{ pragma_select_out_args(Args, OutArgs) },
 
+	{ set__init(DeadVars0) },
+	find_dead_input_vars(InArgs, DeadVars0, DeadVars),
+
 	%
 	% Generate code to <save live variables on stack>
 	%
@@ -359,7 +372,13 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 		% (other than the output args) onto the stack
 		{ get_c_arg_list_vars(OutArgs, OutArgs1) },
 		{ set__list_to_set(OutArgs1, OutArgsSet) },
-		call_gen__save_variables(OutArgsSet, SaveVarsCode)
+		code_info__save_variables(OutArgsSet, _, SaveVarsCode)
+	),
+
+	( { CodeModel = model_semi } ->
+		code_info__reserve_r1(ReserveR1_Code)
+	;
+		{ ReserveR1_Code = empty }
 	),
 
 	%
@@ -374,12 +393,31 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 	% currently in r1 elsewhere, so that the C code can assign to
 	% SUCCESS_INDICATOR without clobbering anything important.
 	%
+	% With --no-lazy-code, the call to code_info__reserve_r1 will have
+	% reserved r1, ensuring that none of InArgs is placed there, and
+	% code_info__clear_r1 just releases r1. This reservation of r1
+	% is not strictly necessary, as we generate assignments from
+	% the input registers to C variables before we invoke code that could
+	% assign to SUCCESS_INDICATOR. However, by not storing an argument in
+	% r1, we don't require the C compiler to generate a copy instruction
+	% from r1 to somewhere else in cases where the last use of the variable
+	% we are trying to pass in r1 is after the first reference to
+	% SUCCESS_INDICATOR. Instead we generate that argument directly
+	% into some other location.
+	%
 	( { CodeModel = model_semi } ->
-		code_info__clear_r1(ShuffleR1_Code)
+		code_info__clear_r1(ClearR1_Code)
 	;
-		{ ShuffleR1_Code = empty }
+		{ ClearR1_Code = empty }
 	),
 
+	%
+	% We cannot kill the forward dead input arguments until we have
+	% finished generating the code producing the input variables.
+	% (The forward dead variables will be dead after the pragma_c_code,
+	% but are live during its input phase.)
+	code_info__make_vars_forward_dead(DeadVars),
+
 	%
 	% Generate <declaration of one local variable for each arg>
 	%
@@ -401,13 +439,13 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 	{ InputComp = pragma_c_inputs(InputDescs) },
 
 	%
-	% save_registers(); /* see notes (1) and (2) above */
+	% MR_save_registers(); /* see notes (1) and (2) above */
 	%
 	{ MayCallMercury = will_not_call_mercury ->
 		SaveRegsComp = pragma_c_raw_code("")
 	;
 		SaveRegsComp = pragma_c_raw_code(
-			"\tsave_registers();\n"
+			"\tMR_save_registers();\n"
 		)
 	},
 
@@ -448,14 +486,14 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 
 	%
 	% #ifndef CONSERVATIVE_GC
-	%   restore_registers(); /* see notes (1) and (3) above */
+	%   MR_restore_registers(); /* see notes (1) and (3) above */
 	% #endif
 	%
 	{ MayCallMercury = will_not_call_mercury ->
 		RestoreRegsComp = pragma_c_noop
 	;
 		RestoreRegsComp = pragma_c_raw_code(
-		    "#ifndef CONSERVATIVE_GC\n\trestore_registers();\n#endif\n"
+		"#ifndef CONSERVATIVE_GC\n\tMR_restore_registers();\n#endif\n"
 		)
 	},
 
@@ -485,7 +523,7 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 			CheckR1_Comp, RestoreRegsComp,
 			OutputComp, ProcLabelHashUndef] },
 	{ PragmaCCode = node([
-		pragma_c(Decls, Components, MayCallMercury, no,
+		pragma_c(Decls, Components, MayCallMercury, no, no,
 			MaybeFailLabel, no)
 			- "Pragma C inclusion"
 	]) },
@@ -521,10 +559,11 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 	%
 	{ Code =
 		tree(SaveVarsCode,
+		tree(ReserveR1_Code, 
 		tree(InputVarsCode,
-		tree(ShuffleR1_Code, 
+		tree(ClearR1_Code, 
 		tree(PragmaCCode,
-		     FailureCode))))
+		     FailureCode)))))
 	}.
 
 :- pred make_proc_label_hash_define(module_info, pred_id, proc_id,
@@ -533,6 +572,14 @@ pragma_c_gen__ordinary_pragma_c_code(CodeModel, Attributes,
 
 make_proc_label_hash_define(ModuleInfo, PredId, ProcId,
 		ProcLabelHashDef, ProcLabelHashUndef) :-
+	ProcLabelHashDef = pragma_c_raw_code(string__append_list([
+		"#define\tMR_PROC_LABEL\t",
+		make_proc_label_string(ModuleInfo, PredId, ProcId), "\n"])),
+	ProcLabelHashUndef = pragma_c_raw_code("#undef\tMR_PROC_LABEL\n").
+
+:- func make_proc_label_string(module_info, pred_id, proc_id) = string.
+
+make_proc_label_string(ModuleInfo, PredId, ProcId) = ProcLabelString :-
 	code_util__make_entry_label(ModuleInfo, PredId, ProcId, no,
 		CodeAddr),
 	( CodeAddr = imported(ProcLabel) ->
@@ -541,15 +588,12 @@ make_proc_label_hash_define(ModuleInfo, PredId, ProcId,
 		llds_out__get_label(ProcLabel, yes, ProcLabelString)
 	;
 		error("unexpected code_addr in make_proc_label_hash_define")
-	),
-	ProcLabelHashDef = pragma_c_raw_code(string__append_list([
-			"#define MR_PROC_LABEL ", ProcLabelString, "\n"])),
-	ProcLabelHashUndef = pragma_c_raw_code("#undef MR_PROC_LABEL\n").
+	).
 
 %-----------------------------------------------------------------------------%
 
 :- pred pragma_c_gen__nondet_pragma_c_code(code_model::in,
-	pragma_c_code_attributes::in, pred_id::in, proc_id::in,
+	pragma_foreign_code_attributes::in, pred_id::in, proc_id::in,
 	list(prog_var)::in, list(maybe(pair(string, mode)))::in, list(type)::in,
 	string::in, maybe(prog_context)::in,
 	string::in, maybe(prog_context)::in,
@@ -578,6 +622,12 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 	{ make_proc_label_hash_define(ModuleInfo, CallerPredId, CallerProcId,
 		ProcLabelDefine, ProcLabelUndef) },
 
+	%
+	% Generate a unique prefix for the C labels that we will define
+	%
+	{ ProcLabelString = make_proc_label_string(ModuleInfo,
+		PredId, ProcId) },
+
 	%
 	% Get a list of input and output arguments
 	%
@@ -653,40 +703,48 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 	},
 
 	%
-	% save_registers(); /* see notes (1) and (2) above */
+	% MR_save_registers(); /* see notes (1) and (2) above */
 	%
 	{ MayCallMercury = will_not_call_mercury ->
 		SaveRegs = ""
 	;
-		SaveRegs = "\tsave_registers();\n"
+		SaveRegs = "\tMR_save_registers();\n"
 	},
 
 	%
-	% restore_registers(); /* see notes (1) and (3) above */
+	% MR_restore_registers(); /* see notes (1) and (3) above */
 	%
 	{ MayCallMercury = will_not_call_mercury ->
 		RestoreRegs = ""
 	;
-		RestoreRegs = "\trestore_registers();\n"
+		RestoreRegs = "\tMR_restore_registers();\n"
 	},
 
 	{
 	Succeed	 = "\tMR_succeed();\n",
 	SucceedDiscard = "\tMR_succeed_discard();\n",
 
-	CallDef1 = "#define\tSUCCEED     \tgoto MR_call_success\n",
-	CallDef2 = "#define\tSUCCEED_LAST\tgoto MR_call_success_last\n",
+	CallDef1 = "#define\tSUCCEED     \tgoto MR_call_success_"
+		++ ProcLabelString ++ "\n",
+	CallDef2 = "#define\tSUCCEED_LAST\tgoto MR_call_success_last_"
+		++ ProcLabelString ++ "\n",
 	CallDef3 = "#define\tFAIL\tMR_fail()\n",
 
-	CallSuccessLabel     = "MR_call_success:\n",
-	CallLastSuccessLabel = "MR_call_success_last:\n",
+	CallSuccessLabel     = "MR_call_success_"
+		++ ProcLabelString ++ ":\n",
+	CallLastSuccessLabel = "MR_call_success_last_"
+		++ ProcLabelString ++ ":\n",
 
-	RetryDef1 = "#define\tSUCCEED     \tgoto MR_retry_success\n",
-	RetryDef2 = "#define\tSUCCEED_LAST\tgoto MR_retry_success_last\n",
+	RetryDef1 = "#define\tSUCCEED     \tgoto MR_retry_success_"
+		++ ProcLabelString ++ "\n",
+	RetryDef2 = "#define\tSUCCEED_LAST\tgoto MR_retry_success_last_"
+		++ ProcLabelString ++ "\n",
 	RetryDef3 = "#define\tFAIL\tMR_fail()\n",
 
-	RetrySuccessLabel     = "MR_retry_success:\n",
-	RetryLastSuccessLabel = "MR_retry_success_last:\n",
+	RetrySuccessLabel     = "MR_retry_success_"
+		++ ProcLabelString ++ ":\n",
+	RetryLastSuccessLabel = "MR_retry_success_last_"
+		++ ProcLabelString ++ ":\n",
 
 	Undef1 = "#undef\tSUCCEED\n",
 	Undef2 = "#undef\tSUCCEED_LAST\n",
@@ -752,7 +810,7 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 		],
 		CallBlockCode = node([
 			pragma_c(CallDecls, CallComponents,
-				MayCallMercury, no, no, yes)
+				MayCallMercury, no, no, no, yes)
 				- "Call and shared pragma C inclusion"
 		]),
 
@@ -781,7 +839,7 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 		],
 		RetryBlockCode = node([
 			pragma_c(RetryDecls, RetryComponents,
-				MayCallMercury, no, no, yes)
+				MayCallMercury, no, no, no, yes)
 				- "Retry and shared pragma C inclusion"
 		]),
 
@@ -801,15 +859,19 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 				"Start of the shared block"
 		]),
 
-		SharedDef1 = "#define\tSUCCEED     \tgoto MR_shared_success\n",
-		SharedDef2 = "#define\tSUCCEED_LAST\tgoto MR_shared_success_last\n",
+		SharedDef1 = "#define\tSUCCEED     \tgoto MR_shared_success_"
+			++ ProcLabelString ++ "\n",
+		SharedDef2 = "#define\tSUCCEED_LAST\tgoto MR_shared_success_last_"
+			++ ProcLabelString ++ "\n",
 		SharedDef3 = "#define\tFAIL\tMR_fail()\n",
 
-		SharedSuccessLabel     = "MR_shared_success:\n",
-		SharedLastSuccessLabel = "MR_shared_success_last:\n",
+		SharedSuccessLabel     = "MR_shared_success_"
+			++ ProcLabelString ++ ":\n",
+		SharedLastSuccessLabel = "MR_shared_success_last_"
+			++ ProcLabelString ++ ":\n",
 
 		llds_out__get_label(SharedLabel, yes, LabelStr),
-		string__format("\tGOTO_LABEL(%s);\n", [s(LabelStr)],
+		string__format("\tMR_GOTO_LABEL(%s);\n", [s(LabelStr)],
 			GotoSharedLabel),
 
 		CallDecls = [SaveStructDecl | Decls],
@@ -838,7 +900,7 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 		],
 		CallBlockCode = node([
 			pragma_c(CallDecls, CallComponents,
-				MayCallMercury, yes(SharedLabel), no, yes)
+				MayCallMercury, yes(SharedLabel), no, no, yes)
 				- "Call pragma C inclusion"
 		]),
 
@@ -867,7 +929,7 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 		],
 		RetryBlockCode = node([
 			pragma_c(RetryDecls, RetryComponents,
-				MayCallMercury, yes(SharedLabel), no, yes)
+				MayCallMercury, yes(SharedLabel), no, no, yes)
 				- "Retry pragma C inclusion"
 		]),
 
@@ -895,7 +957,7 @@ pragma_c_gen__nondet_pragma_c_code(CodeModel, Attributes,
 		],
 		SharedBlockCode = node([
 			pragma_c(SharedDecls, SharedComponents,
-				MayCallMercury, no, no, yes)
+				MayCallMercury, no, no, no, yes)
 				- "Shared pragma C inclusion"
 		]),
 
@@ -1045,24 +1107,39 @@ make_pragma_decls([Arg | Args], Decls) :-
 
 %---------------------------------------------------------------------------%
 
+:- pred find_dead_input_vars(list(c_arg)::in, set(prog_var)::in,
+	set(prog_var)::out, code_info::in, code_info::out) is det.
+
+find_dead_input_vars([], DeadVars, DeadVars) --> [].
+find_dead_input_vars([Arg | Args], DeadVars0, DeadVars) -->
+	{ Arg = c_arg(Var, _MaybeName, _Type, _ArgInfo) },
+	( code_info__variable_is_forward_live(Var) ->
+		{ DeadVars1 = DeadVars0 }
+	;
+		{ set__insert(DeadVars0, Var, DeadVars1) }
+	),
+	find_dead_input_vars(Args, DeadVars1, DeadVars).
+
+%---------------------------------------------------------------------------%
+
 % get_pragma_input_vars returns a list of pragma_c_inputs for the pragma_c
 % data structure in the LLDS. It is essentially a list of the input variables,
 % and the corresponding rvals assigned to those (C) variables.
 
 :- pred get_pragma_input_vars(list(c_arg)::in, list(pragma_c_input)::out,
-		code_tree::out, code_info::in, code_info::out) is det.
+	code_tree::out, code_info::in, code_info::out) is det.
 
 get_pragma_input_vars([], [], empty) --> [].
 get_pragma_input_vars([Arg | Args], Inputs, Code) -->
 	{ Arg = c_arg(Var, MaybeName, Type, _ArgInfo) },
-	(
-		{ var_is_not_singleton(MaybeName, Name) }
-	->
-		code_info__produce_variable(Var, Code0, Rval),
+	( { var_is_not_singleton(MaybeName, Name) } ->
+		code_info__produce_variable(Var, FirstCode, Rval),
+		% code_info__produce_variable_in_reg(Var, FirstCode, Lval),
+		% { Rval = lval(Lval) },
 		{ Input = pragma_c_input(Name, Type, Rval) },
-		get_pragma_input_vars(Args, Inputs1, Code1),
+		get_pragma_input_vars(Args, Inputs1, RestCode),
 		{ Inputs = [Input | Inputs1] },
-		{ Code = tree(Code0, Code1) }
+		{ Code = tree(FirstCode, RestCode) }
 	;
 		% if the variable doesn't occur in the ArgNames list,
 		% it can't be used, so we just ignore it
@@ -1094,17 +1171,22 @@ pragma_acquire_regs([Arg | Args], [Reg | Regs]) -->
 
 place_pragma_output_args_in_regs([], [], []) --> [].
 place_pragma_output_args_in_regs([Arg | Args], [Reg | Regs], Outputs) -->
+	place_pragma_output_args_in_regs(Args, Regs, Outputs0),
 	{ Arg = c_arg(Var, MaybeName, OrigType, _ArgInfo) },
 	code_info__release_reg(Reg),
-	code_info__set_var_location(Var, Reg),
-	{
-		var_is_not_singleton(MaybeName, Name)
-	->
-		Outputs = [pragma_c_output(Reg, OrigType, Name) | Outputs0]
+	( code_info__variable_is_forward_live(Var) ->
+		code_info__set_var_location(Var, Reg),
+		{
+			var_is_not_singleton(MaybeName, Name)
+		->
+			PragmaCOutput = pragma_c_output(Reg, OrigType, Name),
+			Outputs = [PragmaCOutput | Outputs0]
+		;
+			Outputs = Outputs0
+		}
 	;
-		Outputs = Outputs0
-	},
-	place_pragma_output_args_in_regs(Args, Regs, Outputs0).
+		{ Outputs = Outputs0 }
+	).
 place_pragma_output_args_in_regs([_|_], [], _) -->
 	{ error("place_pragma_output_args_in_regs: length mismatch") }.
 place_pragma_output_args_in_regs([], [_|_], _) -->
