@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -91,9 +91,9 @@
 		%	ClassMethods, VarNames
 
 	;	instance(list(class_constraint), class_name, list(type),
-			instance_body, tvarset)
+			instance_body, tvarset, module_name)
 		%	DerivingClass, ClassName, Types, 
-		%	MethodInstances, VarNames
+		%	MethodInstances, VarNames, ModuleContainingInstance
 
 	;	nothing.
 		% used for items that should be ignored (currently only
@@ -104,6 +104,15 @@
 	--->	type_only(type)
 	;	type_and_mode(type, mode).
 
+:- type foreign_language
+	--->	c
+% 	;	cplusplus
+% 	;	csharp
+ 	;	managed_cplusplus
+% 	;	java
+% 	;	il
+	.
+
 :- type pred_or_func
 	--->	predicate
 	;	function.
@@ -134,17 +143,21 @@
 %
 
 :- type pragma_type 
-	--->	c_header_code(string)
+			% a foreign language declaration, such as C
+			% header code.
+	--->	foreign_decl(foreign_language, string)
 
-	;	c_code(string)
+	;	foreign(foreign_language, string)
 
-	;	c_code(pragma_c_code_attributes, sym_name, pred_or_func,
-			list(pragma_var), prog_varset, pragma_c_code_impl)
-			% Set of C code attributes, eg.:
-			%	whether or not the C code may call Mercury,
-			%	whether or not the C code is thread-safe
+	;	foreign(pragma_foreign_code_attributes,
+			sym_name, pred_or_func, list(pragma_var),
+			prog_varset, pragma_foreign_code_impl)
+			% Set of foreign code attributes, eg.:
+			%	what language this code is in
+			%	whether or not the code may call Mercury,
+			%	whether or not the code is thread-safe
 			% PredName, Predicate or Function, Vars/Mode, 
-			% VarNames, C Code Implementation Info
+			% VarNames, Foreign Code Implementation Info
 	
 	;	type_spec(sym_name, sym_name, arity, maybe(pred_or_func),
 			maybe(list(mode)), type_subst, tvarset)
@@ -168,12 +181,12 @@
 			% C function name.
 
 	;	import(sym_name, pred_or_func, list(mode),
-			pragma_c_code_attributes, string)
+			pragma_foreign_code_attributes, string)
 			% Predname, Predicate/function, Modes,
-			% Set of C code attributes, eg.:
-			%	whether or not the C code may call Mercury,
-			%	whether or not the C code is thread-safe
-			% C function name.
+			% Set of foreign code attributes, eg.:
+			%    whether or not the foreign code may call Mercury,
+			%    whether or not the foreign code is thread-safe
+			% foreign function name.
 
 	;	source_file(string)
 			% Source file name.
@@ -269,6 +282,7 @@
 					% evaluation
 	;	eval_loop_check		% loop check only
 	;	eval_memo		% memoing + loop check 
+	;	eval_table_io		% memoing I/O actions for debugging
 	;	eval_minimal.		% minimal model 
 					% evaluation 
 %
@@ -333,35 +347,43 @@
 :- type type_subst == assoc_list(tvar, type).
 
 %
-% Stuff for `c_code' pragma.
+% Stuff for `foreign_code' pragma.
 %
 
 	% This type holds information about the implementation details
-	% of procedures defined via `pragma c_code'.
+	% of procedures defined via `pragma foreign_code'.
 	%
 	% All the strings in this type may be accompanied by the context
 	% of their appearance in the source code. These contexts are
-	% used to tell the C compiler where the included C code comes from,
-	% to allow it to generate error messages that refer to the original
-	% appearance of the code in the Mercury program.
-	% The context is missing if the C code was constructed by the compiler.
-:- type pragma_c_code_impl
-	--->	ordinary(		% This is a C definition of a model_det
-					% or model_semi procedure. (We also
-					% allow model_non, until everyone has
-					% had time to adapt to the new way
+	% used to tell the foreign language compiler where the included
+	% code comes from, to allow it to generate error messages that
+	% refer to the original appearance of the code in the Mercury
+	% program.
+	% The context is missing if the foreign code was constructed by
+	% the compiler.
+	% Note that nondet pragma foreign definitions might not be
+	% possible in all foreign languages.
+:- type pragma_foreign_code_impl
+	--->	ordinary(		% This is a foreign language
+					% definition of a model_det
+					% or model_semi procedure. (We
+					% also allow model_non, until
+					% everyone has had time to adapt
+					% to the new way
 					% of handling model_non pragmas.)
-			string,		% The C code of the procedure.
+			string,		% The code of the procedure.
 			maybe(prog_context)
 		)
-	;	nondet(			% This is a C definition of a model_non
+	;	nondet(			% This is a foreign language
+					% definition of a model_non
 					% procedure.
 			string,
 			maybe(prog_context),
 					% The info saved for the time when
 					% backtracking reenters this procedure
-					% is stored in a C struct. This arg
-					% contains the field declarations.
+					% is stored in a data structure.
+					% This arg contains the field
+					% declarations.
 
 			string,
 			maybe(prog_context),
@@ -385,6 +407,15 @@
 					% Shared code that is executed after
 					% both the previous code fragments.
 					% May not access the input variables.
+		)
+	;	import(
+			string,		% Pragma imported C func name
+			string,		% Code to handle return value
+			string,		% Comma seperated variables which
+					% the import function is called
+					% with.
+
+			maybe(prog_context)
 		).
 
 	% The use of this type is explained in the comment at the top of
@@ -426,21 +457,21 @@
 :- type class_method
 	--->	pred(tvarset, inst_varset, existq_tvars, sym_name,
 			list(type_and_mode), maybe(determinism), condition,
-			class_constraints, prog_context)
+			purity, class_constraints, prog_context)
 		%       TypeVarNames, InstVarNames,
 		%	ExistentiallyQuantifiedTypeVars,
 		%	PredName, ArgTypes, Determinism, Cond
-		%	ClassContext, Context
+		%	Purity, ClassContext, Context
 
 	; 	func(tvarset, inst_varset, existq_tvars, sym_name,
 			list(type_and_mode), type_and_mode,
 			maybe(determinism), condition,
-			class_constraints, prog_context)
+			purity, class_constraints, prog_context)
 		%       TypeVarNames, InstVarNames,
 		%	ExistentiallyQuantfiedTypeVars,
 		%	PredName, ArgTypes, ReturnType,
 		%	Determinism, Cond
-		%	ClassContext, Context
+		%	Purity, ClassContext, Context
 
 	; 	pred_mode(inst_varset, sym_name, list(mode),
 			maybe(determinism), condition,
@@ -459,12 +490,20 @@
 	.
 
 :- type instance_method	
-	--->	func_instance(sym_name, sym_name, arity, prog_context)
-				% Method, Instance, Arity, 
-				% Line number of declaration
-	;	pred_instance(sym_name, sym_name, arity, prog_context)
-				% Method, Instance, Arity, 
+	--->	instance_method(pred_or_func, sym_name, instance_proc_def,
+			arity, prog_context).
+				% PredOrFunc, Method, Instance, Arity, 
 				% Line number of declaration
+
+:- type instance_proc_def
+		% defined using the `pred(...) is <Name>' syntax
+	--->	name(sym_name)	
+
+		% defined using clauses
+	;	clauses(
+			list(item)	% the items must be either
+					% pred_clause or func_clause items
+		)
 	.
 
 :- type instance_body
@@ -480,25 +519,39 @@
 
 		% an abstract type for representing a set of
 		% `pragma_c_code_attribute's.
-:- type pragma_c_code_attributes.
+:- type pragma_foreign_code_attributes.
 
-:- pred default_attributes(pragma_c_code_attributes).
-:- mode default_attributes(out) is det.
+:- pred default_attributes(foreign_language, pragma_foreign_code_attributes).
+:- mode default_attributes(in, out) is det.
 
-:- pred may_call_mercury(pragma_c_code_attributes, may_call_mercury).
+:- pred may_call_mercury(pragma_foreign_code_attributes, may_call_mercury).
 :- mode may_call_mercury(in, out) is det.
 
-:- pred set_may_call_mercury(pragma_c_code_attributes, may_call_mercury,
-		pragma_c_code_attributes).
+:- pred set_may_call_mercury(pragma_foreign_code_attributes, may_call_mercury,
+		pragma_foreign_code_attributes).
 :- mode set_may_call_mercury(in, in, out) is det.
 
-:- pred thread_safe(pragma_c_code_attributes, thread_safe).
+:- pred thread_safe(pragma_foreign_code_attributes, thread_safe).
 :- mode thread_safe(in, out) is det.
 
-:- pred set_thread_safe(pragma_c_code_attributes, thread_safe,
-		pragma_c_code_attributes).
+:- pred set_thread_safe(pragma_foreign_code_attributes, thread_safe,
+		pragma_foreign_code_attributes).
 :- mode set_thread_safe(in, in, out) is det.
 
+:- pred foreign_language(pragma_foreign_code_attributes, foreign_language).
+:- mode foreign_language(in, out) is det.
+
+:- pred set_foreign_language(pragma_foreign_code_attributes, foreign_language,
+		pragma_foreign_code_attributes).
+:- mode set_foreign_language(in, in, out) is det.
+
+:- pred tabled_for_io(pragma_foreign_code_attributes, tabled_for_io).
+:- mode tabled_for_io(in, out) is det.
+
+:- pred set_tabled_for_io(pragma_foreign_code_attributes, tabled_for_io,
+		pragma_foreign_code_attributes).
+:- mode set_tabled_for_io(in, in, out) is det.
+
 	% For pragma c_code, there are two different calling conventions,
 	% one for C code that may recursively call Mercury code, and another
 	% more efficient one for the case when we know that the C code will
@@ -514,6 +567,10 @@
 	--->	not_thread_safe
 	;	thread_safe.
 
+:- type tabled_for_io
+	--->	not_tabled_for_io
+	;	tabled_for_io.
+
 :- type pragma_var    
 	--->	pragma_var(prog_var, string, mode).
 	  	% variable, name, mode
@@ -563,7 +620,7 @@
 
 	% atomic goals
 	;	call(sym_name, list(prog_term), purity)
-	;	unify(prog_term, prog_term).
+	;	unify(prog_term, prog_term, purity).
 
 :- type goals		==	list(goal).
 
@@ -611,7 +668,7 @@
 			% one day we might allow types to take
 			% value parameters as well as type parameters.
 
-% type_defn/3 define above
+% type_defn/3 is defined above as a constructor for item/0
 
 :- type type_defn	
 	--->	du_type(sym_name, list(type_param), list(constructor),
@@ -629,7 +686,13 @@
 			list(constructor_arg)
 		).
 
-:- type constructor_arg	==	pair(string, type).
+:- type constructor_arg	==
+		pair(
+			maybe(ctor_field_name),
+			type
+		).
+
+:- type ctor_field_name == sym_name.
 
 	% An equality_pred specifies the name of a user-defined predicate
 	% used for equality on a type.  See the chapter on them in the
@@ -688,11 +751,8 @@
 % inst_defn/3 defined above
 
 :- type inst_defn	
-	--->	eqv_inst(sym_name, list(inst_param), inst)
-	;	abstract_inst(sym_name, list(inst_param)).
-
-	% probably inst parameters should be variables not terms
-:- type inst_param	==	inst_term.
+	--->	eqv_inst(sym_name, list(inst_var), inst)
+	;	abstract_inst(sym_name, list(inst_var)).
 
 	% An `inst_name' is used as a key for the inst_table.
 	% It is either a user-defined inst `user_inst(Name, Args)',
@@ -753,7 +813,7 @@
 % mode_defn/3 defined above
 
 :- type mode_defn	
-	--->	eqv_mode(sym_name, list(inst_param), mode).
+	--->	eqv_mode(sym_name, list(inst_var), mode).
 
 :- type (mode)		
 	--->	((inst) -> (inst))
@@ -878,31 +938,77 @@
 	--->	must_be_qualified
 	;	may_be_unqualified.
 
+	% Convert the foreign code attributes to their source code
+	% representations suitable for placing in the attributes list of
+	% the pragma (not all attributes have one).
+	% In particular, the foreign language attribute needs to be
+	% handled separately as it belongs at the start of the pragma.
+:- pred attributes_to_strings(pragma_foreign_code_attributes::in,
+		list(string)::out) is det.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- type pragma_c_code_attributes
+:- type pragma_foreign_code_attributes
 	--->	attributes(
-			may_call_mercury,
-			thread_safe
+			foreign_language 	:: foreign_language,
+			may_call_mercury	:: may_call_mercury,
+			thread_safe		:: thread_safe,
+			tabled_for_io		:: tabled_for_io
 		).
 
-default_attributes(attributes(may_call_mercury, not_thread_safe)).
+default_attributes(Language, 
+	attributes(Language, may_call_mercury, not_thread_safe, 
+		not_tabled_for_io)).
 
-may_call_mercury(Attrs, MayCallMercury) :-
-	Attrs = attributes(MayCallMercury, _).
+may_call_mercury(Attrs, Attrs ^ may_call_mercury).
 
-thread_safe(Attrs, ThreadSafe) :-
-	Attrs = attributes(_, ThreadSafe).
+thread_safe(Attrs, Attrs ^ thread_safe).
+
+foreign_language(Attrs, Attrs ^ foreign_language).
+
+tabled_for_io(Attrs, Attrs ^ tabled_for_io).
 
 set_may_call_mercury(Attrs0, MayCallMercury, Attrs) :-
-	Attrs0 = attributes(_, ThreadSafe),
-	Attrs  = attributes(MayCallMercury, ThreadSafe).
+	Attrs = Attrs0 ^ may_call_mercury := MayCallMercury.
 
 set_thread_safe(Attrs0, ThreadSafe, Attrs) :-
-	Attrs0 = attributes(MayCallMercury, _),
-	Attrs  = attributes(MayCallMercury, ThreadSafe).
+	Attrs = Attrs0 ^ thread_safe := ThreadSafe.
+
+set_foreign_language(Attrs0, ForeignLanguage, Attrs) :-
+	Attrs = Attrs0 ^ foreign_language := ForeignLanguage.
+
+set_tabled_for_io(Attrs0, TabledForIo, Attrs) :-
+	Attrs = Attrs0 ^ tabled_for_io := TabledForIo.
+
+attributes_to_strings(Attrs, StringList) :-
+	% We ignore Lang because it isn't an attribute that you can put
+	% in the attribute list -- the foreign language specifier string
+	% is at the start of the pragma.
+	Attrs = attributes(_Lang, MayCallMercury, ThreadSafe, TabledForIO),
+	(
+		MayCallMercury = may_call_mercury,
+		MayCallMercuryStr = "may_call_mercury"
+	;
+		MayCallMercury = will_not_call_mercury,
+		MayCallMercuryStr = "will_not_call_mercury"
+	),
+	(
+		ThreadSafe = not_thread_safe,
+		ThreadSafeStr = "not_thread_safe"
+	;
+		ThreadSafe = thread_safe,
+		ThreadSafeStr = "thread_safe"
+	),
+	(
+		TabledForIO = tabled_for_io,
+		TabledForIOStr = "tabled_for_io"
+	;
+		TabledForIO = not_tabled_for_io,
+		TabledForIOStr = "not_tabled_for_io"
+	),
+	StringList = [MayCallMercuryStr, ThreadSafeStr, TabledForIOStr].
 
 %-----------------------------------------------------------------------------%
