@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -22,7 +22,7 @@
 
 :- implementation.
 
-:- import_module prog_io, prog_io_goal, prog_util.
+:- import_module globals, prog_io, prog_io_goal, prog_util.
 :- import_module term_util, term_errors.
 :- import_module int, map, string, std_util, bool, require.
 
@@ -70,88 +70,153 @@ parse_pragma_type(_, "source_file", PragmaTerms, ErrorTerm, _VarSet, Result) :-
 			ErrorTerm)
 	).
 
-parse_pragma_type(_, "c_header_code", PragmaTerms,
-			ErrorTerm, _VarSet, Result) :-
-    	(
-       	    PragmaTerms = [HeaderTerm]
-        ->
-	    (
-    	        HeaderTerm = term__functor(term__string(HeaderCode), [], _)
-	    ->
-	        Result = ok(pragma(c_header_code(HeaderCode)))
-	    ;
-		Result = error("expected string for C header code", HeaderTerm)
-	    )
+parse_pragma_type(ModuleName, "foreign_decl", PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	parse_pragma_foreign_decl_pragma(ModuleName, "foreign_decl",
+		PragmaTerms, ErrorTerm, VarSet, Result).
+		
+parse_pragma_type(ModuleName, "c_header_code", PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	(
+	    PragmaTerms = [term__functor(_, _, Context) | _]
+	->
+	    LangC = term__functor(term__string("C"), [], Context),
+	    parse_pragma_foreign_decl_pragma(ModuleName, "c_header_code",
+		[LangC | PragmaTerms], ErrorTerm, VarSet, Result)
 	;
-	    Result = error(
-"wrong number of arguments in `:- pragma c_header_code' declaration", 
-			    ErrorTerm)
-        ).
+	    Result = error("wrong number of arguments or unexpected variable in `:- pragma c_header_code' declaration", 
+		    ErrorTerm)
+	).
 
+parse_pragma_type(ModuleName, "foreign_code", PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	parse_pragma_foreign_code_pragma(ModuleName, "foreign_code",
+		    PragmaTerms, ErrorTerm, VarSet, Result).
+
+parse_pragma_type(ModuleName, "foreign_proc", PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	parse_pragma_foreign_proc_pragma(ModuleName, "foreign_proc",
+		    PragmaTerms, ErrorTerm, VarSet, Result).
+
+	% pragma c_code is almost as if we have written foreign_code
+	% or foreign_proc with the language set to "C".
+	% There are a few differences (error messages, some deprecated
+	% syntax is still supported for c_code) so we pass the original
+	% pragma name to parse_pragma_foreign_code_pragma.
 parse_pragma_type(ModuleName, "c_code", PragmaTerms,
 			ErrorTerm, VarSet, Result) :-
 	(
-    	    PragmaTerms = [Just_C_Code_Term]
+			% arity = 1 (same as foreign_code)
+		PragmaTerms = [term__functor(_, _, Context)]
 	->
-	    (
-		Just_C_Code_Term = term__functor(term__string(Just_C_Code), [],
-			_)
-	    ->
-	        Result = ok(pragma(c_code(Just_C_Code)))
-	    ;
-		Result = error("expected string for C code", Just_C_Code_Term)
-	    )
+		LangC = term__functor(term__string("C"), [], Context),
+		parse_pragma_foreign_code_pragma(ModuleName, "c_code",
+			[LangC | PragmaTerms], ErrorTerm, VarSet, Result)
 	;
-    	    PragmaTerms = [PredAndVarsTerm, C_CodeTerm]
+			% arity > 1 (same as foreign_proc)
+		PragmaTerms = [term__functor(_, _, Context) | _]
 	->
-	    % XXX we should issue a warning; this syntax is deprecated.
-	    % Result = error("pragma c_code doesn't say whether it can call mercury", PredAndVarsTerm)
-		    % may_call_mercury is a conservative default.
-	    default_attributes(Attributes),
-	    (
-		C_CodeTerm = term__functor(term__string(C_Code), [], Context)
-	    ->
-	        parse_pragma_c_code(ModuleName, Attributes, PredAndVarsTerm,
-	    	    ordinary(C_Code, yes(Context)), VarSet, Result)
-	    ;
-		Result = error("invalid `:- pragma c_code' declaration -- expecting either `may_call_mercury' or `will_not_call_mercury', and a string for C code",
-		    C_CodeTerm)
-	    )
+		LangC = term__functor(term__string("C"), [], Context),
+		parse_pragma_foreign_proc_pragma(ModuleName, "c_code",
+			[LangC | PragmaTerms], ErrorTerm, VarSet, Result)
 	;
-    	    PragmaTerms = [PredAndVarsTerm, FlagsTerm, C_CodeTerm]
+	    Result = error("wrong number of arguments or unexpected variable in `:- pragma c_code' declaration", 
+		    ErrorTerm)
+	).
+
+:- pred parse_foreign_language(term, foreign_language).
+:- mode parse_foreign_language(in, out) is semidet.
+
+parse_foreign_language(term__functor(term__string(String), _, _), Lang) :-
+	globals__convert_foreign_language(String, Lang).
+
+	% This predicate parses both c_header_code and foreign_decl pragmas.
+:- pred parse_pragma_foreign_decl_pragma(module_name, string,
+		list(term), term, varset, maybe1(item)).
+:- mode parse_pragma_foreign_decl_pragma(in, in, in, in, in, out) is det.
+parse_pragma_foreign_decl_pragma(_ModuleName, Pragma, PragmaTerms,
+			ErrorTerm, _VarSet, Result) :-
+	string__format("invalid `:- pragma %s' declaration ", [s(Pragma)],
+		InvalidDeclStr),
+	( 
+		PragmaTerms = [Lang, HeaderTerm]
 	->
-	    (
-		C_CodeTerm = term__functor(term__string(C_Code), [], Context)
-	    ->
-		( parse_pragma_c_code_attributes_term(FlagsTerm, Flags) ->
-	            parse_pragma_c_code(ModuleName, Flags, PredAndVarsTerm,
-			ordinary(C_Code, yes(Context)), VarSet, Result)
-	        ; parse_pragma_c_code_attributes_term(PredAndVarsTerm, Flags) ->
-		    % XXX we should issue a warning; this syntax is deprecated
-	            parse_pragma_c_code(ModuleName, Flags, FlagsTerm,
-			ordinary(C_Code, yes(Context)), VarSet, Result)
-	        ;
-		    Result = error("invalid second argument in `:- pragma c_code' declaration -- expecting a C code attribute or list of attributes'",
-			FlagsTerm)
+		( 
+			parse_foreign_language(Lang, ForeignLanguage)
+		->
+			(
+				HeaderTerm = term__functor(term__string(
+					HeaderCode), [], _)
+			->
+				Result = ok(pragma(foreign_decl(
+					ForeignLanguage, HeaderCode)))
+			;
+				ErrMsg = "-- expected string for foreign declaration code",
+				Result = error(string__append(InvalidDeclStr,
+					ErrMsg), HeaderTerm)
+			)
+		;
+			ErrMsg = "-- invalid language parameter",
+			Result = error(string__append(InvalidDeclStr, ErrMsg), 
+				Lang)
 		)
-	    ;
-		Result = error("invalid third argument in `:- pragma c_code' declaration -- expecting string for C code",
-		    C_CodeTerm)
-	    )
 	;
-	    (
-    	        PragmaTerms = [PredAndVarsTerm, FlagsTerm,
-		    FieldsTerm, FirstTerm, LaterTerm],
-		term__context_init(DummyContext),
-		SharedTerm = term__functor(term__atom("common_code"),
-			[term__functor(term__string(""), [], DummyContext)],
-			DummyContext)
-	    ;
-    	        PragmaTerms = [PredAndVarsTerm, FlagsTerm,
-		    FieldsTerm, FirstTerm, LaterTerm, SharedTerm]
-	    )
-	->
-	    ( parse_pragma_c_code_attributes_term(FlagsTerm, Flags) ->
+	    string__format("invalid `:- pragma %s' declaration ", [s(Pragma)],
+		ErrorStr),
+	    Result = error(ErrorStr, ErrorTerm)
+        ).
+
+	% This predicate parses both c_code and foreign_code pragmas.
+	% Processing of foreign_proc (or c_code that defines a procedure)
+	% is handled in parse_pragma_foreign_proc_pragma below.
+:- pred parse_pragma_foreign_code_pragma(module_name, string,
+		list(term), term, varset, maybe1(item)).
+:- mode parse_pragma_foreign_code_pragma(in, in, in, in, in, out) is det.
+
+parse_pragma_foreign_code_pragma(ModuleName, Pragma, PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	string__format("invalid `:- pragma %s' declaration ", [s(Pragma)],
+		InvalidDeclStr),
+
+
+	Check1 = (func(PTerms1, ForeignLanguage) = Res is semidet :- 
+		PTerms1 = [Just_Code_Term],
+		(
+			Just_Code_Term = term__functor(term__string(
+				Just_Code), [], _)
+		->
+			Res = ok(pragma(foreign(ForeignLanguage, 
+				Just_Code)))
+		;
+			ErrMsg = "-- expected string for foreign code",
+			Res = error(string__append(InvalidDeclStr, ErrMsg),
+				ErrorTerm)
+		)
+	),
+
+		% After foreign_proc has bootstrapped and the library has
+		% been updated to use foreign_proc where appropriate, we
+		% should uncomment this code and remove Check2, Check3,
+		% Check5, Check6 and the other definition of CheckLength. 
+/*
+	CheckLength = (func(PTermsLen, ForeignLanguage) = Res :- 
+		( 
+			Res0 = Check1(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			ErrMsg = "-- wrong number of arguments",
+			Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				ErrorTerm)
+		)	
+	),
+*/
+
+	Check6 = (func(PTerms6, ForeignLanguage) = Res is semidet :- 
+            PTerms6 = [PredAndVarsTerm, FlagsTerm,
+		    FieldsTerm, FirstTerm, LaterTerm, SharedTerm],
+	    ( parse_pragma_foreign_code_attributes_term(
+	    		ForeignLanguage, FlagsTerm, Flags) ->
 	        ( parse_pragma_keyword("local_vars", FieldsTerm, Fields,
 			FieldsContext) ->
 	            ( parse_pragma_keyword("first_code", FirstTerm, First,
@@ -160,76 +225,407 @@ parse_pragma_type(ModuleName, "c_code", PragmaTerms,
 				LaterContext) ->
 	                    ( parse_pragma_keyword("shared_code", SharedTerm,
 			    		Shared, SharedContext) ->
-	        	        parse_pragma_c_code(ModuleName, Flags,
-				    PredAndVarsTerm,
+	        	        parse_pragma_foreign_code(ModuleName,
+				    Flags, PredAndVarsTerm,
 				    nondet(Fields, yes(FieldsContext),
 				    	First, yes(FirstContext),
 					Later, yes(LaterContext),
 					share, Shared, yes(SharedContext)),
-				    VarSet, Result)
+				    VarSet, Res)
 		            ; parse_pragma_keyword("duplicated_code",
 			    		SharedTerm, Shared, SharedContext) ->
-	        	        parse_pragma_c_code(ModuleName, Flags,
-				    PredAndVarsTerm,
+	        	        parse_pragma_foreign_code(ModuleName,
+				    Flags, PredAndVarsTerm,
 				    nondet(Fields, yes(FieldsContext),
 				    	First, yes(FirstContext),
 					Later, yes(LaterContext),
 					duplicate, Shared, yes(SharedContext)),
-				    VarSet, Result)
+				    VarSet, Res)
 		            ; parse_pragma_keyword("common_code", SharedTerm,
 			    		Shared, SharedContext) ->
-	        	        parse_pragma_c_code(ModuleName, Flags,
-				    PredAndVarsTerm,
+	        	        parse_pragma_foreign_code(ModuleName, 
+				    Flags, PredAndVarsTerm,
 				    nondet(Fields, yes(FieldsContext),
 				    	First, yes(FirstContext),
 					Later, yes(LaterContext),
 					automatic, Shared, yes(SharedContext)),
-				    VarSet, Result)
+				    VarSet, Res)
 		            ;
-		                Result = error("invalid sixth argument in `:- pragma c_code' declaration -- expecting `common_code(<code>)'",
-			            LaterTerm)
+		                ErrMsg = "-- invalid seventh argument, expecting `common_code(<code>)'",
+				Res = error(string__append(InvalidDeclStr,
+					ErrMsg), SharedTerm)
 			    )
 		        ;
-		            Result = error("invalid fifth argument in `:- pragma c_code' declaration -- expecting `retry_code(<code>)'",
-			        LaterTerm)
+		            ErrMsg = "-- invalid sixth argument, expecting `retry_code(<code>)'",
+			    Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	LaterTerm)
 			)
 		    ;
-		        Result = error("invalid fourth argument in `:- pragma c_code' declaration -- expecting `first_code(<code>)'",
-			    FirstTerm)
+		        ErrMsg = "-- invalid fifth argument, expecting `first_code(<code>)'",
+			Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	FirstTerm)
 		    )
 		;
-		    Result = error("invalid third argument in `:- pragma c_code' declaration -- expecting `local_vars(<fields>)'",
-			FieldsTerm)
+		    ErrMsg = "-- invalid fourth argument, expecting `local_vars(<fields>)'",
+		    Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	FieldsTerm)
 		)
 	    ;
-		Result = error("invalid second argument in `:- pragma c_code' declaration -- expecting pragma c_code attribute or list of attributes'",
-			FlagsTerm)
+		ErrMsg = "-- invalid third argument, expecting foreign code attribute or list of attributes",
+		Res = error(string__append(InvalidDeclStr, ErrMsg), FlagsTerm)
 	    )
+	),
+
+	Check5 = (func(PTerms5, ForeignLanguage) = Res is semidet :- 
+		PTerms5 = [PredAndVarsTerm, FlagsTerm,
+		    FieldsTerm, FirstTerm, LaterTerm],
+		term__context_init(DummyContext),
+		SharedTerm = term__functor(term__atom("common_code"),
+			[term__functor(term__string(""), [], DummyContext)],
+			DummyContext),
+		Res = Check6([PredAndVarsTerm, FlagsTerm, FieldsTerm, FirstTerm,
+			LaterTerm, SharedTerm], ForeignLanguage)
+	),
+
+	Check3 = (func(PTerms3, ForeignLanguage) = Res is semidet :- 
+    	    PTerms3 = [PredAndVarsTerm, FlagsTerm, CodeTerm],
+	    (
+		CodeTerm = term__functor(term__string(Code), [], Context)
+	    ->
+		( parse_pragma_foreign_code_attributes_term(ForeignLanguage, 
+			FlagsTerm, Flags) ->
+		    parse_pragma_foreign_code(ModuleName, Flags,
+			PredAndVarsTerm, ordinary(Code, yes(Context)),
+			VarSet, Res)
+	        ; parse_pragma_foreign_code_attributes_term(ForeignLanguage,
+			PredAndVarsTerm, Flags) ->
+		    % XXX we should issue a warning; this syntax is deprecated
+		    % We will continue to accept this if c_code is used, but
+		    % not with foreign_code
+		    ( Pragma = "c_code" ->
+	                parse_pragma_foreign_code(ModuleName,
+			    Flags, FlagsTerm, ordinary(Code, yes(Context)),
+			    VarSet, Res)
+		    ;
+			ErrMsg = "-- invalid second argument, expecting predicate or function mode",
+		        Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	    PredAndVarsTerm)
+		    )	
+	        ;
+		    ErrMsg = "-- invalid third argument, expecting a foreign code attribute or list of attributes",
+		    Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	FlagsTerm)
+		)
+	    ;
+		ErrMsg = "-- invalid fourth argument, expecting string containing foreign code",
+		Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	CodeTerm)
+	    )
+	),
+
+	Check2 = (func(PTerms2, ForeignLanguage) = Res is semidet :- 
+		PTerms2 = [PredAndVarsTerm, CodeTerm],
+	    	% XXX we should issue a warning; this syntax is deprecated.
+		% We will continue to accept this if c_code is used, but
+		% not with foreign_code
+		( 
+			Pragma = "c_code"
+		->
+			% may_call_mercury is a conservative default.
+			default_attributes(ForeignLanguage, Attributes),
+			(
+			    CodeTerm = term__functor(term__string(Code), [],
+				Context)
+			->
+			    parse_pragma_foreign_code(ModuleName, 
+			        Attributes, PredAndVarsTerm, ordinary(Code,
+				yes(Context)), VarSet, Res)
+			;
+			    ErrMsg = "-- expecting either `may_call_mercury' or `will_not_call_mercury', and a string for foreign code",
+			    Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				CodeTerm)
+	    		)
+		;
+			ErrMsg = "-- doesn't say whether it can call mercury",
+			Res = error(string__append(InvalidDeclStr, ErrMsg),
+				ErrorTerm)
+		)
+	),
+
+
+	CheckLength = (func(PTermsLen, ForeignLanguage) = Res :- 
+		( 
+			Res0 = Check1(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check2(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check3(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check5(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check6(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			ErrMsg = "-- wrong number of arguments",
+			Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				ErrorTerm)
+		)	
+	),
+
+	CheckLanguage = (func(PTermsLang) = Res is semidet :- 
+		PTermsLang = [Lang | Rest],
+		( 
+	    		parse_foreign_language(Lang, ForeignLanguage)
+		->
+			Res = CheckLength(Rest, ForeignLanguage)
+		;
+			ErrMsg = "-- invalid language parameter",
+			Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				Lang)
+		)
+	),
+
+	(
+		Result0 = CheckLanguage(PragmaTerms)
+	->
+		Result = Result0
 	;
-	    Result = error(
-	    "wrong number of arguments in `:- pragma c_code' declaration", 
-		    ErrorTerm)
+		ErrMsg0 = "-- wrong number of arguments",
+		Result = error(string__append(InvalidDeclStr, ErrMsg0),
+			ErrorTerm)
 	).
 
+
+	% This predicate parses both c_code and foreign_proc pragmas.
+:- pred parse_pragma_foreign_proc_pragma(module_name, string,
+		list(term), term, varset, maybe1(item)).
+:- mode parse_pragma_foreign_proc_pragma(in, in, in, in, in, out) is det.
+
+parse_pragma_foreign_proc_pragma(ModuleName, Pragma, PragmaTerms,
+			ErrorTerm, VarSet, Result) :-
+	string__format("invalid `:- pragma %s' declaration ", [s(Pragma)],
+		InvalidDeclStr),
+
+	Check6 = (func(PTerms6, ForeignLanguage) = Res is semidet :- 
+            PTerms6 = [PredAndVarsTerm, FlagsTerm,
+		    FieldsTerm, FirstTerm, LaterTerm, SharedTerm],
+	    ( parse_pragma_foreign_code_attributes_term(
+	    		ForeignLanguage, FlagsTerm, Flags) ->
+	        ( parse_pragma_keyword("local_vars", FieldsTerm, Fields,
+			FieldsContext) ->
+	            ( parse_pragma_keyword("first_code", FirstTerm, First,
+		    		FirstContext) ->
+	                ( parse_pragma_keyword("retry_code", LaterTerm, Later,
+				LaterContext) ->
+	                    ( parse_pragma_keyword("shared_code", SharedTerm,
+			    		Shared, SharedContext) ->
+	        	        parse_pragma_foreign_code(ModuleName,
+				    Flags, PredAndVarsTerm,
+				    nondet(Fields, yes(FieldsContext),
+				    	First, yes(FirstContext),
+					Later, yes(LaterContext),
+					share, Shared, yes(SharedContext)),
+				    VarSet, Res)
+		            ; parse_pragma_keyword("duplicated_code",
+			    		SharedTerm, Shared, SharedContext) ->
+	        	        parse_pragma_foreign_code(ModuleName,
+				    Flags, PredAndVarsTerm,
+				    nondet(Fields, yes(FieldsContext),
+				    	First, yes(FirstContext),
+					Later, yes(LaterContext),
+					duplicate, Shared, yes(SharedContext)),
+				    VarSet, Res)
+		            ; parse_pragma_keyword("common_code", SharedTerm,
+			    		Shared, SharedContext) ->
+	        	        parse_pragma_foreign_code(ModuleName, 
+				    Flags, PredAndVarsTerm,
+				    nondet(Fields, yes(FieldsContext),
+				    	First, yes(FirstContext),
+					Later, yes(LaterContext),
+					automatic, Shared, yes(SharedContext)),
+				    VarSet, Res)
+		            ;
+		                ErrMsg = "-- invalid seventh argument, expecting `common_code(<code>)'",
+				Res = error(string__append(InvalidDeclStr,
+					ErrMsg), SharedTerm)
+			    )
+		        ;
+		            ErrMsg = "-- invalid sixth argument, expecting `retry_code(<code>)'",
+			    Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	LaterTerm)
+			)
+		    ;
+		        ErrMsg = "-- invalid fifth argument, expecting `first_code(<code>)'",
+			Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	FirstTerm)
+		    )
+		;
+		    ErrMsg = "-- invalid fourth argument, expecting `local_vars(<fields>)'",
+		    Res = error(string__append(InvalidDeclStr, ErrMsg),
+			    	FieldsTerm)
+		)
+	    ;
+		ErrMsg = "-- invalid third argument, expecting foreign code attribute or list of attributes",
+		Res = error(string__append(InvalidDeclStr, ErrMsg), FlagsTerm)
+	    )
+	),
+
+	Check5 = (func(PTerms5, ForeignLanguage) = Res is semidet :- 
+		PTerms5 = [PredAndVarsTerm, FlagsTerm,
+		    FieldsTerm, FirstTerm, LaterTerm],
+		term__context_init(DummyContext),
+		SharedTerm = term__functor(term__atom("common_code"),
+			[term__functor(term__string(""), [], DummyContext)],
+			DummyContext),
+		Res = Check6([PredAndVarsTerm, FlagsTerm, FieldsTerm, FirstTerm,
+			LaterTerm, SharedTerm], ForeignLanguage)
+	),
+
+	Check3 = (func(PTerms3, ForeignLanguage) = Res is semidet :- 
+    	    PTerms3 = [PredAndVarsTerm, FlagsTerm, CodeTerm],
+	    (
+		CodeTerm = term__functor(term__string(Code), [], Context)
+	    ->
+		( parse_pragma_foreign_code_attributes_term(ForeignLanguage, 
+			FlagsTerm, Flags) ->
+		    parse_pragma_foreign_code(ModuleName, Flags,
+			PredAndVarsTerm, ordinary(Code, yes(Context)),
+			VarSet, Res)
+	        ; parse_pragma_foreign_code_attributes_term(ForeignLanguage,
+			PredAndVarsTerm, Flags) ->
+		    % XXX we should issue a warning; this syntax is deprecated
+		    % We will continue to accept this if c_code is used, but
+		    % not with foreign_code
+		    ( Pragma = "c_code" ->
+	                parse_pragma_foreign_code(ModuleName,
+			    Flags, FlagsTerm, ordinary(Code, yes(Context)),
+			    VarSet, Res)
+		    ;
+			ErrMsg = "-- invalid second argument, expecting predicate or function mode",
+		        Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	    PredAndVarsTerm)
+		    )	
+	        ;
+		    ErrMsg = "-- invalid third argument, expecting a foreign code attribute or list of attributes",
+		    Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	FlagsTerm)
+		)
+	    ;
+		ErrMsg = "-- invalid fourth argument, expecting string containing foreign code",
+		Res = error(string__append(InvalidDeclStr, ErrMsg), 
+		    	CodeTerm)
+	    )
+	),
+
+
+	Check2 = (func(PTerms2, ForeignLanguage) = Res is semidet :- 
+		PTerms2 = [PredAndVarsTerm, CodeTerm],
+	    	% XXX we should issue a warning; this syntax is deprecated.
+		% We will continue to accept this if c_code is used, but
+		% not with foreign_code
+		( 
+			Pragma = "c_code"
+		->
+			% may_call_mercury is a conservative default.
+			default_attributes(ForeignLanguage, Attributes),
+			(
+			    CodeTerm = term__functor(term__string(Code), [],
+				Context)
+			->
+			    parse_pragma_foreign_code(ModuleName, 
+			        Attributes, PredAndVarsTerm, ordinary(Code,
+				yes(Context)), VarSet, Res)
+			;
+			    ErrMsg = "-- expecting either `may_call_mercury' or `will_not_call_mercury', and a string for foreign code",
+			    Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				CodeTerm)
+	    		)
+		;
+			ErrMsg = "-- doesn't say whether it can call mercury",
+			Res = error(string__append(InvalidDeclStr, ErrMsg),
+				ErrorTerm)
+		)
+	),
+
+	CheckLength = (func(PTermsLen, ForeignLanguage) = Res :- 
+		( 
+			Res0 = Check2(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check3(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check5(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			Res0 = Check6(PTermsLen, ForeignLanguage)
+		->	
+			Res = Res0
+		;
+			ErrMsg = "-- wrong number of arguments",
+			Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				ErrorTerm)
+		)	
+	),
+
+	CheckLanguage = (func(PTermsLang) = Res is semidet :- 
+		PTermsLang = [Lang | Rest],
+		( 
+	    		parse_foreign_language(Lang, ForeignLanguage)
+		->
+			Res = CheckLength(Rest, ForeignLanguage)
+		;
+			ErrMsg = "-- invalid language parameter",
+			Res = error(string__append(InvalidDeclStr, ErrMsg), 
+				Lang)
+		)
+	),
+
+	(
+		Result0 = CheckLanguage(PragmaTerms)
+	->
+		Result = Result0
+	;
+		ErrMsg0 = "-- wrong number of arguments",
+		Result = error(string__append(InvalidDeclStr, ErrMsg0),
+			ErrorTerm)
+	).
+
+
 parse_pragma_type(ModuleName, "import", PragmaTerms,
 			ErrorTerm, _VarSet, Result) :-
+		% XXX we assume all imports are C
+	ForeignLanguage = c,
 	(
 	    (
-		PragmaTerms = [PredAndModesTerm, FlagsTerm, C_FunctionTerm],
-		( parse_pragma_c_code_attributes_term(FlagsTerm, Flags) ->
+		PragmaTerms = [PredAndModesTerm, FlagsTerm, FunctionTerm],
+		( parse_pragma_foreign_code_attributes_term(ForeignLanguage,
+				FlagsTerm, Flags) ->
 			FlagsResult = ok(Flags)
 		;
-			FlagsResult = error("invalid second argument in `:- pragma import/3' declaration -- expecting C code attribute or list of attributes'",
+			FlagsResult = error("invalid second argument in `:- pragma import/3' declaration -- expecting foreign code attribute or list of attributes'",
 					FlagsTerm)
 	        )
 	    ;
-		PragmaTerms = [PredAndModesTerm, C_FunctionTerm],
-		default_attributes(Flags),
+		PragmaTerms = [PredAndModesTerm, FunctionTerm],
+		default_attributes(ForeignLanguage, Flags),
 		FlagsResult = ok(Flags)
 	    )	
  	-> 
 	    (
-		C_FunctionTerm = term__functor(term__string(C_Function), [], _)
+		FunctionTerm = term__functor(term__string(Function), [], _)
 	    ->
 		parse_pred_or_func_and_arg_modes(yes(ModuleName),
 			PredAndModesTerm, ErrorTerm,
@@ -241,7 +637,7 @@ parse_pragma_type(ModuleName, "import", PragmaTerms,
 		    (
 			FlagsResult = ok(Attributes),
 			Result = ok(pragma(import(PredName, PredOrFunc,
-				ArgModes, Attributes, C_Function)))
+				ArgModes, Attributes, Function)))
 		    ;
 			FlagsResult = error(Msg, Term),
 			Result = error(Msg, Term)
@@ -252,7 +648,7 @@ parse_pragma_type(ModuleName, "import", PragmaTerms,
 		)
 	    ;
 	    	Result = error(
-	"expected pragma import(PredName(ModeList), C_Function)",
+	"expected pragma import(PredName(ModeList), Function)",
 		     PredAndModesTerm)
 	    )
 	;
@@ -264,11 +660,12 @@ parse_pragma_type(ModuleName, "import", PragmaTerms,
 
 parse_pragma_type(_ModuleName, "export", PragmaTerms,
 		ErrorTerm, _VarSet, Result) :-
+	% XXX we implicitly assume exports are only for C
        (
-	    PragmaTerms = [PredAndModesTerm, C_FunctionTerm]
+	    PragmaTerms = [PredAndModesTerm, FunctionTerm]
        ->
 	    (
-	        C_FunctionTerm = term__functor(term__string(C_Function), [], _)
+	        FunctionTerm = term__functor(term__string(Function), [], _)
 	    ->
 		parse_pred_or_func_and_arg_modes(no, PredAndModesTerm,
 			ErrorTerm, "`:- pragma export' declaration",
@@ -276,14 +673,14 @@ parse_pragma_type(_ModuleName, "export", PragmaTerms,
 		(
 			PredAndModesResult = ok(PredName - PredOrFunc, Modes),
 		    	Result = ok(pragma(export(PredName, PredOrFunc,
-					Modes, C_Function)))
+					Modes, Function)))
 		;    
 			PredAndModesResult = error(Msg, Term),
 			Result = error(Msg, Term)
 		)
 	    ;
 	    	Result = error(
-		     "expected pragma export(PredName(ModeList), C_Function)",
+		     "expected pragma export(PredName(ModeList), Function)",
 		     PredAndModesTerm)
 	    )
 	;
@@ -688,17 +1085,18 @@ parse_pragma_keyword(ExpectedKeyword, Term, StringArg, StartContext) :-
 
 %-----------------------------------------------------------------------------%
 
-:- type collected_pragma_c_code_attribute
+:- type collected_pragma_foreign_code_attribute
 	--->	may_call_mercury(may_call_mercury)
 	;	thread_safe(thread_safe)
-	.
+	;	tabled_for_io(tabled_for_io).
 
-:- pred parse_pragma_c_code_attributes_term(term, pragma_c_code_attributes).
-:- mode parse_pragma_c_code_attributes_term(in, out) is semidet.
+:- pred parse_pragma_foreign_code_attributes_term(foreign_language, term, 
+		pragma_foreign_code_attributes).
+:- mode parse_pragma_foreign_code_attributes_term(in, in, out) is semidet.
 
-parse_pragma_c_code_attributes_term(Term, Attributes) :-
-	default_attributes(Attributes0),
-	parse_pragma_c_code_attributes_term0(Term, AttrList),
+parse_pragma_foreign_code_attributes_term(ForeignLanguage, Term, Attributes) :-
+	default_attributes(ForeignLanguage, Attributes0),
+	parse_pragma_foreign_code_attributes_term0(Term, AttrList),
 	( list__member(may_call_mercury(will_not_call_mercury), AttrList) ->
 		( list__member(may_call_mercury(may_call_mercury), AttrList) ->
 			% XXX an error message would be nice
@@ -715,19 +1113,30 @@ parse_pragma_c_code_attributes_term(Term, Attributes) :-
 			% XXX an error message would be nice
 			fail
 		;
-			set_thread_safe(Attributes1, thread_safe, Attributes)
+			set_thread_safe(Attributes1, thread_safe, Attributes2)
 		)
 	;
-		Attributes = Attributes1
+		Attributes2 = Attributes1
+	),
+	( list__member(tabled_for_io(tabled_for_io), AttrList) ->
+		( list__member(tabled_for_io(not_tabled_for_io), AttrList) ->
+			% XXX an error message would be nice
+			fail
+		;
+			set_tabled_for_io(Attributes2, tabled_for_io,
+				Attributes)
+		)
+	;
+		Attributes = Attributes2
 	).
 
-:- pred parse_pragma_c_code_attributes_term0(term,
-		list(collected_pragma_c_code_attribute)).
-:- mode parse_pragma_c_code_attributes_term0(in, out) is semidet.
+:- pred parse_pragma_foreign_code_attributes_term0(term,
+		list(collected_pragma_foreign_code_attribute)).
+:- mode parse_pragma_foreign_code_attributes_term0(in, out) is semidet.
 
-parse_pragma_c_code_attributes_term0(Term, Flags) :-
+parse_pragma_foreign_code_attributes_term0(Term, Flags) :-
 	(
-		parse_single_pragma_c_code_attribute(Term, Flag)
+		parse_single_pragma_foreign_code_attribute(Term, Flag)
 	->
 		Flags = [Flag]
 	;
@@ -737,20 +1146,22 @@ parse_pragma_c_code_attributes_term0(Term, Flags) :-
 		;
 			Term = term__functor(term__atom("."), [Hd, Tl], _),
 			Flags = [Flag|Flags0],
-			parse_single_pragma_c_code_attribute(Hd, Flag),
-			parse_pragma_c_code_attributes_term0(Tl, Flags0)
+			parse_single_pragma_foreign_code_attribute(Hd, Flag),
+			parse_pragma_foreign_code_attributes_term0(Tl, Flags0)
 		)
 	).
 
-:- pred parse_single_pragma_c_code_attribute(term,
-		collected_pragma_c_code_attribute).
-:- mode parse_single_pragma_c_code_attribute(in, out) is semidet.
+:- pred parse_single_pragma_foreign_code_attribute(term,
+		collected_pragma_foreign_code_attribute).
+:- mode parse_single_pragma_foreign_code_attribute(in, out) is semidet.
 
-parse_single_pragma_c_code_attribute(Term, Flag) :-
+parse_single_pragma_foreign_code_attribute(Term, Flag) :-
 	( parse_may_call_mercury(Term, MayCallMercury) ->
 		Flag = may_call_mercury(MayCallMercury)
 	; parse_threadsafe(Term, ThreadSafe) ->
 		Flag = thread_safe(ThreadSafe)
+	; parse_tabled_for_io(Term, TabledForIo) ->
+		Flag = tabled_for_io(TabledForIo)
 	;
 		fail
 	).
@@ -775,14 +1186,22 @@ parse_threadsafe(term__functor(term__atom("thread_safe"), [], _),
 parse_threadsafe(term__functor(term__atom("not_thread_safe"), [], _),
 	not_thread_safe).
 
-% parse a pragma c_code declaration
+:- pred parse_tabled_for_io(term, tabled_for_io).
+:- mode parse_tabled_for_io(in, out) is semidet.
 
-:- pred parse_pragma_c_code(module_name, pragma_c_code_attributes, term,
-	pragma_c_code_impl, varset, maybe1(item)).
-:- mode parse_pragma_c_code(in, in, in, in, in, out) is det.
+parse_tabled_for_io(term__functor(term__atom("tabled_for_io"), [], _),
+	tabled_for_io).
+parse_tabled_for_io(term__functor(term__atom("not_tabled_for_io"), [], _),
+	not_tabled_for_io).
 
-parse_pragma_c_code(ModuleName, Flags, PredAndVarsTerm0, PragmaImpl,
-	VarSet0, Result) :-
+% parse a pragma foreign_code declaration
+
+:- pred parse_pragma_foreign_code(module_name, pragma_foreign_code_attributes,
+	term, pragma_foreign_code_impl, varset, maybe1(item)).
+:- mode parse_pragma_foreign_code(in, in, in, in, in, out) is det.
+
+parse_pragma_foreign_code(ModuleName, Flags, PredAndVarsTerm0,
+	PragmaImpl, VarSet0, Result) :-
     parse_pred_or_func_and_args(yes(ModuleName), PredAndVarsTerm0,
     	PredAndVarsTerm0, "`:- pragma c_code' declaration", PredAndArgsResult),
     (
@@ -803,7 +1222,7 @@ parse_pragma_c_code(ModuleName, Flags, PredAndVarsTerm0, PragmaImpl,
 	(
 	    Error = no,
 	    varset__coerce(VarSet0, VarSet),
-	    Result = ok(pragma(c_code(Flags, PredName,
+	    Result = ok(pragma(foreign(Flags, PredName,
 		    PredOrFunc, PragmaVars, VarSet, PragmaImpl)))
 	;
 	    Error = yes(ErrorMessage),
