@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1997-1999 University of Melbourne.
+% Copyright (C) 1997-2000 University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -110,8 +110,24 @@ parse_constrained_class(ModuleName, Decl, Constraints, VarSet, Result) :-
 			Result0 = ok(typeclass(_, Name, Vars, Interface, 
 				VarSet0))
 		->
-			Result = ok(typeclass(ConstraintList, Name, Vars,
-				Interface, VarSet0))
+			(
+				%
+				% check for type variables in the constraints
+				% which do not occur in the type class
+				% parameters 
+				% 
+				type_util__constraint_list_get_tvars(
+					ConstraintList, ConstrainedVars),
+				list__member(Var, ConstrainedVars),
+				\+ list__member(Var, Vars)
+			->
+				Result = error(
+"type variable in superclass constraint is not a parameter of this type class",
+					Constraints)
+			;
+				Result = ok(typeclass(ConstraintList, Name,
+					Vars, Interface, VarSet0))
+			)
 		;
 				% if the item we get back isn't a typeclass,
 				% something has gone wrong...
@@ -198,15 +214,13 @@ list_term_to_term_list(Methods, MethodList) :-
 item_to_class_method(error(String, Term), _, error(String, Term)).
 item_to_class_method(ok(Item, Context), Term, Result) :-
 	(
-			% XXX Purity is ignored
-		Item = pred(A, B, C, D, E, F, G, _, I)
+		Item = pred(A, B, C, D, E, F, G, H, I)
 	->
-		Result = ok(pred(A, B, C, D, E, F, G, I, Context))
+		Result = ok(pred(A, B, C, D, E, F, G, H, I, Context))
 	;
-			% XXX Purity is ignored
-		Item = func(A, B, C, D, E, F, G, H, _, J)
+		Item = func(A, B, C, D, E, F, G, H, I, J)
 	->
-		Result = ok(func(A, B, C, D, E, F, G, H, J, Context))
+		Result = ok(func(A, B, C, D, E, F, G, H, I, J, Context))
 	;
 		Item = pred_mode(A, B, C, D, E)
 	->
@@ -341,7 +355,7 @@ parse_instance(ModuleName, VarSet, TypeClassTerm, Result) :-
 	(
 		Arg = term__functor(term__atom("where"), [Name, Methods], _)
 	->
-		parse_non_empty_instance(ModuleName, Name, Methods,
+		parse_non_empty_instance(ModuleName, Name, Methods, VarSet,
 			TVarSet, Result)
 	;
 		parse_instance_name(ModuleName, Arg, TVarSet, Result)
@@ -375,10 +389,11 @@ parse_derived_instance(ModuleName, Decl, Constraints, TVarSet,
 		->
 			Result = Result0
 		;
-			Result0 = ok(instance(_, Name, Types, Body, VarSet0))
+			Result0 = ok(instance(_, Name, Types, Body, VarSet,
+				ModName))
 		->
 			Result = ok(instance(ConstraintList, Name, Types, Body,
-					VarSet0))
+					VarSet, ModName))
 		;
 				% if the item we get back isn't an instance, 
 				% something has gone wrong...
@@ -403,7 +418,7 @@ parse_instance_constraints(ModuleName, Constraints, Result) :-
 :- pred parse_underived_instance(module_name, term, tvarset, maybe1(item)).
 :- mode parse_underived_instance(in, in, in, out) is det.
 
-parse_underived_instance(_ModuleName, Name, TVarSet, Result) :-
+parse_underived_instance(ModuleName, Name, TVarSet, Result) :-
 		% We don't give a default module name here since the instance
 		% declaration could well be for a typeclass defined in another
 		% module
@@ -445,7 +460,7 @@ parse_underived_instance(_ModuleName, Name, TVarSet, Result) :-
 		(
 			ErroneousTypes = [],
 			Result = ok(instance([], ClassName,
-				TermTypes, abstract, TVarSet))
+				TermTypes, abstract, TVarSet, ModuleName))
 		;
 				% XXX We should report an error for _each_
 				% XXX erroneous type
@@ -458,12 +473,12 @@ parse_underived_instance(_ModuleName, Name, TVarSet, Result) :-
 		Result = error(String, Term)
 	).
 
-:- pred parse_non_empty_instance(module_name, term, term, tvarset,
+:- pred parse_non_empty_instance(module_name, term, term, varset, tvarset,
 		maybe1(item)).
-:- mode parse_non_empty_instance(in, in, in, in, out) is det.
+:- mode parse_non_empty_instance(in, in, in, in, in, out) is det.
 
-parse_non_empty_instance(ModuleName, Name, Methods, TVarSet, Result) :-
-	parse_instance_methods(ModuleName, Methods, ParsedMethods),
+parse_non_empty_instance(ModuleName, Name, Methods, VarSet, TVarSet, Result) :-
+	parse_instance_methods(ModuleName, Methods, VarSet, ParsedMethods),
 	(
 		ParsedMethods = ok(MethodList),
 		parse_instance_name(ModuleName, Name, TVarSet,
@@ -474,10 +489,12 @@ parse_non_empty_instance(ModuleName, Name, Methods, TVarSet, Result) :-
 			Result = error(String, Term)
 		;
 			ParsedNameAndTypes = ok(instance(Constraints,
-				NameString, Types, _, _))
+				NameString, Types, _, _, ModName))
 		->
-			Result = ok(instance(Constraints, NameString, Types,
-				concrete(MethodList), TVarSet))
+			Result0 = ok(instance(Constraints, NameString, Types,
+				concrete(MethodList), TVarSet, ModName)),
+			check_tvars_in_instance_constraint(Result0, Name,
+				Result)
 		;
 				% if the item we get back isn't a typeclass,
 				% something has gone wrong...
@@ -488,17 +505,47 @@ parse_non_empty_instance(ModuleName, Name, Methods, TVarSet, Result) :-
 		Result = error(String, Term)
 	).
 
-:- pred parse_instance_methods(module_name, term,
-		maybe1(list(instance_method))).
-:- mode parse_instance_methods(in, in, out) is det.
+:- pred check_tvars_in_instance_constraint(maybe1(item), term, maybe1(item)).
+:- mode check_tvars_in_instance_constraint(in, in, out) is det.
 
-parse_instance_methods(ModuleName, Methods, Result) :-
+check_tvars_in_instance_constraint(error(M,E), _, error(M, E)).
+check_tvars_in_instance_constraint(ok(Item), InstanceTerm, Result) :-
+	(
+		Item = instance(Constraints, _Name, Types, _Methods, _TVarSet,
+			_ModName)
+	->
+		%
+		% check that all of the type variables in the constraints
+		% on the instance declaration also occur in the type class
+		% argument types in the instance declaration
+		%
+		( 
+			type_util__constraint_list_get_tvars(Constraints,
+				TVars),
+			list__member(TVar, TVars),
+			\+ term__contains_var_list(Types, TVar)
+		->
+			Result = error(
+	"unbound type variable(s) in constraints on instance declaration",
+				InstanceTerm)
+		;
+			Result = ok(Item)
+		)
+	;
+		error("check_tvars_in_constraint: expecting instance item")
+	).
+
+:- pred parse_instance_methods(module_name, term, varset,
+		maybe1(list(instance_method))).
+:- mode parse_instance_methods(in, in, in, out) is det.
+
+parse_instance_methods(ModuleName, Methods, VarSet, Result) :-
 	(
 		list_term_to_term_list(Methods, MethodList)
 	->
 			% Convert the list of terms into a list of 
 			% maybe1(class_method)s.
-		list__map(term_to_instance_method(ModuleName),
+		list__map(term_to_instance_method(ModuleName, VarSet),
 			MethodList, Interface),
 		find_errors(Interface, Result)
 	;
@@ -506,10 +553,11 @@ parse_instance_methods(ModuleName, Methods, Result) :-
 	).
 
 	% Turn the term into a method instance
-:- pred term_to_instance_method(module_name, term, maybe1(instance_method)).
-:- mode term_to_instance_method(in, in, out) is det.
+:- pred term_to_instance_method(module_name, varset, term,
+		maybe1(instance_method)).
+:- mode term_to_instance_method(in, in, in, out) is det.
 
-term_to_instance_method(_ModuleName, MethodTerm, Result) :-
+term_to_instance_method(_ModuleName, VarSet, MethodTerm, Result) :-
 	(
 		MethodTerm = term__functor(term__atom("is"), [ClassMethodTerm,
 						InstanceMethod], TermContext)
@@ -532,9 +580,10 @@ term_to_instance_method(_ModuleName, MethodTerm, Result) :-
 					InstanceMethod, "instance method",
 					ok(InstanceMethodName, []))
 			->
-				Result = ok(pred_instance(ClassMethodName,
-					InstanceMethodName, ArityInt,
-					TermContext))
+				Result = ok(instance_method(predicate,
+					ClassMethodName,
+					name(InstanceMethodName),
+					ArityInt, TermContext))
 			;
 				Result = error(
 				    "expected `pred(<Name> / <Arity>) is <InstanceMethod>'",
@@ -558,9 +607,10 @@ term_to_instance_method(_ModuleName, MethodTerm, Result) :-
 					InstanceMethod, "instance method",
 					ok(InstanceMethodName, []))
 			->
-				Result = ok(func_instance(ClassMethodName,
-					InstanceMethodName, ArityInt,
-					TermContext))
+				Result = ok(instance_method(function,
+					ClassMethodName,
+					name(InstanceMethodName),
+					ArityInt, TermContext))
 			;
 				Result = error(
 				    "expected `func(<Name> / <Arity>) is <InstanceMethod>'",
@@ -572,7 +622,46 @@ term_to_instance_method(_ModuleName, MethodTerm, Result) :-
 				MethodTerm)
 		)
 	;
-		Result = error("expected `pred(<Name> / <Arity>) is <InstanceName>'",
-			MethodTerm)
+		% For the clauses in an instance declaration,
+		% the default module name for the clause heads
+		% is the module name of the class that this is an
+		% instance declaration for, but we don't necessarily
+		% know what module that is at this point, since the
+		% class name hasn't been fully qualified yet.
+		% So here we give the special module name ""
+		% as the default, which means that there is no default.
+		% (If the module qualifiers in the clauses don't match
+		% the module name of the class, we will pick that up later,
+		% in check_typeclass.m.)
+		DefaultModuleName = unqualified(""),
+		parse_item(DefaultModuleName, VarSet, MethodTerm, Result0),
+		(
+			Result0 = ok(Item, Context),
+			(
+				Item = pred_clause(_VarNames, ClassMethodName,
+					HeadArgs, _ClauseBody),
+				PredOrFunc = predicate,
+				ArityInt = list__length(HeadArgs)
+			;
+				Item = func_clause(_VarNames, ClassMethodName,
+					FuncArgs, _Result, _ClauseBody),
+				ArityInt = list__length(FuncArgs),
+				PredOrFunc = function
+			)
+		->
+			Result = ok(instance_method(PredOrFunc,
+					ClassMethodName,
+					clauses([Item]),
+					ArityInt, Context))
+		;
+			Result0 = error(ErrorMsg, ErrorTerm)
+		->
+			Result = error(ErrorMsg, ErrorTerm)
+		;
+			% catch-all error message for a syntactically valid item
+			% which is not a clause
+			Result = error("expected clause or `pred(<Name> / <Arity>) is <InstanceName>' or `func(<Name> / <Arity>) is <InstanceName>')",
+				MethodTerm)
+		)
 	).
 
