@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -106,6 +106,12 @@
 :- pred conjunction_to_list(term(T), list(term(T))).
 :- mode conjunction_to_list(in, out) is det.
 
+	% list_to_conjunction(Context, First, Rest, Term).
+	% convert a list to a "conjunction" (bunch of terms separated by ','s)
+
+:- pred list_to_conjunction(prog_context, term(T), list(term(T)), term(T)).
+:- mode list_to_conjunction(in, in, in, out) is det.
+
 	% convert a "sum" (bunch of terms separated by '+' operators) to a list
 
 :- pred sum_to_list(term(T), list(term(T))).
@@ -221,9 +227,26 @@ convert_mode_list([H0|T0], [H|T]) :-
 	convert_mode(H0, H),
 	convert_mode_list(T0, T).
 
+
+	% 
+	% The new operator for mode declarations is >>.
+	% Previously we used ->, but this required a high-precedence
+	% operator such as :: for the :- mode delcaration.
+	%
+	% Using >> allows us to use == for the :- mode declaration.
+	%
+	% Eventually we can stop supporting :: and -> in :- mode
+	% declarations altogether.
+	%
 convert_mode(Term, Mode) :-
 	(
-		Term = term__functor(term__atom("->"), [InstA, InstB], _Context)
+		( 
+			Term = term__functor(term__atom(">>"), 
+				[InstA, InstB], _)
+		;
+			Term = term__functor(term__atom("->"),
+				[InstA, InstB], _)
+		)
 	->
 		convert_inst(InstA, ConvertedInstA),
 		convert_inst(InstB, ConvertedInstB),
@@ -244,7 +267,7 @@ convert_mode(Term, Mode) :-
 		standard_det(DetString, Detism),
 		convert_mode_list(ArgModesTerms, ArgModes),
 		PredInstInfo = pred_inst_info(predicate, ArgModes, Detism),
-		Inst = ground(shared, yes(PredInstInfo)),
+		Inst = ground(shared, higher_order(PredInstInfo)),
 		Mode = (Inst -> Inst)
 	;
 		% Handle higher-order function modes:
@@ -266,7 +289,7 @@ convert_mode(Term, Mode) :-
 		convert_mode(RetModeTerm, RetMode),
 		list__append(ArgModes0, [RetMode], ArgModes),
 		FuncInstInfo = pred_inst_info(function, ArgModes, Detism),
-		Inst = ground(shared, yes(FuncInstInfo)),
+		Inst = ground(shared, higher_order(FuncInstInfo)),
 		Mode = (Inst -> Inst)
 	;
 		parse_qualified_term(Term, Term, "mode definition", R),
@@ -302,15 +325,15 @@ convert_inst(Term, Result) :-
 
 	% `ground' insts
 	; Name = term__atom("ground"), Args0 = [] ->
-		Result = ground(shared, no)
+		Result = ground(shared, none)
 	; Name = term__atom("unique"), Args0 = [] ->
-		Result = ground(unique, no)
+		Result = ground(unique, none)
 	; Name = term__atom("mostly_unique"), Args0 = [] ->
-		Result = ground(mostly_unique, no)
+		Result = ground(mostly_unique, none)
 	; Name = term__atom("clobbered"), Args0 = [] ->
-		Result = ground(clobbered, no)
+		Result = ground(clobbered, none)
 	; Name = term__atom("mostly_clobbered"), Args0 = [] ->
-		Result = ground(mostly_clobbered, no)
+		Result = ground(mostly_clobbered, none)
 	;
 		% The syntax for a higher-order pred inst is
 		%
@@ -326,7 +349,7 @@ convert_inst(Term, Result) :-
 		standard_det(DetString, Detism),
 		convert_mode_list(ArgModesTerm, ArgModes),
 		PredInst = pred_inst_info(predicate, ArgModes, Detism),
-		Result = ground(shared, yes(PredInst))
+		Result = ground(shared, higher_order(PredInst))
 	;
 
 		% The syntax for a higher-order func inst is
@@ -347,7 +370,7 @@ convert_inst(Term, Result) :-
 		convert_mode(RetModeTerm, RetMode),
 		list__append(ArgModes0, [RetMode], ArgModes),
 		FuncInst = pred_inst_info(function, ArgModes, Detism),
-		Result = ground(shared, yes(FuncInst))
+		Result = ground(shared, higher_order(FuncInst))
 
 	% `not_reached' inst
 	; Name = term__atom("not_reached"), Args0 = [] ->
@@ -387,7 +410,14 @@ standard_det("failure", failure).
 parse_bound_inst_list(Disj, Uniqueness, bound(Uniqueness, Functors)) :-
 	disjunction_to_list(Disj, List),
 	convert_bound_inst_list(List, Functors0),
-	list__sort_and_remove_dups(Functors0, Functors).
+	list__sort(Functors0, Functors),
+	% check that the list doesn't specify the same functor twice
+	\+ (
+		list__append(_, SubList, Functors),
+		SubList = [F1, F2 | _],
+		F1 = functor(ConsId, _),
+		F2 = functor(ConsId, _)
+	).
 
 :- pred convert_bound_inst_list(list(term), list(bound_inst)).
 :- mode convert_bound_inst_list(in, out) is semidet.
@@ -420,6 +450,11 @@ disjunction_to_list(Term, List) :-
 conjunction_to_list(Term, List) :-
 	binop_term_to_list(",", Term, List).
 
+list_to_conjunction(_, Term, [], Term).
+list_to_conjunction(Context, First, [Second | Rest], Term) :-
+	list_to_conjunction(Context, Second, Rest, Tail),
+	Term = term__functor(term__atom(","), [First, Tail], Context).
+
 sum_to_list(Term, List) :-
 	binop_term_to_list("+", Term, List).
 
