@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1998-1999 University of Melbourne.
+% Copyright (C) 1998-2000 University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -41,6 +41,11 @@
 		list(prog_var)::in, list(prog_var)::in, index_spec::in,
 		list(rl_var_bounds)::in, list(key_range)::out) is semidet.
 
+	% Succeed if a join is an equi-join, returning the list
+	% of arguments in each tuple which must be equivalent.
+:- pred rl_key__is_equijoin(rl_goal_inputs::in, list(rl_var_bounds)::in,
+		list(int)::out, list(int)::out) is semidet.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -109,6 +114,10 @@ rl_key__remove_useless_info(ModuleInfo,
 	list__map(rl_key__remove_useless_info(ModuleInfo),
 		ArgBounds0, ArgBounds),
 	( 
+		type_is_tuple(Type, _)
+	->
+		Bound = var - Vars
+	;
 		\+ (
 			list__member(ArgBound, ArgBounds),
 			ArgBound \= var - _
@@ -119,7 +128,7 @@ rl_key__remove_useless_info(ModuleInfo,
 		),
 		module_info_types(ModuleInfo, Types),
 		type_to_type_id(Type, TypeId, _),
-		map__lookup(Types, TypeId, TypeDefn),
+		map__search(Types, TypeId, TypeDefn),
 		hlds_data__get_type_defn_body(TypeDefn, Body),
 		Body = du_type(Ctors, _, _, _),
 		Ctors = [_]
@@ -182,47 +191,57 @@ rl_key__bounds_to_key_range(MaybeConstructArgs, Args, VarTypes,
 	assoc_list(int, key_term)::in, bounding_tuple::out) is det.
 
 rl_key__convert_bound(MaybeArgs, Tuple, Bound) :-
+	assoc_list__keys(Tuple, Indexes),
+	assoc_list__values(Tuple, Terms),
+	SeenInfinity0 = no,
+	list__map_foldl(rl_key__convert_key_attr(MaybeArgs),
+		Terms, Attrs, SeenInfinity0, SeenInfinity),
+
+	% If there is an attribute of the key tuple which is
+	% not constrained either by a test against the arguments
+	% of the other input tuple, or by a test against
+	% a constructor, treat the entire bound as infinity.
+	% XXX this is temporary, until Aditi supports infinities
+	% within bounds tuples.
 	(
-		% XXX this is temporary, until Aditi supports infinities
-		% within bounds tuples.
-		(
-			list__member(Attr, Tuple),
-			(
-				MaybeArgs = no,
-				Attr = _ - (var - _)
-			;
-				MaybeArgs = yes(Args),
-				Attr = _ - (var - Vars),
-				\+ (
-					set__member(Var, Vars),
-					list__member(Var, Args)
-				)
-			)
-		)
+		SeenInfinity = yes
 	->
 		Bound = infinity
 	;
-		list__member(Attr, Tuple),
-		\+ (
-			MaybeArgs = yes(Args),
-			Attr = _ - (var - Vars),
-			set__member(Var, Vars),
-			list__member(Var, Args)
-		)
+		% If all attributes are infinity, the entire bound is just
+		% infinity.
+		Attrs = [KeyAttr | _],
+		KeyAttr = infinity
 	->
-		assoc_list__keys(Tuple, Indexes),
-		assoc_list__values(Tuple, Terms),
-		list__map(rl_key__convert_key_attr(MaybeArgs), Terms, Attrs),
+		Bound = infinity
+	;
 		assoc_list__from_corresponding_lists(Indexes, Attrs, KeyTuple),
 		Bound = bound(KeyTuple)
-	;
-		Bound = infinity
 	).
 
 :- pred rl_key__convert_key_attr(maybe(list(prog_var))::in,
-		key_term::in, key_attr::out) is det.
+		key_term::in, key_attr::out, bool::in, bool::out) is det.
 
-rl_key__convert_key_attr(MaybeArgs, var - Vars, Attr) :-
+rl_key__convert_key_attr(MaybeArgs, KeyTerm, KeyAttr,
+		SeenInfinity0, SeenInfinity) :-
+	(
+		% As soon as we've seen one infinity in the key tuple,
+		% the remaining attributes can't affect the result of
+		% the comparison in a B-tree search.
+		SeenInfinity0 = yes,
+		KeyAttr = infinity,
+		SeenInfinity = yes
+	;
+		SeenInfinity0 = no,
+		rl_key__convert_key_attr_2(MaybeArgs, KeyTerm,
+			KeyAttr, SeenInfinity)
+	).
+
+:- pred rl_key__convert_key_attr_2(maybe(list(prog_var))::in,
+		key_term::in, key_attr::out, bool::out) is det.
+
+rl_key__convert_key_attr_2(MaybeArgs, var - Vars, Attr,
+		SeenInfinity) :-
 	(
 		MaybeArgs = yes(Args),
 		set__list_to_set(Args, ArgSet),
@@ -230,13 +249,17 @@ rl_key__convert_key_attr(MaybeArgs, var - Vars, Attr) :-
 		set__to_sorted_list(Intersection, [Arg | _]),
 		list__nth_member_search(Args, Arg, Index)
 	->
-		Attr = input_field(Index)
+		Attr = input_field(Index),
+		SeenInfinity = no
 	;
-		Attr = infinity
+		Attr = infinity,
+		SeenInfinity = yes
 	).
-rl_key__convert_key_attr(MaybeArgs, functor(ConsId, Type, Terms) - _,
-		functor(ConsId, Type, Attrs)) :-
-	list__map(rl_key__convert_key_attr(MaybeArgs), Terms, Attrs).
+rl_key__convert_key_attr_2(MaybeArgs, functor(ConsId, Type, Terms) - _,
+		functor(ConsId, Type, Attrs), SeenInfinity) :-
+	SeenInfinity0 = no,
+	list__map_foldl(rl_key__convert_key_attr(MaybeArgs),
+		Terms, Attrs, SeenInfinity0, SeenInfinity).
 		
 :- pred rl_key__split_key_tuples(assoc_list(int, pair(key_term))::in,
 	assoc_list(int, key_term)::out, assoc_list(int, key_term)::out) is det.
@@ -364,6 +387,93 @@ key_term_less_or_equal(ModuleInfo, UpperLower1, functor(ConsId1, _, Args1),
 
 %-----------------------------------------------------------------------------%
 
+rl_key__is_equijoin(two_inputs(Args1, Args2), [VarBound0 | VarBounds],
+		Attrs1, Attrs2) :-
+	%
+	% For each attribute of the first tuple, work out which
+	% attributes of the second must be equal for the join
+	% condition to succeed.
+	% XXX we don't yet handle cases such as p(X, Y), p(X + 10, Z).
+	%
+	rl_key__restrict_bounds_to_arg_vars(Args1, Args2,
+			VarBound0, EqArgs0),
+	list__foldl(rl_key__intersect_branch_eq_args, VarBounds,
+		EqArgs0, EqArgs),
+	EqArgs \= [],
+
+	%
+	% For each attribute of the first tuple, choose one attribute
+	% of the second which must be equal.
+	%
+	list__map(rl_key__var_and_eq_args_to_attr_pair(Args1, Args2),
+		EqArgs, AttrPairs),
+	assoc_list__keys(AttrPairs, Attrs1),
+	assoc_list__values(AttrPairs, Attrs2).
+
+:- pred rl_key__var_and_eq_args_to_attr_pair(list(prog_var)::in,
+		list(prog_var)::in, pair(prog_var, set(prog_var))::in,
+		pair(int)::out) is det.
+
+rl_key__var_and_eq_args_to_attr_pair(Args1, Args2, Arg1 - EqArgs2,
+		AttrPair) :-
+	(
+		list__nth_member_search(Args1, Arg1, Attr1),
+		set__to_sorted_list(EqArgs2, EqArgsList2),
+		EqArgsList2 = [Arg2 | _],
+		list__nth_member_search(Args2, Arg2, Attr2)
+	->
+		AttrPair = Attr1 - Attr2
+	;	
+		error("rl_key__var_and_eq_args_to_attr_pair")
+	).	
+
+:- pred rl_key__intersect_branch_eq_args(rl_var_bounds::in,
+		assoc_list(prog_var, set(prog_var))::in,
+		assoc_list(prog_var, set(prog_var))::out) is semidet.
+
+rl_key__intersect_branch_eq_args(Bounds, EqArgs0, EqArgs) :-
+	list__filter_map(rl_key__intersect_eq_args(Bounds), EqArgs0, EqArgs),
+	EqArgs \= [].
+
+:- pred rl_key__intersect_eq_args(rl_var_bounds::in,
+		pair(prog_var, set(prog_var))::in,
+		pair(prog_var, set(prog_var))::out) is semidet.
+
+rl_key__intersect_eq_args(Bounds, Var - EqVars0, Var - EqVars) :-
+	map__search(Bounds, Var, VarBounds),
+	rl_key__extract_bounds_eq_vars(VarBounds, BoundsEqArgs),
+	set__intersect(EqVars0, BoundsEqArgs, EqVars),
+	\+ set__empty(EqVars).
+
+:- pred rl_key__restrict_bounds_to_arg_vars(list(prog_var)::in,
+		list(prog_var)::in, rl_var_bounds::in,
+		assoc_list(prog_var, set(prog_var))::out) is det.
+
+rl_key__restrict_bounds_to_arg_vars(Args, ArgsOfOtherTuple,
+		Bounds0, ArgsAndEqOtherArgs) :-
+	set__list_to_set(ArgsOfOtherTuple, ArgsOfOtherTupleSet),
+	list__filter_map(
+		(pred(Arg::in, ArgAndEqOtherArgs::out) is semidet :-
+			map__search(Bounds0, Arg, ArgBounds),
+
+			rl_key__extract_bounds_eq_vars(ArgBounds,
+				BoundsEqArgs),
+			set__intersect(BoundsEqArgs, ArgsOfOtherTupleSet,
+				EqOtherArgs),
+			\+ set__empty(EqOtherArgs),
+			ArgAndEqOtherArgs = Arg - EqOtherArgs
+		), Args, ArgsAndEqOtherArgs).
+
+:- pred rl_key__extract_bounds_eq_vars(pair(key_term)::in,
+		set(prog_var)::out) is det.
+
+rl_key__extract_bounds_eq_vars(LBound - UBound, BoundsEqArgs) :-
+	LBound = _ - LessThanOrEqVars,
+	UBound = _ - GreaterThanOrEqVars,
+	set__intersect(LessThanOrEqVars, GreaterThanOrEqVars, BoundsEqArgs).
+
+%-----------------------------------------------------------------------------%
+
 :- pred rl_key__get_var_bounds(module_info::in, var_map::in,
 		prog_var::in, pair(prog_var, pair(key_term))::out) is det.
 
@@ -683,7 +793,7 @@ rl_key__extract_key_range_unify(assign(Var1, Var2)) -->
 rl_key__extract_key_range_unify(construct(Var, ConsId, Args, _, _, _, _)) -->
 	rl_key__unify_functor(Var, ConsId, Args).
 rl_key__extract_key_range_unify(
-		deconstruct(Var, ConsId, Args, _, _)) -->
+		deconstruct(Var, ConsId, Args, _, _, _)) -->
 	rl_key__unify_functor(Var, ConsId, Args).
 rl_key__extract_key_range_unify(complicated_unify(_, _, _)) -->
 	{ error("rl_key__extract_key_range_unify") }.
@@ -913,7 +1023,7 @@ rl_key__choose_cons_id(ModuleInfo, UpperLower, Type,
 	->	
 		module_info_types(ModuleInfo, Types),
 		type_to_type_id(Type, TypeId, _),
-		map__lookup(Types, TypeId, TypeDefn),
+		map__search(Types, TypeId, TypeDefn),
 		hlds_data__get_type_defn_body(TypeDefn, Body),
 		% If there's a user defined equality pred we're in trouble.
 		Body = du_type(Ctors, _, _, no),
