@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -28,13 +28,12 @@
 :- import_module hlds_module, hlds_pred.
 :- import_module io.
 
-:- pred saved_vars_proc(pred_id, proc_id, proc_info, proc_info,
-		module_info, module_info, io__state, io__state).
-:- mode saved_vars_proc(in, in, in, out, in, out, di, uo) is det.
+:- pred saved_vars_proc(pred_id::in, proc_id::in,
+	proc_info::in, proc_info::out, module_info::in, module_info::out,
+	io__state::di, io__state::uo) is det.
 
-:- pred saved_vars_proc_no_io(pred_id, proc_id, proc_info, proc_info,
-		module_info, module_info).
-:- mode saved_vars_proc_no_io(in, in, in, out, in, out) is det.
+:- pred saved_vars_proc_no_io(proc_info::in, proc_info::out,
+	module_info::in, module_info::out) is det.
 
 %-----------------------------------------------------------------------------%
 
@@ -50,11 +49,10 @@ saved_vars_proc(PredId, ProcId, ProcInfo0, ProcInfo,
 		ModuleInfo0, ModuleInfo) -->
 	write_proc_progress_message("% Minimizing saved vars in ",
 		PredId, ProcId, ModuleInfo0),
-	{ saved_vars_proc_no_io(PredId, ProcId, ProcInfo0, ProcInfo,
+	{ saved_vars_proc_no_io(ProcInfo0, ProcInfo,
 		ModuleInfo0, ModuleInfo) }.
 
-saved_vars_proc_no_io(_PredId, _ProcId, ProcInfo0, ProcInfo,
-		ModuleInfo0, ModuleInfo) :-
+saved_vars_proc_no_io(ProcInfo0, ProcInfo, ModuleInfo0, ModuleInfo) :-
 	proc_info_goal(ProcInfo0, Goal0),
 	proc_info_varset(ProcInfo0, Varset0),
 	proc_info_vartypes(ProcInfo0, VarTypes0),
@@ -71,7 +69,8 @@ saved_vars_proc_no_io(_PredId, _ProcId, ProcInfo0, ProcInfo,
 	implicitly_quantify_clause_body(HeadVars, Goal1, Varset1,
 		VarTypes1, Goal2, Varset, VarTypes, _Warnings),
 	proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0),
-	recompute_instmap_delta(no, Goal2, Goal, VarTypes, InstMap0, 
+	proc_info_inst_varset(ProcInfo0, InstVarSet),
+	recompute_instmap_delta(no, Goal2, Goal, VarTypes, InstVarSet, InstMap0,
 		ModuleInfo0, ModuleInfo),
 
 	% hlds_out__write_goal(Goal, ModuleInfo, Varset, 0, "\n"),
@@ -134,15 +133,14 @@ saved_vars_in_goal(GoalExpr0 - GoalInfo0, SlotInfo0, Goal, SlotInfo) :-
 		Goal = GoalExpr0 - GoalInfo0,
 		SlotInfo = SlotInfo0
 	;
-		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _),
+		GoalExpr0 = pragma_foreign_code(_, _, _, _, _, _, _),
 		Goal = GoalExpr0 - GoalInfo0,
 		SlotInfo = SlotInfo0
 	;
 		GoalExpr0 = bi_implication(_, _),
 		% these should have been expanded out by now
 		error("saved_vars_in_goal: unexpected bi_implication")
-	),
-	!.
+	).
 
 %-----------------------------------------------------------------------------%
 
@@ -294,7 +292,7 @@ saved_vars_delay_goal([Goal0 | Goals0], Construct, Var, IsNonLocal, SlotInfo0,
 				IsNonLocal, SlotInfo1, Goals1, SlotInfo),
 			Goals = [NewConstruct, Goal1 | Goals1]
 		;
-			Goal0Expr = pragma_c_code(_, _, _, _, _, _, _),
+			Goal0Expr = pragma_foreign_code(_, _, _, _, _, _, _),
 			rename_var(SlotInfo0, Var, _NewVar, Subst, SlotInfo1),
 			goal_util__rename_vars_in_goal(Construct, Subst,
 				NewConstruct),
