@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -91,7 +91,6 @@ simplify__pred(Simplifications0, PredId, ModuleInfo0, ModuleInfo,
 	{
 		% Don't warn for compiler-generated procedures.
 		list__member(warn_simple_code, Simplifications0),
-		module_info_pred_info(ModuleInfo0, PredId, PredInfo0),
 		code_util__compiler_generated(PredInfo0)
 	->
 		list__delete_all(Simplifications0, warn_simple_code,
@@ -151,21 +150,29 @@ simplify__proc(Simplifications, PredId, ProcId, ModuleInfo0, ModuleInfo,
 simplify__proc_2(Simplifications, PredId, ProcId, ModuleInfo0, ModuleInfo,
 		ProcInfo0, ProcInfo, Msgs) :-
 	module_info_globals(ModuleInfo0, Globals),
-	det_info_init(ModuleInfo0, PredId, ProcId, Globals, DetInfo0),
+	proc_info_vartypes(ProcInfo0, VarTypes0),
+	det_info_init(ModuleInfo0, VarTypes0, PredId, ProcId, Globals,
+		DetInfo0),
 	proc_info_get_initial_instmap(ProcInfo0, ModuleInfo0, InstMap0),
 	proc_info_varset(ProcInfo0, VarSet0),
-	proc_info_vartypes(ProcInfo0, VarTypes0),
+	proc_info_inst_varset(ProcInfo0, InstVarSet0),
+	proc_info_typeinfo_varmap(ProcInfo0, TVarMap0),
+	proc_info_typeclass_info_varmap(ProcInfo0, TCVarMap0),
 	proc_info_goal(ProcInfo0, Goal0),
 
 	simplify_info_init(DetInfo0, Simplifications, InstMap0,
-		VarSet0, VarTypes0, Info0),
+		VarSet0, InstVarSet0, TVarMap0, TCVarMap0, Info0),
 	simplify__process_goal(Goal0, Goal, Info0, Info),
 	
 	simplify_info_get_varset(Info, VarSet),
 	simplify_info_get_var_types(Info, VarTypes),
+	simplify_info_get_type_info_varmap(Info, TVarMap),
+	simplify_info_get_typeclass_info_varmap(Info, TCVarMap),
 	proc_info_set_varset(ProcInfo0, VarSet, ProcInfo1),
 	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
-	proc_info_set_goal(ProcInfo2, Goal, ProcInfo),
+	proc_info_set_goal(ProcInfo2, Goal, ProcInfo3),
+	proc_info_set_typeinfo_varmap(ProcInfo3, TVarMap, ProcInfo4),
+	proc_info_set_typeclass_info_varmap(ProcInfo4, TCVarMap, ProcInfo),
 	simplify_info_get_module_info(Info, ModuleInfo),
 	simplify_info_get_msgs(Info, Msgs).
 
@@ -208,9 +215,8 @@ simplify__do_process_goal(Goal0, Goal, Info0, Info) :-
 	( simplify_info_requantify(Info1) ->
 		Goal1 = _ - GoalInfo1,
 		goal_info_get_nonlocals(GoalInfo1, NonLocals),
-
-		implicitly_quantify_goal(Goal1, VarSet0, VarTypes0, NonLocals,
-			Goal2, VarSet, VarTypes, _),
+		implicitly_quantify_goal(Goal1, VarSet0, VarTypes0,
+			NonLocals, Goal2, VarSet, VarTypes, _),
 
 		simplify_info_set_varset(Info1, VarSet, Info2),
 		simplify_info_set_var_types(Info2, VarTypes, Info3),
@@ -223,7 +229,8 @@ simplify__do_process_goal(Goal0, Goal, Info0, Info) :-
 
 		simplify_info_get_module_info(Info3, ModuleInfo3),
 		recompute_instmap_delta(RecomputeAtomic, Goal2, Goal3,
-			VarTypes, InstMap0, ModuleInfo3, ModuleInfo4),
+			VarTypes, Info3^inst_varset, InstMap0, ModuleInfo3,
+			ModuleInfo4),
 		simplify_info_set_module_info(Info3, ModuleInfo4, Info4)
 	;
 		Goal3 = Goal1,
@@ -474,7 +481,9 @@ simplify__enforce_invariant(GoalInfo0, GoalInfo, Info0, Info) :-
 
 simplify__goal_2(conj(Goals0), GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 	simplify_info_get_instmap(Info0, InstMap0),
-	simplify__conj(Goals0, [], Goals, GoalInfo0, Info0, Info1),
+	simplify__excess_assigns_in_conj(GoalInfo0,
+		Goals0, Goals1, Info0, Info1a),
+	simplify__conj(Goals1, [], Goals, GoalInfo0, Info1a, Info1),
 	simplify_info_set_instmap(Info1, InstMap0, Info2),
 	( Goals = [] ->
 		goal_info_get_context(GoalInfo0, Context),
@@ -695,7 +704,7 @@ simplify__goal_2(Goal0, GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 		% Don't warn about directly recursive calls.
 		% (That would cause spurious warnings, particularly
 		% with builtin predicates, or preds defined using
-		% pragma c_code.)
+		% pragma foreign.)
 		%
 		simplify_info_get_det_info(Info0, DetInfo0),
 		det_info_get_pred_id(DetInfo0, ThisPredId),
@@ -899,6 +908,15 @@ simplify__goal_2(Goal0, GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 	% other disjunct. (We could also eliminate A, but we leave
 	% that to the recursive invocations.)
 	%
+	% Note however that rerunning determinism analysis, which
+	% we do at the end of simplification, may introduce more
+	% occurrences of these; since we don't iterate simplification
+	% and determinism anaysis until a fixpoint is reached,
+	% we don't guarantee to eliminate all such if-then-elses.
+	% Hence the code generator must be prepared to handle the
+	% case when the condition of an if-then-else has determinism
+	% `det' or `failure'.
+	%
 	% The conjunction operator in the remaining disjunct ought to be
 	% a sequential conjunction, because Mercury's if-then-else always
 	% guarantees sequentiality, whereas conjunction only guarantees
@@ -912,7 +930,6 @@ simplify__goal_2(Goal0, GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 simplify__goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 		GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 	Cond0 = _ - CondInfo0,
-
 	goal_info_get_determinism(CondInfo0, CondDetism0),
 	determinism_components(CondDetism0, CondCanFail0, CondSolns0),
 	( CondCanFail0 = cannot_fail ->
@@ -927,7 +944,15 @@ simplify__goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 	; CondSolns0 = at_most_zero ->
 		% Optimize away the condition and the `then' part.
 		det_negation_det(CondDetism0, MaybeNegDetism),
-		( Cond0 = not(NegCond) - _ ->
+		(
+			Cond0 = not(NegCond) - _,
+			% XXX BUG! This optimization is only safe if it
+			% preserves mode correctness, which means in particular
+			% that the the negated goal must not clobber any
+			% variables.
+			% For now I've just disabled the optimization.
+			semidet_fail
+		->
 			Cond = NegCond
 		;
 			(
@@ -968,6 +993,10 @@ simplify__goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 		simplify__goal(conj(List) - GoalInfo0, Goal - GoalInfo,
 			Info0, Info)
 	;
+		%
+		% recursively simplify the sub-goals,
+		% and rebuild the resulting if-then-else
+		%
 		simplify_info_get_instmap(Info0, InstMap0),
 		simplify__goal(Cond0, Cond, Info0, Info1),
 		simplify_info_update_instmap(Info1, Cond, Info2),
@@ -988,33 +1017,58 @@ simplify__goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
 		merge_instmap_deltas(InstMap0, NonLocals,
 			[CondThenDelta, ElseDelta], NewDelta,
 			ModuleInfo0, ModuleInfo1),
-		simplify_info_set_module_info(Info6, ModuleInfo1, Info),
+		simplify_info_set_module_info(Info6, ModuleInfo1, Info7),
 		goal_info_set_instmap_delta(GoalInfo0, NewDelta, GoalInfo1),
 		IfThenElse = if_then_else(Vars, Cond, Then, Else, SM),
-		%
-		% If-then-elses that are det or semidet may nevertheless
-		% contain nondet or multidet conditions. If this happens, the
-		% if-then-else must be put inside a `some' to appease the code
-		% generator.
-		%
+
 		goal_info_get_determinism(GoalInfo0, IfThenElseDetism0),
 		determinism_components(IfThenElseDetism0, IfThenElseCanFail,
 			IfThenElseNumSolns),
+
+		goal_info_get_determinism(CondInfo, CondDetism),
+		determinism_components(CondDetism, CondCanFail, CondSolns),
 		(
-			simplify_do_once(Info),
-			goal_info_get_determinism(CondInfo, CondDetism),
-			determinism_components(CondDetism, _, at_most_many),
-			IfThenElseNumSolns \= at_most_many
+			%
+			% check again if we can apply one of the above
+			% simplifications after having simplified the 
+			% sub-goals (we need to do this to ensure that
+			% the goal is fully simplified, to maintain the
+			% invariants that the MLDS back-end depends on)
+			%
+			( CondCanFail = cannot_fail
+			; CondSolns = at_most_zero
+			; Else = disj([], _) - _ 
+			)
 		->
-			determinism_components(InnerDetism, IfThenElseCanFail,
-				at_most_many),
-			goal_info_set_determinism(GoalInfo1, InnerDetism,
-				InnerInfo),
-			Goal = some([], can_remove, IfThenElse - InnerInfo)
+			simplify_info_undo_goal_updates(Info0, Info7, Info8),
+			simplify__goal_2(IfThenElse, GoalInfo1,
+				Goal, GoalInfo, Info8, Info)
 		;
-			Goal = IfThenElse
-		),
-		GoalInfo = GoalInfo1
+			(
+				%
+				% If-then-elses that are det or semidet may
+				% nevertheless contain nondet or multidet
+				% conditions. If this happens, the if-then-else
+				% must be put inside a `some' to appease the
+				% code generator.  (Both the MLDS and LLDS
+				% back-ends rely on this.)
+				%
+				simplify_do_once(Info),
+				CondSolns = at_most_many,
+				IfThenElseNumSolns \= at_most_many
+			->
+				determinism_components(InnerDetism,
+					IfThenElseCanFail, at_most_many),
+				goal_info_set_determinism(GoalInfo1,
+					InnerDetism, InnerInfo),
+				Goal = some([], can_remove,
+					IfThenElse - InnerInfo)
+			;
+				Goal = IfThenElse
+			),
+			GoalInfo = GoalInfo1,
+			Info = Info7
+		)
 	).
 
 simplify__goal_2(not(Goal0), GoalInfo0, Goal, GoalInfo, Info0, Info) :-
@@ -1050,7 +1104,12 @@ simplify__goal_2(not(Goal0), GoalInfo0, Goal, GoalInfo, Info0, Info) :-
 		Info = Info3
 	;
 		% remove double negation
-		Goal1 = not(SubGoal - SubGoalInfo) - _
+		Goal1 = not(SubGoal - SubGoalInfo) - _,
+		% XXX BUG! This optimization is only safe if it preserves
+		% mode correctness, which means in particular that the
+		% the negated goal must not clobber any variables.
+		% For now I've just disabled the optimization.
+		semidet_fail
 	->
 		simplify__maybe_wrap_goal(GoalInfo0, SubGoalInfo, SubGoal,
 			Goal, GoalInfo, Info3, Info)
@@ -1083,7 +1142,7 @@ simplify__goal_2(some(Vars1, CanRemove0, Goal1), SomeInfo,
 	).
 
 simplify__goal_2(Goal0, GoalInfo, Goal, GoalInfo, Info0, Info) :-
-	Goal0 = pragma_c_code(_, PredId, ProcId, Args, _, _, _),
+	Goal0 = pragma_foreign_code(_, PredId, ProcId, Args, _, _, _),
 	(
 		simplify_do_calls(Info0),
 		goal_info_is_pure(GoalInfo)
@@ -1124,30 +1183,8 @@ simplify__process_compl_unify(XVar, YVar, UniMode, CanFail, _OldTypeInfoVars,
 		% are being unified.
 		%
 		simplify__type_info_locn(TypeVar, TypeInfoVar, ExtraGoals),
-		{ ArgVars = [TypeInfoVar, XVar, YVar] },
-
-		{ module_info_get_predicate_table(ModuleInfo,
-			PredicateTable) },
-		{ mercury_public_builtin_module(MercuryBuiltin) },
-		{ predicate_table_search_pred_m_n_a(PredicateTable,
-			MercuryBuiltin, "unify", 2, [CallPredId])
-		->
-			PredId = CallPredId
-		;
-			error("simplify.m: can't find `builtin:unify/2'")
-		},
-		% Note: the mode for polymorphic unifications
-		% should be `in, in'. 
-		% (This should have been checked by mode analysis.)
-		{ hlds_pred__in_in_unification_proc_id(ProcId) },
-
-		{ SymName = unqualified("unify") },
-		{ code_util__builtin_state(ModuleInfo, PredId, ProcId,
-			BuiltinState) },
-		{ CallContext = call_unify_context(XVar, var(YVar), Context) },
-		{ Call = call(PredId, ProcId, ArgVars,
-			BuiltinState, yes(CallContext), SymName)
-			- GoalInfo0 }
+		{ simplify__call_generic_unify(TypeInfoVar, XVar, YVar,
+			ModuleInfo, Context, GoalInfo0, Call) }
 
 	; { type_is_higher_order(Type, _, _, _) } ->
 		%
@@ -1176,46 +1213,118 @@ simplify__process_compl_unify(XVar, YVar, UniMode, CanFail, _OldTypeInfoVars,
 		simplify__goal_2(Call0, GoalInfo0, Call1, GoalInfo),
 		{ Call = Call1 - GoalInfo },
 		{ ExtraGoals = [] }
-
-	; { type_to_type_id(Type, TypeId, TypeArgs) } ->
-		%
-		% Convert other complicated unifications into
-		% calls to specific unification predicates,
-		% inserting extra typeinfo arguments if necessary.
-		%
-
-		% generate code to construct the new type_info arguments
-		simplify__make_type_info_vars(TypeArgs, TypeInfoVars,
-			ExtraGoals),
-
-		% create the new call goal
-		{ list__append(TypeInfoVars, [XVar, YVar], ArgVars) },
-		{ module_info_get_special_pred_map(ModuleInfo,
-			SpecialPredMap) },
-		{ map__lookup(SpecialPredMap, unify - TypeId, PredId) },
-		{ determinism_components(Det, CanFail, at_most_one) },
-		{ unify_proc__lookup_mode_num(ModuleInfo, TypeId,
-		 	UniMode, Det, ProcId) },
-		{ SymName = unqualified("__Unify__") },
-		{ CallContext = call_unify_context(XVar, var(YVar), Context) },
-		{ Call0 = call(PredId, ProcId, ArgVars, not_builtin,
-			yes(CallContext), SymName) },
-
-		% add the extra type_info vars to the nonlocals for the call
-		{ goal_info_get_nonlocals(GoalInfo0, NonLocals0) },
-		{ set__insert_list(NonLocals0, TypeInfoVars, NonLocals) },
-		{ goal_info_set_nonlocals(GoalInfo0, NonLocals,
-			CallGoalInfo0) },
-
-		% recursively simplify the call goal
-		simplify__goal_2(Call0, CallGoalInfo0, Call1, CallGoalInfo1),
-		{ Call = Call1 - CallGoalInfo1 }
 	;
-		{ error("simplify: type_to_type_id failed") }
+		{ type_to_type_id(Type, TypeIdPrime, TypeArgsPrime) ->
+			TypeId = TypeIdPrime,
+			TypeArgs = TypeArgsPrime
+		;
+			error("simplify: type_to_type_id failed")
+		},
+		{ determinism_components(Det, CanFail, at_most_one) },
+		{ unify_proc__lookup_mode_num(ModuleInfo, TypeId, UniMode, Det,
+			ProcId) },
+		{ module_info_globals(ModuleInfo, Globals) },
+		{ globals__lookup_bool_option(Globals, special_preds,
+			SpecialPreds) },
+		(
+			{ hlds_pred__in_in_unification_proc_id(ProcId) },
+			{
+				SpecialPreds = no
+			;
+				SpecialPreds = yes,
+
+				%
+				% For most imported types we only generate
+				% unification predicate declarations if they
+				% are needed for complicated unifications
+				% other than proc_id 0.
+				% higher_order.m will specialize these cases
+				% if possible.
+				%
+				special_pred_is_generated_lazily(ModuleInfo,
+					TypeId)
+			}
+		->
+			simplify__make_type_info_vars([Type], TypeInfoVars,
+				ExtraGoals),
+			{ TypeInfoVars = [TypeInfoVarPrime] ->
+				TypeInfoVar = TypeInfoVarPrime
+			;
+				error("simplify__process_compl_unify: more than one typeinfo for one type var")
+			},
+			{ simplify__call_generic_unify(TypeInfoVar, XVar, YVar,
+				ModuleInfo, Context, GoalInfo0, Call) }
+		;
+			%
+			% Convert other complicated unifications into
+			% calls to specific unification predicates,
+			% inserting extra typeinfo arguments if necessary.
+			%
+
+			simplify__make_type_info_vars(TypeArgs,
+				TypeInfoVars, ExtraGoals),
+			{ simplify__call_specific_unify(TypeId, TypeInfoVars,
+				XVar, YVar, ProcId, ModuleInfo, Context,
+				GoalInfo0, Call0, CallGoalInfo0) },
+			simplify__goal_2(Call0, CallGoalInfo0,
+				Call1, CallGoalInfo1),
+			{ Call = Call1 - CallGoalInfo1 }
+		)
 	),
 	{ list__append(ExtraGoals, [Call], ConjList) },
 	{ conj_list_to_goal(ConjList, GoalInfo0, Goal) }.
 
+:- pred simplify__call_generic_unify(prog_var::in, prog_var::in,  prog_var::in, 
+	module_info::in, unify_context::in, hlds_goal_info::in, hlds_goal::out)
+	is det.
+
+simplify__call_generic_unify(TypeInfoVar, XVar, YVar, ModuleInfo, Context,
+		GoalInfo0, Call) :-
+	ArgVars = [TypeInfoVar, XVar, YVar],
+	module_info_get_predicate_table(ModuleInfo, PredicateTable),
+	mercury_public_builtin_module(MercuryBuiltin),
+	( predicate_table_search_pred_m_n_a(PredicateTable,
+		MercuryBuiltin, "unify", 2, [CallPredId])
+	->
+		PredId = CallPredId
+	;
+		error("simplify.m: can't find `builtin:unify/2'")
+	),
+	% Note: the mode for polymorphic unifications
+	% should be `in, in'. 
+	% (This should have been checked by mode analysis.)
+	hlds_pred__in_in_unification_proc_id(ProcId),
+
+	SymName = unqualified("unify"),
+	code_util__builtin_state(ModuleInfo, PredId, ProcId, BuiltinState),
+	CallContext = call_unify_context(XVar, var(YVar), Context),
+	goal_info_get_nonlocals(GoalInfo0, NonLocals0),
+	set__insert(NonLocals0, TypeInfoVar, NonLocals),
+	goal_info_set_nonlocals(GoalInfo0, NonLocals, GoalInfo),
+	Call = call(PredId, ProcId, ArgVars, BuiltinState, yes(CallContext),
+		SymName) - GoalInfo.
+
+:- pred simplify__call_specific_unify(type_id::in, list(prog_var)::in,
+	prog_var::in, prog_var::in, proc_id::in,
+	module_info::in, unify_context::in, hlds_goal_info::in,
+	hlds_goal_expr::out, hlds_goal_info::out) is det.
+
+simplify__call_specific_unify(TypeId, TypeInfoVars, XVar, YVar, ProcId,
+		ModuleInfo, Context, GoalInfo0, CallExpr, CallGoalInfo) :-
+	% create the new call goal
+	list__append(TypeInfoVars, [XVar, YVar], ArgVars),
+	module_info_get_special_pred_map(ModuleInfo, SpecialPredMap),
+	map__lookup(SpecialPredMap, unify - TypeId, PredId),
+	SymName = unqualified("__Unify__"),
+	CallContext = call_unify_context(XVar, var(YVar), Context),
+	CallExpr = call(PredId, ProcId, ArgVars, not_builtin,
+		yes(CallContext), SymName),
+
+	% add the extra type_info vars to the nonlocals for the call
+	goal_info_get_nonlocals(GoalInfo0, NonLocals0),
+	set__insert_list(NonLocals0, TypeInfoVars, NonLocals),
+	goal_info_set_nonlocals(GoalInfo0, NonLocals, CallGoalInfo).
+
 :- pred simplify__make_type_info_vars(list(type)::in, list(prog_var)::out,
 	list(hlds_goal)::out, simplify_info::in, simplify_info::out) is det.
 
@@ -1274,7 +1383,7 @@ simplify__make_type_info_vars(Types, TypeInfoVars, TypeInfoGoals,
 
 simplify__type_info_locn(TypeVar, TypeInfoVar, Goals) -->
 	=(Info0),
-	{ simplify_info_get_typeinfo_map(Info0, TypeInfoMap) },
+	{ simplify_info_get_type_info_varmap(Info0, TypeInfoMap) },
 	{ map__lookup(TypeInfoMap, TypeVar, TypeInfoLocn) },
 	(
 			% If the typeinfo is available in a variable,
@@ -1466,18 +1575,11 @@ simplify__conj([Goal0 | Goals0], RevGoals0, Goals, ConjInfo, Info0, Info) :-
 		),
 		list__reverse(RevGoals, Goals)
 	    ;
-		simplify__excess_assigns(Goal1, ConjInfo,
-			Goals0, Goals1, RevGoals0, RevGoals1,
-			GoalNeeded, Info1, Info2),
-		( GoalNeeded = yes ->
-			simplify__conjoin_goal_and_rev_goal_list(Goal1,
-				RevGoals1, RevGoals2)
-		;
-			RevGoals2 = RevGoals1
-		),
-		simplify_info_update_instmap(Info2, Goal1, Info3),
-		simplify__conj(Goals1, RevGoals2, Goals,
-			ConjInfo, Info3, Info)
+		simplify__conjoin_goal_and_rev_goal_list(Goal1,
+			RevGoals0, RevGoals1),
+		simplify_info_update_instmap(Info1, Goal1, Info2),
+		simplify__conj(Goals0, RevGoals1, Goals,
+			ConjInfo, Info2, Info)
 	    )
 	).
 
@@ -1506,43 +1608,108 @@ simplify__par_conj([Goal0 |Goals0], [Goal | Goals], Info0, Info1, Info) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred simplify__excess_assigns(hlds_goal::in, hlds_goal_info::in,
-		hlds_goals::in, hlds_goals::out,
-		hlds_goals::in, hlds_goals::out, bool::out,
+:- pred simplify__excess_assigns_in_conj(hlds_goal_info::in,
+		list(hlds_goal)::in, list(hlds_goal)::out,
 		simplify_info::in, simplify_info::out) is det.
 
-simplify__excess_assigns(Goal0, ConjInfo, Goals0, Goals,
-		RevGoals0, RevGoals, GoalNeeded, Info0, Info) :-
-	(
-		simplify_do_excess_assigns(Info0),
-		Goal0 = unify(_, _, _, Unif, _) - _,
-		goal_info_get_nonlocals(ConjInfo, NonLocals),
-		Unif = assign(LeftVar, RightVar),
-		( \+ set__member(LeftVar, NonLocals) ->
-			LocalVar = LeftVar, ReplacementVar = RightVar
-		; \+ set__member(RightVar, NonLocals) ->
-			LocalVar = RightVar, ReplacementVar = LeftVar
-		;
-			fail
-		)
-	->
-		GoalNeeded = no,
+simplify__excess_assigns_in_conj(ConjInfo, Goals0, Goals,
+		Info0, Info) :-
+	( simplify_do_excess_assigns(Info0) ->
+		goal_info_get_nonlocals(ConjInfo, ConjNonLocals),
 		map__init(Subn0),
-		map__det_insert(Subn0, LocalVar, ReplacementVar, Subn),
-		goal_util__rename_vars_in_goals(Goals0, no,
-			Subn, Goals),
-		goal_util__rename_vars_in_goals(RevGoals0, no,
-			Subn, RevGoals),
-		simplify_info_get_varset(Info0, VarSet0),
-		varset__delete_var(VarSet0, LocalVar, VarSet),
-		simplify_info_set_varset(Info0, VarSet, Info)
+		simplify__find_excess_assigns_in_conj(ConjNonLocals,
+			Goals0, [], RevGoals, Subn0, Subn1),
+		( map__is_empty(Subn1) ->
+			Goals = Goals0,
+			Info = Info0
+		;
+			renaming_transitive_closure(Subn1, Subn),
+			list__reverse(RevGoals, Goals1),
+			MustSub = no,
+			goal_util__rename_vars_in_goals(Goals1, MustSub,
+				Subn, Goals),
+			simplify_info_get_varset(Info0, VarSet0),
+			map__keys(Subn0, RemovedVars),
+			varset__delete_vars(VarSet0, RemovedVars, VarSet),
+			simplify_info_set_varset(Info0, VarSet, Info1),
+			simplify_info_get_type_info_varmap(Info1, TVarMap0),
+			apply_substitutions_to_var_map(TVarMap0,
+				map__init, map__init, Subn, TVarMap),
+			simplify_info_set_type_info_varmap(Info1, TVarMap,
+				Info2),
+			simplify_info_get_typeclass_info_varmap(Info2,
+				TCVarMap0),
+			apply_substitutions_to_typeclass_var_map(TCVarMap0,
+				map__init, map__init, Subn, TCVarMap),
+			simplify_info_set_typeclass_info_varmap(Info2,
+				TCVarMap, Info)
+		)
 	;
-		GoalNeeded = yes,
 		Goals = Goals0,
-		RevGoals = RevGoals0,
 		Info = Info0
 	).
 
+:- type var_renaming == map(prog_var, prog_var).
+
+:- pred simplify__find_excess_assigns_in_conj(set(prog_var)::in,
+	list(hlds_goal)::in, list(hlds_goal)::in, list(hlds_goal)::out,
+	var_renaming::in, var_renaming::out) is det.
+
+simplify__find_excess_assigns_in_conj(_, [], RevGoals, RevGoals,
+			Subn, Subn).
+simplify__find_excess_assigns_in_conj(ConjNonLocals, [Goal | Goals],
+			RevGoals0, RevGoals, Subn0, Subn) :-
+	( goal_is_excess_assign(ConjNonLocals, Goal, Subn0, Subn1) ->
+		RevGoals1 = RevGoals0,
+		Subn2 = Subn1
+	;
+		RevGoals1 = [Goal | RevGoals0],
+		Subn2 = Subn0
+	),
+	simplify__find_excess_assigns_in_conj(ConjNonLocals, Goals,
+		RevGoals1, RevGoals, Subn2, Subn).
+
+:- pred goal_is_excess_assign(set(prog_var)::in, hlds_goal::in,
+	var_renaming::in, var_renaming::out) is semidet.
+
+goal_is_excess_assign(ConjNonLocals, Goal0, Subn0, Subn) :-
+	Goal0 = unify(_, _, _, Unif, _) - _,
+	Unif = assign(LeftVar0, RightVar0),
+
+	%
+	% Check if we've already substituted
+	% one or both of the variables.
+	%
+	find_renamed_var(Subn0, LeftVar0, LeftVar),
+	find_renamed_var(Subn0, RightVar0, RightVar),
+	( \+ set__member(LeftVar, ConjNonLocals) ->
+		map__det_insert(Subn0, LeftVar, RightVar, Subn)
+	; \+ set__member(RightVar, ConjNonLocals) ->
+		map__det_insert(Subn0, RightVar, LeftVar, Subn)
+	;
+		fail
+	).
+
+:- pred find_renamed_var(var_renaming, prog_var, prog_var).
+:- mode find_renamed_var(in, in, out) is det.
+
+find_renamed_var(Subn, Var0, Var) :-
+	( map__search(Subn, Var0, Var1) ->
+		find_renamed_var(Subn, Var1, Var)
+	;
+		Var = Var0
+	).
+
+	% Collapse chains of renamings.
+:- pred renaming_transitive_closure(var_renaming, var_renaming).
+:- mode renaming_transitive_closure(in, out) is det.
+
+renaming_transitive_closure(VarRenaming0, VarRenaming) :-
+	map__map_values(
+		(pred(_::in, Value0::in, Value::out) is det :-
+			find_renamed_var(VarRenaming0, Value0, Value)
+		), VarRenaming0, VarRenaming).
+
 %-----------------------------------------------------------------------------%
 
 :- pred simplify__switch(prog_var, list(case), list(case), list(case), 
@@ -1638,7 +1805,7 @@ simplify__create_test_unification(Var, ConsId, ConsArity,
 	UniMode = (Inst0 -> Inst0) - (Inst0 -> Inst0),
 	UnifyContext = unify_context(explicit, []),
 	Unification = deconstruct(Var, ConsId,
-		ArgVars, UniModes, can_fail),
+		ArgVars, UniModes, can_fail, no),
 	ExtraGoal = unify(Var, functor(ConsId, ArgVars),
 		UniMode, Unification, UnifyContext),
 	set__singleton_set(NonLocals, Var),
@@ -1752,7 +1919,7 @@ simplify__disj([Goal0 | Goals0], RevGoals0, Goals,  PostBranchInstMaps0,
 	****/
 
 :- pred simplify__fixup_disj(list(hlds_goal), determinism, bool,
-	hlds_goal_info, follow_vars, hlds_goal_expr,
+	hlds_goal_info, store_map, hlds_goal_expr,
 	simplify_info, simplify_info).
 :- mode simplify__fixup_disj(in, in, in, in, in, out, in, out) is det.
 
@@ -1779,7 +1946,7 @@ simplify__fixup_disj(Disjuncts, _, _OutputVars, GoalInfo, SM,
 	%		Disjunct3
 	%	).
 
-:- pred det_disj_to_ite(list(hlds_goal), hlds_goal_info, follow_vars,
+:- pred det_disj_to_ite(list(hlds_goal), hlds_goal_info, store_map,
 	hlds_goal).
 :- mode det_disj_to_ite(in, in, in, out) is det.
 
@@ -1837,42 +2004,54 @@ simplify__contains_multisoln_goal(Goals) :-
 
 :- type simplify_info
 	--->	simplify_info(
-			det_info,
-			set(det_msg),
-			set(simplification),
-			common_info,	% Info about common subexpressions.
-			instmap,
-			prog_varset,
-			map(prog_var, type),
-			bool,		% Does the goal need requantification.
-			bool,		% Do we need to recompute
+			det_info	::	det_info,
+			msgs		::	set(det_msg),
+			simplifications	::	set(simplification),
+			common_info	::	common_info,
+					% Info about common subexpressions.
+			instmap		::	instmap,
+			varset		::	prog_varset,
+			inst_varset	::	inst_varset,
+			requantify	::	bool,
+					% Does the goal need requantification.
+			recompute_atomic ::	bool,
+					% Do we need to recompute
 					% instmap_deltas for atomic goals
-			bool,		% Does determinism analysis need to
+			rerun_det	::	bool,
+					% Does determinism analysis need to
 					% be rerun.
-			int,		% Measure of the improvement in
+			cost_delta	::	int,
+					% Measure of the improvement in
 					% the goal from simplification.
-			int		% Count of the number of lambdas
+			lambdas		::	int,
+					% Count of the number of lambdas
 					% which enclose the current goal.
+			type_info_varmap ::	type_info_varmap,
+			typeclass_info_varmap ::	typeclass_info_varmap
 		).
 
 simplify_info_init(DetInfo, Simplifications0, InstMap,
-		VarSet, VarTypes, Info) :-
+		VarSet, InstVarSet, TVarMap, TCVarMap, Info) :-
 	common_info_init(CommonInfo),
 	set__init(Msgs),
 	set__list_to_set(Simplifications0, Simplifications),
 	Info = simplify_info(DetInfo, Msgs, Simplifications, CommonInfo,
-			InstMap, VarSet, VarTypes, no, no, no, 0, 0). 
+		InstMap, VarSet, InstVarSet, no, no, no, 0, 0,
+		TVarMap, TCVarMap). 
 
 	% Reinitialise the simplify_info before reprocessing a goal.
 :- pred simplify_info_reinit(set(simplification)::in, instmap::in,
 		simplify_info::in, simplify_info::out) is det.
 
-simplify_info_reinit(Simplifications, InstMap0, Info0, Info) :-
-	Info0 = simplify_info(DetInfo, Msgs, _, _, _,
-		VarSet, VarTypes, _, _, _, CostDelta, _),
-	common_info_init(Common),
-	Info = simplify_info(DetInfo, Msgs, Simplifications, Common, InstMap0,
-		VarSet, VarTypes, no, no, no, CostDelta, 0).
+simplify_info_reinit(Simplifications, InstMap0) -->
+	{ common_info_init(Common) },
+	^simplifications := Simplifications,
+	^common_info := Common,
+	^instmap := InstMap0,
+	^requantify := no,
+	^recompute_atomic := no,
+	^rerun_det := no,
+	^lambdas := 0.
 
 	% exported for common.m
 :- interface.
@@ -1880,9 +2059,10 @@ simplify_info_reinit(Simplifications, InstMap0, Info0, Info) :-
 :- import_module prog_data.
 :- import_module set.
 
-:- pred simplify_info_init(det_info, list(simplification), instmap,
-		prog_varset, map(prog_var, type), simplify_info).
-:- mode simplify_info_init(in, in, in, in, in, out) is det.
+:- pred simplify_info_init(det_info::in, list(simplification)::in, instmap::in,
+		prog_varset::in, inst_varset::in, 
+		type_info_varmap::in, typeclass_info_varmap::in,
+		simplify_info::out) is det.
 
 :- pred simplify_info_get_det_info(simplify_info::in, det_info::out) is det.
 :- pred simplify_info_get_msgs(simplify_info::in, set(det_msg)::out) is det.
@@ -1893,38 +2073,51 @@ simplify_info_reinit(Simplifications, InstMap0, Info0, Info) :-
 		common_info::out) is det.
 :- pred simplify_info_get_varset(simplify_info::in, prog_varset::out) is det.
 :- pred simplify_info_get_var_types(simplify_info::in,
-		map(prog_var, type)::out) is det.
+		vartypes::out) is det.
 :- pred simplify_info_requantify(simplify_info::in) is semidet.
 :- pred simplify_info_recompute_atomic(simplify_info::in) is semidet.
 :- pred simplify_info_rerun_det(simplify_info::in) is semidet.
 :- pred simplify_info_get_cost_delta(simplify_info::in, int::out) is det.
+:- pred simplify_info_get_type_info_varmap(simplify_info::in,
+		type_info_varmap::out) is det.
+:- pred simplify_info_get_typeclass_info_varmap(simplify_info::in,
+		typeclass_info_varmap::out) is det.
 
 :- pred simplify_info_get_module_info(simplify_info::in,
 		module_info::out) is det.
+:- pred simplify_info_get_pred_info(simplify_info::in,
+		pred_info::out) is det.
 
 :- implementation.
 
-simplify_info_get_det_info(simplify_info(Det, _,_,_,_,_,_,_,_,_,_,_), Det). 
-simplify_info_get_msgs(simplify_info(_, Msgs, _,_,_,_,_,_,_,_,_,_), Msgs).
-simplify_info_get_simplifications(simplify_info(_,_,Simplify,_,_,_,_,_,_,_,_,_),
-	Simplify). 
-simplify_info_get_common_info(simplify_info(_,_,_,Common, _,_,_,_,_,_,_,_),
-	Common).
-simplify_info_get_instmap(simplify_info(_,_,_,_, InstMap,_,_,_,_,_,_,_),
-	InstMap). 
-simplify_info_get_varset(simplify_info(_,_,_,_,_, VarSet, _,_,_,_,_,_), VarSet).
-simplify_info_get_var_types(simplify_info(_,_,_,_,_,_, VarTypes, _,_,_,_,_),
-	VarTypes). 
-simplify_info_requantify(simplify_info(_,_,_,_,_,_,_, yes, _,_,_,_)).
-simplify_info_recompute_atomic(simplify_info(_,_,_,_,_,_,_,_, yes,_,_,_)).
-simplify_info_rerun_det(simplify_info(_,_,_,_,_,_,_,_,_, yes,_,_)).
-simplify_info_get_cost_delta(simplify_info(_,_,_,_,_,_,_,_,_,_,CostDelta, _),
-	CostDelta).
+simplify_info_get_det_info(SI, SI^det_info).
+simplify_info_get_msgs(SI, SI^msgs).
+simplify_info_get_simplifications(SI, SI^simplifications).
+simplify_info_get_common_info(SI, SI^common_info).
+simplify_info_get_instmap(SI, SI^instmap).
+simplify_info_get_varset(SI, SI^varset).
+simplify_info_get_var_types(SI, VarTypes) :-
+	det_info_get_vartypes(SI^det_info, VarTypes).
+simplify_info_requantify(SI) :-
+	SI^requantify = yes.
+simplify_info_recompute_atomic(SI) :-
+	SI^recompute_atomic = yes.
+simplify_info_rerun_det(SI) :-
+	SI^rerun_det = yes.
+simplify_info_get_cost_delta(SI, SI^cost_delta).
+simplify_info_get_type_info_varmap(SI, SI^type_info_varmap).
+simplify_info_get_typeclass_info_varmap(SI, SI^typeclass_info_varmap).
 
 simplify_info_get_module_info(Info, ModuleInfo) :-
 	simplify_info_get_det_info(Info, DetInfo),
 	det_info_get_module_info(DetInfo, ModuleInfo).
 
+simplify_info_get_pred_info(Info, PredInfo) :-
+	simplify_info_get_det_info(Info, DetInfo),
+	det_info_get_module_info(DetInfo, ModuleInfo),
+	det_info_get_pred_id(DetInfo, PredId),
+	module_info_pred_info(ModuleInfo, PredId, PredInfo).
+
 :- interface.
 
 :- pred simplify_info_set_det_info(simplify_info::in,
@@ -1947,6 +2140,11 @@ simplify_info_get_module_info(Info, ModuleInfo) :-
 		simplify_info::out) is det.
 :- pred simplify_info_set_rerun_det(simplify_info::in,
 		simplify_info::out) is det.
+:- pred simplify_info_set_type_info_varmap(simplify_info::in,
+		type_info_varmap::in, simplify_info::out) is det.
+:- pred simplify_info_set_typeclass_info_varmap(simplify_info::in,
+		typeclass_info_varmap::in, simplify_info::out) is det.
+
 :- pred simplify_info_add_msg(simplify_info::in, det_msg::in,
 		simplify_info::out) is det.
 :- pred simplify_info_do_add_msg(simplify_info::in, det_msg::in, 
@@ -1967,37 +2165,23 @@ simplify_info_get_module_info(Info, ModuleInfo) :-
 
 :- implementation.
 
-simplify_info_set_det_info(simplify_info(_, B, C, D, E, F, G, H, I, J, K, L),
-		Det, simplify_info(Det, B, C, D, E, F, G, H, I, J, K, L)).
-simplify_info_set_msgs(simplify_info(A, _, C, D, E, F, G, H, I, J, K, L), Msgs,
-		simplify_info(A, Msgs, C, D, E, F, G, H, I, J, K, L)). 
-simplify_info_set_simplifications(
-		simplify_info(A, B, _, D, E, F, G, H, I, J, K, L),
-		Simp, simplify_info(A, B, Simp, D, E, F, G, H, I, J, K, L)).
-simplify_info_set_instmap(simplify_info(A, B, C, D, _, F, G, H, I, J, K, L), 
-		InstMap, 
-		simplify_info(A, B, C, D, InstMap, F, G, H, I, J, K, L)). 
-simplify_info_set_common_info(simplify_info(A, B, C, _, E, F, G, H, I, J, K, L),
-		Common, 
-		simplify_info(A, B, C, Common, E, F, G, H, I, J, K, L)). 
-simplify_info_set_varset(simplify_info(A, B, C, D, E, _, G, H, I, J, K, L), 
-		VarSet, 
-		simplify_info(A, B, C, D, E, VarSet, G, H, I, J, K, L)). 
-simplify_info_set_var_types(simplify_info(A, B, C, D, E, F, _, H, I, J, K, L),
-		VarTypes, simplify_info(A, B, C, D, E, F, VarTypes, H,I,J,K,L)).
-simplify_info_set_requantify(simplify_info(A, B, C, D, E, F, G, _, I, J, K, L),
-		simplify_info(A, B, C, D, E, F, G, yes, I, J, K, L)). 
-simplify_info_set_recompute_atomic(simplify_info(A, B, C, D, E, F, G,H,_,J,K,L),
-		simplify_info(A, B, C, D, E, F, G, H, yes, J, K, L)). 
-simplify_info_set_rerun_det(simplify_info(A, B, C, D, E, F, G,H,I,_,K,L),
-		simplify_info(A, B, C, D, E, F, G, H, I, yes, K, L)). 
-simplify_info_set_cost_delta(simplify_info(A, B, C, D, E, F, G, H, I, J, _, L),
-		Delta, simplify_info(A, B, C, D, E, F, G, H, I, J, Delta, L)). 
+simplify_info_set_det_info(SI, Det, SI^det_info := Det).
+simplify_info_set_msgs(SI, Msgs, SI^msgs := Msgs). 
+simplify_info_set_simplifications(SI, Simp, SI^simplifications := Simp).
+simplify_info_set_instmap(SI, InstMap, SI^instmap := InstMap). 
+simplify_info_set_common_info(SI, Common, SI^common_info := Common). 
+simplify_info_set_varset(SI, VarSet, SI^varset := VarSet). 
+simplify_info_set_var_types(SI, VarTypes, SI^det_info := DetInfo) :-
+	det_info_set_vartypes(SI ^ det_info, VarTypes, DetInfo).
+simplify_info_set_requantify(SI, SI^requantify := yes).
+simplify_info_set_recompute_atomic(SI, SI^recompute_atomic := yes).
+simplify_info_set_rerun_det(SI, SI^rerun_det := yes).
+simplify_info_set_cost_delta(SI, Delta, SI^cost_delta := Delta).
+simplify_info_set_type_info_varmap(SI, Map, SI^type_info_varmap := Map).
+simplify_info_set_typeclass_info_varmap(SI, Map,
+		SI^typeclass_info_varmap := Map).
 
-simplify_info_incr_cost_delta(
-		simplify_info(A, B, C, D, E, F,G,H,I,J, Delta0, L),
-		Incr, simplify_info(A, B, C, D, E, F, G, H, I, J, Delta, L)) :-
-	Delta is Delta0 + Incr.
+simplify_info_incr_cost_delta(SI, Incr, SI^cost_delta := SI^cost_delta + Incr).
 
 simplify_info_add_msg(Info0, Msg, Info) :-
 	( simplify_do_warn(Info0) ->
@@ -2011,14 +2195,9 @@ simplify_info_do_add_msg(Info0, Msg, Info) :-
 	set__insert(Msgs0, Msg, Msgs),
 	simplify_info_set_msgs(Info0, Msgs, Info).
 
-simplify_info_enter_lambda(
-		simplify_info(A, B, C, D, E, F, G, H, I, J, K, LambdaCount0),
-		simplify_info(A, B, C, D, E, F, G, H, I, J, K, LambdaCount)) :-
-	LambdaCount is LambdaCount0 + 1.
-simplify_info_leave_lambda(
-		simplify_info(A, B, C, D, E, F, G, H, I, J, K, LambdaCount0),
-		simplify_info(A, B, C, D, E, F, G, H, I, J, K, LambdaCount)) :-
-	LambdaCount1 is LambdaCount0 - 1,
+simplify_info_enter_lambda(SI, SI^lambdas := SI^lambdas + 1).
+simplify_info_leave_lambda(SI, SI^lambdas := LambdaCount) :-
+	LambdaCount1 is SI^lambdas - 1,
 	(
 		LambdaCount1 >= 0
 	->
@@ -2026,9 +2205,8 @@ simplify_info_leave_lambda(
 	;
 		error("simplify_info_leave_lambda: Left too many lambdas")
 	).
-simplify_info_inside_lambda(
-		simplify_info(_,_,_,_,_,_,_,_,_,_,_,LambdaCount)) :-
-	LambdaCount > 0.
+simplify_info_inside_lambda(SI) :-
+	SI^lambdas > 0.
 
 simplify_info_set_module_info(Info0, ModuleInfo, Info) :-
 	simplify_info_get_det_info(Info0, DetInfo0),
@@ -2073,25 +2251,11 @@ simplify_do_more_common(Info) :-
 	simplify_info_get_simplifications(Info, Simplifications),
 	set__member(extra_common_struct, Simplifications).
 
-:- pred simplify_info_get_typeinfo_map(simplify_info::in,
-		map(tvar, type_info_locn)::out) is det.
-
-simplify_info_get_typeinfo_map(Info0, TypeInfoMap) :-
-	simplify_info_get_det_info(Info0, DetInfo0),
-	det_info_get_module_info(DetInfo0, ModuleInfo),
-	det_info_get_pred_id(DetInfo0, ThisPredId),
-	det_info_get_proc_id(DetInfo0, ThisProcId),
-	module_info_pred_proc_info(ModuleInfo, ThisPredId, ThisProcId,
-		_PredInfo, ProcInfo),
-	proc_info_typeinfo_varmap(ProcInfo, TypeInfoMap).
-
 :- pred simplify_info_update_instmap(simplify_info::in, hlds_goal::in,
 		simplify_info::out) is det.
 
-simplify_info_update_instmap(
-		simplify_info(A, B, C, D, InstMap0, F, G, H, I, J, K, L), Goal,
-		simplify_info(A, B, C, D, InstMap, F, G, H, I, J, K, L)) :-
-	update_instmap(Goal, InstMap0, InstMap).
+simplify_info_update_instmap(SI, Goal, SI^instmap := InstMap) :-
+	update_instmap(Goal, SI^instmap, InstMap).
 
 :- type before_after
 	--->	before
@@ -2126,7 +2290,7 @@ simplify_info_maybe_clear_structs(BeforeAfter, Goal, Info0, Info) :-
 			Goal = GoalExpr - _,
 			GoalExpr \= call(_, _, _, _, _, _),
 			GoalExpr \= generic_call(_, _, _, _),
-			GoalExpr \= pragma_c_code(_, _, _, _, _, _, _)
+			GoalExpr \= pragma_foreign_code(_, _, _, _, _, _, _)
 		)
 	->
 		simplify_info_get_common_info(Info0, CommonInfo0),
