@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -178,7 +178,8 @@ first_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated,
 		WholeScc, ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		Error, Module0, Module).
-first_order_check_goal(pragma_c_code(_IsRec, CPred, CProc, _, _, _, _), 
+first_order_check_goal(pragma_foreign_code(_Attributes, CPred,
+			CProc, _, _, _, _), 
 		GoalInfo, Negated, WholeScc, ThisPredProcId, 
 		Error, Module0, Module) -->
 	(
@@ -334,8 +335,8 @@ higher_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated, WholeScc,
 		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
 	higher_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		HighOrderLoops, Error, Module0, Module).
-higher_order_check_goal(pragma_c_code(_IsRec, _, _, _, _, _, _), _GoalInfo, 
-	_Negated, _WholeScc, _ThisPredProcId, _HighOrderLoops, 
+higher_order_check_goal(pragma_foreign_code(_IsRec, _, _, _, _, _, _),
+	_GoalInfo, _Negated, _WholeScc, _ThisPredProcId, _HighOrderLoops, 
 	_, Module, Module) --> [].
 higher_order_check_goal(unify(_Var, _RHS, _Mode, _Uni, _Context), _GoalInfo,
 	_Negated, _WholeScc, _ThisPredProcId, _HighOrderLoops, 
@@ -821,8 +822,8 @@ check_goal1(not(Goal - _GoalInfo), Calls0, Calls, HasAT0, HasAT, CallsHO0,
 		CallsHO) :- 
 	check_goal1(Goal, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
 
-check_goal1(pragma_c_code(_IsRec, _CPred, _CProc, _, _, _, _), Calls, Calls, 
-		HasAT, HasAT, CallsHO, CallsHO).
+check_goal1(pragma_foreign_code(_Attrib, _CPred, _CProc, _, _, _, _),
+		Calls, Calls, HasAT, HasAT, CallsHO, CallsHO).
 
 check_goal1(bi_implication(_, _), _, _, _, _, _, _) :-
 	% these should have been expanded out by now
@@ -912,8 +913,8 @@ get_called_procs(some(_Vars, _, Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
 get_called_procs(not(Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
-get_called_procs(pragma_c_code(_IsRec, _CPred, _CProc, _, _, _, _),
-	Calls, Calls).
+get_called_procs(pragma_foreign_code(_Attrib, _CPred, _CProc,
+		_, _, _, _), Calls, Calls).
 get_called_procs(bi_implication(_, _), _, _) :-
 	% these should have been expanded out by now
 	error("get_called_procs: unexpected bi_implication").
