@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,28 +17,26 @@
 :- interface.
 
 :- import_module hlds_goal, hlds_module, hlds_pred, prog_data.
-:- import_module io, list.
+:- import_module bool, io, list.
 
-:- pred detect_switches(module_info, module_info, io__state, io__state).
-:- mode detect_switches(in, out, di, uo) is det.
+:- pred detect_switches(module_info::in, module_info::out,
+	io__state::di, io__state::uo) is det.
 
-:- pred detect_switches_in_proc(proc_id, pred_id, module_info, module_info).
-:- mode detect_switches_in_proc(in, in, in, out) is det.
+:- pred detect_switches_in_proc(proc_id::in, pred_id::in,
+	module_info::in, module_info::out) is det.
 
 	% find_bind_var(Var, ProcessUnify, Goal0, Goals, Subst0, Subst,
-	%		Result0, Result, Continue):
+	%		Result0, Result, FoundDeconstruct):
 	% 	Used by both switch_detection and cse_detection.
 	%	Searches through `Goal0' looking for the first deconstruction
 	%	unification with `Var' or an alias of `Var'.
-	%	`ProcessUnify' is called if a deconstruction unification with
-	%	`Var' is found, returning either the cons_id for
-	%	switch_detection or the common deconstruction for cse_detection.
 	%	If a deconstruction unification of the variable is found,
 	%	`ProcessUnify' is called to handle it and searching is stopped.
 	%	If not, `Result' is set to `Result0'.
-:- pred find_bind_var(prog_var, process_unify(Result, Info), hlds_goal,
-		hlds_goal, Result, Result, Info, Info).
-:- mode find_bind_var(in, in(process_unify), in, out, in, out, in, out) is det.
+:- pred find_bind_var(prog_var::in,
+	process_unify(Result, Info)::in(process_unify),
+	hlds_goal::in, hlds_goal::out, Result::in, Result::out,
+	Info::in, Info::out, bool::out) is det.
 
 :- type process_unify(Result, Info) ==
 	pred(prog_var, hlds_goal, list(hlds_goal), Result, Result, Info, Info).
@@ -52,7 +50,7 @@
 :- import_module hlds_goal, hlds_data, prog_data, instmap, inst_match.
 :- import_module modes, mode_util, type_util, det_util.
 :- import_module passes_aux, term.
-:- import_module bool, char, int, assoc_list, map, set, std_util, require.
+:- import_module char, int, assoc_list, map, set, std_util, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -216,8 +214,8 @@ detect_switches_in_goal_2(switch(Var, CanFail, Cases0, SM), _, InstMap,
 		VarTypes, ModuleInfo, switch(Var, CanFail, Cases, SM)) :-
 	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases).
 
-detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F,G), _, _, _, _,
-		pragma_c_code(A,B,C,D,E,F,G)).
+detect_switches_in_goal_2(pragma_foreign_code(A,B,C,D,E,F,G), _, _, _, _,
+		pragma_foreign_code(A,B,C,D,E,F,G)).
 detect_switches_in_goal_2(bi_implication(_, _), _, _, _, _, _) :-
 	% these should have been expanded out by now
 	error("detect_switches_in_goal_2: unexpected bi_implication").
@@ -409,7 +407,7 @@ partition_disj(Goals0, Var, GoalInfo, Left, CasesList) :-
 partition_disj_trial([], _Var, Left, Left, Cases, Cases).
 partition_disj_trial([Goal0 | Goals], Var, Left0, Left, Cases0, Cases) :-
 	find_bind_var(Var, find_bind_var_for_switch_in_deconstruct,
-		Goal0, Goal, no, MaybeFunctor, unit, _),
+		Goal0, Goal, no, MaybeFunctor, unit, _, _),
 	(
 		MaybeFunctor = yes(Functor),
 		Left1 = Left0,
@@ -436,14 +434,14 @@ find_bind_var_for_switch_in_deconstruct(_UnifyVar, Goal0, Goals,
 		_Result0, Result, _, unit) :-
 	(
 		Goal0 = unify(A, B, C, UnifyInfo0, E) - GoalInfo,
-		UnifyInfo0 = deconstruct(A, Functor, F, G, _)
+		UnifyInfo0 = deconstruct(A, Functor, F, G, _, I)
 	->
 		Result = yes(Functor),
 			% The deconstruction unification now becomes
 			% deterministic, since the test will get
 			% carried out in the switch.
 		UnifyInfo = deconstruct(A, Functor, F, G,
-			cannot_fail),
+			cannot_fail, I),
 		Goals = [unify(A, B, C, UnifyInfo, E) - GoalInfo]
 	;
 		error("find_bind_var_for_switch_in_deconstruct")
@@ -452,34 +450,50 @@ find_bind_var_for_switch_in_deconstruct(_UnifyVar, Goal0, Goals,
 %-----------------------------------------------------------------------------%
 
 find_bind_var(Var, ProcessUnify, Goal0, Goal,
-		Result0, Result, Info0, Info) :-
+		Result0, Result, Info0, Info, FoundDeconstruct) :-
 	map__init(Substitution),
 	find_bind_var(Var, ProcessUnify, Goal0, Goal, Substitution,
-		_, Result0, Result, Info0, Info, _).
+		_, Result0, Result, Info0, Info, DeconstructSearch),
+	(
+		DeconstructSearch = before_deconstruct,
+		FoundDeconstruct = no
+	;
+		DeconstructSearch = found_deconstruct,
+		FoundDeconstruct = yes
+	;
+		DeconstructSearch = given_up_search,
+		FoundDeconstruct = no
+	).
 
-:- pred find_bind_var(prog_var, process_unify(Result, Info), hlds_goal,
-		hlds_goal, prog_substitution, prog_substitution, Result, Result,
-		Info, Info, bool).
-:- mode find_bind_var(in, in(process_unify), in, out, in,
-	out, in, out, in, out, out) is det.
+:- type deconstruct_search
+	--->	before_deconstruct
+	;	found_deconstruct
+	;	given_up_search.
 
-find_bind_var(Var, ProcessUnify, Goal0 - GoalInfo, Goal, Substitution0,
-		Substitution, Result0, Result, Info0, Info, Continue) :-
+:- pred find_bind_var(prog_var::in,
+	process_unify(Result, Info)::in(process_unify),
+	hlds_goal::in, hlds_goal::out,
+	prog_substitution::in, prog_substitution::out, Result::in, Result::out,
+	Info::in, Info::out, deconstruct_search::out) is det.
+
+find_bind_var(Var, ProcessUnify, Goal0 - GoalInfo, Goal,
+		Substitution0, Substitution, Result0, Result, Info0, Info,
+		FoundDeconstruct) :-
 	( Goal0 = some(Vars, CanRemove, SubGoal0) ->
 		find_bind_var(Var, ProcessUnify, SubGoal0, SubGoal,
 			Substitution0, Substitution, Result0, Result,
-			Info0, Info, Continue),
+			Info0, Info, FoundDeconstruct),
 		Goal = some(Vars, CanRemove, SubGoal) - GoalInfo
 	; Goal0 = conj(SubGoals0) ->
 		conj_find_bind_var(Var, ProcessUnify, SubGoals0, SubGoals,
 			Substitution0, Substitution, Result0, Result,
-			Info0, Info, Continue),
+			Info0, Info, FoundDeconstruct),
 		Goal = conj(SubGoals) - GoalInfo
 	; Goal0 = unify(A, B, _, UnifyInfo0, _) ->
 		(
 			% check whether the unification is a deconstruction
 			% unification on Var or a variable aliased to Var
-			UnifyInfo0 = deconstruct(UnifyVar, _, _, _, _),
+			UnifyInfo0 = deconstruct(UnifyVar, _, _, _, _, _),
 			term__apply_rec_substitution(
 				term__variable(Var),
 				Substitution0, term__variable(Var1)),
@@ -491,11 +505,11 @@ find_bind_var(Var, ProcessUnify, Goal0 - GoalInfo, Goal, Substitution0,
 			call(ProcessUnify, Var, Goal0 - GoalInfo, Goals,
 				Result0, Result, Info0, Info),
 			conj_list_to_goal(Goals, GoalInfo, Goal),
-			Continue = no,
+			FoundDeconstruct = found_deconstruct,
 			Substitution = Substitution0
 		;
 			Goal = Goal0 - GoalInfo,
-			Continue = yes,
+			FoundDeconstruct = before_deconstruct,
 			% otherwise abstractly interpret the unification
 			Result = Result0,
 			Info = Info0,
@@ -511,33 +525,33 @@ find_bind_var(Var, ProcessUnify, Goal0 - GoalInfo, Goal, Substitution0,
 		Substitution = Substitution0,
 		Result = Result0,
 		Info = Info0,
-		Continue = no
+		FoundDeconstruct = given_up_search
 	).
 
-:- pred conj_find_bind_var(prog_var, process_unify(Result, Info), 
-		list(hlds_goal), list(hlds_goal), prog_substitution,
-		prog_substitution, Result, Result, Info, Info, bool).
-:- mode conj_find_bind_var(in, in(process_unify), in, out,
-	in, out, in, out, in, out, out) is det.
+:- pred conj_find_bind_var(prog_var::in,
+	process_unify(Result, Info)::in(process_unify), 
+	list(hlds_goal)::in, list(hlds_goal)::out,
+	prog_substitution::in, prog_substitution::out, Result::in, Result::out,
+	Info::in, Info::out, deconstruct_search::out) is det.
 
 conj_find_bind_var(_Var, _, [], [], Substitution, Substitution,
-		Result, Result, Info, Info, yes).
+		Result, Result, Info, Info, before_deconstruct).
 conj_find_bind_var(Var, ProcessUnify, [Goal0 | Goals0], [Goal | Goals],
 		Substitution0, Substitution, Result0, Result,
-		Info0, Info, Continue) :-
+		Info0, Info, FoundDeconstruct) :-
 	find_bind_var(Var, ProcessUnify, Goal0, Goal, Substitution0,
 		Substitution1, Result0, Result1,
-		Info0, Info1, Continue1),
-	( Continue1 = no ->
-		Continue = no,
+		Info0, Info1, FoundDeconstruct1),
+	( FoundDeconstruct1 = before_deconstruct ->
+		conj_find_bind_var(Var, ProcessUnify, Goals0, Goals,
+			Substitution1, Substitution, Result1, Result,
+			Info1, Info, FoundDeconstruct)
+	;
+		FoundDeconstruct = FoundDeconstruct1,
 		Goals = Goals0,
 		Substitution = Substitution1,
 		Result = Result1,
 		Info = Info1
-	;
-		conj_find_bind_var(Var, ProcessUnify, Goals0, Goals,
-			Substitution1, Substitution, Result1, Result,
-			Info1, Info, Continue)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -592,24 +606,10 @@ cases_to_switch(CasesList, Var, VarTypes, _GoalInfo, SM, InstMap, ModuleInfo,
 :- pred switch_covers_all_cases(sorted_case_list, type, module_info).
 :- mode switch_covers_all_cases(in, in, in) is semidet.
 
-switch_covers_all_cases(CasesList, Type, _ModuleInfo) :-
-	Type = term__functor(term__atom("character"), [], _),
-	% XXX the following code uses the source machine's character size,
-	% not the target's, so it won't work if cross-compiling to a
-	% machine with a different size character.
-	char__max_char_value(MaxChar),
-	char__min_char_value(MinChar),
-	NumChars is MaxChar - MinChar + 1,
-	list__length(CasesList, NumChars).
-
 switch_covers_all_cases(CasesList, Type, ModuleInfo) :-
-	type_to_type_id(Type, TypeId, _),
-	module_info_types(ModuleInfo, TypeTable),
-	map__search(TypeTable, TypeId, TypeDefn),
-	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(_, ConsTable, _, _),
-	map__keys(ConsTable, Constructors),
-	list__same_length(CasesList, Constructors).
+	type_util__switch_type_num_functors(ModuleInfo, Type, NumFunctors),
+	list__length(CasesList, NumCases),
+	NumCases = NumFunctors.
 
 	% convert the assoc_list(cons_id, list(hlds_goal) back into
 	% a plain list(case).
