@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,7 +14,10 @@
 
 :- interface.
 
-:- import_module hlds_goal, hlds_data, prog_data, llds, switch_gen, code_info.
+:- import_module prog_data, hlds_goal, hlds_data.
+:- import_module switch_util, code_model.
+:- import_module llds, code_info.
+
 :- import_module list.
 
 	% Generate intelligent indexing code for tag based switches.
@@ -33,22 +36,6 @@
 :- import_module assoc_list, map, tree, bool, int, string.
 :- import_module require, std_util.
 
-% where is the secondary tag (if any) for this primary tag value
-:- type stag_loc	--->	none ; local ; remote.
-
-% map secondary tag values (-1 stands for none) to their goal
-:- type stag_goal_map	==	map(int, hlds_goal).
-:- type stag_goal_list	==	assoc_list(int, hlds_goal).
-
-% map primary tag values to the set of their goals
-:- type ptag_case_map	==	map(tag_bits, pair(stag_loc, stag_goal_map)).
-:- type ptag_case_list	==	assoc_list(tag_bits,
-					pair(stag_loc, stag_goal_map)).
-
-% map primary tag values to the number of constructors sharing them
-:- type ptag_count_map	==	map(tag_bits, pair(stag_loc, int)).
-:- type ptag_count_list ==	assoc_list(tag_bits, pair(stag_loc, int)).
-
 %-----------------------------------------------------------------------------%
 
 	% The idea is to generate two-level switches, first on the primary
@@ -189,11 +176,11 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 	code_info__get_proc_info(ProcInfo),
 	{ proc_info_vartypes(ProcInfo, VarTypes) },
 	{ map__lookup(VarTypes, Var, Type) },
-	{ tag_switch__get_ptag_counts(Type, ModuleInfo,
+	{ switch_util__get_ptag_counts(Type, ModuleInfo,
 		MaxPrimary, PtagCountMap) },
 	{ map__to_assoc_list(PtagCountMap, PtagCountList) },
 	{ map__init(PtagCaseMap0) },
-	{ tag_switch__group_cases_by_ptag(Cases, PtagCaseMap0, PtagCaseMap) },
+	{ switch_util__group_cases_by_ptag(Cases, PtagCaseMap0, PtagCaseMap) },
 
 	{ map__count(PtagCaseMap, PtagsUsed) },
 	code_info__get_globals(Globals),
@@ -222,7 +209,8 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 	% or if the "register" we get is likely to be slower than
 	% recomputing the tag from scratch.
 
-	code_info__produce_variable_in_reg(Var, VarCode, VarRval),
+	code_info__produce_variable_in_reg(Var, VarCode, VarLval),
+	{ VarRval = lval(VarLval) },
 	code_info__acquire_reg(r, PtagReg),
 	code_info__release_reg(PtagReg),
 	{
@@ -273,7 +261,7 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 
 	(
 		{ PrimaryMethod = binary_search },
-		{ tag_switch__order_ptags_by_value(0, MaxPrimary, PtagCaseMap,
+		{ switch_util__order_ptags_by_value(0, MaxPrimary, PtagCaseMap,
 			PtagCaseList) },
 		tag_switch__generate_primary_binary_search(PtagCaseList,
 			0, MaxPrimary, PtagRval, VarRval, CodeModel, CanFail,
@@ -281,7 +269,7 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 			no, MaybeEnd, CasesCode)
 	;
 		{ PrimaryMethod = jump_table },
-		{ tag_switch__order_ptags_by_value(0, MaxPrimary, PtagCaseMap,
+		{ switch_util__order_ptags_by_value(0, MaxPrimary, PtagCaseMap,
 			PtagCaseList) },
 		tag_switch__generate_primary_jump_table(PtagCaseList,
 			0, MaxPrimary, VarRval, CodeModel, StoreMap,
@@ -294,7 +282,7 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 		{ CasesCode = tree(SwitchCode, TableCode) }
 	;
 		{ PrimaryMethod = try_chain },
-		{ tag_switch__order_ptags_by_count(PtagCountList, PtagCaseMap,
+		{ switch_util__order_ptags_by_count(PtagCountList, PtagCaseMap,
 			PtagCaseList0) },
 		{
 			CanFail = cannot_fail,
@@ -310,7 +298,7 @@ tag_switch__generate(Cases, Var, CodeModel, CanFail, StoreMap, EndLabel,
 			MaybeEnd0, MaybeEnd, CasesCode)
 	;
 		{ PrimaryMethod = try_me_else_chain },
-		{ tag_switch__order_ptags_by_count(PtagCountList, PtagCaseMap,
+		{ switch_util__order_ptags_by_count(PtagCountList, PtagCaseMap,
 			PtagCaseList) },
 		tag_switch__generate_primary_try_me_else_chain(PtagCaseList,
 			PtagRval, VarRval, CodeModel, CanFail, StoreMap,
@@ -1110,247 +1098,3 @@ tag_switch__generate_secondary_binary_search(StagGoals, MinStag, MaxStag,
 	).
 
 %-----------------------------------------------------------------------------%
-
-	% Find out how many secondary tags share each primary tag
-	% of the given variable.
-
-:- pred tag_switch__get_ptag_counts(type, module_info, int, ptag_count_map).
-:- mode tag_switch__get_ptag_counts(in, in, out, out) is det.
-
-tag_switch__get_ptag_counts(Type, ModuleInfo, MaxPrimary, PtagCountMap) :-
-	( type_to_type_id(Type, TypeIdPrime, _) ->
-		TypeId = TypeIdPrime
-	;
-		error("unknown type in tag_switch__get_ptag_counts")
-	),
-	module_info_types(ModuleInfo, TypeTable),
-	map__lookup(TypeTable, TypeId, TypeDefn),
-	hlds_data__get_type_defn_body(TypeDefn, Body),
-	( Body = du_type(_, ConsTable, _, _) ->
-		map__to_assoc_list(ConsTable, ConsList),
-		tag_switch__cons_list_to_tag_list(ConsList, TagList)
-	;
-		error("non-du type in tag_switch__get_ptag_counts")
-	),
-	map__init(PtagCountMap0),
-	tag_switch__get_ptag_counts_2(TagList, -1, MaxPrimary,
-		PtagCountMap0, PtagCountMap).
-
-:- pred tag_switch__get_ptag_counts_2(list(cons_tag), int, int,
-	ptag_count_map, ptag_count_map).
-:- mode tag_switch__get_ptag_counts_2(in, in, out, in, out) is det.
-
-tag_switch__get_ptag_counts_2([], Max, Max, PtagCountMap, PtagCountMap).
-tag_switch__get_ptag_counts_2([ConsTag | TagList], MaxPrimary0, MaxPrimary,
-		PtagCountMap0, PtagCountMap) :-
-	( ConsTag = unshared_tag(Primary) ->
-		int__max(MaxPrimary0, Primary, MaxPrimary1),
-		( map__search(PtagCountMap0, Primary, _) ->
-			error("unshared tag is shared")
-		;
-			map__det_insert(PtagCountMap0, Primary, none - (-1),
-				PtagCountMap1)
-		)
-	; ConsTag = shared_remote_tag(Primary, Secondary) ->
-		int__max(MaxPrimary0, Primary, MaxPrimary1),
-		( map__search(PtagCountMap0, Primary, Target) ->
-			Target = TagType - MaxSoFar,
-			( TagType = remote ->
-				true
-			;
-				error("remote tag is shared with non-remote")
-			),
-			int__max(Secondary, MaxSoFar, Max),
-			map__det_update(PtagCountMap0, Primary, remote - Max,
-				PtagCountMap1)
-		;
-			map__det_insert(PtagCountMap0, Primary,
-				remote - Secondary, PtagCountMap1)
-		)
-	; ConsTag = shared_local_tag(Primary, Secondary) ->
-		int__max(MaxPrimary0, Primary, MaxPrimary1),
-		( map__search(PtagCountMap0, Primary, Target) ->
-			Target = TagType - MaxSoFar,
-			( TagType = local ->
-				true
-			;
-				error("local tag is shared with non-local")
-			),
-			int__max(Secondary, MaxSoFar, Max),
-			map__det_update(PtagCountMap0, Primary, local - Max,
-				PtagCountMap1)
-		;
-			map__det_insert(PtagCountMap0, Primary,
-				local - Secondary, PtagCountMap1)
-		)
-	;
-		error("non-du tag in tag_switch__get_ptag_counts_2")
-	),
-	tag_switch__get_ptag_counts_2(TagList, MaxPrimary1, MaxPrimary,
-		PtagCountMap1, PtagCountMap).
-
-%-----------------------------------------------------------------------------%
-
-	% Group together all the cases that depend on the given variable
-	% having the same primary tag value.
-
-:- pred tag_switch__group_cases_by_ptag(cases_list,
-	ptag_case_map, ptag_case_map).
-:- mode tag_switch__group_cases_by_ptag(in, in, out) is det.
-
-tag_switch__group_cases_by_ptag([], PtagCaseMap, PtagCaseMap).
-tag_switch__group_cases_by_ptag([Case0 | Cases0], PtagCaseMap0, PtagCaseMap) :-
-	Case0 = case(_Priority, Tag, _ConsId, Goal),
-	( Tag = unshared_tag(Primary) ->
-		( map__search(PtagCaseMap0, Primary, _Group) ->
-			error("unshared tag is shared")
-		;
-			map__init(StagGoalMap0),
-			map__det_insert(StagGoalMap0, -1, Goal, StagGoalMap),
-			map__det_insert(PtagCaseMap0, Primary,
-				none - StagGoalMap, PtagCaseMap1)
-		)
-	; Tag = shared_remote_tag(Primary, Secondary) ->
-		( map__search(PtagCaseMap0, Primary, Group) ->
-			Group = StagLoc - StagGoalMap0,
-			( StagLoc = remote ->
-				true
-			;
-				error("remote tag is shared with non-remote")
-			),
-			map__det_insert(StagGoalMap0, Secondary, Goal,
-				StagGoalMap),
-			map__det_update(PtagCaseMap0, Primary,
-				remote - StagGoalMap, PtagCaseMap1)
-		;
-			map__init(StagGoalMap0),
-			map__det_insert(StagGoalMap0, Secondary, Goal,
-				StagGoalMap),
-			map__det_insert(PtagCaseMap0, Primary,
-				remote - StagGoalMap, PtagCaseMap1)
-		)
-	; Tag = shared_local_tag(Primary, Secondary) ->
-		( map__search(PtagCaseMap0, Primary, Group) ->
-			Group = StagLoc - StagGoalMap0,
-			( StagLoc = local ->
-				true
-			;
-				error("local tag is shared with non-local")
-			),
-			map__det_insert(StagGoalMap0, Secondary, Goal,
-				StagGoalMap),
-			map__det_update(PtagCaseMap0, Primary,
-				local - StagGoalMap, PtagCaseMap1)
-		;
-			map__init(StagGoalMap0),
-			map__det_insert(StagGoalMap0, Secondary, Goal,
-				StagGoalMap),
-			map__det_insert(PtagCaseMap0, Primary,
-				local - StagGoalMap, PtagCaseMap1)
-		)
-	;
-		error("non-du tag in tag_switch__group_cases_by_ptag")
-	),
-	tag_switch__group_cases_by_ptag(Cases0, PtagCaseMap1, PtagCaseMap).
-
-%-----------------------------------------------------------------------------%
-
-	% Order the primary tags based on the number of secondary tags
-	% associated with them, putting the ones with the most secondary tags
-	% first. We use selection sort.
-	% Note that it is not an error for a primary tag to have no case list;
-	% this can happen in semideterministic switches, or in det switches
-	% where the initial inst of the switch variable is a bound(...) inst
-	% representing a subtype.
-
-:- pred tag_switch__order_ptags_by_count(ptag_count_list, ptag_case_map,
-	ptag_case_list).
-:- mode tag_switch__order_ptags_by_count(in, in, out) is det.
-
-tag_switch__order_ptags_by_count(PtagCountList0, PtagCaseMap0, PtagCaseList) :-
-	(
-		tag_switch__select_frequent_ptag(PtagCountList0,
-			Primary, _, PtagCountList1)
-	->
-		( map__search(PtagCaseMap0, Primary, PtagCase) ->
-			map__delete(PtagCaseMap0, Primary, PtagCaseMap1),
-			tag_switch__order_ptags_by_count(PtagCountList1,
-				PtagCaseMap1, PtagCaseList1),
-			PtagCaseList = [Primary - PtagCase | PtagCaseList1]
-		;
-			tag_switch__order_ptags_by_count(PtagCountList1,
-				PtagCaseMap0, PtagCaseList)
-		)
-	;
-		( map__is_empty(PtagCaseMap0) ->
-			PtagCaseList = []
-		;
-			error("PtagCaseMap0 is not empty in tag_switch__order_ptags_by_count")
-		)
-	).
-
-	% Select the most frequently used primary tag based on the number of
-	% secondary tags associated with it.
-
-:- pred tag_switch__select_frequent_ptag(ptag_count_list, tag_bits, int,
-	ptag_count_list).
-:- mode tag_switch__select_frequent_ptag(in, out, out, out) is semidet.
-
-tag_switch__select_frequent_ptag([PtagCount0 | PtagCountList1], Primary, Count,
-		PtagCountList) :-
-	PtagCount0 = Primary0 - (_ - Count0),
-	(
-		tag_switch__select_frequent_ptag(PtagCountList1,
-			Primary1, Count1, PtagCountList2),
-		Count1 > Count0
-	->
-		Primary = Primary1,
-		Count = Count1,
-		PtagCountList = [PtagCount0 | PtagCountList2]
-	;
-		Primary = Primary0,
-		Count = Count0,
-		PtagCountList = PtagCountList1
-	).
-
-%-----------------------------------------------------------------------------%
-
-	% Order the primary tags based on their value, lowest value first.
-	% We scan through the primary tags values from zero to maximum.
-	% Note that it is not an error for a primary tag to have no case list,
-	% since this can happen in semideterministic switches.
-
-:- pred tag_switch__order_ptags_by_value(int, int,
-	ptag_case_map, ptag_case_list).
-:- mode tag_switch__order_ptags_by_value(in, in, in, out) is det.
-
-tag_switch__order_ptags_by_value(Ptag, MaxPtag, PtagCaseMap0, PtagCaseList) :-
-	( MaxPtag >= Ptag ->
-		NextPtag is Ptag + 1,
-		( map__search(PtagCaseMap0, Ptag, PtagCase) ->
-			map__delete(PtagCaseMap0, Ptag, PtagCaseMap1),
-			tag_switch__order_ptags_by_value(NextPtag, MaxPtag,
-				PtagCaseMap1, PtagCaseList1),
-			PtagCaseList = [Ptag - PtagCase | PtagCaseList1]
-		;
-			tag_switch__order_ptags_by_value(NextPtag, MaxPtag,
-				PtagCaseMap0, PtagCaseList)
-		)
-	;
-		( map__is_empty(PtagCaseMap0) ->
-			PtagCaseList = []
-		;
-			error("PtagCaseMap0 is not empty in order_ptags_by_value")
-		)
-	).
-
-%-----------------------------------------------------------------------------%
-
-:- pred tag_switch__cons_list_to_tag_list(assoc_list(cons_id, cons_tag),
-	list(cons_tag)).
-:- mode tag_switch__cons_list_to_tag_list(in, out) is det.
-
-tag_switch__cons_list_to_tag_list([], []).
-tag_switch__cons_list_to_tag_list([_ConsId - ConsTag | ConsList],
-		[ConsTag | Tagslist]) :-
-	tag_switch__cons_list_to_tag_list(ConsList, Tagslist).
