@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1997-1999 The University of Melbourne.
+% Copyright (C) 1997-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -130,7 +130,7 @@ traverse_goal_2(unify(_Var, _RHS, _UniMode, Unification, _Context),
 			Info = Info0
 		)
 	;
-		Unification = deconstruct(InVar, ConsId, Args, Modes, _),
+		Unification = deconstruct(InVar, ConsId, Args, Modes, _, _),
 		(
 			unify_change(InVar, ConsId, Args, Modes, Params,
 				Gamma0, InVars0, OutVars)
@@ -183,7 +183,7 @@ traverse_goal_2(if_then_else(_, Cond, Then, Else, _), _, Params, Info0, Info) :-
 	traverse_goal(Else, Params, Info0, Info2),
 	combine_paths(Info1, Info2, Params, Info).
 
-traverse_goal_2(pragma_c_code(_, CallPredId, CallProcId, Args, _, _, _),
+traverse_goal_2(pragma_foreign_code(_, CallPredId, CallProcId, Args, _,_,_),
 		GoalInfo, Params, Info0, Info) :-
 	params_get_module_info(Params, Module),
 	module_info_pred_proc_info(Module, CallPredId, CallProcId, _,
@@ -191,7 +191,7 @@ traverse_goal_2(pragma_c_code(_, CallPredId, CallProcId, Args, _, _, _),
 	proc_info_argmodes(CallProcInfo, CallArgModes),
 	partition_call_args(Module, CallArgModes, Args, _InVars, OutVars),
 	goal_info_get_context(GoalInfo, Context),
-	error_if_intersect(OutVars, Context, pragma_c_code, Info0, Info).
+	error_if_intersect(OutVars, Context, pragma_foreign_code, Info0, Info).
 
 traverse_goal_2(generic_call(_, _, _, _), GoalInfo, Params, Info0, Info) :-
 	%
