@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1993-1997 The University of Melbourne.
+% Copyright (C) 1993-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -23,16 +23,29 @@
 			;	node(T)
 			;	tree(tree(T), tree(T)).
 
+:- func tree__flatten(tree(T)) =  list(T).
+
+	% Make a tree from a list of trees.
+:- func tree__list(list(tree(T))) = tree(T).
+
 :- pred tree__flatten(tree(T), list(T)).
 :- mode tree__flatten(in, out) is det.
 
 :- pred tree__is_empty(tree(T)).
 :- mode tree__is_empty(in) is semidet.
 
+:- pred tree__tree_of_lists_is_empty(tree(list(T))).
+:- mode tree__tree_of_lists_is_empty(in) is semidet.
+
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
+tree__flatten(T) = L :- tree__flatten(T, L).
+
+tree__list([]) = empty.
+tree__list([X | Xs]) = tree(X, tree__list(Xs)).
+
 tree__flatten(T, L) :-
 	tree__flatten_2(T, [], L).
 
@@ -54,3 +67,11 @@ tree__is_empty(tree(L, R)) :-
 	tree__is_empty(R).
 
 %-----------------------------------------------------------------------------%
+
+tree__tree_of_lists_is_empty(empty).
+tree__tree_of_lists_is_empty(node([])).
+tree__tree_of_lists_is_empty(tree(L, R)) :-
+	tree__tree_of_lists_is_empty(L),
+	tree__tree_of_lists_is_empty(R).
+
+%-----------------------------------------------------------------------------%
