@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1994-1999 The University of Melbourne.
+% Copyright (C) 1994-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -20,7 +20,7 @@
 
 :- import_module hlds_module, hlds_pred, hlds_data, prog_data.
 :- import_module term.
-:- import_module list, map.
+:- import_module std_util, list, map.
 
 %-----------------------------------------------------------------------------%
 
@@ -30,6 +30,9 @@
 :- pred type_is_atomic(type, module_info).
 :- mode type_is_atomic(in, in) is semidet.
 
+:- pred type_id_is_atomic(type_id, module_info).
+:- mode type_id_is_atomic(in, in) is semidet.
+
 	% type_is_higher_order(Type, PredOrFunc, ArgTypes) succeeds iff
 	% Type is a higher-order predicate or function type with the specified
 	% argument types (for functions, the return type is appended to the
@@ -39,12 +42,20 @@
 		lambda_eval_method, list(type)).
 :- mode type_is_higher_order(in, out, out, out) is semidet.
 
+	% Succeed if the given type is a tuple type, returning
+	% the argument types.
+:- pred type_is_tuple(type, list(type)).
+:- mode type_is_tuple(in, out) is semidet.
+
 	% type_id_is_higher_order(TypeId, PredOrFunc) succeeds iff
 	% TypeId is a higher-order predicate or function type.
-
 :- pred type_id_is_higher_order(type_id, pred_or_func, lambda_eval_method).
 :- mode type_id_is_higher_order(in, out, out) is semidet.
 
+	% type_id_is_tuple(TypeId) succeeds iff TypeId is a tuple type.
+:- pred type_id_is_tuple(type_id).
+:- mode type_id_is_tuple(in) is semidet.
+
 	% return true iff there was a `where equality is <predname>'
 	% declaration for the specified type, and return the name of
 	% the equality predicate and the context of the type declaration.
@@ -59,6 +70,9 @@
 :- pred type_util__is_dummy_argument_type(type).
 :- mode type_util__is_dummy_argument_type(in) is semidet.
 
+:- pred type_is_io_state(type).
+:- mode type_is_io_state(in) is semidet.
+
 :- pred type_is_aditi_state(type).
 :- mode type_is_aditi_state(in) is semidet.
 
@@ -66,12 +80,13 @@
 :- pred type_util__remove_aditi_state(list(type), list(T), list(T)).
 :- mode type_util__remove_aditi_state(in, in, out) is det.
 
-	% A test for types that are defined by hand (not including
-	% the builtin types).  Don't generate type_ctor_*
-	% for these types.
+	% A test for types that are defined in Mercury, but whose definitions
+	% are `lies', i.e. they are not sufficiently accurate for RTTI
+	% structures describing the types. Since the RTTI will be hand defined,
+	% the compiler shouldn't generate RTTI for these types.
 
-:- pred type_id_is_hand_defined(type_id).
-:- mode type_id_is_hand_defined(in) is semidet.
+:- pred type_id_has_hand_defined_rtti(type_id).
+:- mode type_id_has_hand_defined_rtti(in) is semidet.
 
 	% A test for type_info-related types that are introduced by
 	% polymorphism.m.  Mode inference never infers unique modes
@@ -92,15 +107,19 @@
 :- mode remove_new_prefix(out, in) is det.
 
 	% Given a type, determine what sort of type it is.
-
 :- pred classify_type(type, module_info, builtin_type).
 :- mode classify_type(in, in, out) is det.
 
+	% Given a type_id, determine what sort of type it is.
+:- pred classify_type_id(module_info, type_id, builtin_type).
+:- mode classify_type_id(in, in, out) is det.
+
 :- type builtin_type	--->	int_type
 			;	char_type
 			;	str_type
 			;	float_type
 			;	pred_type
+			;	tuple_type
 			;	enum_type
 			;	polymorphic_type
 			;	user_type.
@@ -160,16 +179,46 @@
 :- pred type_util__type_id_arity(module_info, type_id, arity).
 :- mode type_util__type_id_arity(in, in, out) is det.
 
-	% If the type is a du type, return the list of its constructors.
-
+	% If the type is a du type or a tuple type,
+	% return the list of its constructors.
 :- pred type_constructors(type, module_info, list(constructor)).
 :- mode type_constructors(in, in, out) is semidet.
 
+	% Given a type on which it is possible to have a complete switch,
+	% return the number of alternatives. (It is possible to have a complete
+	% switch on any du type and on the builtin type character. It is not
+	% feasible to have a complete switch on the builtin types integer,
+	% float, and switch. One cannot have a switch on an abstract type,
+	% and equivalence types will have been expanded out by the time
+	% we consider switches.)
+:- pred type_util__switch_type_num_functors(module_info::in, (type)::in,
+	int::out) is semidet.
+
 	% Work out the types of the arguments of a functor.
 	% Aborts if the functor is existentially typed.
+	% The cons_id is expected to be un-module-qualified.
 :- pred type_util__get_cons_id_arg_types(module_info::in, (type)::in,
 		cons_id::in, list(type)::out) is det.
 
+	% The same as type_util__get_cons_id_arg_types except that it
+	% fails rather than aborting if the functor is existentially
+	% typed.
+	% The cons_id is expected to be un-module-qualified.
+:- pred type_util__get_cons_id_non_existential_arg_types(module_info::in,
+		(type)::in, cons_id::in, list(type)::out) is semidet.
+
+	% The same as type_util__get_cons_id_arg_types except that the
+	% cons_id is output non-deterministically.
+	% The cons_id is not module-qualified.
+:- pred type_util__cons_id_arg_types(module_info::in, (type)::in,
+		cons_id::out, list(type)::out) is nondet.
+
+	% Given a type and a cons_id, look up the definitions of that
+	% type and constructor. Aborts if the cons_id is not user-defined.
+:- pred type_util__get_type_and_cons_defn(module_info, (type), cons_id,
+		hlds_type_defn, hlds_cons_defn).
+:- mode type_util__get_type_and_cons_defn(in, in, in, out, out) is det.
+
 	% Given a type and a cons_id, look up the definition of that
 	% constructor; if it is existentially typed, return its definition,
 	% otherwise fail.
@@ -191,15 +240,30 @@
 			(type)		% functor result type
 		).
 
-	% Given a list of constructors for a type,
-	% check whether that type is a no_tag type
+	% Check whether a type is a no_tag type
 	% (i.e. one with only one constructor, and
 	% whose one constructor has only one argument,
 	% and which is not private_builtin:type_info/1),
 	% and if so, return its constructor symbol and argument type.
 
-:- pred type_is_no_tag_type(list(constructor), sym_name, type).
-:- mode type_is_no_tag_type(in, out, out) is semidet.
+:- pred type_is_no_tag_type(module_info, type, sym_name, type).
+:- mode type_is_no_tag_type(in, in, out, out) is semidet.
+
+	% Check whether some constructors are a no_tag type
+	% (i.e. one with only one constructor, and
+	% whose one constructor has only one argument,
+	% and which is not private_builtin:type_info/1),
+	% and if so, return its constructor symbol, argument type,
+	% and the argument's name (if it has one).
+	%
+	% This doesn't do any checks for options that might be set
+	% (such as turning off no_tag_types).  If you want those checks
+	% you should use type_is_no_tag_type/4, or if you really know
+	% what you are doing, perform the checks yourself.
+
+:- pred type_constructors_are_no_tag_type(list(constructor), sym_name, type,
+	maybe(string)).
+:- mode type_constructors_are_no_tag_type(in, out, out, out) is semidet.
 
 	% Unify (with occurs check) two types with respect to a type
 	% substitution and update the type bindings.
@@ -217,6 +281,13 @@
 :- pred type_util__vars(type, list(tvar)).
 :- mode type_util__vars(in, out) is det.
 
+	% Return a list of the type variables of a type,
+	% ignoring any type variables if the variable in
+	% question is a type-info
+
+:- pred type_util__real_vars(type, list(tvar)).
+:- mode type_util__real_vars(in, out) is det.
+
 	% type_list_subsumes(TypesA, TypesB, Subst) succeeds iff the list
 	% TypesA subsumes (is more general than) TypesB, producing a
 	% type substitution which when applied to TypesA will give TypesB.
@@ -224,6 +295,18 @@
 :- pred type_list_subsumes(list(type), list(type), tsubst).
 :- mode type_list_subsumes(in, in, out) is semidet.
 
+	% arg_type_list_subsumes(TVarSet, ArgTypes,
+	%       CalleeTVarSet, CalleeExistQVars, CalleeArgTypes).
+	%
+	% Check that the argument types of the called predicate,
+	% function or constructor subsume the types of the
+	% arguments of the call. This checks that none
+	% of the existentially quantified type variables of
+	% the callee are bound.
+:- pred arg_type_list_subsumes(tvarset, list(type),
+		tvarset, existq_tvars, list(type)).
+:- mode arg_type_list_subsumes(in, in, in, in, in) is semidet.
+
 	% apply a type substitution (i.e. map from tvar -> type)
 	% to all the types in a variable typing (i.e. map from var -> type).
 
@@ -295,6 +378,10 @@
 	map(class_constraint, constraint_proof)).
 :- mode apply_rec_subst_to_constraint_proofs(in, in, out) is det.
 
+:- pred apply_variable_renaming_to_type_map(map(tvar, tvar),
+		vartypes, vartypes).
+:- mode apply_variable_renaming_to_type_map(in, in, out) is det.
+
 :- pred apply_variable_renaming_to_constraints(map(tvar, tvar), 
 	class_constraints, class_constraints).
 :- mode apply_variable_renaming_to_constraints(in, in, out) is det.
@@ -338,12 +425,17 @@
 :- pred constraint_get_tvars(class_constraint, list(tvar)).
 :- mode constraint_get_tvars(in, out) is det.
 
+:- pred get_unconstrained_tvars(list(tvar), list(class_constraint), list(tvar)).
+:- mode get_unconstrained_tvars(in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module bool, int, require, std_util, string.
-:- import_module prog_io, prog_io_goal, prog_util.
+
+:- import_module prog_io, prog_io_goal, prog_util, options, globals.
+:- import_module bool, char, int, string.
+:- import_module assoc_list, require, varset.
 
 type_util__type_id_module(_ModuleInfo, TypeName - _Arity, ModuleName) :-
 	sym_name_get_module_name(TypeName, unqualified(""), ModuleName).
@@ -354,25 +446,26 @@ type_util__type_id_name(_ModuleInfo, Name0 - _Arity, Name) :-
 type_util__type_id_arity(_ModuleInfo, _Name - Arity, Arity).
 
 type_is_atomic(Type, ModuleInfo) :-
-	classify_type(Type, ModuleInfo, BuiltinType),
+	type_to_type_id(Type, TypeId, _),
+	type_id_is_atomic(TypeId, ModuleInfo).
+
+type_id_is_atomic(TypeId, ModuleInfo) :-
+	classify_type_id(ModuleInfo, TypeId, BuiltinType),
 	BuiltinType \= polymorphic_type,
+	BuiltinType \= tuple_type,
 	BuiltinType \= pred_type,
 	BuiltinType \= user_type.
 
 type_util__var(term__variable(Var), Var).
 
-type_id_is_hand_defined(qualified(unqualified("builtin"), "c_pointer") - 0).
-type_id_is_hand_defined(qualified(unqualified("std_util"), "univ") - 0).
-type_id_is_hand_defined(qualified(unqualified("std_util"), "type_info") - 0).
-type_id_is_hand_defined(qualified(unqualified("array"), "array") - 1).
-type_id_is_hand_defined(qualified(PrivateBuiltin, "type_info") - 1) :-
-	mercury_private_builtin_module(PrivateBuiltin).
-type_id_is_hand_defined(qualified(PrivateBuiltin, "type_ctor_info") - 1) :-
-	mercury_private_builtin_module(PrivateBuiltin).
-type_id_is_hand_defined(qualified(PrivateBuiltin, "typeclass_info") - 1) :-
-	mercury_private_builtin_module(PrivateBuiltin).
-type_id_is_hand_defined(qualified(PrivateBuiltin, "base_typeclass_info") - 1) :-
-	mercury_private_builtin_module(PrivateBuiltin).
+type_id_has_hand_defined_rtti(qualified(PB, "type_info") - 1) :-
+	mercury_private_builtin_module(PB).
+type_id_has_hand_defined_rtti(qualified(PB, "type_ctor_info") - 1) :-
+	mercury_private_builtin_module(PB).
+type_id_has_hand_defined_rtti(qualified(PB, "typeclass_info") - 1) :-
+	mercury_private_builtin_module(PB).
+type_id_has_hand_defined_rtti(qualified(PB, "base_typeclass_info") - 1) :-
+	mercury_private_builtin_module(PB).
 
 is_introduced_type_info_type(Type) :-
 	sym_name_and_args(Type, TypeName, _),
@@ -395,63 +488,69 @@ remove_new_prefix(qualified(Module, Name0), qualified(Module, Name)) :-
 
 classify_type(VarType, ModuleInfo, Type) :-
 	( type_to_type_id(VarType, TypeId, _) ->
-		( TypeId = unqualified("character") - 0 ->
-			Type = char_type
-		; TypeId = unqualified("int") - 0 ->
-			Type = int_type
-		; TypeId = unqualified("float") - 0 ->
-			Type = float_type
-		; TypeId = unqualified("string") - 0 ->
-			Type = str_type
-		; type_id_is_higher_order(TypeId, _, _) ->
-			Type = pred_type
-		; type_id_is_enumeration(TypeId, ModuleInfo) ->
-			Type = enum_type
-		;
-			Type = user_type
-		)
+		classify_type_id(ModuleInfo, TypeId, Type)
 	;
 		Type = polymorphic_type
 	).
 
+classify_type_id(ModuleInfo, TypeId, Type) :-
+	( TypeId = unqualified("character") - 0 ->
+		Type = char_type
+	; TypeId = unqualified("int") - 0 ->
+		Type = int_type
+	; TypeId = unqualified("float") - 0 ->
+		Type = float_type
+	; TypeId = unqualified("string") - 0 ->
+		Type = str_type
+	; type_id_is_higher_order(TypeId, _, _) ->
+		Type = pred_type
+	; type_id_is_tuple(TypeId) ->
+		Type = tuple_type
+	; type_id_is_enumeration(TypeId, ModuleInfo) ->
+		Type = enum_type
+	;
+		Type = user_type
+	).
+
 type_is_higher_order(Type, PredOrFunc, EvalMethod, PredArgTypes) :-
 	(
 		Type = term__functor(term__atom("="),
 			[FuncEvalAndArgs, FuncRetType], _)
 	->
-		get_lambda_eval_method(FuncEvalAndArgs, EvalMethod,
-			FuncAndArgs),
-		FuncAndArgs = term__functor(term__atom("func"),
-			FuncArgTypes, _),
+		get_lambda_eval_method_and_args("func", FuncEvalAndArgs,
+			EvalMethod, FuncArgTypes),
 		list__append(FuncArgTypes, [FuncRetType], PredArgTypes),
 		PredOrFunc = function
 	;
-		get_lambda_eval_method(Type, EvalMethod, PredAndArgs),
-		PredAndArgs = term__functor(term__atom("pred"),
-					PredArgTypes, _),
+		get_lambda_eval_method_and_args("pred",
+			Type, EvalMethod, PredArgTypes),
 		PredOrFunc = predicate
 	).
 
-	% From the type of a lambda expression, work out how it should
-	% be evaluated.
-:- pred get_lambda_eval_method((type), lambda_eval_method, (type)) is det.
-:- mode get_lambda_eval_method(in, out, out) is det.
+type_is_tuple(Type, ArgTypes) :-
+	type_to_type_id(Type, TypeId, ArgTypes),
+	type_id_is_tuple(TypeId).
 
-get_lambda_eval_method(Type0, EvalMethod, Type) :-
-	( Type0 = term__functor(term__atom(MethodStr), [Type1], _) ->
-		( MethodStr = "aditi_bottom_up" ->
-			EvalMethod = (aditi_bottom_up),
-			Type = Type1
-		; MethodStr = "aditi_top_down" ->
-			EvalMethod = (aditi_top_down),
-			Type = Type1
-		;
-			EvalMethod = normal,
-			Type = Type0
-		)
-	;
+	% From the type of a lambda expression, work out how it should
+	% be evaluated and extract the argument types.
+:- pred get_lambda_eval_method_and_args(string, (type),
+		lambda_eval_method, list(type)) is det.
+:- mode get_lambda_eval_method_and_args(in, in, out, out) is semidet.
+
+get_lambda_eval_method_and_args(PorFStr, Type0, EvalMethod, ArgTypes) :-
+	Type0 = term__functor(term__atom(Functor), Args, _),
+	( Functor = PorFStr ->
 		EvalMethod = normal,
-		Type = Type0
+		ArgTypes = Args
+	;	
+		Args = [Type1],
+		Type1 = term__functor(term__atom(PorFStr), ArgTypes, _),
+		( Functor = "aditi_bottom_up" ->
+			EvalMethod = (aditi_bottom_up)
+		;
+			Functor = "aditi_top_down",
+			EvalMethod = (aditi_top_down)
+		)
 	).
 
 type_id_is_higher_order(SymName - _Arity, PredOrFunc, EvalMethod) :-
@@ -476,6 +575,8 @@ type_id_is_higher_order(SymName - _Arity, PredOrFunc, EvalMethod) :-
 		PredOrFunc = function
 	).
 
+type_id_is_tuple(unqualified("{}") - _).
+
 type_has_user_defined_equality_pred(ModuleInfo, Type, SymName) :-
 	module_info_types(ModuleInfo, TypeTable),
 	type_to_type_id(Type, TypeId, _TypeArgs),
@@ -503,6 +604,10 @@ type_util__is_dummy_argument_type(Type) :-
 type_util__is_dummy_argument_type_2("io", "state", 0).	 % io:state/0
 type_util__is_dummy_argument_type_2("store", "store", 1). % store:store/1.
 
+type_is_io_state(Type) :-
+        type_to_type_id(Type,
+		qualified(unqualified("io"), "state") - 0, []).
+
 type_is_aditi_state(Type) :-
         type_to_type_id(Type,
 		qualified(unqualified("aditi"), "state") - 0, []).
@@ -531,15 +636,7 @@ type_id_is_enumeration(TypeId, ModuleInfo) :-
 	IsEnum = yes.
 
 type_to_type_id(Type, SymName - Arity, Args) :-
-	sym_name_and_args(Type, SymName0, Args1),
-
-	% `private_builtin:constraint' is introduced by polymorphism, and
-	% should only appear as the argument of a `typeclass:info/1' type.
-	% It behaves sort of like a type variable, so according to the
-	% specification of `type_to_type_id', it should cause failure.
-	% There isn't a definition in the type table.
-	mercury_private_builtin_module(PrivateBuiltin),
-	SymName \= qualified(PrivateBuiltin, "constraint"),
+	Type \= term__variable(_),
 
 	% higher order types may have representations where
 	% their arguments don't directly correspond to the
@@ -566,14 +663,25 @@ type_to_type_id(Type, SymName - Arity, Args) :-
 			EvalMethod = (aditi_top_down),
 			SymName = qualified(unqualified("aditi_top_down"),
 					PorFStr)
-			
 		;
 			EvalMethod = normal,
 			SymName = unqualified(PorFStr)
 		)
 	;
-		SymName = SymName0,
-		Args = Args1,
+		sym_name_and_args(Type, SymName, Args),
+
+		% `private_builtin:constraint' is introduced by polymorphism,
+		% and should only appear as the argument of a
+		% `typeclass:info/1' type.
+		% It behaves sort of like a type variable, so according to the
+		% specification of `type_to_type_id', it should cause failure.
+		% There isn't a definition in the type table.
+		\+ (
+			SymName = qualified(ModuleName, UnqualName),
+			UnqualName = "constraint",
+			mercury_private_builtin_module(PrivateBuiltin),
+			ModuleName = PrivateBuiltin	
+		),
 		list__length(Args, Arity)
 	).
 
@@ -644,44 +752,134 @@ make_type_id(term__atom(Name), Arity, unqualified(Name) - Arity).
 
 type_constructors(Type, ModuleInfo, Constructors) :-
 	type_to_type_id(Type, TypeId, TypeArgs),
-	module_info_types(ModuleInfo, TypeTable),
-	map__search(TypeTable, TypeId, TypeDefn),
-	hlds_data__get_type_defn_tparams(TypeDefn, TypeParams),
-	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(Constructors0, _, _, _),
-	substitute_type_args(TypeParams, TypeArgs, Constructors0,
-		Constructors).
+	( type_id_is_tuple(TypeId) ->
+		% Tuples are never existentially typed.
+		ExistQVars = [],	
+		ClassConstraints = [],
+		CtorArgs = list__map((func(ArgType) = no - ArgType), TypeArgs),
+		Constructors = [ctor(ExistQVars, ClassConstraints,
+				unqualified("{}"), CtorArgs)]
+	;
+		module_info_types(ModuleInfo, TypeTable),
+		map__search(TypeTable, TypeId, TypeDefn),
+		hlds_data__get_type_defn_tparams(TypeDefn, TypeParams),
+		hlds_data__get_type_defn_body(TypeDefn, TypeBody),
+		TypeBody = du_type(Constructors0, _, _, _),
+		substitute_type_args(TypeParams, TypeArgs, Constructors0,
+			Constructors)
+	).
 
 %-----------------------------------------------------------------------------%
 
-type_util__get_cons_id_arg_types(ModuleInfo, VarType, ConsId, ArgTypes) :-
+type_util__switch_type_num_functors(ModuleInfo, Type, NumFunctors) :-
+	type_to_type_id(Type, TypeId, _),
+	( TypeId = unqualified("character") - 0 ->
+		% XXX the following code uses the source machine's character
+		% size, not the target's, so it won't work if cross-compiling
+		% to a machine with a different size character.
+		char__max_char_value(MaxChar),
+		char__min_char_value(MinChar),
+		NumFunctors is MaxChar - MinChar + 1
+	; type_id_is_tuple(TypeId) ->
+		NumFunctors = 1
+	;
+		module_info_types(ModuleInfo, TypeTable),
+		map__search(TypeTable, TypeId, TypeDefn),
+		hlds_data__get_type_defn_body(TypeDefn, TypeBody),
+		TypeBody = du_type(_, ConsTable, _, _),
+		map__count(ConsTable, NumFunctors)
+	).
+
+%-----------------------------------------------------------------------------%
+
+type_util__get_cons_id_arg_types(ModuleInfo, Type, ConsId, ArgTypes) :-
+	type_util__get_cons_id_arg_types_2(abort_on_exist_qvar,
+		ModuleInfo, Type, ConsId, ArgTypes).
+
+type_util__get_cons_id_non_existential_arg_types(ModuleInfo, Type, ConsId,
+		ArgTypes) :-
+	type_util__get_cons_id_arg_types_2(fail_on_exist_qvar,
+		ModuleInfo, Type, ConsId, ArgTypes).
+
+:- type exist_qvar_action
+	--->	fail_on_exist_qvar
+	;	abort_on_exist_qvar.
+
+:- pred type_util__get_cons_id_arg_types_2(exist_qvar_action,
+	module_info, (type), cons_id, list(type)).
+:- mode type_util__get_cons_id_arg_types_2(in(bound(fail_on_exist_qvar)),
+		in, in, in, out) is semidet.
+:- mode type_util__get_cons_id_arg_types_2(in(bound(abort_on_exist_qvar)),
+		in, in, in, out) is det.
+
+type_util__get_cons_id_arg_types_2(EQVarAction, ModuleInfo, VarType, ConsId,
+		ArgTypes) :-
+    (
+	type_to_type_id(VarType, TypeId, TypeArgs)
+    ->
 	(
-		type_to_type_id(VarType, TypeId, TypeArgs),
-		module_info_ctors(ModuleInfo, Ctors),
-		% will fail for builtin cons_ids.
-		map__search(Ctors, ConsId, ConsDefns),
-		CorrectCons = lambda([ConsDefn::in] is semidet, (
-				ConsDefn = hlds_cons_defn(_, _, _, TypeId, _)
-			)),
-		list__filter(CorrectCons, ConsDefns,
-			[hlds_cons_defn(ExistQVars0, _Constraints0, ArgTypes0,
-				_, _)]),
-		ArgTypes0 \= []
+		% The argument types of a tuple cons_id are the
+		% arguments of the tuple type.
+		type_id_is_tuple(TypeId)
+	->
+		ArgTypes = TypeArgs
+	;
+		type_util__do_get_type_and_cons_defn(ModuleInfo, TypeId,
+			ConsId, TypeDefn, ConsDefn),
+		ConsDefn = hlds_cons_defn(ExistQVars0, _Constraints0,
+				Args, _, _),
+		Args \= []
 	->
-		module_info_types(ModuleInfo, Types),
-		map__lookup(Types, TypeId, TypeDefn),
 		hlds_data__get_type_defn_tparams(TypeDefn, TypeDefnParams),
 		term__term_list_to_var_list(TypeDefnParams, TypeDefnVars),
 
 		% XXX handle ExistQVars
-		require(unify(ExistQVars0, []),
-	"type_util__get_cons_id_arg_types: existentially typed cons_id"),
+		( ExistQVars0 = [] ->
+			true
+		; 
+			(
+				EQVarAction = abort_on_exist_qvar,
+				error("type_util__get_cons_id_arg_types: existentially typed cons_id")
+			;
+				EQVarAction = fail_on_exist_qvar,
+				fail
+			)
+		),
 
 		map__from_corresponding_lists(TypeDefnVars, TypeArgs, TSubst),
+		assoc_list__values(Args, ArgTypes0),
 		term__apply_substitution_to_list(ArgTypes0, TSubst, ArgTypes)
 	;
 		ArgTypes = []
-	).
+	)
+    ;
+    	ArgTypes = []
+    ).
+
+type_util__cons_id_arg_types(ModuleInfo, VarType, ConsId, ArgTypes) :-
+	type_to_type_id(VarType, TypeId, TypeArgs),
+	module_info_types(ModuleInfo, Types),
+	map__search(Types, TypeId, TypeDefn),
+	hlds_data__get_type_defn_body(TypeDefn, TypeDefnBody),
+	TypeDefnBody = du_type(_, ConsTags, _, _),
+	map__member(ConsTags, ConsId, _),
+	
+	module_info_ctors(ModuleInfo, Ctors),
+	map__lookup(Ctors, ConsId, ConsDefns),
+	list__member(ConsDefn, ConsDefns),
+	
+	ConsDefn = hlds_cons_defn(ExistQVars0, _, Args, TypeId, _),
+
+	% XXX handle ExistQVars
+	ExistQVars0 = [],
+
+	hlds_data__get_type_defn_tparams(TypeDefn, TypeDefnParams),
+	term__term_list_to_var_list(TypeDefnParams, TypeDefnVars),
+
+	map__from_corresponding_lists(TypeDefnVars, TypeArgs, TSubst),
+	assoc_list__values(Args, ArgTypes0),
+	term__apply_substitution_to_list(ArgTypes0, TSubst, ArgTypes).
+
 
 type_util__is_existq_cons(ModuleInfo, VarType, ConsId) :-
 	type_util__is_existq_cons(ModuleInfo, VarType, ConsId, _). 
@@ -690,14 +888,8 @@ type_util__is_existq_cons(ModuleInfo, VarType, ConsId) :-
 		(type)::in, cons_id::in, hlds_cons_defn::out) is semidet.
 
 type_util__is_existq_cons(ModuleInfo, VarType, ConsId, ConsDefn) :-
-	type_to_type_id(VarType, TypeId, _TypeArgs),
-	module_info_ctors(ModuleInfo, Ctors),
-	% will fail for builtin cons_ids.
-	map__search(Ctors, ConsId, ConsDefns),
-	MatchingCons = lambda([ThisConsDefn::in] is semidet, (
-			ThisConsDefn = hlds_cons_defn(_, _, _, TypeId, _)
-		)),
-	list__filter(MatchingCons, ConsDefns, [ConsDefn]), 
+	type_to_type_id(VarType, TypeId, _),
+	type_util__get_cons_defn(ModuleInfo, TypeId, ConsId, ConsDefn),
 	ConsDefn = hlds_cons_defn(ExistQVars, _, _, _, _),
 	ExistQVars \= [].
 
@@ -706,18 +898,68 @@ type_util__is_existq_cons(ModuleInfo, VarType, ConsId, ConsDefn) :-
 	% otherwise fail.
 type_util__get_existq_cons_defn(ModuleInfo, VarType, ConsId, CtorDefn) :-
 	type_util__is_existq_cons(ModuleInfo, VarType, ConsId, ConsDefn),
-	ConsDefn = hlds_cons_defn(ExistQVars, Constraints, ArgTypes, _, _),
+	ConsDefn = hlds_cons_defn(ExistQVars, Constraints, Args, _, _),
+	assoc_list__values(Args, ArgTypes),
 	module_info_types(ModuleInfo, Types),
 	type_to_type_id(VarType, TypeId, _),
 	map__lookup(Types, TypeId, TypeDefn),
 	hlds_data__get_type_defn_tvarset(TypeDefn, TypeVarSet),
 	hlds_data__get_type_defn_tparams(TypeDefn, TypeDefnParams),
+	type_to_type_id(VarType, TypeId, _),
 	construct_type(TypeId, TypeDefnParams, RetType),
 	CtorDefn = ctor_defn(TypeVarSet, ExistQVars, Constraints,
 		ArgTypes, RetType).
 
+type_util__get_type_and_cons_defn(ModuleInfo, Type, ConsId,
+		TypeDefn, ConsDefn) :-
+	(
+		type_to_type_id(Type, TypeId, _),
+		type_util__do_get_type_and_cons_defn(ModuleInfo,
+			TypeId, ConsId, TypeDefn0, ConsDefn0)
+	->
+		TypeDefn = TypeDefn0,
+		ConsDefn = ConsDefn0
+	;
+		error("type_util__get_type_and_cons_defn")
+	).
+
+:- pred type_util__do_get_type_and_cons_defn(module_info::in,
+		type_id::in, cons_id::in, hlds_type_defn::out,
+		hlds_cons_defn::out) is semidet.
+
+type_util__do_get_type_and_cons_defn(ModuleInfo, TypeId, ConsId,
+		TypeDefn, ConsDefn) :-
+	type_util__get_cons_defn(ModuleInfo, TypeId, ConsId, ConsDefn),
+	module_info_types(ModuleInfo, Types),
+	map__lookup(Types, TypeId, TypeDefn).
+
+:- pred type_util__get_cons_defn(module_info::in, type_id::in, cons_id::in,
+		hlds_cons_defn::out) is semidet.
+
+type_util__get_cons_defn(ModuleInfo, TypeId, ConsId, ConsDefn) :-
+	module_info_ctors(ModuleInfo, Ctors),
+	% will fail for builtin cons_ids.
+	map__search(Ctors, ConsId, ConsDefns),
+	MatchingCons = lambda([ThisConsDefn::in] is semidet, (
+			ThisConsDefn = hlds_cons_defn(_, _, _, TypeId, _)
+		)),
+	list__filter(MatchingCons, ConsDefns, [ConsDefn]).
+	
 %-----------------------------------------------------------------------------%
 
+type_is_no_tag_type(ModuleInfo, Type, Ctor, ArgType) :-
+	type_to_type_id(Type, TypeId, TypeArgs),
+	module_info_no_tag_types(ModuleInfo, NoTagTypes),
+	map__search(NoTagTypes, TypeId, NoTagType),
+	NoTagType = no_tag_type(TypeParams0, Ctor, ArgType0),
+	( TypeParams0 = [] ->
+		ArgType = ArgType0
+	;
+		term__term_list_to_var_list(TypeParams0, TypeParams),
+		map__from_corresponding_lists(TypeParams, TypeArgs, Subn),
+		term__apply_substitution(ArgType0, Subn, ArgType)
+	).
+
 	% The checks for type_info and type_ctor_info
 	% are needed because those types lie about their
 	% arity; it might be cleaner to change that in
@@ -727,17 +969,32 @@ type_util__get_existq_cons_defn(ModuleInfo, VarType, ConsId, CtorDefn) :-
 	% etc. rather than just checking the unqualified type name,
 	% but I found it difficult to verify that the constructors
 	% would always be fully module-qualified at points where
-	% type_is_no_tag_type/3 is called.
+	% type_constructors_are_no_tag_type/3 is called.
 
-type_is_no_tag_type(Ctors, Ctor, Type) :-
+type_constructors_are_no_tag_type(Ctors, Ctor, ArgType, MaybeArgName) :-
 	Ctors = [SingleCtor],
-	SingleCtor = ctor(ExistQVars, _Constraints, Ctor, [_FieldName - Type]),
+	SingleCtor = ctor(ExistQVars, _Constraints, Ctor,
+		[MaybeSymName - ArgType]),
 	ExistQVars = [],
 	unqualify_name(Ctor, Name),
 	Name \= "type_info",
 	Name \= "type_ctor_info",
 	Name \= "typeclass_info",
-	Name \= "base_typeclass_info".
+	Name \= "base_typeclass_info",
+
+	% We don't handle unary tuples as no_tag types --
+	% they are rare enough that it's not worth
+	% the implementation effort.
+	Name \= "{}",
+
+	(
+		MaybeSymName = yes(SymName),
+		unqualify_name(SymName, ArgName),
+		MaybeArgName = yes(ArgName)
+	;
+		MaybeSymName = no,
+		MaybeArgName = no
+	).
 
 %-----------------------------------------------------------------------------%
 
@@ -799,6 +1056,51 @@ type_list_subsumes(TypesA, TypesB, TypeSubst) :-
 	map__init(TypeSubst0),
 	type_unify_list(TypesA, TypesB, TypesBVars, TypeSubst0, TypeSubst).
 
+
+arg_type_list_subsumes(TVarSet, ArgTypes, CalleeTVarSet,
+		CalleeExistQVars0, CalleeArgTypes0) :-
+
+	%
+	% rename the type variables in the callee's argument types.
+	%
+	varset__merge_subst(TVarSet, CalleeTVarSet, _TVarSet1, Subst),
+	term__apply_substitution_to_list(CalleeArgTypes0, Subst,
+				CalleeArgTypes),
+	map__apply_to_list(CalleeExistQVars0, Subst, CalleeExistQTypes0),
+
+	%
+	% check that the types of the candidate predicate/function
+	% subsume the actual argument types
+	% [This is the right thing to do even for calls to
+	% existentially typed preds, because we're using the
+	% type variables from the callee's pred decl (obtained
+	% from the pred_info via pred_info_arg_types) not the types
+	% inferred from the callee's clauses (and stored in the
+	% clauses_info and proc_info) -- the latter
+	% might not subsume the actual argument types.]
+	%
+	type_list_subsumes(CalleeArgTypes, ArgTypes, TypeSubst),
+
+	%
+	% check that the type substitution did not bind any
+	% existentially typed variables to non-ground types
+	%
+	( CalleeExistQTypes0 = [] ->
+		% optimize common case
+		true
+	;
+		term__apply_rec_substitution_to_list(CalleeExistQTypes0,
+			TypeSubst, CalleeExistQTypes),
+		all [T] (list__member(T, CalleeExistQTypes) =>
+				type_util__var(T, _))	
+
+		% it might make sense to also check that
+		% the type substitution did not bind any
+		% existentially typed variables to universally 
+		% quantified type variables in the caller's
+		% argument types
+	).
+
 %-----------------------------------------------------------------------------%
 
 	% Types are represented as terms, but we can't just use term__unify
@@ -902,9 +1204,8 @@ type_unify(term__functor(FX, AsX, _CX), term__functor(FY, AsY, _CY),
 	% XXX Instead of just failing if the functors' name/arity is different,
 	% we should check here if these types have been defined
 	% to be equivalent using equivalence types.  But this
-	% is difficult because (1) it causes typevarset synchronization
-	% problems, and (2) the relevant variables TypeInfo, TVarSet0, TVarSet
-	% haven't been passed in to here.
+	% is difficult because the relevant variable
+	% TypeTable hasn't been passed in to here.
 
 /*******
 	...
@@ -927,16 +1228,15 @@ type_unify(term__functor(FX, AsX, _CX), term__functor(FY, AsY, _CY),
 
 replace_eqv_type(Functor, Arity, Args, EqvType) :-
 
-	% XXX magically_obtain(TypeTable, TVarSet0, TVarSet)
+	% XXX magically_obtain(TypeTable)
 
 	make_type_id(Functor, Arity, TypeId),
 	map__search(TypeTable, TypeId, TypeDefn),
-	TypeDefn = hlds_type_defn(TypeVarSet, TypeParams0,
-			eqv_type(EqvType0), _Condition, Context, _Status),
-	varset__merge(TVarSet0, TypeVarSet, [EqvType0 | TypeParams0],
-			TVarSet, [EqvType1, TypeParams1]),
-	type_param_to_var_list(TypeParams1, TypeParams),
-	term__substitute_corresponding(EqvType1, TypeParams, AsX,
+	get_type_defn_body(TypeDefn, TypeBody),
+	TypeBody = eqv_type(EqvType0),
+	get_type_defn_tparams(TypeDefn, TypeParams0),
+	type_param_to_var_list(TypeParams0, TypeParams),
+	term__substitute_corresponding(EqvType0, TypeParams, AsX,
 		EqvType).
 
 ******/
@@ -970,6 +1270,14 @@ type_unify_head_type_param(Var, HeadVar, HeadTypeParams, Bindings0,
 type_util__vars(Type, Tvars) :-
 	term__vars(Type, Tvars).
 
+type_util__real_vars(Type, Tvars) :-
+	( is_introduced_type_info_type(Type) ->
+		% for these types, we don't add the type parameters
+		Tvars = []
+	;
+		type_util__vars(Type, Tvars)
+	).
+
 %-----------------------------------------------------------------------------%
 
 apply_substitution_to_type_map(VarTypes0, Subst, VarTypes) :-
@@ -1188,6 +1496,12 @@ apply_rec_subst_to_constraint_proofs(Subst, Proofs0, Proofs) :-
 		)),
 	Proofs0, Empty, Proofs).
 
+apply_variable_renaming_to_type_map(Renaming, Map0, Map) :-
+	map__map_values(
+		(pred(_::in, Type0::in, Type::out) is det :-
+			term__apply_variable_renaming(Type0, Renaming, Type)
+		), Map0, Map).
+
 apply_variable_renaming_to_constraints(Renaming,
 		constraints(UniversalCs0, ExistentialCs0),
 		constraints(UniversalCs, ExistentialCs)) :-
@@ -1261,4 +1575,9 @@ constraint_list_get_tvars(Constraints, TVars) :-
 constraint_get_tvars(constraint(_Name, Args), TVars) :-
 	term__vars_list(Args, TVars).
 
+get_unconstrained_tvars(Tvars, Constraints, Unconstrained) :-
+	constraint_list_get_tvars(Constraints, ConstrainedTvars),
+	list__delete_elems(Tvars, ConstrainedTvars, Unconstrained0),
+	list__remove_dups(Unconstrained0, Unconstrained).
+
 %-----------------------------------------------------------------------------%
