@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -112,8 +112,8 @@ unique_modes__check_goal(Goal0, Goal, ModeInfo0, ModeInfo) :-
 	% If the goal is not nondet, then nothing is nondet-live,
 	% so reset the bag of nondet-live vars to be empty.
 	%
-	goal_info_get_code_model(GoalInfo0, CodeModel),
-	( CodeModel = model_non ->
+	goal_info_get_determinism(GoalInfo0, Detism),
+	( determinism_components(Detism, _, at_most_many) ->
 		ModeInfo2 = ModeInfo1
 	;
 		mode_info_set_nondet_live_vars([], ModeInfo1, ModeInfo2)
@@ -128,16 +128,14 @@ unique_modes__check_goal(Goal0, Goal, ModeInfo0, ModeInfo) :-
 	% 
 	% Restore the original bag of nondet-live vars
 	%
-	mode_info_set_nondet_live_vars(NondetLiveVars0, ModeInfo3, ModeInfo),
+	mode_info_set_nondet_live_vars(NondetLiveVars0, ModeInfo3, ModeInfo4),
 
 	%
 	% Grab the final instmap, compute the change in insts
 	% over this goal, and save that instmap_delta in the goal_info.
 	%
-	mode_info_get_instmap(ModeInfo, InstMap),
-	goal_info_get_nonlocals(GoalInfo0, NonLocals),
-	compute_instmap_delta(InstMap0, InstMap, NonLocals, DeltaInstMap),
-	goal_info_set_instmap_delta(GoalInfo0, DeltaInstMap, GoalInfo),
+	compute_goal_instmap_delta(InstMap0, GoalExpr,
+		GoalInfo0, GoalInfo, ModeInfo4, ModeInfo),
 
 	Goal = GoalExpr - GoalInfo.
 
@@ -196,10 +194,12 @@ select_changed_inst_vars([Var | Vars], DeltaInstMap, ModeInfo, ChangedVars) :-
 	mode_info_get_module_info(ModeInfo, ModuleInfo),
 	mode_info_get_instmap(ModeInfo, InstMap0),
 	instmap__lookup_var(InstMap0, Var, Inst0),
+	mode_info_get_var_types(ModeInfo, VarTypes),
+	map__lookup(VarTypes, Var, Type),
 	(
 		instmap_delta_is_reachable(DeltaInstMap),
 		instmap_delta_search_var(DeltaInstMap, Var, Inst),
-		\+ inst_matches_final(Inst, Inst0, ModuleInfo)
+		\+ inst_matches_final(Inst, Inst0, Type, ModuleInfo)
 	->
 		ChangedVars = [Var | ChangedVars1],
 		select_changed_inst_vars(Vars, DeltaInstMap, ModeInfo,
@@ -295,9 +295,9 @@ unique_modes__check_goal_2(disj(List0, SM), GoalInfo0, disj(List, SM)) -->
 		% disjunct, in unique_modes__check_disj.
 		%
 		{ goal_info_get_nonlocals(GoalInfo0, NonLocals) },
-		{ goal_info_get_code_model(GoalInfo0, CodeModel) },
+		{ goal_info_get_determinism(GoalInfo0, Determinism) },
 		% does this disjunction create a choice point?
-		( { CodeModel = model_non } ->
+		( { determinism_components(Determinism, _, at_most_many) } ->
 			mode_info_add_live_vars(NonLocals),
 			make_all_nondet_live_vars_mostly_uniq,
 			mode_info_remove_live_vars(NonLocals)
@@ -309,19 +309,19 @@ unique_modes__check_goal_2(disj(List0, SM), GoalInfo0, disj(List, SM)) -->
 		% Now just modecheck each disjunct in turn, and then
 		% merge the resulting instmaps.
 		%
-		unique_modes__check_disj(List0, CodeModel, NonLocals,
+		unique_modes__check_disj(List0, Determinism, NonLocals,
 			List, InstMapList),
 		instmap__merge(NonLocals, InstMapList, disj)
 	),
 	mode_checkpoint(exit, "disj").
 
-unique_modes__check_goal_2(if_then_else(Vs, A0, B0, C0, SM), GoalInfo0, Goal)
-		-->
+unique_modes__check_goal_2(if_then_else(Vs, Cond0, Then0, Else0, SM),
+		GoalInfo0, Goal) -->
 	mode_checkpoint(enter, "if-then-else"),
 	{ goal_info_get_nonlocals(GoalInfo0, NonLocals) },
-	{ unique_modes__goal_get_nonlocals(A0, A_Vars) },
-	{ unique_modes__goal_get_nonlocals(B0, B_Vars) },
-	{ unique_modes__goal_get_nonlocals(C0, C_Vars) },
+	{ goal_get_nonlocals(Cond0, Cond_Vars) },
+	{ goal_get_nonlocals(Then0, Then_Vars) },
+	{ goal_get_nonlocals(Else0, Else_Vars) },
 	mode_info_dcg_get_instmap(InstMap0),
 	mode_info_lock_vars(if_then_else, NonLocals),
 
@@ -352,38 +352,38 @@ unique_modes__check_goal_2(if_then_else(Vs, A0, B0, C0, SM), GoalInfo0, Goal)
 	%		use(Var)
 	%	).
 	%
-	mode_info_add_live_vars(C_Vars),
+	mode_info_add_live_vars(Else_Vars),
 	=(ModeInfo),
-	{ set__to_sorted_list(A_Vars, A_Vars_List) },
-	{ select_live_vars(A_Vars_List, ModeInfo, A_Live_Vars) },
-	{ A0 = _ - A0_GoalInfo },
-	{ goal_info_get_instmap_delta(A0_GoalInfo, A0_DeltaInstMap) },
-	{ select_changed_inst_vars(A_Live_Vars, A0_DeltaInstMap, ModeInfo,
-				ChangedVars) },
+	{ set__to_sorted_list(Cond_Vars, Cond_Vars_List) },
+	{ select_live_vars(Cond_Vars_List, ModeInfo, Cond_Live_Vars) },
+	{ Cond0 = _ - Cond0_GoalInfo },
+	{ goal_info_get_instmap_delta(Cond0_GoalInfo, Cond0_DeltaInstMap) },
+	{ select_changed_inst_vars(Cond_Live_Vars, Cond0_DeltaInstMap,
+		ModeInfo, ChangedVars) },
 	make_var_list_mostly_uniq(ChangedVars),
-	mode_info_remove_live_vars(C_Vars),
+	mode_info_remove_live_vars(Else_Vars),
 
-	mode_info_add_live_vars(B_Vars),
-	unique_modes__check_goal(A0, A),
-	mode_info_remove_live_vars(B_Vars),
+	mode_info_add_live_vars(Then_Vars),
+	unique_modes__check_goal(Cond0, Cond),
+	mode_info_remove_live_vars(Then_Vars),
 	mode_info_unlock_vars(if_then_else, NonLocals),
-	mode_info_dcg_get_instmap(InstMapA),
-	( { instmap__is_reachable(InstMapA) } ->
-		unique_modes__check_goal(B0, B),
-		mode_info_dcg_get_instmap(InstMapB)
+	mode_info_dcg_get_instmap(InstMapCond),
+	( { instmap__is_reachable(InstMapCond) } ->
+		unique_modes__check_goal(Then0, Then),
+		mode_info_dcg_get_instmap(InstMapThen)
 	;
 		% We should not mode-analyse the goal, since it is unreachable.
 		% Instead we optimize the goal away, so that later passes
 		% won't complain about it not having unique mode information.
-		{ true_goal(B) },
-		{ InstMapB = InstMapA }
+		{ true_goal(Then) },
+		{ InstMapThen = InstMapCond }
 	),
 	mode_info_set_instmap(InstMap0),
-	unique_modes__check_goal(C0, C),
-	mode_info_dcg_get_instmap(InstMapC),
+	unique_modes__check_goal(Else0, Else),
+	mode_info_dcg_get_instmap(InstMapElse),
 	mode_info_set_instmap(InstMap0),
-	instmap__merge(NonLocals, [InstMapB, InstMapC], if_then_else),
-	{ Goal = if_then_else(Vs, A, B, C, SM) },
+	instmap__merge(NonLocals, [InstMapThen, InstMapElse], if_then_else),
+	{ Goal = if_then_else(Vs, Cond, Then, Else, SM) },
 	mode_checkpoint(exit, "if-then-else").
 
 unique_modes__check_goal_2(not(A0), GoalInfo0, not(A)) -->
@@ -398,14 +398,15 @@ unique_modes__check_goal_2(not(A0), GoalInfo0, not(A)) -->
 	%
 	{ goal_info_get_nonlocals(GoalInfo0, NonLocals) },
 	{ set__to_sorted_list(NonLocals, NonLocalsList) },
-	=(ModeInfo),
-	{ select_live_vars(NonLocalsList, ModeInfo, LiveNonLocals) },
+	=(ModeInfo0),
+	{ select_live_vars(NonLocalsList, ModeInfo0, LiveNonLocals) },
 	make_var_list_mostly_uniq(LiveNonLocals),
 	%
 	% But nothing is forward-live for the negated goal, since
 	% if the goal succeeds then execution will immediately backtrack.
 	% So we need to set the live variables set to empty here.
 	%
+	=(ModeInfo),
 	{ mode_info_get_live_vars(ModeInfo, LiveVars0) },
 	mode_info_set_live_vars([]),
 	%
@@ -435,7 +436,6 @@ unique_modes__check_goal_2(generic_call(GenericCall, Args, Modes, Det),
 	;
 		NeverSucceeds = no
 	},
-	{ determinism_to_code_model(Det, CodeModel) },
 
 	{
 		GenericCall = higher_order(_, _, _),
@@ -446,11 +446,14 @@ unique_modes__check_goal_2(generic_call(GenericCall, Args, Modes, Det),
 		GenericCall = class_method(_, _, _, _),
 		ArgOffset = 0
 	;
-		% `aditi_insert' goals have type_info arguments for each
-		% of the arguments of the tuple to insert added to the
-		% start of the argument list by polymorphism.m.
+		% `aditi_insert' and `aditi_delete' goals have type_info
+		% arguments for each of the arguments of the tuple to insert
+		% added to the start of the argument list by polymorphism.m.
 		GenericCall = aditi_builtin(Builtin, UpdatedCallId),
-		( Builtin = aditi_insert(_), UpdatedCallId = _ - _/Arity ->
+		(
+			Builtin = aditi_tuple_insert_delete(_, _),
+			UpdatedCallId = _ - _/Arity
+		->
 			ArgOffset = -Arity
 		;
 			ArgOffset = 0
@@ -458,7 +461,7 @@ unique_modes__check_goal_2(generic_call(GenericCall, Args, Modes, Det),
 	},
 
 	unique_modes__check_call_modes(Args, Modes, ArgOffset,
-		CodeModel, NeverSucceeds),
+		Det, NeverSucceeds),
 	{ Goal = generic_call(GenericCall, Args, Modes, Det) },
 	mode_info_unset_call_context,
 	mode_checkpoint(exit, "generic_call").
@@ -505,15 +508,15 @@ unique_modes__check_goal_2(switch(Var, CanFail, Cases0, SM), GoalInfo0,
 
 	% to modecheck a pragma_c_code, we just modecheck the proc for 
 	% which it is the goal.
-unique_modes__check_goal_2(pragma_c_code(IsRecursive, PredId, ProcId0,
-		Args, ArgNameMap, OrigArgTypes, PragmaCode),
+unique_modes__check_goal_2(pragma_foreign_code(Attributes,
+		PredId, ProcId0, Args, ArgNameMap, OrigArgTypes, PragmaCode),
 		_GoalInfo, Goal) -->
 	mode_checkpoint(enter, "pragma_c_code"),
 	=(ModeInfo),
 	{ mode_info_get_call_id(ModeInfo, PredId, CallId) },
 	mode_info_set_call_context(call(call(CallId))),
 	unique_modes__check_call(PredId, ProcId0, Args, ProcId),
-	{ Goal = pragma_c_code(IsRecursive, PredId, ProcId, Args,
+	{ Goal = pragma_foreign_code(Attributes, PredId, ProcId, Args,
 			ArgNameMap, OrigArgTypes, PragmaCode) },
 	mode_info_unset_call_context,
 	mode_checkpoint(exit, "pragma_c_code").
@@ -542,30 +545,44 @@ unique_modes__check_call(PredId, ProcId0, ArgVars, ProcId,
 	%
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
 	module_info_pred_proc_info(ModuleInfo, PredId, ProcId0,
-		PredInfo, ProcInfo),
+			PredInfo, ProcInfo),
 	compute_arg_offset(PredInfo, ArgOffset),
 	proc_info_argmodes(ProcInfo, ProcArgModes0),
-	proc_info_interface_code_model(ProcInfo, CodeModel),
+	proc_info_interface_determinism(ProcInfo, InterfaceDeterminism),
 	proc_info_never_succeeds(ProcInfo, NeverSucceeds),
 	unique_modes__check_call_modes(ArgVars, ProcArgModes0, ArgOffset,
-			CodeModel, NeverSucceeds, ModeInfo1, ModeInfo2),
+			InterfaceDeterminism, NeverSucceeds,
+			ModeInfo1, ModeInfo2),
+	( ProcInfo ^ mode_errors = [_|_] ->
+		% mode error in callee for this mode
+		WaitingVars = set__list_to_set(ArgVars),
+		mode_info_get_instmap(ModeInfo2, InstMap),
+		instmap__lookup_vars(ArgVars, InstMap, ArgInsts),
+		mode_info_error(WaitingVars,
+			mode_error_in_callee(ArgVars, ArgInsts,
+				PredId, ProcId0,
+				ProcInfo ^ mode_errors),
+			ModeInfo2, ModeInfo3)
+	;
+		ModeInfo3 = ModeInfo2
+	),
 
 	%
 	% see whether or not that worked
 	% (and restore the old error list)
 	%
-	mode_info_get_errors(ModeInfo2, Errors),
-	mode_info_set_errors(OldErrors, ModeInfo2, ModeInfo3),
-	mode_info_get_may_change_called_proc(ModeInfo3, MayChangeCalledProc),
+	mode_info_get_errors(ModeInfo3, Errors),
+	mode_info_set_errors(OldErrors, ModeInfo3, ModeInfo4),
+	mode_info_get_may_change_called_proc(ModeInfo4, MayChangeCalledProc),
 	( Errors = [] ->
 		ProcId = ProcId0,
-		ModeInfo = ModeInfo3
+		ModeInfo = ModeInfo4
 	; MayChangeCalledProc = may_not_change_called_proc ->
 		% We're not allowed to try a different procedure
 		% here, so just return all the errors.
 		ProcId = ProcId0,
 		list__append(OldErrors, Errors, AllErrors),
-		mode_info_set_errors(AllErrors, ModeInfo3, ModeInfo)
+		mode_info_set_errors(AllErrors, ModeInfo4, ModeInfo)
 	;
 		%
 		% If it didn't work, restore the original instmap,
@@ -581,10 +598,10 @@ unique_modes__check_call(PredId, ProcId0, ArgVars, ProcId,
 		% as a result of unique mode analysis.  That is OK,
 		% because uniqueness should not affect determinism.
 		%
-		mode_info_set_instmap(InstMap0, ModeInfo3, ModeInfo4),
+		mode_info_set_instmap(InstMap0, ModeInfo4, ModeInfo5),
 		proc_info_inferred_determinism(ProcInfo, Determinism),
 		modecheck_call_pred(PredId, ProcId0, ArgVars, yes(Determinism),
-			ProcId, NewArgVars, ExtraGoals, ModeInfo4, ModeInfo),
+			ProcId, NewArgVars, ExtraGoals, ModeInfo5, ModeInfo),
 		
 		( NewArgVars = ArgVars, ExtraGoals = no_extra_goals ->
 			true
@@ -604,18 +621,21 @@ unique_modes__check_call(PredId, ProcId0, ArgVars, ProcId,
 	% inst was unique.
 
 :- pred unique_modes__check_call_modes(list(prog_var), list(mode), int,
-		code_model, bool, mode_info, mode_info).
+		determinism, bool, mode_info, mode_info).
 :- mode unique_modes__check_call_modes(in, in, in, in, in,
 			mode_info_di, mode_info_uo) is det.
 
 unique_modes__check_call_modes(ArgVars, ProcArgModes, ArgOffset,
-		CodeModel, NeverSucceeds, ModeInfo0, ModeInfo) :-
+		Determinism, NeverSucceeds, ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
 	mode_list_get_initial_insts(ProcArgModes, ModuleInfo,
-				InitialInsts),
-	modecheck_var_has_inst_list(ArgVars, InitialInsts, ArgOffset,
-				ModeInfo0, ModeInfo1),
-	mode_list_get_final_insts(ProcArgModes, ModuleInfo, FinalInsts),
+			InitialInsts),
+	NeedExactMatch = no,
+	modecheck_var_has_inst_list(ArgVars, InitialInsts,
+			NeedExactMatch, ArgOffset, InstVarSub,
+			ModeInfo0, ModeInfo1),
+	mode_list_get_final_insts(ProcArgModes, ModuleInfo, FinalInsts0),
+	inst_list_apply_substitution(FinalInsts0, InstVarSub, FinalInsts),
 	modecheck_set_var_inst_list(ArgVars, InitialInsts, FinalInsts,
 		ArgOffset, NewArgVars, ExtraGoals, ModeInfo1, ModeInfo2),
 	( NewArgVars = ArgVars, ExtraGoals = no_extra_goals ->
@@ -634,7 +654,7 @@ unique_modes__check_call_modes(ArgVars, ProcArgModes, ArgOffset,
 		% If so, mark all the currently nondet-live variables
 		% whose inst is `unique' as instead being only `mostly_unique'.
 		%
-		( CodeModel = model_non ->
+		( determinism_components(Determinism, _, at_most_many) ->
 			make_all_nondet_live_vars_mostly_uniq(ModeInfo2,
 				ModeInfo)
 		;
@@ -653,7 +673,7 @@ unique_modes__check_call_modes(ArgVars, ProcArgModes, ArgOffset,
 
 unique_modes__check_conj([], []) --> [].
 unique_modes__check_conj([Goal0 | Goals0], [Goal | Goals]) -->
-	{ unique_modes__goal_get_nonlocals(Goal0, NonLocals) },
+	{ goal_get_nonlocals(Goal0, NonLocals) },
 	mode_info_remove_live_vars(NonLocals),
 	unique_modes__check_goal(Goal0, Goal),
 	mode_info_dcg_get_instmap(InstMap),
@@ -672,17 +692,18 @@ unique_modes__check_conj([Goal0 | Goals0], [Goal | Goals]) -->
 
 	% make_par_conj_nonlocal_multiset builds a multiset (bag) of all
 	% the nonlocals of the conjuncts.
-:- pred make_par_conj_nonlocal_multiset(list(hlds_goal), bag(prog_var)).
-:- mode make_par_conj_nonlocal_multiset(in, out) is det.
+:- pred make_par_conj_nonlocal_multiset(list(hlds_goal)::in,
+	bag(prog_var)::out) is det.
 
 make_par_conj_nonlocal_multiset([], Empty) :-
 	bag__init(Empty).
-make_par_conj_nonlocal_multiset([G|Gs], NonLocalsMultiSet) :-
-	make_par_conj_nonlocal_multiset(Gs, NonLocalsMultiSet0),
-	unique_modes__goal_get_nonlocals(G, NonLocals),
+make_par_conj_nonlocal_multiset([Goal | Goals], NonLocalsMultiSet) :-
+	make_par_conj_nonlocal_multiset(Goals, NonLocalsMultiSet0),
+	goal_get_nonlocals(Goal, NonLocals),
 	set__to_sorted_list(NonLocals, NonLocalsList),
 	bag__from_list(NonLocalsList, NonLocalsMultiSet1),
-	bag__union(NonLocalsMultiSet0, NonLocalsMultiSet1, NonLocalsMultiSet).
+	bag__union(NonLocalsMultiSet0, NonLocalsMultiSet1,
+		NonLocalsMultiSet).
 
 	% To unique-modecheck a parallel conjunction, we find the variables
 	% that are nonlocal to more than one conjunct and make them shared,
@@ -735,7 +756,7 @@ unique_modes__check_par_conj_0(NonLocalVarsBag, ModeInfo0, ModeInfo) :-
 unique_modes__check_par_conj_1([], [], []) --> [].
 unique_modes__check_par_conj_1([Goal0 | Goals0], [Goal | Goals],
 		[InstMap - NonLocals|InstMaps]) -->
-	{ unique_modes__goal_get_nonlocals(Goal0, NonLocals) },
+	{ goal_get_nonlocals(Goal0, NonLocals) },
 	mode_info_dcg_get_instmap(InstMap0),
 	unique_modes__check_goal(Goal0, Goal),
 	mode_info_dcg_get_instmap(InstMap),
@@ -748,22 +769,22 @@ unique_modes__check_par_conj_1([Goal0 | Goals0], [Goal | Goals],
 	% the original instmap before processing the next one.
 	% Collect up a list of the resulting instmaps.
 
-:- pred unique_modes__check_disj(list(hlds_goal), code_model, set(prog_var),
+:- pred unique_modes__check_disj(list(hlds_goal), determinism, set(prog_var),
 		list(hlds_goal), list(instmap), mode_info, mode_info).
 :- mode unique_modes__check_disj(in, in, in, out, out,
 		mode_info_di, mode_info_uo) is det.
 
 unique_modes__check_disj([], _, _, [], []) --> [].
-unique_modes__check_disj([Goal0 | Goals0], DisjCodeModel, DisjNonLocals,
+unique_modes__check_disj([Goal0 | Goals0], DisjDetism, DisjNonLocals,
 		[Goal | Goals], [InstMap | InstMaps]) -->
 	mode_info_dcg_get_instmap(InstMap0),
 	(
 		%
 		% If the disjunction was model_nondet, then we already marked
 		% all the non-locals as only being mostly-unique, so we
-		% don't need to do anything speical here...
+		% don't need to do anything special here...
 		%
-		{ DisjCodeModel \= model_non },
+		{ \+ determinism_components(DisjDetism, _, at_most_many) },
 
 		%
 		% ... but for model_semi or model_det disjunctions, if the
@@ -785,7 +806,7 @@ unique_modes__check_disj([Goal0 | Goals0], DisjCodeModel, DisjNonLocals,
 	unique_modes__check_goal(Goal0, Goal),
 	mode_info_dcg_get_instmap(InstMap),
 	mode_info_set_instmap(InstMap0),
-	unique_modes__check_disj(Goals0, DisjCodeModel, DisjNonLocals,
+	unique_modes__check_disj(Goals0, DisjDetism, DisjNonLocals,
 		Goals, InstMaps).
 
 %-----------------------------------------------------------------------------%
@@ -824,11 +845,3 @@ unique_modes__check_case_list([Case0 | Cases0], Var,
 	unique_modes__check_case_list(Cases0, Var, Cases, InstMaps).
 
 %-----------------------------------------------------------------------------%
-
-:- pred unique_modes__goal_get_nonlocals(hlds_goal, set(prog_var)).
-:- mode unique_modes__goal_get_nonlocals(in, out) is det.
-
-unique_modes__goal_get_nonlocals(_Goal - GoalInfo, NonLocals) :-
-	goal_info_get_nonlocals(GoalInfo, NonLocals).
-
-%-----------------------------------------------------------------------------%
