@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996-1999 The University of Melbourne.
+% Copyright (C) 1996-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -247,25 +247,15 @@ setup_pred_args(ModuleInfo, PredId, [ProcId | Rest], UnusedArgInfo, VarUsage0,
 		OptProcs1 = OptProcs0,
 		VarUsage1 = VarUsage0
 	;
-		proc_info_argmodes(ProcInfo, ArgModes),
-		proc_info_headvars(ProcInfo, HeadVars),
 		proc_info_vartypes(ProcInfo, VarTypes),
 		map__keys(VarTypes, Vars),
 		initialise_vardep(VarDep0, Vars, VarDep1),
-		setup_output_args(ModuleInfo, HeadVars,
-			ArgModes, VarDep1, VarDep2),
+		setup_output_args(ModuleInfo, ProcInfo, VarDep1, VarDep2),
 		
 		module_info_globals(ModuleInfo, Globals),
 		proc_interface_should_use_typeinfo_liveness(PredInfo, ProcId,
 			Globals, TypeInfoLiveness),
-		( 
-			TypeInfoLiveness = yes,
-			pred_info_module(PredInfo, PredModule),
-			pred_info_name(PredInfo, PredName),
-			pred_info_arity(PredInfo, PredArity),
-			\+ polymorphism__no_type_info_builtin(PredModule,
-				PredName, PredArity)
-		->
+		( TypeInfoLiveness = yes ->
 			proc_info_typeinfo_varmap(ProcInfo, TVarMap),
 			setup_typeinfo_deps(Vars, VarTypes, 
 				proc(PredId, ProcId), TVarMap, VarDep2,
@@ -275,7 +265,8 @@ setup_pred_args(ModuleInfo, PredId, [ProcId | Rest], UnusedArgInfo, VarUsage0,
 		),
 
 		proc_info_goal(ProcInfo, Goal - _),
-		traverse_goal(ModuleInfo, Goal, VarDep3, VarDep),
+		Info = traverse_info(ModuleInfo, VarTypes),
+		traverse_goal(Info, Goal, VarDep3, VarDep),
 		map__set(VarUsage0, proc(PredId, ProcId), VarDep, VarUsage1),
 		PredProcs1 = [proc(PredId, ProcId) | PredProcs0],
 		OptProcs1 = OptProcs0
@@ -327,32 +318,13 @@ setup_typeinfo_deps([Var | Vars], VarTypeMap, PredProcId, TVarMap, VarDep0,
 
 	% Get output arguments for a procedure given the headvars and the
 	% argument modes, and set them as used.
-:- pred setup_output_args(module_info::in, list(prog_var)::in, list(mode)::in,
+:- pred setup_output_args(module_info::in, proc_info::in,
 			var_dep::in, var_dep::out) is det.
 
-setup_output_args(ModuleInfo, HVars, ArgModes, VarDep0, VarDep) :-
-	(
-		HVars = [Var | Vars], ArgModes = [Mode | Modes]
-	->
-		(
-			% Any argument which has its instantiatedness
-			% changed by the predicate is used.
-			mode_get_insts(ModuleInfo, Mode, Inst1, Inst2),
-			\+ inst_matches_binding(Inst1, Inst2, ModuleInfo)
-		->
-			set_var_used(VarDep0, Var, VarDep1)		
-		;
-			VarDep1 = VarDep0	
-		),
-		setup_output_args(ModuleInfo, Vars, Modes, VarDep1, VarDep)
-	;
-		HVars = [], ArgModes = []
-	->
-		VarDep = VarDep0	
-	;
-		error("setup_output_args: invalid call")
-	).
-
+setup_output_args(ModuleInfo, ProcInfo, VarDep0, VarDep) :-
+	proc_info_instantiated_head_vars(ModuleInfo, ProcInfo,
+		ChangedInstHeadVars),
+	list__foldl(set_var_used, ChangedInstHeadVars, VarDep0, VarDep).
 
 	% searches for the dependencies of a variable, succeeds if the variable
 	%	is definitely used
@@ -391,9 +363,9 @@ add_aliases(UseInf0, Var, Aliases, UseInf) :-
 set_list_vars_used(UseInfo0, Vars, UseInfo) :-
 	map__delete_list(UseInfo0, Vars, UseInfo).
 
-:- pred set_var_used(var_dep::in, prog_var::in, var_dep::out) is det.
+:- pred set_var_used(prog_var::in, var_dep::in, var_dep::out) is det.
 
-set_var_used(UseInfo0, Var, UseInfo) :-
+set_var_used(Var, UseInfo0, UseInfo) :-
 	map__delete(UseInfo0, Var, UseInfo).
 
 
@@ -407,49 +379,56 @@ lookup_local_var(VarDep, Var, UsageInfo) :-
 	% Traversal of goal structure, building up dependencies for all
 	% variables. 
 
-:- pred traverse_goal(module_info::in, hlds_goal_expr::in,
+:- type traverse_info
+	--->	traverse_info(
+			module_info :: module_info,
+			vartypes :: vartypes
+		).
+
+:- pred traverse_goal(traverse_info::in, hlds_goal_expr::in,
 				var_dep::in, var_dep::out) is det.
 
 % handle conjunction
-traverse_goal(ModuleInfo, conj(Goals), UseInf0, UseInf) :-
-	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
+traverse_goal(Info, conj(Goals), UseInf0, UseInf) :-
+	traverse_list_of_goals(Info, Goals, UseInf0, UseInf).
 
 % handle parallel conjunction
-traverse_goal(ModuleInfo, par_conj(Goals, _SM), UseInf0, UseInf) :-
-	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
+traverse_goal(Info, par_conj(Goals, _SM), UseInf0, UseInf) :-
+	traverse_list_of_goals(Info, Goals, UseInf0, UseInf).
 
 % handle disjunction
-traverse_goal(ModuleInfo, disj(Goals, _), UseInf0, UseInf) :-
-	traverse_list_of_goals(ModuleInfo, Goals, UseInf0, UseInf).
+traverse_goal(Info, disj(Goals, _), UseInf0, UseInf) :-
+	traverse_list_of_goals(Info, Goals, UseInf0, UseInf).
 
 % handle switch
-traverse_goal(ModuleInfo, switch(Var, _, Cases, _), UseInf0, UseInf) :-
-	set_var_used(UseInf0, Var, UseInf1),
+traverse_goal(Info, switch(Var, _, Cases, _), UseInf0, UseInf) :-
+	set_var_used(Var, UseInf0, UseInf1),
 	list_case_to_list_goal(Cases, Goals),
-	traverse_list_of_goals(ModuleInfo, Goals, UseInf1, UseInf).
+	traverse_list_of_goals(Info, Goals, UseInf1, UseInf).
 
 % handle predicate call
-traverse_goal(ModuleInfo, call(PredId, ProcId, Args, _, _, _),
+traverse_goal(Info, call(PredId, ProcId, Args, _, _, _),
 						UseInf0, UseInf) :-
-	module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _Pred, Proc),
+	module_info_pred_proc_info(Info^module_info, PredId, ProcId, _Pred,
+		Proc),
 	proc_info_headvars(Proc, HeadVars),
 	add_pred_call_arg_dep(proc(PredId, ProcId), Args, HeadVars,
 		UseInf0, UseInf).
 
 % handle if then else
-traverse_goal(ModuleInfo, if_then_else(_, Cond - _, Then - _, Else - _, _),
+traverse_goal(Info, if_then_else(_, Cond - _, Then - _, Else - _, _),
 			UseInf0, UseInf) :-
-	traverse_goal(ModuleInfo, Cond, UseInf0, UseInf1),
-	traverse_goal(ModuleInfo, Then, UseInf1, UseInf2),
-	traverse_goal(ModuleInfo, Else, UseInf2, UseInf).
+	traverse_goal(Info, Cond, UseInf0, UseInf1),
+	traverse_goal(Info, Then, UseInf1, UseInf2),
+	traverse_goal(Info, Else, UseInf2, UseInf).
 
 % handle negation
-traverse_goal(ModuleInfo, not(Goal - _), UseInf0, UseInf) :-
-	traverse_goal(ModuleInfo, Goal, UseInf0, UseInf).
+traverse_goal(Info, not(Goal - _), UseInf0, UseInf) :-
+	traverse_goal(Info, Goal, UseInf0, UseInf).
 
 % handle quantification
-traverse_goal(ModuleInfo, some(_, _, Goal - _), UseInf0, UseInf) :-
-	traverse_goal(ModuleInfo, Goal, UseInf0, UseInf).
+traverse_goal(Info, some(_, _, Goal - _), UseInf0, UseInf) :-
+	traverse_goal(Info, Goal, UseInf0, UseInf).
 
 % we assume that higher-order predicate calls use all variables involved
 traverse_goal(_, generic_call(GenericCall, Args, _, _), UseInf0, UseInf) :-
@@ -457,9 +436,10 @@ traverse_goal(_, generic_call(GenericCall, Args, _, _), UseInf0, UseInf) :-
 	set_list_vars_used(UseInf0, CallArgs, UseInf1),
 	set_list_vars_used(UseInf1, Args, UseInf).
 
-% handle pragma c_code(...) -
-% only those arguments which have C names can be used in the C code.
-traverse_goal(_, pragma_c_code(_, _, _, Args, Names, _, _), UseInf0, UseInf) :-
+% handle pragma foreign(...) -
+% only those arguments which have names can be used in the foreign code.
+traverse_goal(_, pragma_foreign_code(_, _, _, Args, Names, _, _),
+		UseInf0, UseInf) :-
 	assoc_list__from_corresponding_lists(Args, Names, ArgsAndNames),
 	ArgIsUsed = lambda([ArgAndName::in, Arg::out] is semidet, (
 				ArgAndName = Arg - MaybeName,
@@ -471,21 +451,22 @@ traverse_goal(_, pragma_c_code(_, _, _, Args, Names, _, _), UseInf0, UseInf) :-
 % cases to handle all the different types of unification
 traverse_goal(_, unify(_, _, _, simple_test(Var1, Var2),_), UseInf0, UseInf)
 		:-
-	set_var_used(UseInf0, Var1, UseInf1),
-	set_var_used(UseInf1, Var2, UseInf).
+	set_var_used(Var1, UseInf0, UseInf1),
+	set_var_used(Var2, UseInf1, UseInf).
 		
 traverse_goal(_, unify(_, _, _, assign(Var1, Var2), _), UseInf0, UseInf) :-
 	( local_var_is_used(UseInf0, Var1) ->
 		% if Var1 used to instantiate an output argument, Var2 used
-		set_var_used(UseInf0, Var2, UseInf)
+		set_var_used(Var2, UseInf0, UseInf)
 	;
 		add_aliases(UseInf0, Var2, [Var1], UseInf)
 	).
 
-traverse_goal(ModuleInfo,
-		unify(Var1, _, _, deconstruct(_, _, Args, Modes, CanFail), _),
+traverse_goal(Info,
+		unify(Var1, _, _, 
+			deconstruct(_, _, Args, Modes, CanFail, _), _),
 		UseInf0, UseInf) :-
-	partition_deconstruct_args(ModuleInfo, Args,
+	partition_deconstruct_args(Info, Args,
 		Modes, InputVars, OutputVars),
 		% The deconstructed variable is used if any of the
 		% variables, that the deconstruction binds are used.
@@ -497,7 +478,7 @@ traverse_goal(ModuleInfo,
 		CanFail = can_fail	
 	->
 		% a deconstruction that can_fail uses its left arg
-		set_var_used(UseInf2, Var1, UseInf)
+		set_var_used(Var1, UseInf2, UseInf)
 	;
 		UseInf = UseInf2	
 	).
@@ -517,8 +498,8 @@ traverse_goal(_, unify(Var, Rhs, _, complicated_unify(_, _, _), _),
 	% with --error-check-only and polymorphism has not been run.
 	% Complicated unifications should only be var-var.
 	( Rhs = var(RhsVar) ->
-		set_var_used(UseInf0, RhsVar, UseInf1),
-		set_var_used(UseInf1, Var, UseInf)
+		set_var_used(RhsVar, UseInf0, UseInf1),
+		set_var_used(Var, UseInf1, UseInf)
 	;
 		error("complicated unifications should only be var-var")
 	).
@@ -564,23 +545,27 @@ add_arg_dep(UseInf0, Var, PredProc, Arg, UseInf) :-
 			
 	% Partition the arguments to a deconstruction into inputs
 	% and outputs.
-:- pred partition_deconstruct_args(module_info::in, list(prog_var)::in,
+:- pred partition_deconstruct_args(traverse_info::in, list(prog_var)::in,
 		list(uni_mode)::in, list(prog_var)::out,
 		list(prog_var)::out) is det.
 
-partition_deconstruct_args(ModuleInfo, ArgVars, ArgModes,
-		InputVars, OutputVars) :-
+partition_deconstruct_args(Info, ArgVars, ArgModes, InputVars, OutputVars) :-
 	(
 		ArgVars = [Var | Vars], ArgModes = [Mode | Modes]
 	->
-		partition_deconstruct_args(ModuleInfo,
-			Vars, Modes, InputVars1, OutputVars1),
+		partition_deconstruct_args(Info, Vars, Modes, InputVars1,
+			OutputVars1),
 		Mode = ((InitialInst1 - InitialInst2) ->
 			(FinalInst1 - FinalInst2)),
 
+		map__lookup(Info^vartypes, Var, Type),
+
 		% If the inst of the argument of the LHS is changed,
 		% the argument is input.
-		( inst_matches_binding(InitialInst1, FinalInst1, ModuleInfo) ->
+		(
+			inst_matches_binding(InitialInst1, FinalInst1,
+				Type, Info^module_info)
+		->
 			InputVars = InputVars1
 		;
 			InputVars = [Var | InputVars1]
@@ -588,7 +573,10 @@ partition_deconstruct_args(ModuleInfo, ArgVars, ArgModes,
 
 		% If the inst of the argument of the RHS is changed,
 		% the argument is output.
-		( inst_matches_binding(InitialInst2, FinalInst2, ModuleInfo) ->
+		(
+			inst_matches_binding(InitialInst2, FinalInst2,
+				Type, Info^module_info)
+		->
 			OutputVars = OutputVars1
 		;
 			OutputVars = [Var | OutputVars1]
@@ -627,13 +615,13 @@ list_case_to_list_goal([case(_, Goal) | Cases], [Goal | Goals]) :-
 	list_case_to_list_goal(Cases, Goals).
 
 
-:- pred traverse_list_of_goals(module_info::in, list(hlds_goal)::in,
+:- pred traverse_list_of_goals(traverse_info::in, list(hlds_goal)::in,
 					var_dep::in, var_dep::out) is det.
 
 traverse_list_of_goals(_, [], UseInf, UseInf).
-traverse_list_of_goals(ModuleInfo, [Goal - _ | Goals], UseInf0, UseInf) :-
-	traverse_goal(ModuleInfo, Goal, UseInf0, UseInf1),
-	traverse_list_of_goals(ModuleInfo, Goals, UseInf1, UseInf).  
+traverse_list_of_goals(Info, [Goal - _ | Goals], UseInf0, UseInf) :-
+	traverse_goal(Info, Goal, UseInf0, UseInf1),
+	traverse_list_of_goals(Info, Goals, UseInf1, UseInf).  
 
 
 %-------------------------------------------------------------------------------
@@ -714,7 +702,7 @@ unused_args_check_all_vars(VarUsage, Changed0, Changed, [Var| Vars],
 			)
 		->
 			% set the current variable to used
-			set_var_used(LocalVars0, Var, LocalVars1),
+			set_var_used(Var, LocalVars0, LocalVars1),
 			Changed1 = yes
 		;
 			Changed1 = Changed0,
@@ -987,8 +975,7 @@ make_new_pred_info(ModuleInfo, PredInfo0, UnusedArgs, NameSuffix, Status,
 	% Replace the goal in the procedure with one to call the given
 	% pred_id and proc_id.
 :- pred create_call_goal(list(int)::in, pred_id::in, proc_id::in,
-		module_name::in, string::in, proc_info::in, proc_info::out)
-		is det.
+	module_name::in, string::in, proc_info::in, proc_info::out) is det.
 
 create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 		PredName, OldProc0, OldProc) :-
@@ -1006,13 +993,12 @@ create_call_goal(UnusedArgs, NewPredId, NewProcId, PredModule,
 	GoalExpr = call(NewPredId, NewProcId, NewHeadVars,
 		      not_builtin, no, qualified(PredModule, PredName)),
 	Goal1 = GoalExpr - GoalInfo0,
-	implicitly_quantify_goal(Goal1, Varset0, VarTypes1, NonLocals, 
-			Goal, Varset, VarTypes, _),
+	implicitly_quantify_goal(Goal1, Varset0, VarTypes1,
+		NonLocals, Goal, Varset, VarTypes, _),
 	proc_info_set_goal(OldProc0, Goal, OldProc1),
 	proc_info_set_varset(OldProc1, Varset, OldProc2),
 	proc_info_set_vartypes(OldProc2, VarTypes, OldProc).
 
-
 	% Create a pred_info for an imported pred with a pragma unused_args
 	% in the .opt file.
 :- pred make_imported_unused_args_pred_infos(pred_proc_list::in,
@@ -1173,8 +1159,8 @@ do_fixup_unused_args(VarUsage, proc(OldPredId, OldProcId), ProcCallInfo,
 		Changed = yes,
 			% if anything has changed, rerun quantification
 		set__list_to_set(HeadVars, NonLocals),
-		implicitly_quantify_goal(Goal1, Varset0, VarTypes0, NonLocals,
-						Goal, Varset, VarTypes, _),
+		implicitly_quantify_goal(Goal1, Varset0, VarTypes0,
+			NonLocals, Goal, Varset, VarTypes, _),
 		proc_info_set_goal(FixedProc2, Goal, FixedProc3),
 		proc_info_set_varset(FixedProc3, Varset, FixedProc4),
 		proc_info_set_vartypes(FixedProc4, VarTypes, FixedProc5)
@@ -1289,7 +1275,7 @@ fixup_goal_expr(_ModuleInfo, _UnusedVars, _ProcCallInfo, no,
 
 fixup_goal_expr(_ModuleInfo, _UnusedVars, _ProcCallInfo, no,
 			GoalExpr - GoalInfo, GoalExpr - GoalInfo) :-
-	GoalExpr = pragma_c_code(_, _, _, _, _, _, _).
+	GoalExpr = pragma_foreign_code(_, _, _, _, _, _, _).
 
 fixup_goal_expr(_, _, _, _, bi_implication(_, _) - _, _) :-
 	% these should have been expanded out by now
@@ -1383,7 +1369,7 @@ fixup_unify(_, UnusedVars, no, Unify, Unify) :-
 	\+ list__member(LVar, UnusedVars).
 	
 fixup_unify(ModuleInfo, UnusedVars, Changed, Unify, Unify) :-
-	Unify =	deconstruct(LVar, _, ArgVars, ArgModes, CanFail),
+	Unify =	deconstruct(LVar, _, ArgVars, ArgModes, CanFail, _CanCGC),
 	\+ list__member(LVar, UnusedVars),
 	(
 			% are any of the args unused, if so we need to 	
