@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -15,15 +15,15 @@
 
 :- interface.
 
-:- import_module list, set, bool.
 :- import_module llds, vn_table, vn_type, livemap.
+:- import_module list, set, bool, counter.
 
 :- pred vn_block__divide_into_blocks(list(instruction), set(label),
 	list(list(instruction))).
 :- mode vn_block__divide_into_blocks(in, in, out) is det.
 
 :- pred vn_block__build_block_info(list(instruction), livemap, vn_params,
-	list(parentry), int, vn_tables, vnlvalset, bool, vn_ctrl_tuple).
+	list(parentry), counter, vn_tables, vnlvalset, bool, vn_ctrl_tuple).
 :- mode vn_block__build_block_info(in, in, in, in, in, out, out, out, out)
 	is det.
 
@@ -90,7 +90,7 @@ vn_block__divide_into_blocks_2([Instr0 | Instrs0], BlockInstrs0, PrevBlocks0,
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-vn_block__build_block_info(Instrs, Livemap, Params, ParEntries, LabelNo0,
+vn_block__build_block_info(Instrs, Livemap, Params, ParEntries, C0,
 		VnTables, Liveset, SeenIncr, Tuple) :-
 	vn_table__init_tables(VnTables0),
 	vn_block__build_from_parallel(ParEntries, VnTables0, VnTables1),
@@ -98,7 +98,7 @@ vn_block__build_block_info(Instrs, Livemap, Params, ParEntries, LabelNo0,
 	map__init(Ctrlmap0),
 	map__init(Flushmap0),
 	map__init(Parmap0),
-	Tuple0 = tuple(0, Ctrlmap0, Flushmap0, LabelNo0, Parmap0),
+	Tuple0 = tuple(0, Ctrlmap0, Flushmap0, C0, Parmap0),
 	vn_block__handle_instrs(Instrs, Livemap, Params, VnTables1, VnTables,
 		Liveset0, Liveset, no, SeenIncr, Tuple0, Tuple).
 
@@ -237,11 +237,11 @@ vn_block__handle_instr(assign(Lval, Rval),
 		Specials = LeftSpecials
 	),
 	set__insert_list(Liveset0, Specials, Liveset).
-vn_block__handle_instr(call(Proc, Return, Info, Context, CallModel),
+vn_block__handle_instr(call(Proc, Return, Info, Context, GoalPath, CallModel),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
 	vn_block__new_ctrl_node(vn_call(Proc, Return, Info, Context,
-		CallModel), Livemap, Params, VnTables0, VnTables,
+		GoalPath, CallModel), Livemap, Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(mkframe(NondetFrameInfo, Redoip), Livemap, Params,
 		VnTables0, VnTables, Liveset0, Liveset,
@@ -314,6 +314,13 @@ vn_block__handle_instr(restore_hp(Rval),
 	vn_block__new_ctrl_node(vn_restore_hp(Vn), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
+vn_block__handle_instr(free_heap(Rval),
+		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
+		SeenIncr, SeenIncr, Tuple0, Tuple) :-
+	vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables1),
+	vn_block__new_ctrl_node(vn_free_heap(Vn), Livemap,
+		Params, VnTables1, VnTables,
+		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(store_ticket(Lval),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
@@ -334,6 +341,12 @@ vn_block__handle_instr(discard_ticket,
 	vn_block__new_ctrl_node(vn_discard_ticket, Livemap,
 		Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
+vn_block__handle_instr(prune_ticket,
+		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
+		SeenIncr, SeenIncr, Tuple0, Tuple) :-
+	vn_block__new_ctrl_node(vn_prune_ticket, Livemap,
+		Params, VnTables0, VnTables,
+		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(mark_ticket_stack(Lval),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
@@ -341,11 +354,11 @@ vn_block__handle_instr(mark_ticket_stack(Lval),
 	vn_block__new_ctrl_node(vn_mark_ticket_stack(Vnlval), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(discard_tickets_to(Rval),
+vn_block__handle_instr(prune_tickets_to(Rval),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
 	vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables1),
-	vn_block__new_ctrl_node(vn_discard_tickets_to(Vn), Livemap,
+	vn_block__new_ctrl_node(vn_prune_tickets_to(Vn), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(incr_sp(N, Msg),
@@ -360,7 +373,7 @@ vn_block__handle_instr(decr_sp(N),
 	vn_block__new_ctrl_node(vn_decr_sp(N), Livemap,
 		Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(pragma_c(_, _, _, _, _, _),
+vn_block__handle_instr(pragma_c(_, _, _, _, _, _, _),
 		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
 		SeenIncr, SeenIncr, Tuple, Tuple) :-
 	error("value numbering not supported for pragma_c").
@@ -392,8 +405,8 @@ vn_block__handle_instr(join_and_continue(_, _),
 
 vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		VnTables0, VnTables, Liveset0, Liveset,
-		tuple(Ctrl0, Ctrlmap0, Flushmap0, LabelNo0, Parmap0),
-		tuple(Ctrl,  Ctrlmap,  Flushmap,  LabelNo,  Parmap)) :-
+		tuple(Ctrl0, Ctrlmap0, Flushmap0, C0, Parmap0),
+		tuple(Ctrl,  Ctrlmap,  Flushmap,  C,  Parmap)) :-
 	map__init(FlushEntry0),
 	(
 		VnInstr = vn_livevals(Livevals),
@@ -402,27 +415,27 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 			Liveset0, Liveset),
 		VnTables = VnTables0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
-		VnInstr = vn_call(_, _, _, _, _),
+		VnInstr = vn_call(_, _, _, _, _, _),
 		vn_block__record_at_call(VnTables0, VnTables, Liveset0, Liveset,
 			FlushEntry0, FlushEntry),
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_mkframe(_, _),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_label(_),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_goto(TargetAddr),
@@ -433,37 +446,44 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 			vn_block__record_one_label(Label, Livemap, Params,
 				VnTables0, VnTables, Liveset0, Liveset,
 				FlushEntry0, FlushEntry,
-				LabelNo0, LabelNo, Parallels)
+				C0, C, Parallels)
 		;
 			vn_block__record_at_call(VnTables0, VnTables,
 				Liveset0, Liveset, FlushEntry0, FlushEntry),
-			LabelNo = LabelNo0,
+			C = C0,
 			Parallels = []
 		)
 	;
 		VnInstr = vn_computed_goto(_, Labels),
 		vn_block__record_several_labels(Labels, Livemap, Params,
 			VnTables0, VnTables, Liveset0, Liveset,
-			FlushEntry0, FlushEntry, LabelNo0, LabelNo, Parallels)
+			FlushEntry0, FlushEntry, C0, C, Parallels)
 	;
 		VnInstr = vn_if_val(_, TargetAddr),
 		vn_block__new_if_node(TargetAddr, Livemap, Params,
 			Ctrlmap0, Ctrl0, VnTables0, VnTables, Liveset0, Liveset,
-			FlushEntry0, FlushEntry, LabelNo0, LabelNo, Parallels)
+			FlushEntry0, FlushEntry, C0, C, Parallels)
 	;
 		VnInstr = vn_mark_hp(Vnlval),
 		vn_util__rval_to_vn(lval(hp), Vn, VnTables0, VnTables1),
 		vn_table__set_desired_value(Vnlval, Vn, VnTables1, VnTables),
 		set__insert(Liveset0, Vnlval, Liveset),
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_restore_hp(_),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
+		Parallels = []
+	;
+		VnInstr = vn_free_heap(_),
+		VnTables = VnTables0,
+		Liveset = Liveset0,
+		FlushEntry = FlushEntry0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_store_ticket(Vnlval),
@@ -477,21 +497,28 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		vn_table__set_desired_value(Vnlval, Vn, VnTables1, VnTables),
 		set__insert(Liveset0, Vnlval, Liveset),
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_reset_ticket(_, _),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_discard_ticket,
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
+		Parallels = []
+	;
+		VnInstr = vn_prune_ticket,
+		VnTables = VnTables0,
+		Liveset = Liveset0,
+		FlushEntry = FlushEntry0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_mark_ticket_stack(Vnlval),
@@ -505,28 +532,28 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		vn_table__set_desired_value(Vnlval, Vn, VnTables1, VnTables),
 		set__insert(Liveset0, Vnlval, Liveset),
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
-		VnInstr = vn_discard_tickets_to(_),
+		VnInstr = vn_prune_tickets_to(_),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_incr_sp(_, _),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		VnInstr = vn_decr_sp(_),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	),
 	Ctrl is Ctrl0 + 1,
@@ -536,13 +563,13 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 
 :- pred vn_block__new_if_node(code_addr, livemap, vn_params, ctrlmap, int,
 	vn_tables, vn_tables, vnlvalset, vnlvalset,
-	flushmapentry, flushmapentry, int, int, list(parallel)).
+	flushmapentry, flushmapentry, counter, counter, list(parallel)).
 :- mode vn_block__new_if_node(in, in, in, in, in, in, out, in, out, in, out, in,
 	out, out) is det.
 
 vn_block__new_if_node(TargetAddr, Livemap, Params, Ctrlmap0, Ctrl0,
 		VnTables0, VnTables, Liveset0, Liveset, FlushEntry0, FlushEntry,
-		LabelNo0, LabelNo, Parallels) :-
+		C0, C, Parallels) :-
 	(
 		TargetAddr = label(Label),
 		map__search(Livemap, Label, _)
@@ -550,7 +577,7 @@ vn_block__new_if_node(TargetAddr, Livemap, Params, Ctrlmap0, Ctrl0,
 		vn_block__record_one_label(Label, Livemap, Params,
 			VnTables0, VnTables1, Liveset0, Liveset1,
 			FlushEntry0, FlushEntry1,
-			LabelNo0, LabelNoPrime, ParallelsPrime),
+			C0, CPrime, ParallelsPrime),
 		vn_block__record_compulsory_lvals(VnTables1, Liveset1, Liveset2,
 			FlushEntry1, FlushEntry2),
 		(
@@ -565,13 +592,13 @@ vn_block__new_if_node(TargetAddr, Livemap, Params, Ctrlmap0, Ctrl0,
 				VnTables1, VnTables,
 				Liveset2, Liveset,
 				FlushEntry2, FlushEntry),
-			LabelNo = LabelNo0,
+			C = C0,
 			Parallels = []
 		;
 			VnTables = VnTables1,
 			Liveset = Liveset2,
 			FlushEntry = FlushEntry2,
-			LabelNo = LabelNoPrime,
+			C = CPrime,
 			Parallels = ParallelsPrime
 		)
 	;
@@ -580,7 +607,7 @@ vn_block__new_if_node(TargetAddr, Livemap, Params, Ctrlmap0, Ctrl0,
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		TargetAddr = do_redo
@@ -588,13 +615,13 @@ vn_block__new_if_node(TargetAddr, Livemap, Params, Ctrlmap0, Ctrl0,
 		vn_block__record_compulsory_lvals(VnTables0, Liveset0, Liveset,
 			FlushEntry0, FlushEntry),
 		VnTables = VnTables0,
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	;
 		set__to_sorted_list(Liveset0, Vnlivelist),
 		vn_block__record_livevnlvals(Vnlivelist, VnTables0, VnTables,
 			Liveset0, Liveset, FlushEntry0, FlushEntry),
-		LabelNo = LabelNo0,
+		C = C0,
 		Parallels = []
 	).
 
@@ -631,7 +658,7 @@ vn_block__record_at_call(VnTables0, VnTables, Livevals0, Livevals,
 
 :- pred vn_block__record_several_labels(list(label), livemap, vn_params,
 	vn_tables, vn_tables, vnlvalset, vnlvalset,
-	flushmapentry, flushmapentry, int, int, list(parallel)).
+	flushmapentry, flushmapentry, counter, counter, list(parallel)).
 % :- mode vn_block__record_several_labels(in, in, in, di, uo, di, uo, di, uo,
 %	in, out, out) is det.
 :- mode vn_block__record_several_labels(in, in, in, in, out, in, out, in, out,
@@ -639,25 +666,25 @@ vn_block__record_at_call(VnTables0, VnTables, Livevals0, Livevals,
 
 vn_block__record_several_labels(Labels, Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals, FlushEntry0, FlushEntry,
-		LabelNo0, LabelNo, Parallels) :-
+		C0, C, Parallels) :-
 	vn_block__record_labels(Labels, Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals1, FlushEntry0, FlushEntry1,
-		LabelNo0, LabelNo, Parallels),
+		C0, C, Parallels),
 	vn_block__record_compulsory_lvals(VnTables, Livevals1, Livevals,
 		FlushEntry1, FlushEntry).
 
 :- pred vn_block__record_one_label(label, livemap, vn_params,
 	vn_tables, vn_tables, vnlvalset, vnlvalset,
-	flushmapentry, flushmapentry, int, int, list(parallel)).
+	flushmapentry, flushmapentry, counter, counter, list(parallel)).
 :- mode vn_block__record_one_label(in, in, in, in, out, in, out, in, out, in,
 	out, out) is det.
 
 vn_block__record_one_label(Label, Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals, FlushEntry0, FlushEntry,
-		LabelNo0, LabelNo, Parallels) :-
+		C0, C, Parallels) :-
 	vn_block__record_label(Label, Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals1, FlushEntry0, FlushEntry1,
-		LabelNo0, LabelNo, Parallels),
+		C0, C, Parallels),
 	vn_block__record_compulsory_lvals(VnTables, Livevals1, Livevals,
 		FlushEntry1, FlushEntry).
 
@@ -665,28 +692,28 @@ vn_block__record_one_label(Label, Livemap, Params, VnTables0, VnTables,
 
 :- pred vn_block__record_labels(list(label), livemap, vn_params,
 	vn_tables, vn_tables, vnlvalset, vnlvalset,
-	flushmapentry, flushmapentry, int, int, list(parallel)).
+	flushmapentry, flushmapentry, counter, counter, list(parallel)).
 % :- mode vn_block__record_labels(in, in, in, di, uo, di, uo, di, uo, in,
 %	 out, out) is det.
 :- mode vn_block__record_labels(in, in, in, in, out, in, out, in, out, in,
 	out, out) is det.
 
 vn_block__record_labels([], _, _, VnTables, VnTables, Livevals, Livevals,
-	FlushEntry, FlushEntry, LabelNo, LabelNo, []).
+	FlushEntry, FlushEntry, C, C, []).
 vn_block__record_labels([Label | Labels], Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals, FlushEntry0, FlushEntry,
-		LabelNo0, LabelNo, Parallels) :-
+		C0, C, Parallels) :-
 	vn_block__record_label(Label, Livemap, Params, VnTables0, VnTables1,
 		Livevals0, Livevals1, FlushEntry0, FlushEntry1,
-		LabelNo0, LabelNo1, Parallels0),
+		C0, C1, Parallels0),
 	vn_block__record_labels(Labels, Livemap, Params, VnTables1, VnTables,
 		Livevals1, Livevals, FlushEntry1, FlushEntry,
-		LabelNo1, LabelNo, Parallels1),
+		C1, C, Parallels1),
 	list__append(Parallels0, Parallels1, Parallels).
 
 :- pred vn_block__record_label(label, livemap, vn_params, vn_tables, vn_tables,
 	vnlvalset, vnlvalset, flushmapentry, flushmapentry,
-	int, int, list(parallel)).
+	counter, counter, list(parallel)).
 % :- mode vn_block__record_label(in, in, in, di, uo, di, uo, di, uo, in,
 %	 out, out) is det.
 :- mode vn_block__record_label(in, in, in, in, out, in, out, in, out, in,
@@ -694,23 +721,23 @@ vn_block__record_labels([Label | Labels], Livemap, Params, VnTables0, VnTables,
 
 vn_block__record_label(Label, Livemap, Params, VnTables0, VnTables,
 		Livevals0, Livevals, FlushEntry0, FlushEntry,
-		LabelNo0, LabelNo, Parallels) :-
+		C0, C, Parallels) :-
 	( map__search(Livemap, Label, Liveset) ->
 		set__to_sorted_list(Liveset, Livelist),
 		vn_block__record_livevals(Livelist, Params, VnTables0, VnTables,
 			Livevals0, Livevals, FlushEntry0, FlushEntry,
 			ParEntries),
 		( ParEntries = [] ->
-			LabelNo = LabelNo0,
+			C = C0,
 			Parallels = []
 		;
-			( Label = local(ProcLabel, _) ->
-				LabelNo is LabelNo0 + 1,
-				NewLabel = local(ProcLabel, LabelNo),
+			( Label = local(_, ProcLabel) ->
+				counter__allocate(N, C0, C),
+				NewLabel = local(N, ProcLabel),
 				Parallels = [parallel(Label, NewLabel,
 					ParEntries)]
 			;
-				LabelNo = LabelNo0,
+				C = C0,
 				Parallels = []
 			)
 		)
@@ -897,7 +924,7 @@ vn_block__is_ctrl_instr(comment(_), no).
 vn_block__is_ctrl_instr(livevals(_), yes).
 vn_block__is_ctrl_instr(block(_, _, _), no).
 vn_block__is_ctrl_instr(assign(_, _), no).
-vn_block__is_ctrl_instr(call(_, _, _, _, _), yes).
+vn_block__is_ctrl_instr(call(_, _, _, _, _, _), yes).
 vn_block__is_ctrl_instr(mkframe(_, _), yes).
 vn_block__is_ctrl_instr(label(_), yes).
 vn_block__is_ctrl_instr(goto(_), yes).
@@ -907,18 +934,20 @@ vn_block__is_ctrl_instr(if_val(_, _), yes).
 vn_block__is_ctrl_instr(incr_hp(_, _, _, _), no).
 vn_block__is_ctrl_instr(mark_hp(_), yes).
 vn_block__is_ctrl_instr(restore_hp(_), yes).
+vn_block__is_ctrl_instr(free_heap(_), yes).
 vn_block__is_ctrl_instr(store_ticket(_), yes).
 vn_block__is_ctrl_instr(reset_ticket(_, _), yes).
 vn_block__is_ctrl_instr(discard_ticket, yes).
+vn_block__is_ctrl_instr(prune_ticket, yes).
 vn_block__is_ctrl_instr(mark_ticket_stack(_), yes).
-vn_block__is_ctrl_instr(discard_tickets_to(_), yes).
+vn_block__is_ctrl_instr(prune_tickets_to(_), yes).
 vn_block__is_ctrl_instr(incr_sp(_, _), yes).
 vn_block__is_ctrl_instr(decr_sp(_), yes).
 vn_block__is_ctrl_instr(init_sync_term(_, _), no).
 vn_block__is_ctrl_instr(fork(_, _, _), yes).
 vn_block__is_ctrl_instr(join_and_terminate(_), yes).
 vn_block__is_ctrl_instr(join_and_continue(_, _), yes).
-vn_block__is_ctrl_instr(pragma_c(_, _, _, _, _, _), no).
+vn_block__is_ctrl_instr(pragma_c(_, _, _, _, _, _, _), no).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
