@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -16,7 +16,7 @@
 :- interface.
 
 :- import_module vn_type, vn_table, vn_temploc.
-:- import_module llds, list.
+:- import_module llds, list, io.
 
 	% Flush the given nodes in the given order.
 
@@ -186,11 +186,11 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		Instrs = [livevals(Livevals) - ""]
 	;
 		Vn_instr = vn_call(ProcAddr, RetAddr, LiveInfo,
-			Context, CodeModel),
+			Context, GoalPath, CodeModel),
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		Instrs = [call(ProcAddr, RetAddr, LiveInfo,
-			Context, CodeModel) - ""]
+			Context, GoalPath, CodeModel) - ""]
 	;
 		Vn_instr = vn_mkframe(NondetFrameInfo, Redoip),
 		vn_util__rval_to_vn(const(code_addr_const(Redoip)), AddrVn,
@@ -241,6 +241,12 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		vn_table__set_current_value(vn_hp, Vn, VnTables1, VnTables),
 		Instr = restore_hp(Rval) - "",
 		list__append(FlushInstrs, [Instr], Instrs)
+	;
+		Vn_instr = vn_free_heap(Vn),
+		vn_flush__vn(Vn, [src_ctrl(N)], [], Rval, VnTables0, VnTables,
+			Templocs0, Templocs, Params, FlushInstrs),
+		Instr = free_heap(Rval) - "",
+		list__append(FlushInstrs, [Instr], Instrs)
 	;
 		Vn_instr = vn_store_ticket(Vnlval),
 		vn_flush__access_path(Vnlval, [src_ctrl(N)], [], Lval,
@@ -262,6 +268,11 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		Instrs = [discard_ticket - ""]
+	;
+		Vn_instr = vn_prune_ticket,
+		VnTables = VnTables0,
+		Templocs = Templocs0,
+		Instrs = [prune_ticket - ""]
 	;
 		Vn_instr = vn_mark_ticket_stack(Vnlval),
 		vn_flush__access_path(Vnlval, [src_ctrl(N)], [], Lval,
@@ -273,10 +284,10 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		Instr = mark_ticket_stack(Lval) - "",
 		list__append(FlushInstrs, [Instr], Instrs)
 	;
-		Vn_instr = vn_discard_tickets_to(Vn),
+		Vn_instr = vn_prune_tickets_to(Vn),
 		vn_flush__vn(Vn, [src_ctrl(N)], [], Rval, VnTables0, VnTables,
 			Templocs0, Templocs, Params, FlushInstrs),
-		Instr = discard_tickets_to(Rval) - "",
+		Instr = prune_tickets_to(Rval) - "",
 		list__append(FlushInstrs, [Instr], Instrs)
 	;
 		Vn_instr = vn_incr_sp(Incr, Msg),
@@ -693,7 +704,9 @@ vn_flush__vn_value(Vn, Srcs, Forbidden, Rval, VnTables0, VnTables,
 	;
 		Vnrval = vn_create(Tag, MaybeRvals, ArgTypes, StatDyn,
 			Label, Msg),
-		Rval = create(Tag, MaybeRvals, ArgTypes, StatDyn, Label, Msg),
+		Reuse = no,
+		Rval = create(Tag, MaybeRvals, ArgTypes, StatDyn,
+			Label, Msg, Reuse),
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		Instrs = []
