@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -148,7 +148,7 @@ vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables) :-
 		Rval = var(_),
 		error("value_number should never get rval: var")
 	;
-		Rval = create(Tag, Args, ArgTypes, StatDyn, Label, Msg),
+		Rval = create(Tag, Args, ArgTypes, StatDyn, Label, Msg, _),
 		vn_util__vnrval_to_vn(vn_create(Tag, Args, ArgTypes,
 			StatDyn, Label, Msg), Vn, VnTables0, VnTables)
 	;
@@ -217,14 +217,10 @@ vn_util__vnrval_to_vn(Vnrval, Vn, VnTables0, VnTables) :-
 :- pred vn_util__vnrval_may_share_vn(vnrval).
 :- mode vn_util__vnrval_may_share_vn(in) is semidet.
 
-vn_util__vnrval_may_share_vn(Vnrval) :-
-	(
-		Vnrval = vn_unop(cast_to_unsigned, _)
-	->
-		fail
-	;
-		true
-	).
+vn_util__vnrval_may_share_vn(_Vnrval) :-
+	% Currently any vnrval may share vn.
+	% We used to fail here for `cast_to_unsigned' operators.
+	semidet_succeed.
 
 	% Simplify the vnrval by partially evaluating expressions involving
 	% constants. To make this simpler, swap the arguments of commutative
@@ -773,6 +769,11 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 		vn_util__simplify_int_compare_op(<, Vn1, Vnrval1, Vn2, Vnrval2,
 			false, Vnrval),
 		VnTables = VnTables0
+	;
+		Binop = unsigned_le,
+		vn_util__simplify_int_compare_op(unsigned_le, Vn1, Vnrval1,
+			Vn2, Vnrval2, true, Vnrval),
+		VnTables = VnTables0
 	;
 		Binop = (float_eq),
 		vn_util__simplify_float_compare_op(float_eq, Vn1, Vnrval1,
@@ -907,6 +908,25 @@ vn_util__simplify_float_compare_op(ComparePred, Vn1, Vnrval1, Vn2, Vnrval2,
 		vn_util__const_if_equal_vns(Vn1, Vn2, ResultIfEqual, Vnrval)
 	).
 
+	% unsigned_le(X, Y) performs an unsigned less-than-or-equal
+	% comparison.  It should succeed iff the C expression
+	% `(MR_Unsigned)(X) < (MR_Unsigned)(Y)' evaluates to true.
+:- pred unsigned_le(int::in, int::in) is semidet.
+unsigned_le(X, Y) :-
+	( X < 0, Y >= 0 ->
+		% when cast to unsigned, X will be a big positive number,
+		% and Y will be smaller, so we must fail
+		fail
+	; X >= 0, Y < 0 ->
+		% when cast to unsigned, Y will be a big positive number,
+		% and X will be smaller, so we must succeed
+		true
+	;
+		% if they both have the same sign, then we can just do
+		% an ordinary comparison.
+		X =< Y
+	).
+
 :- pred float_eq(float::in, float::in) is semidet.
 float_eq(X, X).
 
@@ -1076,7 +1096,7 @@ vn_util__find_lvals_in_rval(Rval, Lvals) :-
 		Rval = var(_),
 		error("var found in vn_util__find_lvals_in_rval")
 	;
-		Rval = create(_, _, _, _, _, _),
+		Rval = create(_, _, _, _, _, _, _),
 		Lvals = []
 	;
 		Rval = mkword(_, Rval1),
@@ -1228,7 +1248,7 @@ vn_util__build_uses_from_ctrl(Ctrl, Ctrlmap, VnTables0, VnTables) :-
 			VnInstr = vn_livevals(_),
 			VnTables1 = VnTables0
 		;
-			VnInstr = vn_call(_, _, _, _, _),
+			VnInstr = vn_call(_, _, _, _, _, _),
 			VnTables1 = VnTables0
 		;
 			VnInstr = vn_mkframe(_, _),
@@ -1256,6 +1276,10 @@ vn_util__build_uses_from_ctrl(Ctrl, Ctrlmap, VnTables0, VnTables) :-
 			VnInstr = vn_restore_hp(Vn),
 			vn_util__record_use(Vn, src_ctrl(Ctrl),
 				VnTables0, VnTables1)
+		;
+			VnInstr = vn_free_heap(Vn),
+			vn_util__record_use(Vn, src_ctrl(Ctrl),
+				VnTables0, VnTables1)
 		;
 			VnInstr = vn_store_ticket(Vnlval),
 			vn_util__vnlval_access_vns(Vnlval, Vns),
@@ -1268,13 +1292,16 @@ vn_util__build_uses_from_ctrl(Ctrl, Ctrlmap, VnTables0, VnTables) :-
 		;
 			VnInstr = vn_discard_ticket,
 			VnTables1 = VnTables0
+		;
+			VnInstr = vn_prune_ticket,
+			VnTables1 = VnTables0
 		;
 			VnInstr = vn_mark_ticket_stack(Vnlval),
 			vn_util__vnlval_access_vns(Vnlval, Vns),
 			vn_util__record_use_list(Vns, src_ctrl(Ctrl),
 				VnTables0, VnTables1)
 		;
-			VnInstr = vn_discard_tickets_to(Vn),
+			VnInstr = vn_prune_tickets_to(Vn),
 			vn_util__record_use(Vn, src_ctrl(Ctrl),
 				VnTables0, VnTables1)
 		;
