@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1996-1997 The University of Melbourne.
+% Copyright (C) 1996-1997,2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -17,7 +17,7 @@
 
 :- interface.
 
-:- import_module list.
+:- import_module enum, list.
 
 %-----------------------------------------------------------------------------%
 
@@ -28,18 +28,25 @@
 
 :- type bool ---> no ; yes.
 
+:- instance enum(bool).
+
+:- func bool__or(bool, bool) = bool.
 :- pred bool__or(bool, bool, bool).
 :- mode bool__or(in, in, out) is det.
 
+:- func bool__or_list(list(bool)) = bool.
 :- pred bool__or_list(list(bool), bool).
 :- mode bool__or_list(in, out) is det.
 
+:- func bool__and(bool, bool) = bool.
 :- pred bool__and(bool, bool, bool).
 :- mode bool__and(in, in, out) is det.
 
+:- func bool__and_list(list(bool)) = bool.
 :- pred bool__and_list(list(bool), bool).
 :- mode bool__and_list(in, out) is det.
 
+:- func bool__not(bool) = bool.
 :- pred bool__not(bool, bool).
 :- mode bool__not(in, out) is det.
 
@@ -47,9 +54,25 @@
 
 :- implementation.
 
+:- instance enum(bool) where [
+	to_int(Bool) = bool_to_int(Bool),
+	from_int(bool_to_int(Bool)) = Bool
+].
+
+:- func bool_to_int(bool) = int.
+:- mode bool_to_int(in) = out is det.
+:- mode bool_to_int(out) = in is semidet.
+
+bool_to_int(no) = 0.
+bool_to_int(yes) = 1.
+
+bool__or(X, Y) = Result :- bool__or(X, Y, Result).
+
 bool__or(yes, _, yes).
 bool__or(no, Bool, Bool).
 
+bool__or_list(List) = Result :- bool__or_list(List, Result).
+
 bool__or_list([], no).
 bool__or_list([Bool | Bools], Result) :-
 	( Bool = yes ->
@@ -58,9 +81,13 @@ bool__or_list([Bool | Bools], Result) :-
 		bool__or_list(Bools, Result)
 	).
 
+bool__and(X, Y) = Result :- bool__and(X, Y, Result).
+
 bool__and(no, _, no).
 bool__and(yes, Bool, Bool).
 
+bool__and_list(List) = Result :- bool__and_list(List, Result).
+
 bool__and_list([], yes).
 bool__and_list([Bool | Bools], Result) :-
 	( Bool = no ->
@@ -69,6 +96,8 @@ bool__and_list([Bool | Bools], Result) :-
 		bool__and_list(Bools, Result)
 	).
 
+bool__not(X) = Result :- bool__not(X, Result).
+
 bool__not(no, yes).
 bool__not(yes, no).
 
