@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995-1999 The University of Melbourne.
+% Copyright (C) 1995-2001 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -77,7 +77,7 @@
 :- mode math__truncate(in) = out is det.
 
 %---------------------------------------------------------------------------%
-% Power/logarithm operations
+% Polynomial roots
 
 	% math__sqrt(X) = Sqrt is true if Sqrt is the positive square
 	% root of X.
@@ -86,6 +86,21 @@
 :- func math__sqrt(float) = float.
 :- mode math__sqrt(in) = out is det.
 
+:- type math__quadratic_roots
+	--->	no_roots
+	;	one_root(float)
+	;	two_roots(float, float).
+
+	% math__solve_quadratic(A, B, C) = Roots is true if Roots are
+	% the solutions to the equation Ax^2 + Bx + C.
+	%
+	% Domain restriction: A \= 0
+:- func math__solve_quadratic(float, float, float) = quadratic_roots.
+:- mode math__solve_quadratic(in, in, in) = out is det.
+
+%---------------------------------------------------------------------------%
+% Power/logarithm operations
+
 	% math__pow(X, Y) = Res is true if Res is X raised to the
 	% power of Y.
 	%
@@ -187,10 +202,11 @@
 %---------------------------------------------------------------------------%
 
 :- implementation.
+:- import_module float.
 
-% These operations are all implemented using the C interface.
+% These operations are mostly implemented using the C interface.
 
-:- pragma c_header_code("
+:- pragma foreign_decl("C", "
 
 	#include <math.h>
 
@@ -209,9 +225,17 @@
 
 	void ML_math_domain_error(const char *where);
 
-"). % end pragma c_header_code
+"). % end pragma foreign_decl
 
-:- pragma c_code("
+:- pragma foreign_decl("MC++", "
+
+	// This is not defined in the .NET Frameworks.
+	// For pi and e we use the constants defined in System::Math.
+
+	#define	ML_FLOAT_LN2		0.69314718055994530941
+").
+
+:- pragma foreign_code("C", "
 
 	#include ""mercury_trace_base.h""
 	#include <stdio.h>
@@ -227,58 +251,104 @@
 			""Software error: Domain error in call to `%s'\\n"",
 			where);
 		MR_trace_report(stderr);
+	#ifndef MR_HIGHLEVEL_CODE
 		MR_dump_stack(MR_succip, MR_sp, MR_curfr, FALSE);
+	#endif
 		exit(1);
 	}
 
-"). % end pragma c_code
+"). % end pragma foreign_code
+
+:- pragma foreign_code("MC++", "
+
+/*
+** Handle domain errors.
+*/
+static void
+ML_math_domain_error(MR_String where)
+{
+	throw new mercury::runtime::Exception(where);
+}
+
+"). % end pragma foreign_code
 
 %
 % Mathematical constants from math.m
 %
 	% Pythagoras' number
-:- pragma c_code(math__pi = (Pi::out), [will_not_call_mercury, thread_safe],"
+:- pragma foreign_proc("C", 
+	math__pi = (Pi::out), [will_not_call_mercury, thread_safe],"
 	Pi = ML_FLOAT_PI;
 ").
+:- pragma foreign_proc("MC++", 
+	math__pi = (Pi::out), [will_not_call_mercury, thread_safe],"
+	Pi = System::Math::PI;
+").
 
 	% Base of natural logarithms
-:- pragma c_code(math__e = (E::out), [will_not_call_mercury, thread_safe],"
+:- pragma foreign_proc("C", 
+	math__e = (E::out), [will_not_call_mercury, thread_safe],"
 	E = ML_FLOAT_E;
 ").
+:- pragma foreign_proc("MC++", 
+	math__e = (E::out), [will_not_call_mercury, thread_safe],"
+	E = System::Math::E;
+").
 
 %
 % math__ceiling(X) = Ceil is true if Ceil is the smallest integer
 % not less than X.
 %
-:- pragma c_code(math__ceiling(Num::in) = (Ceil::out),
+:- pragma foreign_proc("C", 
+	math__ceiling(Num::in) = (Ceil::out),
 		[will_not_call_mercury, thread_safe],"
 	Ceil = ceil(Num);
 ").
+:- pragma foreign_proc("MC++", 
+	math__ceiling(Num::in) = (Ceil::out),
+		[will_not_call_mercury, thread_safe],"
+	Ceil = System::Math::Ceil(Num);
+").
 
 %
 % math__floor(X) = Floor is true if Floor is the largest integer
 % not greater than X.
 %
-:- pragma c_code(math__floor(Num::in) = (Floor::out),
+:- pragma foreign_proc("C", 
+	math__floor(Num::in) = (Floor::out),
 		[will_not_call_mercury, thread_safe],"
 	Floor = floor(Num);
 ").
+:- pragma foreign_proc("MC++", 
+	math__floor(Num::in) = (Floor::out),
+		[will_not_call_mercury, thread_safe],"
+	Floor = System::Math::Floor(Num);
+").
 
 %
 % math__round(X) = Round is true if Round is the integer
 % closest to X.  If X has a fractional component of 0.5,
 % it is rounded up.
 %
-:- pragma c_code(math__round(Num::in) = (Rounded::out),
+:- pragma foreign_proc("C", 
+	math__round(Num::in) = (Rounded::out),
 		[will_not_call_mercury, thread_safe],"
 	Rounded = floor(Num+0.5);
 ").
+:- pragma foreign_proc("MC++", 
+	math__round(Num::in) = (Rounded::out),
+		[will_not_call_mercury, thread_safe],"
+	// XXX the semantics of System::Math::Round() are not the same as ours.
+	// Unfortunately they are better (round to nearest even number).
+	Rounded = System::Math::Floor(Num+0.5);
+").
 
 %
 % math__truncate(X) = Trunc is true if Trunc is the integer
 % closest to X such that |Trunc| =< |X|.
 %
-:- pragma c_code(math__truncate(X::in) = (Trunc::out),
+:- pragma foreign_proc("C",
+	math__truncate(X::in) = (Trunc::out),
 		[will_not_call_mercury, thread_safe],"
 	if (X < 0.0) {
 		Trunc = ceil(X);
@@ -286,6 +356,15 @@
 		Trunc = floor(X);
 	}
 ").
+:- pragma foreign_proc("MC++",
+	math__truncate(X::in) = (Trunc::out),
+		[will_not_call_mercury, thread_safe],"
+	if (X < 0.0) {
+		Trunc = System::Math::Ceil(X);
+	} else {
+		Trunc = System::Math::Floor(X);
+	}
+").
 
 %
 % math__sqrt(X) = Sqrt is true if Sqrt is the positive square
@@ -294,7 +373,7 @@
 % Domain restrictions:
 %		X >= 0
 %
-:- pragma c_code(math__sqrt(X::in) = (SquareRoot::out),
+:- pragma foreign_proc("C", math__sqrt(X::in) = (SquareRoot::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
@@ -303,6 +382,59 @@
 #endif
 	SquareRoot = sqrt(X);
 ").
+:- pragma foreign_proc("MC++", math__sqrt(X::in) = (SquareRoot::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X < 0.0) {
+		ML_math_domain_error(""math__sqrt"");
+	}
+#endif
+	SquareRoot = System::Math::Sqrt(X);
+").
+
+
+%
+% math__solve_quadratic(A, B, C) = Roots is true if Roots are
+% the solutions to the equation Ax^2 + Bx + C.
+%
+% Domain restrictions:
+% 		A \= 0
+%
+math__solve_quadratic(A, B, C) = Roots :-
+	%
+	% This implementation is designed to minimise numerical errors;
+	% it is adapted from "Numerical recipes in C".
+	%
+	DSquared = B * B - 4.0 * A * C,
+	compare(CmpD, DSquared, 0.0),
+	(
+		CmpD = (<),
+		Roots = no_roots
+	;
+		CmpD = (=),
+		Root = -0.5 * B / A,
+		Roots = one_root(Root)
+	;
+		CmpD = (>),
+		D = sqrt(DSquared),
+		compare(CmpB, B, 0.0),
+		(
+			CmpB = (<),
+			Q = -0.5 * (B - D),
+			Root1 = Q / A,
+			Root2 = C / Q
+		;
+			CmpB = (=),
+			Root1 = -0.5 * D / A,
+			Root2 = -Root1
+		;
+			CmpB = (>),
+			Q = -0.5 * (B + D),
+			Root1 = Q / A,
+			Root2 = C / Q
+		),
+		Roots = two_roots(Root1, Root2)
+	).
 
 %
 % math__pow(X, Y) = Res is true if Res is X raised to the
@@ -312,7 +444,7 @@
 %		X >= 0
 %		X = 0 implies Y > 0
 %
-:- pragma c_code(math__pow(X::in, Y::in) = (Res::out),
+:- pragma foreign_proc("C", math__pow(X::in, Y::in) = (Res::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < 0.0) {
@@ -331,14 +463,38 @@
 #endif
 ").
 
+:- pragma foreign_proc("MC++", math__pow(X::in, Y::in) = (Res::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X < 0.0) {
+		ML_math_domain_error(""math__pow"");
+	}
+	if (X == 0.0) {
+		if (Y <= 0.0) {
+			ML_math_domain_error(""math__pow"");
+		}
+		Res = 0.0;
+	} else {
+		Res = System::Math::Pow(X, Y);
+	}
+#else
+	Res = System::Math::Pow(X, Y);
+#endif
+").
+
+
 %
 % math__exp(X) = Exp is true if Exp is X raised to the
 % power of e.
 %
-:- pragma c_code(math__exp(X::in) = (Exp::out),
+:- pragma foreign_proc("C", math__exp(X::in) = (Exp::out),
 		[will_not_call_mercury, thread_safe],"
 	Exp = exp(X);
 ").
+:- pragma foreign_proc("MC++", math__exp(X::in) = (Exp::out),
+		[will_not_call_mercury, thread_safe],"
+	Exp = System::Math::Exp(X);
+").
 
 %
 % math__ln(X) = Log is true if Log is the natural logarithm
@@ -347,7 +503,7 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__ln(X::in) = (Log::out),
+:- pragma foreign_proc("C", math__ln(X::in) = (Log::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
@@ -356,6 +512,15 @@
 #endif
 	Log = log(X);
 ").
+:- pragma foreign_proc("MC++", math__ln(X::in) = (Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0) {
+		ML_math_domain_error(""math__ln"");
+	}
+#endif
+	Log = System::Math::Log(X);
+").
 
 %
 % math__log10(X) = Log is true if Log is the logarithm to
@@ -364,7 +529,7 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__log10(X::in) = (Log10::out),
+:- pragma foreign_proc("C", math__log10(X::in) = (Log10::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
@@ -373,6 +538,15 @@
 #endif
 	Log10 = log10(X);
 ").
+:- pragma foreign_proc("MC++", math__log10(X::in) = (Log10::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0) {
+		ML_math_domain_error(""math__log10"");
+	}
+#endif
+	Log10 = System::Math::Log10(X);
+").
 
 %
 % math__log2(X) = Log is true if Log is the logarithm to
@@ -381,7 +555,7 @@
 % Domain restrictions:
 %		X > 0
 %
-:- pragma c_code(math__log2(X::in) = (Log2::out),
+:- pragma foreign_proc("C", math__log2(X::in) = (Log2::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0) {
@@ -390,6 +564,15 @@
 #endif
 	Log2 = log(X) / ML_FLOAT_LN2;
 ").
+:- pragma foreign_proc("MC++", math__log2(X::in) = (Log2::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0) {
+		ML_math_domain_error(""math__log2"");
+	}
+#endif
+	Log2 = System::Math::Log(X) / ML_FLOAT_LN2;
+").
 
 %
 % math__log(B, X) = Log is true if Log is the logarithm to
@@ -400,7 +583,7 @@
 %		B > 0
 %		B \= 1
 %
-:- pragma c_code(math__log(B::in, X::in) = (Log::out),
+:- pragma foreign_proc("C", math__log(B::in, X::in) = (Log::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X <= 0.0 || B <= 0.0) {
@@ -412,30 +595,56 @@
 #endif
 	Log = log(X)/log(B);
 ").
+:- pragma foreign_proc("MC++", math__log(B::in, X::in) = (Log::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X <= 0.0 || B <= 0.0) {
+		ML_math_domain_error(""math__log"");
+	}
+	if (B == 1.0) {
+		ML_math_domain_error(""math__log"");
+	}
+#endif
+	Log = System::Math::Log(X,B);
+").
+
 
 %
 % math__sin(X) = Sin is true if Sin is the sine of X.
 %
-:- pragma c_code(math__sin(X::in) = (Sin::out),
+:- pragma foreign_proc("C", math__sin(X::in) = (Sin::out),
 		[will_not_call_mercury, thread_safe],"
 	Sin = sin(X);
 ").
+:- pragma foreign_proc("MC++", math__sin(X::in) = (Sin::out),
+		[will_not_call_mercury, thread_safe],"
+	Sin = System::Math::Sin(X);
+").
+
 
 %
 % math__cos(X) = Sin is true if Cos is the cosine of X.
 %
-:- pragma c_code(math__cos(X::in) = (Cos::out),
+:- pragma foreign_proc("C", math__cos(X::in) = (Cos::out),
 		[will_not_call_mercury, thread_safe],"
 	Cos = cos(X);
 ").
+:- pragma foreign_proc("MC++", math__cos(X::in) = (Cos::out),
+		[will_not_call_mercury, thread_safe],"
+	Cos = System::Math::Cos(X);
+").
 
 %
 % math__tan(X) = Tan is true if Tan is the tangent of X.
 %
-:- pragma c_code(math__tan(X::in) = (Tan::out),
+:- pragma foreign_proc("C", math__tan(X::in) = (Tan::out),
 		[will_not_call_mercury, thread_safe],"
 	Tan = tan(X);
 ").
+:- pragma foreign_proc("MC++", math__tan(X::in) = (Tan::out),
+		[will_not_call_mercury, thread_safe],"
+	Tan = System::Math::Tan(X);
+").
 
 %
 % math__asin(X) = ASin is true if ASin is the inverse
@@ -444,7 +653,7 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma c_code(math__asin(X::in) = (ASin::out),
+:- pragma foreign_proc("C", math__asin(X::in) = (ASin::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
@@ -453,6 +662,15 @@
 #endif
 	ASin = asin(X);
 ").
+:- pragma foreign_proc("MC++", math__asin(X::in) = (ASin::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X < -1.0 || X > 1.0) {
+		ML_math_domain_error(""math__asin"");
+	}
+#endif
+	ASin = System::Math::Asin(X);
+").
 
 %
 % math__acos(X) = ACos is true if ACos is the inverse
@@ -461,7 +679,7 @@
 % Domain restrictions:
 %		X must be in the range [-1,1]
 %
-:- pragma c_code(math__acos(X::in) = (ACos::out),
+:- pragma foreign_proc("C", math__acos(X::in) = (ACos::out),
 		[will_not_call_mercury, thread_safe], "
 #ifndef ML_OMIT_MATH_DOMAIN_CHECKS
 	if (X < -1.0 || X > 1.0) {
@@ -470,51 +688,81 @@
 #endif
 	ACos = acos(X);
 ").
+:- pragma foreign_proc("MC++", math__acos(X::in) = (ACos::out),
+		[will_not_call_mercury, thread_safe], "
+#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
+	if (X < -1.0 || X > 1.0) {
+		ML_math_domain_error(""math__acos"");
+	}
+#endif
+	ACos = System::Math::Acos(X);
+").
+
 
 %
 % math__atan(X) = ATan is true if ATan is the inverse
 % tangent of X, where ATan is in the range [-pi/2,pi/2].
 %
-:- pragma c_code(math__atan(X::in) = (ATan::out),
+:- pragma foreign_proc("C", math__atan(X::in) = (ATan::out),
 		[will_not_call_mercury, thread_safe],"
 	ATan = atan(X);
 ").
+:- pragma foreign_proc("MC++", math__atan(X::in) = (ATan::out),
+		[will_not_call_mercury, thread_safe],"
+	ATan = System::Math::Atan(X);
+").
 
 %
 % math__atan2(Y, X) = ATan is true if ATan is the inverse
 % tangent of Y/X, where ATan is in the range [-pi,pi].
 %
-:- pragma c_code(math__atan2(Y::in, X::in) = (ATan2::out), 
+:- pragma foreign_proc("C", math__atan2(Y::in, X::in) = (ATan2::out), 
 		[will_not_call_mercury, thread_safe], "
 	ATan2 = atan2(Y, X);
 ").
+:- pragma foreign_proc("MC++", math__atan2(Y::in, X::in) = (ATan2::out), 
+		[will_not_call_mercury, thread_safe], "
+	ATan2 = System::Math::Atan2(Y, X);
+").
 
 %
 % math__sinh(X) = Sinh is true if Sinh is the hyperbolic
 % sine of X.
 %
-:- pragma c_code(math__sinh(X::in) = (Sinh::out),
+:- pragma foreign_proc("C", math__sinh(X::in) = (Sinh::out),
 		[will_not_call_mercury, thread_safe],"
 	Sinh = sinh(X);
 ").
+:- pragma foreign_proc("MC++", math__sinh(X::in) = (Sinh::out),
+		[will_not_call_mercury, thread_safe],"
+	Sinh = System::Math::Sinh(X);
+").
 
 %
 % math__cosh(X) = Cosh is true if Cosh is the hyperbolic
 % cosine of X.
 %
-:- pragma c_code(math__cosh(X::in) = (Cosh::out),
+:- pragma foreign_proc("C", math__cosh(X::in) = (Cosh::out),
 		[will_not_call_mercury, thread_safe],"
 	Cosh = cosh(X);
 ").
+:- pragma foreign_proc("MC++", math__cosh(X::in) = (Cosh::out),
+		[will_not_call_mercury, thread_safe],"
+	Cosh = System::Math::Cosh(X);
+").
 
 %
 % math__tanh(X) = Tanh is true if Tanh is the hyperbolic
 % tangent of X.
 %
-:- pragma c_code(math__tanh(X::in) = (Tanh::out),
+:- pragma foreign_proc("C", math__tanh(X::in) = (Tanh::out),
 		[will_not_call_mercury, thread_safe],"
 	Tanh = tanh(X);
 ").
+:- pragma foreign_proc("MC++", math__tanh(X::in) = (Tanh::out),
+		[will_not_call_mercury, thread_safe],"
+	Tanh = System::Math::Tanh(X);
+").
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
@@ -699,287 +947,32 @@
 
 :- implementation.
 
-% These operations are all implemented using the C interface.
+% These operations are all implemented in terms of the functional versions.
 
 
-%
-% Mathematical constants from math.m
-%
-	% Pythagoras' number
-:- pragma c_code(math__pi(Pi::out), [will_not_call_mercury, thread_safe],
-		"Pi = ML_FLOAT_PI;").
-
-	% Base of natural logarithms
-:- pragma c_code(math__e(E::out), [will_not_call_mercury, thread_safe],
-		"E = ML_FLOAT_E;").
-
-%
-% math__ceiling(X, Ceil) is true if Ceil is the smallest integer
-% not less than X.
-%
-:- pragma c_code(math__ceiling(Num::in, Ceil::out),
-		[will_not_call_mercury, thread_safe],
-	"Ceil = ceil(Num);").
-
-%
-% math__floor(X, Floor) is true if Floor is the largest integer
-% not greater than X.
-%
-:- pragma c_code(math__floor(Num::in, Floor::out),
-		[will_not_call_mercury, thread_safe],
-	"Floor = floor(Num);").
-
-%
-% math__round(X, Round) is true if Round is the integer
-% closest to X.  If X has a fractional component of 0.5,
-% it is rounded up.
-%
-:- pragma c_code(math__round(Num::in, Rounded::out),
-		[will_not_call_mercury, thread_safe], "
-	Rounded = floor(Num+0.5);
-").
-
-%
-% math__truncate(X, Trunc) is true if Trunc is the integer
-% closest to X such that |Trunc| =< |X|.
-%
-:- pragma c_code(math__truncate(X::in, Trunc::out),
-		[will_not_call_mercury, thread_safe], "
-	if (X < 0.0) {
-		Trunc = ceil(X);
-	} else {
-		Trunc = floor(X);
-	}
-").
-
-%
-% math__sqrt(X, Sqrt) is true if Sqrt is the positive square
-% root of X.  
-%
-% Domain restrictions:
-%		X >= 0
-%
-:- pragma c_code(math__sqrt(X::in, SquareRoot::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X < 0.0) {
-		ML_math_domain_error(""math__sqrt"");
-	}
-#endif
-	SquareRoot = sqrt(X);
-").
-
-%
-% math__pow(X, Y, Res) is true if Res is X raised to the
-% power of Y.
-%
-% Domain restrictions:
-%		X >= 0
-%		X = 0 implies Y > 0
-%
-:- pragma c_code(math__pow(X::in, Y::in, Res::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X < 0.0) {
-		ML_math_domain_error(""math__pow"");
-	}
-	if (X == 0.0) {
-		if (Y <= 0.0) {
-			ML_math_domain_error(""math__pow"");
-		}
-		Res = 0.0;
-	} else {
-		Res = pow(X, Y);
-	}
-#else
-	Res = pow(X, Y);
-#endif
-").
-
-%
-% math__exp(X, Exp) is true if Exp is X raised to the
-% power of e.
-%
-:- pragma c_code(math__exp(X::in, Exp::out),
-		[will_not_call_mercury, thread_safe], "
-	Exp = exp(X);
-").
-
-%
-% math__ln(X, Log) is true if Log is the natural logarithm
-% of X.
-%
-% Domain restrictions:
-%		X > 0
-%
-:- pragma c_code(math__ln(X::in, Log::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X <= 0.0) {
-		ML_math_domain_error(""math__ln"");
-	}
-#endif
-	Log = log(X);
-").
-
-%
-% math__log10(X, Log) is true if Log is the logarithm to
-% base 10 of X.
-%
-% Domain restrictions:
-%		X > 0
-%
-:- pragma c_code(math__log10(X::in, Log10::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X <= 0.0) {
-		ML_math_domain_error(""math__log10"");
-	}
-#endif
-	Log10 = log10(X);
-").
-
-%
-% math__log2(X, Log) is true if Log is the logarithm to
-% base 2 of X.
-%
-% Domain restrictions:
-%		X > 0
-%
-:- pragma c_code(math__log2(X::in, Log2::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X <= 0.0) {
-		ML_math_domain_error(""math__log2"");
-	}
-#endif
-	Log2 = log(X) / ML_FLOAT_LN2;
-").
-
-%
-% math__log(B, X, Log) is true if Log is the logarithm to
-% base B of X.
-%
-% Domain restrictions:
-%		X > 0
-%		B > 0
-%		B \= 1
-%
-:- pragma c_code(math__log(B::in, X::in, Log::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X <= 0.0 || B <= 0.0) {
-		ML_math_domain_error(""math__log"");
-	}
-	if (B == 1.0) {
-		ML_math_domain_error(""math__log"");
-	}
-#endif
-	Log = log(X)/log(B);
-").
-
-%
-% math__sin(X, Sin) is true if Sin is the sine of X.
-%
-:- pragma c_code(math__sin(X::in, Sin::out),
-		[will_not_call_mercury, thread_safe], "
-	Sin = sin(X);
-").
-
-%
-% math__cos(X, Cos) is true if Cos is the cosine of X.
-%
-:- pragma c_code(math__cos(X::in, Cos::out),
-		[will_not_call_mercury, thread_safe], "
-	Cos = cos(X);
-").
-
-%
-% math__tan(X, Tan) is true if Tan is the tangent of X.
-%
-:- pragma c_code(math__tan(X::in, Tan::out),
-		[will_not_call_mercury, thread_safe], "
-	Tan = tan(X);
-").
-
-%
-% math__asin(X, ASin) is true if ASin is the inverse
-% sine of X, where ASin is in the range [-pi/2,pi/2].
-%
-% Domain restrictions:
-%		X must be in the range [-1,1]
-%
-:- pragma c_code(math__asin(X::in, ASin::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X < -1.0 || X > 1.0) {
-		ML_math_domain_error(""math__asin"");
-	}
-#endif
-	ASin = asin(X);
-").
-
-%
-% math__acos(X, ACos) is true if ACos is the inverse
-% cosine of X, where ACos is in the range [0, pi].
-%
-% Domain restrictions:
-%		X must be in the range [-1,1]
-%
-:- pragma c_code(math__acos(X::in, ACos::out),
-		[will_not_call_mercury, thread_safe], "
-#ifndef ML_OMIT_MATH_DOMAIN_CHECKS
-	if (X < -1.0 || X > 1.0) {
-		ML_math_domain_error(""math__acos"");
-	}
-#endif
-	ACos = asin(X);
-").
-
-%
-% math__atan(X, ATan) is true if ATan is the inverse
-% tangent of X, where ATan is in the range [-pi/2,pi/2].
-%
-:- pragma c_code(math__atan(X::in, ATan::out),
-		[will_not_call_mercury, thread_safe], "
-	ATan = atan(X);
-").
-
-%
-% math__atan2(Y, X, ATan) is true if ATan is the inverse
-% tangent of Y/X, where ATan is in the range [-pi,pi].
-%
-:- pragma c_code(math__atan2(Y::in, X::in, ATan2::out),
-		[will_not_call_mercury, thread_safe], "
-	ATan2 = atan2(Y, X);
-").
-
-%
-% math__sinh(X, Sinh) is true if Sinh is the hyperbolic
-% sine of X.
-%
-:- pragma c_code(math__sinh(X::in, Sinh::out),
-		[will_not_call_mercury, thread_safe], "
-	Sinh = sinh(X);
-").
-
-%
-% math__cosh(X, Cosh) is true if Cosh is the hyperbolic
-% cosine of X.
-%
-:- pragma c_code(math__cosh(X::in, Cosh::out),
-		[will_not_call_mercury, thread_safe], "
-	Cosh = cosh(X);
-").
-
-%
-% math__tanh(X, Tanh) is true if Tanh is the hyperbolic
-% tangent of X.
-%
-:- pragma c_code(math__tanh(X::in, Tanh::out),
-		[will_not_call_mercury, thread_safe], "
-	Tanh = tanh(X);
-").
+pi(pi).
+e(e).
+ceiling(X, ceiling(X)).
+floor(X, floor(X)).
+round(X, round(X)).
+truncate(X, truncate(X)).
+sqrt(X, sqrt(X)).
+pow(X, Y, pow(X, Y)).
+exp(X, exp(X)).
+ln(X, ln(X)).
+log10(X, log10(X)).
+log2(X, log2(X)).
+log(X, Y, log(X, Y)).
+sin(X, sin(X)).
+cos(X, cos(X)).
+tan(X, tan(X)).
+asin(X, asin(X)).
+acos(X, acos(X)).
+atan(X, atan(X)).
+atan2(X, Y, atan2(X, Y)).
+sinh(X, sinh(X)).
+cosh(X, cosh(X)).
+tanh(X, tanh(X)).
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
