@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1999 The University of Melbourne.
+% Copyright (C) 1993-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -17,7 +17,7 @@
 
 :- module term.
 :- interface.
-:- import_module list, map, std_util.
+:- import_module enum, list, map, std_util.
 
 %-----------------------------------------------------------------------------%
 
@@ -55,6 +55,7 @@
 
 :- type term_to_type_result(T) == term_to_type_result(T, generic).
 
+:- func term__try_term_to_type(term(U)) = term_to_type_result(T, U).
 :- pred term__try_term_to_type(term(U), term_to_type_result(T, U)).
 :- mode term__try_term_to_type(in, out) is det.
 	% term__try_term_to_type(Term, Result):
@@ -72,7 +73,7 @@
 	% to the offending subterm.
 
 :- type term_to_type_error(T)
-	--->	type_error(term(T), type_info, term__context,
+	--->	type_error(term(T), type_desc, term__context,
 			term_to_type_context)
 	;	mode_error(var(T), term_to_type_context).
 
@@ -89,14 +90,17 @@
 :- mode term__term_to_type(in, out) is semidet.
 	% term_to_type(Term, Type) :- try_term_to_type(Term, ok(Type)).
 
+:- func term__det_term_to_type(term(_)) = T.
 :- pred term__det_term_to_type(term(_), T).
 :- mode term__det_term_to_type(in, out) is det.
 	% like term_to_type, but calls error/1 rather than failing.
 
+:- func term__type_to_term(T) = term(_).
 :- pred term__type_to_term(T, term(_)).
 :- mode term__type_to_term(in, out) is det.
 	% converts a value to a term representation of that value
 
+:- func term__univ_to_term(univ) = term(_).
 :- pred term__univ_to_term(univ, term(_)).
 :- mode term__univ_to_term(in, out) is det.
 	% calls term__type_to_term on the value stored in the univ
@@ -104,16 +108,19 @@
 
 %-----------------------------------------------------------------------------%
 
+:- func term__vars(term(T)) = list(var(T)).
 :- pred term__vars(term(T), list(var(T))).
 :- mode term__vars(in, out) is det.
 %	term__vars(Term, Vars)
 %		Vars is the list of variables contained in Term, in the order 
 %		obtained by traversing the term depth first, left-to-right.
 
+:- func term__vars_2(term(T), list(var(T))) = list(var(T)).
 :- pred term__vars_2(term(T), list(var(T)), list(var(T))).
 :- mode term__vars_2(in, in, out) is det.
 %		As above, but with an accumulator.
 
+:- func term__vars_list(list(term(T))) = list(var(T)).
 :- pred term__vars_list(list(term(T)), list(var(T))).
 :- mode term__vars_list(in, out) is det.
 %	term__vars_list(TermList, Vars)
@@ -144,16 +151,20 @@
 %		unify (with occur check) two terms with respect to a set
 %	 	of bindings and possibly update the set of bindings
 
+:- func term__substitute(term(T), var(T), term(T)) = term(T).
 :- pred term__substitute(term(T), var(T), term(T), term(T)).
 :- mode term__substitute(in, in, in, out) is det.
 %	term__substitute(Term0, Var, Replacement, Term) :
 %		replace all occurrences of Var in Term0 with Replacement,
 %		and return the result in Term.
 
+:- func term__substitute_list(list(term(T)), var(T), term(T)) = list(term(T)).
 :- pred term__substitute_list(list(term(T)), var(T), term(T), list(term(T))).
 :- mode term__substitute_list(in, in, in, out) is det.
 %		as above, except for a list of terms rather than a single term
 
+:- func term__substitute_corresponding(list(var(T)),
+		list(term(T)), term(T)) = term(T).
 :- pred term__substitute_corresponding(list(var(T)), list(term(T)),
 		term(T), term(T)).
 :- mode term__substitute_corresponding(in, in, in, out) is det.
@@ -164,6 +175,8 @@
 %	        length as Repls, the behaviour is undefined and probably
 %		harmful.
 
+:- func term__substitute_corresponding_list(list(var(T)),
+		list(term(T)), list(term(T))) = list(term(T)).
 :- pred term__substitute_corresponding_list(list(var(T)), list(term(T)),
 		list(term(T)), list(term(T))).
 :- mode term__substitute_corresponding_list(in, in, in, out) is det.
@@ -171,6 +184,7 @@
 %		As above, except applies to a list of terms rather than a
 %		single term.
 
+:- func term__apply_rec_substitution(term(T), substitution(T)) = term(T).
 :- pred term__apply_rec_substitution(term(T), substitution(T), term(T)).
 :- mode term__apply_rec_substitution(in, in, out) is det.
 %	term__apply_rec_substitution(Term0, Substitution, Term) :
@@ -178,15 +192,20 @@
 %		no more substitions can be applied, and then
 %		return the result in Term.
 
+:- func term__apply_rec_substitution_to_list(list(term(T)),
+		substitution(T)) = list(term(T)).
 :- pred term__apply_rec_substitution_to_list(list(term(T)), substitution(T),
 						list(term(T))).
 :- mode term__apply_rec_substitution_to_list(in, in, out) is det.
 
+:- func term__apply_substitution(term(T), substitution(T)) = term(T).
 :- pred term__apply_substitution(term(T), substitution(T), term(T)).
 :- mode term__apply_substitution(in, in, out) is det.
 %	term__apply_substitution(Term0, Substitution, Term) :
 %		apply substitution to Term0 and return the result in Term.
 
+:- func term__apply_substitution_to_list(list(term(T)),
+		substitution(T)) = list(term(T)).
 :- pred term__apply_substitution_to_list(list(term(T)), substitution(T),
 						list(term(T))).
 :- mode term__apply_substitution_to_list(in, in, out) is det.
@@ -204,23 +223,28 @@
 :- mode term__occurs_list(in, in, in) is semidet.
 %		as above, except for a list of terms rather than a single term
 
+:- func term__relabel_variable(term(T), var(T), var(T)) = term(T).
 :- pred term__relabel_variable(term(T), var(T), var(T), term(T)).
 :- mode term__relabel_variable(in, in, in, out) is det.
 %	term__relabel_variable(Term0, OldVar, NewVar, Term) :
 %		replace all occurences of OldVar in Term0 with
 %		NewVar and put the result in Term.
 
+:- func term__relabel_variables(list(term(T)), var(T), var(T)) = list(term(T)).
 :- pred term__relabel_variables(list(term(T)), var(T), var(T), list(term(T))).
 :- mode term__relabel_variables(in, in, in, out) is det.
 %	term__relabel_variables(Terms0, OldVar, NewVar, Terms) :
 %		same as term__relabel_variable but for a list of terms.
 
+:- func term__apply_variable_renaming(term(T), map(var(T), var(T))) = term(T).
 :- pred term__apply_variable_renaming(term(T), map(var(T), var(T)), term(T)).
 :- mode term__apply_variable_renaming(in, in, out) is det.
 % 		same as term__relabel_variable, except relabels
 % 		multiple variables. If a variable is not in the
 % 		map, it is not replaced.
 
+:- func term__apply_variable_renaming_to_list(list(term(T)),
+		map(var(T), var(T))) = list(term(T)).
 :- pred term__apply_variable_renaming_to_list(list(term(T)),
 		map(var(T), var(T)), list(term(T))).
 :- mode term__apply_variable_renaming_to_list(in, in, out) is det.
@@ -241,6 +265,8 @@
 	% To manage a supply of variables, use the following 2 predicates.
 	% (We might want to give these a unique mode later.)
 
+
+:- func term__init_var_supply = var_supply(T).
 :- pred term__init_var_supply(var_supply(T)).
 :- mode term__init_var_supply(out) is det.
 :- mode term__init_var_supply(in) is semidet. % implied
@@ -253,6 +279,15 @@
 %		create a fresh variable (var) and return the
 %		updated var_supply.
 
+%-----------------------------------------------------------------------------%
+
+	% from_int/1 should only be applied to integers returned
+	% by to_int/1. This instance declaration is needed to
+	% allow sets of variables to be represented using
+	% sparse_bitset.m.
+:- instance enum(var(_)).
+
+:- func term__var_to_int(var(T)) = int.
 :- pred term__var_to_int(var(T), int).
 :- mode term__var_to_int(in, out) is det.
 %		Convert a variable to an int.
@@ -266,20 +301,28 @@
 :- pred term__context_line(term__context, int).
 :- mode term__context_line(in, out) is det.
 
+:- func term__context_line(term__context) = int.
+
 	% Given a term context, return the source file.
 
 :- pred term__context_file(term__context, string).
 :- mode term__context_file(in, out) is det.
 
+:- func term__context_file(term__context) = string.
+
 	% Used to initialize the term context when reading in
 	% (or otherwise constructing) a term.
 
 :- pred term__context_init(term__context).
 :- mode term__context_init(out) is det.
 
+:- func term__context_init = term__context.
+
 :- pred term__context_init(string, int, term__context).
 :- mode term__context_init(in, in, out) is det.
 
+:- func term__context_init(string, int) = term__context.
+
 	% Convert a list of terms which are all vars into a list
 	% of vars.  Abort (call error/1) if the list contains
 	% any non-variables.
@@ -287,6 +330,8 @@
 :- pred term__term_list_to_var_list(list(term(T)), list(var(T))).
 :- mode term__term_list_to_var_list(in, out) is det.
 
+:- func term__term_list_to_var_list(list(term(T))) = list(var(T)).
+
 	% Convert a list of terms which are all vars into a list
 	% of vars (or vice versa).
 
@@ -294,6 +339,8 @@
 :- mode term__var_list_to_term_list(in, out) is det.
 :- mode term__var_list_to_term_list(out, in) is semidet.
 
+:- func term__var_list_to_term_list(list(var(T))) = list(term(T)).
+
 %-----------------------------------------------------------------------------%
 	
 	% term__generic_term(Term) is true iff `Term' is a term of type
@@ -308,14 +355,20 @@
 :- pred term__coerce(term(T), term(U)).
 :- mode term__coerce(in, out) is det.
 
+:- func term__coerce(term(T)) = term(U).
+
 	% Coerce a var of type `T' into a var of type `U'.
 :- pred term__coerce_var(var(T), var(U)).
 :- mode term__coerce_var(in, out) is det.
 
+:- func term__coerce_var(var(T)) = var(U).
+
 	% Coerce a var_supply of type `T' into a var_supply of type `U'.
 :- pred term__coerce_var_supply(var_supply(T), var_supply(U)).
 :- mode term__coerce_var_supply(in, out) is det.
 
+:- func term__coerce_var_supply(var_supply(T)) = var_supply(U).
+
 %-----------------------------------------------------------------------------%
 
 :- implementation.
@@ -327,6 +380,15 @@
 
 :- interface.
 
+% This is the same as term_to_type, except that an integer
+% is allowed where a character is expected. This is needed by 
+% extras/aditi/aditi.m because Aditi does not have a builtin
+% character type. This also allows an integer where a float
+% is expected.
+
+:- pred term__term_to_type_with_int_instead_of_char(term(U), T).
+:- mode term__term_to_type_with_int_instead_of_char(in, out) is semidet.
+
 	% This predidicate is being phased out, because of the problem
 	% mentioned in the "BEWARE:" below.
 :- pragma obsolete(term__compare/4).
@@ -345,7 +407,7 @@
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module std_util, require, array, int, string.
+:- import_module bool, char, std_util, require, array, int, string.
 
 %-----------------------------------------------------------------------------%
 
@@ -359,8 +421,20 @@
 term__term_to_type(Term, Val) :-
 	term__try_term_to_type(Term, ok(Val)).
 
+term__term_to_type_with_int_instead_of_char(Term, Val) :-
+	IsAditiTuple = yes,
+	term__try_term_to_type(IsAditiTuple, Term, ok(Val)).
+
 term__try_term_to_type(Term, Result) :-
-	term__try_term_to_univ(Term, type_of(ValTypedVar), UnivResult),
+	IsAditiTuple = no,
+	term__try_term_to_type(IsAditiTuple, Term, Result).
+
+:- pred term__try_term_to_type(bool, term(U), term_to_type_result(T, U)).
+:- mode term__try_term_to_type(in, in, out) is det.
+
+term__try_term_to_type(IsAditiTuple, Term, Result) :-
+	term__try_term_to_univ(IsAditiTuple, Term,
+		type_of(ValTypedVar), UnivResult),
 	(
 		UnivResult = ok(Univ),
 		det_univ_to_type(Univ, Val),
@@ -371,23 +445,23 @@ term__try_term_to_type(Term, Result) :-
 		Result = error(Error)
 	).
 
-:- pred term__try_term_to_univ(term(T)::in, type_info::in,
+:- pred term__try_term_to_univ(bool::in, term(T)::in, type_desc::in,
 		term_to_type_result(univ, T)::out) is det.
 
-term__try_term_to_univ(Term, Type, Result) :-
-	term__try_term_to_univ_2(Term, Type, [], Result).
+term__try_term_to_univ(IsAditiTuple, Term, Type, Result) :-
+	term__try_term_to_univ_2(IsAditiTuple, Term, Type, [], Result).
 	
-:- pred term__try_term_to_univ_2(term(T)::in, type_info::in,
+:- pred term__try_term_to_univ_2(bool::in, term(T)::in, type_desc::in,
 		term_to_type_context::in,
 		term_to_type_result(univ, T)::out) is det.
 
-term__try_term_to_univ_2(term__variable(Var), _Type, Context,
+term__try_term_to_univ_2(_, term__variable(Var), _Type, Context,
 		error(mode_error(Var, Context))).
-term__try_term_to_univ_2(Term, Type, Context, Result) :-
+term__try_term_to_univ_2(IsAditiTuple, Term, Type, Context, Result) :-
 	Term = term__functor(Functor, ArgTerms, TermContext),
 	(
 		type_ctor_and_args(Type, TypeCtor, TypeArgs),
-		term__term_to_univ_special_case(
+		term__term_to_univ_special_case(IsAditiTuple,
 			type_ctor_module_name(TypeCtor),
 			type_ctor_name(TypeCtor), 
 			TypeArgs, Term, Type, Context, SpecialCaseResult)
@@ -397,8 +471,8 @@ term__try_term_to_univ_2(Term, Type, Context, Result) :-
 		Functor = term__atom(FunctorName),
 		list__length(ArgTerms, Arity),
 		find_functor(Type, FunctorName, Arity, FunctorNumber, ArgTypes),
-		term__term_list_to_univ_list(ArgTerms, ArgTypes,
-			Functor, 1, Context, TermContext, ArgsResult)
+		term__term_list_to_univ_list(IsAditiTuple, ArgTerms,
+			ArgTypes, Functor, 1, Context, TermContext, ArgsResult)
 	->
 		(
 			ArgsResult = ok(ArgValues),
@@ -418,31 +492,44 @@ term__try_term_to_univ_2(Term, Type, Context, Result) :-
 		Result = error(type_error(Term, Type, TermContext, RevContext))
 	).
 
-:- pred term__term_to_univ_special_case(string::in, string::in, 
-		list(type_info)::in, 
+:- pred term__term_to_univ_special_case(bool::in, string::in, string::in, 
+		list(type_desc)::in, 
 		term(T)::in(bound(term__functor(ground, ground, ground))),
-		type_info::in, term_to_type_context::in,
+		type_desc::in, term_to_type_context::in,
 		term_to_type_result(univ, T)::out) is semidet.
 
-term__term_to_univ_special_case("builtin", "character", [],
+term__term_to_univ_special_case(IsAditiTuple, "builtin", "character", [],
 		Term, _, _, ok(Univ)) :-
-	Term = term__functor(term__atom(FunctorName), [], _),
-	string__first_char(FunctorName, Char, ""),
+	(
+		IsAditiTuple = no,
+		Term = term__functor(term__atom(FunctorName), [], _),
+		string__first_char(FunctorName, Char, "")
+	;
+		IsAditiTuple = yes,
+		Term = term__functor(term__integer(Int), [], _),
+		char__to_int(Char, Int)
+	),
 	type_to_univ(Char, Univ).
-term__term_to_univ_special_case("builtin", "int", [],
+term__term_to_univ_special_case(_, "builtin", "int", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__integer(Int), [], _),
 	type_to_univ(Int, Univ).
-term__term_to_univ_special_case("builtin", "string", [],
+term__term_to_univ_special_case(_, "builtin", "string", [],
 		Term, _, _, ok(Univ)) :-
 	Term = term__functor(term__string(String), [], _),
 	type_to_univ(String, Univ).
-term__term_to_univ_special_case("builtin", "float", [],
+term__term_to_univ_special_case(IsAditiTuple, "builtin", "float", [],
 		Term, _, _, ok(Univ)) :-
-	Term = term__functor(term__float(Float), [], _),
-	type_to_univ(Float, Univ).
-term__term_to_univ_special_case("array", "array", [ElemType], Term, _Type,
-		PrevContext, Result) :-
+	( Term = term__functor(term__float(Float), [], _) ->
+		type_to_univ(Float, Univ)
+	;
+		IsAditiTuple = yes,
+		Term = term__functor(term__integer(Int), [], _),
+		int__to_float(Int, Float),
+		type_to_univ(Float, Univ)
+	).
+term__term_to_univ_special_case(IsAditiTuple, "array", "array", [ElemType],
+		Term, _Type, PrevContext, Result) :-
 	%
 	% arrays are represented as terms of the form
 	%	array([elem1, elem2, ...])
@@ -457,7 +544,8 @@ term__term_to_univ_special_case("array", "array", [ElemType], Term, _Type,
 	ListType = type_of([Elem]),
 	ArgContext = arg_context(term__atom("array"), 1, TermContext),
 	NewContext = [ArgContext | PrevContext],
-	term__try_term_to_univ_2(ArgList, ListType, NewContext, ArgResult),
+	term__try_term_to_univ_2(IsAditiTuple, ArgList, ListType,
+		NewContext, ArgResult),
 	(
 		ArgResult = ok(ListUniv),
 		has_type(Elem2, ElemType),
@@ -469,10 +557,11 @@ term__term_to_univ_special_case("array", "array", [ElemType], Term, _Type,
 		ArgResult = error(Error),
 		Result = error(Error)
 	).
-term__term_to_univ_special_case("builtin", "c_pointer", _, _, _, 
+term__term_to_univ_special_case(_, "builtin", "c_pointer", _, _, _, 
 		_, _) :-
 	fail.
-term__term_to_univ_special_case("std_util", "univ", [], Term, _, _, Result) :-
+term__term_to_univ_special_case(_, "std_util", "univ", [],
+		Term, _, _, Result) :-
 	% Implementing this properly would require keeping a
 	% global table mapping from type names to type_infos
 	% for all of the types in the program...
@@ -496,24 +585,27 @@ term__term_to_univ_special_case("std_util", "univ", [], Term, _, _, Result) :-
 	% like all the other results returned from this procedure.
 	Result = ok(univ(Univ)).
 
-term__term_to_univ_special_case("std_util", "type_info", _, _, _, _, _) :-
+term__term_to_univ_special_case(_, "std_util", "type_info", _, _, _, _, _) :-
 	% ditto
 	fail.
 
-:- pred term__term_list_to_univ_list(list(term(T))::in, list(type_info)::in,
-		term__const::in, int::in, term_to_type_context::in,
-		term__context::in, term_to_type_result(list(univ), T)::out)
-		is semidet.
-term__term_list_to_univ_list([], [], _, _, _, _, ok([])).
-term__term_list_to_univ_list([ArgTerm|ArgTerms], [Type|Types],
-		Functor, ArgNum, PrevContext, TermContext, Result) :-
+:- pred term__term_list_to_univ_list(bool::in, list(term(T))::in,
+		list(type_desc)::in, term__const::in, int::in,
+		term_to_type_context::in, term__context::in,
+		term_to_type_result(list(univ), T)::out) is semidet.
+term__term_list_to_univ_list(_, [], [], _, _, _, _, ok([])).
+term__term_list_to_univ_list(IsAditiTuple, [ArgTerm|ArgTerms],
+		[Type|Types], Functor, ArgNum, PrevContext,
+		TermContext, Result) :-
 	ArgContext = arg_context(Functor, ArgNum, TermContext),
 	NewContext = [ArgContext | PrevContext],
-	term__try_term_to_univ_2(ArgTerm, Type, NewContext, ArgResult),
+	term__try_term_to_univ_2(IsAditiTuple, ArgTerm, Type,
+		NewContext, ArgResult),
 	(
 		ArgResult = ok(Arg),
-		term__term_list_to_univ_list(ArgTerms, Types, Functor,
-			ArgNum + 1, PrevContext, TermContext, RestResult),
+		term__term_list_to_univ_list(IsAditiTuple, ArgTerms, Types,
+			Functor, ArgNum + 1, PrevContext,
+			TermContext, RestResult),
 		(
 			RestResult = ok(Rest),
 			Result = ok([Arg | Rest])
@@ -526,14 +618,14 @@ term__term_list_to_univ_list([ArgTerm|ArgTerms], [Type|Types],
 		Result = error(Error)
 	).
 
-:- pred term__find_functor(type_info::in, string::in, int::in, int::out,
-		list(type_info)::out) is semidet.
+:- pred term__find_functor(type_desc::in, string::in, int::in, int::out,
+		list(type_desc)::out) is semidet.
 term__find_functor(Type, Functor, Arity, FunctorNumber, ArgTypes) :-
 	N = num_functors(Type),
 	term__find_functor_2(Type, Functor, Arity, N, FunctorNumber, ArgTypes).
         
-:- pred term__find_functor_2(type_info::in, string::in, int::in, int::in, 
-	int::out, list(type_info)::out) is semidet.
+:- pred term__find_functor_2(type_desc::in, string::in, int::in, int::in, 
+	int::out, list(type_desc)::out) is semidet.
 term__find_functor_2(TypeInfo, Functor, Arity, Num, FunctorNumber, ArgTypes) :-
 	Num >= 0,
 	Num1 = Num - 1,
@@ -566,6 +658,8 @@ term__det_term_to_type(Term, X) :-
 term__type_to_term(Val, Term) :- type_to_univ(Val, Univ),
 	term__univ_to_term(Univ, Term).
 
+	% convert the value stored in the univ (as distinct from
+	% the univ itself) to a term.
 term__univ_to_term(Univ, Term) :-
 	term__context_init(Context),
 	Type = univ_type(Univ),
@@ -588,14 +682,15 @@ term__univ_to_term(Univ, Term) :-
 			error(Message)
 		)
 	;
-		deconstruct(Univ, FunctorString, _FunctorArity, FunctorArgs),
+		deconstruct(univ_value(Univ), FunctorString, _FunctorArity,
+			FunctorArgs),
 		term__univ_list_to_term_list(FunctorArgs, TermArgs),
 		Term = term__functor(term__atom(FunctorString), TermArgs,
 			Context)
 	).
 
 :- pred term__univ_to_term_special_case(string::in, string::in, 
-		list(type_info)::in, univ::in, term__context::in,
+		list(type_desc)::in, univ::in, term__context::in,
 		term(T)::out) is semidet.
 
 term__univ_to_term_special_case("builtin", "int", [], Univ, Context,
@@ -616,15 +711,16 @@ term__univ_to_term_special_case("std_util", "type_info", [], Univ, Context,
 	det_univ_to_type(Univ, TypeInfo),
 	type_info_to_term(Context, TypeInfo, Term).
 term__univ_to_term_special_case("std_util", "univ", [], Univ, Context, Term) :-
+	det_univ_to_type(Univ, NestedUniv),
 	Term = term__functor(term__atom("univ"),
 			% XXX what operator should we use for type
 			% qualification?
 			[term__functor(term__atom(":"),	 % TYPE_QUAL_OP
 				[ValueTerm, TypeTerm],
 				Context)], Context),
-	type_info_to_term(Context, univ_type(Univ), TypeTerm),
-	UnivValue = univ_value(Univ),
-	term__type_to_term(UnivValue, ValueTerm).
+	type_info_to_term(Context, univ_type(NestedUniv), TypeTerm),
+	NestedUnivValue = univ_value(NestedUniv),
+	term__type_to_term(NestedUnivValue, ValueTerm).
 
 term__univ_to_term_special_case("array", "array", [ElemType], Univ, Context, 
 		Term) :-
@@ -647,7 +743,7 @@ term__univ_list_to_term_list([Value|Values], [Term|Terms]) :-
 	term__univ_list_to_term_list(Values, Terms).
 
 % given a type_info, return a term that represents the name of that type.
-:- pred type_info_to_term(term__context::in, type_info::in,
+:- pred type_info_to_term(term__context::in, type_desc::in,
 		term(T)::out) is det.
 type_info_to_term(Context, TypeInfo, Term) :-
 	type_ctor_and_args(TypeInfo, TypeCtor, ArgTypes),
@@ -980,8 +1076,17 @@ term__create_var(var_supply(V0), var(V), var_supply(V)) :-
 
 %-----------------------------------------------------------------------------%
 
+:- instance enum(var(_)) where [
+	to_int(X) = term__var_to_int(X),
+	from_int(X) = term__unsafe_int_to_var(X)
+].
+
 term__var_to_int(var(Var), Var).
 
+	% Cast an integer to a var(T), subverting the type-checking.
+:- func unsafe_int_to_var(int) = var(T).
+term__unsafe_int_to_var(Var) = var(Var).
+
 %-----------------------------------------------------------------------------%
 
 	% substitute a variable name in a term.
@@ -1079,10 +1184,11 @@ term__generic_term(_).
 
 %-----------------------------------------------------------------------------%
 
-term__coerce(term__variable(var(V)), term__variable(var(V))).
-term__coerce(term__functor(Cons, Args0, Ctxt),
-		term__functor(Cons, Args, Ctxt)) :-
-	list__map(term__coerce, Args0, Args).
+term__coerce(A, B) :-
+	% Normally calls to this predicate should only be
+	% generated by the compiler, but type coercion by
+	% copying was taking about 3% of the compiler's runtime.
+	private_builtin__unsafe_type_cast(A, B).
 
 term__coerce_var(var(V), var(V)).
 
@@ -1093,73 +1199,6 @@ term__coerce_var_supply(var_supply(Supply), var_supply(Supply)).
 % Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
 % 	Function forms added.
 
-:- interface.
-
-:- func term__context_init = term__context.
-
-:- func term__init_var_supply = var_supply(T).
-
-:- func term__try_term_to_type(term(U)) = term_to_type_result(T, U).
-
-:- func term__det_term_to_type(term(_)) = T.
-
-:- func term__type_to_term(T) = term(_).
-
-:- func term__univ_to_term(univ) = term(_).
-
-:- func term__vars(term(T)) = list(var(T)).
-
-:- func term__vars_2(term(T), list(var(T))) = list(var(T)).
-
-:- func term__vars_list(list(term(T))) = list(var(T)).
-
-:- func term__substitute(term(T), var(T), term(T)) = term(T).
-
-:- func term__substitute_list(list(term(T)), var(T), term(T)) = list(term(T)).
-
-:- func term__substitute_corresponding(list(var(T)), list(term(T)), term(T)) = term(T).
-
-:- func term__substitute_corresponding_list(list(var(T)), list(term(T)), list(term(T))) = list(term(T)).
-
-:- func term__apply_rec_substitution(term(T), substitution(T)) = term(T).
-
-:- func term__apply_rec_substitution_to_list(list(term(T)), substitution(T)) = list(term(T)).
-
-:- func term__apply_substitution(term(T), substitution(T)) = term(T).
-
-:- func term__apply_substitution_to_list(list(term(T)), substitution(T)) = list(term(T)).
-
-:- func term__relabel_variable(term(T), var(T), var(T)) = term(T).
-
-:- func term__relabel_variables(list(term(T)), var(T), var(T)) = list(term(T)).
-
-:- func term__apply_variable_renaming(term(T), map(var(T), var(T))) = term(T).
-
-:- func term__apply_variable_renaming_to_list(list(term(T)), map(var(T), var(T))) = list(term(T)).
-
-:- func term__var_to_int(var(T)) = int.
-
-:- func term__context_line(term__context) = int.
-
-:- func term__context_file(term__context) = string.
-
-:- func term__context_init(string, int) = term__context.
-
-:- func term__term_list_to_var_list(list(term(T))) = list(var(T)).
-
-:- func term__var_list_to_term_list(list(var(T))) = list(term(T)).
-
-:- func term__coerce(term(T)) = term(U).
-
-:- func term__coerce_var(var(T)) = var(U).
-
-:- func term__coerce_var_supply(var_supply(T)) = var_supply(U).
-
-% ---------------------------------------------------------------------------- %
-% ---------------------------------------------------------------------------- %
-
-:- implementation.
-
 term__context_init = C :-
 	term__context_init(C).
 
