@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1994-1997, 1999 The University of Melbourne.
+% Copyright (C) 1994-1997, 1999-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -17,12 +17,18 @@
 :- interface.
 
 :- import_module list, assoc_list.
+:- import_module term.	% for var/1.
 
 :- type tree234(K, V).
 
 :- pred tree234__init(tree234(K, V)).
 :- mode tree234__init(uo) is det.
 
+:- func tree234__init = tree234(K, V).
+
+:- pred tree234__is_empty(tree234(K, V)).
+:- mode tree234__is_empty(in) is semidet.
+
 :- pred tree234__member(tree234(K, V), K, V).
 :- mode tree234__member(in, out, out) is nondet.
 
@@ -32,6 +38,8 @@
 :- pred tree234__lookup(tree234(K, V), K, V).
 :- mode tree234__lookup(in, in, out) is det.
 
+:- func tree234__lookup(tree234(K, V), K) = V.
+
 :- pred tree234__lower_bound_search(tree234(K, V), K, K, V).
 :- mode tree234__lower_bound_search(in, in, out, out) is semidet.
 
@@ -54,11 +62,15 @@
 % :- mode tree234__set(di_tree234, in, in, uo_tree234) is det.
 :- mode tree234__set(in, in, in, out) is det.
 
+:- func tree234__set(tree234(K, V), K, V) = tree234(K, V).
+
 :- pred tree234__delete(tree234(K, V), K, tree234(K, V)).
 :- mode tree234__delete(di, in, uo) is det.
 % :- mode tree234__delete(di_tree234, in, uo_tree234) is det.
 :- mode tree234__delete(in, in, out) is det.
 
+:- func tree234__delete(tree234(K, V), K) = tree234(K, V).
+
 :- pred tree234__remove(tree234(K, V), K, V, tree234(K, V)).
 :- mode tree234__remove(di, in, uo, uo) is semidet.
 % :- mode tree234__remove(di_tree234, in, out, uo_tree234) is semidet.
@@ -70,12 +82,18 @@
 %	is semidet.
 :- mode tree234__remove_smallest(in, out, out, out) is semidet.
 
+	% Given a tree234, return a list of all the keys in the tree.
+	% The list that is returned is in sorted order.
 :- pred tree234__keys(tree234(K, V), list(K)).
 :- mode tree234__keys(in, out) is det.
 
+:- func tree234__keys(tree234(K, V)) = list(K).
+
 :- pred tree234__values(tree234(K, V), list(V)).
 :- mode tree234__values(in, out) is det.
 
+:- func tree234__values(tree234(K, V)) = list(V).
+
 :- pred tree234__update(tree234(K, V), K, V, tree234(K, V)).
 :- mode tree234__update(in, in, in, out) is semidet.
 % :- mode tree234__update(di_tree234, in, in, uo_tree234) is det.
@@ -85,27 +103,74 @@
 :- pred tree234__count(tree234(K, V), int).
 :- mode tree234__count(in, out) is det.
 
+:- func tree234__count(tree234(K, V)) = int.
+
 :- pred tree234__assoc_list_to_tree234(assoc_list(K, V), tree234(K, V)).
 :- mode tree234__assoc_list_to_tree234(in, out) is det.
 
+:- func tree234__assoc_list_to_tree234(assoc_list(K, V)) = tree234(K, V).
+
+	% Given a tree234, return an association list of all the
+	% keys and values in the tree.  The association list that
+	% is returned is sorted on the keys.
 :- pred tree234__tree234_to_assoc_list(tree234(K, V), assoc_list(K, V)).
 :- mode tree234__tree234_to_assoc_list(in, out) is det.
 
+:- func tree234__tree234_to_assoc_list(tree234(K, V)) = assoc_list(K, V).
+
 :- pred tree234__foldl(pred(K, V, T, T), tree234(K, V), T, T).
 :- mode tree234__foldl(pred(in, in, in, out) is det, in, in, out) is det.
 :- mode tree234__foldl(pred(in, in, in, out) is semidet, in, in, out)
 		is semidet.
 :- mode tree234__foldl(pred(in, in, di, uo) is det, in, di, uo) is det.
 
+:- func tree234__foldl(func(K, V, T) = T, tree234(K, V), T) = T.
+
+:- pred tree234__foldl2(pred(K, V, T, T, U, U), tree234(K, V), T, T, U, U).
+:- mode tree234__foldl2(pred(in, in, in, out, in, out) is det, 
+		in, in, out, in, out) is det.
+:- mode tree234__foldl2(pred(in, in, in, out, in, out) is semidet, 
+		in, in, out, in, out) is semidet.
+:- mode tree234__foldl2(pred(in, in, in, out, di, uo) is det,
+		in, in, out, di, uo) is det.
+:- mode tree234__foldl2(pred(in, in, di, uo, di, uo) is det,
+		in, di, uo, di, uo) is det.
+
+
 :- pred tree234__map_values(pred(K, V, W), tree234(K, V), tree234(K, W)).
 :- mode tree234__map_values(pred(in, in, out) is det, in, out) is det.
 :- mode tree234__map_values(pred(in, in, out) is semidet, in, out) is semidet.
 
+:- func tree234__map_values(func(K, V) = W, tree234(K, V)) = tree234(K, W).
+
+:- pred tree234__map_foldl(pred(K, V, W, A, A), tree234(K, V), tree234(K, W),
+	A, A).
+:- mode tree234__map_foldl(pred(in, in, out, in, out) is det,
+	in, out, in, out) is det.
+:- mode tree234__map_foldl(pred(in, in, out, in, out) is semidet,
+	in, out, in, out) is semidet.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
 :- implementation.
 
+% Everything below here is not intended to be part of the public interface,
+% and will not be included in the Mercury library reference manual.
+
+:- interface.
+
+:- pragma type_spec(tree234__search/3, K = var(_)).
+:- pragma type_spec(tree234__search/3, K = int).
+
+:- pragma type_spec(tree234__lookup/3, K = var(_)).
+
+:- pragma type_spec(tree234__set(in, in, in, out), K = var(_)).
+
+%-----------------------------------------------------------------------------%
+
+:- implementation.
+
 :- import_module int, require, bool, std_util.
 
 :- type tree234(K, V)	--->
@@ -149,6 +214,9 @@
 
 tree234__init(empty).
 
+tree234__is_empty(Tree) :-
+	Tree = empty.
+
 %------------------------------------------------------------------------------%
 
 tree234__member(empty, _K, _V) :- fail.
@@ -848,8 +916,8 @@ tree234__insert(Tin, K, V, Tout) :-
 
 tree234__insert2(two(K0, V0, T0, T1), K, V, Tout) :-
 	(
-		T0 = empty,
-		T1 = empty
+		T0 = empty
+		% T1 = empty implied by T0 = empty
 	->
 		compare(Result, K, K0),
 		(
@@ -943,9 +1011,9 @@ tree234__insert2(two(K0, V0, T0, T1), K, V, Tout) :-
 
 tree234__insert3(three(K0, V0, K1, V1, T0, T1, T2), K, V, Tout) :-
 	(
-		T0 = empty,
-		T1 = empty,
-		T2 = empty
+		T0 = empty
+		% T1 = empty implied by T0 = empty
+		% T2 = empty implied by T0 = empty
 	->
 		compare(Result0, K, K0),
 		(
@@ -1141,11 +1209,12 @@ tree234__set(Tin, K, V, Tout) :-
 :- mode tree234__set2(di_two, di, di, uo) is det.
 % :- mode tree234__set2(sdi_two, in, in, uo_tree234) is det.
 :- mode tree234__set2(in_two, in, in, out) is det.
+:- pragma type_spec(tree234__set2(in_two, in, in, out), K = var(_)).
 
 tree234__set2(two(K0, V0, T0, T1), K, V, Tout) :-
 	(
-		T0 = empty,
-		T1 = empty
+		T0 = empty
+		% T1 = empty implied by T0 = empty
 	->
 		compare(Result, K, K0),
 		(
@@ -1238,12 +1307,13 @@ tree234__set2(two(K0, V0, T0, T1), K, V, Tout) :-
 :- mode tree234__set3(di_three, di, di, uo) is det.
 % :- mode tree234__set3(sdi_three, in, in, uo_tree234) is det.
 :- mode tree234__set3(in_three, in, in, out) is det.
+:- pragma type_spec(tree234__set3(in_three, in, in, out), K = var(_)).
 
 tree234__set3(three(K0, V0, K1, V1, T0, T1, T2), K, V, Tout) :-
 	(
-		T0 = empty,
-		T1 = empty,
-		T2 = empty
+		T0 = empty
+		% T1 = empty implied by T0 = empty
+		% T2 = empty implied by T0 = empty
 	->
 		compare(Result0, K, K0),
 		(
@@ -2422,6 +2492,26 @@ tree234__foldl(Pred, four(K0, V0, K1, V1, K2, V2, T0, T1, T2, T3), Acc0, Acc) :-
 	call(Pred, K2, V2, Acc5, Acc6),
 	tree234__foldl(Pred, T3, Acc6, Acc).
 
+tree234__foldl2(_Pred, empty, A, A) --> [].
+tree234__foldl2(Pred, two(K, V, T0, T1), A0, A) -->
+	tree234__foldl2(Pred, T0, A0, A1),
+	call(Pred, K, V, A1, A2),
+	tree234__foldl2(Pred, T1, A2, A).
+tree234__foldl2(Pred, three(K0, V0, K1, V1, T0, T1, T2), A0, A) -->
+	tree234__foldl2(Pred, T0, A0, A1),
+	call(Pred, K0, V0, A1, A2),
+	tree234__foldl2(Pred, T1, A2, A3),
+	call(Pred, K1, V1, A3, A4),
+	tree234__foldl2(Pred, T2, A4, A).
+tree234__foldl2(Pred, four(K0, V0, K1, V1, K2, V2, T0, T1, T2, T3), A0, A) -->
+	tree234__foldl2(Pred, T0, A0, A1),
+	call(Pred, K0, V0, A1, A2),
+	tree234__foldl2(Pred, T1, A2, A3),
+	call(Pred, K1, V1, A3, A4),
+	tree234__foldl2(Pred, T2, A4, A5),
+	call(Pred, K2, V2, A5, A6),
+	tree234__foldl2(Pred, T3, A6, A).
+
 %------------------------------------------------------------------------------%
 
 tree234__map_values(_Pred, empty, empty).
@@ -2450,6 +2540,34 @@ tree234__map_values(Pred, Tree0, Tree) :-
 	tree234__map_values(Pred, RMid0, RMid),
 	tree234__map_values(Pred, Right0, Right).
 
+%------------------------------------------------------------------------------%
+
+tree234__map_foldl(_Pred, empty, empty, A, A).
+tree234__map_foldl(Pred, Tree0, Tree, A0, A) :-
+	Tree0 = two(K0, V0, Left0, Right0),
+	Tree  = two(K0, W0, Left, Right),
+	tree234__map_foldl(Pred, Left0, Left, A0, A1),
+	call(Pred, K0, V0, W0, A1, A2),
+	tree234__map_foldl(Pred, Right0, Right, A2, A).
+tree234__map_foldl(Pred, Tree0, Tree, A0, A) :-
+	Tree0 = three(K0, V0, K1, V1, Left0, Middle0, Right0),
+	Tree  = three(K0, W0, K1, W1, Left, Middle, Right),
+	tree234__map_foldl(Pred, Left0, Left, A0, A1),
+	call(Pred, K0, V0, W0, A1, A2),
+	tree234__map_foldl(Pred, Middle0, Middle, A2, A3),
+	call(Pred, K1, V1, W1, A3, A4),
+	tree234__map_foldl(Pred, Right0, Right, A4, A).
+tree234__map_foldl(Pred, Tree0, Tree, A0, A) :-
+	Tree0 = four(K0, V0, K1, V1, K2, V2, Left0, LMid0, RMid0, Right0),
+	Tree  = four(K0, W0, K1, W1, K2, W2, Left, LMid, RMid, Right),
+	tree234__map_foldl(Pred, Left0, Left, A0, A1),
+	call(Pred, K0, V0, W0, A1, A2),
+	tree234__map_foldl(Pred, LMid0, LMid, A2, A3),
+	call(Pred, K1, V1, W1, A3, A4),
+	tree234__map_foldl(Pred, RMid0, RMid, A4, A5),
+	call(Pred, K2, V2, W2, A5, A6),
+	tree234__map_foldl(Pred, Right0, Right, A6, A).
+
 %------------------------------------------------------------------------------%
 
 	% count the number of elements in a tree
@@ -2470,40 +2588,11 @@ tree234__count(four(_, _, _, _, _, _, T0, T1, T2, T3), N) :-
 	tree234__count(T3, N3),
 	N is 3 + N0 + N1 + N2 + N3.
 
-%------------------------------------------------------------------------------%
-%------------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
 % Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
 %       Function forms added.
 
-:- interface.
-
-:- func tree234__init = tree234(K, V).
-
-:- func tree234__lookup(tree234(K, V), K) = V.
-
-:- func tree234__set(tree234(K, V), K, V) = tree234(K, V).
-
-:- func tree234__delete(tree234(K, V), K) = tree234(K, V).
-
-:- func tree234__keys(tree234(K, V)) = list(K).
-
-:- func tree234__values(tree234(K, V)) = list(V).
-
-:- func tree234__count(tree234(K, V)) = int.
-
-:- func tree234__assoc_list_to_tree234(assoc_list(K, V)) = tree234(K, V).
-
-:- func tree234__tree234_to_assoc_list(tree234(K, V)) = assoc_list(K, V).
-
-:- func tree234__foldl(func(K, V, T) = T, tree234(K, V), T) = T.
-
-:- func tree234__map_values(func(K, V) = W, tree234(K, V)) = tree234(K, W).
-
-% ---------------------------------------------------------------------------- %
-% ---------------------------------------------------------------------------- %
-
-:- implementation.
-
 tree234__init = T :-
 	tree234__init(T).
 
@@ -2538,4 +2627,3 @@ tree234__foldl(F, T, A) = B :-
 tree234__map_values(F, T1) = T2 :-
 	P = ( pred(X::in, Y::in, Z::out) is det :- Z = F(X, Y) ),
 	tree234__map_values(P, T1, T2).
-
