@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1993-1999 The University of Melbourne.
+% Copyright (C) 1993-2000 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -38,6 +38,8 @@
 :- pred varset__init(varset(T)).
 :- mode varset__init(out) is det.
 
+:- func varset__init = varset(T).
+
 	% check whether a varset is empty.
 :- pred varset__is_empty(varset(T)).
 :- mode varset__is_empty(in) is semidet.
@@ -58,28 +60,40 @@
 :- pred varset__delete_var(varset(T), var(T), varset(T)).
 :- mode varset__delete_var(in, in, out) is det.
 
+:- func varset__delete_var(varset(T), var(T)) = varset(T).
+
 	% delete the names and values for a list of variables
 :- pred varset__delete_vars(varset(T), list(var(T)), varset(T)).
 :- mode varset__delete_vars(in, in, out) is det.
 
+:- func varset__delete_vars(varset(T), list(var(T))) = varset(T).
+
 	% return a list of all the variables in a varset
 :- pred varset__vars(varset(T), list(var(T))).
 :- mode varset__vars(in, out) is det.
 
+:- func varset__vars(varset(T)) = list(var(T)).
+
 	% set the name of a variable
 :- pred varset__name_var(varset(T), var(T), string, varset(T)).
 :- mode varset__name_var(in, in, in, out) is det.
 
+:- func varset__name_var(varset(T), var(T), string) = varset(T).
+
 	% lookup the name of a variable;
 	% create one if it doesn't have one using V_ as a prefix
 :- pred varset__lookup_name(varset(T), var(T), string).
 :- mode varset__lookup_name(in, in, out) is det.
 
+:- func varset__lookup_name(varset(T), var(T)) = string.
+
 	% lookup the name of a variable;
 	% create one if it doesn't have one using the specified prefix
 :- pred varset__lookup_name(varset(T), var(T), string, string).
 :- mode varset__lookup_name(in, in, in, out) is det.
 
+:- func varset__lookup_name(varset(T), var(T), string) = string.
+
 	% lookup the name of a variable;
 	% fail if it doesn't have one
 :- pred varset__search_name(varset(T), var(T), string).
@@ -90,10 +104,14 @@
 :- pred varset__bind_var(varset(T), var(T), term(T), varset(T)).
 :- mode varset__bind_var(in, in, in, out) is det.
 
+:- func varset__bind_var(varset(T), var(T), term(T)) = varset(T).
+
 	% bind a set of terms to a set of variables.
 :- pred varset__bind_vars(varset(T), substitution(T), varset(T)).
 :- mode varset__bind_vars(in, in, out) is det.
 
+:- func varset__bind_vars(varset(T), substitution(T)) = varset(T).
+
 	% lookup the value of a variable
 :- pred varset__search_var(varset(T), var(T), term(T)).
 :- mode varset__search_var(in, in, out) is semidet.
@@ -102,6 +120,8 @@
 :- pred varset__lookup_vars(varset(T), substitution(T)).
 :- mode varset__lookup_vars(in, out) is det.
 
+:- func varset__lookup_vars(varset(T)) = substitution(T).
+
 	% Combine two different varsets, renaming apart:
 	% varset__merge(VarSet0, NewVarSet, Terms0, VarSet, Terms) is
 	% true iff VarSet is the varset that results from joining
@@ -139,10 +159,14 @@
 :- pred varset__get_bindings(varset(T), substitution(T)).
 :- mode varset__get_bindings(in, out) is det.
 
+:- func varset__get_bindings(varset(T)) = substitution(T).
+
 	% set the bindings for all the bound variables.
 :- pred varset__set_bindings(varset(T), substitution(T), varset(T)).
 :- mode varset__set_bindings(in, in, out) is det.
 
+:- func varset__set_bindings(varset(T), substitution(T)) = varset(T).
+
 	% Create a map from names to variables.
 	% Each name is mapped to only one variable, even if a name is
 	% shared by more than one variable. Therefore this predicate
@@ -151,37 +175,50 @@
 :- pred varset__create_name_var_map(varset(T), map(string, var(T))).
 :- mode varset__create_name_var_map(in, out) is det.
 
+:- func varset__create_name_var_map(varset(T)) = map(string, var(T)).
+
 	% Return an association list giving the name of each variable.
 	% Every variable has an entry in the returned association list,
 	% even if it shares its name with another variable.
 :- pred varset__var_name_list(varset(T), assoc_list(var(T), string)).
 :- mode varset__var_name_list(in, out) is det.
 
+:- func varset__var_name_list(varset(T)) = assoc_list(var(T), string).
+
 	% Given a list of variable and varset in which some variables have
 	% no name but some other variables may have the same name,
 	% return another varset in which every variable has a unique name.
 	% If necessary, names will have suffixes added on the end;
 	% the second argument gives the suffix to use.
-:- pred varset__ensure_unique_names(list(var(T)), string, varset(T), varset(T)).
+:- pred varset__ensure_unique_names(list(var(T)),
+		string, varset(T), varset(T)).
 :- mode varset__ensure_unique_names(in, in, in, out) is det.
 
+:- func varset__ensure_unique_names(list(var(T)),
+		string, varset(T)) = varset(T).
+
 	% Given a varset and a set of variables, remove the names
 	% and values of any other variables stored in the varset.
 :- pred varset__select(varset(T), set(var(T)), varset(T)).
 :- mode varset__select(in, in, out) is det.
 
+:- func varset__select(varset(T), set(var(T))) = varset(T).
+
 	% Given a varset and a list of variables, construct a new varset
 	% containing one variable for each one in the list (and no others).
 	% Also return a substitution mapping the selected variables in the
 	% original varset into variables in the new varset. The relative
 	% ordering of variables in the original varset is maintained.
-:- pred varset__squash(varset(T), list(var(T)), varset(T), map(var(T), var(T))).
+:- pred varset__squash(varset(T), list(var(T)),
+		varset(T), map(var(T), var(T))).
 :- mode varset__squash(in, in, out, out) is det.
 
 	% Coerce the types of the variables in a varset.
 :- pred varset__coerce(varset(T), varset(U)).
 :- mode varset__coerce(in, out) is det.
 
+:- func varset__coerce(varset(T)) = varset(U).
+
 :- implementation.
 %-----------------------------------------------------------------------------%
 
@@ -530,75 +567,17 @@ copy_var_names([OldVar - Name | Rest], Subst, NewVarSet0, NewVarSet) :-
 
 %-----------------------------------------------------------------------------%
 
-varset__coerce(varset(S0, N0, B0), varset(S, N, B)) :-
-		% Coerce the var_supply
-	term__coerce_var_supply(S0, S),
-
-		% Coerce all the vars in the mapping from vars to names
-	map__to_assoc_list(N0, NList0),
-	list__map(lambda([P0::in, P::out] is det, (
-		P0 = V0 - Name,
-		P = V - Name,
-		term__coerce_var(V0, V)
-	)), NList0, NList),
-	map__from_assoc_list(NList, N),
-
-		% Coerce all the vars and terms in the substitution
-	map__to_assoc_list(B0, BList0),
-	list__map(lambda([Q0::in, Q::out] is det, (
-		Q0 = V0 - Binding0,
-		Q = V - Binding,
-		term__coerce_var(V0, V),
-		term__coerce(Binding0, Binding)
-	)), BList0, BList),
-	map__from_assoc_list(BList, B).
+varset__coerce(A, B) :-
+	% Normally calls to this predicate should only be
+	% generated by the compiler, but type coercion by 
+	% copying was taking about 3% of the compiler's runtime.
+	private_builtin__unsafe_type_cast(A, B).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 % Ralph Becket <rwab1@cl.cam.ac.uk> 30/04/99
 %	Function forms added.
 
-:- interface.
-
-:- func varset__init = varset(T).
-
-:- func varset__delete_var(varset(T), var(T)) = varset(T).
-
-:- func varset__delete_vars(varset(T), list(var(T))) = varset(T).
-
-:- func varset__vars(varset(T)) = list(var(T)).
-
-:- func varset__name_var(varset(T), var(T), string) = varset(T).
-
-:- func varset__lookup_name(varset(T), var(T)) = string.
-
-:- func varset__lookup_name(varset(T), var(T), string) = string.
-
-:- func varset__bind_var(varset(T), var(T), term(T)) = varset(T).
-
-:- func varset__bind_vars(varset(T), substitution(T)) = varset(T).
-
-:- func varset__lookup_vars(varset(T)) = substitution(T).
-
-:- func varset__get_bindings(varset(T)) = substitution(T).
-
-:- func varset__set_bindings(varset(T), substitution(T)) = varset(T).
-
-:- func varset__create_name_var_map(varset(T)) = map(string, var(T)).
-
-:- func varset__var_name_list(varset(T)) = assoc_list(var(T), string).
-
-:- func varset__ensure_unique_names(list(var(T)), string, varset(T)) = varset(T).
-
-:- func varset__select(varset(T), set(var(T))) = varset(T).
-
-:- func varset__coerce(varset(T)) = varset(U).
-
-% ---------------------------------------------------------------------------- %
-% ---------------------------------------------------------------------------- %
-
-:- implementation.
-
 varset__init = VS :-
 	varset__init(VS).
 
