@@ -3,7 +3,7 @@ INIT mercury_sys_init_engine
 ENDINIT
 */
 /*
-** Copyright (C) 1993-1999 The University of Melbourne.
+** Copyright (C) 1993-2001 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
@@ -15,8 +15,8 @@ ENDINIT
 #include	<setjmp.h>
 
 #include	"mercury_engine.h"
-#include	"mercury_memory_zones.h"	/* for create_zone() */
-#include	"mercury_memory_handlers.h"	/* for default_handler() */
+#include	"mercury_memory_zones.h"	/* for MR_create_zone() */
+#include	"mercury_memory_handlers.h"	/* for MR_default_handler() */
 
 #include	"mercury_dummy.h"
 
@@ -28,11 +28,11 @@ ENDINIT
 
 #endif
 
-static	void	call_engine_inner(Code *entry_point) NO_RETURN;
+static	void	call_engine_inner(MR_Code *entry_point) NO_RETURN;
 
 #ifndef USE_GCC_NONLOCAL_GOTOS
-  static Code	*engine_done(void);
-  static Code	*engine_init_registers(void);
+  static MR_Code	*engine_done(void);
+  static MR_Code	*engine_init_registers(void);
 #endif
 
 bool	MR_debugflag[MR_MAXFLAG];
@@ -49,20 +49,20 @@ bool	MR_debugflag[MR_MAXFLAG];
 ** the zone structures and context structures).
 */
 void 
-init_engine(MercuryEngine *eng)
+MR_init_engine(MercuryEngine *eng)
 {
 	/*
 	** First, ensure that the truly global stuff has been initialized
 	** (if it was already initialized, this does nothing).
 	*/
 
-	init_memory();
+	MR_init_memory();
 
 #ifndef USE_GCC_NONLOCAL_GOTOS
 	{
 		static bool made_engine_done_label = FALSE;
 		if (!made_engine_done_label) {
-			make_label("engine_done", LABEL(engine_done),
+			MR_make_label("engine_done", MR_LABEL(engine_done),
 				engine_done);
 			made_engine_done_label = TRUE;
 		}
@@ -75,42 +75,34 @@ init_engine(MercuryEngine *eng)
 	*/
 
 #ifndef	CONSERVATIVE_GC
-	eng->heap_zone = create_zone("heap", 1, heap_size, next_offset(),
-			heap_zone_size, default_handler);
+	eng->heap_zone = MR_create_zone("heap", 1, MR_heap_size,
+			MR_next_offset(), MR_heap_zone_size,
+			MR_default_handler);
 	eng->e_hp = eng->heap_zone->min;
 
 #ifdef	NATIVE_GC
-	eng->heap_zone2 = create_zone("heap2", 1, heap_size, next_offset(),
-			heap_zone_size, default_handler);
+	eng->heap_zone2 = MR_create_zone("heap2", 1, MR_heap_size,
+			MR_next_offset(), MR_heap_zone_size,
+			MR_default_handler);
 
   #ifdef MR_DEBUG_AGC_PRINT_VARS
-	eng->debug_heap_zone = create_zone("debug_heap", 1, debug_heap_size,
-			next_offset(), debug_heap_zone_size, default_handler);
+	eng->debug_heap_zone = MR_create_zone("debug_heap", 1,
+			MR_debug_heap_size, MR_next_offset(),
+			MR_debug_heap_zone_size, MR_default_handler);
   #endif
 #endif
 
-	eng->solutions_heap_zone = create_zone("solutions_heap", 1,
-			solutions_heap_size, next_offset(),
-			solutions_heap_zone_size, default_handler);
+	eng->solutions_heap_zone = MR_create_zone("solutions_heap", 1,
+			MR_solutions_heap_size, MR_next_offset(),
+			MR_solutions_heap_zone_size, MR_default_handler);
 	eng->e_sol_hp = eng->solutions_heap_zone->min;
 
-	eng->global_heap_zone = create_zone("global_heap", 1,
-			global_heap_size, next_offset(),
-			global_heap_zone_size, default_handler);
+	eng->global_heap_zone = MR_create_zone("global_heap", 1,
+			MR_global_heap_size, MR_next_offset(),
+			MR_global_heap_zone_size, MR_default_handler);
 	eng->e_global_hp = eng->global_heap_zone->min;
 #endif
 
-#ifdef MR_LOWLEVEL_DEBUG
-	/*
-	** Create the dumpstack, used for debugging stack traces.
-	** Note that we can just make the dumpstack the same size as
-	** the detstack and we never have to worry about the dumpstack
-	** overflowing.
-	*/
-	dumpstack_zone = create_zone("dumpstack", 1, detstack_size,
-		next_offset(), detstack_zone_size, default_handler);
-#endif
-
 #ifdef	MR_THREAD_SAFE
 	eng->owner_thread = pthread_self();
 	eng->c_depth = 0;
@@ -121,46 +113,57 @@ init_engine(MercuryEngine *eng)
 	** Finally, allocate an initialize context (Mercury thread)
 	** in the engine and initialize the per-context stuff.
 	*/
-	eng->this_context = create_context();
+	eng->this_context = MR_create_context();
 }
 
 /*---------------------------------------------------------------------------*/
 
-void finalize_engine(MercuryEngine *eng)
+void MR_finalize_engine(MercuryEngine *eng)
 {
-	
+	/*
+	** XXX there are lots of other resources in MercuryEngine that
+	** might need to be finalized.  
+	*/
+	MR_destroy_context(eng->this_context);
 }
 
 /*---------------------------------------------------------------------------*/
 
-MercuryEngine *create_engine(void)
+MercuryEngine *
+MR_create_engine(void)
 {
 	MercuryEngine *eng;
 
-	eng = MR_GC_NEW(MercuryEngine);
-	init_engine(eng);
+	/*
+	** We need to use MR_GC_NEW_UNCOLLECTABLE() here,
+	** rather than MR_GC_NEW(), since the engine pointer
+	** will normally be stored in thread-local storage, which is
+	** not traced by the conservative garbage collector.
+	*/
+	eng = MR_GC_NEW_UNCOLLECTABLE(MercuryEngine);
+	MR_init_engine(eng);
 	return eng;
 }
 
 void
-destroy_engine(MercuryEngine *eng)
+MR_destroy_engine(MercuryEngine *eng)
 {
-	finalize_engine(eng);
+	MR_finalize_engine(eng);
 	MR_GC_free(eng);
 }
 
 /*---------------------------------------------------------------------------*/
 
 /*
-** Word *
-** MR_call_engine(Code *entry_point, bool catch_exceptions)
+** MR_Word *
+** MR_call_engine(MR_Code *entry_point, bool catch_exceptions)
 **
 **	This routine calls a Mercury routine from C.
 **
 **	The called routine should be det/semidet/cc_multi/cc_nondet.
 **
 **	If the called routine returns normally (this includes the case of a
-**	semidet/cc_nondet routine failing, i.e. returning with r1 = FALSE),
+**	semidet/cc_nondet routine failing, i.e. returning with MR_r1 = FALSE),
 **	then MR_call_engine() will return NULL.
 **
 **	If the called routine exits by throwing an exception, then the
@@ -175,14 +178,14 @@ destroy_engine(MercuryEngine *eng)
 **	to MR_call_engine().  Specifically, the non-transient real registers
 **	must have valid values, and the fake_reg copies of the transient
 **	(register window) registers must have valid values; call_engine()
-**	will call restore_transient_registers() and will then assume that
+**	will call MR_restore_transient_registers() and will then assume that
 **	all the registers have been correctly set up.
 **
-**	call_engine() will call save_registers() before returning.
+**	call_engine() will call MR_save_registers() before returning.
 **	That will copy the real registers we use to the fake_reg array.
 **
-**	Beware, however, that if you are planning to return to C code
-**	that did not #include "mercury_regs.h" (directly or via e.g. "mercury_imp.h"),
+**	Beware, however, that if you are planning to return to C code that did
+**	not #include "mercury_regs.h" (directly or via e.g. "mercury_imp.h"),
 **	and you have fiddled with the Mercury registers or invoked
 **	call_engine() or anything like that, then you will need to
 **	save the real registers that C is using before modifying the
@@ -196,21 +199,21 @@ destroy_engine(MercuryEngine *eng)
 **	invoke call_engine() to invoke invoke Mercury routines (which
 **	in turn invoke C functions which ... etc. ad infinitum.)
 **
-**	call_engine() calls setjmp() and then invokes call_engine_inner()
+**	MR_call_engine() calls setjmp() and then invokes call_engine_inner()
 **	which does the real work.  call_engine_inner() exits by calling
-**	longjmp() to return to call_engine().  There are two 
+**	longjmp() to return to MR_call_engine().  There are two 
 **	different implementations of call_engine_inner(), one for gcc,
 **	and another portable version that works on standard ANSI C compilers.
 */
 
-Word *
-MR_call_engine(Code *entry_point, bool catch_exceptions)
+MR_Word *
+MR_call_engine(MR_Code *entry_point, bool catch_exceptions)
 {
 
 	jmp_buf		curr_jmp_buf;
 	jmp_buf		* volatile prev_jmp_buf;
 #if defined(PROFILE_TIME)
-	Code		* volatile prev_proc;
+	MR_Code		* volatile prev_proc;
 #endif
 
 	/*
@@ -219,7 +222,7 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 	** will work.
 	*/
 
-	restore_transient_registers();
+	MR_restore_transient_registers();
 
 	prev_jmp_buf = MR_ENGINE(e_jmp_buf);
 	MR_ENGINE(e_jmp_buf) = &curr_jmp_buf;
@@ -230,17 +233,27 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 	*/
 	if (catch_exceptions) {
 		MR_create_exception_handler("call_engine",
-			MR_C_LONGJMP_HANDLER, 0, ENTRY(do_fail));
+			MR_C_LONGJMP_HANDLER, 0, MR_ENTRY(MR_do_fail));
 	}
-		
+
 	/*
 	** Mark this as the spot to return to.
 	*/
-	if (setjmp(curr_jmp_buf)) {
-		Word	* this_frame;
-		Word	* exception;
 
-		debugmsg0("...caught longjmp\n");
+#ifdef	MR_DEBUG_JMPBUFS
+	printf("engine setjmp %p\n", curr_jmp_buf);
+#endif
+
+	if (setjmp(curr_jmp_buf)) {
+		MR_Word	* this_frame;
+		MR_Word	* exception;
+
+#ifdef	MR_DEBUG_JMPBUFS
+		printf("engine caught jmp %p %p\n",
+			prev_jmp_buf, MR_ENGINE(e_jmp_buf));
+#endif
+
+		MR_debugmsg0("...caught longjmp\n");
 		/*
 		** On return,
 		** set MR_prof_current_proc to be the caller proc again
@@ -248,8 +261,8 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 		** restore the registers (since longjmp may clobber them),
 		** and restore the saved value of MR_ENGINE(e_jmp_buf).
 		*/
-		update_prof_current_proc(prev_proc);
-		restore_registers();
+		MR_update_prof_current_proc(prev_proc);
+		MR_restore_registers();
 		MR_ENGINE(e_jmp_buf) = prev_jmp_buf;
 		if (catch_exceptions) {
 			/*
@@ -267,14 +280,14 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 			** If we added an exception hander, but we didn't
 			** get an exception, then we need to remove the
 			** exception handler frames from the nondet stack
-			** and deallocate the trail ticket allocated by
+			** and prune the trail ticket allocated by
 			** MR_create_exception_handler().
 			*/
 			this_frame = MR_curfr;
 			MR_maxfr = MR_prevfr_slot(this_frame);
 			MR_curfr = MR_succfr_slot(this_frame);
 #ifdef MR_USE_TRAIL
-			MR_discard_ticket();
+			MR_prune_ticket();
 #endif
 		}
 		return NULL;
@@ -292,7 +305,7 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 #ifdef PROFILE_CALLS
   #ifdef PROFILE_TIME
 	if (MR_prof_current_proc != NULL) {
-		PROFILE(entry_point, MR_prof_current_proc);
+		MR_PROFILE(entry_point, MR_prof_current_proc);
 	}
   #else
 	/*
@@ -322,7 +335,7 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 	*/
 #ifdef PROFILE_TIME
 	prev_proc = MR_prof_current_proc;
-	set_prof_current_proc(entry_point);
+	MR_set_prof_current_proc(entry_point);
 #endif
 
 	call_engine_inner(entry_point);
@@ -333,7 +346,7 @@ MR_call_engine(Code *entry_point, bool catch_exceptions)
 /* The gcc-specific version */
 
 static void 
-call_engine_inner(Code *entry_point)
+call_engine_inner(MR_Code *entry_point)
 {
 	/*
 	** Allocate some space for local variables in other
@@ -368,7 +381,8 @@ call_engine_inner(Code *entry_point)
 
 	if (!initialized)
 	{
-		make_label("engine_done", LABEL(engine_done), engine_done);
+		MR_make_label("engine_done", MR_LABEL(engine_done),
+			engine_done);
 		initialized = TRUE;
 	}
 }
@@ -379,19 +393,19 @@ call_engine_inner(Code *entry_point)
 	** call mechanism
 	*/
 
-	restore_transient_registers();
+	MR_restore_transient_registers();
 
 	/*
 	** We save the address of the locals in a global pointer to make
 	** sure that gcc can't optimize them away.
 	*/
 
-	global_pointer = locals;
+	MR_global_pointer = locals;
 
 #ifdef MR_LOWLEVEL_DEBUG
 	memset((void *)locals, MAGIC_MARKER, LOCALS_SIZE);
 #endif
-	debugmsg1("in `call_engine_inner', locals at %p\n", (void *)locals);
+	MR_debugmsg1("in `call_engine_inner', locals at %p\n", (void *)locals);
 
 	/*
 	** We need to ensure that there is at least one
@@ -408,7 +422,7 @@ call_engine_inner(Code *entry_point)
 	** Also for gcc versions >= egcs1.1, we need to ensure that
 	** there is at least one jump to an unknown label.
 	*/
-	goto *dummy_identify_function(&&dummy_label);
+	goto *MR_dummy_identify_function(&&dummy_label);
 dummy_label:
 
 	/*
@@ -435,9 +449,9 @@ dummy_label:
 	** Now just call the entry point
 	*/
 
-	noprof_call(entry_point, LABEL(engine_done));
+	MR_noprof_call(entry_point, MR_LABEL(engine_done));
 
-Define_label(engine_done);
+MR_define_label(engine_done);
 
 	/*
 	** Decrement the number of times we've entered this
@@ -466,7 +480,7 @@ Define_label(engine_done);
 }
 #endif
 
-	debugmsg1("in label `engine_done', locals at %p\n", locals);
+	MR_debugmsg1("in label `engine_done', locals at %p\n", locals);
 
 #ifdef MR_LOWLEVEL_DEBUG
 	/*
@@ -474,7 +488,7 @@ Define_label(engine_done);
 	** was actually used.
 	*/
 
-	if (check_space) {
+	if (MR_check_space) {
 		int	low = 0, high = LOCALS_SIZE;
 		int	used_low, used_high;
 
@@ -503,14 +517,19 @@ Define_label(engine_done);
 	** save them first.
 	*/
 	MR_ENGINE(e_exception) = NULL;
-	save_registers();
-	debugmsg0("longjmping out...\n");
+	MR_save_registers();
+
+#ifdef	MR_DEBUG_JMPBUFS
+	printf("engine longjmp %p\n", MR_ENGINE(e_jmp_buf));
+#endif
+
+	MR_debugmsg0("longjmping out...\n");
 	longjmp(*(MR_ENGINE(e_jmp_buf)), 1);
 }} /* end call_engine_inner() */
 
 /* with nonlocal gotos, we don't save the previous locations */
 void 
-dump_prev_locations(void) {}
+MR_dump_prev_locations(void) {}
 
 #else /* not USE_GCC_NONLOCAL_GOTOS */
 
@@ -529,20 +548,20 @@ dump_prev_locations(void) {}
 ** would get mucked up because of the function call from call_engine_inner().
 */
 
-static Code *
+static MR_Code *
 engine_done(void)
 {
 	MR_ENGINE(e_exception) = NULL;
-	save_registers();
-	debugmsg0("longjmping out...\n");
+	MR_save_registers();
+	MR_debugmsg0("longjmping out...\n");
 	longjmp(*(MR_ENGINE(e_jmp_buf)), 1);
 }
 
-static Code *
+static MR_Code *
 engine_init_registers(void)
 {
-	restore_transient_registers();
-	MR_succip = (Code *) engine_done;
+	MR_restore_transient_registers();
+	MR_succip = (MR_Code *) engine_done;
 	return NULL;
 }
 
@@ -555,13 +574,13 @@ engine_init_registers(void)
 
 #define NUM_PREV_FPS	40
 
-typedef Code	*Func(void);
+typedef MR_Code	*Func(void);
 
-static Code 	*prev_fps[NUM_PREV_FPS];
+static MR_Code 	*prev_fps[NUM_PREV_FPS];
 static int	prev_fp_index = 0;
 
 void 
-dump_prev_locations(void)
+MR_dump_prev_locations(void)
 {
 	int i, pos;
 
@@ -572,15 +591,15 @@ dump_prev_locations(void)
 		printf("previous %d locations:\n", NUM_PREV_FPS);
 		for (i = 0; i < NUM_PREV_FPS; i++) {
 			pos = (i + prev_fp_index) % NUM_PREV_FPS;
-			printlabel(prev_fps[pos]);
+			MR_printlabel(stdout, prev_fps[pos]);
 		}
 	}
 }
 
 static void 
-call_engine_inner(Code *entry_point)
+call_engine_inner(MR_Code *entry_point)
 {
-	reg	Func	*fp;
+	register Func	*fp;
 
 	/*
 	** Start up the actual engine.
@@ -608,13 +627,13 @@ if (!MR_tracedebug) {
 #endif
 	for (;;)
 	{
-		prev_fps[prev_fp_index] = (Code *) fp;
+		prev_fps[prev_fp_index] = (MR_Code *) fp;
 
 		if (++prev_fp_index >= NUM_PREV_FPS)
 			prev_fp_index = 0;
 
-		debuggoto(fp);
-		debugsreg();
+		MR_debuggoto(fp);
+		MR_debugsreg();
 		fp = (Func *) (*fp)();
 	}
 } /* end call_engine_inner() */
@@ -623,7 +642,7 @@ if (!MR_tracedebug) {
 /*---------------------------------------------------------------------------*/
 
 void
-terminate_engine(void)
+MR_terminate_engine(void)
 {
 	/*
 	** we don't bother to deallocate memory...
@@ -633,48 +652,48 @@ terminate_engine(void)
 
 /*---------------------------------------------------------------------------*/
 
-Define_extern_entry(do_redo);
-Define_extern_entry(do_fail);
-Define_extern_entry(do_succeed);
-Define_extern_entry(do_last_succeed);
-Define_extern_entry(do_not_reached);
-Define_extern_entry(exception_handler_do_fail);
+MR_define_extern_entry(MR_do_redo);
+MR_define_extern_entry(MR_do_fail);
+MR_define_extern_entry(MR_do_succeed);
+MR_define_extern_entry(MR_do_last_succeed);
+MR_define_extern_entry(MR_do_not_reached);
+MR_define_extern_entry(MR_exception_handler_do_fail);
 
-BEGIN_MODULE(special_labels_module)
-	init_entry_ai(do_redo);
-	init_entry_ai(do_fail);
-	init_entry_ai(do_succeed);
-	init_entry_ai(do_last_succeed);
-	init_entry_ai(do_not_reached);
-	init_entry_ai(exception_handler_do_fail);
-BEGIN_CODE
+MR_BEGIN_MODULE(special_labels_module)
+	MR_init_entry_ai(MR_do_redo);
+	MR_init_entry_ai(MR_do_fail);
+	MR_init_entry_ai(MR_do_succeed);
+	MR_init_entry_ai(MR_do_last_succeed);
+	MR_init_entry_ai(MR_do_not_reached);
+	MR_init_entry_ai(MR_exception_handler_do_fail);
+MR_BEGIN_CODE
 
-Define_entry(do_redo);
+MR_define_entry(MR_do_redo);
 	MR_redo();
 
-Define_entry(do_fail);
+MR_define_entry(MR_do_fail);
 	MR_fail();
 
-Define_entry(do_succeed);
+MR_define_entry(MR_do_succeed);
 	MR_succeed();
 
-Define_entry(do_last_succeed);
+MR_define_entry(MR_do_last_succeed);
 	MR_succeed_discard();
 
-Define_entry(do_not_reached);
-	fatal_error("reached not_reached\n");
+MR_define_entry(MR_do_not_reached);
+	MR_fatal_error("reached not_reached\n");
 
-Define_entry(exception_handler_do_fail);
+MR_define_entry(MR_exception_handler_do_fail);
 	/*
-	** `exception_handler_do_fail' is the same as `do_fail':
-	** it just invokes fail().  The reason we don't just use
-	** `do_fail' for this is that when unwinding the stack we
-	** check for a redoip of `exception_handler_do_fail' and
+	** `MR_exception_handler_do_fail' is the same as `MR_do_fail':
+	** it just invokes MR_fail().  The reason we don't just use
+	** `MR_do_fail' for this is that when unwinding the stack we
+	** check for a redoip of `MR_exception_handler_do_fail' and
 	** handle it specially.
 	*/
 	MR_fail();
 
-END_MODULE
+MR_END_MODULE
 
 void mercury_sys_init_engine(void); /* suppress gcc warning */
 void mercury_sys_init_engine(void) {
