@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 1998-1999 The University of Melbourne.
+** Copyright (C) 1998-2001 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
@@ -12,40 +12,45 @@
 
 #include "mercury_imp.h"
 #include "mercury_stack_trace.h"
+#include "mercury_debug.h"
 #include <stdio.h>
 
-static	void	MR_dump_stack_record_init(bool include_contexts);
+static	void	MR_dump_stack_record_init(bool include_trace_data,
+			bool include_contexts);
 static	void	MR_dump_stack_record_frame(FILE *fp,
-			const MR_Stack_Layout_Label *,
-			Word *base_sp, Word *base_curfr, 
+			const MR_Label_Layout *label_layout,
+			MR_Word *base_sp, MR_Word *base_curfr, 
 			MR_Print_Stack_Record print_stack_record);
 static	void	MR_dump_stack_record_flush(FILE *fp, 
 			MR_Print_Stack_Record print_stack_record);
 
+static	void	MR_print_proc_id_internal(FILE *fp,
+			const MR_Proc_Layout *entry, bool spec);
+
 static	void	MR_maybe_print_context(FILE *fp,
 			const char *filename, int lineno);
 static	void	MR_maybe_print_parent_context(FILE *fp, bool print_parent,
 			bool verbose, const char *filename, int lineno);
 
 /* see comments in mercury_stack_trace.h */
-Code	*MR_stack_trace_bottom;
-Word	*MR_nondet_stack_trace_bottom;
+MR_Code	*MR_stack_trace_bottom;
+MR_Word	*MR_nondet_stack_trace_bottom;
 
 void
-MR_dump_stack(Code *success_pointer, Word *det_stack_pointer,
-	Word *current_frame, bool include_trace_data)
+MR_dump_stack(MR_Code *success_pointer, MR_Word *det_stack_pointer,
+	MR_Word *current_frame, bool include_trace_data)
 {
 #ifndef	MR_STACK_TRACE
 	fprintf(stderr, "Stack dump not available in this grade.\n");
 #else
 
-	const MR_Internal		*label;
-	const MR_Stack_Layout_Label	*layout;
-	const char			*result;
+	const MR_Internal	*label;
+	const MR_Label_Layout	*layout;
+	const char		*result;
 
 	fprintf(stderr, "Stack dump follows:\n");
 
-	do_init_modules();
+	MR_do_init_modules();
 	label = MR_lookup_internal_by_addr(success_pointer);
 	if (label == NULL) {
 		fprintf(stderr, "internal label not found\n");
@@ -63,22 +68,23 @@ MR_dump_stack(Code *success_pointer, Word *det_stack_pointer,
 }
 
 const char *
-MR_dump_stack_from_layout(FILE *fp, const MR_Stack_Layout_Label *label_layout,
-	Word *det_stack_pointer, Word *current_frame, bool include_trace_data,
-	bool include_contexts, MR_Print_Stack_Record print_stack_record)
+MR_dump_stack_from_layout(FILE *fp, const MR_Label_Layout *label_layout,
+	MR_Word *det_stack_pointer, MR_Word *current_frame,
+	bool include_trace_data, bool include_contexts,
+	MR_Print_Stack_Record print_stack_record)
 {
 	MR_Stack_Walk_Step_Result	result;
-	const MR_Stack_Layout_Entry	*entry_layout;
-	const MR_Stack_Layout_Label	*cur_label_layout;
-	const MR_Stack_Layout_Label	*prev_label_layout;
+	const MR_Proc_Layout		*entry_layout;
+	const MR_Label_Layout		*cur_label_layout;
+	const MR_Label_Layout		*prev_label_layout;
 	const char			*problem;
-	Word				*stack_trace_sp;
-	Word				*stack_trace_curfr;
-	Word				*old_trace_sp;
-	Word				*old_trace_curfr;
+	MR_Word				*stack_trace_sp;
+	MR_Word				*stack_trace_curfr;
+	MR_Word				*old_trace_sp;
+	MR_Word				*old_trace_curfr;
 
-	do_init_modules();
-	MR_dump_stack_record_init(include_contexts);
+	MR_do_init_modules();
+	MR_dump_stack_record_init(include_trace_data, include_contexts);
 
 	stack_trace_sp = det_stack_pointer;
 	stack_trace_curfr = current_frame;
@@ -98,30 +104,16 @@ MR_dump_stack_from_layout(FILE *fp, const MR_Stack_Layout_Label *label_layout,
 			MR_dump_stack_record_flush(fp, print_stack_record);
 			return problem;
 		} else if (result == STEP_ERROR_AFTER) {
-			if (include_trace_data) {
-				MR_dump_stack_record_frame(fp,
-					prev_label_layout,
-					old_trace_sp, old_trace_curfr, 
-					print_stack_record);
-			} else {
-				MR_dump_stack_record_frame(fp,
-					prev_label_layout,
-					NULL, NULL, print_stack_record);
-			}
+			MR_dump_stack_record_frame(fp, prev_label_layout,
+				old_trace_sp, old_trace_curfr, 
+				print_stack_record);
 
 			MR_dump_stack_record_flush(fp, print_stack_record);
 			return problem;
 		} else {
-			if (include_trace_data) {
-				MR_dump_stack_record_frame(fp,
-					prev_label_layout,
-					old_trace_sp, old_trace_curfr, 
-					print_stack_record);
-			} else {
-				MR_dump_stack_record_frame(fp,
-					prev_label_layout,
-					NULL, NULL, print_stack_record);
-			}
+			MR_dump_stack_record_frame(fp, prev_label_layout,
+				old_trace_sp, old_trace_curfr, 
+				print_stack_record);
 		}
 	} while (cur_label_layout != NULL);
 
@@ -129,13 +121,13 @@ MR_dump_stack_from_layout(FILE *fp, const MR_Stack_Layout_Label *label_layout,
 	return NULL;
 }
 
-const MR_Stack_Layout_Label *
-MR_find_nth_ancestor(const MR_Stack_Layout_Label *label_layout,
-	int ancestor_level, Word **stack_trace_sp, Word **stack_trace_curfr,
+const MR_Label_Layout *
+MR_find_nth_ancestor(const MR_Label_Layout *label_layout, int ancestor_level,
+	MR_Word **stack_trace_sp, MR_Word **stack_trace_curfr,
 	const char **problem)
 {
 	MR_Stack_Walk_Step_Result	result;
-	const MR_Stack_Layout_Label	*return_label_layout;
+	const MR_Label_Layout		*return_label_layout;
 	int				i;
 
 	if (ancestor_level < 0) {
@@ -143,7 +135,7 @@ MR_find_nth_ancestor(const MR_Stack_Layout_Label *label_layout,
 		return NULL;
 	}
 
-	do_init_modules();
+	MR_do_init_modules();
 	*problem = NULL;
 	for (i = 0; i < ancestor_level && label_layout != NULL; i++) {
 		result = MR_stack_walk_step(label_layout->MR_sll_entry,
@@ -166,9 +158,9 @@ MR_find_nth_ancestor(const MR_Stack_Layout_Label *label_layout,
 }
 
 MR_Stack_Walk_Step_Result
-MR_stack_walk_step(const MR_Stack_Layout_Entry *entry_layout,
-	const MR_Stack_Layout_Label **return_label_layout,
-	Word **stack_trace_sp_ptr, Word **stack_trace_curfr_ptr,
+MR_stack_walk_step(const MR_Proc_Layout *entry_layout,
+	const MR_Label_Layout **return_label_layout,
+	MR_Word **stack_trace_sp_ptr, MR_Word **stack_trace_curfr_ptr,
 	const char **problem_ptr)
 {
 	MR_Internal		*label;
@@ -176,7 +168,7 @@ MR_stack_walk_step(const MR_Stack_Layout_Entry *entry_layout,
 	MR_Long_Lval_Type	type;
 	int			number;
 	int			determinism;
-	Code			*success;
+	MR_Code			*success;
 
 	*return_label_layout = NULL;
 
@@ -201,7 +193,7 @@ MR_stack_walk_step(const MR_Stack_Layout_Entry *entry_layout,
 			return STEP_ERROR_AFTER;
 		}
 
-		success = (Code *) MR_based_stackvar(*stack_trace_sp_ptr,
+		success = (MR_Code *) MR_based_stackvar(*stack_trace_sp_ptr,
 					number);
 		*stack_trace_sp_ptr = *stack_trace_sp_ptr -
 			entry_layout->MR_sle_stack_slots;
@@ -230,11 +222,11 @@ MR_stack_walk_step(const MR_Stack_Layout_Entry *entry_layout,
 }
 
 void
-MR_dump_nondet_stack_from_layout(FILE *fp, Word *base_maxfr)
+MR_dump_nondet_stack_from_layout(FILE *fp, MR_Word *base_maxfr)
 {
 	int	frame_size;
 
-	do_init_modules();
+	MR_do_init_modules();
 
 	/*
 	** The comparison operator in the condition of the while loop
@@ -250,7 +242,7 @@ MR_dump_nondet_stack_from_layout(FILE *fp, Word *base_maxfr)
 			MR_print_nondstackptr(fp, base_maxfr);
 			fprintf(fp, ": temp\n");
 			fprintf(fp, " redoip: ");
-			printlabel(MR_redoip_slot(base_maxfr));
+			MR_printlabel(fp, MR_redoip_slot(base_maxfr));
 			fprintf(fp, " redofr: ");
 			MR_print_nondstackptr(fp, MR_redofr_slot(base_maxfr));
 			fprintf(fp, " \n");
@@ -258,7 +250,7 @@ MR_dump_nondet_stack_from_layout(FILE *fp, Word *base_maxfr)
 			MR_print_nondstackptr(fp, base_maxfr);
 			fprintf(fp, ": temp\n");
 			fprintf(fp, " redoip: ");
-			printlabel(MR_redoip_slot(base_maxfr));
+			MR_printlabel(fp, MR_redoip_slot(base_maxfr));
 			fprintf(fp, " redofr: ");
 			MR_print_nondstackptr(fp, MR_redofr_slot(base_maxfr));
 			fprintf(fp, " \n");
@@ -270,12 +262,12 @@ MR_dump_nondet_stack_from_layout(FILE *fp, Word *base_maxfr)
 			fprintf(fp, ": ordinary, %d words\n",
 				frame_size);
 			fprintf(fp, " redoip: ");
-			printlabel(MR_redoip_slot(base_maxfr));
+			MR_printlabel(fp, MR_redoip_slot(base_maxfr));
 			fprintf(fp, " redofr: ");
 			MR_print_nondstackptr(fp, MR_redofr_slot(base_maxfr));
 			fprintf(fp, " \n");
 			fprintf(fp, " succip: ");
-			printlabel(MR_succip_slot(base_maxfr));
+			MR_printlabel(fp, MR_succip_slot(base_maxfr));
 			fprintf(fp, " succfr: ");
 			MR_print_nondstackptr(fp, MR_succfr_slot(base_maxfr));
 			fprintf(fp, " \n");
@@ -285,36 +277,39 @@ MR_dump_nondet_stack_from_layout(FILE *fp, Word *base_maxfr)
 	}
 }
 
-static	const MR_Stack_Layout_Entry	*prev_entry_layout;
-static	int				prev_entry_layout_count;
-static	int				prev_entry_start_level;
-static	Word				*prev_entry_base_sp;
-static	Word				*prev_entry_base_curfr;
-static	const char			*prev_entry_filename;
-static	int				prev_entry_linenumber;
-static	bool				prev_entry_context_mismatch;
-static	int				current_level;
-static	bool				contexts_enabled;
+static	const MR_Proc_Layout	*prev_entry_layout;
+static	int			prev_entry_layout_count;
+static	int			prev_entry_start_level;
+static	MR_Word			*prev_entry_base_sp;
+static	MR_Word			*prev_entry_base_curfr;
+static	const char		*prev_entry_filename;
+static	int			prev_entry_linenumber;
+static	const char		*prev_entry_goal_path;
+static	bool			prev_entry_context_mismatch;
+static	int			current_level;
+static	bool			trace_data_enabled;
+static	bool			contexts_enabled;
 
 static void
-MR_dump_stack_record_init(bool include_contexts)
+MR_dump_stack_record_init(bool include_trace_data, bool include_contexts)
 {
 	prev_entry_layout = NULL;
 	prev_entry_layout_count = 0;
 	prev_entry_start_level = 0;
 	current_level = 0;
 	contexts_enabled = include_contexts;
+	trace_data_enabled = include_trace_data;
 }
 
 static void
-MR_dump_stack_record_frame(FILE *fp, const MR_Stack_Layout_Label *label_layout,
-	Word *base_sp, Word *base_curfr,
+MR_dump_stack_record_frame(FILE *fp, const MR_Label_Layout *label_layout,
+	MR_Word *base_sp, MR_Word *base_curfr,
 	MR_Print_Stack_Record print_stack_record)
 {
-	const MR_Stack_Layout_Entry	*entry_layout;
-	const char			*filename;
-	int				linenumber;
-	bool				must_flush;
+	const MR_Proc_Layout	*entry_layout;
+	const char		*filename;
+	int			linenumber;
+	bool			must_flush;
 
 	entry_layout = label_layout->MR_sll_entry;
 	if (! MR_find_context(label_layout, &filename, &linenumber)
@@ -333,17 +328,16 @@ MR_dump_stack_record_frame(FILE *fp, const MR_Stack_Layout_Label *label_layout,
 		(entry_layout != prev_entry_layout) ||
 
 		/*
-		** We cannot merge two calls even to the same procedure.
+		** We cannot merge two calls even to the same procedure
 		** if we are printing trace data, since this will differ
 		** between the calls.
 		**
 		** Note that it is not possible for two calls to the same
 		** procedure to differ on whether the procedure has trace
-		** layout data or not; this is why we don't have to test
-		** prev_entry_base_sp and prev_entry_base_curfr.
+		** layout data or not.
 		*/
 
-		((base_sp != NULL) || (base_curfr != NULL));
+		trace_data_enabled;
 
 	if (must_flush) {
 		MR_dump_stack_record_flush(fp, print_stack_record);
@@ -355,6 +349,7 @@ MR_dump_stack_record_frame(FILE *fp, const MR_Stack_Layout_Label *label_layout,
 		prev_entry_base_curfr = base_curfr;
 		prev_entry_filename = filename;
 		prev_entry_linenumber = linenumber;
+		prev_entry_goal_path = MR_label_goal_path(label_layout);
 		prev_entry_context_mismatch = FALSE;
 	} else {
 		prev_entry_layout_count++;
@@ -376,20 +371,21 @@ MR_dump_stack_record_flush(FILE *fp, MR_Print_Stack_Record print_stack_record)
 			prev_entry_layout_count, prev_entry_start_level,
 			prev_entry_base_sp, prev_entry_base_curfr,
 			prev_entry_filename, prev_entry_linenumber,
-			prev_entry_context_mismatch);
+			prev_entry_goal_path, prev_entry_context_mismatch);
 	}
 }
 
 void
-MR_dump_stack_record_print(FILE *fp, const MR_Stack_Layout_Entry *entry_layout,
-	int count, int start_level, Word *base_sp, Word *base_curfr,
-	const char *filename, int linenumber, bool context_mismatch)
+MR_dump_stack_record_print(FILE *fp, const MR_Proc_Layout *entry_layout,
+	int count, int start_level, MR_Word *base_sp, MR_Word *base_curfr,
+	const char *filename, int linenumber, const char *goal_path,
+	bool context_mismatch)
 {
 	fprintf(fp, "%4d ", start_level);
 
 	if (count > 1) {
 		fprintf(fp, " %3d* ", count);
-	} else if ((base_sp == NULL) && (base_curfr == NULL)) {
+	} else if (! trace_data_enabled) {
 		fprintf(fp, "%5s ", "");
 	} else {
 		/*
@@ -399,21 +395,30 @@ MR_dump_stack_record_print(FILE *fp, const MR_Stack_Layout_Entry *entry_layout,
 		*/
 	}
 
-	MR_print_call_trace_info(fp, entry_layout, base_sp, base_curfr);
+	MR_maybe_print_call_trace_info(fp, trace_data_enabled, entry_layout,
+		base_sp, base_curfr);
 	MR_print_proc_id(fp, entry_layout);
 	if (strdiff(filename, "") && linenumber > 0) {
 		fprintf(fp, " (%s:%d%s)", filename, linenumber,
 			context_mismatch ? " and others" : "");
 	}
 
+	if (trace_data_enabled) {
+		if (strdiff(goal_path, "")) {
+			fprintf(fp, " %s", goal_path);
+		} else {
+			fprintf(fp, " (empty)");
+		}
+	}
+
 	fprintf(fp, "\n");
 }
 
 bool
-MR_find_context(const MR_Stack_Layout_Label *label, const char **fileptr,
+MR_find_context(const MR_Label_Layout *label, const char **fileptr,
 	int *lineptr)
 {
-	const MR_Stack_Layout_Entry	*proc;
+	const MR_Proc_Layout		*proc;
 	const MR_Module_Layout		*module;
 	const MR_Module_File_Layout	*file_layout;
 	int				i, j;
@@ -439,73 +444,107 @@ MR_find_context(const MR_Stack_Layout_Label *label, const char **fileptr,
 }
 
 void
-MR_print_call_trace_info(FILE *fp, const MR_Stack_Layout_Entry *entry,
-	Word *base_sp, Word *base_curfr)
+MR_maybe_print_call_trace_info(FILE *fp, bool include_trace_data,
+	const MR_Proc_Layout *entry, MR_Word *base_sp, MR_Word *base_curfr)
+{
+	if (include_trace_data) {
+		MR_print_call_trace_info(fp, entry, base_sp, base_curfr);
+	}
+}
+
+/*
+** Note that MR_print_call_trace_info is more permissive than its documentation
+** in the header file.
+*/
+
+void
+MR_print_call_trace_info(FILE *fp, const MR_Proc_Layout *entry,
+	MR_Word *base_sp, MR_Word *base_curfr)
 {
 	bool	print_details;
 
-	if (base_sp != NULL && base_curfr != NULL) {
-		if (MR_ENTRY_LAYOUT_HAS_EXEC_TRACE(entry)) {
-			Integer maybe_from_full =
-				entry->MR_sle_maybe_from_full;
-			if (maybe_from_full > 0) {
-				/*
-				** For procedures compiled with shallow
-				** tracing, the details will be valid only
-				** if the value of MR_from_full saved in
-				** the appropriate stack slot was TRUE.
-			    	*/
-				if (MR_DETISM_DET_STACK(entry->MR_sle_detism)) {
-					print_details = MR_based_stackvar(
-						base_sp, maybe_from_full);
-				} else {
-					print_details = MR_based_framevar(
-						base_curfr, maybe_from_full);
-				}
-			} else {
-				/*
-				** For procedures compiled with full tracing,
-				** we can always print out the details.
-				*/
-				print_details = TRUE;
-			}
-		} else {
-			print_details = FALSE;
-		}
+	if (base_sp == NULL || base_curfr == NULL) {
+		return;
+	}
 
-		if (print_details) {
+	if (MR_ENTRY_LAYOUT_HAS_EXEC_TRACE(entry)) {
+		MR_Integer maybe_from_full =
+			entry->MR_sle_maybe_from_full;
+		if (maybe_from_full > 0) {
+			/*
+			** For procedures compiled with shallow
+			** tracing, the details will be valid only
+			** if the value of MR_from_full saved in
+			** the appropriate stack slot was TRUE.
+			*/
 			if (MR_DETISM_DET_STACK(entry->MR_sle_detism)) {
-				fprintf(fp, "%7lu %7lu %4lu ",
-					(unsigned long)
-					MR_event_num_stackvar(base_sp) + 1,
-					(unsigned long)
-					MR_call_num_stackvar(base_sp),
-					(unsigned long)
-					MR_call_depth_stackvar(base_sp));
+				print_details = MR_based_stackvar(
+					base_sp, maybe_from_full);
 			} else {
-				fprintf(fp, "%7lu %7lu %4lu ",
-					(unsigned long)
-					MR_event_num_framevar(base_curfr) + 1,
-					(unsigned long)
-					MR_call_num_framevar(base_curfr),
-					(unsigned long)
-					MR_call_depth_framevar(base_curfr));
+				print_details = MR_based_framevar(
+					base_curfr, maybe_from_full);
 			}
 		} else {
-			/* ensure that the remaining columns line up */
-			fprintf(fp, "%21s", "");
+			/*
+			** For procedures compiled with full tracing,
+			** we can always print out the details.
+			*/
+			print_details = TRUE;
 		}
+	} else {
+		print_details = FALSE;
+	}
+
+	if (print_details) {
+		if (MR_DETISM_DET_STACK(entry->MR_sle_detism)) {
+			fprintf(fp, "%7lu %7lu %4lu ",
+				(unsigned long)
+				MR_event_num_stackvar(base_sp) + 1,
+				(unsigned long)
+				MR_call_num_stackvar(base_sp),
+				(unsigned long)
+				MR_call_depth_stackvar(base_sp));
+		} else {
+			fprintf(fp, "%7lu %7lu %4lu ",
+				(unsigned long)
+				MR_event_num_framevar(base_curfr) + 1,
+				(unsigned long)
+				MR_call_num_framevar(base_curfr),
+				(unsigned long)
+				MR_call_depth_framevar(base_curfr));
+		}
+	} else {
+		/* ensure that the remaining columns line up */
+		fprintf(fp, "%21s", "");
 	}
 }
 
 void
-MR_print_proc_id(FILE *fp, const MR_Stack_Layout_Entry *entry)
+MR_print_proc_id(FILE *fp, const MR_Proc_Layout *entry)
+{
+	MR_print_proc_id_internal(fp, entry, FALSE);
+}
+
+void
+MR_print_proc_spec(FILE *fp, const MR_Proc_Layout *entry)
+{
+	MR_print_proc_id_internal(fp, entry, TRUE);
+}
+
+static void
+MR_print_proc_id_internal(FILE *fp, const MR_Proc_Layout *entry,
+	bool spec)
 {
 	if (! MR_ENTRY_LAYOUT_HAS_PROC_ID(entry)) {
-		fatal_error("cannot print procedure id without layout");
+		MR_fatal_error("cannot print procedure id without layout");
 	}
 
 	if (MR_ENTRY_LAYOUT_COMPILER_GENERATED(entry)) {
+		if (spec) {
+			MR_fatal_error("cannot generate specifications "
+				"for compiler generated procedures");
+		}
+
 		fprintf(fp, "%s for %s:%s/%ld-%ld",
 			entry->MR_sle_comp.MR_comp_pred_name,
 			entry->MR_sle_comp.MR_comp_type_module,
@@ -527,16 +566,22 @@ MR_print_proc_id(FILE *fp, const MR_Stack_Layout_Entry *entry)
 		{
 			fprintf(fp, "func");
 		} else {
-			fatal_error("procedure is not pred or func");
+			MR_fatal_error("procedure is not pred or func");
 		}
 
-		fprintf(fp, " %s:%s/%ld-%ld",
+		if (spec) {
+			fprintf(fp, "*");
+		} else {
+			fprintf(fp, " ");
+		}
+
+		fprintf(fp, "%s:%s/%ld-%ld",
 			entry->MR_sle_user.MR_user_decl_module,
 			entry->MR_sle_user.MR_user_name,
 			(long) entry->MR_sle_user.MR_user_arity,
 			(long) entry->MR_sle_user.MR_user_mode);
 
-		if (strcmp(entry->MR_sle_user.MR_user_decl_module,
+		if (!spec && strcmp(entry->MR_sle_user.MR_user_decl_module,
 				entry->MR_sle_user.MR_user_def_module) != 0)
 		{
 			fprintf(fp, " {%s}",
@@ -544,12 +589,15 @@ MR_print_proc_id(FILE *fp, const MR_Stack_Layout_Entry *entry)
 		}
 	}
 
-	fprintf(fp, " (%s)", MR_detism_names[entry->MR_sle_detism]);
+	if (! spec) {
+		fprintf(fp, " (%s)", MR_detism_names[entry->MR_sle_detism]);
+	}
 }
 
 void
-MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
-	const MR_Stack_Layout_Entry *entry, Word *base_sp, Word *base_curfr,
+MR_print_proc_id_trace_and_context(FILE *fp, bool include_trace_data,
+	MR_Context_Position pos, const MR_Proc_Layout *entry,
+	MR_Word *base_sp, MR_Word *base_curfr,
 	const char *path, const char *filename, int lineno, bool print_parent, 
 	const char *parent_filename, int parent_lineno, int indent)
 {
@@ -557,8 +605,8 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 	switch (pos) {
 		case MR_CONTEXT_NOWHERE:
 			fprintf(fp, " ");
-			MR_print_call_trace_info(fp, entry,
-				base_sp, base_curfr);
+			MR_maybe_print_call_trace_info(fp, include_trace_data,
+				entry, base_sp, base_curfr);
 			MR_print_proc_id(fp, entry);
 			if (strlen(path) > 0) {
 				fprintf(fp, " %s", path);
@@ -571,8 +619,8 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 			MR_maybe_print_parent_context(fp, print_parent,
 				FALSE, parent_filename, parent_lineno);
 			fprintf(fp, " ");
-			MR_print_call_trace_info(fp, entry,
-				base_sp, base_curfr);
+			MR_maybe_print_call_trace_info(fp, include_trace_data,
+				entry, base_sp, base_curfr);
 			MR_print_proc_id(fp, entry);
 			if (strlen(path) > 0) {
 				fprintf(fp, " %s", path);
@@ -582,8 +630,8 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 
 		case MR_CONTEXT_AFTER:
 			fprintf(fp, " ");
-			MR_print_call_trace_info(fp, entry,
-				base_sp, base_curfr);
+			MR_maybe_print_call_trace_info(fp, include_trace_data,
+				entry, base_sp, base_curfr);
 			MR_print_proc_id(fp, entry);
 			if (strlen(path) > 0) {
 				fprintf(fp, " %s", path);
@@ -599,8 +647,8 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 			MR_maybe_print_parent_context(fp, print_parent,
 				TRUE, parent_filename, parent_lineno);
 			fprintf(fp, "\n%*s ", indent, "");
-			MR_print_call_trace_info(fp, entry,
-				base_sp, base_curfr);
+			MR_maybe_print_call_trace_info(fp, include_trace_data,
+				entry, base_sp, base_curfr);
 			MR_print_proc_id(fp, entry);
 			if (strlen(path) > 0) {
 				fprintf(fp, " %s", path);
@@ -610,8 +658,8 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 
 		case MR_CONTEXT_NEXTLINE:
 			fprintf(fp, " ");
-			MR_print_call_trace_info(fp, entry,
-				base_sp, base_curfr);
+			MR_maybe_print_call_trace_info(fp, include_trace_data,
+				entry, base_sp, base_curfr);
 			MR_print_proc_id(fp, entry);
 			if (strlen(path) > 0) {
 				fprintf(fp, " %s", path);
@@ -624,7 +672,7 @@ MR_print_proc_id_trace_and_context(FILE *fp, MR_Context_Position pos,
 			break;
 
 		default:
-			fatal_error("invalid MR_Context_Position");
+			MR_fatal_error("invalid MR_Context_Position");
 	}
 }
 
