@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 1997-1999 The University of Melbourne.
+** Copyright (C) 1997-2000 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
@@ -40,42 +40,62 @@
 **	  this goal being solvable] in an if-then-else with a nondet condition,
 **	  or in solutions/2 (MR_solve);
 **	- when executing a `retry' command in the debugger (MR_retry).
+** MR_prune_ticket()
+**	called when cutting away the topmost choice point
 ** MR_discard_ticket()
-**	called when cutting away or failing over the topmost choice point
+**	called when failing over the topmost choice point
 ** MR_mark_ticket_stack()
-**	called before a commit, and when entering an execution traced procedure
-** MR_discard_tickets_to()
+**	called before a commit,
+**	when establishing an exception handler,
+**	or when entering an execution traced procedure
+** MR_prune_tickets_to()
 **	called after a commit
+** MR_discard_tickets_to()
+**	called when an exception is thrown,
+**	or when doing a retry in the debugger
 */
 /*---------------------------------------------------------------------------*/
 
-/* void MR_mark_ticket_stack(Word &); */
+/* void MR_mark_ticket_stack(MR_Word &); */
 #define MR_mark_ticket_stack(save_ticket_counter)		\
 	do {							\
-		save_ticket_counter = MR_ticket_counter;	\
+		(save_ticket_counter) = MR_ticket_counter;	\
+	} while(0)
+
+/* void MR_prune_ticket(void); */
+#define MR_prune_ticket()					\
+	do {							\
+		--MR_ticket_counter;				\
 	} while(0)
 
 /* void MR_discard_ticket(void); */
 #define MR_discard_ticket()					\
 	do {							\
-		--MR_ticket_counter;				\
+		MR_ticket_high_water = --MR_ticket_counter;	\
 	} while(0)
 
-/* void MR_discard_tickets_to(Word); */
+/* void MR_prune_tickets_to(MR_Word); */
+#define MR_prune_tickets_to(save_ticket_counter)		\
+	do {							\
+		MR_ticket_counter = (save_ticket_counter);	\
+	} while(0)
+
+/* void MR_discard_tickets_to(MR_Word); */
 #define MR_discard_tickets_to(save_ticket_counter)		\
 	do {							\
-		MR_ticket_counter = save_ticket_counter;	\
+		MR_ticket_high_water = MR_ticket_counter =	\
+			(save_ticket_counter);			\
 	} while(0)
 
 	/* 
 	** Called when we create a choice point
 	** (including semidet choice points).
 	*/
-/* void MR_store_ticket(Word &); */
+/* void MR_store_ticket(MR_Word &); */
 #define MR_store_ticket(save_trail_ptr)				\
 	do {							\
-		(save_trail_ptr) = (Word) MR_trail_ptr; 	\
-		++MR_ticket_counter;				\
+		(save_trail_ptr) = (MR_Word) MR_trail_ptr; 	\
+		MR_ticket_counter = ++MR_ticket_high_water;	\
 	} while(0)
 
 	/*
@@ -86,15 +106,15 @@
 	** transient registers are not needed because
 	** MR_trail_ptr is never a real register.
 	*/
-/* void MR_reset_ticket(Word, MR_untrail_reason); */
+/* void MR_reset_ticket(MR_Word, MR_untrail_reason); */
 #define MR_reset_ticket(old, kind)				\
 	do {							\
 		MR_TrailEntry *old_trail_ptr =  		\
-			(MR_TrailEntry *)old;			\
+			(MR_TrailEntry *) (old);		\
 		if (MR_trail_ptr != old_trail_ptr) {		\
-			/* save_transient_registers(); */	\
-			MR_untrail_to(old_trail_ptr, kind);	\
-			/* restore_transient_registers(); */	\
+			/* MR_save_transient_registers(); */	\
+			MR_untrail_to(old_trail_ptr, (kind));	\
+			/* MR_restore_transient_registers(); */	\
 		}						\
 	} while(0)
 
@@ -141,7 +161,6 @@ typedef enum {
 
 	/*
 	** MR_exception:
-	** (reserved for future use)
 	** An exception was thrown.
 	** Behaves as MR_undo, except that function trail entries may
 	** choose to behave differently for exceptions than for failure.
@@ -186,8 +205,8 @@ typedef struct {
 #endif
 	union {
 		struct {
-			Word *MR_address;
-			Word MR_value;
+			MR_Word *MR_address;
+			MR_Word MR_value;
 		} MR_val;
 		struct {
 			MR_untrail_func_type *MR_untrail_func;
@@ -210,13 +229,13 @@ typedef struct {
   */
   #define MR_get_trail_entry_kind(entry)				\
 	((MR_trail_entry_kind)						\
-	  (MR_tag((Word) (entry)->MR_union.MR_val.MR_address)))
+	  (MR_tag((MR_Word) (entry)->MR_union.MR_val.MR_address)))
 
   /*
-  ** Word * MR_get_trail_entry_address(const MR_trail_entry *);
+  ** MR_Word * MR_get_trail_entry_address(const MR_trail_entry *);
   */
   #define MR_get_trail_entry_address(entry) \
-	((Word *)							\
+	((MR_Word *)							\
 	  MR_body((entry)->MR_union.MR_val.MR_address, MR_value_trail_tag))
 
   /*
@@ -225,17 +244,17 @@ typedef struct {
   */
   #define MR_get_trail_entry_untrail_func(entry)			\
 	((MR_untrail_func_type *)					\
-	    MR_body((Word) (entry)->MR_union.MR_func.MR_untrail_func,	\
+	    MR_body((MR_Word) (entry)->MR_union.MR_func.MR_untrail_func,\
 		     MR_func_trail_tag))
 
   /*
   ** void MR_store_value_trail_entry(
-  **		MR_trail_entry *entry, Word *address, Word value);
+  **		MR_trail_entry *entry, MR_Word *address, MR_Word value);
   */
   #define MR_store_value_trail_entry(entry, address, value)		\
 	  do {								\
 		(entry)->MR_union.MR_val.MR_address =			\
-			(Word *) (Word)					\
+			(MR_Word *) (MR_Word)				\
 			  MR_mkword(MR_value_trail_tag, (address));	\
 		(entry)->MR_union.MR_val.MR_value = (value);		\
 	  } while (0)
@@ -247,7 +266,7 @@ typedef struct {
   #define MR_store_function_trail_entry(entry, func, datum)		\
 	  do {								\
 		(entry)->MR_union.MR_func.MR_untrail_func =		\
-			(MR_untrail_func_type *) (Word)			\
+			(MR_untrail_func_type *) (MR_Word)		\
 			  MR_mkword(MR_func_trail_tag, (func));		\
 		(entry)->MR_union.MR_func.MR_datum = (datum);		\
 	  } while (0)
@@ -262,7 +281,7 @@ typedef struct {
 
   /*
   ** void MR_store_value_trail_entry(
-  **		MR_trail_entry *entry, Word *address, Word value);
+  **		MR_trail_entry *entry, MR_Word *address, MR_Word value);
   */
   #define MR_store_value_trail_entry(entry, address, value)		\
 	  do {								\
@@ -284,7 +303,7 @@ typedef struct {
 #endif
 
 /*
-** Word MR_get_trail_entry_value(const MR_trail_entry *);
+** MR_Word MR_get_trail_entry_value(const MR_trail_entry *);
 */
 #define MR_get_trail_entry_value(entry) \
 	((entry)->MR_union.MR_val.MR_value)
@@ -298,21 +317,38 @@ typedef struct {
 /*---------------------------------------------------------------------------*/
 
 /* The Mercury trail */
-extern MemoryZone *MR_trail_zone;
+extern MR_MemoryZone *MR_trail_zone;
 
 /* Pointer to the current top of the Mercury trail */
-/* N.B. Use `MR_trail_ptr', defined in mercury_regorder.h, not `MR_trail_ptr_var'. */
+/*
+** N.B. Use `MR_trail_ptr', defined in mercury_regorder.h,
+** not `MR_trail_ptr_var'.
+*/
 extern MR_TrailEntry *MR_trail_ptr_var;
 
 /*
-** An integer variable that is incremented whenever we create a choice
+** An integer variable that holds the current choice point identifier;
+** it is allocated a new value whenever we create a choice
 ** point (including semidet choice points, e.g. in an if-then-else)
-** and decremented whenever we remove one.
+** and it is reset whenever a choice point is backtracked over
+** or pruned away.
 **
 ** N.B.  Use `MR_ticket_counter', defined in mercury_regorder.h,
 ** not `MR_ticket_counter_var'.
 */
-extern Unsigned MR_ticket_counter_var;
+extern MR_Unsigned MR_ticket_counter_var;
+
+/*
+** An integer variable that is incremented whenever we create a choice
+** point (including semidet choice points, e.g. in an if-then-else)
+** and decremented or reset whenever we backtrack over one,
+** but which is _not_ decremented or reset when a choice point is
+** pruned away with a commit.
+**
+** N.B.  Use `MR_ticket_high_water', defined in mercury_regorder.h,
+** not `MR_ticket_high_water_var'.
+*/
+extern MR_Unsigned MR_ticket_high_water_var;
 
 /*---------------------------------------------------------------------------*/
 /*
@@ -325,7 +361,7 @@ extern Unsigned MR_ticket_counter_var;
 /*---------------------------------------------------------------------------*/
 
 /*
-** void  MR_trail_value(Word *address, Word value);
+** void  MR_trail_value(MR_Word *address, MR_Word value);
 **
 ** Make sure that when the current execution is
 ** backtracked over, `value' is placed in `address'.
@@ -338,7 +374,7 @@ extern Unsigned MR_ticket_counter_var;
 	} while(0);
 
 /*
-** void  MR_trail_current_value(Word *address);
+** void  MR_trail_current_value(MR_Word *address);
 **
 ** Make sure that when the current execution is
 ** backtracked over, the value currently in `address'
@@ -371,13 +407,18 @@ extern Unsigned MR_ticket_counter_var;
 void MR_untrail_to(MR_TrailEntry *old_trail_ptr, MR_untrail_reason reason);
 
 /* abstract type */
-typedef Unsigned MR_ChoicepointId;
+typedef MR_Unsigned MR_ChoicepointId;
 
 /*
 ** MR_ChoicepointId MR_current_choicepoint_id(void);
 **
-** Returns a value indicative of the current
-** choicepoint.  If we execute
+** Returns a value indicative of the current choicepoint. 
+** The value remains meaningful if the choicepoint is pruned
+** away by a commit, but is not meaningful after backtracking
+** past the point where the choicepoint was created (since
+** choicepoint ids may be reused after backtracking).
+**
+** If we execute
 ** 
 ** 	oldcp = MR_current_choicepoint_id();
 ** 	... and a long time later ...
@@ -389,13 +430,27 @@ typedef Unsigned MR_ChoicepointId;
 ** then code A will be executed if and only if the
 ** current choicepoint is the same in both calls.
 */
-#define MR_current_choicepoint_id() ((const MR_ChoicepointId)MR_ticket_counter)
+#define MR_current_choicepoint_id() ((const MR_ChoicepointId) MR_ticket_counter)
 
 /*
 ** MR_ChoicepointId MR_null_choicepoint_id(void);
 **
 ** A macro defining a "null" ChoicepointId.
+** This is suitable for use in static initializers.
 */
 #define MR_null_choicepoint_id() ((const MR_ChoicepointId)0)
 
+/*
+** bool MR_choicepoint_newer(MR_ChoicepointId x, MR_ChoicepointId y);
+**
+** Returns true iff the choicepoint indicated by `x'
+** is newer than (i.e. was created more recently than)
+** the choicepoint indicated by `y'.
+** The null ChoicepointId is considered older than
+** any non-null ChoicepoindId.
+** If either of the choice points have been
+** backtracked over, the behaviour is undefined.
+*/
+#define MR_choicepoint_newer(x, y) ((x) > (y))
+
 #endif /* not MERCURY_TRAIL_H */
