@@ -5,7 +5,7 @@
 
 #include "cpp_main.h"
 
-typedef Word MercuryList;
+typedef MR_Word MercuryList;
 
 static void print_list(MercuryList);
 
@@ -25,7 +25,7 @@ void cpp_main() {
 	** to the Mercury predicate foo/1 in mode
 	** 	:- mode foo(out) is cc_multi.
 	*/
-	Integer value;
+	MR_Integer value;
 	one_foo(&value);
 	printf("one_foo(&value) gives value = %ld\n", (long) value);
 
@@ -76,15 +76,15 @@ void cpp_main() {
 }
 
 static void print_list(MercuryList list) {
-	if (list_is_empty(list)) {
+	if (MR_list_is_empty(list)) {
 		printf("[]");
 	} else {
 		printf("[");
 		printf("%ld", (long) list_head(list));
-		list = list_tail(list);
-		while (!list_is_empty(list)) {
-			printf(", %ld", (long) list_head(list));
-			list = list_tail(list);
+		list = MR_list_tail(list);
+		while (!MR_list_is_empty(list)) {
+			printf(", %ld", (long) MR_list_head(list));
+			list = MR_list_tail(list);
 		}
 		printf("]");
 	}
