@@ -6,7 +6,7 @@
 ;; Keywords: unix, tools
 
 ;; Copyright (C) 1992, 1993, 1994, 1995, 1996 Free Software Foundation, Inc.
-;; Copyright (C) 1998-1999 Fergus Henderson.
+;; Copyright (C) 1998-2000 Fergus Henderson.
 
 ;; This file is part of GNU Emacs.
 
@@ -1146,12 +1146,12 @@ containing the executable being debugged.")
 	    (let
 		((msg (substring result (match-beginning 1) (match-end 1)))
 		 ;; match 2 is for the optional path
-		 (caller-file
+		 (file
 		   (substring result (match-beginning 3) (match-end 3)))
-		 (caller-line
+		 (line
 		   (substring result (match-beginning 4) (match-end 4)))
-		 (file (substring result (match-beginning 5) (match-end 5)))
-		 (line (substring result (match-beginning 6) (match-end 6))))
+		 (caller-file (substring result (match-beginning 5) (match-end 5)))
+		 (caller-line (substring result (match-beginning 6) (match-end 6))))
 	      (message "%s" msg)
               (save-excursion
                 (let ((abs-caller-file (gud-mdb-file-name caller-file))
@@ -1374,17 +1374,22 @@ and source-file directory for your debugger."
   (setq gud-menu-name "MDB")
   (setq gud-buffer-name-prefix "mdb")
   (setq gud-show-commands t)
-  (setq gud-redirect-io t)
+  ;; Setting gud-redirect-io is broken in recent versions of
+  ;; GNU Emacs (e.g. 20.4.1), since Emacs apparently now uses
+  ;; pipes rather than ttys.  So we don't enable this by
+  ;; default anymore.
+  ;; (setq gud-redirect-io t)
 
   (gud-common-init command-line 'gud-mdb-massage-args
 		   'gud-mdb-marker-filter 'gud-mdb-find-file)
 
   (gud-mdb-define-menu-entries)
   (gud-mdb-add-menu-entries)
-  (save-excursion
-    (set-buffer gud-prog-buffer)
-    (display-buffer gud-prog-buffer)
-    (gud-mdb-add-menu-entries))
+  (if gud-prog-buffer
+      (save-excursion
+        (set-buffer gud-prog-buffer)
+        (display-buffer gud-prog-buffer)
+        (gud-mdb-add-menu-entries)))
 
   (setq comint-prompt-regexp "^mdb> ")
   (setq paragraph-start comint-prompt-regexp)
@@ -1593,8 +1598,9 @@ comint mode, which see."
     (apply 'make-comint prog-buf-name program nil
 	   (funcall massage-args file args))
     (if gud-redirect-io
-	(setq gud-prog-buffer (concat "*" prog-buf-name "*"))
-	(set-buffer debugger-buf)))
+        (progn
+	  (setq gud-prog-buffer (concat "*" prog-buf-name "*"))
+	  (set-buffer debugger-buf))))
 
   ;; Since comint clobbered the mode, we don't set it until now.
   (gud-mode)
